{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
// Generic material form: shows materials of a supplied materiallist as frames on a pagecontrol
// and allows you to manage them.
//
// Used by various other tools in DeleD (e.g. Material Picker, Replace Materials, Material Editor)

unit frmMaterialsForm;

{$MODE Delphi}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, StdCtrls, dglOpenGL, Valedit, Contnrs, Types, ExtCtrls,
  StrUtils, Buttons, Menus, ToolWin, ImgList, FileUtil,
  fraTexture,
  fraMaterial,
  unit_Scene,
  unit_Variables,
  unit_Textures,
  unit_Materials,
  unit_MaterialList,
  unit_Functions;

type
  TfrmMaterials = class(TForm)
    pcMaterials: TPageControl;
    lblTitle: TLabel;
    SaveDialog1: TSaveDialog;
    ctbButtons: TControlBar;
    tlbEdit: TToolBar;
    tbAddMaterial: TToolButton;
    tbRemoveMaterial: TToolButton;
    ToolButton3: TToolButton;
    tbAddCategory: TToolButton;
    tbRemoveCategory: TToolButton;
    pmMaterials: TPopupMenu;
    miAddMaterial: TMenuItem;
    miRemoveMaterial: TMenuItem;
    N1: TMenuItem;
    miAddCategory: TMenuItem;
    miRemoveCategory: TMenuItem;
    ilMaterial: TImageList;
    tbSaveToDML: TToolButton;
    ToolButton7: TToolButton;
    miRenameCategory: TMenuItem;
    N2: TMenuItem;
    miSavetoDML: TMenuItem;
    tbRenameCategory: TToolButton;
    procedure Clear;
    procedure pcMaterialsResize(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure pcMaterialsChanging(Sender: TObject; var AllowChange: Boolean);
    procedure pcMaterialsChange(Sender: TObject);
    procedure pcMaterialsDragDrop(Sender, Source: TObject; X, Y: Integer);
    procedure pcMaterialsDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
    procedure FormDblClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure AddMaterialClick(Sender: TObject);
    procedure RemoveMaterialClick(Sender: TObject);
    procedure AddCategoryClick(Sender: TObject);
    procedure RemoveCategoryClick(Sender: TObject);
    procedure RenameCategoryClick(Sender: TObject);
    procedure SaveToDMLClick(Sender: TObject);
  private
    fScene: TScene;
    fMaterials: TMaterialList;
    fSelectedMaterial: TMaterial;
    fOnSelect: TNotifyEvent;
    fOnOpen: TNotifyEvent;
    fOnShow: TNotifyEvent;
    fCanAcceptTextures: boolean;
    function FrameForMaterial(const aMaterial: TMaterial): TMaterialFrame;
    procedure UpdateButtonState;
    procedure OnSelectHandler(Sender: TObject);     // handler for when a material frame (!) is selected
    procedure SetTitle(value: string);
    procedure SetShowButtons(value: boolean);
    procedure SetSelectedMaterial(const aMaterial: TMaterial);
    function GetSelectedCategory: string;
    procedure SetSelectedCategory(aCategory: string);
    function GetFirstMaterialFromCategory(aCategory: string): TMaterial;
  public
    constructor Create(AOwner: TComponent; AMaterialList: TMaterialList; AScene: TScene); reintroduce;
    procedure DeselectFrames;
    procedure CreateCategory(ACategoryName: string);
    procedure CreateCategories;
    procedure ClearCategory;
    procedure RemoveCategory;
    procedure FillCategory;
    procedure UpdateCategory;
    procedure ClearGUI;
    procedure UpdateGUI;
    property Title: string write SetTitle;
    property ShowButtons: boolean write SetShowButtons;
    property OnSelect: TNotifyEvent read fOnSelect write fOnSelect;
    property OnOpen: TNotifyEvent read fOnOpen write fOnOpen;
    property OnShow: TNotifyEvent read fOnShow write fOnShow;
    property CanAcceptTextures: boolean read fCanAcceptTextures write fCanAcceptTextures;
    property Materials: TMaterialList read fMaterials write fMaterials;
    property SelectedMaterial: TMaterial read fSelectedMaterial write SetSelectedMaterial;
    property SelectedCategory: string read GetSelectedCategory write SetSelectedCategory;
  end;

implementation

uses
  frmMaterialEditorForm,
  frmTextureBrowserForm, // for drag object
  unit_Enums,
  unit_ParamList,
  unit_DMLWriter;

{$R *.lfm}

// ### TfrmMaterials ###########################################################

constructor TfrmMaterials.Create(AOwner: TComponent; AMaterialList: TMaterialList; AScene: TScene);
begin
    inherited Create(AOwner);
    fScene := AScene;
    fMaterials := AMaterialList;
    CanAcceptTextures := false;
    Title := '';
    ShowButtons := false;
end;

procedure TfrmMaterials.FormCreate(Sender: TObject);
begin
    // INFO, PJ: Only enable the popup menu if we are inside the material editor.
    ///          Not the cleanest solution, but nice and easy. :)
    if (Owner is TfrmMaterialEditor) then
        pcMaterials.PopupMenu := pmMaterials
    else
        pcMaterials.PopupMenu := nil;
end;

procedure TfrmMaterials.OnSelectHandler(Sender: TObject);
var materialFrame: TMaterialFrame;
begin
    DeselectFrames;
    if Assigned(Sender) then begin
        materialFrame := Sender as TMaterialFrame;
        // set the selected material; NOTE: this triggers an OnSelect event
        // which is handled by other forms such as MainForm and MaterialEditorForm
        SelectedMaterial := materialFrame.Material;
        // only select this frame as the selected one
        materialFrame.Selected := true;
        // fire event; handled by other forms (such as MainForm and MaterialEditorForm)
        // NOTE: turned off because setting the SelectedMaterial also does this?
        //if Assigned(fOnSelect) then
        //    fOnSelect(materialFrame);
    end else begin
        SelectedMaterial := nil;    //Materials[0];       // JCC, 29-02-2008: use system material
        if Assigned(fOnSelect) then
            fOnSelect(nil);
    end;
    UpdateButtonState;
end;

procedure TfrmMaterials.SetTitle(value: string);
begin
    if length(value)=0 then
        lblTitle.Visible := false
    else begin
        lblTitle.Visible := true;
        lblTitle.Caption := value;
    end;
end;

procedure TfrmMaterials.SetShowButtons(value: boolean);
begin
    ctbButtons.Visible := value;
end;

procedure TfrmMaterials.SetSelectedMaterial(const aMaterial: TMaterial);
var frame: TMaterialFrame;
begin
    fMaterials.SelectedMaterial := aMaterial;
    fSelectedMaterial := aMaterial;

    if Assigned(aMaterial) then
        SelectedCategory := aMaterial.Category; // this clears and refills the category if the category changed

    // setting the select material triggers an onSelect (common VCL methodology)
    // JCC: causes its OnSelectHandler to be executed twice, causing the Undo buffer to give erronous results
    frame := frameForMaterial(aMaterial);
    if Assigned(fOnSelect) then
        fOnSelect(frame);

    DeselectFrames;

    // make sure the frame is selected (SelectedCategory might not have done this if category didn't change)
    if Assigned(frame) then
        frame.Selected := true;

    UpdateButtonState;
end;

function TfrmMaterials.GetSelectedCategory: string;
begin
    if Assigned(pcMaterials.ActivePage) then
        Result := pcMaterials.ActivePage.Caption
    else
        Result := '';
end;

procedure TfrmMaterials.SetSelectedCategory(aCategory: string);
var counter: integer;
begin
    if (pcMaterials.PageCount>0) and not (Assigned(pcMaterials.ActivePage) and (aCategory=pcMaterials.ActivePage.Caption)) then begin
        ClearCategory;
        counter := 0;
        aCategory := Uppercase(aCategory);
        with pcMaterials do begin
            // fill the correct tab; if we can't find the correct tab,
            // set the first tab to be the selectede one and fill it
            while (counter<PageCount) and (Uppercase(Pages[counter].Caption)<>aCategory) do
                inc(counter);
            if counter<PageCount then
                ActivePage := Pages[counter]
            else
                ActivePage := Pages[0];
            FillCategory;
        end;
    end;
end;

function TfrmMaterials.GetFirstMaterialFromCategory(aCategory: string): TMaterial;
var i: integer;
    material,resultMaterial: TMaterial;
begin
    resultMaterial := nil;
    for i := 0 to materials.numMaterials-1 do begin
        material := materials.Materials[i];
        if (material.Category = aCategory) and material.IsVisible then begin
            resultMaterial := material;
            break;
        end;
    end;
    result := resultMaterial;
end;

procedure TfrmMaterials.Clear;
begin
    pcMaterials.Visible := false;
    while pcMaterials.PageCount>0 do
        pcMaterials.Pages[0].Free;
    pcMaterials.Visible := true;
end;

procedure TfrmMaterials.pcMaterialsResize(Sender: TObject);
begin
    UpdateCategory;
end;

procedure TfrmMaterials.pcMaterialsChanging(Sender: TObject; var AllowChange: Boolean);
begin
    // remove material-frames on the current tab (which is about to be changed to another tab!)
    ClearCategory;

    // set the selected material to nil (automatically informs application)
    SelectedMaterial := nil;
end;

procedure TfrmMaterials.pcMaterialsChange(Sender: TObject);
begin
    FillCategory;
    //UpdateCategory;       // also called by FillCategory
    UpdateButtonState;
end;

procedure TfrmMaterials.pcMaterialsDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
begin
    Accept := (Source is TTextureDragObject) and (CanAcceptTextures);
end;

procedure TfrmMaterials.pcMaterialsDragDrop(Sender, Source: TObject; X, Y: Integer);
var textureFrame: TTextureFrame;
    texture: TTexture;
    material: TMaterial;
    name: string;
begin
    if Source is TTextureDragObject then begin
        textureFrame := TTextureDragObject(Source).TextureFrame;
        texture := Materials.Textures.AddTexture(textureFrame.Texture);
        material := TMaterial.Create;
        name := ExtractFileName(textureFrame.Texture);
        name := LeftStr(name,length(name)-length(ExtractFileExt(name)));
        if not Materials.NameIsUnique(name) then name := Materials.CreateUniqueName;
        material.Name := name;
        material.Category := pcMaterials.ActivePage.Caption;
        material.AddLayerForTexture(texture,boReplace);
        Materials.Add(material);
        FillCategory;
        SelectedMaterial := material;        // nil;
    end;
end;

// adds a standard material to the current category
procedure TfrmMaterials.AddMaterialClick(Sender: TObject);
var texture: TTexture;
    material: TMaterial;
begin
    texture := Materials.Textures.AddTexture(dir_texture+'System\system.bmp');
    material := TMaterial.Create;
    material.Name := Materials.CreateUniqueName;
    material.Category := pcMaterials.ActivePage.Caption;
    material.AddLayerForTexture(texture,boReplace);
    Materials.Add(material);
    FillCategory;
    SelectedMaterial := material;
end;

// removes the selected material (removing the standard (0) material is not allowed)
procedure TfrmMaterials.RemoveMaterialClick(Sender: TObject);
var similarMaterials: TMaterialList;
    i: integer;
    fParamList: TParamList;
begin
    if SelectedMaterial.IsSystem then
        raise Exception.Create('Cannot delete the system-material.')
    else begin
        similarMaterials := Materials.SimilarMaterialsForMaterial(SelectedMaterial);
        try
            fParamList := TParamList.Create(false);
            fParamList.Add(similarMaterials,'currentMaterials',true);
            fParamList.Add(Materials[0],'newMaterial');
            fParamList.Add(false,'doSelectionOnly');
            fScene.ToolExecuter.Execute(ptFaceMaterialReplacer,taModification,fParamList);
            FreeAndNil(fParamList);

            for i := 0 to similarMaterials.Count - 1 do
                Materials.Remove(similarMaterials[i]);
        finally
            FreeAndNil(similarMaterials);
        end;
        FillCategory;
        //SelectedMaterial := nil;
        SelectedMaterial := GetFirstMaterialFromCategory(SelectedCategory);
        if SelectedMaterial = nil then          // nil is no good here..
            SelectedMaterial := Materials[0];

        // Notify the material editor (material detail) that the selected material was deleted
        //if Assigned(fOnSelect) then
        //    fOnSelect(nil);
    end;
end;

// adds a category (simply a new tabsheet)
procedure TfrmMaterials.AddCategoryClick(Sender: TObject);
var name: string;
begin
    name := InputBox('New Category', 'Enter a category name', '');
    if length(name)>0 then begin
        CreateCategory(name);
        SelectedCategory := name;
    end;
end;

// renames a category (tabsheet+all materials in that category)
procedure TfrmMaterials.RenameCategoryClick(Sender: TObject);
var oldname, newname: string;
    i: integer;
    material: TMaterial;
begin
    newname := InputBox('Rename Category', 'Enter a category name', '');
    if length(newname)>0 then begin
        oldname := pcMaterials.ActivePage.Caption;
        pcMaterials.ActivePage.Caption := newname;
        for i := 0 to Materials.Count-1 do begin
            material := Materials[i];
            if material.Category=oldname then
                material.Category := newname;
        end;
    end;
end;

// removes an entire category but only after each individual material has been removed first
procedure TfrmMaterials.RemoveCategoryClick(Sender: TObject);
begin
    RemoveCategory;
end;

procedure TfrmMaterials.SaveToDMLClick(Sender: TObject);
var filename: string;
    originalMaterials: TMaterialList;
    dmlWriter: TDMLWriter;
begin
    try
        Screen.Cursor := crHourGlass;
        SaveDialog1.DefaultExt := '.dml';
        SaveDialog1.InitialDir := dir_texture;
        SaveDialog1.Filter := 'DeleD Material List (*.dml)|*.dml';

        SaveDialog1.FileName := RemoveExtension(fScene.filename)+'.dml';
        if SaveDialog1.Execute then begin
            filename := SaveDialog1.FileName;
            originalMaterials := materials.GetOriginalMaterials;
            dmlWriter := TDMLWriter.Create;
            try
                if FileExistsUTF8(filename) then begin
                    if messagedlg(filename+' exists, overwrite?', mtConfirmation, [mbYes, mbNo], 0) = mrYes then begin
                        dmlWriter.Save(filename,fScene,originalMaterials);
                    end
                end else begin
                    dmlWriter.Save(filename,fScene,originalMaterials);
                end;
            finally
                FreeAndNil(originalMaterials);
                FreeAndNil(dmlWriter);
            end;
        end;
    finally
        Screen.Cursor := crDefault;
    end;
end;

procedure TfrmMaterials.DeselectFrames;
var i,j: integer;
    tab: TTabSheet;
    sb: TScrollbox;
    pframe: TMaterialFrame;
begin
    for i := 0 to pcMaterials.PageCount-1 do begin                              // loop through all tabs
        tab := pcMaterials.Pages[i];
        sb := TScrollbox(tab.FindChildControl('scrollbox'));
        if Assigned(sb) then begin                                              // each tab has a scrollbox with textureframes
            for j := 0 to sb.ControlCount-1 do begin
                pframe := TMaterialFrame(sb.FindComponent('picture'+IntToStr(j)));
                pframe.Selected := false;
            end;
        end;
    end;
end;

function TfrmMaterials.FrameForMaterial(const aMaterial: TMaterial): TMaterialFrame;
var i,j: integer;
    tab: TTabSheet;
    sb: TScrollbox;
    frame: TMaterialFrame;
begin
    result := nil;
    if Assigned(aMaterial) then begin
        for i := 0 to pcMaterials.PageCount-1 do begin                          // loop through all tabs
            tab := pcMaterials.Pages[i];
            sb := TScrollbox(tab.FindChildControl('scrollbox'));
            if Assigned(sb) then begin                                          // each tab has a scrollbox with textureframes
                for j := 0 to sb.ControlCount-1 do begin
                    frame := TMaterialFrame(sb.FindComponent('picture'+IntToStr(j)));
                    if frame.Material = aMaterial then begin
                        result := frame;
                        exit;
                    end;
                end;
            end;
        end;
    end;
end;

procedure TfrmMaterials.UpdateButtonState;
var isEnabled: boolean;
begin
    isEnabled := Assigned(fSelectedMaterial) and not fSelectedMaterial.IsSystem;
    tbRemoveMaterial.Enabled := isEnabled;
    miRemoveMaterial.Enabled := isEnabled;

    isEnabled := Assigned(pcMaterials.ActivePage) and (Uppercase(SelectedCategory)<>'SYSTEM');
    tbRemoveCategory.Enabled := isEnabled;
    miRemoveCategory.Enabled := isEnabled;
    tbRenameCategory.Enabled := isEnabled;
    miRenameCategory.Enabled := isEnabled;
end;

procedure TfrmMaterials.FormShow(Sender: TObject);
begin
    UpdateButtonState;
    // fire our own OnShow event
    if Assigned(fOnShow) then fOnShow(self);
end;

procedure TfrmMaterials.CreateCategory(ACategoryName: string);
var tab: TTabSheet;
    sb: TScrollbox;
begin
    tab := pcMaterials.AddTabSheet; //TTabSheet.CreateParented(pcMaterials.Handle); // create a new tab on the pagecontrol
    tab.Caption := ACategoryName;
    tab.Align := alClient;
    sb := TScrollbox.Create(tab);  //CreateParented(tab.Handle);  // create a scrollbox on the new tab
    sb.Parent := tab;
    sb.VertScrollBar.Smooth := true;
    sb.VertScrollBar.Tracking := true;
    sb.HorzScrollBar.Smooth := true;
    sb.HorzScrollBar.Tracking := true;
    sb.Name := 'scrollbox';
    sb.Parent := tab;
//    sb.Ctl3D := false;
    sb.BorderStyle := bsSingle;
    sb.Align := alClient;
    sb.OnDragOver := pcMaterials.OnDragOver;
    sb.OnDragDrop := pcMaterials.OnDragDrop;
end;

procedure TfrmMaterials.CreateCategories;
var i: integer;
    categories: TStringList;
begin
    pcMaterials.Visible := false;

    categories := materials.ExtractCategories;
    for i := 0 to categories.Count-1 do
        CreateCategory(categories.Strings[i]);

    FreeAndNil(categories);

    pcMaterials.Visible := true;
end;

procedure TfrmMaterials.ClearCategory;
var tab: TTabSheet;
    counter: integer;
    sb: TScrollbox;
    control: TControl;
begin
    tab := pcMaterials.ActivePage;
    if Assigned(tab) then begin
        sb := TScrollbox(tab.FindChildControl('scrollbox'));
        if Assigned(sb) then begin
            counter := 0;
            while counter<sb.ControlCount do begin
                control := sb.Controls[counter];
                if control is TMaterialFrame then
                    FreeAndNil(control)
                else
                    inc(counter);
            end;
        end;
    end;
end;

procedure TfrmMaterials.RemoveCategory;
var shouldRemove: Boolean;
    i,j: integer;
    material: TMaterial;
    similarMaterials: TMaterialList;
    fParamList: TParamList;
begin
    if (UpperCase(SelectedCategory) = 'SYSTEM') then
        MessageDlg('Removing the System category is not allowed.',mtInformation,[mbOK],0)
    else begin

        shouldRemove := true;
        if (Materials.CategoryIsUsed(SelectedCategory)) then begin

            shouldRemove := MessageDlg('Warning'
                ,'Removing this category will delete all materials in the category. Are you sure?'
                ,mtWarning, mbYesNo, 0 ) = mrYes;
            if shouldRemove then begin
                // remove all materials from this category
                for i := Materials.NumMaterials-1 downto 0 do begin
                    material := Materials.Materials[i];
                    if (material.Category = SelectedCategory) then begin
                        similarMaterials := Materials.SimilarMaterialsForMaterial(material);
                        try
                            fParamList := TParamList.Create(false);
                            fParamList.Add(similarMaterials,'currentMaterials',true);
                            fParamList.Add(Materials[0],'newMaterial');
                            fParamList.Add(false,'doSelectionOnly');
                            fScene.ToolExecuter.Execute(ptFaceMaterialReplacer,taModification,fParamList);
                            FreeAndNil(fParamList);
                            for j := 0 to similarMaterials.Count - 1 do
                                Materials.Remove(similarMaterials[j]);
                        finally
                            FreeAndNil(similarMaterials);
                        end;
                    end;
                end;
            end;
        end;
        if shouldRemove then begin
            pcMaterials.ActivePage.Free;
            SelectedMaterial := Materials[0];   // JCC, 29-02-2008: select the system material when removing categories
            //pcMaterials.ActivePageIndex := 0;
            //FillCategory;
            UpdateButtonState;
        end;
    end;
end;

procedure TfrmMaterials.FillCategory;
var i,count: integer;
    material: TMaterial;
    tab: TTabSheet;
    sb: TScrollbox;
    pic: TMaterialFrame;
begin
    tab := pcMaterials.ActivePage;
    if Assigned(tab) then begin
        ClearCategory;
        sb := TScrollbox(tab.FindChildControl('scrollbox'));
        count := 0;
        for i := 0 to materials.numMaterials-1 do begin                         // place material-frames on the scrollbox
            material := materials.Materials[i];
            if (material.Category = tab.Caption) and material.IsVisible then begin
                pic := TMaterialFrame.Create(sb);
                pic.OnSelect := OnSelectHandler;
                pic.Material := material;
                pic.Name := 'picture'+IntToStr(count);                          // identifier
                pic.Parent := sb;
                pic.Cursor := crHandPoint;
                pic.Tag := i;
                pic.UpdateGUI;
                //if SelectedMaterial=material then
                //    pic.Selected := true;
                inc(count);
            end;
        end;
        // this scrolls the material frames to their correct position
        // according to the size (width/height) of the parent window
        UpdateCategory;
    end;
end;

procedure TfrmMaterials.UpdateCategory;
var i,x,y: integer;
    tab: TTabSheet;
    sb: TScrollbox;
    materialFrame: TMaterialFrame;
begin
    tab := pcMaterials.ActivePage;
    if Assigned(tab) then begin
        sb := TScrollbox(tab.FindChildControl('scrollbox'));
        if Assigned(sb) then begin
            sb.VertScrollBar.Position := 0;
            x := 5;
            y := 5;
            for i := 0 to sb.ControlCount-1 do begin
                materialFrame := TMaterialFrame(sb.FindComponent('picture'+IntToStr(i)));
                if Assigned(materialFrame) then begin
                    materialFrame.Left := x;
                    materialFrame.Top := y;
                    x := x + materialFrame.Width + 5;
                    if sb.Width-5-x < 65 then begin
                        x := 5;
                        y := y + 85;
                    end;
                end;
            end;
        end;
    end;
end;

procedure TfrmMaterials.ClearGUI;
var i: integer;
begin
    for i := 0 to pcMaterials.PageCount-1 do
        pcMaterials.Pages[0].Free;
    SelectedCategory := '';     // JCC 06-03-2008
end;

procedure TfrmMaterials.UpdateGUI;
var currentCategory: string;
begin
    currentCategory := SelectedCategory;
    ClearGUI;
    CreateCategories;
    SelectedCategory := currentCategory;
    FillCategory;
end;

procedure TfrmMaterials.FormDblClick(Sender: TObject);
// Due to destroy-in-event problems, the materialFrame cannot handle it's own doubleclick
// So that leaves us to handle it...
var tab: TTabSheet;
    sb: TScrollbox;
    position: TPoint;
    i: integer;
    frame: TMaterialFrame;
begin
    if Assigned(fOnOpen) then begin
        tab := pcMaterials.ActivePage;
        sb := TScrollbox(tab.FindChildControl('scrollbox'));
        position := sb.ScreenToClient(Mouse.CursorPos);
        for i := 0 to sb.ControlCount-1 do begin
            frame := TMaterialFrame(sb.FindComponent('picture'+IntToStr(i)));
            if (position.X >= frame.Left + frame.Image1.Left) and
               (position.X < frame.Left + frame.Image1.Left + frame.Image1.Width) and
               (position.Y >= frame.Top + frame.Image1.Left) and
               (position.Y < frame.Top + frame.Image1.Left + frame.Image1.Height) then begin
                fOnOpen(frame);
                break;        // Important, after fOnOpen the remaining frames might have been destroyed.
            end;
        end;
    end;
end;

end.
