{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit frmRaytracingForm;

{$MODE Delphi}

interface

uses
  GR32,
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, frmDockableForm, Grids,
  ComCtrls, ToolWin, ExtCtrls, ImgList,
  unit_RaytracingThread, unit_Scene, unit_ViewPortManager, unit_RaytraceManager,
  unit_Options, unit_TextureTypes;

type
  TfrmRaytracing = class(TfrmDockable)
    tbRaytracing: TToolBar;
    tbAdd: TToolButton;
    ilRaytracing: TImageList;
    tbRemove: TToolButton;
    tmRefresh: TTimer;
    dgItems: TDrawGrid;
    ToolButton1: TToolButton;
    tbSave: TToolButton;
    tbPlay: TToolButton;
    tbStop: TToolButton;
    ToolButton5: TToolButton;
    procedure tbAddClick(Sender: TObject);
    procedure tmRefreshTimer(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure dgItemsDrawCell(aSender: TObject; aCol, aRow: Integer; aRect: TRect; aState: TGridDrawState);
    procedure tbRemoveClick(Sender: TObject);
    procedure dgItemsSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure tbSaveClick(Sender: TObject);
    procedure tbPlayClick(Sender: TObject);
    procedure tbStopClick(Sender: TObject);
    procedure dgItemsDblClick(Sender: TObject);
    procedure dgItemsKeyPress(Sender: TObject; var Key: Char);
  private
    fScene: TScene;
    fViewportManager: TViewportManager;
    fRaytracingThread: TRaytracingThread;
    fCameraOptions: TCameraOptions;

    fPreviousSelectedItem: TRaytracingItem;
    fPreviousItemStatus: TRaytracingItemStatus;
    fThumbnailBuffer: TBitmap32;
    fProcessingBitmap: TBitmap32;
    fPausedBitmap: TBitmap32;
    fDoneBitmap: TBitmap32;

    function SelectedItem: TRaytracingItem;

    procedure OpenSelectedItem;
    procedure UpdateItems;
    procedure UpdateColumnSizes;
    procedure UpdateButtonState;

    function MakeThumbnailRect( const aSourceRect, aTargetRect: TRect ): TRect;
    function MakeStatusBitmap( const aResourceName: String ): TBitmap32;
    function MaxColorToOne( const aColor: TFloatPixel ): TFloatPixel;
    function ConvertTextureData( const aTextureData: TFloatTextureData ): TTextureData;

  public
    constructor Create(
      const aScene: TScene;
      const aViewportManager: TViewportManager;
      const aRaytracingThread: TRaytracingThread;
      const aCameraOptions: TCameraOptions ); reintroduce;
    destructor Destroy; override;
    property Thread: TRaytracingThread read fRaytracingThread;
  end;

implementation

{$R *.lfm}

uses
    Math, 
    unit_Variables,
    unit_Cameras,
    unit_Constants,
    unit_TextureManipulator,
    unit_TextureWriter,
    unit_TextureLoader,
    unit_Functions,
    frmViewPortForm,
    frm3dView,
    frmRaytracingPreviewForm,
    frmRaytracingOptionsForm,
    RTMathFn;

const
    DRAWGRID_NO_SELECTION: TGridRect = (Left:-1; Top:-1; Right:-1; Bottom:-1 );

constructor TfrmRaytracing.Create(
    const aScene: TScene;
    const aViewportManager: TViewportManager;
    const aRaytracingThread: TRaytracingThread;
    const aCameraOptions: TCameraOptions );
begin
    inherited Create( nil );

    fScene := aScene;
    fViewportManager := aViewportManager;
    fRaytracingThread := aRaytracingThread;
    fCameraOptions := aCameraOptions;

    dgItems.RowHeights[0] := 17;
    dgItems.DoubleBuffered := TRUE;
    dgItems.Font.Name := 'Arial';
    dgItems.Selection := DRAWGRID_NO_SELECTION;

    fThumbnailBuffer := TBitmap32.Create;
    fThumbnailBuffer.Width := RAYTRACER_THUMBNAIL_SIZE;
    fThumbnailBuffer.Height := RAYTRACER_THUMBNAIL_SIZE;

    fProcessingBitmap := MakeStatusBitmap( 'imgRaytracerProcessing' );
    fPausedBitmap := MakeStatusBitmap( 'imgRaytracerPaused' );
    fDoneBitmap := MakeStatusBitmap( 'imgRaytracerDone' );
end;

destructor TfrmRaytracing.Destroy;
begin
    FreeAndNil( fDoneBitmap );
    FreeAndNil( fPausedBitmap );
    FreeAndNil( fProcessingBitmap );
    FreeAndNil( fThumbnailBuffer );

    inherited Destroy;
end;

function TfrmRaytracing.MakeStatusBitmap( const aResourceName: String ): TBitmap32;
var loader: TBitmap32TextureLoader;
begin
    result := TBitmap32.Create;
    result.DrawMode := dmBlend;
    result.MasterAlpha := 128;
    loader := TBitmap32TextureLoader.Create( result );
    try
        loader.LoadPNGFromResource( aResourceName );
    finally
        FreeAndNil( loader );
    end;
end;

procedure TfrmRaytracing.tbAddClick(Sender: TObject);
var optionsForm: TFrmRaytracingOptions;
    viewport: Tfrm3D;
    options: TRaytracingOptions;
    thumbnail: TTextureData;
    camera: TCamera3D;
begin
  inherited;

  viewport := fViewportManager.Active3DViewport;

  if viewPort = nil then
      raise Exception.create( 'Raytracing without a 3D viewport not supported yet (WIP).'  );

  optionsForm := TfrmRaytracingOptions.create(viewPort);
  if optionsForm.ShowModal = mrOK then begin
      options := optionsForm.Options;
      thumbnail := optionsForm.Thumbnail;

      // INFO, PJ: Temporary hack, until we truely support storing FOV's in camera.
      camera := viewPort.Camera3D;
      camera.FOV := fCameraOptions.FOV;
      fRaytracingThread.AddToQueue( fScene, camera, options, thumbnail );
  end;
  FreeAndNil( optionsForm );

  UpdateItems;
end;

procedure TfrmRaytracing.UpdateItems;
begin
    dgItems.ColCount := 2;
    // INFO, PJ: Little workaround for Delphi's TCustomGrid FixedRows < NumRows restriction
    dgItems.RowCount := Max( fRaytracingThread.NumItems + 1, dgItems.FixedRows + 1 );

    if fRaytracingThread.NumItems = 0 then
        dgItems.Selection := DRAWGRID_NO_SELECTION;

    UpdateColumnSizes;
    UpdateButtonState;
end;

procedure TfrmRaytracing.tmRefreshTimer(Sender: TObject);
begin
  UpdateItems;
  dgItems.invalidate;
end;

procedure TfrmRaytracing.FormResize(Sender: TObject);
begin
    UpdateColumnSizes;
end;

procedure TfrmRaytracing.UpdateColumnSizes;
var i, totalWidth: integer;
begin
    totalWidth := dgItems.ClientWidth;

    // The size of the second column auto-adjusts
    for i := 0 to dgItems.ColCount - 2 do begin
        totalWidth := totalWidth - dgItems.ColWidths[i];
    end;

    dgItems.ColWidths[1] := totalWidth;
end;

function TfrmRaytracing.MakeThumbnailRect( const aSourceRect, aTargetRect: TRect ): TRect;
var sourceWidth, sourceHeight, targetWidth, targetHeight: Integer;
    widthRatio, heightRatio: Single;
begin
    sourceWidth := aSourceRect.Right - aSourceRect.Left;
    sourceHeight := aSourceRect.Bottom - aSourceRect.Top;

    targetWidth := aTargetRect.Right - aTargetRect.Left;
    targetHeight := aTargetRect.Bottom - aTargetRect.Top;


    widthRatio := targetWidth / sourceWidth;
    heightRatio := targetHeight / sourceHeight;

    if widthRatio > heightRatio then
    begin
        result := Rect(
            aTargetRect.Left + Trunc(( targetWidth - sourceWidth * heightRatio ) / 2 ),
            aTargetRect.Top,
            aTargetRect.Left + Trunc(( targetWidth + sourceWidth * heightRatio ) / 2 ),
            aTargetRect.Bottom );
    end
    else begin
        result := Rect(
            aTargetRect.Left,
            aTargetRect.Top + Trunc(( targetHeight - sourceHeight * widthRatio ) / 2 ),
            aTargetRect.Right,
            aTargetRect.Top + Trunc(( targetHeight + sourceHeight * widthRatio ) / 2 ));
    end;

end;

procedure TfrmRaytracing.dgItemsDrawCell(aSender: TObject; aCol, aRow: Integer; aRect: TRect; aState: TGridDrawState);
var isDummyRow: Boolean;
    item: TRaytracingItem;
    text: string;
    backgroundRect: TRect;

begin
    dgItems.Canvas.Font := dgItems.Font;

    isDummyRow := (aRow = 1) and (fRaytracingThread.NumItems = 0);

    // Draw the background... we do this in the first column only to enable
    // drawing text overlapping both columns.
    if (aCol = 0) and not isDummyRow then begin
        backgroundRect := Rect( 0, aRect.Top, dgItems.ClientWidth, aRect.Bottom );

        if gdFocused in aState then
        begin
            dgItems.Canvas.Brush.Color := $D4AF99;
            dgItems.Canvas.FillRect( backgroundRect );
        end
        else if gdSelected in aState then
        begin
            dgItems.Canvas.Brush.Color := $EEE5E2;
            dgItems.Canvas.FillRect( backgroundRect );
        end
        else
            FillGradientRect( dgItems.Canvas, backgroundRect, $FBFAFB, $EDEAEB );

        dgItems.Canvas.Pen.Color := $A7A6AA;
        dgItems.Canvas.MoveTo( backgroundRect.Left, backgroundRect.Bottom - 1 );
        dgItems.Canvas.LineTo( backgroundRect.Right, backgroundRect.Bottom - 1 );
    end;

    //INFO, PJ: Header
    if aRow = 0 then begin
        if aCol = 0 then begin
            SetBKMode( dgItems.Canvas.Handle, TRANSPARENT );
            dgItems.Canvas.Font.Style := [fsBold];
            InflateRect( aRect, -2, -1 );
            dgItems.Canvas.TextOut( 3, 1, 'Raytracing Queue' );
        end
        else
            dgItems.Canvas.Font.Style := [];

    end
    else if isDummyRow then begin
        // INFO, PJ: Do nothing. Workaround for Delphi's TCustomGrid FixedRows < NumRows restriction
    end
    else if aCol = 0 then begin
        dgItems.Canvas.Brush.Color := $EDE9EB;
        dgItems.Canvas.FrameRect( Rect( aRect.Left + 2, aRect.Top + 2, aRect.Left + RAYTRACER_THUMBNAIL_SIZE + 4, aRect.Top + RAYTRACER_THUMBNAIL_SIZE + 4 ));
        fThumbnailBuffer.FillRect( 0, 0, RAYTRACER_THUMBNAIL_SIZE, RAYTRACER_THUMBNAIL_SIZE, clBlack32 );
        item := fRaytracingThread.Items[aRow - 1];
        item.Thumbnail.DrawTo( fThumbnailBuffer, MakeThumbnailRect( item.Thumbnail.ClipRect, fThumbnailBuffer.ClipRect ));

        case item.Status of
            risProcessing: begin
                if GetTickCount mod 1500 < 1000 then
                    fProcessingBitmap.DrawTo( fThumbnailBuffer, fThumbnailBuffer.width - fProcessingBitmap.Width, fThumbnailBuffer.Height - fProcessingBitmap.Height );
            end;

            risPaused:
                fPausedBitmap.DrawTo( fThumbnailBuffer, fThumbnailBuffer.width - fPausedBitmap.Width, fThumbnailBuffer.Height - fPausedBitmap.Height );

            risDone:
                fDoneBitmap.DrawTo( fThumbnailBuffer, fThumbnailBuffer.width - fDoneBitmap.Width, fThumbnailBuffer.Height - fDoneBitmap.Height );

        end;

        fThumbnailBuffer.DrawTo( dgItems.Canvas.Handle, aRect.Left + 3, aRect.Top + 3 );
    end
    else begin
        item := fRaytracingThread.Items[aRow - 1]; // REFA, PJ: Not thread-safe
        text := Format(
            '%s' + CRLF +
            '%d x %d pixels',
            [item.SceneName, item.Options.Width, item.Options.Height] );

        if item.Status in [risProcessing, risPaused] then
            text := text + Format( CRLF + 'Progress: %d%%', [item.PercentageDone] );

    end;

    if text <> '' then begin
        SetBKMode( dgItems.Canvas.Handle, TRANSPARENT ); 
        InflateRect( aRect, -2, -1 );
        DrawText( dgItems.Canvas.Handle, PChar( text ), -1, aRect, DT_LEFT or DT_WORDBREAK );
    end;
end;

procedure TfrmRaytracing.tbRemoveClick(Sender: TObject);
begin  
    fRaytracingThread.RemoveFromQueue( dgItems.Selection.Top - dgItems.FixedRows );

    UpdateItems;
end;

procedure TfrmRaytracing.dgItemsSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
begin
    CanSelect := fRaytracingThread.NumItems > 0;

    UpdateButtonState;
end;

function TfrmRaytracing.SelectedItem: TRaytracingItem;
var itemIndex: integer;
begin
    itemIndex := dgItems.Selection.Top - dgItems.FixedRows;
    if  ( itemIndex >= 0 ) and
        ( itemIndex < fRaytracingThread.NumItems ) then begin
        result := fRaytracingThread.Items[itemIndex];
    end
    else
        result := nil;
end;

procedure TfrmRaytracing.UpdateButtonState;
var
    //itemIndex: integer;
    item: TRaytracingItem;
begin
    item := selectedItem;

    if item = nil then
    begin
        tbSave.Enabled := false;
        tbRemove.Enabled := false;
        tbPlay.Enabled := false;
        tbStop.Enabled := false;
    end
    else if (item <> fPreviousSelectedItem) or (item.Status <> fPreviousItemStatus) then begin
        tbSave.Enabled := (item.Status = risDone);
        tbRemove.Enabled := true;
        tbPlay.Enabled := (item.Status in [risNew, risPaused]);
        tbStop.Enabled := (item.Status in [risPaused, risProcessing]);

        fPreviousSelectedItem := item;
        fPreviousItemStatus := item.Status;
    end;
end;

function TfrmRaytracing.MaxColorToOne( const aColor: TFloatPixel ): TFloatPixel;
var
  maxValue: Single;

begin
  maxValue := Max( aColor.R, aColor.G, aColor.B );

  if maxValue > 1 then
  begin
      result.R := aColor.R / maxValue;
      result.G := aColor.G / maxValue;
      result.B := aColor.B / maxValue;
  end
  else
      result := aColor;
end;

function TfrmRaytracing.ConvertTextureData( const aTextureData: TFloatTextureData ): TTextureData;
var i: Integer;
    source: PFloatPixel;
    color: TFloatPixel;
    target: PTexturePixel24;

begin
    // INFO, PJ: Simlar conversions now occur:
    //  - in the generation of the thumbnails within the raytracing mechanism
    //  - in the raytracing preview
    //  - here. Change some day.
    result := TextureData( aTextureData.Width, aTextureData.Height, 24 );

    source := @aTextureData.Pixels[0];
    target := @result.Pixels[0];
    for i := 0 to aTextureData.Width * aTextureData.Height - 1 do begin
        color := MaxColorToOne( source^ );

        target[0] := Trunc( 255 * color.B );
        target[1] := Trunc( 255 * color.G );
        target[2] := Trunc( 255 * color.R );

        inc( source );
        inc( target );
    end;
end;

procedure TfrmRaytracing.tbSaveClick(Sender: TObject);
var item: TRaytracingItem;
    saveDialog: TSaveDialog;
    textureWriter: TTextureWriter;
begin
    item := selectedItem;
    if item = nil then exit;

    saveDialog := TSaveDialog.create( nil );
    try
        saveDialog.filter := TTextureWriter.SupportedFilesMask;

        if saveDialog.execute then begin
            textureWriter := TTextureWriter.Create;

            try
                textureWriter.SaveTextureData( ConvertTextureData( item.Output ), saveDialog.Filename);
            finally
                FreeAndNil( textureWriter );
            end;
        end;
    finally
        FreeAndNil( saveDialog );
    end;
end;

procedure TfrmRaytracing.tbPlayClick(Sender: TObject);
var item: TRaytracingItem;
begin
    item := SelectedItem;

    if Assigned( item ) then
        fRaytracingThread.StartItem( item );
end;

procedure TfrmRaytracing.tbStopClick(Sender: TObject);
var item: TRaytracingItem;
begin
    item := SelectedItem;

    if Assigned( item ) then
        fRaytracingThread.StopItem( item );
end;

procedure TfrmRaytracing.dgItemsDblClick(Sender: TObject);
begin
  OpenSelectedItem;
end;

procedure TfrmRaytracing.dgItemsKeyPress(Sender: TObject; var Key: Char);
begin
  if key = 'a' then OpenSelectedItem;
end;

procedure TfrmRaytracing.OpenSelectedItem;
var item: TRaytracingItem;
    previewForm: TfrmRaytracingPreview;
begin
    item := SelectedItem;

    if Assigned( item ) then begin
        previewForm := TfrmRaytracingPreview.Create( item, fRaytracingThread );
        try
            previewForm.ShowModal;
            
        finally
            FreeAndNil( previewForm );
        end;
    end;

    // INFO, PJ: Flush cached output data
    item.Invalidate;
end;

end.
