unit fraInspectorPath;

{$MODE Delphi}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, ValEdit, fraInspectorDefault,
  unit_Enums,
  unit_Variables,
  unit_Paths;

type
  TfraInspectorPaths = class(TfraInspectorDefaultFrame)
    vlePath: TValueListEditor;
    procedure vlePathStringsChange(Sender: TObject);
    procedure vlePathEditButtonClick(Sender: TObject);
    procedure vlePrimitiveDrawCell(Sender: TObject; ACol, ARow: Integer; ARect: TRect; State: TGridDrawState);
    procedure vlePathSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
  private
    fControlPointPropertiesVisible: Boolean;
    procedure UpdateControlPointProperties(const aCP: TPathControlPoint);
    procedure DeleteRow( const anIndex: integer );
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure UpdateInspector; override;
  end;

var
  fraInspectorPaths: TfraInspectorPaths;

implementation

uses
    frmCommentsForm,
    frmPropertiesForm;

{$R *.lfm}


{ TfraInspectorPaths }


constructor TfraInspectorPaths.Create(AOwner: TComponent);
var
  Strings: TStringList;
begin
    inherited Create(AOwner);
    Width := 220;
    Height := 310;
    Strings := TStringList.Create;
    try
      //Boolean options
      Strings.Add('True');
      Strings.Add('False');

      //Define values for properties
      vlePath.ItemProps['Closed'].EditStyle := esPickList;
      vlePath.ItemProps['Closed'].PickList := Strings;
      vlePath.ItemProps['ID'].ReadOnly := true;
      vlePath.ItemProps['Controlpoints'].ReadOnly := true;
      vlePath.ItemProps['Comments'].EditStyle := esEllipsis;
      vlePath.ItemProps['Tag'].EditStyle := esEllipsis;

      //Set options for snap method selection
      Strings.Clear;
      Strings.add('None');
      Strings.add('Vertex');
      Strings.add('Center');
      vlePath.ItemProps['Snap method'].EditStyle := esPickList;
      vlePath.ItemProps['Snap method'].PickList := Strings;

      //Set options for path-type selection
      Strings.Clear;
      Strings.add('Linear');
      Strings.add('CubicBezier');
      Strings.add('CatmullRom');
      Strings.add('Hermite');
      vlePath.ItemProps['Path-type'].EditStyle := esPickList;
      vlePath.ItemProps['Path-type'].PickList := Strings;

      vlePath.Col := 1;
    finally
      Strings.Free;
    end;

    fControlPointPropertiesVisible := False;
end;

destructor TfraInspectorPaths.Destroy;
begin
  vlePath.Strings.Clear;
  inherited;
end;

procedure TfraInspectorPaths.UpdateInspector;
var
  Path: TPath;
begin
    if Primitives.Count>0 then
    begin
        nowUpdating := true;
        if not (Primitives[0] is TPath) then Exit;
        Path := TPath(Primitives[0]);

        with vlePath do
        begin
            Enabled := true;
            Values['Name'] := Path.Name;
            Values['Comments'] := Path.Comments;
            Values['Tag'] := Path.Tag;
            Values['Closed'] := BoolToStr(Path.IsClosed, true);
            Values['Path-Type'] := PathTypeToString(Path.PathType);
            Values['ID'] := IntToStr(Path.ID);
            Values['Controlpoints'] := IntToStr(Path.ControlPointCount);
            case Path.SnapMethod of
                smNone:     Values['Snap method'] := 'None';
                smVertex:   Values['Snap method'] := 'Vertex';
                smCenter:   Values['Snap method'] := 'Center';
            end;
        end;

        //Pass the first selected controlpoint to the controlpoint property updater
        UpdateControlPointProperties(  Path.FirstSelectedControlPoint );

        nowUpdating := false;
    end;
end;

procedure TfraInspectorPaths.vlePathStringsChange(Sender: TObject);
var prop: string;
    i: integer;
    path: TPath;
    value: string;
  J: Integer;
begin
  if Assigned(Primitives) and not (nowupdating) then
  begin

    prop  := vlePath.Keys[vlePath.Row];
    value := vlePath.Values[prop];

    for i := 0 to Primitives.Count-1 do
      if Primitives[I] is TPath then
      begin
        Path := TPath(Primitives[i]);

        if prop='Name' then
        begin
          Path.name := value;
          if Path.name='' then
              Path.name := 'path';
          if Assigned(Path.treenode) then
              Path.treenode.Text := Path.name;
        end;

        if prop='Comments' then
        begin
          Path.Comments := Value;
        end;

        if prop='Tag' then
        begin
          Path.Tag := Value;
        end;

        if prop='Snap method' then
        begin
          if value='None' then Path.snapmethod := smNone;
          if value='Vertex' then Path.snapmethod := smVertex;
          if value='Center' then Path.snapmethod := smCenter;
        end;

        if prop='Path-type' then
          Path.PathType := StringToPathType(value);

        if prop='Closed' then
          Path.IsClosed := StringToBoolean(value);

        if prop='Controlpoint Tag' then
          for J := 0 to Path.SelectedControlPoints.Count - 1 do
            TPathControlPoint(Path.SelectedControlPoints[J]).Tag := Value;
      end;
  end;
end;


procedure TfraInspectorPaths.vlePathEditButtonClick(Sender: TObject);
var prop: string;
    value: string;
    commentsForm: TfrmComments;
    propertiesForm: TfrmProperties;
    description: string;
begin
    prop := vlePath.Keys[vlePath.Row];
    value := vlePath.Values[prop];

    if prop = 'Comments' then
    begin
        description := Format('primitive "%s"', [vlePath.Values['Name']]);
        commentsForm := TfrmComments.Create(value, description);
        try
            if commentsForm.ShowModal = mrOK then
                vlePath.Values[prop] := commentsForm.Comments;
        finally
            FreeAndNil(commentsForm);
        end;
    end
    else if prop = 'Tag' then
    begin
        description := Format('primitive "%s"', [vlePath.Values['Name']]);
        propertiesForm := TfrmProperties.Create(value, description);
        try
            if propertiesForm.ShowModal = mrOK then
                vlePath.Values[prop] := propertiesForm.Tag;
        finally
            FreeAndNil(propertiesForm);
        end;
    end
    else if prop = 'Controlpoint Tag' then
    begin
        description := Format('Path "%s"', [vlePath.Values['Name']]);
        propertiesForm := TfrmProperties.Create(value, description);
        try
            if propertiesForm.ShowModal = mrOK then
                vlePath.Values[prop] := propertiesForm.Tag;
        finally
            FreeAndNil(propertiesForm);
        end;
    end
end;

procedure TfraInspectorPaths.UpdateControlPointProperties(const aCP: TPathControlPoint);
var
  shouldShow: boolean;
begin
    nowupdating := true;

    shouldShow := ( editMode in [emVertex] ) and Assigned( aCP );

    if shouldShow then begin
        if fControlPointPropertiesVisible then begin
            vlePath.Values['Controlpoint Tag'] := aCP.Tag;
        end else begin
            vlePath.Strings.Add( 'Controlpoint Properties=' );
            vlePath.ItemProps['Controlpoint Properties'].EditMask := 'header';
            vlePath.ItemProps['Controlpoint Properties'].ReadOnly := TRUE;

            vlePath.Strings.Add( 'Controlpoint Tag=' + aCP.Tag);
            vlePath.ItemProps['Controlpoint Tag'].EditStyle := esEllipsis;
        end;
    end
    else begin
        if fControlPointPropertiesVisible then begin
            DeleteRow( vlePath.Strings.Count - 1 );
            DeleteRow( vlePath.Strings.Count - 1 );
        end;
    end;

    fControlPointPropertiesVisible := shouldShow;
    nowUpdating := false;
end;

type
    THackGrid = class( TCustomGrid );

procedure TfraInspectorPaths.DeleteRow( const anIndex: integer );
var selection: TGridRect;
begin
    if vlePath.Row = anIndex then
    begin
        selection.Left := 1;
        selection.Top :=  0;
        selection.Right := 1;
        selection.Bottom := 0;

        vlePath.Selection := selection;
    end;

    vlePath.DeleteRow( anIndex );
end;

procedure TfraInspectorPaths.vlePrimitiveDrawCell(Sender: TObject; ACol, ARow: Integer; ARect: TRect; State: TGridDrawState);
// TODO, PJ: Make this xxxDrawCell handlers (copied 5(!) times in the current setup)
//           generic. For example, by introducing a TValueListEditorController class.
var key: string;
    value: string;
    TextRect: TRect;
begin
    with Sender as TValueListEditor do begin
        key := Keys[aRow];
        value := Values[key];

        if ItemProps[ARow].EditMask = 'header' then begin
            // Draw header cells
            if ACol = 0 then
            begin
                Canvas.Brush.Color := clBtnFace;
                Canvas.FillRect(Rect( Parent.ClientRect.Left, ARect.Top, Parent.ClientRect.Right - 1, ARect.Bottom ));
                Canvas.Font.Style := [fsBold];
                Canvas.Font.Color := clBlack;
                TextRect := Rect( Parent.ClientRect.Left, ARect.Top, Parent.ClientRect.Right - 1, ARect.Bottom );

                Canvas.TextRect( TextRect, ARect.Left, ARect.Top, Cells[ACol,ARow]);
                //DrawText(Canvas.Handle,PChar(Cells[ACol,ARow]),-1,TextRect,DT_SINGLELINE);

            end;
        end
        else begin
            // Draw ordinary cells
            Canvas.Font.Style := [];
            if not Enabled then
                Canvas.Font.Color := clGray
            else begin
                if Odd(ACol) then begin
                    if ItemProps[ARow].ReadOnly then
                        Canvas.Font.Color := clGrayText
                    else
                        Canvas.Font.Color := clNavy;
                end else
                    Canvas.Font.Color := clBlack;
            end;
            Canvas.Brush.Color := clBtnFace;
            Canvas.FillRect(ARect);


            DrawText(Canvas.Handle,PChar(Cells[ACol,ARow]),-1,ARect,DT_SINGLELINE);
        end;
        if gdFocused in State then
            Canvas.DrawFocusRect(ARect);
    end;
end;

//Makes sure that some cells are not selectable
procedure TfraInspectorPaths.vlePathSelectCell(Sender: TObject; ACol, ARow: Integer; var CanSelect: Boolean);
var itemProp: TItemProp;
begin
    itemProp := vlePath.ItemProps[ARow];

    // INFO, PJ: The 'assigned' prop is to workaround a VCL bug when the valueListEditor is cleared.
    if ( ARow in [0,6,7] ) or
       (Assigned(itemProp) and (itemProp.editMask = 'header')) then begin
        CanSelect := false;
    end;
end;

end.
