{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit fraMaterial;

{$MODE Delphi}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, FileUtil, Contnrs,
  unit_Materials,
  unit_Variables,
  unit_TextureTypes,
  unit_Types,
  unit_Enums;

type
  TMaterialFrame = class(TFrame)
    Shape1: TShape;
    Image1: TImage;
    Label1: TLabel;
    procedure Image1Click(Sender: TObject);
    procedure Image1DblClick(Sender: TObject);
  private
    fOnSelect: TNotifyEvent;
    fMaterial: TMaterial;
    fSelected: boolean;
    procedure SetSelected(value: boolean);
    procedure SetMaterial(const aMaterial: TMaterial);
    procedure SetMaterialName(const aMaterialName: string);
    procedure BlendLayerTexture(const aSourceLayer: TTextureData; aTargetLayer: TTextureData; aBlendOperation: TBlendOperation);
    procedure BlendLayerColor(const aColor: TColor; aTargetLayer: TTextureData; aBlendOperation: TBlendOperation);
    procedure BlendSpecular(aTargetLayer: TTextureData);
    procedure DrawBackground(aTargetLayer: TTextureData);
  public
    constructor Create(AOwner: TComponent); reintroduce;
    destructor Destroy; override;
    procedure LoadFromFile(AFileName: string);
    procedure UpdateGUI;
    property OnSelect: TNotifyEvent read fOnSelect write fOnSelect;
    property Material: TMaterial read fMaterial write SetMaterial;
    property MaterialName: string write SetMaterialName;
    property Selected: boolean read fSelected write SetSelected;
  end;

implementation

uses math,
     unit_Functions,
     unit_TextureLoader;

{$R *.lfm}

destructor TMaterialFrame.Destroy;
begin
    inherited destroy;
end;


// ### TMaterialFrame ##########################################################

procedure TMaterialFrame.SetSelected(value: boolean);
begin
    fselected := value;
    if fselected then begin
        Shape1.Brush.Color := clRed;
        Label1.Font.Color := clRed;
    end else begin
        Shape1.Brush.Color := clBtnFace;
        Shape1.Pen.Color := clBtnFace;
        Label1.Font.Color := clWindowText;
    end;
end;

procedure TMaterialFrame.SetMaterial(const aMaterial: TMaterial);
begin
    fMaterial := aMaterial;
    MaterialName := fMaterial.Name;
end;

procedure TMaterialFrame.SetMaterialName(const aMaterialName: string);
begin
    Label1.Caption := Lowercase(aMaterialName);
    Label1.Hint := Label1.Caption;
    Image1.Hint := Label1.Caption;
end;

procedure TMaterialFrame.Image1Click(Sender: TObject);
begin
    Selected := true;
    if Assigned(fOnSelect) then
        fOnSelect(self);
end;

procedure TMaterialFrame.UpdateGUI;
var layer: TTextureLayer;
    textureLoader: TTextureLoader;
    previewData,textureData: TTextureData;
    i: integer;
    bitmapInfo: TBitmapInfo;
begin
    // update the material reference picture by blending each layer into the final picture
    // using the current settings of each layer

    image1.Picture.Bitmap.Width := Image1.Width;
    image1.Picture.Bitmap.Height := Image1.Height;
    image1.Picture.Bitmap.PixelFormat := pf32bit;
    previewData.Width := Image1.Width;
    previewData.Height := Image1.Height;
    previewData.BitsPerPixel := 32;
    SetLength(previewData.Pixels, previewData.Width * previewData.Height * ( previewData.BitsPerPixel div 8 ));

    textureLoader := TTextureLoader.Create;
    try
        if Assigned(Material) then begin
            if (Material.NumTextureLayers > 0) and Material.TextureLayers[0].isTransparent then
                DrawBackground(previewData);

            for i := 0 to Material.NumTextureLayers-1 do begin
                layer := Material.TextureLayers[i];

                if Assigned(layer) and (layer.BlendOperation <> boIgnore ) and ( i <> LIGHTMAP_TEXTURE_LAYER_INDEX) then begin
                    //if Assigned(layer.Texture) then begin
                    if layer.Texture.isColorTexture then
                        BlendLayerColor( layer.Color, previewData, layer.BlendOperation)
                    else begin
                        textureData := textureLoader.LoadTextureData(layer.Texture.filename);
                        BlendLayerTexture( textureData, previewData, layer.BlendOperation );
                    end;
                end;
            end;


            if toggleMaterialPreviewShowSpecular and (Material.NumTextureLayers > 0) then
                BlendSpecular( previewData );
        end;
    finally
        FreeAndNil(textureLoader);
    end;

    FillChar(bitmapInfo, SizeOf(bitmapInfo), 0);
    bitmapInfo.bmiHeader.biSize := SizeOf( TBitmapInfoHeader );
    bitmapInfo.bmiHeader.biWidth := previewData.Width;
    bitmapInfo.bmiHeader.biHeight := -previewData.Height;
    bitmapInfo.bmiHeader.biPlanes := 1;
    bitmapInfo.bmiHeader.biBitCount := 32;
    bitmapInfo.bmiHeader.biCompression := BI_RGB;
    SetDIBits( Image1.Canvas.Handle, Image1.Picture.Bitmap.Handle, 0, previewData.Height, @previewData.Pixels[0], bitmapInfo, DIB_RGB_COLORS );
    image1.Invalidate;
end;

// blends sourceLayer into targetLayer using the given blendOperation
procedure TMaterialFrame.BlendLayerTexture(const aSourceLayer: TTextureData; aTargetLayer: TTextureData; aBlendOperation: TBlendOperation);
var targetX, targetY, sourceX, sourceY, sourceXPlus1, sourceYPlus1, width, height, targetIndex, targetLUTIndex, i, component: integer;
    nrBytesPerSourcePixel, nrBytesPerTargetPixel, nrBytesToCopy, alphaValue: integer;
    x, y, alpha, beta: single;
    sourcePixel: TTexturePixel;
    targetPixel, diffuseColor: PTexturePixel;
    lutEntry: PLookupTableEntry;
    blendOperation: TBlendOperation;
begin
    nrBytesPerSourcePixel := aSourceLayer.BitsPerPixel div 8;
    nrBytesPerTargetPixel := aTargetLayer.BitsPerPixel div 8;
    nrBytesToCopy := Min(nrBytesPerSourcePixel, nrBytesPerTargetPixel);

    if (aBlendOperation = boAlphaBlend) and (aSourceLayer.BitsPerPixel <> 32) then
    begin
        blendOperation := boReplace;
    end
    else
        blendOperation := aBlendOperation;

    width := aSourceLayer.Width;
    height := aSourceLayer.Height;

    targetLUTIndex := 0;
    targetIndex := 0;
    for targetY := 0 to aTargetLayer.Height - 1 do begin

        for targetX := 0 to aTargetLayer.Width - 1 do begin
            lutEntry := @MaterialPreviewLUT[targetLUTIndex];
            if ( lutEntry.U <> -1 ) and ( lutEntry.V <> -1 ) then begin
                // bilinear filtered texture lookup
                x := lutEntry.U;
                y := lutEntry.V;

                x := x * width - 0.5;
                y := y * height - 0.5;

                if x < 0 then x := x + width;
                if y < 0 then y := y + height;


                alpha := Frac(x);
                beta := Frac(y);

                sourceX := Trunc(x);
                sourceY := Trunc(y);

                sourceXPlus1 := sourceX + 1;
                if sourceXPlus1 >= width then sourceXPlus1 := sourceXPlus1 - width;

                sourceYPlus1 := sourceY + 1;
                if sourceYPlus1 >= height then sourceYPlus1 := sourceYPlus1 - height;

                for i := 0 to nrBytesPerSourcePixel - 1 do begin
                    component := Trunc(
                        (1-alpha) * (1 - beta) * aSourceLayer.Pixels[(sourceY * width + sourceX) * nrBytesPerSourcePixel + i] +
                        alpha * (1 - beta) * aSourceLayer.Pixels[(sourceY * width + sourceXPlus1) * nrBytesPerSourcePixel + i] +
                        (1-alpha) * beta * aSourceLayer.Pixels[(sourceYPlus1 * width + sourceX) * nrBytesPerSourcePixel + i] +
                        alpha * beta * aSourceLayer.Pixels[(sourceYPlus1 * width + sourceXPlus1) * nrBytesPerSourcePixel + i] );

                    sourcePixel[i] := component;
                end;
            end
            else begin
                FillChar( sourcePixel, SizeOf(sourcePixel), 0 );
            end;

            targetPixel := @aTargetLayer.Pixels[targetIndex];

            case blendOperation of
                boReplace:
                begin
                    for i := 0 to nrBytesToCopy - 1 do begin
                        targetPixel[i] := sourcePixel[i];
                    end;
                end;

                boAdd:
                begin
                    for i := 0 to nrBytesToCopy - 1 do begin
                        component := targetPixel[i] + sourcePixel[i];

                        if component > 255 then
                            component := 255;

                        targetPixel[i] := component;
                    end;
                end;

                boModulate:
                begin
                    for i := 0 to nrBytesToCopy - 1 do begin
                        component := ( targetPixel[i] * sourcePixel[i] ) shr 8;

                        targetPixel[i] := component;
                    end;
                end;

                boAlphaBlend:
                begin
                    alphaValue := sourcePixel[3];
                    for i := 0 to nrBytesToCopy - 1 do begin
                        targetPixel[i] := (alphaValue * sourcePixel[i] + (255 - alphaValue) * targetPixel[i]) div 256;
                    end;
                end;

                boNormalMap:
                begin
                    alphaValue := (sourcePixel[0] + sourcePixel[1] + sourcePixel[2]) div 3;
                    for i := 0 to nrBytesToCopy - 1 do begin
                        targetPixel[i] := targetPixel[i] * alphaValue div 256;
                    end;
                end;
            end;

            // Fill any uncopied bytes (at this moment, this is always the alpha) with white
            for i := nrBytesToCopy to nrBytesPerTargetPixel - 1 do begin
                targetPixel[i] := 255;
            end;

            // Apply colors from color map
            diffuseColor := @lutEntry.Diffuse;

            for i := 0 to nrBytesToCopy - 1 do begin
                component := ( targetPixel[i] * diffuseColor[i] ) shr 8;

                targetPixel[i] := component;
            end;

            targetLUTIndex := targetLUTIndex + 1;
            targetIndex := targetIndex + nrBytesPerTargetPixel;
        end;
    end;
end;

// blends sourceLayer into targetLayer using the given blendOperation
procedure TMaterialFrame.BlendLayerColor(const aColor: TColor; aTargetLayer: TTextureData; aBlendOperation: TBlendOperation);
var targetX, targetY, lutIndex, targetIndex, i, component: integer;
    sourcePixel: TTexturePixel;
    targetPixel, diffuseColor: PTexturePixel;
    lutEntry: PLookupTableEntry;
    bytesPerPixel: integer;
begin
    sourcePixel[0] := (aColor shr 16) and $000000FF;
    sourcePixel[1] := (aColor shr 8) and $000000FF;
    sourcePixel[2] := aColor and $000000FF;
    sourcePixel[3] := aColor shr 24;

    bytesPerPixel := aTargetLayer.BitsPerPixel div 8;

    lutIndex := 0;
    targetIndex := 0;
    for targetY := 0 to aTargetLayer.Height - 1 do begin
        for targetX := 0 to aTargetLayer.Width - 1 do begin
            lutEntry := @MaterialPreviewLUT[lutIndex];
            targetPixel := @aTargetLayer.Pixels[targetIndex];

            case aBlendOperation of
                boReplace:
                begin
                    for i := 0 to bytesPerPixel - 1 do
                    begin
                        targetPixel[i] := sourcePixel[i];
                    end;
                end;

                boAdd:
                begin
                    for i := 0 to bytesPerPixel - 1 do
                    begin
                        component := targetPixel[i] + sourcePixel[i];

                        if component > 255 then
                            component := 255;

                        targetPixel[i] := component;
                    end;
                end;

                boModulate:
                begin
                    for i := 0 to bytesPerPixel - 1 do
                    begin
                        component := ( targetPixel[i] * sourcePixel[i] ) shr 8;

                        targetPixel[i] := component;
                    end;
                end;

                boAlphaBlend:
                begin
                    for i := 0 to bytesPerPixel - 1 do begin
                        targetPixel[i] := (sourcePixel[3] * sourcePixel[i] + (255 - sourcePixel[3]) * targetPixel[i]) div 256;
                    end;
                end;
            end;

            // Apply colors from color map
            diffuseColor := @lutEntry.Diffuse;

            for i := 0 to 2 do begin
                component := ( targetPixel[i] * diffuseColor[i] ) shr 8;

                targetPixel[i] := component;
            end;

            lutIndex := lutIndex + 1;
            targetIndex := targetIndex + bytesPerPixel;
        end;
    end;
end;

procedure TMaterialFrame.BlendSpecular(aTargetLayer: TTextureData);
var targetX, targetY, sourceIndex, targetIndex, i, component, numBytesPerPixel: integer;
    targetPixel, specularColor: PTexturePixel;
    lutEntry: PLookupTableEntry;
begin
    numBytesPerPixel :=  aTargetLayer.BitsPerPixel div 8;
    sourceIndex := 0;
    targetIndex := 0;
    for targetY := 0 to aTargetLayer.Height - 1 do begin
        for targetX := 0 to aTargetLayer.Width - 1 do begin
            lutEntry := @MaterialPreviewLUT[sourceIndex];
            targetPixel := @aTargetLayer.Pixels[targetIndex];

            specularColor := @lutEntry.Specular;
            for i := 0 to 2 do begin
                component := targetPixel[i] + specularColor[i];

                if component > 255 then
                    component := 255;

                targetPixel[i] := component;
            end;

            sourceIndex := sourceIndex + 1;
            targetIndex := targetIndex + numBytesPerPixel;
        end;
    end;
end;

procedure TMaterialFrame.DrawBackground(aTargetLayer: TTextureData);
var targetIndex, targetX, targetY, i: integer;
    backgroundPixel, targetPixel: PTexturePixel;
begin
    targetIndex := 0;
    for targetY := 0 to aTargetLayer.Height - 1 do begin
        for targetX := 0 to aTargetLayer.Width - 1 do begin
            targetPixel := @aTargetLayer.Pixels[targetIndex];

            if (targetX div 8 + targetY div 8 ) mod 2 = 0 then
                backgroundPixel := @colorTransparencyGrid1
            else
                backgroundPixel := @colorTransparencyGrid2;

            for i := 0 to 2 do begin
                targetPixel[i] := backgroundPixel[i];
            end;

            targetIndex := targetIndex + 4;
        end;
    end;
end;

procedure TMaterialFrame.LoadFromFile(AFileName: string);
begin
    if not FileExistsUTF8(AFileName) then
        Label1.Caption := 'Image not found.'
    else
        Image1.Picture.LoadFromFile(AFileName);
end;

constructor TMaterialFrame.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    EnableDragImage(self);
end;

procedure TMaterialFrame.Image1DblClick(Sender: TObject);
// Possibly opens up the ME. However, this could eventually destroy this frame (refresh),
// so we need to make sure it happens outside this event. Very much a hacklike solution
begin
    PostMessage( parent.parent.parent.parent.handle, WM_LBUTTONDBLCLK, 0, 0 );
end;

end.

