{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTGlossySpecularBRDFCl;

{$MODE Delphi}

interface

uses
  RTBRDFCl, RTMiscTp, RTSamplerCl;

type
  TRTGlossySpecularBRDF = class( TRTBRDF )
  private
      fKs: Single;
      fCs: TRGBColor;
      fExp: Single;
      fSampler: TRTSampler;

  public
      constructor Create; overload;
      constructor Create( const aNumSamples: Integer; const aSamplingExponent: Single ); overload;
      destructor Destroy; override;

      function F( const aShadeRec: TRTShadeRec; const aWI, aWO: TRTVector3D ): TRGBColor; override;
      function SampleF( const aShadeRec: TRTShadeRec; const aWO: TRTVector3D; out aWI: TRTVector3D ): TRGBColor; override;

      property SpecularReflection: Single read fKs write fKs;
      property SpecularColor: TRGBColor read fCs write fCs;
      property SpecularExponent: Single read fExp write fExp;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  Math, SysUtils,
  RTMathFn, RTMiscCn, RTMultiJitteredSamplerCl;

constructor TRTGlossySpecularBRDF.Create( const aNumSamples: Integer; const aSamplingExponent: Single );
begin
  inherited Create;

  fKs := 0.2;
  fCs := COLOR_WHITE;
  fExp := 1;

  if aNumSamples > 0 then
  begin
      fSampler := TRTMultiJitteredSampler.Create( aNumSamples );
      fSampler.MapSamplesToHemisphere( aSamplingExponent );
  end;
end;

// ----------------------------------------------------------------------------

constructor TRTGlossySpecularBRDF.Create;
begin
  Create( 0, 0 );
end;

// ----------------------------------------------------------------------------

destructor TRTGlossySpecularBRDF.Destroy;
begin
  FreeAndNil( fSampler );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

function TRTGlossySpecularBRDF.F( const aShadeRec: TRTShadeRec; const aWI, aWO: TRTVector3D ): TRGBColor;
var
  nDotWi, rDotWo: Single;
  r: TRTVector3D;

begin
  nDotWi := VectorDotProduct( aShadeRec.Normal, aWI );
  r := VectorAdd( VectorMultiply( aWi, -1 ), VectorMultiply( aShadeRec.Normal, 2 * nDotWi ));

  rDotWo := VectorDotProduct( r, aWO );

  if rDotWo > 0 then
  begin
      result := ColorMultiply( fCs, fKs * Power( rDotWo, fExp ));
  end
  else
      result := COLOR_BLACK;
end;

// ----------------------------------------------------------------------------

function TRTGlossySpecularBRDF.SampleF( const aShadeRec: TRTShadeRec; const aWO: TRTVector3D; out aWI: TRTVector3D ): TRGBColor; 
var
  phongLobe: Single;
  nDotWo: Single;
  u, v, w: TRTVector3D;
  sp: TRTPoint3D;

begin
  nDotWo := VectorDotProduct( aShadeRec.Normal, aWO );

  // INFO, PJ: Direction of mirror reflection (and orthonormal base around it
  w := VectorSubtract( VectorMultiply( aShadeRec.Normal, 2 * nDotWo ), aWO );
  u := VectorCrossProduct( Vector( 0.00424, 1, 0.00654 ), w );
  VectorNormalize( u );
  v := VectorCrossProduct( u, w );

  sp := fSampler.NextSampleHemisphere;

  aWI := VectorAdd(
      VectorMultiply( u, sp.X ),
      VectorMultiply( v, sp.Y ),
      VectorMultiply( w, sp.Z ));

  // INFO, PJ: Reflected ray is below surface, mirror
  if VectorDotProduct( aShadeRec.Normal, aWI ) < 0 then
  begin
      aWI := VectorAdd(
          VectorMultiply( u, -sp.X ),
          VectorMultiply( v, -sp.Y ),
          VectorMultiply( w, sp.Z ));
  end;

  phongLobe := Power( VectorDotProduct( w, aWI ), fExp );
  phongLobe := phongLobe * VectorDotProduct( aShadeRec.Normal, aWi );
  result := ColorMultiply( fCs, fKs * phongLobe );
end;

// ----------------------------------------------------------------------------

end.
