{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTCameraCl;

{$MODE Delphi}

interface

uses
  RTWorldCl, RTMiscTp;

type
  TRTCamera = class( TObject )
  private
      fEye: TRTPoint3D;
      fLookAt: TRTPoint3D;
      fUp: TRTVector3D;

      fRoll: Single;
      fPitch: Single;
      fYaw: Single;

      fExposureTime: Single;

      fFrustumOffset: Single;
      fPixelOffset: Integer;

      // INFO, PJ: Local Axes of orthonormal base
      fU: TRTVector3D;
      fV: TRTVector3D;
      fW: TRTVector3D;

      fShouldStop: Boolean;
      fRenderPosition: Integer;
      fIsRunning: Boolean;

      procedure SetViewDirection( const aDirection: TRTVector3D );

  protected
      procedure ComputeUVW;
      procedure RenderScene( const aWorld: TRTWorld ); virtual; abstract;

      property U: TRTVector3D read fU;
      property V: TRTVector3D read fV;
      property W: TRTVector3D read fW;

  public
      constructor Create;

      procedure Render( const aWorld: TRTWorld );
      procedure Move( const anX, aY, aZ: Double );

      property Eye: TRTPoint3D read fEye write fEye;
      property LookAt: TRTPoint3D read fLookAt write fLookAt;
      property Up: TRTVector3D read fUp write fUp;

      property Roll: Single read fRoll write fRoll;
      property Yaw: Single read fYaw write fYaw;
      property Pitch: Single read fPitch write fPitch;

      property ExposureTime: Single read fExposureTime write fExposureTime;

      // INFO, PJ: Alternative ways to set the above properties.
      property ViewDirection: TRTVector3D write SetViewDirection;

      // INFO, PJ: Properties for optional stereoscopic viewing (both horizontal offsets)
      property FrustumOffset: Single read fFrustumOffset write fFrustumOffset;
      property PixelOffset: Integer read fPixelOffset write fPixelOffset;

      // INFO, PJ: Properties for managing start/stop/position
      property ShouldStop: Boolean read fShouldStop write fShouldStop;
      property RenderPosition: Integer read fRenderPosition write fRenderPosition;
      property IsRunning: Boolean read fIsRunning;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  RTMathFn;

constructor TRTCamera.Create;
begin
  inherited Create;

  fUp := Vector( 0, 1, 0 );
  fExposureTime := 1;
end;

// ----------------------------------------------------------------------------

procedure TRTCamera.SetViewDirection( const aDirection: TRTVector3D );
begin
  // INFO, PJ: For convenience only. For a more practical approach, it should continually adapt
  //           the fLookAt when the
  fLookAt := VectorAdd( fEye, aDirection );
end;

// ----------------------------------------------------------------------------

procedure TRTCamera.Render( const aWorld: TRTWorld );
begin
    fIsRunning := TRUE;

    try
        RenderScene( aWorld );
        
    finally
        fIsRunning := FALSE;
    end;
end;

// ----------------------------------------------------------------------------

procedure TRTCamera.Move( const anX, aY, aZ: Double );
begin
  fEye.X := fEye.X + anX;
  fEye.Y := fEye.Y + aY;
  fEye.Z := fEye.Z + aZ;
end;

// ----------------------------------------------------------------------------

procedure TRTCamera.ComputeUVW;
var
  matrix: TRTMatrix;

begin
  // INFO, PJ: Handle singularities
  if (fEye.X = fLookAt.X) and (fEye.Z = fLookat.Z) then
  begin
      if fEye.Y > fLookAt.Y then
      begin // INFO, PJ: Camera looking vertically down
          fU := Vector( 0, 0, 1 );
          fV := Vector( 1, 0, 0 );
          fW := Vector( 0, 1, 0 );
      end
      else
      begin // INFO, PJ: Camera looking vertically up
		    fU := Vector( 1, 0, 0 );
		    fV := Vector( 0, 0, 1 );
		    fW := Vector( 0, -1, 0 );
      end;
  end
  else
  begin
      // INFO, PJ: Construct Orthonormal Base.
      fW := VectorSubtract( fEye, fLookAt );
      VectorNormalize( fW );

      fU := VectorCrossProduct( fUp, fW );
      VectorNormalize( fU );

      fV := VectorCrossProduct( fW, fU );
  end;

  if fYaw <> 0 then
  begin
      matrix := MatrixForAxisRotation( fV, fYaw );
      fU := MatrixMultiplyVector( matrix, fU );
      fW := MatrixMultiplyVector( matrix, fW );
  end;

  if fPitch <> 0 then
  begin
      matrix := MatrixForAxisRotation( fU, fPitch );
      fV := MatrixMultiplyVector( matrix, fV );
      fW := MatrixMultiplyVector( matrix, fW );
  end;

  // INFO, PJ: We apply yaw, pitch and roll
  if fRoll <> 0 then
  begin
      matrix := MatrixForAxisRotation( fW, fRoll );
      fU := MatrixMultiplyVector( matrix, fU );
      fV := MatrixMultiplyVector( matrix, fV );
  end;
end;

end.
