{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTPinholeCameraCl;

{$MODE Delphi}

interface

uses
  RTCameraCl, RTWorldCl, RTMiscTp, RTViewPlaneCl;

type
  TRTPinholeCamera = class( TRTCamera )
  private
      fD: Single;     // INFO, PJ: View-plane distance
      fZoom: Single;  // INFO, PJ: Zoom factor

      fFOV: Single;   // INFO, PJ: Field of view. Replaces fD if used. 

      function DirectionForPoint( const aPoint: TRTPoint2D ): TRTVector3D;

      procedure ComputeViewingDistance( const aViewPlane: TRTViewPlane );
            
      property Zoom: Single read fZoom;
      property D: Single read fD;

  public
      constructor Create;
      procedure RenderScene( const aWorld: TRTWorld ); override;

      property ViewDistance: Single write fD;

      // INFO, PJ: Alternative ways to set the above properties.
      property FOV: Single write fFOV; // INFO, PJ: FOV in degrees.
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  Forms, Math, SysUtils,
  RTMiscCn, RTMathFn;

constructor TRTPinholeCamera.Create;
begin
  inherited Create;

  fZoom := 1;
end;

// ----------------------------------------------------------------------------

procedure TRTPinholeCamera.ComputeViewingDistance( const aViewPlane: TRTViewPlane );
begin
  // INFO, PJ: If a FOV is provided instead of a D, we calculate D from FOV and the ViewPlane.
  //           We use OpenGL gluPerspective compatible FOV-Y
  if ( fD = 0 ) and ( fFOV <> 0 ) then
  begin
      fD := aViewPlane.VRes / ( 2 * Tan( DegToRad( fFOV ) / 2 ));
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTPinholeCamera.RenderScene( const aWorld: TRTWorld );
var
  l: TRGBColor;
  ray: TRTRay;
  vp: TRTViewPlane;
  pixelSize: Single;
  r, c, j, depth: Integer;
  sp, pp: TRTPoint2D;
  startPosition: Integer;
  startRow, startColumn: Integer;

begin
  if not Assigned( aWorld.Tracer ) then
  begin
      raise Exception.Create( 'No tracer is assigned.' );
  end;

  ComputeViewingDistance( aWorld.ViewPlane );

  ComputeUVW;

  vp := aWorld.ViewPlane;
  aWorld.OpenWindow( vp.HRes, vp.VRes );

  pixelSize := vp.PixelSize / Zoom;
  ray.Origin := Eye;
  depth := 0;

  startPosition := RenderPosition;
  startRow := vp.VRes - 1 - (startPosition div vp.HRes);
  startColumn := startPosition mod vp.HRes;

  // INFO, PJ: Keep-this-side-up fix for reversed UVW
  for r := startRow downto 0 do
  begin
      for c := startColumn to vp.HRes - 1 do
      begin
          l := COLOR_BLACK;

          for j := 0 to vp.NumSamples - 1 do
          begin
              sp := vp.Sampler.NextSampleUnitSquare;
              pp.X := pixelSize * ( c - 0.5 * vp.HRes + sp.X ) + FrustumOffset;
              pp.Y := pixelSize * ( r - 0.5 * vp.VRes + sp.Y );
              ray.Direction := DirectionForPoint( pp );

              l := ColorAdd( l, aWorld.Tracer.TraceRay( ray, depth ));
          end;

          l := ColorDivide( l, vp.NumSamples );
          aWorld.DisplayPixel( r, c + PixelOffset, l );

          inc( startPosition );
          if ShouldStop then break;
      end;

      startColumn := 0;

      if ShouldStop then break;
  end;

  RenderPosition := startPosition;
end;

// ----------------------------------------------------------------------------

function TRTPinholeCamera.DirectionForPoint( const aPoint: TRTPoint2D ): TRTVector3D;
begin
  result := VectorAdd(
      VectorMultiply( U, aPoint.X ),
      VectorMultiply( V, aPoint.Y ),
      VectorMultiply( W, -D ));

  VectorNormalize( result );
end;


end.
