{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTSphericalCameraCl;

{$MODE Delphi}

interface

uses
  RTCameraCl, RTWorldCl, RTMiscTp;

type
  TRTSphericalCamera = class( TRTCamera )
  private
      fHFOV: Single;
      fVFOV: Single;

      function MakeDirection( const aPoint: TRTPoint2D; const aHRes, aVRes: Integer; const aPixelSize: single ): TRTVector3D;
      procedure ComputeFOV( const aHRes, aVRes: Integer );

  public
      constructor Create;
      procedure RenderScene( const aWorld: TRTWorld ); override;

      property HorizontalFOV: Single read fHFOV write fHFOV;
      property VerticalFOV: Single read fVFOV write fVFOV;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  Math,
  RTViewPlaneCl, RTMiscCn, RTMathFn;

constructor TRTSphericalCamera.Create;
begin
  inherited Create;

  fHFOV := 180;
end;

// ----------------------------------------------------------------------------

procedure TRTSphericalCamera.RenderScene( const aWorld: TRTWorld );
var
  vp: TRTViewPlane;
  ray: TRTRay;
  r, c, j: Integer;
  l: TRGBColor;
  sp, pp: TRTPoint2D;
  
begin
  ComputeUVW;
  ComputeFOV( aWorld.ViewPlane.HRes, aWorld.ViewPlane.VRes );

  vp := aWorld.ViewPlane;
  aWorld.OpenWindow( vp.HRes, vp.VRes );

  ray.Origin := eye;

  for r := 0 to vp.VRes - 1 do
  begin
      for c := 0 to vp.HRes - 1 do
      begin
          l := COLOR_BLACK;

          for j := 0 to vp.NumSamples - 1 do
          begin
              sp := vp.Sampler.NextSampleUnitSquare;
              pp.X := vp.PixelSize * ( c - 0.5 * vp.HRes + sp.X ) + FrustumOffset;
              pp.Y := vp.PixelSize * ( r - 0.5 * vp.VRes + sp.Y );
              ray.Direction := MakeDirection( pp, vp.HRes, vp.VRes, vp.PixelSize );

              l := ColorAdd( l, aWorld.Tracer.TraceRay( ray ));
          end;

          l := ColorDivide( l, vp.NumSamples );
          l := ColorMultiply( l, ExposureTime );
          aWorld.DisplayPixel( r, c + PixelOffset, l );
      end;
  end;

end;

// ----------------------------------------------------------------------------

procedure TRTSphericalCamera.ComputeFOV( const aHRes, aVRes: Integer );
begin
  // INFO, PJ: If either Horizontal or Vertical FOV is not specified, use the resolutions to calculate it.
  if fHFOV = 0 then
  begin
      fHFOV := fVFOV * aHRes / aVRes;  
  end;

  if fVFOV = 0 then
  begin
      fVFOV := fHFOV * aVRes / aHRes;
  end;
end;

// ----------------------------------------------------------------------------

function TRTSphericalCamera.MakeDirection( const aPoint: TRTPoint2D; const aHRes, aVRes: Integer; const aPixelSize: single ): TRTVector3D;
var
  pn: TRTPoint2D;
  lambda, psi: Single;
  phi, theta: Single;
  sinPhi, cosPhi, sinTheta, cosTheta: Single;

begin
  // INFO, PJ: Normalized device coordinates (should HRes be leading, combined with with an aspect ratio setting?)
  pn := Point( aPoint.X * 2 / ( aPixelSize * aHRes ), aPoint.Y * 2 / ( aPixelSize * aVRes ));

  // INFO, PJ: Angles lamda and phi in radians
  lambda := pn.X * DegToRad( fHFOV / 2 );
  psi := pn.Y * DegToRad( fVFOV / 2 );

  // INFO, PJ: Spherical azimuth and polar angle
  phi := PI - lambda;
  theta := 0.5 * PI - psi;

  sinPhi := Sin( phi );
  cosPhi := Cos( phi );
  sinTheta := Sin( theta );
  cosTheta := Cos( theta );

  result := VectorAdd(
      VectorMultiply( U, sinTheta * sinPhi ),
      VectorMultiply( V, cosTheta ),
      VectorMultiply( W, sinTheta * cosPhi ));
end;

end.

