{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTThinLensCameraCl;

{$MODE Delphi}

interface

uses
  RTCameraCl, RTWorldCl, RTSamplerCl, RTMiscTp;

type
  TRTThinLensCamera = class( TRTCamera )
  private
      fLensRadius: Single;
      fD: Single;             // INFO, PJ: View-plane distance
      fF: Single;             // INFO, PJ: Focal plane distance
      fZoom: Single;          // INFO, PJ: Zoom factor
      fSampler: TRTSampler;

      procedure SetSampler( const aSampler: TRTSampler );
      function MakeDirection( const aPixelPoint, aLensPoint: TRTPoint2D ): TRTVector3D;

      property D: Single read fD;
      property F: Single read fF;

  public
      constructor Create;
      destructor Destroy; override;

      procedure RenderScene( const aWorld: TRTWorld ); override;

      property ViewDistance: Single read fD write fD;
      property FocalDistance: Single read fF write fF;
      property LensRadius: Single read fLensRadius write fLensRadius;
      property Sampler: TRTSampler read fSampler write SetSampler;
      property Zoom: Single read fZoom write fZoom;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils,
  RTViewPlaneCl, RTMiscCn, RTMathFn;

constructor TRTThinLensCamera.Create;
begin
  inherited Create;

  fZoom := 1;
end;

// ----------------------------------------------------------------------------

destructor TRTThinLensCamera.Destroy;
begin
  FreeAndNil( fSampler );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure TRTThinLensCamera.SetSampler( const aSampler: TRTSampler );
begin
  FreeAndNil( fSampler );

  fSampler := aSampler;
  fsampler.MapSamplesToUnitDisk;
end;

// ----------------------------------------------------------------------------

function TRTThinLensCamera.MakeDirection( const aPixelPoint, aLensPoint: TRTPoint2D ): TRTVector3D;
var
  p: TRTPoint2D;  // INFO, PJ: Hit point on focal plane

begin
  p.X := aPixelPoint.X * f / d;
  p.Y := aPixelPoint.Y * f / d;

  result := VectorAdd(
      VectorMultiply( U, p.X - aLensPoint.X ),
      VectorMultiply( V, p.Y - aLensPoint.Y ),
      VectorMultiply( W, -F ));

  VectorNormalize( result );
end;

// ----------------------------------------------------------------------------

procedure TRTThinLensCamera.RenderScene( const aWorld: TRTWorld );
var
  vp: TRTViewPlane;
  pixelSize: Single;
  r, c, n: Integer;
  l: TRGBColor;
  ray: TRTRay;
  sp: TRTPoint2D;     // INFO, PJ: Sample point (unit square)
  dp: TRTPoint2D;     // INFO, PJ: Sample point (unit disk)
  pp: TRTPoint2D;     // INFO, PJ: Sample point on pixel
  lp: TRTPoint2D;     // INFO, PJ: Sample point on lens

begin
  ComputeUVW;

  vp := aWorld.ViewPlane;
  aWorld.OpenWindow( vp.HRes, vp.VRes );
  pixelSize := vp.PixelSize / zoom;
  for r := 0 to vp.VRes - 1 do
  begin
      for c := 0 to vp.HRes - 1 do
      begin
          l := COLOR_BLACK;

          for n := 0 to vp.NumSamples - 1 do
          begin
              sp := vp.Sampler.NextSampleUnitSquare;
              pp.X := pixelSize * ( c - 0.5 * vp.HRes + sp.X ) + FrustumOffset;
              pp.Y := pixelSize * ( r - 0.5 * vp.VRes + sp.Y );

              dp := Sampler.NextSampleUnitDisk;
              lp  := PointMultiply( dp, LensRadius );

              ray.Origin := VectorAdd( eye,
                  VectorMultiply( U, lp.X ),
                  VectorMultiply( V, lp.Y ));

              ray.Direction := MakeDirection( pp, lp );
              l := ColorAdd( l, aWorld.Tracer.TraceRay( ray ));
          end;

          l := ColorDivide( l, vp.NumSamples );
          l := ColorMultiply( l, ExposureTime );
          aWorld.DisplayPixel( r, c + PixelOffset, l );
      end;
  end;
end;

end.

