{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTBVHCl;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  Generic class for constructing a Boundin Volume Hierarchy, useful for
  raytracing, lightmapping and whatnot.

  Based on Ingo Wald's "On fast Construction of SAH-based Bounding Volume
  Hierarchies". See

  http://www.sci.utah.edu/~wald/Publications/2007/FastBuild/download/fastbuild.pdf

  Implementation phases:
  [.] Naieve Implementation
  [ ] Optimization for memory allocation
  [ ] SIMD Optimization (only if present)
  [ ] Parallelization

  Misc
  [ ] Move to a generic implementation that works with generic data as opposed
      to Geometric Objects, so we can apply this structure to the lightmapper.
  ---------------------------------------------------------------------------- }
interface

uses
  Forms, Controls, ComCtrls,
  RTMiscTp, RTCompoundObjectCl, RTGeometricObjectCl;

type
  TRTBVHNode = class;
  TRTObjectInfo = record
      Centroid: TRTPoint3D;
      Min, Max: TRTPoint3D;
      GeometricObject: TRTGeometricObject;
  end;

// ----------------------------------------------------------------------------

  TRTBVH = class( TRTCompoundObject )
  private
      fPrimitives: Array of TRTObjectInfo;
      fRootNode: TRTBVHNode;

      fDebugForm: TForm;
      fDebugTree: TTreeView;

      procedure Resolve;
      procedure Initialize;
      procedure Build;

      function InfoForObject( const anObject: TRTGeometricObject ): TRTObjectInfo;
      procedure InitializeNode( const aNode: TRTBVHNode );
      procedure SplitNode( const aNode: TRTBVHNode );
      procedure SwapPrimitives( const anIndex1, anIndex2: Integer );
      function ConvertRay( const aRay: TRTRay ): TRTRay;

      procedure DrawAABB( const aMin, aMax: TRTVector3D );
      procedure DrawNodeTree( const aNode: TRTBVHNode; const aLeafsOnly: Boolean );
      procedure DrawNode( const aNode: TRTBVHNode );
      procedure AddTreeNodesForNode( const aNode: TRTBVHNode; const aParentNode: TTreeNode; const aNodes: TTreeNodes );
      function HitNode( const aNode: TRTBVHNode; const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
      function ShadowHitNode( const aNode: TRTBVHNode; const aRay: TRTRay; var aMinT: Double ): Boolean;
      function IntersectsNode( const aRay: TRTRay; const aNode: TRTBVHNode; const aT0, aT1: Single ): Boolean;

  public
      destructor Destroy; override;
      function Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean; override;
      function ShadowHit( const aRay: TRTRay; var aMinT: Double ): Boolean; override;

      // INFO, PJ: For debugging purposes
      procedure DrawPrimitives;
      procedure DrawNodes( const aLeafsOnly: Boolean );
      procedure DrawSelectedNode;
      procedure ShowDebugTree;
  end;

// ----------------------------------------------------------------------------

  TRTBVHNode = class
  private
      fStartIndex: Integer;
      fEndIndex: Integer;
      fLeft: TRTBVHNode;
      fRight: TRTBVHNode;
      fVoxelMin: TRTPoint3D;
      fVoxelMax: TRTPoint3D;
      fCentroidMin: TRTPoint3D;
      fCentroidMax: TRTPoint3D;

  public
      constructor Create( const aStartIndex, anEndIndex: Integer );
      destructor Destroy; override;

      function NumPrimitives: Integer;
      function IsLeaf: Boolean;

      property Left: TRTBVHNode read fLeft write fLeft;
      property Right: TRTBVHNode read fRight write fRight;

      property StartIndex: Integer read fStartIndex;
      property EndIndex: Integer read fEndIndex;
      property VoxelMin: TRTPoint3D read fVoxelMin write fVoxelMin;
      property VoxelMax: TRTPoint3D read fVoxelMax write fVoxelMax;
      property CentroidMin: TRTPoint3D read fCentroidMin write fCentroidMin;
      property CentroidMax: TRTPoint3D read fCentroidMax write fCentroidMax;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  dglOpenGL, Math, SysUtils,
  RTMathFn, RTMiscCn;

const
  NUM_BINS = 2;
  EPSILON = 0.0001;
  MIN_PRIMITIVES_PER_NODE = 4;


destructor TRTBVH.Destroy;
begin
  if Assigned( fDebugForm ) then
  begin
      fDebugForm.Release;
  end;

  FreeAndNil( fRootNode );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.Resolve;
begin
  if not Assigned( fPrimitives ) then
  begin
      Initialize;
      Build;
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.Initialize;
var
  i: Integer;

begin
  SetLength( fPrimitives, NumObjects );

  for i := 0 to NumObjects - 1 do
  begin
      fPrimitives[i] := InfoForObject( Objects[i] );
  end;

  fRootNode := TRTBVHNode.Create( Low( fPrimitives ), High( fPrimitives ));
  InitializeNode( fRootNode );
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.Build;
begin
  if (fRootNode.NumPrimitives > MIN_PRIMITIVES_PER_NODE) then
  begin
      try
          SplitNode( fRootNode );
      except
          on e: exception do
          begin
              fRootNode := nil;
          end;
      end;
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.InitializeNode( const aNode: TRTBVHNode );
var
  i: Integer;
  info: TRTObjectInfo;

begin
  for i := aNode.StartIndex to aNode.EndIndex do
  begin
      info := fPrimitives[i];

      aNode.VoxelMin := VectorMin( aNode.VoxelMin, info.Min );
      aNode.VoxelMax := VectorMax( aNode.VoxelMax, info.Max );
      aNode.CentroidMin := VectorMin( aNode.CentroidMin, info.Centroid );
      aNode.CentroidMax := VectorMax( aNode.CentroidMax, info.Centroid );
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.SwapPrimitives( const anIndex1, anIndex2: Integer );
var
  temp: TRTObjectInfo;

begin
  temp := fPrimitives[anIndex1];
  fPrimitives[anIndex1] := fPrimitives[anIndex2];
  fPrimitives[anIndex2] := temp;
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.SplitNode( const aNode: TRTBVHNode );
var
  axis: TRTAxis3D;
  k0, k1: Single;
  nodeDimensions: TRTVector3D;
  leftIndex, rightIndex: Integer;
  majorSize: Double;

begin
  // INFO, PJ: Select the dominant axis
  nodeDimensions := VectorSubtract( aNode.CentroidMax, aNode.CentroidMin );
  axis := VectorMaxAxis( nodeDimensions );

  majorSize := nodeDimensions.Elements[axis];

  if majorSize = 0 then
  begin
      // INFO, PJ:
      //   The node doesn't have dimensions, making it impossible to split.
      //   A probable cause is multiple polyons at the exact same position.
      exit;
  end;

  // INFO, PJ: For now (testing purposes), we simply split half-way (2-bin).

  // INFO, PJ:  We sort the triangles (left = bin 0, right = bin 1)
  k0 := aNode.CentroidMin.Elements[axis];
  k1 := NUM_BINS * ( 1 - EPSILON ) / ( majorSize );
  leftIndex := aNode.StartIndex;
  rightIndex := aNode.EndIndex;
  while leftIndex < rightIndex do
  begin
      // INFO, PJ: We apply <= and >= to prevent rounding errors from locking the algorithm.
      while Trunc( k1 * ( fPrimitives[leftIndex].Centroid.Elements[axis] - k0 )) <= 0 do inc( leftIndex );
      while Trunc( k1 * ( fPrimitives[rightIndex].Centroid.Elements[axis] - k0 )) >= 1 do dec( rightIndex );

      if rightIndex > leftIndex then
      begin
          SwapPrimitives( leftIndex, rightIndex );
      end;
  end;

  // INFO, PJ: Create left and right subnodes and recurse.
  if rightIndex >= aNode.StartIndex then
  begin
      aNode.Left := TRTBVHNode.Create( aNode.StartIndex, leftIndex - 1);
      InitializeNode( aNode.Left );

      if aNode.Left.NumPrimitives > MIN_PRIMITIVES_PER_NODE then
      begin
          SplitNode( aNode.Left );
      end;
  end;

  if leftIndex <= aNode.EndIndex then
  begin
      aNode.Right := TRTBVHNode.Create( rightIndex + 1, aNode.EndIndex );

      InitializeNode( aNode.Right );

      if aNode.Right.NumPrimitives > MIN_PRIMITIVES_PER_NODE then
      begin
          SplitNode( aNode.Right );
      end;
  end;
end;

// ----------------------------------------------------------------------------

function TRTBVH.InfoForObject( const anObject: TRTGeometricObject ): TRTObjectInfo;
begin
  result.Min := anObject.BoundingBox.Min;
  result.Max := anObject.BoundingBox.Max;

  result.Centroid := anObject.Centroid;
  result.GeometricObject := anObject;
end;

// ----------------------------------------------------------------------------

function TRTBVH.ConvertRay( const aRay: TRTRay ): TRTRay;
begin
  result := aRay;

  // INFO, PJ: Fill out optimization fields for result.
  result.invDirection := Vector( 1 / result.Direction.X, 1 / result.Direction.Y, 1 / result.Direction.Z );
  result.Signed[rtA3X] := ( result.invDirection.X < 0 );
  result.Signed[rtA3Y] := ( result.invDirection.Y < 0 );
  result.Signed[rtA3Z] := ( result.invDirection.Z < 0 );
end;

// ----------------------------------------------------------------------------

function TRTBVH.Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
begin
  Resolve;

  result := HitNode( fRootNode, convertRay( aRay ), aMinT, aShadeRec );
end;

// ----------------------------------------------------------------------------

function TRTBVH.ShadowHit( const aRay: TRTRay; var aMinT: Double ): Boolean; 
begin
  Resolve;

  result := ShadowHitNode( fRootNode, convertRay( aRay ), aMinT );
end;

// ----------------------------------------------------------------------------

function TRTBVH.HitNode( const aNode: TRTBVHNode; const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
var
  t: Double;
  normal: TRTNormal;
  localHitPoint: TRTPoint3D;
  i: Integer;
  geometricObject: TRTGeometricObject;
  uvCoordinate: TRTUVCoordinate;
  material: TObject;

begin
  result := FALSE;
  material := nil;

  // TODO, PJ: Find sensible number for t1... (rootnode.size * 2 or similar)
  if not IntersectsNode( aRay, aNode, 0, 10000000 ) then
  begin
      exit;
  end;

  aMinT := HUGE_VALUE;

  if aNode.IsLeaf then
  begin
      for i := aNode.StartIndex to aNode.EndIndex do
      begin
          geometricObject := fPrimitives[i].GeometricObject;

          // INFO, PJ: The "Store local copy of most but not all properties" construction is not too good.
          if geometricObject.Hit( aRay, t, aShadeRec ) and ( t < aMinT ) then
          begin
              result := TRUE;
              aMinT := t;
              normal := aShadeRec.Normal;
              material := aShadeRec.Material;
              localHitPoint := aShadeRec.LocalHitPoint;
              uvCoordinate := aShadeRec.UVCoordinate;
          end;
      end;
  end
  else
  begin
      if HitNode( aNode.Left, aRay, t, aShadeRec ) and ( t < aMinT ) then
      begin
          result := TRUE;
          aMinT := t;
          normal := aShadeRec.Normal;
          material := aShadeRec.Material;
          localHitPoint := aShadeRec.LocalHitPoint;
          uvCoordinate := aShadeRec.UVCoordinate;
      end;

      if HitNode( aNode.Right, aRay, t, aShadeRec ) and ( t < aMinT ) then
      begin
          result := TRUE;
          aMinT := t;
          normal := aShadeRec.Normal;
          material := aShadeRec.Material;
          localHitPoint := aShadeRec.LocalHitPoint;
          uvCoordinate := aShadeRec.UVCoordinate;
      end;
  end;

  if result then
  begin
      aShadeRec.T := aMinT;
      aShadeRec.Normal := normal;
      aShadeRec.Material := material;
      aShadeRec.LocalHitPoint := localHitPoint;
      aShadeRec.UVCoordinate := uvCoordinate;
  end;
end;

// ----------------------------------------------------------------------------

function TRTBVH.ShadowHitNode( const aNode: TRTBVHNode; const aRay: TRTRay; var aMinT: Double ): Boolean;
var
  i: Integer;
  geometricObject: TRTGeometricObject;

begin
  result := FALSE;

  // TODO, PJ: Find sensible number for t1... (rootnode.size * 2 or similar)
  if not IntersectsNode( aRay, aNode, 0, 10000000 ) then
  begin
      exit;
  end;

  if aNode.IsLeaf then
  begin
      for i := aNode.StartIndex to aNode.EndIndex do
      begin
          geometricObject := fPrimitives[i].GeometricObject;

          // INFO, PJ: The "Store local copy of most but not all properties" construction is not too good.
          if geometricObject.ShadowHit( aRay, aMinT ) then
          begin
              result := TRUE;
              exit;
          end;
      end;
  end
  else
  begin
      if  ShadowHitNode( aNode.Left, aRay, aMinT ) or
          ShadowHitNode( aNode.Right, aRay, aMinT ) then
      begin
          result := TRUE;
          exit;
      end;
  end;
end;

// ----------------------------------------------------------------------------

// INFO, PJ:
//   Based on "An Efficient and Robust Ray–Box Intersection Algorithm", by Williams, Barrus, Morley and Shirley
//   See http://www.cs.utah.edu/~awilliam/box/box.pdf
// NOTE, PJ: Probably more efficient than the cohen-sutherland algorithm as used in our Octree implementation.
function TRTBVH.IntersectsNode( const aRay: TRTRay; const aNode: TRTBVHNode; const aT0, aT1: Single ): Boolean;
var
  bounds: Array[Boolean] of TRTVector3D;
  tMin, tMax, tyMin, tyMax, tzMin, tzMax: Single;

begin
  bounds[FALSE] := aNode.VoxelMin;
  bounds[TRUE] := aNode.VoxelMax;

  tmin := ( bounds[aRay.Signed[rtA3X]].X - aRay.Origin.X ) * aRay.invDirection.X;
  tmax := ( bounds[not aRay.Signed[rtA3X]].X - aRay.Origin.X ) * aRay.InvDirection.X;

  tymin := ( bounds[aRay.Signed[rtA3Y]].Y - aRay.Origin.Y ) * aRay.InvDirection.Y;
  tymax := ( bounds[not aRay.Signed[rtA3Y]].Y - aRay.Origin.Y ) * aRay.InvDirection.Y;

  if (tMin > tyMax) or (tyMin > tMax ) then
  begin
      result := FALSE;
      exit;
  end;

  if tyMin > tMin then tMin := tyMin;
  if tyMax < tMax then tMax := tyMax;

  tzmin := ( bounds[aRay.Signed[rtA3Z]].Z - aRay.Origin.Z ) * aRay.invDirection.Z;
  tzmax := ( bounds[not aRay.Signed[rtA3Z]].Z - aRay.Origin.Z ) * aRay.invDirection.Z;
  if (tMin > tzMax ) or ( tzMin > tMax ) then
  begin
      result := FALSE;
      exit;
  end;

  result := ( tMin < aT1 ) and ( tMax > aT0 );
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.DrawPrimitives;
var
  i: Integer;
  info: TRTObjectInfo;

begin
  glColor3f( 1, 1, 0 );

  glDisable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);
  glDisable(GL_CULL_FACE);

  for i := Low( fPrimitives ) to High( fPrimitives ) do
  begin
      info := fPrimitives[i];

      DrawAABB( info.Min, info.Max );
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.DrawAABB( const aMin, aMax: TRTVector3D );
const
  EPSILON = 1;

begin
  glBegin(GL_LINE_LOOP);

  glVertex3f( aMin.X - EPSILON, aMin.Y - EPSILON, aMin.Z - EPSILON );
  glVertex3f( aMin.X - EPSILON, aMax.Y + EPSILON, aMin.Z - EPSILON );
  glVertex3f( aMax.X + EPSILON, aMax.Y + EPSILON, aMin.Z - EPSILON );
  glVertex3f( aMax.X + EPSILON, aMin.Y - EPSILON, aMin.Z - EPSILON );

  glVertex3f( aMin.X - EPSILON, aMin.Y - EPSILON, aMin.Z - EPSILON );
  glVertex3f( aMin.X - EPSILON, aMin.Y - EPSILON, aMax.Z + EPSILON );
  glVertex3f( aMax.X + EPSILON, aMin.Y - EPSILON, aMax.Z + EPSILON );
  glVertex3f( aMax.X + EPSILON, aMin.Y - EPSILON, aMin.Z - EPSILON );

  glVertex3f( aMin.X - EPSILON, aMin.Y - EPSILON, aMin.Z - EPSILON );
  glVertex3f( aMin.X - EPSILON, aMax.Y + EPSILON, aMin.Z - EPSILON );
  glVertex3f( aMin.X - EPSILON, aMax.Y + EPSILON, aMax.Z + EPSILON );
  glVertex3f( aMin.X - EPSILON, aMin.Y - EPSILON, aMax.Z + EPSILON );

  glEnd;

  glBegin(GL_LINE_LOOP);

  glVertex3f( aMax.X + EPSILON, aMax.Y + EPSILON, aMax.Z + EPSILON );
  glVertex3f( aMax.X + EPSILON, aMin.Y - EPSILON, aMax.Z + EPSILON );
  glVertex3f( aMin.X - EPSILON, aMin.Y - EPSILON, aMax.Z + EPSILON );
  glVertex3f( aMin.X - EPSILON, aMax.Y + EPSILON, aMax.Z + EPSILON );

  glVertex3f( aMax.X + EPSILON, aMax.Y + EPSILON, aMax.Z + EPSILON );
  glVertex3f( aMax.X + EPSILON, aMax.Y + EPSILON, aMin.Z - EPSILON );
  glVertex3f( aMin.X - EPSILON, aMax.Y + EPSILON, aMin.Z - EPSILON );
  glVertex3f( aMin.X - EPSILON, aMax.Y + EPSILON, aMax.Z + EPSILON );

  glVertex3f( aMax.X + EPSILON, aMax.Y + EPSILON, aMax.Z + EPSILON );
  glVertex3f( aMax.X + EPSILON, aMin.Y - EPSILON, aMax.Z + EPSILON );
  glVertex3f( aMax.X + EPSILON, aMin.Y - EPSILON, aMin.Z - EPSILON );
  glVertex3f( aMax.X + EPSILON, aMax.Y + EPSILON, aMin.Z - EPSILON );

  glEnd;
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.DrawNodes( const aLeafsOnly: Boolean );
begin
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_LIGHTING);
  glEnable(GL_DEPTH_TEST);
  glDisable(GL_CULL_FACE);

  DrawNodeTree( fRootNode, aLeafsOnly );
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.DrawNodeTree( const aNode: TRTBVHNode; const aLeafsOnly: Boolean );
begin
  if (not aLeafsOnly) or aNode.IsLeaf then
  begin
      DrawNode( aNode );
  end;

  if aNode.Left <> nil then DrawNodeTree( aNode.Left, aLeafsOnly );
  if aNode.Right <> nil then DrawNodeTree( aNode.Right, aLeafsOnly );
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.DrawNode( const aNode: TRTBVHNode );
var
  info: TRTObjectInfo;
  i: Integer;

begin
  glColor3f( 1, 1, 0 );
  DrawAABB( aNode.VoxelMin, aNode.VoxelMax );

  glColor3f( 0, 0, 1 );
  DrawAABB( aNode.CentroidMin, aNode.CentroidMax );

  glColor3f( 0, 0, 0 );
  for i := aNode.StartIndex to aNode.EndIndex do
  begin
      info := fPrimitives[i];

      DrawAABB( info.Min, info.Max );
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.DrawSelectedNode;
var
  node: TRTBVHNode;
  
begin
  if Assigned( fDebugTree ) and Assigned( fDebugTree.Selected ) then
  begin
      node := TRTBVHNode( fDebugTree.Selected.Data );

      DrawNode( node );
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.ShowDebugTree;
var
  parentNode: TTreeNode;
  
begin
  if fDebugForm = nil then
  begin
      fDebugForm := TForm.Create( nil );
      fDebugForm.Position := poMainFormCenter;
      fDebugForm.Width := 256;
      fDebugForm.Height := 512;
      fDebugForm.Caption := 'BVH Debug Tree';
      fDebugForm.FormStyle := fsStayOnTop;

      fDebugTree := TTreeView.Create( fDebugForm );
      fDebugTree.Parent := fDebugForm;
      fDebugTree.Align := alClient;

      parentNode := fDebugTree.Items.AddChildObject( nil, 'root', fRootNode );
      AddTreeNodesForNode( fRootNode, parentNode, fDebugTree.Items );

      fDebugTree.FullExpand;
  end;

  if not fDebugForm.Visible then
  begin
      fDebugForm.Show;
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTBVH.AddTreeNodesForNode( const aNode: TRTBVHNode; const aParentNode: TTreeNode; const aNodes: TTreeNodes );
var
  parentNode: TTreeNode;

begin
  if Assigned( aNode.Left ) then
  begin
      parentNode := aNodes.AddChildObject( aParentNode, 'left', aNode.Left );

      AddTreeNodesForNode( aNode.Left, parentNode, aNodes );
  end;

  if Assigned( aNode.Right ) then
  begin
      parentNode := aNodes.AddChildObject( aParentNode, 'right', aNode.Right );

      AddTreeNodesForNode( aNode.Right, parentNode, aNodes );
  end;
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TRTBVHNode.Create( const aStartIndex, anEndIndex: Integer );
begin
  inherited Create;

  fStartIndex := aStartIndex;
  fEndIndex := anEndIndex;

  fVoxelMin := Vector( Infinity );
  fVoxelMax := Vector( -Infinity );
  fCentroidMin := Vector( Infinity );
  fCentroidMax := Vector( -Infinity );
end;

// ----------------------------------------------------------------------------

destructor TRTBVHNode.Destroy;
begin
  FreeAndNil( fLeft );
  FreeAndNil( fRight );
  
  inherited Destroy;
end;

// ----------------------------------------------------------------------------

function TRTBVHNode.NumPrimitives: Integer;
begin
  result := fEndIndex - fStartIndex + 1;

end;

  // ----------------------------------------------------------------------------

function TRTBVHNode.IsLeaf: Boolean;
begin
  result := ( fLeft = nil ) and ( fRight = nil );
end;

end.
