{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTBoxCl;

{$MODE Delphi}

interface

uses
  RTGeometricObjectCl, RTMiscTp, RTAABoundingBoxCl;

type
  TRTBox = class( TRTGeometricObject )
  private
      fX0: Double;
      fY0: Double;
      fZ0: Double;

      fX1: Double;
      fY1: Double;
      fZ1: Double;

      function MakeNormal( const anIndex: Integer ): TRTNormal;

  protected
      function MakeBoundingBox: TRTAABoundingBox; override;

  public
      constructor Create( const aMinPoint, aMaxPoint: TRTPoint3D ); 

      function Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean; override;

      property X0: Double read fX0;
      property Y0: Double read fY0;
      property Z0: Double read fZ0;

      property X1: Double read fX1;
      property Y1: Double read fY1;
      property Z1: Double read fZ1;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  Math, SysUtils,
  RTMathFn, RTMiscCn;

const
  EPSILON = 0.001;

constructor TRTBox.Create( const aMinPoint, aMaxPoint: TRTPoint3D );
begin
  inherited Create;

  fX0 := aMinPoint.X;
  fY0 := aMinPoint.Y;
  fZ0 := aMinPoint.Z;

  fX1 := aMaxPoint.X;
  fY1 := aMaxPoint.Y;
  fZ1 := aMaxPoint.Z;
end;

// ----------------------------------------------------------------------------

function TRTBox.MakeBoundingBox: TRTAABoundingBox;
begin
  result := TRTAABoundingBox.Create( Vector( fX0, fY0, fZ0 ), Vector( fX1, fY1, fZ1 ));
end;

// ----------------------------------------------------------------------------

function TRTBox.Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
var
  ox, oy, oz: Double;
  dx, dy, dz: Double;
  a, b, c: Double;
  txMin, txMax, tyMin, tyMax, tzMin, tzMax, t0, t1: Double;
  inFace, outFace: integer;

begin
  ox := aRay.Origin.X; oy := aRay.Origin.Y; oz := aRay.Origin.Z;
  dx := aRay.Direction.X; dy := aRay.Direction.Y; dz := aRay.Direction.Z;

  if dx = 0 then
  begin
      a := HUGE_VALUE;
  end
  else
      a := 1 / dx;

  if a > 0 then
  begin
      txMin := ( X0 - ox ) * a;
      txMax := ( X1 - ox ) * a;
  end
  else
  begin
      txMin := ( X1 - ox ) * a;
      txMax := ( X0 - ox ) * a;
  end;

  if dy = 0 then
  begin
      b := HUGE_VALUE;
  end
  else
      b := 1 / dy;

  if b > 0 then
  begin
      tyMin := ( Y0 - oy ) * b;
      tyMax := ( Y1 - oy ) * b;
  end
  else
  begin
      tyMin := ( Y1 - oy ) * b;
      tyMax := ( Y0 - oy ) * b;
  end;

  if dz = 0 then
  begin
      c := HUGE_VALUE;
  end
  else
      c := 1 / dz;
      
  if c > 0 then
  begin
      tzMin := ( Z0 - oz ) * c;
      tzMax := ( Z1 - oz ) * c;
  end
  else
  begin
      tzMin := ( Z1 - oz ) * c;
      tzMax := ( Z0 - oz ) * c;
  end;

  if txMin > tyMin then
  begin
      t0 := txMin;
      if a >= 0 then inFace := 0 else inFace := 3;
  end
  else
  begin
      t0 := tyMin;
      if b >= 0 then inFace := 1 else inFace := 4;
  end;

  if tzMin > t0 then
  begin
      t0 := tzMin;
      if c >= 0 then inFace := 2 else inFace := 5;
  end;

  if txMax < tyMax then
  begin
      t1 := txMax;
      if a>= 0 then outFace := 3 else outFace := 0;
  end
  else
  begin
      t1 := tyMax;
      if b >= 0 then outFace := 4 else outFace := 0;
  end;

  if tzMax < t1 then
  begin
      t1 := tzMax;
      if c >= 0 then outFace := 5 else outFace  := 2;
  end;

  if ( t0 < t1 ) and ( t1 > EPSILON ) then
  begin
      if t0 > EPSILON then
      begin
          aMinT := t0;
          aShadeRec.Normal := MakeNormal( inFace );
      end
      else
      begin
          aMinT := t1;
          aShadeRec.Normal := MakeNormal( outFace );
      end;

      aShadeRec.LocalHitPoint := VectorAdd( aRay.Origin, VectorMultiply( aRay.Direction, aMinT ));
      result := TRUE;
  end
  else
      result := FALSE;
end;

// ----------------------------------------------------------------------------

function TRTBox.MakeNormal( const anIndex: Integer ): TRTNormal;
begin
  case anIndex of
      0: result := Vector( -1, 0, 0 );
      1: result := Vector( 0, -1, 0 );
      2: result := Vector( 0, 0, -1 );
      3: result := Vector( 1, 0, 0 );
      4: result := Vector( 0, 1, 0 );
      5: result := Vector( 0, 0, 1 );
  else
      raise Exception.Create( Format( 'Invalid face index %d', [anIndex] ));
  end;
end;

end.

