{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTCompoundObjectCl;

{$MODE Delphi}

interface

uses
  Contnrs,
  RTGeometricObjectCl, RTMiscTp, RTMaterialCl, RTAABoundingBoxCl;

type
  TRTCompoundObject = class( TRTGeometricObject )
  private
      fObjects: TObjectList;

  protected
      function GetObject( const anIndex: Integer ): TRTGeometricObject; 
      procedure SetMaterial( const aMaterial: TRTMaterial ); override;
      function MakeBoundingBox: TRTAABoundingBox; override;

  public
      constructor Create( const anOwnsObjects: Boolean = TRUE );
      destructor Destroy; override;

      procedure AddObject( const anObject: TRTGeometricObject );
      function Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean; override;
      function ShadowHit( const aRay: TRTRay; var aMinT: Double ): Boolean; override;

      // INFO, PJ: Interal properties, only use for debugging purposes.
      property Objects[const anIndex: Integer]: TRTGeometricObject read GetObject;
      function NumObjects: Integer;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils,
  RTMiscCn, RTMathFn;

const
  EPSILON = 0.00001;

constructor TRTCompoundObject.Create( const anOwnsObjects: Boolean = TRUE );
begin
  inherited Create;

  fObjects := TObjectList.Create( anOwnsObjects );
end;

// ----------------------------------------------------------------------------

destructor TRTCompoundObject.Destroy;
begin
  FreeAndNil( fObjects );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

function TRTCompoundObject.GetObject( const anIndex: Integer ): TRTGeometricObject;
begin
  result := fObjects[anIndex] as TRTGeometricObject;
end;

// ----------------------------------------------------------------------------

procedure TRTCompoundObject.SetMaterial( const aMaterial: TRTMaterial );
var
  i: Integer;
  geometricObject: TRTGeometricObject;

begin
  inherited SetMaterial( aMaterial );

  for i := 0 to fObjects.Count - 1 do
  begin
      geometricObject := fObjects[i] as TRTGeometricObject;

      geometricObject.Material := aMaterial;
      geometricObject.OwnsMaterial := FALSE;
  end;
end;

// ----------------------------------------------------------------------------

function TRTCompoundObject.MakeBoundingBox: TRTAABoundingBox;
var
  i: Integer;
  min, max: TRTPoint3D;
  box: TRTAABoundingBox;

begin
  min := Vector( HUGE_VALUE );
  max := Vector( -HUGE_VALUE );

  for i := 0 to NumObjects - 1 do
  begin
      box := Objects[i].BoundingBox;

      if box.Min.X < min.X then min.X := box.Min.X;
      if box.Min.Y < min.Y then min.Y := box.Min.Y;
      if box.Min.Z < min.Z then min.Z := box.Min.Z;

      if box.Max.X > max.X then max.X := box.Max.X;
      if box.Max.Y > max.Y then max.Y := box.Max.Y;
      if box.Max.Z > max.Z then max.Z := box.Max.Z;
  end;

  // INFO, PJ: Slightly enlarge the bounding box to make sure we include all geometry.
  min := VectorAdd( min, -EPSILON );
  max := VectorAdd( max, EPSILON );

  result := TRTAABoundingBox.Create( min, max );
end;

// ----------------------------------------------------------------------------

function TRTCompoundObject.NumObjects: Integer;
begin
  result := fObjects.Count;
end;

// ----------------------------------------------------------------------------

procedure TRTCompoundObject.AddObject( const anObject: TRTGeometricObject );
begin
  fObjects.Add( anObject );
end;

// ----------------------------------------------------------------------------

function TRTCompoundObject.Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
var
  t: Double;
  normal: TRTNormal;
  localHitPoint: TRTPoint3D;
  i: Integer;
  geometricObject: TRTGeometricObject;
  uvCoordinate: TRTUVCoordinate;

begin
  result := FALSE;
  aMinT := HUGE_VALUE;

  for i := 0 to fObjects.Count - 1 do
  begin
      geometricObject := fObjects[i] as TRTGeometricObject;

      // INFO, PJ: The "Store local copy of most but not all properties" construction is not too good.
      //           Plus, it's a direct copy from TRTWorld.HitObjects. Ugh.
      if geometricObject.Hit( aRay, t, aShadeRec ) and ( t < aMinT ) then
      begin
          result := TRUE;
          aMinT := t;
          RegisterMaterial( geometricObject.Material );
          normal := aShadeRec.Normal;
          localHitPoint := aShadeRec.LocalHitPoint;
          uvCoordinate := aShadeRec.UVCoordinate;
      end;
  end;

  if result then
  begin
      aShadeRec.T := aMinT;
      aShadeRec.Normal := normal;
      aShadeRec.LocalHitPoint := localHitPoint;
      aShadeRec.UVCoordinate := uvCoordinate;
  end
end;

// ----------------------------------------------------------------------------

function TRTCompoundObject.ShadowHit( const aRay: TRTRay; var aMinT: Double ): Boolean;
var
  t: Double;
  i: Integer;
  geometricObject: TRTGeometricObject;

begin
  result := FALSE;
  aMinT := HUGE_VALUE;

  for i := 0 to fObjects.Count - 1 do
  begin
      geometricObject := fObjects[i] as TRTGeometricObject;

      if geometricObject.ShadowHit( aRay, t ) and ( t < aMinT ) then
      begin
          result := TRUE;
          aMinT := t;
      end;
  end;
end;

end.
