{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTGeometricObjectCl;

{$MODE Delphi}

interface

uses
  RTMiscTp, RTMaterialCl, RTAABoundingBoxCl;

type
  TRTGeometricObject = class( TObject )
  private
      fColor: TRGBColor;
      fMaterial: TRTMaterial;
      fCastsShadows: Boolean;
      fOwnsMaterial: Boolean;
      fBoundingBox: TRTAABoundingBox;

      function GetBoundingBox: TRTAABoundingBox;

  protected
      function GetCentroid: TRTPoint3D; virtual;
      function GetMaterial: TRTMaterial; virtual;
      procedure SetMaterial( const aMaterial: TRTMaterial ); virtual;
      procedure RegisterMaterial( const aMaterial: TRTMaterial );

      function MakeBoundingBox: TRTAABoundingBox; virtual;

      property LocalMaterial: TRTMaterial read fMaterial write fMaterial;

  public
      constructor Create;
      destructor Destroy; override;

      function Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean; virtual; abstract;
      function ShadowHit( const aRay: TRTRay; var aMinT: Double ): Boolean; virtual;

      property BoundingBox: TRTAABoundingBox read GetBoundingBox write fBoundingBox;

      property Color: TRGBColor read fColor write fColor; // REFA, PJ: Remove?
      property Material: TRTMaterial read GetMaterial write SetMaterial;
      property Centroid: TRTPoint3D read GetCentroid;  // INFO, PJ: Approximate centroid
      property CastsShadows: Boolean read fCastsShadows write fCastsShadows;
      property OwnsMaterial: Boolean read fOwnsMaterial write fOwnsMaterial; // REFA, PJ: Remove
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils,
  RTMathFn;

constructor TRTGeometricObject.Create;
begin
  inherited Create;

  fCastsShadows := TRUE;
end;

// ----------------------------------------------------------------------------

destructor TRTGeometricObject.Destroy;
begin
  if fOwnsMaterial then
  begin
      FreeAndNil( fMaterial );
  end;

  FreeAndNil( fBoundingBox );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

function TRTGeometricObject.MakeBoundingBox: TRTAABoundingBox;
begin
  raise Exception.Create( Format( 'Class %s does not implement MakeBoundingBox', [ClassName] ));
end;

// ----------------------------------------------------------------------------

function TRTGeometricObject.GetBoundingBox: TRTAABoundingBox;
begin
  if fBoundingBox = nil then
  begin
      fBoundingBox := MakeBoundingBox;
  end;

  result := fBoundingBox;
end;

// ----------------------------------------------------------------------------

function TRTGeometricObject.GetMaterial: TRTMaterial;
begin
  result := fMaterial;
end;

// ----------------------------------------------------------------------------

function TRTGeometricObject.GetCentroid: TRTPoint3D; 
begin
  // INFO, PJ: As a quick approximation, we return the center of the bounding box.
  //           For some non-symmetric objects, like a triangle, a more accurate
  //           answer is beter. Simply override this method in such classes. 
  result := VectorCentroid( [BoundingBox.Min, BoundingBox.Max] );
end;

// ----------------------------------------------------------------------------

procedure TRTGeometricObject.SetMaterial( const aMaterial: TRTMaterial );
begin
  fMaterial := aMaterial;

  fOwnsMaterial := TRUE;
end;

// ----------------------------------------------------------------------------

procedure TRTGeometricObject.RegisterMaterial( const aMaterial: TRTMaterial );
begin
  fMaterial := aMaterial;
end;

// ----------------------------------------------------------------------------

// INFO, PJ:
//   More efficient version of hit (doesn't update shaderec), dedicated to shadowcasting.
//   This default implementation simply calls hit. Override in child classes.
function TRTGeometricObject.ShadowHit( const aRay: TRTRay; var aMinT: Double ): Boolean;
var
  shade: TRTShadeRec;

begin
  result := CastsShadows;

  if result then
  begin
      shade := ShadeRec( nil );
      result := Hit( aRay, aMinT, shade );
  end;
end;

end.
