{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTGridCl;

{$MODE Delphi}

interface

uses
  Contnrs,
  RTCompoundObjectCl, RTIntegerListCl, RTMiscTp;

type
  TRTGrid = class( TRTCompoundObject )
  private
      fCells: TObjectList;
      fCellCounts: TRTIntegerList;
      fCompoundObjects: TObjectList; // INFO, PJ: Release pool

      // INFO, PJ: Number of cells in the X, Y and Z direction
      fNumX: Integer;
      fNumY: Integer;
      fNumZ: Integer;

      procedure InitializeCells;

  public
      constructor Create;
      destructor Destroy; override;

      function Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean; override;
      function ShadowHit( const aRay: TRTRay; var aMinT: Double ): Boolean; override;

      function DebugString: String;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils, Math, Dialogs,
  RTMiscCn, RTAABoundingBoxCl, RTMathFn, RTGeometricObjectCl;

constructor TRTGrid.Create;
begin
  inherited Create;

  fCells := TObjectList.Create( FALSE );
  fCellCounts := TRTIntegerList.Create;
  fCompoundObjects := TObjectList.Create( TRUE );

  OwnsMaterial := FALSE;
end;

// ----------------------------------------------------------------------------

destructor TRTGrid.Destroy;
begin
  FreeAndNil( fCompoundObjects );
  FreeAndNil( fCellCounts );
  FreeAndNil( fCells );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure TRTGrid.InitializeCells;
var
  min, max: TRTPoint3D;
  indexMinX, indexMinY, indexMinZ, indexMaxX, indexMaxY, indexMaxZ: Integer;
  wx, wy, wz, multiplier, s: Single;
  numCells: Integer;
  i, index, ix, iy, iz: integer;
  geometricObject: TRTGeometricObject;
  compoundObject: TRTCompoundObject;
  box: TRTAABoundingBox;

begin
  min := BoundingBox.Min;
  max := BoundingBox.Max;

	// INFO, PJ: Dimensions of the grid in the x, y, and z directions
  wx := max.X - min.X;
  wy := max.Y - min.Y;
  wz := max.Z - min.Z;

  // INFO, PJ: multiplyer scales the number of grid cells relative to the number of objects
  multiplier := 2;

  s := Power( wx * wy * wz / NumObjects, 1/3 );
  fNumX := Trunc( multiplier * wx / s ) + 1;
  fNumY := Trunc( multiplier * wy / s ) + 1;
  fNumZ := Trunc( multiplier * wz / s ) + 1;

  // INFO, PJ: Initialize cells
  numCells := fNumX * fNumY * fNumZ;
  fCells.Capacity := numCells;
  for i := 0 to numCells - 1 do
  begin
      fCells.Add( nil );
  end;

  // INFO, PJ: Initialize cell statistics
  fCellCounts.Capacity := numCells;
  for i := 0 to numCells - 1 do
  begin
      fCellCounts.Add( 0 );
  end;

  fCompoundObjects.Capacity := numCells;

  // INFO, PJ: Put the objects into the grid
  for i := 0 to NumObjects - 1 do
  begin
      geometricObject := Objects[i];
      box := geometricObject.BoundingBox;

		// INFO, PJ: Compute the cell indices at the corners of the bounding box of the object
		indexMinX := Clamp( Trunc((box.Min.X - min.X) * fNumX / (max.X - min.X)), 0, fNumX - 1);
		indexMinY := Clamp( Trunc((box.Min.Y - min.Y) * fNumY / (max.Y - min.Y)), 0, fNumY - 1);
		indexMinZ := Clamp( Trunc((box.Min.Z - min.Z) * fNumZ / (max.Z - min.Z)), 0, fNumZ - 1);
		indexMaxX := Clamp( Trunc((box.Max.X - min.X) * fNumX / (max.X - min.X)), 0, fNumX - 1);
		indexMaxY := Clamp( Trunc((box.Max.Y - min.Y) * fNumY / (max.Y - min.Y)), 0, fNumY - 1);
		indexMaxZ := Clamp( Trunc((box.Max.Z - min.Z) * fNumZ / (max.Z - min.Z)), 0, fNumZ - 1);

      for iz := indexMinZ to indexMaxZ do
      begin
          for iy := indexMinY to indexMaxY do
          begin
              for ix := indexMinX to indexMaxX do
              begin
                  index := fNumX * fNumY * iz + fNumX * iy + ix;

                  if fCellCounts[index] = 0 then
                  begin
                      // INFO, PJ: We simply insert the single object
                      fCells[index] := geometricObject;
                  end
                  else if fCellCounts[index] = 1 then
                  begin
                      // INFO, PJ: We replace the existing object with a compound object
                      compoundObject := TRTCompoundObject.Create( FALSE );
                      compoundObject.AddObject( fCells[index] as TRTGeometricObject );
                      compoundObject.AddObject( geometricObject );
                      fCompoundObjects.Add( compoundObject );
                      fCells[index] := compoundObject;
                  end
                  else
                  begin
                      // INFO, PJ: We extend the compound object already in the cell
                      compoundObject := fCells[index] as TRTCompoundObject;
                      compoundObject.AddObject( geometricObject );
                  end;

                  fCellCounts[index] := fCellCounts[index] + 1;
              end;
          end;
      end;
  end;
end;

// ----------------------------------------------------------------------------

function TRTGrid.Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
var
  box: TRTAABoundingBox;
  geometricObject: TRTGeometricObject;
  p: TRTPoint3D;
  a, b, c: Double;
  t0, t1: Double;
  ox, oy, oz: Double;
  dx, dy, dz: Double;
  x0, y0, z0, x1, y1, z1: Double;
  txMin, tyMin, tzMin: Double;
  txMax, tyMax, tzMax: Double;
  ix, iy, iz: Integer;
  dtx, dty, dtz: Double;
  txNext, tyNext, tzNext: Double;
  ixStep, iyStep, izStep: Integer;
  ixStop, iyStop, izStop: Integer;

begin
  if fCells.Count < NumObjects then
  begin
      InitializeCells;
  end;

  ox := aRay.Origin.X;
  oy := aRay.Origin.Y;
  oz := aRay.Origin.Z;
  dx := aRay.Direction.X;
  dy := aRay.Direction.Y;
  dz := aRay.Direction.Z;

  box := BoundingBox;
  x0 := box.Min.X;
  y0 := box.Min.Y;
  z0 := box.Min.Z;
  x1 := box.Max.X;
  y1 := box.Max.Y;
  z1 := box.Max.Z;

  if dx = 0 then a := HUGE_VALUE else a := 1 / dx;
  if a >= 0 then
  begin
      txMin := (x0 - ox) * a;
      txMax := (x1 - ox) * a;
  end
  else
  begin
      txMin := (x1 - ox) * a;
      txMax := (x0 - ox) * a;
  end;

  if dy = 0 then b := HUGE_VALUE else b := 1 / dy;
  if b >= 0 then
  begin
      tyMin := (y0 - oy) * b;
      tyMax := (y1 - oy) * b;
  end
  else
  begin
      tyMin := (y1 - oy) * b;
      tyMax := (y0 - oy) * b;
  end;

  if dz = 0 then c := HUGE_VALUE else c := 1 / dz;
  if c >= 0 then
  begin
      tzMin := (z0 - oz) * c;
      tzMax := (z1 - oz) * c;
  end
  else
  begin
      tzMin := (z1 - oz) * c;
      tzMax := (z0 - oz) * c;
  end;

  t0 := Max( txMin, tyMin, tzMin );
  t1 := Min( txMax, tyMax, tzMax );

  if t0 > t1 then
  begin
      result := FALSE;
      exit;
  end;

	// INFO, PJ: Initial cell coordinates

	if box.Contains( aRay.Origin ) then
  begin
		ix := Clamp( Trunc((ox - x0) * fNumX / (x1 - x0)), 0, fNumX - 1 );
		iy := Clamp( Trunc((oy - y0) * fNumY / (y1 - y0)), 0, fNumY - 1 );
		iz := Clamp( Trunc((oz - z0) * fNumZ / (z1 - z0)), 0, fNumZ - 1 );
  end
  else
  begin
      // INFO, PJ: Calculate initial hit point with grid's bounding box
		p := VectorAdd( aRay.Origin, VectorMultiply( aRay.Direction, t0 ));
		ix := Clamp( Trunc((p.X - x0) * fNumX / (x1 - x0)), 0, fNumX - 1 );
		iy := Clamp( Trunc((p.Y - y0) * fNumY / (y1 - y0)), 0, fNumY - 1 );
		iz := Clamp( Trunc((p.Z - z0) * fNumZ / (z1 - z0)), 0, fNumZ - 1 );
  end;

	// INFO, PJ: Ray parameter increments per cell in the x, y, and z directions

	dtx := ( txMax - txMin ) / fNumX;
	dty := ( tyMax - tyMin ) / fNumY;
	dtz := ( tzMax - tzMin ) / fNumZ;

  if dx > 0 then
  begin
      txNext := txMin + (ix + 1) * dtx;
      ixStep := 1;
      ixStop := fNumX;
  end
  else
  begin
      txNext := txMin + (fNumX - ix) * dtx;
      ixStep := -1;
      ixStop := -1;
  end;

  if dx = 0 then
  begin
      txNext := HUGE_VALUE;
      ixStep := -1;
      ixStop := 1;
  end;

  if dy > 0 then
  begin
      tyNext := tyMin + (iy + 1) * dty;
      iyStep := 1;
      iyStop := fNumY;
  end
  else
  begin
      tyNext := tyMin + (fNumY - iy) * dty;
      iyStep := -1;
      iyStop := -1;
  end;

  if dy = 0 then
  begin
      tyNext := HUGE_VALUE;
      iyStep := -1;
      iyStop := 1;
  end;

  if dz > 0 then
  begin
      tzNext := tzMin + (iz + 1) * dtz;
      izStep := 1;
      izStop := fNumZ;
  end
  else
  begin
      tzNext := tzMin + (fNumZ - iz) * dtz;
      izStep := -1;
      izStop := -1;
  end;

  if dz = 0 then
  begin
      tzNext := HUGE_VALUE;
      izStep := -1;
      izStop := 1;
  end;

  // INFO, PJ: Traverse the grid
  while TRUE do
  begin
      geometricObject := fCells[ix + fNumX * iy + fNumX * fNumY * iz] as TRTGeometricObject;

      if (txNext < tyNext) and (txNext < tzNext ) then
      begin
          if Assigned( geometricObject ) and geometricObject.Hit( aRay, aMinT, aShadeRec ) and ( aMinT < txNext ) then
          begin
              RegisterMaterial( geometricObject.Material );
              result := TRUE;
              exit;
          end;

          txNext := txNext + dtx;
          ix := ix + ixStep;

          if ix = ixStop then
          begin
              result := FALSE;
              exit;
          end;
      end
      else
      begin
          if tyNext < tzNext then
          begin
              if Assigned( geometricObject ) and geometricObject.Hit( aRay, aMinT, aShadeRec ) and ( aMinT < tyNext ) then
              begin
                  RegisterMaterial( geometricObject.Material );
                  result := TRUE;
                  exit;
              end;

              tyNext := tyNext + dty;
              iy := iy + iyStep;

              if iy = iyStop then
              begin
                  result := FALSE;
                  exit;
              end;
          end
          else
          begin
              if Assigned( geometricObject ) and geometricObject.Hit( aRay, aMinT, aShadeRec ) and ( aMinT < tzNext ) then
              begin
                  RegisterMaterial( geometricObject.Material );
                  result := TRUE;
                  exit;
              end;

              tzNext := tzNext + dtz;
              iz := iz + izStep;

              if iz = izStop then
              begin
                  result := FALSE;
                  exit;
              end;
          end;
      end;
  end;
end;

// ----------------------------------------------------------------------------

function TRTGrid.ShadowHit( const aRay: TRTRay; var aMinT: Double ): Boolean;
var
  shade: TRTShadeRec;

begin
  result := CastsShadows;

  if result then
  begin
      shade := ShadeRec( nil );
      result := Hit( aRay, aMinT, shade );
  end;
end;

// ----------------------------------------------------------------------------

function TRTGrid.DebugString: String;
var
  numZeroes, numOnes, numTwos, numThrees, numGreater: Integer;
  i, count: Integer;
  
begin
  numZeroes := 0;
  numOnes := 0;
  numTwos := 0;
  numThrees := 0;
  numGreater := 0;

  for i := 0 to fCellCounts.Count - 1 do
  begin
      count := fCellCounts[i];

      case count of
          0:  numZeroes := numZeroes + 1;
          1:  numOnes := numOnes + 1;
          2:  numTwos := numTwos + 1;
          3:  numThrees := numThrees + 1;
      else
          numGreater := numGreater + 1;
      end;
  end;

  result :=
      'Objects:     ' + IntToStr( NumObjects ) + CRLF +
      'Cells:       ' + IntToStr( fCells.Count ) + CRLF +
      '  0 objects: ' + IntToStr( numZeroes ) + CRLF +
      '  1 object:  ' + IntToStr( numOnes ) + CRLF +
      '  2 objects: ' + IntToStr( numTwos ) + CRLF +
      '  3 objects: ' + IntToStr( numThrees ) + CRLF +
      '  >3 objects: ' + IntToStr( numGreater );
end;

end.
