{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTInstanceCl;

{$MODE Delphi}

interface

uses
  RTGeometricObjectCl, RTMiscTp;

type
  TRTInstance = class( TRTGeometricObject )
  private
      fObject: TRTGeometricObject;
      fTransformTexture: Boolean;

      fInvMatrix: TRTMatrix;

      property InvMatrix: TRTMatrix read fInvMatrix;

  public
      constructor Create( const anObject: TRTGeometricObject );

      function Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean; override;

      procedure Translate( const aDX, aDY, aDZ: Single );
      procedure Scale( const anX, aY, aZ: Single );
      procedure RotateX( const anAngle: Single );
      procedure RotateY( const anAngle: Single );
      procedure RotateZ( const anAngle: Single );

      property TransformTexture: Boolean read fTransformTexture write fTransformTexture;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  RTMathFn;

constructor TRTInstance.Create( const anObject: TRTGeometricObject );
begin
  inherited Create;

  fObject := anObject;
  fInvMatrix := MatrixIdentity;
  OwnsMaterial := FALSE;
end;

// ----------------------------------------------------------------------------

function TRTInstance.Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
var
  invRay: TRTRay;

begin
  invRay.Origin := MatrixMultiplyPoint( InvMatrix, aRay.Origin );
  invRay.Direction := MatrixMultiplyVector( InvMatrix, aRay.Direction );

  if fObject.Hit( invRay, aMinT, aShadeRec ) then
  begin
      aShadeRec.Normal := MatrixMultiplyNormal( InvMatrix, aShadeRec.Normal );
      VectorNormalize( aShadeRec.Normal );

      if Assigned( fObject.Material ) then
      begin
          LocalMaterial := fObject.Material;
      end;

      if not TransformTexture then
      begin
          aShadeRec.LocalHitPoint := VectorAdd( aRay.Origin, VectorMultiply( aRay.Direction, aMinT ));
      end;

      result := TRUE;
  end
  else
      result := FALSE;
end;

// ----------------------------------------------------------------------------

procedure TRTInstance.Translate( const aDX, aDY, aDZ: Single );
var
  invTransform: TRTMatrix;
  
begin
  invTransform := Matrix(
      1, 0, 0, -aDX,
      0, 1, 0, -aDY,
      0, 0, 1, -aDZ,
      0, 0, 0, 1
  );

  fInvMatrix := MatrixMultiply( fInvMatrix, invTransform );
end;

// ----------------------------------------------------------------------------

procedure TRTInstance.Scale( const anX, aY, aZ: Single );
var
  invTransform: TRTMatrix;
  
begin
  invTransform := Matrix(
      1 / anX, 0, 0, 0,
      0, 1 / aY, 0, 0,
      0, 0, 1 / aZ, 0,
      0, 0, 0, 1
  );

  fInvMatrix := MatrixMultiply( fInvMatrix, invTransform );
end;

// ----------------------------------------------------------------------------

procedure TRTInstance.RotateX( const anAngle: Single );
var
  sinAngle: Single;
  cosAngle: Single;
  invTransform: TRTMatrix;

begin
  sinAngle := Sin( anAngle );
  cosAngle := Cos( anAngle );

  invTransform := Matrix(
      1, 0, 0, 0,
      0, cosAngle, sinAngle, 0,
      0, -sinAngle, cosAngle, 0,
      0, 0, 0, 1
  );

  fInvMatrix := MatrixMultiply( fInvMatrix, invTransform );
end;

// ----------------------------------------------------------------------------

procedure TRTInstance.RotateY( const anAngle: Single );
var
  sinAngle: Single;
  cosAngle: Single;
  invTransform: TRTMatrix;

begin
  sinAngle := Sin( anAngle );
  cosAngle := Cos( anAngle );

  invTransform := Matrix(
      cosAngle, 0, -sinAngle, 0,
      0, 1, 0, 0,
      sinAngle, 0, cosAngle, 0,
      0, 0, 0, 1
  );

  fInvMatrix := MatrixMultiply( fInvMatrix, invTransform );
end;

// ----------------------------------------------------------------------------

procedure TRTInstance.RotateZ( const anAngle: Single );
var
  sinAngle: Single;
  cosAngle: Single;
  invTransform: TRTMatrix;

begin
  sinAngle := Sin( anAngle );
  cosAngle := Cos( anAngle );

  invTransform := Matrix(
      cosAngle, sinAngle, 0, 0,
      -sinAngle, cosAngle, 0, 0,
      0, 0, 1, 0,
      0, 0, 0, 1
  );

  fInvMatrix := MatrixMultiply( fInvMatrix, invTransform );
end;

// ----------------------------------------------------------------------------

end.
