{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTMeshTriangleCl;

{$MODE Delphi}

interface

uses
  RTGeometricObjectCl, RTMeshCl, RTMiscTp, RTAABoundingBoxCl, RTMaterialCl;

type
  TRTTriangleType = ( rtTTFlat, rtTTSmooth );

// ----------------------------------------------------------------------------

  TRTMeshTriangle = class( TRTGeometricObject )
  private
      fMesh: TRTMesh;
      fType: TRTTriangleType;

      fVertex0: TRTPoint3D;
      fVertex1: TRTPoint3D;
      fVertex2: TRTPoint3D;

      fNormals: Array[0..2] of TRTNormal;
      fUVCoordinates: Array[0..2] of TRTUVCoordinate;

      procedure SetNormals( const anIndex: Integer; const aNormal: TRTNormal );
      procedure SetUVCoordinate( const anIndex: Integer; const aCoordinate: TRTUVCoordinate );

      function CalculateSmoothNormal( const aBeta, aGamma: Single ): TRTNormal;
      function CalculateUVCoordinate( const aBeta, aGamma: Single ): TRTUVCoordinate;

  protected
      function MakeBoundingBox: TRTAABoundingBox; override;
      function GetCentroid: TRTPoint3D; override;
      function GetMaterial: TRTMaterial; override;
      procedure SetMaterial( const aMaterial: TRTMaterial ); override;

  public
      constructor Create( const aMesh: TRTMesh; const anIndex1, anIndex2, anIndex3: Integer; const aType: TRTTriangleType ); overload;
      constructor Create( const aMesh: TRTMesh; const aVertex1, aVertex2, aVertex3: TRTPoint3D; const aType: TRTTriangleType ); overload;

      function Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean; override;

      property Normal: TRTNormal write fNormals[0];
      property Normals[const anIndex: Integer]: TRTNormal write SetNormals;
      property UVCoordinates[const anIndex: Integer]: TRTUVCoordinate write SetUVCoordinate;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  RTMathFn, RTMiscCn;

const
  EPSILON = 0.000001;

constructor TRTMeshTriangle.Create( const aMesh: TRTMesh; const anIndex1, anIndex2, anIndex3: Integer; const aType: TRTTriangleType );
var
  v0, v1, v2: TRTPoint3D;

begin
  v0 := fMesh.Vertices[anIndex1]^;
  v1 := fMesh.Vertices[anIndex2]^;
  v2 := fMesh.Vertices[anIndex3]^;

  Create( aMesh, v0, v1, v2, aType );
end;

// ----------------------------------------------------------------------------

constructor TRTMeshTriangle.Create( const aMesh: TRTMesh; const aVertex1, aVertex2, aVertex3: TRTPoint3D; const aType: TRTTriangleType );
begin
  inherited Create;

  fMesh := aMesh;
  fType := aType;

  fVertex0 := aVertex1;
  fVertex1 := aVertex2;
  fVertex2 := aVertex3;

  fNormals[0] := VectorCrossProduct( VectorSubtract( fVertex1, fVertex0 ), VectorSubtract( fVertex2, fVertex0 ));
  VectorNormalize( fNormals[0] );
end;

// ----------------------------------------------------------------------------

function TRTMeshTriangle.Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
var
  v0, v1, v2: PRTPoint3D;
  a, b, c, d, e, f, g, h, i, j, k, l: Double;
  m, n, p, q, r, s, t: Double;
  invDenom, e1, e2, e3, beta, gamma: Double;
  denom: Double;

begin
  a := fVertex0.X - fVertex1.X; b := fVertex0.X - fVertex2.X; c := aRay.Direction.X; d := fVertex0.X - aRay.Origin.X;
  e := fVertex0.Y - fVertex1.Y; f := fVertex0.Y - fVertex2.Y; g := aRay.Direction.Y; h := fVertex0.Y - aRay.Origin.Y;
  i := fVertex0.Z - fVertex1.Z; j := fVertex0.Z - fVertex2.Z; k := aRay.Direction.Z; l := fVertex0.Z - aRay.Origin.Z;

  m := f * k - g * j; n := h * k - g * l; p := f * l - h * j;
  q := g * i - e  * k; s := e * j - f * i;

  denom := ( a * m + b * q + c * s );

  if denom <> 0 then
  begin
      invDenom := 1 / denom;
  end
  else
      invDenom := HUGE_VALUE;

  e1 := d * m - b * n - c * p;
  beta := e1 * invDenom;

  if beta < 0 then
  begin
      result := FALSE;
      exit;
  end;

  r := e * l - h * i;
  e2 := a * n + d * q + c * r;
  gamma := e2 * invDenom;

  if gamma < 0 then
  begin
      result := FALSE;
      exit;
  end;

  if beta + gamma > 1 then
  begin
      result := FALSE;
      exit;
  end;

  e3 := a * p - b * r + d * s;
  t := e3 * invDenom;

  if t < EPSILON then
  begin
      result := FALSE;
      exit;
  end;

  aMinT := t;

  if fType = rtTTSmooth then
  begin
      aShadeRec.Normal := CalculateSmoothNormal( beta, gamma );
  end
  else
      aShadeRec.Normal := fNormals[0];

  aShadeRec.UVCoordinate := CalculateUVCoordinate( beta, gamma );
  aShadeRec.LocalHitPoint := VectorAdd( aRay.Origin, VectorMultiply( aRay.Direction, t ));
  aShadeRec.Material := Material;

  result := TRUE;
end;

// ----------------------------------------------------------------------------

procedure TRTMeshTriangle.SetNormals( const anIndex: Integer; const aNormal: TRTNormal );
begin
  fNormals[anIndex] := aNormal;
end;

// ----------------------------------------------------------------------------

procedure TRTMeshTriangle.SetUVCoordinate( const anIndex: Integer; const aCoordinate: TRTUVCoordinate );
begin
  fUVCoordinates[anIndex] := aCoordinate;
end;

// ----------------------------------------------------------------------------

function TRTMeshTriangle.CalculateSmoothNormal( const aBeta, aGamma: Single ): TRTNormal;
begin
  result := VectorAdd(
      VectorMultiply( fNormals[0], ( 1 - aBeta - aGamma )),
      VectorMultiply( fNormals[1], aBeta ),
      VectorMultiply( fNormals[2], aGamma ));
end;

// ----------------------------------------------------------------------------

function TRTMeshTriangle.CalculateUVCoordinate( const aBeta, aGamma: Single ): TRTUVCoordinate;
begin
    result.U :=
        ( 1 - aBeta - aGamma ) * fUVCoordinates[0].U +
        aBeta * fUVCoordinates[1].U + aGamma * fUVCoordinates[2].U;

    result.V :=
        ( 1 - aBeta - aGamma ) * fUVCoordinates[0].V +
        aBeta * fUVCoordinates[1].V + aGamma * fUVCoordinates[2].V;
end;

// ----------------------------------------------------------------------------

function TRTMeshTriangle.MakeBoundingBox: TRTAABoundingBox;
var
  minPoint, maxPoint: TRTPoint3D;

begin
  minPoint := Vector(
      Min( fVertex0.X, fVertex1.X, fVertex2.X ) - EPSILON,
      Min( fVertex0.Y, fVertex1.Y, fVertex2.Y ) - EPSILON,
      Min( fVertex0.Z, fVertex1.Z, fVertex2.Z ) - EPSILON
  );

  maxPoint := Vector(
      Max( fVertex0.X, fVertex1.X, fVertex2.X ) + EPSILON,
      Max( fVertex0.Y, fVertex1.Y, fVertex2.Y ) + EPSILON,
      Max( fVertex0.Z, fVertex1.Z, fVertex2.Z ) + EPSILON
  );

  result := TRTAABoundingBox.Create( minPoint, maxPoint );
end;

// ----------------------------------------------------------------------------

function TRTMeshTriangle.GetCentroid: TRTPoint3D;
begin
  result := VectorCentroid( [fVertex0, fVertex1, fVertex2] ); 
end;

// ----------------------------------------------------------------------------

function TRTMeshTriangle.GetMaterial: TRTMaterial;
begin
  result := LocalMaterial;

  if not Assigned( result ) then
  begin
      result := fMesh.Material;
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTMeshTriangle.SetMaterial( const aMaterial: TRTMaterial );
begin
  inherited SetMaterial( aMaterial );

  OwnsMaterial := FALSE;
end;

end.
