{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTOpenCylinderCl;

{$MODE Delphi}

interface

uses
  RTGeometricObjectCl, RTMiscTp, RTDiscCl;

type
  TRTOpenCylinder = class( TRTGeometricObject )
  private
      fBottom: Double;
      fTop: Double;
      fRadius: Double;
      fInvRadius: Double;

      procedure SetRadius( const aRadius: Double );

  public
      constructor Create; overload;
      constructor Create( const aBottom, aTop, aRadius: Double ); overload;

      function Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean; override;

      property Bottom: Double read fBottom write fBottom;
      property Top: Double read fTop write fTop;
      property Radius: Double read fRadius write SetRadius;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils,
  RTMathFn;

const
  EPSILON = 0.001;

constructor TRTOpenCylinder.Create;
begin
  Create( -1, 1, 1 );
end;

// ----------------------------------------------------------------------------

constructor TRTOpenCylinder.Create( const aBottom, aTop, aRadius: Double );
begin
  inherited Create;

  fBottom := aBottom;
  fTop := aTop;
  fRadius := aRadius;
  fInvRadius := 1 / fRadius;
end;

// ----------------------------------------------------------------------------

procedure TRTOpenCylinder.SetRadius( const aRadius: Double );
begin
  fRadius := aRadius;
  fInvRadius := 1 / fRadius;
end;

// ----------------------------------------------------------------------------

function TRTOpenCylinder.Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
var
  ox, oy, oz, dx, dy, dz: Double;
  a, b, c, disc, e, denom, t, yHit: Double;

begin
  ox := aRay.Origin.X;
  oy := aRay.Origin.Y;
  oz := aRay.Origin.Z;
  dx := aRay.Direction.X;
  dy := aRay.Direction.Y;
  dz := aRay.Direction.Z;

  a := dx * dx + dz * dz;
  b := 2 * ( ox * dx + oz * dz );
  c := ox * ox + oz * oz - radius * radius;
  disc := b * b - 4 * a * c;

  result := FALSE;

  if disc > 0 then
  begin
      e := Sqrt( disc );
      denom := 2 * a;
      t := ( -b - e ) / denom;  // INFO, PJ: Smaller root

      if t > EPSILON then
      begin
          yHit := oy + t * dy;

          if ( yHit > bottom ) and ( yHit < top ) then
          begin
              aMinT := t;
              aShadeRec.Normal := Vector( ( ox + t * dx ) * fInvRadius, 0, ( oz + t * dz ) * fInvRadius );

				// INFO, PJ: Test for hitting from inside (which doesn't work with transparency at all)
              if VectorDotProduct( aRay.Direction, aShadeRec.Normal ) > 0 then
              begin
                  aShadeRec.Normal := VectorMultiply( aShadeRec.Normal, -1 );
              end;

              aShadeRec.LocalHitPoint := VectorAdd( aRay.Origin, VectorMultiply( aRay.Direction, aMinT ));

              result := TRUE;
          end;
      end
      else
      begin
		    t := ( -b + e ) / denom;    // INFO, PJ: Larger root

          if t > EPSILON then
          begin
              yHit := oy + t * dy;

              if ( yHit > bottom ) and ( yHit < top ) then
              begin
                  aMinT := t;
                  aShadeRec.Normal := Vector(( ox + t * dx ) * fInvRadius, 0, ( oz + t * dz ) * fInvRadius );

                  // INFO, PJ: Test for hitting from inside (which doesn't work with transparency at all)
                  if VectorDotProduct( aRay.Direction, aShadeRec.Normal ) > 0 then
                  begin
                      aShadeRec.Normal := VectorMultiply( aShadeRec.Normal, -1 );
                  end;

                  aShadeRec.LocalHitPoint := VectorAdd( aRay.Origin, VectorMultiply( aRay.Direction, aMinT ));

                  result := TRUE;
              end;
          end;
      end;
  end;
end;


end.

