{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTSphereCl;

{$MODE Delphi}

interface

uses
  RTGeometricObjectCl, RTMiscTp, RTAABoundingBoxCl;

type
  TRTSphereType = ( rtSTConvex, rtSTConcave );

  TRTSphere = class( TRTGeometricObject )
  private
      fCenter: TRTPoint3D;
      fRadius: Double;
      fType: TRTSphereType;

      function CalculateUVCoordinate( const aPoint: TRTPoint3D ): TRTUVCoordinate;

  protected
      function MakeBoundingBox: TRTAABoundingBox; override;

  public
      constructor Create; overload;
      constructor Create( const aCenter: TRTPoint3D; const aRadius: Double; const aType: TRTSphereType = rtSTConvex ); overload;

      function Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean; override;

      property Center: TRTPoint3D write fCenter;
      property Radius: Double write fRadius;
      property SphereType: TRTSphereType read fType write fType;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  Math,
  RTMathFn;

const
  EPSILON = 0.001;

constructor TRTSphere.Create;
begin
  Create( Vector( 0, 0, 0 ), 1, rtSTConvex );
end;

// ----------------------------------------------------------------------------

constructor TRTSphere.Create( const aCenter: TRTPoint3D; const aRadius: Double; const aType: TRTSphereType ); 
begin
  inherited Create;

  fCenter := aCenter;
  fRadius := aRadius;
  fType := aType;
end;

// ----------------------------------------------------------------------------

function TRTSphere.MakeBoundingBox: TRTAABoundingBox;
begin
  result := TRTAABoundingBox.Create( VectorAdd( fCenter, -fRadius ), VectorAdd( fCenter, fRadius ));
end;

// ----------------------------------------------------------------------------

function TRTSphere.Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
var
  t, a, b, c, d, e, denom: Double;
  diff: TRTVector3D;

begin
  diff := VectorSubtract( aRay.Origin, fCenter );

  a := VectorDotProduct( aRay.Direction, aRay.Direction );
  b := 2 * VectorDotProduct( diff, aRay.Direction );
  c := VectorDotProduct( diff, diff ) - fRadius * fRadius;
  d := b * b - 4 * a * c;

  result := FALSE;

  if d >= 0 then
  begin
      e := Sqrt( d );
      denom := 2 * a;

      // INFO, PJ: Smaller root.
      t := ( -b - e ) / denom;
      if t > EPSILON then
      begin
          aMinT := t;
          aShadeRec.Normal := VectorDivide( VectorAdd( diff, VectorMultiply( aRay.Direction, t )), fRadius );
          aShadeRec.LocalHitPoint := VectorAdd( aRay.Origin, VectorMultiply( aRay.Direction, t ));
          result := TRUE;
      end
      else
      begin
          // INFO, PJ: Larger root
          t := ( -b + e ) / denom;

          if t > EPSILON then
          begin
              aMinT := t;
              aShadeRec.Normal := VectorDivide( VectorAdd( diff, VectorMultiply( aRay.Direction, t )), fRadius );
              aShadeRec.LocalHitPoint := VectorAdd( aRay.Origin, VectorMultiply( aRay.Direction, t ));
              result := TRUE;
          end;
      end;
  end;

  if result then
  begin
      aShadeRec.Material := Material;

      if fType = rtSTConcave then
      begin
          aShadeRec.Normal := VectorMultiply( aShadeRec.Normal, -1 );
      end;

      if Material.ContainsTextures then
      begin
          aShadeRec.UVCoordinate := CalculateUVCoordinate( aShadeRec.LocalHitPoint );
      end;
  end;
end;

// ----------------------------------------------------------------------------

function TRTSphere.CalculateUVCoordinate( const aPoint: TRTPoint3D ): TRTUVCoordinate;
var
  position: TRTPoint3D;
  theta: Single;
  phi: Single;

begin
  position := VectorDivide( VectorSubtract( aPoint, fCenter ), fRadius );

  theta := ArcCos( position.Y );
  phi := ArcTan2( position.Z, position.X );

  result.U := 1 - phi / ( 2 * PI );
  result.V := theta / PI;
end;

end.
