{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTTorusCl;

{$MODE Delphi}

interface

uses
  RTGeometricObjectCl, RTMiscTp, RTBoxCl;

type
  TRTTorus = class( TRTGeometricObject )
  private
      fA: Double;
      fB: Double;
      fBoundingBox: TRTBox;

      function ComputeNormal( const aLocation: TRTPoint3D ): TRTNormal;

  public
      constructor Create; overload;
      constructor Create( const aRadius, aThickness: Double ); overload;
      destructor Destroy; override;

      function Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean; override;

      property Radius: Double write fA;
      property Thickness: Double write fB;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils,
  RTMathFn, RTMiscCn;

const
  EPSILON = 0.001;

constructor TRTTorus.Create;
begin
  Create( 1, 0.5 );
end;

// ----------------------------------------------------------------------------

constructor TRTTorus.Create( const aRadius, aThickness: Double );
begin
  inherited Create;

  fA := aRadius;
  fB := aThickness;
  fBoundingBox := TRTBox.Create( Vector( -fA - fB, -fB, -fA -fB ), Vector( fA + fB, fB, fA + fB ));
end;

// ----------------------------------------------------------------------------

destructor TRTTorus.Destroy;
begin
  FreeAndNil( fBoundingBox );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

function TRTTorus.ComputeNormal( const aLocation: TRTPoint3D ): TRTNormal;
var
  x, y, z: Double;
  xSqr, ySqr, zSqr, aSqr, bSqr: Double;

begin
  x := aLocation.X;
  y := aLocation.Y;
  z := aLocation.Z;

  xSqr := x * x;
  ySqr := y * y;
  zSqr := z * z;

  aSqr := fA * fA;
  bSqr := fB * fB;

  result.X := 4 * x * ( xSqr + ySqr + zSqr - ( aSqr + bSqr ));
  result.Y := 4 * y * ( xSqr + ySqr + zSqr - ( aSqr + bSqr ) + 2 * aSqr );
  result.Z := 4 * z * ( xSqr + ySqr + zSqr - ( aSqr + bSqr ));

  VectorNormalize( result );
end;

// ----------------------------------------------------------------------------

function TRTTorus.Hit( const aRay: TRTRay; var aMinT: Double; var aShadeRec: TRTShadeRec ): Boolean;
var
  x1, y1, z1, d1, d2, d3: Double;
  t, e, f, sumDSqr, fourASqr: Double;
  i, numRealRoots: Integer;
  isIntersected: Boolean;
  coeffs: TQuarticCoefficients;
  roots: TQuarticRoots;

begin
  if not fBoundingBox.Hit( aRay, aMinT, aShadeRec ) then
  begin
      result := FALSE;
      exit;
  end;

  x1 := aRay.Origin.X; y1 := aRay.Origin.Y; z1 := aRay.Origin.Z;
  d1 := aRay.Direction.X; d2 := aRay.Direction.Y; d3 := aRay.Direction.Z;

  sumDSqr := d1 * d1 + d2 * d2 + d3 * d3;
  e := x1 * x1 + y1 * y1 + z1 * z1 - fA * fA - fB * fB;
  f := x1 * d1 + y1 * d2 + z1 * d3;
  fourASqr := 4 * fA * fA;

  coeffs[0] := e * e - fourASqr * ( fB * fB - y1 * y1 );
  coeffs[1] := 4 * f * e + 2 * fourASqr * y1 * d2;
  coeffs[2] := 2 * sumDSqr * e + 4 * f * f + fourASqr * d2 * d2;
  coeffs[3] := 4 * sumDSqr * f;
  coeffs[4] := sumDSqr * sumDSqr;

  numRealRoots := SolveQuartic( coeffs, roots );

  if numRealRoots = 0 then
  begin
      result := FALSE;
      exit;
  end;

  isIntersected := FALSE;
  t:= HUGE_VALUE;

  // INFO, PJ: e find the smallest root > EPSILON (if any)
  for i := 0 to numRealRoots - 1 do
  begin
      if ( roots[i] > EPSILON ) and ( roots[i] < t ) then
      begin
          isIntersected := TRUE;
          t := roots[i];
      end;
  end;

  if not isIntersected then
  begin
      result := FALSE;
      exit;
  end;

  aMinT := t;
  aShadeRec.LocalHitPoint := VectorAdd( aRay.Origin, VectorMultiply( aRay.Direction, t ));
  aShadeRec.Normal := ComputeNormal( aShadeRec.LocalHitPoint );

  result := TRUE;
end;


end.

