{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTPhongMaterialCl;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  INFO, PJ: Currently, only supports texture in diffuse part of BRDF.
  ---------------------------------------------------------------------------- }

interface

uses
  RTMaterialCl, RTGlossySpecularBRDFCl, RTMiscTp,
  RTGenericLambertianBRDFCl, RTTextureCl;

type
  TRTPhongMaterial = class( TRTMaterial )
  private
      fAmbientBRDF: TRTGenericLambertianBRDF;
      fDiffuseBRDF: TRTGenericLambertianBRDF;
      fSpecularBRDF: TRTGlossySpecularBRDF;

      procedure SetColor( const aColor: TRGBColor );
      procedure SetTexture( const aTexture: TRTTexture );
      procedure SetAmbientReflection( const aValue: Single );
      procedure SetDiffuseReflection( const aValue: Single );
      procedure SetSpecularReflection( const aValue: Single );
      procedure SetSpecularExponent( const aValue: Single );
      procedure SetSpecularColor( const aColor: TRGBColor );

  public
      constructor Create;
      destructor Destroy; override;

      function Shade( const aShadeRec: TRTShadeRec ): TRGBColor; override;
      function ContainsTextures: Boolean; override;

      property AmbientReflection: Single write SetAmbientReflection;
      property DiffuseReflection: Single write SetDiffuseReflection;
      property SpecularReflection: Single write SetSpecularReflection;
      property SpecularExponent: Single write SetSpecularExponent;
      property SpecularColor: TRGBColor write SetSpecularColor;

      // INFO, PJ: Mutually exclusive: Color vs. Texture
      property Color: TRGBColor write SetColor;
      property Texture: TRTTexture write SetTexture;

  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils,
  RTLightCl, RTWorldCl, RTMathFn, RTMiscCn, RTLambertianBRDFCl, RTTexturedLambertianBRDFCl;
  
constructor TRTPhongMaterial.Create;
begin
  inherited Create;

  fambientBRDF := TRTLambertianBRDF.Create;
  fDiffuseBRDF := TRTLambertianBRDF.Create;
  fSpecularBRDF := TRTGlossySpecularBRDF.Create;
end;

// ----------------------------------------------------------------------------

destructor TRTPhongMaterial.Destroy;
begin
  FreeAndNil( fSpecularBRDF );
  FreeAndNil( fDiffuseBRDF );
  FreeAndNil( fAmbientBRDF );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure TRTPhongMaterial.SetAmbientReflection( const aValue: Single );
begin
  fAmbientBRDF.DiffuseReflection := aValue;
end;

// ----------------------------------------------------------------------------

procedure TRTPhongMaterial.SetDiffuseReflection( const aValue: Single );
begin
  fDiffuseBRDF.DiffuseReflection := aValue;
end;

// ----------------------------------------------------------------------------

procedure TRTPhongMaterial.SetSpecularReflection( const aValue: Single );
begin
  fSpecularBRDF.SpecularReflection := aValue;
end;

// ----------------------------------------------------------------------------

procedure TRTPhongMaterial.SetSpecularExponent( const aValue: Single );
begin
  fSpecularBRDF.SpecularExponent := aValue;
end;

// ----------------------------------------------------------------------------

procedure TRTPhongMaterial.SetTexture( const aTexture: TRTTexture );
var
  reflection: Single;
  
begin
  if not ( fAmbientBRDF is TRTTexturedLambertianBRDF ) then
  begin
      // INFO, PJ: Not very elegant, change some day.
      reflection := fAmbientBRDF.DiffuseReflection;
      FreeAndNil( fAmbientBRDF );

      fAmbientBRDF := TRTTexturedLambertianBRDF.Create;
      fAmbientBRDF.DiffuseReflection := reflection;

      reflection := fDiffuseBRDF.DiffuseReflection;
      FreeAndNil( fDiffuseBRDF );

      fDiffuseBRDF := TRTTexturedLambertianBRDF.Create;
      fDiffuseBRDF.DiffuseReflection := reflection;
  end;

  TRTTexturedLambertianBRDF(fAmbientBRDF).DiffuseTexture := aTexture;
  TRTTexturedLambertianBRDF(fDiffuseBRDF).DiffuseTexture := aTexture;
end;

// ----------------------------------------------------------------------------

procedure TRTPhongMaterial.SetColor( const aColor: TRGBColor );
var
  reflection: Single;

begin
  if not ( fAmbientBRDF is TRTLambertianBRDF ) then
  begin
      // INFO, PJ: Not very elegant, change some day.
      reflection := fAmbientBRDF.DiffuseReflection;
      FreeAndNil( fAmbientBRDF );

      fAmbientBRDF := TRTLambertianBRDF.Create;
      fAmbientBRDF.DiffuseReflection := reflection;

      reflection := fDiffuseBRDF.DiffuseReflection;
      FreeAndNil( fDiffuseBRDF );

      fDiffuseBRDF := TRTLambertianBRDF.Create;
      fDiffuseBRDF.DiffuseReflection := reflection;
  end;

  TRTLambertianBRDF(fAmbientBRDF).DiffuseColor := aColor;
  TRTLambertianBRDF(fDiffuseBRDF).DiffuseColor := aColor;
end;

// ----------------------------------------------------------------------------

procedure TRTPhongMaterial.SetSpecularColor( const aColor: TRGBColor );
begin
  fSpecularBRDF.SpecularColor := aColor;
end;

// ----------------------------------------------------------------------------

function TRTPhongMaterial.Shade( const aShadeRec: TRTShadeRec ): TRGBColor;
var
  wi, wo: TRTVector3D;
  world: TRTWorld;
  i: Integer;
  nDotWi: Single;
  radiance: TRGBColor;
  light: TRTLight;
  isInShadow: Boolean;
  shadowRay: TRTRay;

begin
  world := TRTWorld( aShadeRec.World );
  wo := VectorMultiply( aShadeRec.Ray.Direction, -1 );
          
  // INFO, PJ: Ambient color
  result := ColorMultiply( fAmbientBRDF.Rho( aShadeRec, wo ), world.AmbientLight.L( aShadeRec ));

  for i := 0 to world.NumLights - 1 do
  begin
      light := world.Lights[i];

      wi := light.MakeDirection( aShaderec );
      nDotWi := VectorDotProduct( aShadeRec.Normal, wi );

      if nDotWi > 0 then
      begin
          isInShadow := FALSE;

          if ReceivesShadows and light.CastsShadows then
          begin
              shadowRay := Ray( aShadeRec.HitPoint, wi );

              isInShadow := light.IsInShadow( shadowRay, aShadeRec );
          end;

          if not isInShadow then
          begin
              radiance := ColorAdd( fDiffuseBRDF.F( aShadeRec, wi, wo ), fSpecularBRDF.F( aShadeRec, wo, wi ));
              radiance := ColorMultiply( radiance, ColorMultiply( world.Lights[i].L( aShadeRec ), nDotWi ));

              result := ColorAdd( result, radiance );
          end;
      end;
  end;
end;

// ----------------------------------------------------------------------------

function TRTPhongMaterial.ContainsTextures: Boolean;
begin
  // INFO, PJ: Not very elegant or powerful. Change some day.
  result := ( fDiffuseBRDF is TRTTexturedLambertianBRDF );
end;

end.

