{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTSamplerCl;

{$MODE Delphi}

interface

uses
  RTMiscTp,
  RTPoint2DListCl, RTVector3DListCl, RTIntegerListCl;

type
  TRTSampler = class( TObject )
  private
      fNumSamples: Integer;
      fNumSets: Integer;
      fIndexOffset: Integer;
      fSamples: TRTPoint2DList;
      fDiskSamples: TRTPoint2DList;
      fHemisphereSamples: TRTVector3DList;
      fShuffledIndices: TRTIntegerList;
      fCount: Integer;

      procedure InitializeShuffledIndices;
      function NextIndex: Integer;

  protected
      procedure GenerateSamples; virtual; abstract;

      property NumSets: Integer read fNumSets;
      property Samples: TRTPoint2DList read fSamples;

  public
      constructor Create( const aNumSamples: Integer; const aNumSets: Integer = 83 );
      destructor Destroy; override;

      procedure MapSamplesToUnitDisk;
      procedure MapSamplesToHemisphere( const anExponent: Single );
      
      function NextSampleUnitSquare: TRTPoint2D;
      function NextSampleUnitDisk: TRTPoint2D;
      function NextSampleHemisphere: TRTPoint3D;

      property NumSamples: Integer read fNumSamples;
      property Count: Integer read fCount;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils, Math;

constructor TRTSampler.Create( const aNumSamples: Integer; const aNumSets: Integer );
begin
  inherited Create;

  fNumSamples := aNumSamples;
  fNumSets := aNumSets;

  fSamples := TRTPoint2DList.Create;
  fShuffledIndices := TRTIntegerList.Create;
  fShuffledIndices.Capacity := fNumSamples * fNumSets;

  GenerateSamples;
  InitializeShuffledIndices;
end;

// ----------------------------------------------------------------------------

destructor TRTSampler.Destroy;
begin
  FreeAndNil( fSamples );
  FreeAndNil( fDiskSamples );
  FreeAndNil( fHemisphereSamples );
  FreeAndNil( fShuffledIndices );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure TRTSampler.InitializeShuffledIndices;
var
  indices: TRTIntegerList;
  j, p: Integer;

begin
  indices := TRTIntegerList.Create;
  for j := 0 to NumSamples - 1 do
  begin
      indices.Add( j );
  end;

  for p := 0 to NumSets - 1 do
  begin
      indices.Shuffle;

      for j := 0 to NumSamples - 1 do
      begin
          fShuffledIndices.Add( indices[j] );
      end;
  end;

  FreeAndNil( indices );
end;

// ----------------------------------------------------------------------------

procedure TRTSampler.MapSamplesToUnitDisk;
var
  i: integer;
  sp: TRTPoint2D;
  r, phi: Single;
  
begin
  FreeAndNil( fDiskSamples );
  fDiskSamples := TRTPoint2DList.Create;
  fDiskSamples.Capacity := fSamples.Count;

  for i := 0 to fSamples.Count - 1 do
  begin
      // INFO, PJ: Map sample points onto [-1,1] [-1, 1] region.
      sp.X := 2 * fSamples[i].X - 1;
      sp.Y := 2 * fSamples[i].Y - 1;

      if sp.X > -sp.Y then
      begin
          if sp.X > sp.Y then
          begin   // INFO, PJ: Sector 1
              r := sp.X;
              phi := sp.Y / sp.X;
          end
          else
          begin   // INFO, PJ: Sector 2
              r := sp.Y;
              phi := 2 - sp.X / sp.Y;
          end;
      end
      else
      begin
          if sp.X < sp.Y then
          begin   // INFO, PJ: Sector 3
              r := -sp.X;
              phi := 4 + sp.Y / sp.X;
          end
          else
          begin   // INFO, PJ: Sector 4
              r := -sp.y;
              if sp.Y <> 0 then
              begin   // INFO, PJ: Avoid division by zero at origin
                  phi := 6 - sp.X / sp.Y;
              end
              else
                  phi := 0;
          end;
      end;

      phi := phi * PI / 4;

      sp.X := r * Cos( phi );
      sp.Y := r * Sin( phi );

      fDiskSamples.Add( sp );
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTSampler.MapSamplesToHemisphere( const anExponent: Single );
var
  sp: TRTPoint2D;
  point: TRTPoint3D;
  cosPhi, sinPhi, cosTheta, sinTheta: single;
  j: Integer;
  
begin
  FreeAndNil( fHemisphereSamples );
  fHemisphereSamples := TRTVector3DList.Create;
  fHemisphereSamples.Capacity := fSamples.Count;

  for j := 0 to fSamples.Count - 1 do
  begin
      sp := fSamples[j]^;

      cosPhi := Cos( 2 * PI * sp.X );
      sinPhi := Sin( 2 * PI * sp.X );
      cosTheta := Power( 1 - sp.Y, 1 / ( anExponent + 1 ));
      sinTheta := Sqrt( 1 - cosTheta * cosTheta );

      point.X := sinTheta * cosPhi;
      point.Y := sinTheta * sinPhi;
      point.Z := cosTheta;
      fHemisphereSamples.Add( point );
  end;
end;

// ----------------------------------------------------------------------------

function TRTSampler.NextSampleUnitSquare: TRTPoint2D;
begin
  result := fSamples[NextIndex]^;
end;

// ----------------------------------------------------------------------------

function TRTSampler.NextSampleUnitDisk: TRTPoint2D;
begin
  result := fDiskSamples[NextIndex]^;
end;

// ----------------------------------------------------------------------------

function TRTSampler.NextSampleHemisphere: TRTPoint3D;
begin
  result := fHemisphereSamples[NextIndex]^;
end;

// ----------------------------------------------------------------------------

function TRTSampler.NextIndex: integer;
begin
  // INFO, PJ: Isn't all this modding a tad expensive? Test some day!
  if (fCount mod fNumSamples) = 0 then
  begin
      // INFO, PJ: Jump to the next random set
      fIndexOffset := random( fNumSets ) * fNumSamples;
  end;

  result := fIndexOffset + fShuffledIndices[fIndexOffset + fCount mod fNumSamples];

  fCount := fCount + 1;
end;

end.
