{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTPLYMeshLoaderCl;

{$MODE Delphi}

interface

uses
  RTMeshLoaderCl, RTMeshCl, RTTextFileStreamCl;

type
  TRTPLYHeader = record
      NumVertices: Integer;
      NumFaces: Integer;
  end;

// ----------------------------------------------------------------------------

  TRTPLYMeshLoader = class( TRTMeshLoader )
  private
      function StrHasPrefix( const aString, aPrefix: String ): Boolean;

      function ParseString( const aString: String; const anIndex: Integer; out aValue: String ): Integer;
      function ParseFloat( const aString: String; const anIndex: Integer; out aFloat: Double ): Integer;
      function ParseInteger( const aString: String; const anIndex: Integer; out anInteger: Integer ): Integer;

      function ReadHeader( const aStream: TRTTextFileStream ): TRTPLYHeader;
      procedure AppendVertices( const aStream: TRTTextFileStream; const aHeader: TRTPLYHeader; const aMesh: TRTMesh );
      procedure AppendFaces( const aStream: TRTTextFileStream; const aHeader: TRTPLYHeader; const aMesh: TRTMesh );
  public
      function Load( const aFileName: String ): TRTMesh; override;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils,
  RTMiscTp;

function TRTPLYMeshLoader.Load( const aFileName: String ): TRTMesh;
var
  header: TRTPlyHeader;
  stream: TRTTextFileStream;

begin
  result := TRTMesh.Create;

  stream := TRTTextFileStream.Create( aFileName, fmOpenRead and fmShareDenyNone );
  try
      stream.LineSize := 80;
      
      header := ReadHeader( stream );

      AppendVertices( stream, header, result );
      AppendFaces( stream, header, result );

  finally
      FreeAndNil( stream );
  end;
end;

// ----------------------------------------------------------------------------

function TRTPLYMeshLoader.ReadHeader( const aStream: TRTTextFileStream ): TRTPLYHeader;
var
  line: String;
  
begin
  FillChar( result, SizeOf( result ), 0 );

  while aStream.Position < aStream.Size do
  begin
      line := aStream.ReadLine;

      if line = 'end_header' then
      begin
          break;
      end
      else if StrHasPrefix( line, 'element vertex ' ) then
      begin
          result.NumVertices := StrToInt( Copy( line, Length( 'element vertex ' ) + 1, Length( line )));
      end
      else if StrHasPrefix( line, 'element face ' ) then
      begin
          result.NumFaces := StrToInt( Copy( line, Length( 'element face ' ) + 1, Length( line )));
      end;
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTPLYMeshLoader.AppendVertices( const aStream: TRTTextFileStream; const aHeader: TRTPLYHeader; const aMesh: TRTMesh );
var
  i, index: Integer;
  vertex: TRTPoint3D;
  line: String;

begin
  for i := 0 to aHeader.NumVertices - 1 do
  begin
      line := aStream.ReadLine;

      index := ParseFloat( line, 1, vertex.X );
      index := ParseFloat( line, index, vertex.Y );
      ParseFloat( line, index, vertex.Z );

      aMesh.AddVertex( vertex );
  end;
end;

// ----------------------------------------------------------------------------

procedure TRTPLYMeshLoader.AppendFaces( const aStream: TRTTextFileStream; const aHeader: TRTPLYHeader; const aMesh: TRTMesh );
var
  i, index: Integer;
  numVertices, v0, v1, v2: Integer;
  line: String;
  
begin
  for i := 0 to aHeader.NumFaces - 1 do
  begin
      line := aStream.ReadLine;

      index := ParseInteger( line, 1, numVertices );
      index := ParseInteger( line, index, v0 );
      index := ParseInteger( line, index, v1 );
      ParseInteger( line, index, v2 );

      if numVertices <> 3 then
      begin
          raise Exception.Create( Format( 'Face with %d vertices found, only triangles are currently supported.', [numVertices] ));
      end;

      aMesh.AddTriangle( v0, v1, v2, nil );
  end;
end;

// ----------------------------------------------------------------------------

function TRTPLYMeshLoader.StrHasPrefix( const aString, aPrefix: String ): Boolean;
begin
   // INFO, PJ: Typical string library function...

   result := Copy( aString, 1, Length( aPrefix )) = aPrefix;
end;

// ----------------------------------------------------------------------------

function TRTPLYMeshLoader.ParseString( const aString: String; const anIndex: Integer; out aValue: String ): Integer;
begin
  result := anIndex;
  while aString[result] <> ' ' do
  begin
      result := result + 1;
  end;

  aValue := Copy( aString, anIndex, result - anIndex );

  result := result + 1;
end;

// ----------------------------------------------------------------------------

function TRTPLYMeshLoader.ParseFloat( const aString: String; const anIndex: Integer; out aFloat: Double ): Integer;
var
  value: String;

begin
  result := ParseString( aString, anIndex, value );

  aFloat := StrToFloat( value );
end;

// ----------------------------------------------------------------------------

function TRTPLYMeshLoader.ParseInteger( const aString: String; const anIndex: Integer; out anInteger: Integer ): Integer;
var
  value: String;

begin
  result := ParseString( aString, anIndex, value );

  anInteger := StrToInt( value );
end;


end.

