{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTMathFn;

{$MODE Delphi}

interface

uses
  RTMiscTp;

// INFO, PJ: All these Vector functions can be applied to Vectors, Points and Normals.
function VectorAdd( const aV1, aV2: TRTVector3D ): TRTVector3D; overload;
function VectorAdd( const aV1, aV2, aV3: TRTVector3D ): TRTVector3D; overload;
function VectorAdd( const aV: TRTVector3D; const aScalar: Double ): TRTVector3D; overload;
function VectorSubtract( const aV1, aV2: TRTVector3D ): TRTVector3D; overload;
function VectorSubtract( const aV: TRTVector3D; const aScalar: Double ): TRTVector3D; overload;
function VectorMultiply( const aV: TRTVector3D; const aScalar: Double ): TRTVector3D;
function VectorDivide( const aV: TRTVector3D; const aScalar: Double ): TRTVector3D;
function VectorEquals( const aV1, aV2: TRTVector3D ): Boolean; overload;
function VectorEquals( const aV1: TRTVector3D; const aValue: Double ): Boolean; overload;
function VectorLength( const aV: TRTVector3D ): Double;
function VectorSquareLength( const aV: TRTVector3D ): Double;
function VectorDotProduct( const aV1, aV2: TRTVector3D ): Double;
function VectorCrossProduct( const aV1, aV2: TRTVector3D ): TRTVector3D;
procedure VectorNormalize( var aV: TRTVector3D );
function VectorHat( const aV: TRTVector3D ): TRTVector3D;
function VectorDistance( const aV1, aV2: TRTVector3D ): Double;
function VectorSquareDistance( const aV1, aV2: TRTVector3D ): Double;
function VectorMin( const aV1, aV2: TRTVector3D ): TRTVector3D;
function VectorMax( const aV1, aV2: TRTVector3D ): TRTVector3D;
function VectorMaxAxis( const aV: TRTVector3D ): TRTAxis3D;
function VectorCentroid( const aVectors: Array of TRTVector3D ): TRTVector3D;

// ----------------------------------------------------------------------------

function PointMultiply( const aP1: TRTPoint2D; const aScalar: Double ): TRTPoint2D;

// ----------------------------------------------------------------------------

function ColorAdd( const aC1, aC2: TRGBColor ): TRGBColor;
function ColorMultiply( const aC: TRGBColor; const aScalar: Double ): TRGBColor; overload;
function ColorMultiply( const aC1, aC2: TRGBColor ): TRGBColor; overload;
function ColorDivide( const aC: TRGBColor; const aScalar: Double ): TRGBColor;
function ColorPower( const aC: TRGBColor; const aPower: Double ): TRGBColor;

// ----------------------------------------------------------------------------

function MatrixMultiply( const aMatrix: TRTMatrix; const aScalar: Double ): TRTMatrix; overload;
function MatrixMultiply( const aMatrix1, aMatrix2: TRTMatrix ): TRTMatrix; overload;
function MatrixMultiplyPoint( const aMatrix: TRTMatrix; const aPoint: TRTPoint3D ): TRTPoint3D;
function MatrixMultiplyVector( const aMatrix: TRTMatrix; const aVector: TRTVector3D ): TRTVector3D;
function MatrixMultiplyNormal( const aMatrix: TRTMatrix; const aNormal: TRTNormal ): TRTVector3D;
function MatrixDivide( const aMatrix: TRTMatrix; const aScalar: Double ): TRTMatrix;
function MatrixForAxisRotation( const anAxis: TRTVector3D; const anAngle: Double ): TRTMatrix;

// ----------------------------------------------------------------------------

type
  TQuadricCoefficients = Array[0..2] of Double;
  TQuadricRoots = Array[0..1] of Double;
  TCubicCoefficients = Array[0..3] of Double;
  TCubicRoots = Array[0..2] of Double;
  TQuarticCoefficients = Array[0..4] of Double;
  TQuarticRoots = Array[0..3] of Double;

function Max( const aValue1, aValue2, aValue3: Double ): Double; overload;
function Min( const aValue1, aValue2, aValue3: Double ): Double; overload;

procedure Swap( var aValue1, aValue2: Single ); overload;
procedure Swap( var aValue1, aValue2: Integer ); overload;

function Clamp( const aValue, aMin, aMax: Integer ): Integer;

function SolveQuadric( const aCoefficients: TQuadricCoefficients; out aRoots: TQuadricRoots ): Integer;
function SolveCubic( const aCoefficients: TCubicCoefficients; out aRoots: TCubicRoots ): Integer;
function SolveQuartic( const aCoefficients: TQuarticCoefficients; out aRoots: TQuarticRoots ): Integer;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  Math, SysUtils;

var
  hackTop: Boolean;

function VectorAdd( const aV1, aV2: TRTVector3D ): TRTVector3D;
begin
  result.X := aV1.X + aV2.X;
  result.Y := aV1.Y + aV2.Y;
  result.Z := aV1.Z + aV2.Z;
end;

// ----------------------------------------------------------------------------

function VectorAdd( const aV1, aV2, aV3: TRTVector3D ): TRTVector3D; overload;
begin
  result.X := aV1.X + aV2.X + aV3.X;
  result.Y := aV1.Y + aV2.Y + aV3.Y;
  result.Z := aV1.Z + aV2.Z + aV3.Z;
end;

// ----------------------------------------------------------------------------

function VectorAdd( const aV: TRTVector3D; const aScalar: Double ): TRTVector3D;
begin
  result.X := aV.X + aScalar;
  result.Y := aV.Y + aScalar;
  result.Z := aV.Z + aScalar;
end;

// ----------------------------------------------------------------------------

function VectorSubtract( const aV1, aV2: TRTVector3D ): TRTVector3D;
begin
  result.X := aV1.X - aV2.X;
  result.Y := aV1.Y - aV2.Y;
  result.Z := aV1.Z - aV2.Z;
end;

// ----------------------------------------------------------------------------

function VectorSubtract( const aV: TRTVector3D; const aScalar: Double ): TRTVector3D;
begin
  result.X := aV.X - aScalar;
  result.Y := aV.Y - aScalar;
  result.Z := aV.Z - aScalar;
end;

// ----------------------------------------------------------------------------

function VectorMultiply( const aV: TRTVector3D; const aScalar: Double ): TRTVector3D;
begin
  result.X := aV.X * aScalar;
  result.Y := aV.Y * aScalar;
  result.Z := aV.Z * aScalar;
end;

// ----------------------------------------------------------------------------

function VectorDivide( const aV: TRTVector3D; const aScalar: Double ): TRTVector3D;
var
  divisor: Double;

begin
  divisor := 1 / aScalar;

  result.X := aV.X * divisor;
  result.Y := aV.Y * divisor;
  result.Z := aV.Z * divisor;
end;

// ----------------------------------------------------------------------------

function VectorLength( const aV: TRTVector3D ): Double;
begin
  result := Sqrt( VectorSquareLength( aV ));
end;

// ----------------------------------------------------------------------------

function VectorSquareLength( const aV: TRTVector3D ): Double;
begin
  result := aV.X * aV.X + aV.Y * aV.Y + aV.Z * aV.Z;
end;

// ----------------------------------------------------------------------------

function VectorDotProduct( const aV1, aV2: TRTVector3D ): Double;
begin
  result := aV1.X * aV2.X + aV1.Y * aV2.Y + aV1.Z * aV2.Z;
end;

// ----------------------------------------------------------------------------

function VectorCrossProduct( const aV1, aV2: TRTVector3D ): TRTVector3D;
begin
  result.X := aV1.Y * aV2.Z - aV2.Y * aV1.Z;
  result.Y := aV1.Z * aV2.X - aV2.Z * aV1.X;
  result.Z := aV1.X * aV2.Y - aV2.X * aV1.Y;
end;

// ----------------------------------------------------------------------------

procedure VectorNormalize( var aV: TRTVector3D );
var
  divisor: Double;
  length: Double;

begin
  length := VectorLength( aV );
  if length <> 0 then
  begin
      divisor := 1 / VectorLength( aV );
  end
  else
      divisor := 1;

  aV.X := aV.X * divisor;
  aV.Y := aV.Y * divisor;
  aV.Z := aV.Z * divisor;
end;

// ----------------------------------------------------------------------------

function VectorHat( const aV: TRTVector3D ): TRTVector3D;
var
  divisor: Double;

begin
  divisor := 1 / VectorLength( aV );

  result.X := aV.X * divisor;
  result.Y := aV.Y * divisor;
  result.Z := aV.Z * divisor;
end;

// ----------------------------------------------------------------------------

function VectorDistance( const aV1, aV2: TRTVector3D ): Double;
begin
  result := Sqrt( VectorSquareDistance( aV1, aV2 ));
end;

// ----------------------------------------------------------------------------

function VectorSquareDistance( const aV1, aV2: TRTVector3D ): Double;
begin
  result := VectorSquareLength( VectorSubtract( aV1, aV2 ));
end;

// ----------------------------------------------------------------------------

function VectorMin( const aV1, aV2: TRTVector3D ): TRTVector3D;
begin
  result.X := Min( aV1.X, aV2.X );
  result.Y := Min( aV1.Y, aV2.Y );
  result.Z := Min( aV1.Z, aV2.Z );
end;

// ----------------------------------------------------------------------------

function VectorMax( const aV1, aV2: TRTVector3D ): TRTVector3D;
begin
  result.X := Max( aV1.X, aV2.X );
  result.Y := Max( aV1.Y, aV2.Y );
  result.Z := Max( aV1.Z, aV2.Z );
end;

// ----------------------------------------------------------------------------

function VectorMaxAxis( const aV: TRTVector3D ): TRTAxis3D;
begin
  if ( aV.X >= aV.Y ) and ( aV.X >= aV.Z ) then
  begin
      result := rtA3X;
  end
  else if ( aV.Y >= aV.X ) and ( aV.Y >= aV.Z ) then
  begin
      result := rtA3Y;
  end
  else
      result := rtA3Z;
end;

// ----------------------------------------------------------------------------

function VectorCentroid( const aVectors: Array of TRTVector3D ): TRTVector3D;
var
  i: Integer;
  
begin
  result := Vector( 0 );

  for i := Low( aVectors ) to High( aVectors )  do
  begin
      result := VectorAdd( result, aVectors[i] );
  end;

  result := VectorMultiply( result, 1 / Length( aVectors ));
end;

// ----------------------------------------------------------------------------

function VectorEquals( const aV1, aV2: TRTVector3D ): Boolean;
begin
  result := ( aV1.X = aV2.X ) and ( aV1.Y = aV2.Y ) and ( aV1.Z = aV2.Z );
end;

// ----------------------------------------------------------------------------

function VectorEquals( const aV1: TRTVector3D; const aValue: Double ): Boolean;
begin
  result := ( aV1.X = aValue ) and ( aV1.Y = aValue ) and ( aV1.Z = aValue );
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

function PointMultiply( const aP1: TRTPoint2D; const aScalar: Double ): TRTPoint2D;
begin
  result.X := aP1.X * aScalar;
  result.Y := aP1.Y * aScalar;
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

function ColorAdd( const aC1, aC2: TRGBColor ): TRGBColor;
begin
  result.R := aC1.R + aC2.R;
  result.G := aC1.G + aC2.G;
  result.B := aC1.B + aC2.B;
end;

// ----------------------------------------------------------------------------

function ColorMultiply( const aC: TRGBColor; const aScalar: Double ): TRGBColor;
begin
  result.R := aC.R * aScalar;
  result.G := aC.G * aScalar;
  result.B := aC.B * aScalar;
end;

// ----------------------------------------------------------------------------

function ColorMultiply( const aC1, aC2: TRGBColor ): TRGBColor;
begin
  result.R := aC1.R * aC2.R;
  result.G := aC1.G * aC2.G;
  result.B := aC1.B * aC2.B;
end;

// ----------------------------------------------------------------------------

function ColorDivide( const aC: TRGBColor; const aScalar: Double ): TRGBColor;
begin
  result.R := aC.R / aScalar;
  result.G := aC.G / aScalar;
  result.B := aC.B / aScalar;
end;

// ----------------------------------------------------------------------------

function ColorPower( const aC: TRGBColor; const aPower: Double ): TRGBColor;
begin
  result.R := Power( aC.R, aPower );
  result.G := Power( aC.G, aPower );
  result.B := Power( aC.B, aPower );
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

function MatrixMultiply( const aMatrix: TRTMatrix; const aScalar: Double ): TRTMatrix;
var
  x, y: Integer;

begin
  for y := 0 to 3 do
  begin
      for x := 0 to 3 do
      begin
          result.Elements[y, x] := aMatrix.Elements[y, x] * aScalar;
      end;
  end;
end;

// ----------------------------------------------------------------------------

function MatrixMultiply( const aMatrix1, aMatrix2: TRTMatrix ): TRTMatrix;
var
  x, y, i: Integer;
  value: Double;
  
begin
  for y := 0 to 3 do
  begin
      for x := 0 to 3 do
      begin
          value := 0;

          for i := 0 to 3 do
          begin     
              value := value + aMatrix1.Elements[y, i] * aMatrix2.Elements[i, x];
          end;

          result.Elements[y, x] := value;
      end;
  end;
end;

// ----------------------------------------------------------------------------

function MatrixMultiplyVector( const aMatrix: TRTMatrix; const aVector: TRTVector3D ): TRTVector3D;
begin
  result.X := aMatrix.Elements[0, 0] * aVector.X + aMatrix.Elements[0, 1] * aVector.Y + aMatrix.Elements[0, 2] * aVector.Z;
  result.Y := aMatrix.Elements[1, 0] * aVector.X + aMatrix.Elements[1, 1] * aVector.Y + aMatrix.Elements[1, 2] * aVector.Z;
  result.Z := aMatrix.Elements[2, 0] * aVector.X + aMatrix.Elements[2, 1] * aVector.Y + aMatrix.Elements[2, 2] * aVector.Z;
end;

// ----------------------------------------------------------------------------

function MatrixMultiplyPoint( const aMatrix: TRTMatrix; const aPoint: TRTPoint3D ): TRTPoint3D; 
begin
  result.X := aMatrix.Elements[0, 0] * aPoint.X + aMatrix.Elements[0, 1] * aPoint.Y + aMatrix.Elements[0, 2] * aPoint.Z + aMatrix.Elements[0, 3];
  result.Y := aMatrix.Elements[1, 0] * aPoint.X + aMatrix.Elements[1, 1] * aPoint.Y + aMatrix.Elements[1, 2] * aPoint.Z + aMatrix.Elements[1, 3];
  result.Z := aMatrix.Elements[2, 0] * aPoint.X + aMatrix.Elements[2, 1] * aPoint.Y + aMatrix.Elements[2, 2] * aPoint.Z + aMatrix.Elements[2, 3];
end;

// ----------------------------------------------------------------------------

// INFO, PJ: Transform using the _transposed_ matrix. Make sure to normalize afterwards at the appropriate time. 
function MatrixMultiplyNormal( const aMatrix: TRTMatrix; const aNormal: TRTNormal ): TRTVector3D;
begin
  result.X := aMatrix.Elements[0, 0] * aNormal.X + aMatrix.Elements[1, 0] * aNormal.Y + aMatrix.Elements[2, 0] * aNormal.Z;
  result.Y := aMatrix.Elements[0, 1] * aNormal.X + aMatrix.Elements[1, 1] * aNormal.Y + aMatrix.Elements[2, 1] * aNormal.Z;
  result.Z := aMatrix.Elements[0, 2] * aNormal.X + aMatrix.Elements[1, 2] * aNormal.Y + aMatrix.Elements[2, 2] * aNormal.Z;
end;

// ----------------------------------------------------------------------------

function MatrixDivide( const aMatrix: TRTMatrix; const aScalar: Double ): TRTMatrix;
begin
  result := MatrixMultiply( aMatrix, 1 / aScalar );
end;

// ----------------------------------------------------------------------------

function MatrixForAxisRotation( const anAxis: TRTVector3D; const anAngle: Double ): TRTMatrix;
var
  c, s: Double;

begin
  s := Sin( DegToRad( aNAngle ));
  c := Cos( DegToRad( anAngle ));

  result.Elements[0, 0] := c + ( 1 - c ) * anAxis.X * anAxis.X;
  result.Elements[0, 1] := ( 1 - c ) * anAxis.X * anAxis.Y - s * anAxis.Z;
  result.Elements[0, 2] := ( 1 - c ) * anAxis.X * anAxis.Z + s * anAxis.Y;
  result.Elements[1, 0] := ( 1 - c ) * anAxis.X * anAxis.Y + s * anAxis.Z;
  result.Elements[1, 1] := c + ( 1 - c ) * anAxis.Y * anAxis.Y;
  result.Elements[1, 2] := ( 1 - c ) * anAxis.Y * anAxis.Z - s * anAxis.X;
  result.Elements[2, 0] := ( 1 - c ) * anAxis.X * anAxis.Z - s * anAxis.Y;
  result.Elements[2, 1] := ( 1 - c ) * anAxis.Y * anAxis.Z + s * anAxis.X;
  result.Elements[2, 2] := c + ( 1 - c ) * anAxis.Z * anAxis.Z;
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

function Max( const aValue1, aValue2, aValue3: Double ): Double;
begin
  if aValue1 > aValue2 then
  begin
      result := aValue1;
  end
  else
      result := aValue2;

  if aValue3 > result then
  begin
      result := aValue3;
  end;
end;

// ----------------------------------------------------------------------------

function Min( const aValue1, aValue2, aValue3: Double ): Double;
begin
  if aValue1 < aValue2 then
  begin
      result := aValue1;
  end
  else
      result := aValue2;

  if aValue3 < result then
  begin
      result := aValue3;
  end;
end;

// ----------------------------------------------------------------------------

procedure Swap( var aValue1, aValue2: Single );
var
  temp: Single;

begin
  temp := aValue1;
  aValue1 := aValue2;
  aValue2 := temp;
end;

// ----------------------------------------------------------------------------

procedure Swap( var aValue1, aValue2: Integer );
var
  temp: Integer;

begin
  temp := aValue1;
  aValue1 := aValue2;
  aValue2 := temp;
end;

// ----------------------------------------------------------------------------

function Clamp( const aValue, aMin, aMax: Integer ): Integer;
begin
  if aValue < aMin then result := aMin
  else if aValue > aMax then result := aMax
  else result := aValue;
end;

// ----------------------------------------------------------------------------

function CubicRoot( const aValue: Double ): Double;
begin
  if aValue > 0 then
  begin
      result := Power( aValue, 1 / 3 );
  end
  else if aValue < 0 then
  begin
      result := -Power( -aValue, 1 / 3 );
  end
  else
      result := 0;
end;

// ----------------------------------------------------------------------------

function IsZero( const aValue: Double ): Boolean; 
const
  EPSILON: Double = 1E-8;

begin
   Result := Abs( aValue ) <= EPSILON;
end;

// ----------------------------------------------------------------------------

// REFA, PJ: Fix identation
function SolveQuadric( const aCoefficients: TQuadricCoefficients; out aRoots: TQuadricRoots ): Integer;
var
   p, q, D, sqrt_D : Double;

begin
  if aCoefficients[2] = 0 then
  begin
      result := 0;
      exit;
  end;
  
  p := aCoefficients[1]/(2*aCoefficients[2]);
  q := aCoefficients[0]/aCoefficients[2];

  D := Sqr(p)-q;

  if IsZero(D) then
  begin
    Result := 1;
    aRoots[0]:=-p;
  end
  else if D>0 then
  begin
    sqrt_D:=Sqrt(D);
    Result := 2;
    aRoots[0]:=sqrt_D-p;
    aRoots[1]:=-sqrt_D-p;
  end
  else
  begin
    // if (D < 0)
    Result := 0;
  end;
end;

// ----------------------------------------------------------------------------

// REFA, PJ: Fix identation
function SolveCubic( const aCoefficients: TCubicCoefficients; out aRoots: TCubicRoots ): Integer;
var
   i : Integer;
   sub : Double;
   A, B, Cc : Double;
   sq_A, p, q : Double;
   cb_p, D : Double;
   u, v, phi, t, sqrt_D, invC3 : Double;
   quadricCoefficients: TQuadricCoefficients;
   quadricRoots: TQuadricRoots;

begin
   if IsZero(aCoefficients[3]) then
   begin
      Move( aCoefficients[0], quadricCoefficients, SizeOf( quadricCoefficients ));
      Result := SolveQuadric( quadricCoefficients, quadricRoots );
      Move( quadricRoots[0], aRoots[0], SizeOf( quadricRoots ));

      Exit;
   end;

   // normal form: x^3 + Ax^2 + Bx + C = 0

  invC3:=1/aCoefficients[3];
  A := aCoefficients[2]*invC3;
  B := aCoefficients[1]*invC3;
  Cc:= aCoefficients[0]*invC3;

  // substitute x = y - A/3 to eliminate quadric term:
  // x^3 +px + q = 0

  sq_A := Sqr(A);
  p := ( 1 / 3 ) * (B-( 1 / 3 ) *sq_A);
  q := ( 1 / 2 ) * (2.0/27*A*sq_A-( 1 / 3 ) *A*B+Cc);

  // use Cardano's formula

  cb_p := Sqr(p)*p;
  D := Sqr(q)+cb_p;

  if IsZero(D) then
  begin
      if IsZero(q) then
      begin // one triple solution
          Result := 1;
	        aRoots[0]:=0;

      end
      else
      begin // one single and one double solution
          u := CubicRoot(-q);
          Result := 2;
          aRoots[0]:=2*u;
          aRoots[1]:=-u;
      end;
  end
  else if D<0 then
  begin // Casus irreducibilis: three real solutions
      phi:=( 1 / 3 ) *ArcCos(-q * (1 / Sqrt(-cb_p)));
      t:=2*Sqrt(-p);
      Result := 3;
      aRoots[0]:= t*Cos(phi);
      aRoots[1]:=-t*Cos(phi+PI/3);
      aRoots[2]:=-t*Cos(phi-PI/3);
  end
  else
  begin // one real solution

      sqrt_D:=Sqrt(D);
      u:=CubicRoot(sqrt_D-q);
      v:=-CubicRoot(sqrt_D+q);
      Result := 1;
      aRoots[0]:= u + v;
  end;

   // resubstitute

  sub := ( 1 / 3 ) * A;

  for i := 0 to Result - 1 do
  begin
      aRoots[i]:= aRoots[i] - sub;
  end;
end;

// ----------------------------------------------------------------------------

function SolveQuartic( const aCoefficients: TQuarticCoefficients; out aRoots: TQuarticRoots ): Integer;
var
  cubicCoefficients: TCubicCoefficients;
  cubicRoots: TCubicRoots;
  quadricCoefficients: TQuadricCoefficients;
  quadricRoots: TQuadricRoots;
  a, aSqr, b, c, d, p, q, r, z, u, v, sub: Double;
  n, i: Integer;

begin
  if hackTop or IsZero(aCoefficients[4]) then begin
      Move( aCoefficients[0], cubicCoefficients, SizeOf( cubicCoefficients ));
      Result := SolveCubic( cubicCoefficients, cubicRoots );
      Move( cubicRoots[0], aRoots[0], SizeOf( aRoots ));

      Exit;
  end;

  a := aCoefficients[3] / aCoefficients[4];
  b := aCoefficients[2] / aCoefficients[4];
  c := aCoefficients[1] / aCoefficients[4];
  d := aCoefficients[0] / aCoefficients[4];

  aSqr := a * a;
  p := -3/8 * aSqr + b;
  q := 1/8 * aSqr * a - 1/2 * a * b + c;
  r := -3/256 * aSqr * aSqr + 1/16 * aSqr * b - 1/4 * a * c + d;

  if IsZero( r ) then
  begin
      cubicCoefficients[0] := q;
      cubicCoefficients[1] := p;
      cubicCoefficients[2] := 0;
      cubicCoefficients[3] := 1;

      Result := SolveCubic( cubicCoefficients, cubicRoots );
      Move( cubicRoots[0], aRoots[0], Result * SizeOf(Double));
      aRoots[3] := 0;

  end else
  begin
		cubicCoefficients[ 0 ] := 1.0/2 * r * p - 1.0/8 * q * q;
		cubicCoefficients[ 1 ] := - r;
		cubicCoefficients[ 2 ] := - 1.0/2 * p;
		cubicCoefficients[ 3 ] := 1;

		result := SolveCubic( cubicCoefficients, cubicRoots );

		// ... and take the one real solution ...

      Assert( result > 0 );
		z := cubicRoots[0];

		// ... to build two quadric equations

		u := z * z - r;
		v := 2 * z - p;

		if (IsZero(u)) then
		    u := 0
		else if (u > 0) then
		    u := sqrt(u)
		else
      begin
          result := 0;
          exit;
      end;


		if (IsZero(v)) then
      begin
		    v := 0;
      end
		else if (v > 0) then
      begin
		    v := sqrt(v)
      end
		else
      begin
		    result := 0;
          exit;
      end;


      if q < 0 then
      begin
          v := -v;
      end;

		quadricCoefficients[ 0 ] := z - u;
      quadricCoefficients[ 1 ] := v;
		quadricCoefficients[ 2 ] := 1;

		result := SolveQuadric(quadricCoefficients, quadricRoots);
      Move( quadricRoots, aRoots[0], result * SizeOf( double ));

		quadricCoefficients[ 0 ] := z + u;
      quadricCoefficients[ 1 ] := -v;
		quadricCoefficients[ 2 ] := 1;

      n := SolveQuadric(quadricCoefficients, quadricRoots );
      Move( quadricRoots, aRoots[result], n * SizeOf( double ));

      result := result + n;
  end;

  sub := 1/4 * A;
  for i := 0 to result - 1 do
  begin
      aRoots[i] := aRoots[i] - sub;
  end;
end;

end.
