{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTMeshCl;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  After constructing the mesh, add the mesh.Grid to your scene.
  ---------------------------------------------------------------------------- }

interface

uses
  RTVector3DListCl, RTCompoundObjectCl, RTMiscTp, RTMaterialCl;

type
  TRTMesh = class( TObject )
  private
      fVertices: TRTVector3DList;
      fMaterial: TRTMaterial;
      fCompoundObject: TRTCompoundObject;
      fOwnsCompound: Boolean;

      function GetVertex( const anIndex: Integer ): PRTPoint3D;

  public
      constructor Create( const aCompoundObject: TRTCompoundObject = nil );
      destructor Destroy; override;

      procedure AddVertex( const aVertex: TRTPoint3D );

      procedure AddTriangle( const anIndex1, anIndex2, anIndex3: Integer; const aMaterial: TRTMaterial ); overload;
      procedure AddTriangle( const anIndex1, anIndex2, anIndex3: Integer; const aNormal: TRTNormal ); overload;
      procedure AddTriangle( const anIndex1, anIndex2, anIndex3: Integer; const aNormal1, aNormal2, aNormal3: TRTNormal; const aMaterial: TRTMaterial ); overload;
      procedure AddTriangle( const aVertex1, aVertex2, aVertex3: TRTPoint3D; const aNormal1, aNormal2, aNormal3: TRTNormal; const aMaterial: TRTMaterial ); overload;
      procedure AddTriangle(
          const aVertex1, aVertex2, aVertex3: TRTPoint3D;
          const aNormal1, aNormal2, aNormal3: TRTNormal;
          const aUVCoordinate1, aUVCoordinate2, aUVCoordinate3: TRTUVCoordinate;
          const aMaterial: TRTMaterial ); overload;

      property Vertices[const anIndex: Integer]: PRTPoint3D read GetVertex;
      property Material: TRTMaterial read fMaterial write fMaterial;
      property CompoundObject: TRTCompoundObject read fCompoundObject;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils,
  RTMathFn, RTMeshTriangleCl, RTGridCl;

constructor TRTMesh.Create( const aCompoundObject: TRTCompoundObject = nil );
begin
  inherited Create;

  fVertices := TRTVector3DList.Create;
  fCompoundObject := aCompoundObject;

  if fCompoundObject = nil then
  begin
      fOwnsCompound := TRUE;
      fCompoundObject := TRTGrid.Create;
  end;
end;

// ----------------------------------------------------------------------------

destructor TRTMesh.Destroy;
begin
  if fOwnsCompound then
  begin
      FreeAndNil( fCompoundObject );
  end;
  
  FreeAndNil( fMaterial );
  FreeAndNil( fVertices );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

function TRTMesh.GetVertex( const anIndex: Integer ): PRTPoint3D;
begin
  result := fVertices[anIndex];
end;

// ----------------------------------------------------------------------------

procedure TRTMesh.AddVertex( const aVertex: TRTPoint3D );
begin
  fVertices.Add( aVertex );
end;

// ----------------------------------------------------------------------------

procedure TRTMesh.AddTriangle( const anIndex1, anIndex2, anIndex3: Integer; const aMaterial: TRTMaterial );
var
  triangle: TRTMeshTriangle;

begin
  triangle := TRTMeshTriangle.Create( self, anIndex1, anIndex2, anIndex3, rtTTFlat );
  triangle.Material := aMaterial;
  fCompoundObject.AddObject( triangle );
end;

// ----------------------------------------------------------------------------

procedure TRTMesh.AddTriangle( const anIndex1, anIndex2, anIndex3: Integer; const aNormal: TRTNormal );
var
  triangle: TRTMeshTriangle;

begin
  triangle := TRTMeshTriangle.Create( self, anIndex1, anIndex2, anIndex3, rtTTFlat );
  triangle.Normal := aNormal;
  fCompoundObject.AddObject( triangle );
end;

// ----------------------------------------------------------------------------

procedure TRTMesh.AddTriangle( const anIndex1, anIndex2, anIndex3: Integer; const aNormal1, aNormal2, aNormal3: TRTNormal; const aMaterial: TRTMaterial );
var
  triangle: TRTMeshTriangle;

begin
  triangle := TRTMeshTriangle.Create( self, anIndex1, anIndex2, anIndex3, rtTTSmooth );
  triangle.Normals[0] := aNormal1;
  triangle.Normals[1] := aNormal2;
  triangle.Normals[2] := aNormal3;
  triangle.Material := aMaterial;
  fCompoundObject.AddObject( triangle );
end;

// ----------------------------------------------------------------------------

procedure TRTMesh.AddTriangle( const aVertex1, aVertex2, aVertex3: TRTPoint3D; const aNormal1, aNormal2, aNormal3: TRTNormal; const aMaterial: TRTMaterial );
var
  triangle: TRTMeshTriangle;

begin
  triangle := TRTMeshTriangle.Create( self, aVertex1, aVertex2, aVertex3, rtTTSmooth );
  triangle.Normals[0] := aNormal1;
  triangle.Normals[1] := aNormal2;
  triangle.Normals[2] := aNormal3;
  triangle.Material := aMaterial;
  fCompoundObject.AddObject( triangle );
end;

// ----------------------------------------------------------------------------

procedure TRTMesh.AddTriangle(
    const aVertex1, aVertex2, aVertex3: TRTPoint3D;
    const aNormal1, aNormal2, aNormal3: TRTNormal;
    const aUVCoordinate1, aUVCoordinate2, aUVCoordinate3: TRTUVCoordinate;
    const aMaterial: TRTMaterial );
var
  triangle: TRTMeshTriangle;
  normal: TRTVector3D;
begin
  // JCC: added some code to check for degenerate triangles (just average the vertices
  // of a polygon to get degenerate triangles)
  // NOTE: this probably needs to be added to the overloaded AddTriangles too someday
  // but as they aren't used now, I decided to leave them as they are
  normal := VectorCrossProduct( VectorSubtract( aVertex2, aVertex1 ), VectorSubtract( aVertex3, aVertex1 ));
  VectorNormalize( normal );
  if (normal.X=0) and (normal.Y=0) and (normal.Z=0) then

  else begin
      triangle := TRTMeshTriangle.Create( self, aVertex1, aVertex2, aVertex3, rtTTSmooth );
      triangle.Normals[0] := aNormal1;
      triangle.Normals[1] := aNormal2;
      triangle.Normals[2] := aNormal3;
      triangle.UVCoordinates[0] := aUVCoordinate1;
      triangle.UVCoordinates[1] := aUVCoordinate2;
      triangle.UVCoordinates[2] := aUVCoordinate3;
      triangle.Material := aMaterial;
      fCompoundObject.AddObject( triangle );
  end;
end;


end.
 
