{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTMiscTp;

{$MODE Delphi}

interface

type
  TRTAxis3D = ( rtA3X, rtA3Y, rtA3Z );

  TRTVector3D = record
      case Boolean of
          FALSE: (
              X: Double;
              Y: Double;
              Z: Double;
          );
          TRUE: (
              Elements: Array[TRTAxis3D] of Double;
          );
  end;
  PRTVector3D = ^TRTVector3D;

  // INFO, PJ: These aliases improve code readability 
  TRTPoint3D = TRTVector3D;
  PRTPoint3D = PRTVector3D;
  TRTNormal = TRTPoint3D;

  TRTPoint2D = record
      case Boolean of
      FALSE:  ( X: Double; Y: Double );
      TRUE:   ( Elements: Array[0..1] of Double );
  end;
  PRTPoint2D = ^TRTPoint2D;

  TRTUVCoordinate = record
      U, V: Double;
  end;

  TRTMatrix = record
      Elements: Array[0..3, 0..3] of Double;
  end;

  TRGBColor = record
      R, G, B: Single;
  end;

  TRTRay = record
      Origin: TRTPoint3D;
      Direction: TRTVector3D;

      // INFO, PJ: These fields are part of the optimization of the ray-tree intersection
      invDirection: TRTVector3D;
      signed: Array[TRTAxis3D] of Boolean;
  end;

  TRTShadeRec = record
      HitAnObject: Boolean;           // Did the ray hit an object?
      HitPoint: TRTPoint3D;           // World coordinates of hit point
      LocalHitPoint: TRTPoint3D;      // World coordinates of hit point (untransformed)
      Normal: TRTNormal;              // Normal at hint point
      Color: TRGBColor;               // REFA, PJ: Remove (no longer needed)
      World: TObject;                 // World reference for shading (REFA, PJ: generic object to avoid circular reference)
      Material: TObject;              // Nearest object's material (REFA, PJ: generic object to avoid circular reference)
      Ray: TRTRay;                    // Used for specular highlights
      Depth: Integer;                 // Recursion Depth
      Direction: TRTVector3D;         // For area lights
      T: Double;                      // Hit distance.
      UVCoordinate: TRTUVCoordinate;  // UV Coordinate at hit point
  end;

  TRTToneMapping = ( tmNone, tmShowOutOfGamut, tmMaxToOne );

// INFO, PJ: "constructor" functions (we threat these records as light-weight objects)
function Vector( const anX, aY, aZ: Double ): TRTVector3D; overload;
function Vector( const aValue: Double ): TRTVector3D; overload;
function Point( const anX, aY: Double ): TRTPoint2D;
function MatrixIdentity: TRTMatrix;
function Matrix( const a11, a12, a13, a14, a21, a22, a23, a24, a31, a32, a33, a34, a41, a42, a43, a44: Double ): TRTMatrix;
function Ray( const anOrigin: TRTPoint3D; const aDirection: TRTVector3D ): TRTRay;
function RGBColor( const aC: Single ): TRGBColor; overload;
function RGBColor( const anR, aG, aB: Single ): TRGBColor; overload;
function ShadeRec( const aWorld: TObject ): TRTShadeRec;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

function Vector( const anX, aY, aZ: Double ): TRTVector3D;
begin
  result.X := anX;
  result.Y := aY;
  result.Z := aZ;
end;

// ----------------------------------------------------------------------------

function Vector( const aValue: Double ): TRTVector3D;
begin
  result.X := aValue;
  result.Y := aValue;
  result.Z := aValue;
end;

// ----------------------------------------------------------------------------

function Point( const anX, aY: Double ): TRTPoint2D;
begin
  result.X := anX;
  result.Y := aY;
end;

// ----------------------------------------------------------------------------

function MatrixIdentity: TRTMatrix;
begin
  result.Elements[0, 0] := 1; result.Elements[0, 1] := 0; result.Elements[0, 2] := 0; result.Elements[0, 3] := 0;
  result.Elements[1, 0] := 0; result.Elements[1, 1] := 1; result.Elements[1, 2] := 0; result.Elements[1, 3] := 0;
  result.Elements[2, 0] := 0; result.Elements[2, 1] := 0; result.Elements[2, 2] := 1; result.Elements[2, 3] := 0;
  result.Elements[3, 0] := 0; result.Elements[3, 1] := 0; result.Elements[3, 2] := 0; result.Elements[3, 3] := 1;
end;

// ----------------------------------------------------------------------------

function Matrix( const a11, a12, a13, a14, a21, a22, a23, a24, a31, a32, a33, a34, a41, a42, a43, a44: Double ): TRTMatrix;
begin
  result.Elements[0, 0] := a11; result.Elements[0, 1] := a12; result.Elements[0, 2] := a13; result.Elements[0, 3] := a14;
  result.Elements[1, 0] := a21; result.Elements[1, 1] := a22; result.Elements[1, 2] := a23; result.Elements[1, 3] := a24;
  result.Elements[2, 0] := a31; result.Elements[2, 1] := a32; result.Elements[2, 2] := a33; result.Elements[2, 3] := a34;
  result.Elements[3, 0] := a41; result.Elements[3, 1] := a42; result.Elements[3, 2] := a43; result.Elements[3, 3] := a44;
end;

// ----------------------------------------------------------------------------

function Ray( const anOrigin: TRTPoint3D; const aDirection: TRTVector3D ): TRTRay;
begin
  result.Origin := anOrigin;
  result.Direction := aDirection;
end;

// ----------------------------------------------------------------------------

function RGBColor( const aC: Single ): TRGBColor;
begin
  result.R := aC;
  result.G := aC;
  result.B := aC;
end;

// ----------------------------------------------------------------------------

function RGBColor( const anR, aG, aB: Single ): TRGBColor;
begin
  result.R := anR;
  result.G := aG;
  result.B := aB;
end;

// ----------------------------------------------------------------------------

function ShadeRec( const aWorld: TObject ): TRTShadeRec;
begin
  FillChar( result, SizeOf( result ), 0 );
  result.World := aWorld;
end;



end.

