{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit RTWorldCl;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  Slightly abstract world base class.
  Implement own child class that implements
  - Initialization of the render window (OpenWindow( ... ));
  - actual rendering ( RenderPixel( ... )).
  ---------------------------------------------------------------------------- }

interface

uses
  Contnrs,
  RTGeometricObjectCl, RTViewPlaneCl, RTMiscTp, RTSphereCl, RTTracerCl, RTLightCl;

type
  TRTWorld = class( TObject )
  private
      fViewPlane: TRTViewPlane;
      fBackgroundColor: TRGBColor;
      fAmbientLight: TRTLight;
      fTracer: TRTTracer;
      fObjects: TObjectList;
      fLights: TObjectList;
      fObjectPool: TObjectList;

      function ClampToColor( const aColor: TRGBColor ): TRGBColor;
      function MaxToOne( const aColor: TRGBColor ): TRGBColor;

      function GetObject( const anIndex: Integer ): TRTGeometricObject;
      function GetLight( const anIndex: Integer ): TRTLight;

      procedure SetAmbientLight( const aLight: TRTLight );

  protected
      procedure RenderPixel( const aRow, aColumn: Integer; const aColor: TRGBColor ); virtual; abstract;
      procedure RegisterObject( const anObject: TObject );

  public
      constructor Create;
      destructor Destroy; override;

      procedure AddObject( const anObject: TRTGeometricObject );
      procedure AddLight( const aLight: TRTLight );

      function NumObjects: Integer;
      function NumLights: Integer;

      procedure DisplayPixel( const aRow, aColumn: Integer; const aColor: TRGBColor );
      procedure OpenWindow( const aHRes, aVRes: Integer ); virtual;

      function HitBareBonesObject( const aRay: TRTRay ): TRTShadeRec; // REFA, PJ: Remove?
      function HitObjects( const aRay: TRTRay ): TRTShadeRec;

      property ViewPlane: TRTViewPlane read fViewPlane;
      property Tracer: TRTTracer read fTracer write fTracer;
      property BackgroundColor: TRGBColor read fBackgroundColor write fBackgroundColor;
      property AmbientLight: TRTLight read fAmbientLight write SetAmbientLight;

      property Objects[const anIndex: Integer]: TRTGeometricObject read GetObject;
      property Lights[const anIndex: Integer]: TRTLight read GetLight;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  Forms, Math, SysUtils,
  RTMathFn, RTMiscCn, RTPlaneCl, RTAmbientLightCl, 
  RTSingleSphereTracerCl, RTMultipleObjectsTracerCl, RT2DFunctionTracerCl;

constructor TRTWorld.Create;
begin
  inherited Create;

  fObjects := TObjectList.Create( {$IFDEF RAYTRACER_DEBUG}FALSE{$ELSE}TRUE{$ENDIF} );
  fLights := TObjectList.Create( TRUE );

  fViewPlane := TRTViewPlane.Create;
  fAmbientLight := TRTAmbientLight.Create;

  // INFO, PJ: Poor mans release pool
  fObjectPool := TObjectList.Create( TRUE );
end;

// ----------------------------------------------------------------------------

destructor TRTWorld.Destroy;
begin
  FreeAndNil( fAmbientLight );
  FreeAndNil( fViewPlane );
  FreeAndNil( fAmbientLight );
  FreeAndNil( fTracer );
  FreeAndNil( fLights );
  FreeAndNil( fObjects );
  FreeAndNil( fObjectPool );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

function TRTWorld.ClampToColor( const aColor: TRGBColor ): TRGBColor;
begin
  result := aColor;

  // INFO, PJ: Clamps color to red if any of the colors exceeds 1
  if ( aColor.R > 1 ) or ( aColor.G > 1 ) or ( aColor.B > 1 ) then
  begin
      result.R := 1;
      result.G := 0;
      result.B := 0;
  end;
end;

// ----------------------------------------------------------------------------

function TRTWorld.MaxToOne( const aColor: TRGBColor ): TRGBColor;
var
  maxValue: Single;

begin
  maxValue := Max( aColor.R, aColor.G, aColor.B );

  if maxValue > 1 then
  begin                           
      result := ColorDivide( aColor, maxValue );
  end
  else
      result := aColor;
end;

// ----------------------------------------------------------------------------

procedure TRTWorld.RegisterObject( const anObject: TObject );
begin
  fObjectPool.Add( anObject );
end;

// ----------------------------------------------------------------------------

function TRTWorld.GetObject( const anIndex: Integer ): TRTGeometricObject;
begin
  result := fObjects[anIndex] as TRTGeometricObject;
end;

// ----------------------------------------------------------------------------

function TRTWorld.GetLight( const anIndex: Integer ): TRTLight;
begin
  result := fLights[anIndex] as TRTLight;
end;

// ----------------------------------------------------------------------------

procedure TRTWorld.SetAmbientLight( const aLight: TRTLight );
begin
  FreeAndNil( fAmbientLight );
  fAmbientLight := aLight;
end;

// ----------------------------------------------------------------------------

procedure TRTWorld.DisplayPixel( const aRow, aColumn: Integer; const aColor: TRGBColor );
var
  mappedColor: TRGBColor;

begin
  case fViewPlane.ToneMapping of
      tmShowOutOfGamut: mappedColor := ClampToColor( aColor );
      tmMaxToOne:       mappedColor := MaxToOne( aColor );
  else
      mappedColor := aColor;
  end;

  if fViewPlane.Gamma <> 1 then
  begin
      mappedColor := ColorPower( mappedColor, fViewPlane.InvGamma );
  end;

  RenderPixel( aRow, aColumn, mappedColor );
end;

// ----------------------------------------------------------------------------

procedure TRTWorld.OpenWindow( const aHRes, aVRes: Integer );
begin
    // INFO, PJ: Empty on purpose. Override in TRTRenderWorld if your
    //           render target requires specific initialization.
end;

// ----------------------------------------------------------------------------

procedure TRTWorld.AddObject( const anObject: TRTGeometricObject );
begin
  fObjects.Add( anObject );
end;

// ----------------------------------------------------------------------------

procedure TRTWorld.AddLight( const aLight: TRTLight );
begin
  fLights.Add( aLight );
end;

// ----------------------------------------------------------------------------

function TRTWorld.NumObjects: Integer;
begin
  result := fObjects.Count;
end;

// ----------------------------------------------------------------------------

function TRTWorld.NumLights: Integer;
begin
  result := fLights.Count;
end;

// ----------------------------------------------------------------------------

function TRTWorld.HitBareBonesObject( const aRay: TRTRay ): TRTShadeRec;
var
  i: Integer;
  t, tMin: Double;
  geometricObject: TRTGeometricObject;

begin
  result := ShadeRec( self );
  tMin := HUGE_VALUE;

  for i := 0 to fObjects.Count - 1 do
  begin
      geometricObject := fObjects[i] as TRTGeometricObject;
      if geometricObject.Hit( aRay, t, result ) and ( t < tMin ) then
      begin
          result.HitAnObject := TRUE;
          tMin := t;
          result.Color := geometricObject.Color;
      end;
  end;
end;

// ----------------------------------------------------------------------------

function TRTWorld.HitObjects( const aRay: TRTRay ): TRTShadeRec;
var
  i: Integer;
  geometricObject: TRTGeometricObject;
  t, tMin: Double;
  localHitPoint: TRTVector3D;
  uvCoordinate: TRTUVCoordinate;
  normal: TRTNormal;
  material: TObject;

begin
  result := ShadeRec( self );
  tMin := HUGE_VALUE;
  material := nil;

  for i := 0 to fObjects.Count - 1 do
  begin
      geometricObject := fObjects[i] as TRTGeometricObject;

      // REFA, PJ:
      //   The ".hit modifies some stuff so we store some but not all in local variables"
      //   approach isn't very nice...
      if geometricObject.Hit( aRay, t, result ) and ( t < tMin ) then
      begin
          result.HitAnObject := TRUE;
          tMin := t;
          material := result.Material;
          result.HitPoint := VectorAdd( aRay.Origin, VectorMultiply( aRay.Direction, t ));
          normal := result.Normal;
          localHitPoint := result.LocalHitPoint;
          uvCoordinate := result.UVCoordinate;
      end;
  end;

  // INFO, PJ: Some properties might be changed since we actually hit that object, we restore them.
  if result.HitAnObject then
  begin
      result.T := tMin;
      result.Normal := normal;
      result.Material := material;
      result.LocalHitPoint := localHitPoint;
      result.UVCoordinate := uvCoordinate;
  end;
end;

end.
