{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_AABB;

{$MODE Delphi}

interface

uses SysUtils,
     unit_Types;

type
  TAxisAlignedBoundingBox = class(TObject)
  private
    fAABB: array[0..31] of TVertice;
    fVertices: TVerticeList;
    fMinVertice,fMaxVertice: TVertice;
    fOwnList: boolean;
    function GetWidth: double;
    function GetHeight: double;
    function GetDepth: double;
    function GetAABB(anIndex: integer): TVertice;
    procedure SetMinMaxVertices;
    property OwnList: boolean read fOwnList write fOwnList;
  public
    constructor Create(verticeList: TVerticeList); virtual;
    destructor Destroy; override;
    procedure Update;
    property Vertices: TVerticeList read fVertices write fVertices;
    property AABB[anIndex: integer]: TVertice read GetAABB; default;
    property MinVertice: TVertice read fMinVertice;
    property MaxVertice: TVertice read fMaxVertice;
    property Width: double read GetWidth;
    property Height: double read GetHeight;
    property Depth: double read GetDepth;
  end;

implementation

// ### TAxisAlignedBoundingBox #################################################

// name: Constructor
// desc: creates the actual object
// note: if a verticelist is provided, the AABB object will point to that
//       verticelist to get its vertices from; if not, it will create its
//       own verticelist which will have to be filled with data from outside
//       this class
constructor TAxisAlignedBoundingBox.Create(verticeList: TVerticeList);
var i: integer;
begin
    for i := 0 to 31 do
        fAABB[i] := TVertice.Create(0,0,0);

    fMinVertice := TVertice.Create(0,0,0);
    fMaxVertice := TVertice.Create(0,0,0);

    if Assigned(verticeList) then begin
        fVertices := verticeList;
        OwnList := false;
    end else begin
        fVertices := TVerticeList.Create(false,true);
        OwnList := true;
    end;
    Update;
end;

destructor TAxisAlignedBoundingBox.Destroy;
var i: integer;
begin
    for i := 0 to 31 do
        FreeAndNil(fAABB[i]);
    if OwnList then FreeAndNil(fVertices);
    FreeAndNil(fMinVertice);
    FreeAndNil(fMaxVertice);
    inherited;
end;

// NOTE: might have to use Abs here
function TAxisAlignedBoundingBox.GetWidth: double;
begin
    Result := fmaxVertice.x-fminVertice.x;
end;

// NOTE: might have to use Abs here
function TAxisAlignedBoundingBox.GetHeight: double;
begin
    Result := fmaxVertice.y-fminVertice.y;
end;

// NOTE: might have to use Abs here
function TAxisAlignedBoundingBox.GetDepth: double;
begin
    Result := fmaxVertice.z-fminVertice.z;
end;

function TAxisAlignedBoundingBox.GetAABB(anIndex: integer): TVertice;
begin
    Result := fAABB[anIndex];
end;

procedure TAxisAlignedBoundingBox.SetMinMaxVertices;
var i: integer;
    tempV: TVertice;
begin
    if Vertices.Count>0 then begin
        tempV := Vertices[0];
        fMinVertice.x := tempV.x;
        fMinVertice.y := tempV.y;
        fMinVertice.z := tempV.z;
        fMaxVertice.x := tempV.x;
        fMaxVertice.y := tempV.y;
        fMaxVertice.z := tempV.z;
        for i := 1 to Vertices.Count-1 do begin
            tempV := Vertices[i];
            if tempV.x < fMinVertice.x then fMinVertice.x := tempV.x;
            if tempV.y < fMinVertice.y then fMinVertice.y := tempV.y;
            if tempV.z < fMinVertice.z then fMinVertice.z := tempV.z;
            if tempV.x > fMaxVertice.x then fMaxVertice.x := tempV.x;
            if tempV.y > fMaxVertice.y then fMaxVertice.y := tempV.y;
            if tempV.z > fMaxVertice.z then fMaxVertice.z := tempV.z;
        end;
    end else begin
        fMinVertice.x := 0;
        fMinVertice.y := 0;
        fMinVertice.z := 0;
        fMaxVertice.x := 0;
        fMaxVertice.y := 0;
        fMaxVertice.z := 0;
    end;
end;

procedure TAxisAlignedBoundingBox.Update;
var t: double;
begin
    // calculate extreme min & max vertices
    SetMinMaxVertices;

    // now set vertices of the 8 corners of the AABB
    AABB[0].x  := fMinVertice.x-2;
    AABB[0].y  := fMinVertice.y-2;
    AABB[0].z  := fMaxVertice.z+2;
    AABB[4].x  := fMinVertice.x-2;
    AABB[4].y  := fMinVertice.y-2;
    AABB[4].z  := fMinVertice.z-2;
    AABB[8].x  := fMaxVertice.x+2;
    AABB[8].y  := fMinVertice.y-2;
    AABB[8].z  := fMinVertice.z-2;
    AABB[12].x := fMaxVertice.x+2;
    AABB[12].y := fMinVertice.y-2;
    AABB[12].z := fMaxVertice.z+2;
    AABB[16].x := fMinVertice.x-2;
    AABB[16].y := fMaxVertice.y+2;
    AABB[16].z := fMaxVertice.z+2;
    AABB[20].x := fMinVertice.x-2;
    AABB[20].y := fMaxVertice.y+2;
    AABB[20].z := fMinVertice.z-2;
    AABB[24].x := fMaxVertice.x+2;
    AABB[24].y := fMaxVertice.y+2;
    AABB[24].z := fMinVertice.z-2;
    AABB[28].x := fMaxVertice.x+2;
    AABB[28].y := fMaxVertice.y+2;
    AABB[28].z := fMaxVertice.z+2;

    // and calculate the extra vertices needed to complete the corners of the AABB (3 per corner)
    t := 1/6;

    AABB[1].Assign(AABB[0].x + (AABB[12].x-AABB[0].x)*t,AABB[0].y,AABB[0].z);
    AABB[2].Assign(AABB[0].x,AABB[0].y + (AABB[16].y-AABB[0].y)*t,AABB[0].z);
    AABB[3].Assign(AABB[0].x,AABB[0].y,AABB[0].z + (AABB[4].z-AABB[0].z)*t);

    AABB[5].Assign(AABB[4].x + (AABB[8].x-AABB[4].x)*t,AABB[4].y,AABB[4].z);
    AABB[6].Assign(AABB[4].x,AABB[4].y + (AABB[20].y-AABB[4].y)*t,AABB[4].z);
    AABB[7].Assign(AABB[4].x,AABB[4].y,AABB[4].z + (AABB[0].z-AABB[4].z)*t);

    AABB[9].Assign(AABB[8].x + (AABB[4].x-AABB[8].x)*t,AABB[8].y,AABB[8].z);
    AABB[10].Assign(AABB[8].x,AABB[8].y + (AABB[24].y-AABB[8].y)*t,AABB[8].z);
    AABB[11].Assign(AABB[8].x,AABB[8].y,AABB[8].z + (AABB[12].z-AABB[8].z)*t);

    AABB[13].Assign(AABB[12].x + (AABB[0].x-AABB[12].x)*t,AABB[12].y,AABB[12].z);
    AABB[14].Assign(AABB[12].x,AABB[12].y + (AABB[28].y-AABB[12].y)*t,AABB[12].z);
    AABB[15].Assign(AABB[12].x,AABB[12].y,AABB[12].z + (AABB[8].z-AABB[12].z)*t);

    AABB[17].Assign(AABB[16].x + (AABB[28].x-AABB[16].x)*t,AABB[16].y,AABB[16].z);
    AABB[18].Assign(AABB[16].x,AABB[16].y + (AABB[0].y-AABB[16].y)*t,AABB[16].z);
    AABB[19].Assign(AABB[16].x,AABB[16].y,AABB[16].z + (AABB[20].z-AABB[16].z)*t);

    AABB[21].Assign(AABB[20].x + (AABB[24].x-AABB[20].x)*t,AABB[20].y,AABB[20].z);
    AABB[22].Assign(AABB[20].x,AABB[20].y + (AABB[4].y-AABB[20].y)*t,AABB[20].z);
    AABB[23].Assign(AABB[20].x,AABB[20].y,AABB[20].z + (AABB[16].z-AABB[20].z)*t);

    AABB[25].Assign(AABB[24].x + (AABB[20].x-AABB[24].x)*t,AABB[24].y,AABB[24].z);
    AABB[26].Assign(AABB[24].x,AABB[24].y + (AABB[8].y-AABB[24].y)*t,AABB[24].z);
    AABB[27].Assign(AABB[24].x,AABB[24].y,AABB[24].z + (AABB[28].z-AABB[24].z)*t);

    AABB[29].Assign(AABB[28].x + (AABB[16].x-AABB[28].x)*t,AABB[28].y,AABB[28].z);
    AABB[30].Assign(AABB[28].x,AABB[28].y + (AABB[12].y-AABB[28].y)*t,AABB[28].z);
    AABB[31].Assign(AABB[28].x,AABB[28].y,AABB[28].z + (AABB[24].z-AABB[28].z)*t);
end;

end.
