{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Animation;

{$MODE Delphi}

interface

uses
  Classes,
  ComCtrls,
  unit_Types,
  unit_IntegerList,
  unit_AnimationFrame;

type
  // An animation contains multiple animation frames. Each frame contains the
  // location of all joints in the skeleton to which this animation belongs.
  // At any given time, only one frame can be the selected frame.
  TAnimation = class(TObject)
  private
    fID: integer;
    fName: string;
    fFPS: integer;
    fAuthor: string;
    fSkeleton: TObject;
    fHasChanged: boolean;
    fAnimationFrames: TAnimationFrameList;
    fKeyframeIndices: TIntegerList;
    fSelectedFrame: TAnimationFrame;
    fTreeNode: TTreeNode;
    fTag: string;
    fOnChange: TNotifyEvent;
    function GetNumberOfFrames: integer;
    function GetNumberOfKeyFrames: integer;
    function GetSelectedFrame: TAnimationFrame;
    function GetSelectedFrameIndex: integer;
    function GetKeyframeIndices: TIntegerList;
    procedure SetHasChanged(AValue: boolean);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(AAnimation: TAnimation);
    procedure GotoFrame(AFrameIndex: integer);
    procedure SwapFrame(AFrameIndex: integer; ASwapWithPrevious: boolean);
    function GotoFirstFrame: integer;
    function GotoPreviousKeyFrame: integer;
    function GotoPreviousFrame: integer;
    function GotoNextFrame: integer;
    function GotoNextKeyFrame: integer;
    function GotoLastFrame: integer;
    property ID: integer read fID write fID;
    property Name: string read fName write fName;
    property FPS: integer read fFPS write fFPS;
    property Author: string read fAuthor write fAuthor;
    property Tag: string read fTag write fTag;
    property Skeleton: TObject read fSkeleton write fSkeleton;      // TODO: make strongly typed (note: circular reference with unit_Skeleton)
    property NumberOfFrames: integer read GetNumberOfFrames;
    property NumberOfKeyFrames: integer read GetNumberOfKeyFrames;
    property KeyframeIndices: TIntegerList read GetKeyframeIndices;
    property HasChanged: boolean read fHasChanged write SetHasChanged;
    property AnimationFrames: TAnimationFrameList read fAnimationFrames write fAnimationFrames;
    property SelectedFrame: TAnimationFrame read GetSelectedFrame write fSelectedFrame;   // might want to have a GetSelectedFrame that returns the first frame if no frame was selected yet...
    property SelectedFrameIndex: integer read GetSelectedFrameIndex;
    property TreeNode: TTreeNode read fTreeNode write fTreeNode;
    property OnChange: TNotifyEvent read fOnChange write fOnChange;
  end;

  TAnimationList = class(TDeleDObjectList)
  private
    fAnimationIDValue: integer;
    function GetAnimation(anIndex: integer): TAnimation;
  public
    procedure AddAnimation(AAnimation: TAnimation);
    property Animations[anIndex: integer]: TAnimation read GetAnimation; default;
  end;

implementation

uses
  SysUtils,
  unit_ParamList,
  unit_Enums,
  unit_AnimationFramesInterpolatorTool;

// ### TAnimation ##############################################################

constructor TAnimation.Create;
begin
    fAnimationFrames := TAnimationFrameList.Create(true,true);                  // the animation owns its frames
    fName := 'animation';
    fAuthor := 'DeleD';
    fFPS := 30;
    fTag := '';
end;

destructor TAnimation.Destroy;
begin
    FreeAndNil(fAnimationFrames);
    FreeAndNil(fKeyframeIndices);
    inherited Destroy;
end;

procedure TAnimation.Assign(AAnimation: TAnimation);
var i: integer;
begin
    fName := AAnimation.Name;
    fAuthor := AAnimation.Author;
    fFPS := AAnimation.fFPS;
    fTag := AAnimation.Tag;

    fAnimationFrames.Clear;
    for i := 0 to AAnimation.AnimationFrames.Count-1 do
        fAnimationFrames.AddAnimationFrame(AAnimation.AnimationFrames[i].Copy);
end;

function TAnimation.GetNumberOfFrames: integer;
begin
    if Assigned(fAnimationFrames) then
        Result := fAnimationFrames.Count
    else
        Result := 0;
end;

function TAnimation.GetNumberOfKeyFrames: integer;
var i, counter: integer;
begin
    counter := 0;
    for i := 0 to AnimationFrames.Count-1 do begin
        if AnimationFrames[i].IsKeyFrame then
            inc(counter);
    end;
    Result := counter;
end;

function TAnimation.GetSelectedFrame: TAnimationFrame;
begin
    if not Assigned(fSelectedFrame) then
        fSelectedFrame := AnimationFrames[0];       // NOTE: if we keep using this, we need to make sure an animation has at least one frame
    Result := fSelectedFrame;
end;

function TAnimation.GetSelectedFrameIndex: integer;
begin
    Result := AnimationFrames.IndexOf(SelectedFrame);
end;

function TAnimation.GetKeyframeIndices: TIntegerList;
var i: integer;
begin
    if not Assigned(fKeyframeIndices) then
        fKeyframeIndices := TIntegerList.Create;

    fKeyframeIndices.Clear;
    for i := 0 to AnimationFrames.Count-1 do begin
        if AnimationFrames[i].IsKeyFrame then
            fKeyframeIndices.Add(i);
    end;

    Result := fKeyframeIndices;
end;

procedure TAnimation.SetHasChanged(AValue: boolean);
var paramList: TParamList;
    frameInterpolator: TAnimationFramesInterpolatorTool;
begin
    fHasChanged := AValue;
    if AValue then begin
        // the animation has changed so rebuild the keyframe indexlist
        GetKeyframeIndices;
        // and interpolate rotation angles of all joints over all frames in the animation
        paramList := TParamList.Create(false);
        paramList.Add(self, 'selectedAnimation');
        frameInterpolator := TAnimationFramesInterpolatorTool.Create(taNotApplicable);
        frameInterpolator.Execute(paramList);
        FreeAndNil(frameInterpolator);
        FreeAndNil(paramList);
        // set to False again
        HasChanged := FALSE;
    end;
    // now notify the world the animation was changed;
    if Assigned(fOnChange) then
        fOnChange(self);
end;

procedure TAnimation.GotoFrame(AFrameIndex: integer);
begin
    if (AFrameIndex>-1) and (AFrameIndex<AnimationFrames.Count) then
        SelectedFrame := AnimationFrames[AFrameIndex];
end;

procedure TAnimation.SwapFrame(AFrameIndex: integer; ASwapWithPrevious: boolean);
begin
    if ASwapWithPrevious then begin
        if (AFrameIndex>0) and (not AnimationFrames[AFrameIndex-1].IsKeyFrame) then begin
            AnimationFrames.Move(AFrameIndex, AFrameIndex-1);
            SelectedFrame := AnimationFrames[AFrameIndex];
            HasChanged := TRUE;
        end;
    end else begin
        if (AFrameIndex<AnimationFrames.Count-1) and (not AnimationFrames[AFrameIndex+1].IsKeyFrame) then begin
            AnimationFrames.Move(AFrameIndex, AFrameIndex+1);
            SelectedFrame := AnimationFrames[AFrameIndex];
            HasChanged := TRUE;
        end;
    end;
end;

function TAnimation.GotoFirstFrame: integer;
var index: integer;
begin
    index := 0;
    SelectedFrame := AnimationFrames[index];
    Result := index;
end;

function TAnimation.GotoPreviousKeyFrame: integer;
var index,keyFrameIndex: integer;
    frame: TAnimationFrame;
begin
    index := AnimationFrames.IndexOf(SelectedFrame);
    keyFrameIndex := index;
    if (index>0) then begin
        dec(index);
        frame := AnimationFrames[index];
        while (index>0) and (not frame.IsKeyFrame) do begin
            dec(index);
            frame := AnimationFrames[index];
        end;
        if frame.IsKeyFrame then begin
            keyFrameIndex := index;
            SelectedFrame := AnimationFrames[keyFrameIndex];
        end;
    end;
    Result := keyFrameIndex;
end;

function TAnimation.GotoPreviousFrame: integer;
var index: integer;
begin
    index := AnimationFrames.IndexOf(SelectedFrame);
    if index>0 then
        dec(index);
    SelectedFrame := AnimationFrames[index];
    Result := index;
end;

function TAnimation.GotoNextFrame: integer;
var index: integer;
begin
    index := AnimationFrames.IndexOf(SelectedFrame);
    if index<AnimationFrames.Count-1 then
        inc(index);
    SelectedFrame := AnimationFrames[index];
    Result := index;
end;

function TAnimation.GotoNextKeyFrame: integer;
var index,keyFrameIndex: integer;
    frame: TAnimationFrame;
begin
    index := AnimationFrames.IndexOf(SelectedFrame);
    keyFrameIndex := index;
    if (index<AnimationFrames.Count-1) then begin
        inc(index);
        frame := AnimationFrames[index];
        while (index<AnimationFrames.Count-1) and (not frame.IsKeyFrame) do begin
            inc(index);
            frame := AnimationFrames[index];
        end;
        if frame.IsKeyFrame then begin
            keyFrameIndex := index;
            SelectedFrame := AnimationFrames[keyFrameIndex];
        end;
    end;
    Result := keyFrameIndex;
end;

function TAnimation.GotoLastFrame: integer;
var index: integer;
begin
    index := AnimationFrames.Count-1;
    SelectedFrame := AnimationFrames[index];
    Result := index;
end;

// ### TAnimationList ##########################################################

function TAnimationList.GetAnimation(anIndex: integer): TAnimation;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TAnimation(Items[anIndex])
    else
        Result := nil;
end;

procedure TAnimationList.AddAnimation(AAnimation: TAnimation);
begin
    inc(fAnimationIDValue);
    AAnimation.ID := fAnimationIDValue;
    Add(AAnimation);
    if Assigned(OnChange) then
        OnChange(nil);
end;

end.
 
