{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_AnimationFrame;

{$MODE Delphi}

interface

uses
  unit_Types,
  unit_JointLocation;

type
  // A skeleton consists out of multiple (linked) joints.
  // An animation frame contains the locations of all the joints of the skeleton;
  // this represents the skeleton in a certain pose
  // An animation consists out of multiple animation frames.
  TAnimationFrame = class(TObject)
  private
    fID: integer;
    fTag: string;
    fJointLocations: TJointLocationList;
    fIsBaseFrame: boolean;
    fIsKeyFrame: boolean;
    fAnimation: TObject;    // the animation to which this frame belongs; typed as TObject because of circular reference with unit_Animation
  public
    constructor Create;
    destructor Destroy; override;
    function Copy: TAnimationFrame;
    procedure Assign(AAnimationFrame: TAnimationFrame);
    function GetJointLocationByID(AID: integer): TJointLocation;
    procedure AddJointLocation(AJointLocation, AParentJointLocation: TJointLocation);
    procedure RemoveJointLocationByID(AID: integer);
    procedure RemoveJointLocation(AJointLocation: TJointLocation);
    property ID: integer read fID write fID;
    property Tag: string read fTag write fTag;
    property Animation: TObject read fAnimation write fAnimation;
    property JointLocations: TJointLocationList read fJointLocations write fJointLocations;
    property IsBaseFrame: boolean read fIsBaseFrame write fIsBaseFrame;
    property IsKeyFrame: boolean read fIsKeyFrame write fIsKeyFrame;
  end;

  TAnimationFrameList = class(TDeleDObjectList)
  private
    fAnimationFrameIDValue: integer;
    function GetAnimationFrame(anIndex: integer): TAnimationFrame;
  public
    function GetAnimationFrameByID(aID: integer): TAnimationFrame;
    procedure AddAnimationFrame(aAnimationFrame: TAnimationFrame; aIndex: integer = -1);
    procedure ResetIDValues;
    property AnimationFrames[anIndex: integer]: TAnimationFrame read GetAnimationFrame; default;
  end;

implementation

uses
  SysUtils,
  unit_Skeleton,
  unit_Animation, Contnrs;

// ### TAnimationFrame #########################################################

constructor TAnimationFrame.Create;
begin
    fJointLocations := TJointLocationList.Create(true,true);
    fIsBaseFrame := false;
    fIsKeyFrame := false;
end;

destructor TAnimationFrame.Destroy;
begin
    FreeAndNil(fJointLocations);
    inherited;
end;

function TAnimationFrame.Copy: TAnimationFrame;
var animationFrame: TAnimationFrame;
begin
    // might want to check IsBaseFrame here?
    animationFrame := TAnimationFrame.Create;
    animationFrame.Assign(self);
    Result := animationFrame;
end; 

procedure TAnimationFrame.Assign(AAnimationFrame: TAnimationFrame);
begin
    //IsBaseFrame := AAnimationFrame.IsBaseFrame;
    IsKeyFrame := AAnimationFrame.IsKeyFrame;
    Animation := AAnimationFrame.Animation;
    Tag := AAnimationFrame.Tag;
    JointLocations.Assign(AAnimationFrame.JointLocations);
end;

function TAnimationFrame.GetJointLocationByID(AID: integer): TJointLocation;
var counter: integer;
    jointLocation: TJointLocation;
begin
    counter := 0;
    jointLocation := JointLocations[counter];
    while (jointLocation.ID <> AID) and (counter < JointLocations.Count-1) do begin
        inc(counter);
        jointLocation := JointLocations[counter];
    end;

    if jointLocation.ID=AID then
        Result := jointLocation
    else
        Result := nil;
end;

procedure TAnimationFrame.AddJointLocation(AJointLocation, AParentJointLocation: TJointLocation);
begin
    fJointLocations.Add(AJointLocation);
    if Assigned(AParentJointLocation) then begin
        AParentJointLocation.ChildLocations.Add(AJointLocation);
        AJointLocation.ParentLocation := AParentJointLocation;
    end;
end;

// note: also removes childs of the given jointlocation
procedure TAnimationFrame.RemoveJointLocationByID(AID: integer);
var jointLocation, parentLocation: TJointLocation;
begin
    jointLocation := GetJointLocationByID(AID);
    if Assigned(jointLocation) then begin
        // remove from the parent
        parentLocation := jointLocation.ParentLocation;
        if Assigned(parentLocation) then
            parentLocation.ChildLocations.Remove(jointLocation);
        // recursively remove the jointlocations and its childs
        RemoveJointLocation(jointLocation);
    end;
end;

procedure TAnimationFrame.RemoveJointLocation(AJointLocation: TJointLocation);
var i: integer;
begin
    for i := 0 to AJointLocation.ChildLocations.Count-1 do
        RemoveJointLocation(AJointLocation.ChildLocations[i]);
    JointLocations.Remove(AJointLocation);
end;

// ### TAnimationFrameList #####################################################

function TAnimationFrameList.GetAnimationFrame(anIndex: integer): TAnimationFrame;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TAnimationFrame(Items[anIndex])
    else
        Result := nil;
end;

function TAnimationFrameList.GetAnimationFrameByID(aID: integer): TAnimationFrame;
var counter: integer;
    isFound: boolean;
    animationFrame: TAnimationFrame;
begin
    counter := 0;
    isFound := false;
    animationFrame := nil;
    while (counter<Count) and (not isFound) do begin
        animationFrame := AnimationFrames[counter];
        if animationFrame.ID = aID then
            isFound := true
        else
            inc(counter);
    end;
    Result := animationFrame;
end;

procedure TAnimationFrameList.AddAnimationFrame(aAnimationFrame: TAnimationFrame; aIndex: integer);
begin
    // NOTE: when inserting (!) a frame, it actually gets an ID that differs from
    // its position in the list. When saving the animation to the DXS (using
    // unit_XMLDataWriter), we give each frame in the list an ID which equals
    // its position in the list by using the ResetIDValues method listed below.
    inc(fAnimationFrameIDValue);
    aAnimationFrame.ID := fAnimationFrameIDValue;
    if aIndex=-1 then
        Add(aAnimationFrame)
    else
        Insert(aIndex,aAnimationFrame);

end;

procedure TAnimationFrameList.ResetIDValues;
var i: integer;
begin
    for i := 0 to Count-1 do
        AnimationFrames[i].ID := i+1;   // ID's are one-based
end;

end.
 
