{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
// Called when Animation.HasChanged is set to TRUE. See TAnimation.SetHasChanged.
//
// If the user makes changes to an arbitrary frame in the animation,
// that frame becomes a keyframe (IsKeyFrame is set to true). When
// such change occurs, the HasChanged property of the animation is
// set to true, signaling we need to reflect the change(s) made to
// the (possibly new) keyframe upon the other frames in the animation.
//
// For example, if you have an animation consisting of 30 frames and
// you modify frame 15 and 23, the software will automatically call the frames
// interpolator tool. The interpolator will interpolate joint rotation angles
// of frames between frame 15 and 23. Thus, you only need to make changes to a
// few frames and DeleD will calculate the positions of the joints in the other
// frames automatically.

unit unit_AnimationFramesInterpolatorTool;

{$MODE Delphi}

interface

uses
  unit_SkeletonTool,
  unit_ParamList,
  unit_Types,
  unit_Skeleton,
  unit_AnimationFrame,
  unit_Joint,
  unit_JointLocation,
  unit_Animation;

type
  TAnimationFramesInterpolatorTool = class(TSkeletonTool)
  private
    fAnimation: TAnimation;
    function GetNextKeyFrameIndex(AStartIndex: integer): integer;
    procedure InterpolateJointLocation(AKeyFrameIndex, ANextKeyFrameIndex: integer; AKeyFrame, ANextKeyFrame: TAnimationFrame; AJointLocation, ANextJointLocation: TJointLocation);
    procedure InterpolateBetweenFrames(AKeyFrameIndex, ANextKeyFrameIndex: integer);
    procedure DoInterpolate(AAnimation: TAnimation);
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  SysUtils,Dialogs,
  unit_Math;

// ### TAnimationFramesInterpolator ############################################

function TAnimationFramesInterpolatorTool.GetNextKeyFrameIndex(AStartIndex: integer): integer;
var animationFrame: TAnimationFrame;
    isFound: boolean;
    nextIndex: integer;
begin
    isFound := false;
    nextIndex := AStartIndex;
    while (nextIndex<fAnimation.AnimationFrames.Count-1) and (not isFound) do begin
        inc(nextIndex);
        animationFrame := fAnimation.AnimationFrames[nextIndex];
        isFound := animationFrame.IsKeyFrame;
    end;

    if not isFound then
        Result := -1
    else
        Result := nextIndex;
end;

procedure TAnimationFramesInterpolatorTool.InterpolateJointLocation(AKeyFrameIndex, ANextKeyFrameIndex: integer;
                                                                    AKeyFrame, ANextKeyFrame: TAnimationFrame;
                                                                    AJointLocation, ANextJointLocation: TJointLocation);
var i,numberOfFrames: integer;
    interpolatedRotationVector, rotationVector: TVertice;
    interpolatedLocationVector, locationVector: TVertice;
    jointLocation, nextJointLocation: TJointLocation;
    animationFrame: TAnimationFrame;
begin
    numberOfFrames := ANextKeyFrameIndex-AKeyFrameIndex;

    interpolatedRotationVector := vectorSub(ANextJointLocation.Rotation,AJointLocation.Rotation);
    interpolatedRotationVector.DivideByScalar(numberOfFrames);
    rotationVector := AJointLocation.Rotation.Copy;

    interpolatedLocationVector := vectorSub(ANextJointLocation.Location,AJointLocation.Location);
    interpolatedLocationVector.DivideByScalar(numberOfFrames);
    locationVector := AJointLocation.Location.Copy;

    for i := AKeyFrameIndex+1 to ANextKeyFrameIndex-1 do begin
        animationFrame := fAnimation.AnimationFrames[i];
        jointLocation := animationFrame.GetJointLocationByID(AJointLocation.ID);

        // interpolate rotation vector over jointlocations
        rotationVector.Add(interpolatedRotationVector);
        jointLocation.Rotation.Assign(rotationVector);

        // interpolate location vector over jointlocations
        locationVector.Add(interpolatedLocationVector);
        jointLocation.Location.Assign(locationVector);
    end;

    FreeAndNil(interpolatedRotationVector);
    FreeAndNil(rotationVector);

    FreeAndNil(interpolatedLocationVector);
    FreeAndNil(locationVector);

    for i := 0 to AJointLocation.ChildLocations.Count-1 do begin
        jointLocation := AJointLocation.ChildLocations[i];
        nextJointLocation := ANextJointLocation.ChildLocations[i];
        InterpolateJointLocation(AKeyFrameIndex, ANextKeyFrameIndex, AKeyFrame, ANextKeyFrame, jointLocation, nextJointLocation);
    end;
end;

procedure TAnimationFramesInterpolatorTool.InterpolateBetweenFrames(AKeyFrameIndex, ANextKeyFrameIndex: integer);
var keyFrame,nextKeyFrame: TAnimationFrame;
    jointLocation, nextJointLocation: TJointLocation;
begin
    keyFrame := fAnimation.AnimationFrames[AKeyFrameIndex];
    nextKeyFrame := fAnimation.AnimationFrames[ANextKeyFrameIndex];

    jointLocation := keyFrame.JointLocations[0];
    nextJointLocation := nextKeyFrame.JointLocations[0];

    InterpolateJointLocation(AKeyFrameIndex, ANextKeyFrameIndex, keyFrame, nextKeyFrame, jointLocation, nextJointLocation);
end;

procedure TAnimationFramesInterpolatorTool.DoInterpolate(AAnimation: TAnimation);
var i,index,nextIndex: integer;
    animationFrame: TAnimationFrame;
begin
    index := GetNextKeyFrameIndex(-1);
    if index <> -1 then begin

        // if the first keyframe isn't the first frame in the animation,
        // we always reset all frames up to the first keyframe so they
        // resemble the baseframe
        if index > 0 then begin
            for i := 0 to index-1 do begin
                animationFrame := AAnimation.AnimationFrames[i];
                animationFrame.Assign(TSkeleton(AAnimation.Skeleton).BaseFrame);
                // we have assigned the baseframe to this frame, which destroyed
                // the Animation property - set it here again
                animationFrame.Animation := AAnimation;
            end;
        end;

        // we interpolate between keyframes
        nextIndex := GetNextKeyFrameIndex(index);
        while nextIndex<>-1 do begin
            InterpolateBetweenFrames(index,nextIndex);
            index := nextIndex;
            nextIndex := GetNextKeyFrameIndex(nextIndex);
        end;

        // if the last keyframe isn't the last frame of the animation,
        // reset all frames from the last keyframe to the last frame
        if index < AAnimation.NumberOfFrames-2 then begin
            for i := index+1 to AAnimation.NumberOfFrames-1 do begin
                animationFrame := AAnimation.AnimationFrames[i];
                animationFrame.Assign(TSkeleton(AAnimation.Skeleton).BaseFrame);
                // we have assigned the baseframe to this frame, which destroyed
                // the Animation property - set it here again
                animationFrame.Animation := AAnimation;
            end;
        end;
    end;
end;

procedure TAnimationFramesInterpolatorTool.DoPrepare(AParamList: TParamList);
begin
end;

procedure TAnimationFramesInterpolatorTool.DoExecute(AParamList: TParamList);
begin
    fAnimation := TAnimation(AParamList[0]);
    DoInterpolate(fAnimation);
end;



end.
