{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_BSP;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Graphics, Controls, Dialogs, Contnrs, StrUtils, dglOpenGL,
  frmProgressForm,
  unit_Types,
  unit_Generics,
  unit_Enums,
  unit_Math;

type
  TBSPNode = class(TObject)
  private
    fSplitter: TGenericPolygon;
    fBackNode: TBSPNode;
    fFrontNode: TBSPNode;
    function Intersection(anA, aB: TGenericVertex; plane: TPlane): TGenericVertex;
    procedure SplitPolygon(originalPolygon: TGenericPolygon; plane: TPlane; var left,right: TGenericPolygon);
    function FindSplitterIndex(list: TGenericPolygonList; optimizeMethod: TBSPOptimize): integer;
  public
    constructor CreateBSP(list: TGenericPolygonList; progressForm: TfrmProgress);
    destructor Destroy; override;
    procedure Traverse;
    property Splitter: TGenericPolygon read fSplitter write fSplitter;
    property BackNode: TBSPNode read fBackNode write fBackNode;
    property FrontNode: TBSPNode read fFrontNode write fFrontNode;
  end;

implementation

// ### TBSPNode ################################################################

procedure TBSPNode.Traverse;    // testfunction
var i: integer;
begin
    if Assigned(BackNode) then
        BackNode.Traverse;

    glBegin(GL_LINE_LOOP);
        for i := 0 to Splitter.vertices.Count-1 do
            glVertex3f(Splitter.vertices[i].x, Splitter.vertices[i].y, Splitter.vertices[i].z);
    glEnd;

    if Assigned(FrontNode) then
        FrontNode.Traverse;
end;

function TBSPNode.Intersection(anA, aB: TGenericVertex; plane: TPlane): TGenericVertex;
var a, b, v, ms: TVertex;
    f: double;
    uvA, uvB, uvCoord: TUVCoord;
    i: integer;

begin
    a := VerticeToVertex( anA );
    b := VerticeToVertex( aB );

    v := VertexSubtract( b, a );
    f := -DistToPlane( anA, plane ) / VertexDotProduct( v, VerticeToVertex( plane.Normal ));
    ms := VertexMultiply( v, f );
    Result := TGenericVertex.Create( VertexAdd( a, ms ));

    // Intersect the UV Coordinates as well.
    for i := 0 to anA.UVCoordinates.Count - 1 do begin
        uvA := anA.UVCoordinates[i];
        uvB := aB.UVCoordinates[i];

        uvCoord := TUVCoord.Create(
            uvA.U + f * ( uvB.U - uvA.U ),
            uvA.V + f * ( uvB.V - uvA.V ));
        Result.UVCoordinates.Add( uvCoord );
    end;
end;
                                   
procedure TBSPNode.SplitPolygon(originalPolygon: TGenericPolygon; plane: TPlane; var left, right: TGenericPolygon);
const MinDist = 0.0001;
var i,index: integer;
    A,B,X: TGenericVertex;
    planeDA, planeDB: double;
begin
    // loop through the vertices of the original polygon, checking each vertex against the plane
    for i := 0 to originalPolygon.vertices.Count-1 do begin

        index := (i+1) mod originalPolygon.vertices.Count;  // next vertex in the list

        A := originalPolygon.vertices[i];
        B := originalPolygon.vertices[index];
        planeDA := DistToPlane(A,plane);
        planeDB := DistToPlane(B,plane);

        if planeDB > MinDist then begin             // vector B is to the right

            if planeDA < -MinDist then begin        // vector A is to the left

                // calculate intersection point
                X := Intersection(A,B,plane);

                // store it in both polygons
                left.Vertices.Add(X);
                right.Vertices.Add(X.Copy);
            end;

            if (planeDA > -MinDist) and (planeDA < MinDist) then	// vector A is on the plane
                right.Vertices.Add(A.Copy);

            right.Vertices.Add(B.Copy);

        end else if planeDB < -MinDist then begin   // vector B is to the left

            if planeDA > MinDist then begin         // vector A is to the right

                // calculate intersection point
                X := Intersection(A,B,plane);

                // store it in both polygons
                left.Vertices.Add(X);
                right.Vertices.Add(X.Copy);
            end;

            if (planeDA > -MinDist) and (planeDA < MinDist) then	// vector A is on the plane
                right.Vertices.Add(A.Copy);

            left.Vertices.Add(B.Copy);

        end else begin      // both vertices are on the same side or ON the plane

            left.Vertices.Add(B.Copy);
            right.Vertices.Add(B.Copy);

        end;

    end;
end;

// determine which polygon is most suitable to act as splitter
function TBSPNode.FindSplitterIndex(list: TGenericPolygonList; optimizeMethod: TBSPOptimize): integer;
var i,j,searchValue,newValue,polyIndex,back,front,spanning: integer;
    polygon: TGenericPolygon;
begin
    // always return the first polygon if the method is boUnoptimized
    if optimizeMethod=boUnoptimized then begin
        Result := 0;
        exit;
    end;

    searchValue := 32767;
    polyIndex := 0;

    // for each polygon, find the # of splits + the # of back/front nodes it would create when used as Splitter
    for i := 0 to list.Count-1 do begin
        polygon := list.Polygons[i];

        back := 0;
        front := 0;
        spanning := 0;

        for j := 0 to list.Count-1 do begin
            if j<>i then begin

                case list.Polygons[j].RelativePosition(polygon) of
                    rpBack:
                        inc(back);
                    rpFront,rpCoincident:
                        inc(front);
                    rpSpanning:
                        inc(spanning);
                end;
            end;
        end;

        case optimizeMethod of
            boBalanced:
                begin
                    newValue := abs(front - back);
                    if newValue<SearchValue then begin
                        SearchValue := newValue;
                        polyIndex := i;
                    end;
                end;
            boMinimumSplits:
                begin
                    if spanning<SearchValue then begin
                        SearchValue := spanning;
                        polyIndex := i;
                    end;
                end;
        end;
    end;

    Result := polyIndex;
end;

constructor TBSPNode.CreateBSP(list: TGenericPolygonList; progressForm: TfrmProgress);
var leftList,rightList: TGenericPolygonList;
    i,splitterIndex: integer;
    polygon,leftPolygon,rightPolygon: TGenericPolygon;
begin
    // find the best suitable Splitter
    Splitter := TGenericPolygon.Create;
    splitterIndex := FindSplitterIndex(list,boMinimumSplits);
    Splitter.Assign(list[splitterIndex]);

    // create left/right polygonlists
    leftList := TGenericPolygonList.Create(true);
    rightList := TGenericPolygonList.Create(true);

    // increase the position of the progress form; its Max property has been set
    // in the calling routine (CSG); due to the recursive nature of this routine,
    // each increase of the position property actually means that a list of polygons
    // (for a node) has been handled - these lists get shorter while the tree is
    // being build so the progress form shows an increase in speed during build
    progressForm.Position := progressForm.Position+1;

    // assign all polygons (except Splitter) to left/right lists
    for i := 0 to list.Count-1 do begin

        if i<>splitterIndex then begin

            polygon := list.Polygons[i];

            case polygon.RelativePosition(Splitter) of
                rpBack:
                    begin   // left
                        leftList.Add(polygon.Copy);
                    end;
                rpSpanning:
                    begin   // split
                        leftPolygon := TGenericPolygon.Create;
                        rightPolygon := TGenericPolygon.Create;
                        SplitPolygon(polygon,Splitter.Plane,leftPolygon,rightPolygon);
                        leftList.Add(leftPolygon);
                        rightList.Add(rightPolygon);
                    end;
                rpFront:
                    begin   // right
                        rightList.Add(polygon.Copy);
                    end;
                rpCoincident:
                    begin
                        if polygon.IsEqualNormal(Splitter) then
                            rightList.Add(polygon.Copy)
                        else
                            leftList.Add(polygon.Copy);
                    end;
            end;

        end;
    end;

    // now create new nodes based on the left/right lists
    if leftList.Count>0 then
        BackNode := TBSPNode.CreateBSP(leftList,progressForm);
    if rightList.Count>0 then
        FrontNode := TBSPNode.CreateBSP(rightList,progressForm);

    // free stuff
    FreeAndNil(leftList);
    FreeAndNil(rightList);
end;

// traverse the tree, destroying nodes in the process
destructor TBSPNode.Destroy;
begin
    FreeAndNil(fSplitter);
    if Assigned(fBackNode) then
        FreeAndNil(fBackNode);
    if Assigned(fFrontNode) then
        FreeAndNil(fFrontNode);
end;

end.
