{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_CSGSplitterTool;

{$MODE Delphi}

interface

uses
  unit_CSGTool,
  unit_ParamList,
  unit_Generics;

type
  TCSGSplitter = class(TCSGTool)
  private
    procedure Intersect(ABasePolygonList: TGenericPolygonListList;
                        ADoOptimizeGeometry,
                        ADoShowProgress: boolean;
                        AResultPolygonList: TGenericPolygonListList);
    procedure Subtract(ABasePolygonList: TGenericPolygonListList;
                       ADoOptimizeGeometry,
                       ADoShowProgress: boolean;
                       AResultPolygonList: TGenericPolygonListList);
  protected
    procedure DoCSGExecute(ABasePolygonList: TGenericPolygonListList;
                           ADoOptimizeGeometry,
                           ADoShowProgress: boolean;
                           AResultPolygonList: TGenericPolygonListList); override;
    procedure DoPrepare(AParamList: TParamList); override;
  public
  end;

implementation

uses
  SysUtils,
  unit_BSP;

// ### TCSGSplitter ############################################################

procedure TCSGSplitter.Intersect(ABasePolygonList: TGenericPolygonListList;
                                 ADoOptimizeGeometry,
                                 ADoShowProgress: boolean;
                                 AResultPolygonList: TGenericPolygonListList);
var i: integer;
    treeObject, treeOperator: TBSPNode;
    operator, currentObject, finalObject, pushA, pushB: TGenericPolygonList;
begin
    // create a generic polygonlist to store the CSG result in
    finalObject := TGenericPolygonList.Create(true);

    // use the first object in the baseList as operator and create a BSP tree for it
    operator := ABasePolygonList.GenericPolygonList[0];
    SetProgressFormProperties('Building BSP','Building Binary Space Partition Trees...','Object: '+operator.Name,operator.Count,0,ADoShowProgress);
    treeOperator := TBSPNode.CreateBSP(operator,fProgressForm);

    // loop through baseList, performing CSG on each object
    for i := 1 to ABasePolygonList.Count-1 do begin

        // create BSP tree for current object
        currentObject := ABasePolygonList.GenericPolygonList[i];
        SetProgressFormProperties('Building BSP','Building Binary Space Partition Trees...','Object: '+currentObject.Name,currentObject.Count,0,ADoShowProgress);
        treeObject := TBSPNode.CreateBSP(currentObject,fProgressForm);

        // prepare progress form for CSG operation
        SetProgressFormProperties('Performing CSG','Performing Constructive Solid Geometry...','',0,0,ADoShowProgress);

        // push objects through BSP trees, thus performing CSG
        pushA := PushList(currentObject,treeOperator,false,true,false,ADoOptimizeGeometry);   // push polygons of A through BSP of B, keeping only those that are in B
        pushB := PushList(operator,treeObject,false,false,false,ADoOptimizeGeometry);         // push polygons of B through BSP of A, keeping only those that are in A

        // add the lists together to create the final object
        finalObject.Clear;
        finalObject.AddList(pushA);
        finalObject.AddList(pushB);

        // free the operator
        FreeAndNil(treeOperator);
        FreeAndNil(treeObject);
        FreeAndNil(pushA);
        FreeAndNil(pushB);

        // create a new operator based on the previous CSG action
        operator := finalObject;
        treeOperator := TBSPNode.CreateBSP(operator,fProgressForm);
    end;

    // final cleanup
    FreeAndNil(treeOperator);

    // store result
    AResultPolygonList.Add(finalObject);
end;

procedure TCSGSplitter.Subtract(ABasePolygonList: TGenericPolygonListList;
                                ADoOptimizeGeometry,
                                ADoShowProgress: boolean;
                                AResultPolygonList: TGenericPolygonListList);
var i: integer;
    treeObject, treeOperator: TBSPNode;
    operator, currentObject, finalObject, pushA, pushB: TGenericPolygonList;
begin
    // use the last object in the ABasePolygonList as operator and create a BSP tree for it
    operator := ABasePolygonList.GenericPolygonList[ABasePolygonList.Count-1];

    if operator.Count > 0 then begin

        SetProgressFormProperties('Building BSP','Building Binary Space Partition Trees...','Object: '+operator.Name,operator.Count,0,ADoShowProgress);
        try
            treeOperator := TBSPNode.CreateBSP(operator,fProgressForm);

            // loop through ABasePolygonList, performing CSG on each object
            for i := 0 to ABasePolygonList.Count-2 do begin

                // create a generic polygonlist to store the CSG result in
                finalObject := TGenericPolygonList.Create(true);

                // create BSP tree for current object
                currentObject := ABasePolygonList.GenericPolygonList[i];
                SetProgressFormProperties('Building BSP','Building Binary Space Partition Trees...','Object: '+currentObject.Name,currentObject.Count,0,ADoShowProgress);
                treeObject := TBSPNode.CreateBSP(currentObject,fProgressForm);

                // prepare progress form for CSG operation
                SetProgressFormProperties('Performing CSG','Performing Constructive Solid Geometry...','',0,0,ADoShowProgress);

                // perform CSG
                if fRemoveCoplanar then begin
                    pushA := PushList(currentObject,treeOperator,fStopAtCoplanar,true,true,ADoOptimizeGeometry);     // push polygons of A through BSP of B, keeping only those that are NOT in B
                    pushB := PushList(operator,treeObject,fStopAtCoplanar,true,false,ADoOptimizeGeometry);           // push polygons of B through BSP of A, keeping only those that are in A
                end else begin
                    pushA := PushList(currentObject,treeOperator,fStopAtCoplanar,true,true,ADoOptimizeGeometry);     // push polygons of A through BSP of B, keeping only those that are NOT in B
                    pushB := PushList(operator,treeObject,fStopAtCoplanar,false,false,ADoOptimizeGeometry);          // push polygons of B through BSP of A, keeping only those that are in A
                end;

                {
                if removeCoplanar then begin
                    pushA := PushList(currentObject,treeOperator,true,true,optimizeGeometry);     // push polygons of A through BSP of B, keeping only those that are NOT in B
                    pushB := PushList(operator,treeObject,true,false,optimizeGeometry);          // push polygons of B through BSP of A, keeping only those that are in A
                end else begin
                    pushA := PushList(currentObject,treeOperator,true,true,optimizeGeometry);     // push polygons of A through BSP of B, keeping only those that are NOT in B
                    pushB := PushList(operator,treeObject,false,false,optimizeGeometry);          // push polygons of B through BSP of A, keeping only those that are in A
                end;
                }

                // invert polygons of B
                pushB.InvertPolygons;

                // add the lists together to create the final object
                finalObject.AddList(pushA);
                finalObject.AddList(pushB);

                // store in result
                AResultPolygonList.Add(finalObject);

                // clean up for this run
                FreeAndNil(treeObject);
                FreeAndNil(pushA);
                FreeAndNil(pushB);
            end;

        finally
            FreeAndNil(treeOperator);
        end;
    end;
end;

procedure TCSGSplitter.DoCSGExecute(ABasePolygonList: TGenericPolygonListList;
                                    ADoOptimizeGeometry,
                                    ADoShowProgress: boolean;
                                    AResultPolygonList: TGenericPolygonListList);
var intersectResultList: TGenericPolygonListList;
begin
    intersectResultList := TGenericPolygonListList.Create(false);

    try
        // first perform an intersection on the 2 objects
        Intersect(ABasePolygonList,ADoOptimizeGeometry,ADoShowProgress,intersectResultList);

        if intersectResultList[0].Count>0 then begin

            // remove the current operator from the baselist and add the result
            // from the intersection as new operator to the baselist
            ABasePolygonList.Delete(1);
            intersectResultList[0].Name := 'intersection result';
            ABasePolygonList.Add(intersectResultList[0]);

            // subtract the intersection result from the first object
            fRemoveCoplanar := true;
            fStopAtCoplanar := true;
            Subtract(ABasePolygonList,ADoOptimizeGeometry,ADoShowProgress,AResultPolygonList);

            // the intersection is part of the result as well
            AResultPolygonList.Add(intersectResultList[0].Copy);
        end;

    finally
        FreeAndNil(intersectResultList);
    end;
end;

procedure TCSGSplitter.DoPrepare(AParamList: TParamList);
begin
    Hint := 'CSG Split';
    Description := 'CSG Split';
end;

end.
