{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
// baseclass for CSG tools

unit unit_CSGTool;

{$MODE Delphi}

interface

uses
  SysUtils,
  unit_Types,
  unit_Generics,
  unit_Enums,
  unit_Math,
  unit_Primitives,
  unit_PrimitiveList,
  unit_BSP,
  unit_ParamList,
  frmProgressForm,
  unit_PrimitiveTool;

type
  TCSGTool = class(TPrimitiveTool)
  protected
    fProgressForm: TfrmProgress;
    fRemoveCoplanar: boolean;
    fStopAtCoplanar: boolean;
    function Intersection(anA, aB: TGenericVertex; plane: TPlane): TGenericVertex;
    procedure SplitPolygon(originalPolygon: TGenericPolygon; plane: TPlane; var left,right: TGenericPolygon);
    procedure PushPolygon(polygon: TGenericPolygon; treeParent,treeChild: TBSPNode; stopAtCoplanar,removeCoplanar,keepFront: boolean; var numSplitPolys,numKeepPolys: integer; var polyList: TGenericPolygonList);
    function PushList(list: TGenericPolygonList; tree: TBSPNode; stopAtCoplanar,removeCoplanar,keepFront,optimizeGeometry: boolean): TGenericPolygonList;
    procedure SetProgressFormProperties(title,subTitle,description: string; max,position: integer; showForm: boolean);
    procedure DoCSGExecute(ABasePolygonList: TGenericPolygonListList;
                           ADoOptimizeGeometry,
                           ADoShowProgress: boolean;
                           AResultPolygonList: TGenericPolygonListList); virtual; abstract;
  public
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  Dialogs,
  unit_Options,
  unit_IntegerList,
  unit_PrimitiveRemoverTool,
  unit_PrimitiveOptimizerTool;

// INFO, PJ: Copied from TBSPNode.Intersection. Change some day.
function TCSGTool.Intersection(anA, aB: TGenericVertex; plane: TPlane): TGenericVertex;
var a, b, v, ms: TVertex;
    f: double;
    uvA, uvB, uvCoord: TUVCoord;
    i: integer;

begin
    a := VerticeToVertex( anA );
    b := VerticeToVertex( aB );

    v := VertexSubtract( b, a );
    f := -DistToPlane( anA, plane ) / VertexDotProduct( v, VerticeToVertex( plane.Normal ));
    ms := VertexMultiply( v, f );
    Result := TGenericVertex.Create( VertexAdd( a, ms ));

    // Intersect the UV Coordinates as well.
    for i := 0 to anA.UVCoordinates.Count - 1 do begin
        uvA := anA.UVCoordinates[i];
        uvB := aB.UVCoordinates[i];

        uvCoord := TUVCoord.Create( 
            uvA.U + f * ( uvB.U - uvA.U ),
            uvA.V + f * ( uvB.V - uvA.V ));
        Result.UVCoordinates.Add( uvCoord );
    end;
end;

procedure TCSGTool.SplitPolygon(originalPolygon: TGenericPolygon; plane: TPlane; var left, right: TGenericPolygon);
const MinDist = 0.0001;
var i,index: integer;
    A,B,X: TGenericVertex;
    planeDA, planeDB: double;
begin
    // loop through the vertices of the original polygon, checking each vertex against the plane
    for i := 0 to originalPolygon.Vertices.Count-1 do begin

        index := (i+1) mod originalPolygon.Vertices.Count;  // next vertex in the list

        A := originalPolygon.Vertices[i];
        B := originalPolygon.Vertices[index];
        planeDA := DistToPlane(A,plane);
        planeDB := DistToPlane(B,plane);

        if planeDB > MinDist then begin             // vector B is to the right

            if planeDA < -MinDist then begin        // vector A is to the left

                // calculate intersection point
                X := Intersection(A,B,plane);

                // store it in both polygons
                left.Vertices.Add(X);
                right.Vertices.Add(X.Copy);
            end;

            if (planeDA > -MinDist) and (planeDA < MinDist) then	// vector A is on the plane
                left.Vertices.Add(A.Copy);

            right.Vertices.Add(B.Copy);

        end else if planeDB < -MinDist then begin   // vector B is to the left

            if planeDA > MinDist then begin         // vector A is to the right

                // calculate intersection point
                X := Intersection(A,B,plane);

                // store it in both polygons
                left.Vertices.Add(X);
                right.Vertices.Add(X.Copy);
            end;

            if (planeDA > -MinDist) and (planeDA < MinDist) then	// vector A is on the plane
                right.Vertices.Add(A.Copy);

            left.Vertices.Add(B.Copy);

        end else begin      // both vertices are on the same side or ON the plane

            left.Vertices.Add(B.Copy);
            right.Vertices.Add(B.Copy);

        end;
    end;
end;

procedure TCSGTool.PushPolygon(polygon: TGenericPolygon; treeParent,treeChild: TBSPNode; stopAtCoplanar,removeCoplanar,keepFront: boolean; var numSplitPolys,numKeepPolys: integer; var polyList: TGenericPolygonList);
var backPolygon,frontPolygon,resPolygon: TGenericPolygon;
    polygonPosition: TRelativePosition;
    node: TBSPNode;
begin
    if treeChild=nil then begin	// it's a leaf, so we stop here

        polygonPosition := polygon.RelativePosition(treeParent.Splitter);

        resPolygon := nil;
        case polygonPosition of
            rpBack:
                if not keepFront then resPolygon := polygon.Copy;
            rpFront:
                if keepFront then resPolygon := polygon.Copy;
            rpCoincident:
                if (not removeCoplanar) then resPolygon := polygon.Copy;
        end;

        if Assigned(resPolygon) then begin
            polyList.Add(resPolygon);
            inc(numKeepPolys);
        end;

    end else begin

        // determine position of the polygon relative to the splitter and push it (or it's fragments by now) through relevant BSP node(s)
        case polygon.RelativePosition(treeChild.Splitter) of

            rpBack:
                PushPolygon(polygon,treeChild,treeChild.BackNode,stopAtCoplanar,removeCoplanar,keepFront,numSplitPolys,numKeepPolys,polyList);

            rpFront:
                PushPolygon(polygon,treeChild,treeChild.FrontNode,stopAtCoplanar,removeCoplanar,keepFront,numSplitPolys,numKeepPolys,polyList);

            rpSpanning:
                begin
                    backPolygon := TGenericPolygon.Create;
                    frontPolygon := TGenericPolygon.Create;

                    backPolygon.Material := polygon.Material;
                    frontPolygon.Material := polygon.Material;

                    SplitPolygon(polygon,treeChild.Splitter.Plane,backPolygon,frontPolygon);

                    inc(numSplitPolys);

                    PushPolygon(frontPolygon,treeChild,treeChild.FrontNode,stopAtCoplanar,removeCoplanar,keepFront,numSplitPolys,numKeepPolys,polyList);
                    PushPolygon(backPolygon,treeChild,treeChild.BackNode,stopAtCoplanar,removeCoplanar,keepFront,numSplitPolys,numKeepPolys,polyList);

                    FreeAndNil(backPolygon);
                    FreeAndNil(frontPolygon);
                end;

            rpCoincident:   // polygon is coplanar with the splitter
                begin
                    if polygon.IsEqualNormal(treeChild.Splitter) then begin     // polygon is pointing in the same direction as the splitter

                        // should probably figure out if the polygon overlaps the splitter;
                        // if not, continue; if it does, and if removeCoplanar=true, then stop

                        if removeCoplanar then begin
                            if stopAtCoplanar then              // dirty trick; used for Split command in relation with Subtract
                                node := nil                     // when performing a Subtract on itself, this routine should continue
                            else                                // but if Subtract is used when performing a Split, it should stop when
                                node := treeChild.BackNode;     // a coplanar polygon encountered
                            PushPolygon(polygon,treeChild,node,stopAtCoplanar,removeCoplanar,keepFront,numSplitPolys,numKeepPolys,polyList) // treeChild.backNode
                        end else
                            PushPolygon(polygon,treeChild,treeChild.FrontNode,stopAtCoplanar,removeCoplanar,keepFront,numSplitPolys,numKeepPolys,polyList);
                    end else
                        PushPolygon(polygon,treeChild,treeChild.BackNode,stopAtCoplanar,removeCoplanar,keepFront,numSplitPolys,numKeepPolys,polyList);
                end;
        end;
    end;
end;

function TCSGTool.PushList(list: TGenericPolygonList; tree: TBSPNode; stopAtCoplanar,removeCoplanar,keepFront,optimizeGeometry: boolean): TGenericPolygonList;
var i,j,numSplitPolys,numKeepPolys: integer;
begin
    Result := TGenericPolygonList.Create(true);

    fProgressForm.Max := list.Count;

    for i := 0 to list.Count-1 do begin
        fProgressForm.Position := i;
        numSplitPolys := 0;
        numKeepPolys := 0;
        PushPolygon(list.Polygons[i],tree,tree,stopAtCoplanar,removeCoplanar,keepFront,numSplitPolys,numKeepPolys,Result);
        if optimizeGeometry then begin
            if (numSplitPolys+1)=numKeepPolys then begin                        // splitting was not necessary, so keep original polygon
                // remove all polygon-fragments
                for j := 0 to numKeepPolys-1 do
                    Result.Delete(Result.Count-1);

                // and replace them with the original polygon
                Result.Add(list.Polygons[i].Copy);
            end else begin                                                      // polygon was split; now try merging fragments


            end;
        end;
    end;
end;

procedure TCSGTool.SetProgressFormProperties(title,subTitle,description: string; max,position: integer; showForm: boolean);
begin
    if Assigned(fProgressForm) then begin
        fProgressForm.Title := title;
        fProgressForm.SubTitle := subTitle;
        fProgressForm.Description := description;
        fProgressForm.Max := max;
        fProgressForm.Position := position;
        if showForm then
            fProgressForm.Show;
    end;
end;

procedure TCSGTool.DoExecute(AParamList: TParamList);
var i: integer;
    baseList,resultList: TGenericPolygonListList;
    polyList,fixedResult: TGenericPolygonList;
    primitive,operator: TBasePrimitive;
    //keepOperator,autoOptimize,
    optimizeGeometry,showProgress: boolean;
    csgOptions: TCSGOptions;
begin
    if SelectedPrimitives.Count<2 then
        MessageDlg('Please select 2 or more objects for CSG operations.'+#13#10+#13#10+
                   'The last selected object will be used as CSG operator.',mtError,[mbOK],0)
    else begin
        // get parameters from ParamList
        csgOptions := TCSGOptions(AParamList.NamedParameters['csgOptions']);

        //autoOptimize := false;
        //keepOperator := TBooleanObject(AParamList.NamedParameters['keepOperator']).Value;
        optimizeGeometry := false;
        fProgressForm := TfrmProgress.Create(nil);
        showProgress := false;

        // create lists to work with
        baseList := TGenericPolygonListList.Create(true);
        resultList := TGenericPolygonListList.Create(true); // generics created by CSG methods

        try
            // convert each baseprimitive into a generic polygonlist, stored in a list of lists
            for i := 0 to SelectedPrimitives.Count-1 do begin
                polyList := TGenericPolygonList.Create(true);
                SelectedPrimitives[i].BasePrimitiveToGenericPolygonList(polyList);
                // if 1 of the objects being used during CSG has over a certain amount of polygons,
                // the progress form must be made visible; this amount can be set in the Options screen
                // and is fed to the CSG tool at startup
                if not showProgress then
                    showProgress := polyList.Count>csgOptions.MinAmountOfPolygonsForProgressForm;
                baseList.Add(polyList);
            end;

            // execute actual CSG routine (implemented in derived classes)
            DoCSGExecute(baseList, optimizeGeometry, showProgress, resultList);

            // first delete the current selection from the scene, then add CSG result(s)
            operator := nil;
            if csgOptions.KeepOperator then begin
                // if we want to keep the operator, we remove it from the selection first
                // so it doesn't get deleted when the selection is remove from the scene
                operator := SelectedPrimitives[SelectedPrimitives.Count-1];
                SelectedPrimitives.Remove(operator);
                // we also remove it from the Undo buffer because we don't want it
                // copied back into the scene again
                UndoPrimitiveList.DeleteByID(operator.ID);
                PrimitiveIDList.Remove(operator.ID,true);
            end;
            TSelectedPrimitivesRemoverFunction.Execute(ScenePrimitives,SelectedPrimitives);

            // fix geometry and create a DeleD primitive out of each object in the result list + add them to the scene+selection
            for i := 0 to resultList.Count-1 do begin

                // fix T-junctions and connect polygons for each object in the result list
                fixedResult := resultList.GenericPolygonList[i].FixTJunctions(fProgressForm);

                // create a new DeleD primitive for each object in the result list and add it to the scene+selection
                primitive := TUserDefined.CreateUserDefined;
                primitive.GenericPolygonListToBasePrimitive(fixedResult);
                ScenePrimitives.AddPrimitive(primitive,true);

                // add ID for UndoRedo purposes
                PrimitiveIDList.Add(primitive.ID);

                // auto-optimize the geometry if wanted
                if csgOptions.AutoOptimize then
                    TPrimitiveOptimizeFunction.Execute(primitive);

                // free temporary list
                FreeAndNil(fixedResult);
            end;

            // add the operator to the scene if wanted
            if csgOptions.KeepOperator then
                SelectedPrimitives.Add(operator);

        finally
            FreeAndNil(baseList);
            FreeAndNil(resultList);
            FreeAndNil(fProgressForm);
        end;
    end;
end;

end.
