{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Cameras;

{$MODE Delphi}

interface

uses
    unit_Types,
    unit_Math,
    unit_Variables;

type
  TCamera = class( TObject)
  public
      function Clone: TCamera; virtual; abstract;
      procedure Assign( const aCamera: TCamera ); virtual; abstract;
  end;

  TCamera2d = class(TCamera)                                                    // camera for 2d views
  private
    fPosition: TVertice;
    fDepth: integer;                                                            // depth (alternative z)
    fMinZoom: integer;
    fMaxZoom: integer;
    procedure SetDepth(AValue: integer);
  public
    constructor Create(x,y,z: double);
    destructor Destroy; override;
    function Clone: TCamera; override;
    procedure Assign( const aCamera: TCamera ); override;
    procedure Move(vector: TVertice);
    procedure MoveTo(APosition: TVertice);
    procedure Reset;
    property Position: TVertice read fPosition write fPosition;
    property Depth: integer read fDepth write SetDepth;
    property MinZoom: integer read fMinZoom write fMinZoom;
    property MaxZoom: integer read fMaxZoom write fMaxZoom;
  end;

  // INFO, PJ: We should probably merge this with TCamera2D some day, pretty similar
  TCameraUV = class(TCamera)
  private
    fCenterU: single;
    fCenterV: single;
    fScale: single;
  public
    function Clone: TCamera; override;
    procedure Assign( const aCamera: TCamera ); override;
    procedure Zoom(const aDelta: integer);
    procedure Move( const aDeltaX, aDeltaY: integer );
    procedure FitToWindow( const aWidth, aHeight: integer );
    property CenterU: single read fCenterU write fCenterU;
    property CenterV: single read fCenterV write fCenterV;
    property Scale: single read fScale write fScale;
  end;


  TCamera3d = class(TCamera)                                                    // camera for 3d view
  private
    fPosition: TVertice;
    fRotation: TVertice;
    fFOV: Single;

    function GetMatrix: TMatrix4x4;

  public
    constructor Create(x,y,z: double);
    destructor Destroy; override;
    function Clone: TCamera; override;
    procedure Assign( const aCamera: TCamera ); override;
    procedure Move(vector: TVertice);
    procedure MoveRelative(const aVector: TVertice); overload;
    procedure MoveRelative(const anX, aY, aZ: single); overload;
    procedure Orbit(const anX, aY, aZ: Single; const aOrbitSelectedObject: Boolean;
                    const aSelectedObjectX, aSelectedObjectY, aSelectedObjectZ : Single;
                    const aConstrainToX,aConstrainToY: Boolean); overload;
    procedure Rotate(const anX, anY, aZ: integer);
    procedure RotateWithAdj(const anX, anY, aZ: Single);
    procedure Reset;
    property Position: TVertice read fPosition write fPosition;
    property Rotation: TVertice read fRotation write fRotation;              // pitch, yaw and roll
    property FOV: Single read fFOV write fFOV;
    property Matrix: TMatrix4x4 read GetMatrix;
  end;

implementation

uses
    SysUtils,
    unit_Constants;

// ### TCamera2d ###############################################################

procedure TCamera2d.SetDepth(AValue: integer);
begin
    if AValue < minzoom then AValue := minzoom;
    if AValue > maxzoom then AValue := maxzoom;
    fDepth := AValue;
end;

procedure TCamera2d.Move(vector: TVertice);
begin
    Position.Subtract(vector);
end;

procedure TCamera2d.MoveTo(APosition: TVertice);
begin
    Position.x := APosition.x;
    Position.y := APosition.y;
    Position.z := APosition.z;
end;

procedure TCamera2d.Reset;
begin
    Position.Reset;
    Depth := 512;
end;

constructor TCamera2d.Create(x,y,z: double);
begin
    fPosition := TVertice.Create(x,y,z);
    MinZoom := 1;
    MaxZoom := 4096;
    Depth := 512;
end;

destructor TCamera2d.Destroy;
begin
    FreeAndNil(fPosition);
    inherited;
end;

function TCamera2d.Clone: TCamera;
begin
    result := TCamera2d.Create(0,0,0);
    result.Assign(self);
end;

procedure TCamera2d.Assign( const aCamera: TCamera );
var camera: TCamera2D;
begin
    if Assigned( aCamera ) then begin
        if aCamera is TCamera2D then
        begin
            camera := TCamera2D( aCamera );
            fPosition.Assign( camera.Position );
            fDepth := camera.Depth;
            fMinZoom := camera.MinZoom;
            fMaxZoom := camera.MaxZoom;
        end
        else
            raise Exception.Create( Format( 'Cannot assign a %s to a 2D camera.', [aCamera.ClassName] ));
    end;
end;

// ### TCameraUV ###############################################################

function TCameraUV.Clone: TCamera;
begin
    result := TCameraUV.Create;
    result.Assign( self );
end;

procedure TCameraUV.Assign( const aCamera: TCamera );
var camera: TCameraUV;
begin
    if Assigned( aCamera ) then begin
        if aCamera is TCameraUV then
        begin
            camera := TCameraUV( aCamera );
            fCenterU := camera.CenterU;
            fCenterV := camera.CenterV;
            fScale := camera.Scale;
        end
        else
            raise Exception.Create( Format( 'Cannot assign a %s to a UV camera.', [aCamera.ClassName] ));
    end;
end;

procedure TCameraUV.Zoom(const aDelta: integer);
var delta: integer;
begin
    delta := aDelta;
    if toggleMouse then delta := -delta;

    // Wrap this in a convenience function...
    fScale := fScale * ( 1 + delta / 256 );

    if fScale < DEFAULT_UV_EDITOR_SCALE then fScale := DEFAULT_UV_EDITOR_SCALE;
    if fScale > 1 / DEFAULT_UV_EDITOR_SCALE then fScale := 1 / DEFAULT_UV_EDITOR_SCALE;
end;

procedure TCameraUV.FitToWindow( const aWidth, aHeight: integer );
const
    FIT_MARGIN = 1.03;
begin
    fCenterU := 0.5;
    fCenterV := 0.5;

    if aWidth < aHeight then
        fScale := aWidth * DEFAULT_UV_EDITOR_SCALE / FIT_MARGIN
    else
        fScale := aHeight * DEFAULT_UV_EDITOR_SCALE / FIT_MARGIN;
end;

procedure TCameraUV.Move( const aDeltaX, aDeltaY: integer );
var delta: TUVCoordinate;
begin
    delta.U := aDeltaX * DEFAULT_UV_EDITOR_SCALE / fScale;
    delta.V := aDeltaY * DEFAULT_UV_EDITOR_SCALE / fScale;
    fCenterU := fCenterU - delta.U;
    fCenterV := fCenterV - delta.V;
end;



// ### TCamera3d ###############################################################

function TCamera3d.Clone: TCamera;
begin
    result := TCamera3D.Create( 0, 0, 0 );
    result.Assign( self );
end;

procedure TCamera3d.Assign(const aCamera: TCamera);
var camera: TCamera3D;
begin
    if Assigned( aCamera ) then begin
        if aCamera is TCamera3D then
        begin
            camera := TCamera3D( aCamera );
            fPosition.Assign( camera.Position );
            fRotation.Assign( camera.Rotation );
            fFOV := camera.FOV;
        end
        else
            raise Exception.Create( Format( 'Cannot assign a %s to a 3D camera.', [aCamera.ClassName] ));
    end;
end;

function TCamera3d.GetMatrix: TMatrix4x4;
begin
    result := Matrix4x4EulerSetup( Rotation.X, Rotation.Y, Rotation.Z, -Position.X, -Position.Y, -Position.Z );
end;

procedure TCamera3d.Move(vector: TVertice);
begin
    Position.Subtract(vector);
end;

procedure TCamera3d.MoveRelative(const aVector: TVertice);
var vector: TVertice;
    matrix: TMatrix;
begin
    matrix := MatrixEulerSetup(Round(Rotation.x), Round(Rotation.y), Round(Rotation.z), 0, 0, 0);

    // Multiply direction _transpose_ of rotation matrix
    vector := TVertice.Create(
        matrix[0, 0] * aVector.X + matrix[1, 0] * aVector.Y + matrix[2, 0] * aVector.Z,
        matrix[0, 1] * aVector.X + matrix[1, 1] * aVector.Y + matrix[2, 1] * aVector.Z,
        matrix[0, 2] * aVector.X + matrix[1, 2] * aVector.Y + matrix[2, 2] * aVector.Z
    );

    Move(vector);

    FreeAndNil(vector);
    FreeAndNil(matrix);
end;

procedure TCamera3d.Rotate(const anX, anY, aZ: integer);
begin
    Rotation.x := Rotation.x+anX;
    if Rotation.x > 359 then Rotation.x := 0;
    if Rotation.x < 0 then Rotation.x := 359;

    Rotation.y := Rotation.y+anY;
    if Rotation.y > 359 then Rotation.y := 0;
    if Rotation.y < 0 then Rotation.y := 359;

    Rotation.z := Rotation.z+aZ;
    if Rotation.z > 359 then Rotation.z := 0;
    if Rotation.z < 0 then Rotation.z := 359;
end;

procedure TCamera3d.RotateWithAdj(const anX, anY, aZ: Single);
begin
    Rotation.x := Rotation.x+anX;
    if Rotation.x >= 360 then Rotation.x := Rotation.x - 360;
    if Rotation.x < 0 then Rotation.x := 360 + Rotation.x;

    Rotation.y := Rotation.y+anY;
    if Rotation.y >= 360 then Rotation.y := Rotation.y - 360;
    if Rotation.y < 0 then Rotation.y := 360 + Rotation.y;

    Rotation.z := Rotation.z+aZ;
    if Rotation.z >= 360 then Rotation.z := Rotation.z - 360;
    if Rotation.z < 0 then Rotation.z := 360 + Rotation.z;
end;

procedure TCamera3d.MoveRelative(const anX, aY, aZ: single);
var vector: TVertice;
begin
    vector := TVertice.Create(anX, aY, aZ);
    try
        MoveRelative(vector);
    finally
        FreeAndNil(vector);
    end;
end;

procedure TCamera3d.Orbit(const anX, aY, aZ: Single;
                          const aOrbitSelectedObject: Boolean;
                          const aSelectedObjectX, aSelectedObjectY, aSelectedObjectZ : Single;
                          const aConstrainToX,aConstrainToY: Boolean);
var
    x,{y,}z    : Single;
    rx,{ry,}rz : Single;
    
    procedure RotatePoint_YAxis(angle,x,z,ox,oz : Single; var rx,rz : Single);
    begin
        angle := angle * PI / 180;
        x     := x - ox;
        z     := z - oz;
        rx    := x * Cos(angle) - z * Sin(angle) + ox;
        rz    := x * Sin(angle) + z * Cos(angle) + oz;
    end;

    //todo: include this when x-axis orbit is working along with y-axis orbit
{    procedure RotatePoint_XAxis(angle,z,y,oz,oy : Single; Var rz,ry : Single);
    begin
        angle := angle * PI / 180;
        z     := z - oz;
        y     := y - oy;
        rz    := z * Cos(angle) - y * Sin(angle) + oz;
        ry    := z * Sin(angle) + y * Cos(angle) + oy;
    end;}
begin
    // default to world origin for orbit point
    if aOrbitSelectedObject then begin // use selected object center as orbit point
        x := aSelectedObjectX;
        // y := aSelectedObjectY;
        z := aSelectedObjectZ;
    end else begin
        x := 0;
        // y := 0;
        z := 0;
    end;

    //todo: include this when x-axis orbit is working along with y-axis orbit
//    if aConstrainToY or (not aConstrainToX and not aConstrainToY) then begin
    begin
        RotatePoint_YAxis(-anX,Position.x,Position.z,x,z,rx,rz);
        Position.x := rx;
        Position.z := rz;
        RotateWithAdj(0, -anX, 0);
    end;

    //todo: include this when x-axis orbit is working along with y-axis orbit
    {
    if aConstrainToX or (not aConstrainToX and not aConstrainToY) then begin
         RotatePoint_XAxis(-aY,Position.z,Position.y,z,y,rz,ry);
         Position.z := rz;
         Position.y := ry;
         RotateWithAdj(-aY, 0, 0);
    end;}
end;

procedure TCamera3d.Reset;
begin
    Position.x := 0;
    Position.y := 512;
    Position.z := 512;

    fRotation.X := 0;
    fRotation.Y := 0;
    fRotation.Z := 0;
end;

constructor TCamera3d.Create(x,y,z: double);
begin
    fPosition := TVertice.Create(x,y,z);
    fRotation := TVertice.Create(0,0,0);
end;

destructor TCamera3d.Destroy;
begin
    FreeAndNil(fPosition);
    FreeAndNil(fRotation);
    inherited;
end;

end.

