{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_DDSSupport;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  Native Software DDS loader. S3TC (DXTc) decoding based on specifications
  on wikipedia.

  TODO:
  [/] Support for R8G8B8
  [/] Support for A8R8G8B8
  [/] Support for DXT1
  [/] Support for DXT3
  [/] Support for DXT5
  [ ] Support for DXT2 - find sample file
  [ ] Support for DXT4 - find sample file
  [ ] Support for overige formaten - find sample file
  ---------------------------------------------------------------------------- }


interface

uses
    Classes, Windows,
    unit_TextureTypes;

type
    TPixelFormat = (pfUnknown, pfR8G8B8, pfA8R8G8B8, pfDXT1, pfDXT2, pfDXT3, pfDXT4, pfDXT5);
    TDXT1SourceBlock = packed array[0..15] of TTexturePixel24;
    TDXT1Block = packed record
        C0: Word;
        C1: Word;
        LUT: Array[0..3] of Byte;
    end;
    PDXT1Block = ^TDXT1Block;


const
    COMPRESSED_PIXEL_FORMATS: set of TPixelFormat = [pfDXT1, pfDXT2, pfDXT3, pfDXT4, pfDXT5];
    COMPRESSED_FORMATS_24_BIT: set of TPixelFormat = [pfDXT1];
    COMPRESSED_FORMATS_32_BIT: set of TPixelFormat = [pfDXT2, pfDXT3, pfDXT4, pfDXT5];
    BLOCK_SIZES: Array[TPixelFormat] of Integer = ( 0, 1, 1, 8, 8, 16, 16, 16);

type
    TDDSLoader = class(TObject)
    private
        fStream: TStream;

        // Various properties cached from the header
        fPixelFormat: TPixelFormat;
        fBufferSize: Integer;

        function R5G6B5ToColor(const aValue: Word): Cardinal;
        
        procedure UnpackBufferDXT1(
            const aBuffer: Pointer; const aBufferSize: integer;
            const aPixelFormat: TPixelFormat;
            const aTextureData: TTextureData );

        procedure UnpackBufferDXT3(
            const aBuffer: Pointer; const aBufferSize: integer;
            const aPixelFormat: TPixelFormat;
            const aTextureData: TTextureData );

        procedure UnpackBufferDXT5(
            const aBuffer: Pointer; const aBufferSize: integer;
            const aPixelFormat: TPixelFormat;
            const aTextureData: TTextureData );


    public
        constructor Create(aStream: TStream);

        function LoadHeader: TTextureData;
        function Load: TTextureData;
    end;

// ----------------------------------------------------------------------------

    TDDSWriter = class( TObject )
    private
        fStream: TStream;

        procedure WriteHeader(const aTextureData: TTextureData);
        procedure CompressDXT1(const aBlock: TDXT1SourceBlock; const aTarget: PDXT1Block); 

    public
        constructor Create(aStream: TStream);

        procedure Write(const aTextureData: TTextureData);
    end;


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
    SysUtils;

type
  TFourCC = Packed Array[0..3] of Char;

const
    DDS_HEADER_MAGIC = $20534444;           // "DDS "

    DDSD_CAPS               = $00000001;
    DDSD_HEIGHT             = $00000002;
    DDSD_WIDTH              = $00000004;
    DDSD_PITCH              = $00000008;
    DDSD_PIXELFORMAT        = $00001000;
    DDSD_MIPMAPCOUNT        = $00020000;
    DDSD_LINEARSIZE         = $00080000;
    DDSD_DEPTH              = $00800000;

    DDPF_ALPHAPIXELS        = $00000001;
    DDPF_FOURCC             = $00000004;
    DDPF_INDEXED            = $00000020;
    DDPF_RGB                = $00000040;

    DDSCAPS_COMPLEX         = $00000008;
    DDSCAPS_TEXTURE         = $00001000;
    DDSCAPS_MIPMAP          = $00400000;

    DDSCAPS2_CUBEMAP            = $00000200;
    DDSCAPS2_CUBEMAP_POSITIVEX  = $00000400;
    DDSCAPS2_CUBEMAP_NEGATIVEX  = $00000800;
    DDSCAPS2_CUBEMAP_POSITIVEY  = $00001000;
    DDSCAPS2_CUBEMAP_NEGATIVEY  = $00002000;
    DDSCAPS2_CUBEMAP_POSITIVEZ  = $00004000;
    DDSCAPS2_CUBEMAP_NEGATIVEZ  = $00008000;
    DDSCAPS2_VOLUME             = $00200000;

const
  D3DFMT_DXT1: TFourCC = 'DXT1';
  D3DFMT_DXT2: TFourCC = 'DXT2';
  D3DFMT_DXT3: TFourCC = 'DXT3';
  D3DFMT_DXT4: TFourCC = 'DXT4';
  D3DFMT_DXT5: TFourCC = 'DXT5';


type
    TDDPixelFormat = record
        Size: DWord;
        Flags: DWord;
        FourCC: DWord;
        BPP: DWord;
        RedMask: DWord;
        GreenMask: DWord;
        BlueMask: DWord;
        AlphaMask: DWord;
    end;
    PDDPixelFormat = ^TDDPixelFormat;

    TDDSCaps = record
        Caps1: DWord;
        Caps2: DWord;
        Caps3: DWord;
        Caps4: DWord;
    end;

    TDDColorKey = record
        LowValue: DWord;
        HighValue: DWord;
    end;

    TDDSurfaceDescription = record
        Size: DWord;
        Flags: DWord;
        Height: Dword;
        Width: DWord;
        Pitch: DWord;  // Contains size of buffer at largest mipmap size
        Depth: DWord;
        MipMapLevels: DWord;
        AlphaBitDepth: DWord;
        Reserved: Array[0..9] of DWord;
        PixelFormat: TDDPixelFormat;
        Caps: TDDSCaps;
        Reserved2: Array[0..0] of DWord;
    end;

constructor TDDSLoader.Create(aStream: TStream);
begin
    inherited Create;

    fStream := aStream;
end;

function TDDSLoader.LoadHeader: TTextureData;
var magicValue: DWord;
    surfaceDescription: TDDSurfaceDescription;
    pf: PDDPixelFormat;
    formatCode: String;
begin
    fStream.Read(magicValue, SizeOf(magicValue));
    if magicValue <> DDS_HEADER_MAGIC then
        raise Exception.Create( 'Invalid DDS header.' );

    fStream.Read(surfaceDescription, SizeOf(surfaceDescription));

    with surfaceDescription.Caps do begin
        if ( caps1 and DDSCAPS_TEXTURE ) = 0 then
            raise Exception.Create( 'DDS File does not contain a texture.' );

        if ( caps2 and DDSCAPS2_CUBEMAP ) <> 0 then
            raise Exception.Create( 'Cubemaps not supported.' );

        if ( caps2 and DDSCAPS2_VOLUME ) <> 0 then
            raise Exception.Create( 'Volume textures not supported.' );

    end;

    formatCode := '';
    pf := @surfaceDescription.PixelFormat;
    if ( pf.Flags and DDPF_FOURCC ) <> 0 then begin
        formatCode := TFourCC(pf.FourCC);
        if pf.FourCC = DWord(D3DFMT_DXT1) then
            fPixelFormat := pfDXT1
        else if pf.FourCC = DWord(D3DFMT_DXT3) then
            fPixelFormat := pfDXT3
        else if pf.FourCC = DWord(D3DFMT_DXT5) then
            fPixelFormat := pfDXT5;
    end
    else if ( pf.Flags and DDPF_RGB ) <> 0 then begin
        if  ( pf.RedMask = $FF0000 ) and
            ( pf.Greenmask = $FF00 ) and
            ( pf.BlueMask = $FF ) then begin
            if ( pf.BPP = 32 ) and ( pf.AlphaMask = $FF000000 ) then
                fPixelFormat := pfA8R8G8B8
            else if pf.BPP = 24 then
                fPixelFormat := pfR8G8B8;
        end;
    end;
    if fPixelFormat = pfUnknown then begin
        if formatCode = '' then formatCode := 'unknown';
        raise Exception.Create( Format( 'DDS pixel format (%s) not supported.', [formatCode] ));
    end;
    
    result.Width := surfaceDescription.Width;
    result.Height := surfaceDescription.Height;
    if fPixelFormat in COMPRESSED_PIXEL_FORMATS then
        result.BitsPerPixel := 32
    else
        result.BitsPerPixel := surfaceDescription.PixelFormat.BPP;

    if surfaceDescription.Flags and DDSD_LINEARSIZE = 0 then
        raise Exception.Create( 'DDS does not contain pitch information.' );

    fBufferSize := surfaceDescription.Pitch;

    result.Pixels := nil;
end;

function TDDSLoader.Load: TTextureData;
var buffer: PChar;
begin
    result := LoadHeader;

    Setlength(result.Pixels, result.Width * result.Height * result.BitsPerPixel div 8);

    GetMem(buffer, fBufferSize);
    fStream.Read(buffer^, fBufferSize);
    case fPixelFormat of
        pfR8G8B8,
        pfA8R8G8B8: move(buffer^, result.Pixels[0], fBufferSize);
        pfDXT1:     UnpackBufferDXT1(buffer, fBufferSize, fPixelFormat, result);
        pfDXT3:     UnpackBufferDXT3(buffer, fBufferSize, fPixelFormat, result);
        pfDXT5:     UnpackBufferDXT5(buffer, fBufferSize, fPixelFormat, result);
    else
        raise Exception.Create( 'Unsupported pixelformat.' );
    end;

    FreeMem(buffer);
end;

type
    T32BitColor = packed record
        B, G, R, A: Byte;
    end;
    P32BitColor = ^T32BitColor;

    T24BitColor = packed record
        B, G, R: Byte;
    end;
    P24BitColor = ^T24BitColor;

    TDXT3Block = packed record
        Alpha: Array[0..3] of Word;
        C0: Word;
        C1: Word;
        LUT: Array[0..3] of Byte;
    end;
    PDXT3Block = ^TDXT3Block;

    TDXT5Block = packed record
        A0: Byte;
        A1: Byte;
        AlphaLUT: Array[0..2] of Word;
        C0: Word;
        C1: Word;
        LUT: Array[0..3] of Byte;
    end;
    PDXT5Block = ^TDXT5Block;

function TDDSLoader.R5G6B5ToColor(const aValue: Word): Cardinal;
begin
    result :=
        ( aValue and $F800 ) shl 8 or
        ( aValue and $7E0 ) shl 5 or
        ( aValue and $1F ) shl 3;
end;

// Upacks the contents of the buffer according to the S3TC (DXT1-5) specs
procedure TDDSLoader.UnpackBufferDXT1(
    const aBuffer: Pointer; const aBufferSize: integer;
    const aPixelFormat: TPixelFormat;
    const aTextureData: TTextureData );
var block: PDXT1Block;
    x, y, subY: integer;
    p0, p1, p2, p3: P32BitColor;
    lutRow: byte;
    colors: array[0..3] of Cardinal;
    targetPixel: ^Cardinal;

begin
    block := aBuffer;

    targetPixel := @aTextureData.Pixels[0];

    for y := 0 to aTextureData.Height div 4 - 1 do begin
        for x := 0 to aTextureData.Width div 4 - 1 do begin
            colors[0] := R5G6B5ToColor(block.C0);
            colors[1] := R5G6B5ToColor(block.C1);

            p0 := @colors[0];
            p1 := @colors[1];
            p2 := @colors[2];
            p3 := @colors[3];

            p0.A := 255;
            p1.A := 255;

            if block.C0 > block.C1 then begin
                p2.A := 255;
                p2.R := (2 * p0.R + p1.R) div 3;
                p2.G := (2 * p0.G + p1.G) div 3;
                p2.B := (2 * p0.B + p1.B) div 3;

                p3.A := 255;
                p3.R := (p0.R + 2 * p1.R) div 3;
                p3.G := (p0.G + 2 * p1.G) div 3;
                p3.B := (p0.B + 2 * p1.B) div 3;
            end
            else begin
                p2.A := 255;
                p2.R := ( p0.R + p1.R ) div 2;
                p2.G := ( p0.G + p1.G ) div 2;
                p2.B := ( p0.B + p1.B ) div 2;

                colors[3] := 0;
            end;

            for subY := 0 to 3 do begin
                lutRow := block.LUT[subY];

                targetPixel^ := colors[lutRow and 3];
                inc(targetPixel);

                targetPixel^ := colors[(lutRow shr 2) and 3];
                inc(targetPixel);

                targetPixel^ := colors[(lutRow shr 4) and 3];
                inc(targetPixel);

                targetPixel^ := colors[lutRow shr 6];
                inc(targetPixel);

                inc(targetPixel, ( aTextureData.Width - 4 ));
            end;

            dec(targetPixel, aTextureData.Width * 4 - 4);

            inc(block);
        end;

        inc(targetPixel, aTextureData.Width * 3);
    end;
end;

procedure TDDSLoader.UnpackBufferDXT3(
    const aBuffer: Pointer; const aBufferSize: integer;
    const aPixelFormat: TPixelFormat;
    const aTextureData: TTextureData );
var block: PDXT3Block;
    x, y, subY: integer;
    p0, p1, p2, p3: P32BitColor;
    lutRow: byte;
    colors: array[0..3] of cardinal;
    targetPixel: ^cardinal;
    alphaRow: word;
    alphaMapping: array[0..15] of cardinal;

begin
    block := aBuffer;

    for x := 0 to 15 do begin
        alphaMapping[x] := cardinal(x) shl 28;
    end;

    // Special correction: alpha of $F means $FF, not $F0.
    alphaMapping[15] := $FF000000;

    targetPixel := @aTextureData.Pixels[0];

    for y := 0 to aTextureData.Height div 4 - 1 do begin
        for x := 0 to aTextureData.Width div 4 - 1 do begin
            colors[0] := R5G6B5ToColor(block.C0);
            colors[1] := R5G6B5ToColor(block.C1);

            p0 := @colors[0];
            p1 := @colors[1];
            p2 := @colors[2];
            p3 := @colors[3];

            p2.R := (2 * p0.R + p1.R) div 3;
            p2.G := (2 * p0.G + p1.G) div 3;
            p2.B := (2 * p0.B + p1.B) div 3;

            p3.R := (p0.R + 2 * p1.R) div 3;
            p3.G := (p0.G + 2 * p1.G) div 3;
            p3.B := (p0.B + 2 * p1.B) div 3;

            for subY := 0 to 3 do begin
                lutRow := block.LUT[subY];
                alphaRow := block.Alpha[subY];

                targetPixel^ := colors[lutRow and 3] or alphaMapping[alphaRow and 15];
                inc(targetPixel);

                targetPixel^ := colors[(lutRow shr 2) and 3] or alphaMapping[(alphaRow shr 4) and 15];
                inc(targetPixel);

                targetPixel^ := colors[(lutRow shr 4) and 3] or alphaMapping[(alphaRow shr 8) and 15];
                inc(targetPixel);

                targetPixel^ := colors[lutRow shr 6] or alphaMapping[alphaRow shr 12];
                inc(targetPixel);

                inc(targetPixel, ( aTextureData.Width - 4 ));
            end;

            dec(targetPixel, aTextureData.Width * 4 - 4);

            inc(block);
        end;

        inc(targetPixel, aTextureData.Width * 3);
    end;
end;

procedure TDDSLoader.UnpackBufferDXT5(
    const aBuffer: Pointer; const aBufferSize: integer;
    const aPixelFormat: TPixelFormat;
    const aTextureData: TTextureData );
var block: PDXT5Block;
    x, y, subY: integer;
    p0, p1, p2, p3: P32BitColor;
    lutRow: byte;
    colors: array[0..3] of Cardinal;
    targetPixel: ^Cardinal;
    alpha: array[0..7] of byte;
    totalAlphaLut: Int64;
    alphaLUTRow: Word;


begin
    block := aBuffer;

    targetPixel := @aTextureData.Pixels[0];

    for y := 0 to aTextureData.Height div 4 - 1 do begin
        for x := 0 to aTextureData.Width div 4 - 1 do begin
            colors[0] := R5G6B5ToColor(block.C0);
            colors[1] := R5G6B5ToColor(block.C1);

            p0 := @colors[0];
            p1 := @colors[1];
            p2 := @colors[2];
            p3 := @colors[3];

            p2.R := (2 * p0.R + p1.R) div 3;
            p2.G := (2 * p0.G + p1.G) div 3;
            p2.B := (2 * p0.B + p1.B) div 3;

            p3.R := (p0.R + 2 * p1.R) div 3;
            p3.G := (p0.G + 2 * p1.G) div 3;
            p3.B := (p0.B + 2 * p1.B) div 3;

            totalAlphaLUT := 0;
            Move( block.AlphaLut[0], totalAlphaLut, 6);
            if totalAlphaLUT > 0 then
            begin
                totalAlphaLUT := totalALphaLut;
            end;

            alpha[0] := block.A0;
            alpha[1] := block.A1;

            if alpha[0] > alpha[1] then begin
                alpha[2] := (( 6 * alpha[0] + 1 * alpha[1]) div 7);
                alpha[3] := (( 5 * alpha[0] + 2 * alpha[1]) div 7);
                alpha[4] := (( 4 * alpha[0] + 3 * alpha[1]) div 7);
                alpha[5] := (( 3 * alpha[0] + 4 * alpha[1]) div 7);
                alpha[6] := (( 2 * alpha[0] + 5 * alpha[1]) div 7);
                alpha[7] := (( 1 * alpha[0] + 6 * alpha[1]) div 7);
            end
            else begin
                alpha[2] := (( 4 * alpha[0] + 1 * alpha[1]) div 5);
                alpha[3] := (( 3 * alpha[0] + 2 * alpha[1]) div 5);
                alpha[4] := (( 2 * alpha[0] + 3 * alpha[1]) div 5);
                alpha[5] := (( 1 * alpha[0] + 4 * alpha[1]) div 5);
                alpha[6] := 0;
                alpha[7] := $FF;
            end;

            for subY := 0 to 3 do begin
                lutRow := block.LUT[subY];
                alphaLutRow := totalAlphaLut and $0FFF;
                totalAlphaLut := totalAlphaLut shr 12;

                targetPixel^ := colors[lutRow and 3] or alpha[alphaLutRow and 7] shl 24;
                inc(targetPixel);

                targetPixel^ := colors[(lutRow shr 2) and 3] or alpha[(alphaLutRow shr 3) and 7] shl 24;
                inc(targetPixel);

                targetPixel^ := colors[(lutRow shr 4) and 3] or alpha[(alphaLutRow shr 6) and 7] shl 24;
                inc(targetPixel);

                targetPixel^ := colors[lutRow shr 6] or alpha[(alphaLutRow shr 9) and 7] shl 24;
                inc(targetPixel);

                inc(targetPixel, ( aTextureData.Width - 4 ));
            end;

            dec(targetPixel, aTextureData.Width * 4 - 4);

            inc(block);
        end;

        inc(targetPixel, aTextureData.Width * 3);
    end;
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TDDSWriter.Create(aStream: TStream);
begin
    inherited Create;

    fStream := aStream;
end;

// ----------------------------------------------------------------------------

procedure TDDSWriter.Write(const aTextureData: TTextureData);
var y, y1, x, startIndex, numBytesPerPixel: integer;
    block: TDXT1SourceBlock;
    buffer: Array of Byte;
    target: PDXT1Block;

begin
    WriteHeader( aTextureData );

    SetLength( buffer, aTextureData.Width * aTextureData.Height * 2 );
    target := @buffer[0];

    numBytesPerPixel := aTextureData.BitsPerPixel div 8;

    // Encode all 4x4 blocks in the image.
    startIndex := 0;
    for y := 0 to aTextureData.Height div 4 - 1 do begin
        for x := 0 to aTextureData.Width div 4 - 1 do begin
            for y1 := 0 to 3 do begin
                Move( aTextureData.Pixels[startIndex + y1 * aTextureData.Width * NumBytesPerPixel], block[y1], 4 * NumBytesPerPixel );
            end;

            CompressDXT1( block, target );

            startIndex := startIndex + 4 * NumBytesPerPixel;
            Inc( target );
        end;

        startIndex := startIndex + 3 * aTextureData.Width * NumBytesPerPixel;
    end;

    fStream.Write( buffer[0], Length( buffer ));
end;

// ----------------------------------------------------------------------------

procedure TDDSWriter.WriteHeader(const aTextureData: TTextureData);
var
  magicValue: DWord;
  surfaceDescription: TDDSurfaceDescription;

begin
  Assert( aTextureData.BitsPerPixel = 24 );

  magicValue := DDS_HEADER_MAGIC;
  fStream.Write( magicValue, SizeOf( DWord ));

  FillChar( surfaceDescription, SizeOf( surfaceDescription ), 0 );
  surfaceDescription.Size := SizeOf( surfaceDescription );
  surfaceDescription.Flags :=
      DDSD_LINEARSIZE or DDSD_CAPS or DDSD_PIXELFORMAT or DDSD_WIDTH or DDSD_HEIGHT;
  surfaceDescription.Height := aTextureData.Height;
  surfaceDescription.Width := aTextureData.Width;
  surfaceDescription.Pitch := aTextureData.Height * aTextureData.Width * 2;
  surfaceDescription.Caps.Caps1 := DDSCAPS_TEXTURE;

  surfaceDescription.PixelFormat.Size := SizeOf( TDDPixelFormat );
  surfaceDescription.PixelFormat.Flags := DDPF_FOURCC;
  surfaceDescription.PixelFormat.FourCC := DWord(D3DFMT_DXT1);
  surfaceDescription.PixelFormat.BPP := 24;
  surfaceDescription.PixelFormat.RedMask := $FF0000;
  surfaceDescription.PixelFormat.GreenMask := $FF00;
  surfaceDescription.PixelFormat.BlueMask := $FF;

  fStream.Write( surfaceDescription, SizeOf( surfaceDescription ));
end;

// ----------------------------------------------------------------------------

procedure TDDSWriter.CompressDXT1(const aBlock: TDXT1SourceBlock; const aTarget: PDXT1Block);
var i: integer;
begin
    aTarget.C0 :=
        ( aBlock[0][2] shr 3 ) shl 11 +
        ( aBlock[0][1] shr 2 ) shl 5 +
        ( aBlock[0][0] shr 3 );
    aTarget.C1 := 0;
    for i := 0 to 3 do aTarget.Lut[i] := 1;
end;

end.

