{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_DMFSceneReader;

{$MODE Delphi}

interface

uses
  Classes, SysUtils, StrUtils, FileUtil,
  frmProgressForm,
  unit_Enums,
  unit_Scene,
  unit_MaterialManager;

type
  TDMFSceneReader = class(TObject)
  private
    // misc methods
    function Validate090Polygon(aStringList: TStringList; aStartPosition: integer): boolean;
    procedure GenerateDummyTexture(const aFilename: string);
    function MakeValidTextureFileName(const aFileName: string; const aWarnings: TStringList): string;
    function HexToRGBColor(const aHexValue: string): TRGBColor;
    // format based Set functions (used by plugin-system+load routines)
    function SetObjects090(numObjects: integer; strings: TStringList; offset: longint; aScene: TScene; shouldReplace: boolean; aProgressForm: TfrmProgress; materials: TMaterialManager = nil): longint;
    function SetObjects110(numObjects: integer; strings: TStringList; offset: longint; aScene: TScene; shouldReplace: boolean; aProgressForm: TfrmProgress; materials: TMaterialManager = nil): longint;
    function SetLights090(numLights: integer; strings: TStringList; offset: longint; aScene: TScene; shouldReplace: boolean; aProgressForm: TfrmProgress): longint;
    function SetMaterials090(numMaterials: integer; strings: TStringList; offset: longint; aScene: TScene; aProgressForm: TfrmProgress; newMaterials: TMaterialManager = nil): longint;
    function SetMaterials090Ex(materials: TMaterialManager; numMaterials: integer; strings: TStringList; offset: longint; aScene: TScene; aProgressForm: TfrmProgress; newMaterials: TMaterialManager = nil): longint;
    function SetMaterials091(numMaterials: integer; strings: TStringList; offset: longint; aScene: TScene; aProgressForm: TfrmProgress; aWarnings: TStringList; newMaterials: TMaterialManager = nil): longint;
    function SetMaterials091Ex(materials: TMaterialManager; numMaterials: integer; strings: TStringList; offset: longint; aScene: TScene; aProgressForm: TfrmProgress; aWarnings: TStringList; newMaterials: TMaterialManager = nil): longint;
    // load methods for specific DMF (DeleD Map File) files
    procedure Load091(aFileName: string; aScene: TScene; aProgressForm: TfrmProgress; aWarnings: TStringList);
    procedure Load110(aFileName: string; aScene: TScene; aProgressForm: TfrmProgress; aWarnings: TStringList);
  public
    // generic Set functions; they write data to the scene; this data is provided
    // in stringlists and has a specific format
    function SetObjects(format: string; numObjects: integer; strings: TStringList; offset: longint; aScene: TScene; shouldReplace: boolean; aProgressForm: TfrmProgress; materials: TMaterialManager): boolean;
    function SetLights(format: string; numLights: integer; strings: TStringList; offset: longint; aScene: TScene; shouldReplace: boolean; aProgressForm: TfrmProgress): boolean;
    function SetMaterials(format: string; numMaterials: integer; strings: TStringList; offset: longint; aScene: TScene; aProgressForm: TfrmProgress; newMaterials: TMaterialManager): boolean;
    // generic Load method
    procedure ReadFromFile(const aFileName: string; const aScene: TScene);
  end;

implementation

uses
  Forms, Dialogs,
  unit_Variables,
  unit_Types,
  unit_Materials,
  unit_Primitives,
  unit_PrimitiveList,
  unit_Lights,
  unit_DPFLoader,
  unit_Functions,
  unit_TextureTypes,
  unit_TextureWriter,
  unit_TextureLoader,
  unit_MaterialMatcher;

// ### TDMFSceneReader #########################################################

function TDMFSceneReader.Validate090Polygon(aStringList: TStringList; aStartPosition: integer): boolean;
var i,j,numVertices,numEntries: integer;
begin
// valid polygon
// 4;2;8;9;10;11;-2.2500;2.8130;-2.2500;-2.1250;2.2500;-2.1250;2.2500;2.8130;

// invalid polygon because vertices are references more than once
// 4;2;11;10;10;11;2.2500;0.0000;2.2500;0.0000;2.2500;0.0000;2.2500;0.0000;

    Result := false;
    if Assigned(aStringList) then begin
        try
            numVertices := StrToInt(aStringList[aStartPosition]);
            numEntries := 2+numVertices+(numVertices*2);
            if aStringList.Count-aStartPosition>=numEntries then begin  // we have enough entries to check
                for i := 0 to numVertices - 2 do begin
                    for j := i + 1 to numVertices - 1 do begin
                        if aStringList[i+2+aStartPosition] = aStringList[j+2+aStartPosition] then begin
                            Result := false;
                            exit;
                        end;
                    end;
                end;
                Result := true;
            end;
        except
            Result := false;
        end;
    end;
end;

procedure TDMFSceneReader.GenerateDummyTexture(const aFileName: string);
var dir: string;
    extension: string;
    textureData: TTextureData;
    textureWriter: TTextureWriter;
begin
    dir := ExtractFileDir(aFileName);
    if ( dir <> '' ) and not ForceDirectoriesUTF8(dir) then
        raise Exception.Create(Format('Can''t auto-create texture directory "%s".',[dir]));

    extension := ExtractFileExt(aFileName);
    if TTextureWriter.SupportsFileFormat(extension) then begin
        textureData.Width := 2;
        textureData.Height := 2;
        textureData.BitsPerPixel := 24;
        SetLength(textureData.Pixels, textureData.Width * textureData.Height *
            textureData.BitsPerPixel div 8);
        FillChar(textureData.Pixels[0], Length(textureData.Pixels), Random(256));

        textureWriter := TTextureWriter.Create;
        try
            textureWriter.SaveTextureData(textureData,aFileName);
        finally
            FreeAndNil(textureWriter);
        end;
    end
    else
        raise Exception.Create(Format('Can''t generate texture for %s format',[extension]));
end;

function TDMFSceneReader.MakeValidTextureFileName(const aFileName: string; const aWarnings: TStringList): string;
var shouldGenerate: boolean;
    extension: string;
begin
    Result := aFileName;
    shouldGenerate := FALSE;

    extension := LowerCase(ExtractFileExt(aFileName));
    if not TTextureLoader.SupportsFileFormat( extension ) then begin
        Result := ChangeFileExt(aFileName, '.jpg' );

        if FileExistsUTF8(Result) then begin
            if Assigned(aWarnings) then
                aWarnings.Add( Format('Texture format "%s" not supported, using texture "%s" instead.', [extension, Result]));
        end
        else begin
            if Assigned(aWarnings) then
                aWarnings.Add( Format('Texture format "%s" not supported, generated dummy texture "%s"', [extension, Result]));

            shouldGenerate := TRUE;
        end;
    end
    else if not FileExistsUTF8(aFileName) then begin
        if Assigned(aWarnings) then
            aWarnings.Add( Format('Texture "%s" not found, generated dummy texture.', [aFileName]));

        shouldGenerate := TRUE;
    end;

    if shouldGenerate then
        GenerateDummyTexture(Result);
end;

function TDMFSceneReader.HexToRGBColor(const aHexValue:string): TRGBColor;
var rString, gString, bString: string;
    rValue, gValue, bValue: byte;
begin
    bString := Copy( aHexValue, 3, 2 );
    gString := Copy( aHexValue, 5, 2 );
    rString := Copy( aHexValue, 7, 2 );

    HexToBin(PChar(rString),@rValue,2);
    HexToBin(PChar(gString),@gValue,2);
    HexToBin(PChar(bString),@bValue,2);

    Result.R := rValue / 255;
    Result.G := gValue / 255;
    Result.B := bValue / 255;
end;

function TDMFSceneReader.SetObjects090(numobjects: integer; strings: TStringList; offset: longint; aScene: TScene; shouldReplace: boolean; aProgressForm: TfrmProgress; materials: TMaterialManager = nil): longint;
var obj: TBasePrimitive;
    tempPL: TPrimitiveList;
    i,j,k,l,numPolys,numVerts,numEntries,mid: integer;
    position: longint;
    templockUV: boolean;
    polygon: TPolygon;
    uvcoord: TUVCoord;
    uvCoords: TUVList;
    material: TMaterial;
begin
    polygon := nil;
    position := -1;

    if Assigned(aProgressForm) then begin
        aProgressForm.Description := 'Loading objects.';
        aProgressForm.Max := numobjects;
    end;

    tempPL := TPrimitiveList.Create(false);

    if not Assigned(materials) then
        materials := aScene.MaterialManager;

    try
        try
            // load all objects into a temporary primitivelist
            position := offset+1;
            for i := 0 to numobjects-1 do begin
                if Assigned( aProgressForm ) then
                    aProgressForm.Position := i+1;

                case StrToInt(strings[position+2]) of    // should be handled with classtype?
                    0:  obj := TRectangle.Create;
                    1:  obj := TCube.Create;
                    2:  obj := TPyramid.Create;
                    3:  obj := TCylinder.Create;
                    4:  obj := TPolyline.Create;
                    6:  obj := TSphere.Create;
                    9:  obj := TUserDefined.Create;
                    12: obj := TGrid.Create;
                    13: obj := TPointLight.Create;
                    14: obj := TDirectionalLight.Create;
                    15: obj := TSphere.Create;
                    16: obj := TPrimitive.Create;
                    17: obj := TLight.Create;
                    18: obj := TEllipse.Create;
                    19: obj := TText.Create;
                    //22: obj := TPath.Create;
                    else
                    obj := TUserDefined.Create;
                end;
                tempPL.AddPrimitive(obj);
                obj.ID := StrToInt(strings[position+1]);
                obj.Name := strings[position];
                obj.ObjectType := TObjectType(StrToInt(strings[position+2]));
                if obj.ObjectType=otUndefined then obj.objecttype := otUserDefined;
                obj.Visible := StrToBool(strings[position+3]);
                obj.SnapMethod := TSnapMethod(StrToInt(strings[position+4]));
                obj.Groupid := StrToInt(strings[position++5]);
                templockUV := StrToBool(strings[position+6]);
                obj.AutoUV := false;                                                // false so that UV coords aren't recalculated when loading map
                obj.Tag := strings[position+7];

                // read vertices
                inc(position,8);
                numverts := StrToInt(strings[position]);
                for j := 0 to numverts-1 do
                    obj.Vertices.Add(TVertice.Create(StrToFloat(strings[position+(j*3)+1]),StrToFloat(strings[position+(j*3)+2]),StrToFloat(strings[position+(j*3)+3])));

                // read polygons
                inc(position,(numverts*3)+1);
                numpolys := StrToInt(strings[position]);
                inc(position);

                for j := 0 to numpolys-1 do begin

                    numVerts := StrToInt(strings[position]);
                    numEntries := 2+numVerts+(numVerts*2);

                    if not Validate090Polygon(strings,position) then
                        inc(position,numEntries)
                    else begin
                        polygon := TPolygon.Create;

                        // materialID, used to assign a material to this polygon
                        // assign a material from the (provided or default) materialmanager to this polygon
                        mid := StrToInt(strings[position+1]);
                        if mid<0 then mid := 0;
                        material := materials.Materials[mid];
                        polygon.Material := material;

                        for k := 0 to numverts-1 do
                            obj.Polygon_AddVertex(polygon,obj.Vertices[StrToInt(strings[position+2+k])]);

                        inc(position,2+numverts);

                        for l := 0 to material.NumTextureLayers - 1 do begin
                            uvCoords := TUVList.Create;
                            polygon.UVCoordinates[l] := uvCoords;
                            if Assigned(material.TextureLayers[l]) then begin
                                for k := 0 to polygon.numvertices-1 do begin
                                    uvcoord := TUVcoord.Create(StrToFloat(strings[position]),1-StrToFloat(strings[position+1]));
                                    uvCoords.Add(uvcoord);
                                    inc(position,2);
                                end;
                            end;
                        end;
                        obj.Polygons.Add(polygon);
                    end;
                end;
                obj.AutoUV := templockUV;                                       // workaround
            end;

        except
            on E: Exception do begin
                if Assigned(tempPL) then begin                                  // free temporary primitivelist and its objects
                    tempPL.OwnsObjects := true;
                    FreeAndNil(tempPL);
                    FreeAndNil(polygon);                                        // there might be a floating poly
                end;
                raise Exception.Create('TDMFSceneReader.SetObjects090: '+E.message);
            end;
        end;

        // if everything went OK, copy pointers from tempPM to scene primitivelist
        // With large amounts of objects, Extract(0) is extremely slow! Change some day.
        for i := 0 to tempPL.Count-1 do begin
            obj := tempPL.Extract(0);
            if shouldReplace then
                aScene.Primitives.AddUniquePrimitive(obj,true)
            else
                aScene.Primitives.AddPrimitive(obj,false);
        end;

    finally
        FreeAndNil(tempPL);         // free temp primitivelist; it didn't own the objects so objects are still alive in the real pm
    end;

    Result := position;
end;

function TDMFSceneReader.SetObjects110(numobjects: integer; strings: TStringList; offset: longint; aScene: TScene; shouldReplace: boolean; aProgressForm: TfrmProgress; materials: TMaterialManager = nil): longint;
var obj: TBasePrimitive;
    tempPL: TPrimitiveList;
    i,j,k,l,numPolys,numVerts,numEntries,mid: integer;
    position: longint;
    templockUV: boolean;
    polygon: TPolygon;
    uvcoord: TUVCoord;
    uvCoords: TUVList;
    material: TMaterial;
begin
    polygon := nil;
    position := -1;

    if Assigned(aProgressForm) then begin
        aProgressForm.Description := 'Loading objects.';
        aProgressForm.Max := numobjects;
    end;

    tempPL := TPrimitiveList.Create(false);

    if not Assigned(materials) then
        materials := aScene.MaterialManager;

    try
        try
            // load all objects into a temporary primitivelist
            position := offset+1;
            for i := 0 to numobjects-1 do begin
                if Assigned( aProgressForm ) then
                    aProgressForm.Position := i+1;

                case StrToInt(strings[position+2]) of    // should be handled with classtype?
                    0:  obj := TRectangle.Create;
                    1:  obj := TCube.Create;
                    2:  obj := TPyramid.Create;
                    3:  obj := TCylinder.Create;
                    4:  obj := TPolyline.Create;
                    6:  obj := TSphere.Create;
                    9:  obj := TUserDefined.Create;
                    12: obj := TGrid.Create;
                    13: obj := TPointLight.Create;
                    14: obj := TDirectionalLight.Create;
                    15: obj := TSphere.Create;
                    16: obj := TPrimitive.Create;
                    17: obj := TLight.Create;
                    18: obj := TEllipse.Create;
                    19: obj := TText.Create;
                    //22: obj := TPath.Create;
                    else
                    obj := TUserDefined.Create;
                end;
                tempPL.AddPrimitive(obj);
                obj.ID := StrToInt(strings[position+1]);
                obj.Name := strings[position];
                obj.ObjectType := TObjectType(StrToInt(strings[position+2]));
                if obj.ObjectType=otUndefined then obj.objecttype := otUserDefined;
                obj.Visible := StrToBool(strings[position+3]);
                obj.SnapMethod := TSnapMethod(StrToInt(strings[position+4]));
                obj.Groupid := StrToInt(strings[position++5]);
                templockUV := StrToBool(strings[position+6]);
                obj.AutoUV := false;                                                // false so that UV coords aren't recalculated when loading map
                obj.Tag := strings[position+7];

                // read vertices
                inc(position,8);
                numverts := StrToInt(strings[position]);
                for j := 0 to numverts-1 do
                    obj.Vertices.Add(TVertice.Create(StrToFloat(strings[position+(j*3)+1]),StrToFloat(strings[position+(j*3)+2]),StrToFloat(strings[position+(j*3)+3])));

                // read polygons
                inc(position,(numverts*3)+1);
                numpolys := StrToInt(strings[position]);
                inc(position);

                for j := 0 to numpolys-1 do begin

                    numVerts := StrToInt(strings[position]);

                    mid := StrToInt(strings[position+1]);
                    if mid>materials.numMaterials then mid := 0;                // TODO: do we really want this functionality?
                    material := materials.Materials[mid];

                    if not Validate090Polygon(strings,position) then begin
                        numEntries := 2+numVerts+((numVerts*material.NumUsedTextureLayers)*2);
                        inc(position,numEntries)
                    end
                    else begin

                        polygon := TPolygon.Create;

                        // materialID, used to assign a material to this polygon
                        // assign a material from the (provided or default) materialmanager to this polygon
                        polygon.Material := material;

                        for k := 0 to numverts-1 do
                            obj.Polygon_AddVertex(polygon,obj.Vertices[StrToInt(strings[position+2+k])]);

                        inc(position,2+numverts);

                        for l := 0 to material.NumTextureLayers - 1 do begin
                            uvCoords := TUVList.Create;
                            polygon.UVCoordinates[l] := uvCoords;
                            if Assigned(material.TextureLayers[l]) then begin
                                for k := 0 to polygon.numvertices-1 do begin
                                    uvcoord := TUVcoord.Create(StrToFloat(strings[position]),StrToFloat(strings[position+1]));
                                    uvCoords.Add(uvcoord);
                                    inc(position,2);
                                end;
                            end;
                        end;
                        obj.Polygons.Add(polygon);
                        polygon := nil;
                    end;
                end;
                obj.AutoUV := templockUV;                                       // workaround
            end;

        except
            on E: Exception do begin
                if Assigned(tempPL) then begin                                  // free temporary primitivelist and its objects
                    tempPL.OwnsObjects := true;
                    FreeAndNil(tempPL);
                    FreeAndNil(polygon);                                        // there might be a floating poly
                end;
                raise Exception.Create('TDMFSceneReader.SetObjects110: '+E.message);
            end;
        end;

        // if everything went OK, copy pointers from tempPM to scene primitivelist
        // we only trigger a refresh for the last object (because it updates the gui... but it also sets scene.hasChanged!) 
        // NOTE JC: why don't we use something like Assign here?
        for i := 0 to tempPL.Count-1 do begin
            obj := tempPL.Extract(0);
            if shouldReplace then
                aScene.Primitives.AddUniquePrimitive(obj, true, (i = tempPL.Count-1))
            else
                aScene.Primitives.AddPrimitive(obj, false, (i = tempPL.Count-1));
        end;

    finally
        FreeAndNil(tempPL);         // free temp primitivelist; it didn't own the objects so objects are still alive in the real pm
    end;

    Result := position;
end;

function TDMFSceneReader.SetLights090(numlights: integer; strings: TStringList; offset: longint; aScene: TScene; shouldReplace: boolean; aProgressForm: TfrmProgress): longint;
var i:integer;
    position: longint;
    tempPL: TPrimitiveList;
    pointlight: TPointlight;
    obj: TBasePrimitive;
    aDPFLoader: TDPFLoader;
    dirlight: TDirectionalLight;
    spotlight: TSpotlight;
begin
    position := -1;

    if Assigned(aProgressForm) and (numlights>0) then begin
        aProgressForm.Description := 'Loading lights.';
        aProgressForm.Max := numlights;
    end;

    tempPL := TPrimitiveList.Create(false);

    try

        try

            position := offset;
            for i := 0 to numlights-1 do begin
                if Assigned( aProgressForm ) then
                    aProgressForm.Position := i+1;

                inc(position);

                case StrToInt(strings[position]) of
                    1:  begin
                            pointlight := TPointLight.Create;
                            pointlight.constant := StrToFloat(strings[position+1]);         // "with pointlight do" doesn't work?
                            pointlight.linear := StrToFloat(strings[position+2]);
                            pointlight.quadratic := StrToFloat(strings[position+3]);
                            pointlight.range := StrToInt(strings[position+4]);
                            pointlight.position.x := StrToFloat(strings[position+5]);
                            pointlight.position.y := StrToFloat(strings[position+6]);
                            pointlight.position.z := StrToFloat(strings[position+7]);
                            pointlight.flash := StrToInt(strings[position+8]);
                            pointlight.active := StrToBool(strings[position+9]);
                            pointlight.diffuse := StringComponentsToRGBColor(strings[position+10],strings[position+11],strings[position+12]);
                            pointlight.specular := StringComponentsToRGBColor(strings[position+13],strings[position+14],strings[position+15]);
                            pointlight.intensity := StrToInt(strings[position+16]);
                            pointlight.id := StrToInt(strings[position+17]);
                            aDPFLoader := TDPFLoader.Create;
                            aDPFLoader.LoadDPF('pointlight',true,dir_texture,pointlight.position,aScene.MaterialManager,TPrimitive(pointLight));
                            FreeAndNil(aDPFLoader);
                            //LoadPrefab(TBasePrimitive(pointlight),true,'pointlight',pointlight.position,false,false,materialManager);
                            pointlight.name := strings[position+18];            // LoadPrefab Assigned default object name
                            tempPL.AddPrimitive(pointlight);
                            inc(position,18);
                        end;

                    2:  begin
                            dirlight := TDirectionalLight.Create;
                            dirlight.direction.x := StrToFloat(strings[position+1]);
                            dirlight.direction.y := StrToFloat(strings[position+2]);
                            dirlight.direction.z := StrToFloat(strings[position+3]);
                            dirlight.position.x := StrToFloat(strings[position+4]);
                            dirlight.position.y := StrToFloat(strings[position+5]);
                            dirlight.position.z := StrToFloat(strings[position+6]);
                            dirlight.flash := StrToInt(strings[position+7]);
                            dirlight.active := StrToBool(strings[position+8]);
                            dirlight.diffuse := StringComponentsToRGBColor(strings[position+9],strings[position+10],strings[position+11]);
                            dirlight.specular := StringComponentsToRGBColor(strings[position+12],strings[position+13],strings[position+14]);
                            dirlight.intensity := StrToInt(strings[position+15]);
                            dirlight.id := StrToInt(strings[position+16]);
                            // now load prefab from resourcefile, adding position to vertices
                            aDPFLoader := TDPFLoader.Create;
                            aDPFLoader.LoadDPF('directionallight',true,dir_texture,dirlight.position,aScene.MaterialManager,TPrimitive(dirLight));
                            FreeAndNil(aDPFLoader);
                            //LoadPrefab(TBasePrimitive(dirlight),true,'directionallight',dirlight.position,false,false,materialManager);
                            dirlight.name := strings[position+17];              // LoadPrefab Assigned default object name
                            dirLight.FaceDirection;
                            tempPL.AddPrimitive(dirlight);
                            inc(position,17);
                        end;

                    3:  begin
                            spotlight := TSpotlight.Create;
                            spotlight.constant := StrToFloat(strings[position+1]);
                            spotlight.linear := StrToFloat(strings[position+2]);
                            spotlight.quadratic := StrToFloat(strings[position+3]);
                            spotlight.range := StrToInt(strings[position+4]);
                            spotlight.direction.x := StrToFloat(strings[position+5]);
                            spotlight.direction.y := StrToFloat(strings[position+6]);
                            spotlight.direction.z := StrToFloat(strings[position+7]);
                            spotlight.innerangle := StrToInt(strings[position+8]);
                            spotlight.outerangle := StrToInt(strings[position+9]);
                            spotlight.falloff := StrToInt(strings[position+10]);
                            spotlight.position.x := StrToFloat(strings[position+11]);
                            spotlight.position.y := StrToFloat(strings[position+12]);
                            spotlight.position.z := StrToFloat(strings[position+13]);
                            spotlight.flash := StrToInt(strings[position+14]);
                            spotlight.active := StrToBool(strings[position+15]);
                            spotlight.diffuse := StringComponentsToRGBColor(strings[position+16],strings[position+17],strings[position+18]);
                            spotlight.specular := StringComponentsToRGBColor(strings[position+19],strings[position+20],strings[position+21]);
                            spotlight.intensity := StrToInt(strings[position+22]);
                            spotlight.id := StrToInt(strings[position+23]);
                            aDPFLoader := TDPFLoader.Create;
                            aDPFLoader.LoadDPF('spotlight',true,dir_texture,spotlight.position,aScene.MaterialManager,TPrimitive(spotLight));
                            FreeAndNil(aDPFLoader);
                            //LoadPrefab(TBasePrimitive(spotlight),true,'spotlight',spotlight.position,false,false,materialManager);
                            spotlight.name := strings[position+24];             // LoadPrefab Assigned default object name
                            spotlight.FaceDirection;
                            tempPL.AddPrimitive(spotlight);
                            inc(position,24);
                        end;
                end;
            end;

        except
            on E: Exception do begin
                if Assigned(tempPL) then begin                                  // free temporary primitivelist and its objects
                    tempPL.OwnsObjects := true;
                    FreeAndNil(tempPL);
                end;
                raise Exception.Create('TDMFSceneReader.SetLights090: '+E.message);
            end;
        end;

        // if everything went OK, copy pointers from tempPL to scene primitivelist
        for i := 0 to tempPL.Count-1 do begin
            obj := tempPL.Extract(0);
            if shouldReplace then
                aScene.Primitives.AddUniquePrimitive(obj,true)
            else
                aScene.Primitives.AddPrimitive(obj, false);
        end;

    finally
        FreeAndNil(tempPL);     // free temp primitivelist; it didn't own the objects so objects are still accessible via the real pm
    end;

    Result := position;
end;

function TDMFSceneReader.SetMaterials090(nummaterials: integer; strings: TStringList; offset: longint; aScene: TScene; aProgressForm: TfrmProgress; newMaterials: TMaterialManager = nil): longint;
begin
   Result :=  SetMaterials090Ex(aScene.MaterialManager, nummaterials, strings, offset, aScene, aProgressForm, newMaterials);
end;

// returns a set of newMaterials
function TDMFSceneReader.SetMaterials090Ex(materials: TMaterialManager; nummaterials: integer; strings: TStringList; offset: longint; aScene: TScene; aProgressForm: TfrmProgress; newMaterials: TMaterialManager = nil): longint;
var i,j,numlayers:integer;
    position: longint;
    tempMaterial,addedMaterial: TMaterial;
    tempLayer: TTextureLayer;
    layerinfo: TStringList;
    fileName: String;
begin
    position := offset+1;

    if Assigned(aProgressForm) then begin
        aProgressForm.Description := 'Loading materials.';
        aProgressForm.Max := nummaterials;
    end;

    for i := 0 to nummaterials-1 do begin
        if Assigned( aProgressForm ) then
            aProgressForm.Position := i+1;

        tempMaterial := TMaterial.Create;
        inc(position,1); // skip index
        tempMaterial.Name := strings[position];
        tempMaterial.ID := i;
        tempMaterial.Category := strings[position+1];
        tempMaterial.IsUsed := StrToBool(strings[position+2]);
        numlayers := StrToInt(strings[position+3]);
        inc(position,4);
        for j := 0 to numlayers-1 do begin
            if strings[position] <> '' then begin
                tempLayer := TTextureLayer.Create;
                with tempLayer do begin
                    layerinfo := SubdivideStringToList(strings[position],',');
                    fileName := dir_texture + layerinfo[0];
                    // Little hack to support colors in prefabs
                    if ExtractFileExt(filename)='' then begin
                        LayerType := mlColor;
                        Color := HexStringToColor(layerinfo[0]);
                        If (Color and $FF000000) = 0 then
                            Color := Color or integer($FF000000);
                        Texture := materials.Textures.AddTexture(Color);
                    end
                    else begin
                        LayerType := mlTexture;
                        if not FileExistsUTF8(fileName) then
                            GenerateDummyTexture(fileName);
                        if j = LIGHTMAP_TEXTURE_LAYER_INDEX then
                            texture := materials.Lightmaps.AddTexture(fileName)
                        else
                            texture := materials.Textures.AddTexture(fileName);
                    end;
                    blendOperation := TBlendOperation(StrToInt(layerinfo[1])+1);
                    tempMaterial.AddLayer(tempLayer);
                    FreeAndNil(layerinfo);
                end
            end else
                tempMaterial.AddEmptyLayer;

            inc(position);
        end;

        addedMaterial := materials.AddMaterial(tempMaterial);

        if Assigned(newMaterials) then
            newMaterials.Add(addedMaterial);        // add is slightly safer than addMaterial here

        if addedMaterial<>tempMaterial then
            FreeAndNil(tempMaterial);
    end;

    TMaterialMatcher.Match(materials);
    //materials.ResolveOriginalMaterials;

    Result := position;
end;

function TDMFSceneReader.SetMaterials091(nummaterials: integer; strings: TStringList; offset: longint; aScene: TScene; aProgressForm: TfrmProgress; aWarnings: TStringList; newMaterials: TMaterialManager = nil): longint;
begin
   Result :=  SetMaterials091Ex(aScene.MaterialManager, nummaterials, strings, offset, aScene, aProgressForm, aWarnings, newMaterials);
end;

function TDMFSceneReader.SetMaterials091Ex(materials: TMaterialManager; nummaterials: integer; strings: TStringList; offset: longint; aScene: TScene; aProgressForm: TfrmProgress; aWarnings: TStringList; newMaterials: TMaterialManager = nil): longint;
var i,j,numlayers:integer;
    position: longint;
    tempMaterial,addedMaterial: TMaterial;
    tempLayer: TTextureLayer;
    layerinfo: TStringList;
    FileName: string;
begin
    position := offset+1;

    if Assigned(aProgressForm) then begin
        aProgressForm.Description := 'Loading materials.';
        aProgressForm.Max := nummaterials;
    end;

    for i := 0 to nummaterials-1 do begin
        if Assigned( aProgressForm ) then
            aProgressForm.Position := i+1;

        tempMaterial := TMaterial.Create;
        inc(position,1); // skip index
        tempMaterial.Name := strings[position];
        tempMaterial.ID := i;
        tempMaterial.Category := strings[position+1];
        tempMaterial.IsUsed := StrToBool(strings[position+2]);
        numlayers := StrToInt(strings[position+3]);
        inc(position,4);
        for j := 0 to numlayers-1 do begin
            if strings[position] <> '' then begin
                layerinfo := SubdivideStringToList(strings[position],',');

                tempLayer := TTextureLayer.Create;
                with tempLayer do begin
                    LayerType := TMaterialLayerType(StrToInt(layerinfo[0]));
                    if LayerType=mlTexture then begin
                        FileName := dir_texture + layerinfo[1];
                        FileName := MakeValidTextureFileName(FileName, aWarnings);
                        if j=LIGHTMAP_TEXTURE_LAYER_INDEX then
                            Texture := materials.Lightmaps.AddTexture(FileName)
                        else
                            Texture := materials.Textures.AddTexture(FileName);

                    end else begin
                        Color := HexStringToColor(layerinfo[1]);

                        // NOTE, PJ: Pre-1.3 DeleD's always stored an alpha component of zero. We correct that here.
                        // When we introduce a new fileformat, this ad-hoc correction won't be needed there.
                        If (Color and $FF000000) = 0 then
                            Color := Color or integer($FF000000);

                        Texture := materials.Textures.AddTexture(Color);
                    end;

                    BlendOperation := TBlendOperation(StrToInt(layerinfo[2]));
                end;
                tempMaterial.AddLayer(tempLayer);
                FreeAndNil(layerinfo);
            end else
                tempMaterial.AddEmptyLayer;

            inc(position);
        end;

        addedMaterial := materials.AddMaterial(tempMaterial);

        if Assigned(newMaterials) then
            newMaterials.Add(addedMaterial);        // add is slightly safer than addMaterial here

        if addedMaterial<>tempMaterial then
            FreeAndNil(tempMaterial);
    end;

    TMaterialMatcher.Match(materials);
    //materials.ResolveOriginalMaterials;

    Result := position;
end;

procedure TDMFSceneReader.Load091(aFileName: string; aScene: TScene; aProgressForm: TfrmProgress; aWarnings: TStringList);
var txtSeperated: string;
    strings,values: TStringlist;
    nummaterials, numobjects, numlights: integer;
    position: longint;
    newMaterials: TMaterialManager;
begin
    try
        strings := TStringList.Create;
        try
            strings.LoadFromFile(aFileName);            // load the DMF file into a stringlist
            txtSeperated := FastReplace(strings.Text,#13#10,'');
        finally
            FreeAndNil(strings);
        end;

        values := TStringList.Create;
        newMaterials := TMaterialManager.Create(false);
        try
            SubdivideStringToList2(txtSeperated,';',values);

            // load scene properties
            aScene.Name := values[2];
            aScene.AmbientColor := HexToRGBColor(values[3]);
            aScene.ShadowOpacity := StrToInt(values[4]);

            // read material information
            nummaterials := StrToInt(values[5]);
            position := SetMaterials091(nummaterials,values,5,aScene,aProgressForm,aWarnings,newMaterials);

            // read object information
            numobjects := StrToInt(values[position]);
            position := SetObjects090(numobjects,values,position,aScene,false,aProgressForm,newMaterials);

            // load lights
            numlights := StrToInt(values[position]);
            SetLights090(numlights,values,position,aScene,false,aProgressForm);

        finally
            FreeAndNil(values);
            FreeAndNil(newMaterials);
        end;

    except
        on E: Exception do raise Exception.Create('TDMFSceneReader.Load091: '+E.Message);
    end;
end;

procedure TDMFSceneReader.Load110(aFileName: string; aScene: TScene; aProgressForm: TfrmProgress; aWarnings: TStringList);
var txtSeperated: string;
    strings,values: TStringlist;
    nummaterials, numobjects, numlights: integer;
    position: longint;
    newMaterials: TMaterialManager;
begin
    try
        strings := TStringList.Create;
        try
            strings.LoadFromFile(aFileName);            // load the DMF file into a stringlist
            txtSeperated := FastReplace(strings.Text,#13#10,'');
        finally
            FreeAndNil(strings);
        end;

        values := TStringList.Create;
        newMaterials := TMaterialManager.Create(false);
        try
            SubdivideStringToList2(txtSeperated,';',values);

            // load scene properties
            aScene.name := values[2];
            aScene.AmbientColor := HexToRGBColor(values[3]);
            aScene.ShadowOpacity := StrToInt(values[4]);

            // read material information
            nummaterials := StrToInt(values[5]);
            position := SetMaterials091(nummaterials,values,5,aScene,aProgressForm,aWarnings,newMaterials);

            // read object information
            numobjects := StrToInt(values[position]);
            position := SetObjects110(numobjects,values,position,aScene,false,aProgressForm,newMaterials);

            // load lights
            numlights := StrToInt(values[position]);
            SetLights090(numlights,values,position,aScene,false,aProgressForm);

        finally
            FreeAndNil(values);
            FreeAndNil(newMaterials);
        end;

    except
        on E: Exception do raise Exception.Create('TDMFSceneReader.Load110: '+E.Message);
    end;
end;

function TDMFSceneReader.SetObjects(format: string; numobjects: integer; strings: TStringList; offset: longint; aScene: TScene; shouldReplace: boolean; aProgressForm: TfrmProgress; materials: TMaterialManager): boolean;
begin
    try
        if format='090' then begin
            SetObjects090(numobjects,strings,offset,aScene,shouldReplace,aProgressForm,materials);
            Result := true;
        end else
        if format='091' then begin
            SetObjects090(numobjects,strings,offset,aScene,shouldReplace,aProgressForm,materials);
            Result := true;
        end else
        if format='110' then begin
            SetObjects110(numobjects,strings,offset,aScene,shouldReplace,aProgressForm,materials);
            Result := true;
        end else begin
            raise Exception.Create('Invalid dataformat specified.');
            Result := false;
        end;
    except
        on E: Exception do raise Exception.Create('TDMFSceneReader.SetObjects: '+E.message);
    end;
end;

function TDMFSceneReader.SetLights(format: string; numlights: integer; strings: TStringList; offset: longint; aScene: TScene; shouldReplace: boolean; aProgressForm: TfrmProgress): boolean;
begin
    try
        if format='090' then begin
            SetLights090(numlights,strings,offset,aScene,shouldReplace,aProgressForm);
            Result := true;
        end else
        if format='091' then begin
            SetLights090(numlights,strings,offset,aScene,shouldReplace,aProgressForm);
            Result := true;
        end else
        if format='110' then begin
            SetLights090(numlights,strings,offset,aScene,shouldReplace,aProgressForm);
            Result := true;
        end else begin
            raise Exception.Create('Invalid dataformat specified.');
            Result := false;
        end;
    except
        on E: Exception do raise Exception.Create('TDMFSceneReader.SetLights: '+E.message);
    end;
end;

function TDMFSceneReader.SetMaterials(format: string; nummaterials: integer; strings: TStringList; offset: longint; aScene: TScene; aProgressForm: TfrmProgress; newMaterials: TMaterialManager): boolean;
begin
    try
        if format='090' then begin
            SetMaterials090(nummaterials,strings,offset,aScene,aProgressForm,newMaterials);
            Result := true;
        end else
        if format='091' then begin
            SetMaterials091(nummaterials,strings,offset,aScene,aProgressForm,nil,newMaterials);
            Result := true;
        end else
        if format='110' then begin
            SetMaterials091(nummaterials,strings,offset,aScene,aProgressForm,nil,newMaterials);
            Result := true;
        end else begin
            raise Exception.Create('Invalid dataformat specified.');
            Result := false;
        end;
    except
        on E: Exception do raise Exception.Create('TDMFSceneReader.SetMaterials: '+E.message);
    end;
end;

procedure TDMFSceneReader.ReadFromFile(const aFileName: string; const aScene: TScene);
var F1: TextFile;
    idstr,value: string;
    progressForm: TfrmProgress;
    warnings: TStringList;
begin
    try
        warnings := TStringList.Create;
        warnings.Sorted := true;
        warnings.Duplicates := dupIgnore;

        try
            {$IFNDEF CONSOLE_TEST}
            progressForm := TfrmProgress.Create(nil);
            progressForm.Title := 'Loading...';
            progressForm.SubTitle := 'Loading ' + ExtractFileName(aFileName);
            progressForm.Description := 'Loading ' + ExtractFileName(aFileName);
            progressForm.Show;
            {$ELSE}
            progressForm := nil;
            {$ENDIF}

            try
                // set filename and open file
                aScene.FileName := aFileName;

                AssignFile(F1, aFileName);
                try
                    Reset(F1);
                    Readln(F1,idstr);                                       // DeleD #.# Map File string
                    Readln(F1,value);                                       // Version info
                finally
                    CloseFile(F1);
                end;

                if LeftStr(idstr,14) <> 'DeleD Map File' then
                    raise Exception.Create('no DeleD Map file')
                else begin
                    value := Trim(FastReplace(value,';',''));
                    if value='Version 1.0' then
                        Load091(aFileName,aScene,progressForm,warnings)
                    else if value='Version 1.1' then
                        Load110(aFileName,aScene,progressForm,warnings)
                    else
                        raise Exception.Create('DMF format not recognized');
                end;

                aScene.UpdateHandlers;
            finally
                if Assigned(progressForm) then progressForm.Release;
            end;

            if warnings.Count > 0 then
                MessageDlg(PChar(warnings.Text), mtWarning, [mbOK], 0);

        finally
            FreeAndNil(warnings);
        end;

    except
        on E: Exception do begin
            aScene.FileName := 'no-scene-yet.dmf';
            raise Exception.Create('TDMFSceneReader.Load: '+E.Message);
        end;
    end;
end;


end.
