{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_DMFSceneWriter;

{$MODE Delphi}

interface

uses Classes,
  unit_Enums,
  unit_MaterialList,
  unit_Scene;

type
  TDMFSceneWriter = class(TObject)
  private
    fErrorMessages: TStringList;
    procedure ShowErrorMessages;
    {$HINTS OFF}
    procedure AddUniqueErrorMessage(const aMessage: string);
    {$HINTS ON}
    property ErrorMessages: TStringList read fErrorMessages;
    function FloatToString(const aFloat: double; aPrecision: integer = 7; aNumDigits: integer = 7): string;
    function RGBColorToHex(const aColor:TRGBColor): string;
    // format based Get functions
    procedure GetObjects090(aScene: TScene; selected: boolean; var strings: TStringList);
    procedure GetLights090(aScene: TScene; selected: boolean; var strings: TStringList);
    procedure GetMaterials090Ex(materials: TMaterialList; selected: boolean; var strings: TStringList; options: TGetMaterialsOptions = []);
    procedure GetMaterials090(aScene: TScene; selected: boolean; var strings: TStringList; options: TGetMaterialsOptions = []);
    procedure GetMaterials091(aScene: TScene; selected: boolean; var strings: TStringList; options: TGetMaterialsOptions = []);
    procedure GetMaterials091Ex(materials: TMaterialList; selected: boolean; var strings: TStringList; options: TGetMaterialsOptions = []);
  public
    constructor Create;
    destructor Destroy; override;
    // generic Get functions; they return specific (selected) items from a scene
    // or provided list in a specific format which is returned in a string or stringlist;
    // used in plugin architecture
    function GetObjects(aScene: TScene; format: string; selected: boolean): string;
    function GetLights(aScene: TScene; format: string; selected: boolean): string;
    function GetMaterials(aScene: TScene; format: string; selected: boolean): string; overload;
    procedure GetMaterials(aScene: TScene; format: string; selected: boolean; strings: TStringList); overload;
    procedure GetMaterials(aMaterialList: TMaterialList; format: string; selected: boolean; strings: TStringList; options: TGetMaterialsOptions = []); overload;
    // generic Save method
    procedure WriteToFile(const aScene: TScene; const aFileName: string; const aSelectedOnly: boolean);
  end;

implementation

uses
  Dialogs, SysUtils, Math,
  unit_Types,
  unit_Functions,
  unit_Constants,
  unit_Primitives,
  unit_PrimitiveList,
  unit_Lights,
  unit_Materials;

constructor TDMFSceneWriter.Create;
begin
    fErrorMessages := TStringList.Create;
end;

destructor TDMFSceneWriter.Destroy;
begin
    FreeAndNil(fErrorMessages);
    inherited Destroy;
end;

procedure TDMFSceneWriter.ShowErrorMessages;
var index: integer;
    errors: string;
begin
    if ErrorMessages.Count > 0 then begin
        errors := '';
        for index := 0 to ErrorMessages.Count - 1 do
            errors := errors + '- ' + ErrorMessages[index] + CRLF;

        MessageDlg( 'The following errors occured while performing this action:' + CRLF + CRLF + errors + CRLF +
            'Please inform the Delgine Team about this error, providing as much details as possible. Your help in solving this problem is very much appreciated!',mtWarning,[mbOK],0);
    end;
end;

procedure TDMFSceneWriter.AddUniqueErrorMessage(const aMessage: string);
begin
    if ErrorMessages.IndexOf(aMessage) = -1 then
        ErrorMessages.Add(aMessage);
end;

function TDMFSceneWriter.FloatToString(const aFloat: double; aPrecision: integer; aNumDigits: integer): string;
var value: double;
begin
    value := aFloat;

    if IsNaN(value) then begin
        AddUniqueErrorMessage( 'An invalid floating point value (NaN) was encountered and replaced by zero.' );
        value := 0;
    end
    else if IsInfinite(value) then
    begin
        AddUniqueErrorMessage( 'An invalid floating point value (Infinity) was encountered and replaced by zero.' );
        value := 0;
    end;

    Result := FloatToStrF(value, ffFixed, aPrecision, aNumDigits);
end;

function TDMFSceneWriter.RGBColorToHex(const aColor:TRGBColor): string;
begin
    Result := '00' + IntToHex(Round(aColor.B * 255), 2) + IntToHex(Round(aColor.G * 255), 2) + IntToHex(Round(aColor.R * 255), 2);
end;

procedure TDMFSceneWriter.GetObjects090(aScene: TScene; selected: boolean; var strings: TStringList);
var i,j,k,l: integer;
    list: TPrimitiveList;
    obj: TBasePrimitive;
    poly: TPolygon;
    vertex: TVertice;
    uvcoord: TUVCoord;
    str: string;
begin
    list := aScene.Primitives.GetItems(TPrimitive);
    for i := 0 to list.Count-1 do begin
        obj := list[i];

        if ((selected) and (obj.selected)) or (not selected) then begin
            strings.Add(obj.name+';'+
                        IntToStr(obj.id)+';'+
                        IntToStr(integer(obj.objecttype))+';'+
                        BoolToStr(obj.visible)+';'+
                        IntToStr(integer(obj.SnapMethod))+';'+
                        IntToStr(obj.groupid)+';'+
                        BoolToStr(obj.AutoUV)+';'+
                        obj.tag + ';');

            // write vertices
            strings.Add(IntToStr(obj.numvertices)+';');
            for j := 0 to obj.numvertices-1 do begin
                vertex := obj.Vertices[j];
                strings.Add(FloatToString(vertex.x)+';'+
                            FloatToString(vertex.y)+';'+
                            FloatToString(vertex.z)+';');
            end;

            // write polygons
            strings.Add(IntToStr(obj.numpolygons)+';');
            for j := 0 to obj.numpolygons-1 do begin
                poly := obj.Polygons[j];

                str := IntToStr(poly.numvertices)+';'+IntToStr(aScene.MaterialManager.IndexOfMaterial(poly.material))+';';

                for k := 0 to poly.numvertices-1 do
                    str := str+IntToStr(obj.Vertices.IndexOf(poly.VerticeList[k]))+';';

                for l := 0 to poly.Material.NumTextureLayers - 1 do begin

                    if Assigned(poly.Material.TextureLayers[l]) then begin
                        for k := 0 to poly.numvertices-1 do begin
                            uvcoord := poly.UVCoordinates[l][k];
                            str := str+FloatToString(uvcoord.u)+';'+FloatToString(uvcoord.v)+';'
                        end;
                    end;
                end;
                strings.Add(str);
            end;
        end;
    end;
    FreeAndNil(list);
end;

procedure TDMFSceneWriter.GetLights090(aScene: TScene; selected: boolean; var strings: TStringList);
var i: integer;
    list: TPrimitiveList;
    str: string;
    light: TLight;
    pointlight: TPointLight;
    dirlight: TDirectionalLight;
    spotlight: TSpotLight;
begin
    list := aScene.Primitives.GetItems(TLight);
    for i := 0 to list.Count-1 do begin
        light := TLight(list[i]);

        if ((selected) and (light.selected)) or (not selected) then begin

            // save properties based on type
            if light.objecttype=otPointLight then begin
                pointlight := TPointLight(light);
                str := '1;'+FloatToStr(pointlight.constant)+';'+
                            FloatToStr(pointlight.linear)+';'+
                            FloatToStr(pointlight.quadratic)+';'+
                            IntToStr(pointlight.range)+';'
            end;
            if light.objecttype=otDirectionalLight then begin
                dirlight := TDirectionalLight(light);
                str := '2;'+FloatToStr(dirlight.direction.x)+';'+FloatToStr(dirlight.direction.y)+';'+FloatToStr(dirlight.direction.z)+';';
            end;
            if light.objecttype=otSpotLight then begin
                spotlight := TSpotLight(light);
                str := '3;'+FloatToStr(spotlight.constant)+';'+
                            FloatToStr(spotlight.linear)+';'+
                            FloatToStr(spotlight.quadratic)+';'+
                            IntToStr(spotlight.range)+';'+
                            FloatToStr(spotlight.direction.x)+';'+FloatToStr(spotlight.direction.y)+';'+FloatToStr(spotlight.direction.z)+';'+
                            IntToStr(spotlight.innerangle)+';'+
                            IntToStr(spotlight.outerangle)+';'+
                            IntToStr(spotlight.falloff)+';';
            end;
            // save default light properties (PJ: Note how position != center (especially after rotation)
            str := str + FloatToString(light.position.x)+';'+FloatToString(light.position.y)+';'+FloatToString(light.position.z)+';'+
                         IntToStr(light.flash)+';'+
                         BoolToStr(light.active)+';'+
                         RGBColorToString(light.diffuse,';')+';'+
                         RGBColorToString(light.specular,';')+';'+
                         IntToStr(light.intensity)+';'+
                         IntToStr(light.id)+';'+
                         light.name + ';';
            strings.Add(str);
        end;
    end;
    FreeAndNil(list);
end;

procedure TDMFSceneWriter.GetMaterials090(aScene: TScene; selected: boolean; var strings: TStringList; options: TGetMaterialsOptions);
begin
    GetMaterials090Ex(aScene.MaterialManager,selected,strings,options);
end;

procedure TDMFSceneWriter.GetMaterials090Ex(materials: TMaterialList; selected: boolean; var strings: TStringList; options: TGetMaterialsOptions);
var i,j: integer;
    material: TMaterial;
    layer: TTextureLayer;
    str: string;
begin
    // save materiallist to the stringlist
    for i := 0 to materials.nummaterials-1 do begin
        material := materials.Materials[i];

        if material.IsOriginal or not (gmoOriginalsOnly in options) then begin

            str := IntToStr(i)+';'+material.Name+';'+material.Category+';'+ BoolToStr(material.IsUsed)+';' +IntToStr(material.NumTextureLayers)+';';
            for j := 0 to material.NumTextureLayers-1 do begin
                layer := material.TextureLayers[j];

                if Assigned(layer) then
                    str := str + layer.Texture.ShortFilename+','+IntToStr(integer(layer.BlendOperation)-1)+';'      // blendop-1 because 090 format has one less blendop option (no Ignore)
                else
                    str := str + ';';

            end;
            strings.Add(str);
        end;
    end;
end;

procedure TDMFSceneWriter.GetMaterials091(aScene: TScene; selected: boolean; var strings: TStringList; options: TGetMaterialsOptions);
begin
    GetMaterials091Ex(aScene.MaterialManager,selected,strings,options);
end;

procedure TDMFSceneWriter.GetMaterials091Ex(materials: TMaterialList; selected: boolean; var strings: TStringList; options: TGetMaterialsOptions);
var i,j: integer;
    material: TMaterial;
    layer: TTextureLayer;
    str: string;
begin
    // save materiallist to the stringlist
    for i := 0 to materials.nummaterials-1 do begin
        material := materials.Materials[i];

        if material.IsOriginal or not (gmoOriginalsOnly in options) then begin

            str := IntToStr(i)+';'+material.Name+';'+material.Category+';'+ BoolToStr(material.IsUsed)+';' +IntToStr(material.NumTextureLayers)+';';
            for j := 0 to material.NumTextureLayers-1 do begin
                layer := material.TextureLayers[j];

                if Assigned(layer) then begin

                    if layer.LayerType=mlColor then
                        str := str + IntToStr(integer(layer.LayerType))+','+ColorToHexString(layer.Color)+','+IntToStr(integer(layer.BlendOperation))+';'
                    else
                        str := str + IntToStr(integer(layer.LayerType))+','+layer.Texture.ShortFilename+','+IntToStr(integer(layer.BlendOperation))+';';

                end else
                    str := str + ';';

            end;
            strings.Add(str);
        end;
    end;
end;

// generic function that returns a string that consists of object data
function TDMFSceneWriter.GetObjects(aScene: TScene; format: string; selected: boolean): string;
var strings: TStringList;
begin
    strings := TStringList.Create;
    try
        try
            if format='090' then begin
                GetObjects090(aScene,selected,strings);
                Result := FastReplace(strings.Text,#13#10,'');
            end else
            if format='091' then begin
                GetObjects090(aScene,selected,strings);
                Result := FastReplace(strings.Text,#13#10,'');
            end else begin
                raise Exception.Create('Invalid dataformat specified.');
                Result := 'Invalid dataformat specified.';
            end;
        except
            on E: Exception do raise Exception.Create('TDMFSceneWriter.GetObjects: '+E.message);
        end;
    finally
        FreeAndNil(strings);
    end;
end;

// generic function that returns a string that consists of light data
function TDMFSceneWriter.GetLights(aScene: TScene; format: string; selected: boolean): string;
var strings: TStringList;
begin
    strings := TStringList.Create;
    try
        try
            if format='090' then begin
                GetLights090(aScene,selected,strings);
                Result := FastReplace(strings.Text,#13#10,'');
            end else
            if format='091' then begin
                GetLights090(aScene,selected,strings);
                Result := FastReplace(strings.Text,#13#10,'');
            end else begin
                raise Exception.Create('Invalid dataformat specified.');
                Result := 'Invalid dataformat specified.';
            end;
        except
            on E: Exception do raise Exception.Create('TDMFSceneWriter.GetLights: '+E.message);
        end;
    finally
        FreeAndNil(strings);
    end;
end;

// generic function that returns a string that consists of material data
function TDMFSceneWriter.GetMaterials(aScene: TScene; format: string; selected: boolean): string;
var strings: TStringList;
begin
    strings := TStringList.Create;
    try
        try
            if format='090' then begin
                GetMaterials090(aScene,selected,strings);
                Result := FastReplace(strings.Text,#13#10,'');
            end else
            if format='091' then begin
                GetMaterials091(aScene,selected,strings);
                Result := FastReplace(strings.Text,#13#10,'');
            end else begin
                raise Exception.Create('Invalid dataformat specified.');
                Result := 'Invalid dataformat specified.';
            end;
        except
            on E: Exception do raise Exception.Create('TDMFSceneWriter.GetMaterials: '+E.message);
        end;
    finally
        FreeAndNil(strings);
    end;
end;

procedure TDMFSceneWriter.GetMaterials(aScene: TScene; format: string; selected: boolean; strings: TStringList);
begin
    try
        if format='090' then begin
            GetMaterials090(aScene,selected,strings);
        end else
        if format='091' then begin
            GetMaterials091(aScene,selected,strings);
        end else begin
            raise Exception.Create('Invalid dataformat specified.');
        end;
    except
        on E: Exception do raise Exception.Create('TDMFSceneWriter.GetMaterials: '+E.message);
    end;
end;

procedure TDMFSceneWriter.GetMaterials(aMaterialList: TMaterialList; format: string; selected: boolean; strings: TStringList; options: TGetMaterialsOptions = []);
begin
    try
        if format='090' then begin
            GetMaterials090Ex(aMaterialList,selected,strings);
        end else
        if format='091' then begin
            GetMaterials091Ex(aMaterialList,selected,strings);
        end else begin
            raise Exception.Create('Invalid dataformat specified.');
        end;
    except
        on E: Exception do raise Exception.Create('TDMFSceneWriter.GetMaterials: '+E.message);
    end;
end;

procedure TDMFSceneWriter.WriteToFile(const aScene: TScene; const aFileName: string; const aSelectedOnly: boolean);
var strings: TStringList;
    i,numSelectedLights: integer;
begin
    ErrorMessages.Clear;

    strings := TStringList.Create;
    try
        try
            if not aSelectedOnly then begin                                         // only write scene info if wanted
                strings.Add('DeleD Map File;');                                     // DeleD Map File string
                strings.Add('Version 1.1;');                                        // Version info
                strings.Add(aScene.Name+';'+
                            RGBColorToHex(aScene.AmbientColor)+';'+
                            IntToStr(aScene.ShadowOpacity)+';');
            end else    // only save version info
                strings.Add('Version 1.1;');                                        // Version info

            // save materials to stringlist in 091 format
            if not aSelectedOnly then
                strings.Add(IntToStr(aScene.MaterialManager.nummaterials)+';')
            else
                raise Exception.Create('Saving of selected materials currently not supported.');
            GetMaterials091(aScene,aSelectedOnly,strings);

            // save objects to stringlist in 090 format
            if not aSelectedOnly then
                strings.Add(IntToStr(aScene.Primitives.Count)+';')
            else
                strings.Add(IntToStr(aScene.SelectedPrimitives.Count)+';');
            GetObjects090(aScene,aSelectedOnly,strings);

            // save lights to stringlist in 090 format
            if not aSelectedOnly then
                strings.Add(IntToStr(aScene.Primitives.NumLights)+';')
            else begin
                numSelectedLights := 0;
                for i := 0 to aScene.SelectedPrimitives.Count-1 do begin
                    if aScene.SelectedPrimitives[i] is TLight then
                        inc(numSelectedLights);
                end;
                strings.Add(IntToStr(numSelectedLights)+';');
            end;
            GetLights090(aScene,aSelectedOnly,strings);

            strings.SaveToFile(aFileName);

        finally
            FreeAndNil(strings);
        end;

    except
        on E: Exception do raise Exception.Create('TDMFSceneWriter.Save: '+E.Message);
    end;
    ShowErrorMessages;          // only shows error messages if present
end;




end.
