{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_EdgeBevellerTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TVertexInfo = class(TObject)
  private
    fEdge: TEdge;
    fEdgeVertex: TVertice;
    fVertex: TVertice;
    fDirection: TVertice;
  public
    constructor Create;
    destructor Destroy; override;
    property Edge: TEdge read fEdge write fEdge;
    property EdgeVertex: TVertice read fEdgeVertex write fEdgeVertex;
    property Vertex: TVertice read fVertex write fVertex;
    property Direction: TVertice read fDirection write fDirection;
  end;

  TEdgeBeveller = class(TPrimitiveTool)
  private
    fEPS: double;
    fDefaultMaterial: TMaterial;
    fVertexInfoList: TObjectList;

    procedure DoBevel(APrimitive: TBasePrimitive; AVertexInfoList: TObjectList);
    procedure GenerateVertexInfo(APrimitive: TBasePrimitive; AVertexInfoList: TObjectList);
    procedure GetAdjacentPolygons(ASourcePrimitive: TBasePrimitive; APolygonList: TPolygonList);
    procedure RemoveOriginalGeometry(APolygonList: TPolygonList; ASourcePrimitive: TBasePrimitive);

    procedure GeneratePolygon(APolygon: TPolygon; AVertexInfoList: TObjectList; ADestPrimitive: TBasePrimitive); overload;
    procedure GeneratePolygon(AEdge: TEdge; AVertexInfoList: TObjectList; ADestPrimitive: TBasePrimitive; ASelectionList: TEdgeList); overload;
    procedure GeneratePolygon(AEdge: TEdge; AVertex: TVertice; AVertexInfoList: TObjectList; ADestPrimitive: TBasePrimitive); overload;

    function GetEdgeFromPolygonByVertex(polygon: TPolygon; edge: TEdge; vertex: TVertice): TEdge;
    function GetVertexInfo(edge: TEdge; vertex: TVertice; vertexInfoList: TObjectList): TVertexInfo; overload;
    function GetVertexInfo(vertex: TVertice; vertexInfoList: TObjectList): TVertexInfo; overload;
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
    function CanExecute(AParamList: TParamList): boolean; override;
    constructor Create(AToolAction: TToolAction);
    destructor Destroy; override;
  end;

implementation

uses unit_PrimitiveList;

constructor TEdgeBeveller.Create(AToolAction: TToolAction);
begin
    inherited Create(AToolAction);
    fVertexInfoList := TObjectList.Create(true);
end;

destructor TEdgeBeveller.Destroy;
begin
    FreeAndNil(fVertexInfoList);
    inherited Destroy;
end;

// we create new geometry in the Prepare method
procedure TEdgeBeveller.DoPrepare(AParamList: TParamList);
var i: integer;
begin
    Hint := 'Bevel edges';
    Description := 'Bevels edges';

    fEPS := TDoubleObject(AParamList[0]).Value;
    fDefaultMaterial := TMaterial(AParamList[1]);

    for i := 0 to SelectedPrimitives.Count-1 do begin
        if SelectedPrimitives[i].ModifyGeometry then
            DoBevel(SelectedPrimitives[i],fVertexInfoList);
    end;
end;

// the new geometry is scaled along edges in the Execute method
procedure TEdgeBeveller.DoExecute(AParamList: TParamList);
var i: integer;
    numberOfUnits: integer;
    dirVector: TVertice;
    vertexInfo: TVertexInfo;
begin
    numberOfUnits := TIntegerObject(AParamList.NamedParameters['numberOfUnits']).Value;

    for i := 0 to fVertexInfoList.Count-1 do begin
        vertexInfo := fVertexInfoList[i] as TVertexInfo;
        dirVector := TVertice.Create(0,0,0);
        dirVector.Assign(vertexInfo.Direction);
        // multiply with # of units to move to get the final direction vector
        dirVector.MultiplyByScalar(-numberOfUnits);
        vertexInfo.Vertex.Add(dirVector);
        FreeAndNil(dirVector);
    end;
end;

function TEdgeBeveller.CanExecute(AParamList: TParamList): boolean;
var i,j,k: integer;
    primitive: TBasePrimitive;
    edge: TEdge;
    edgeList: TEdgeList;
    canBevel: boolean;
begin
    canBevel := true;
    edgeList := TEdgeList.Create(false,true);
    for i := 0 to SelectedPrimitives.Count-1 do begin
        primitive := SelectedPrimitives[i];
        // check if any of the selected edges are adjacent; if so, beveling is not allowed.
        for j := 0 to primitive.SelectedEdges.Count-1 do begin
            edge := primitive.SelectedEdges[j];
            edge.GetAdjacent(edgeList);
            edgeList.Remove(edge);

            for k := 0 to edgeList.Count-1 do begin
                if primitive.SelectedEdges.IsInList(edgeList[k]) then begin
                    canBevel := false;
                    Break;
                end;
            end;
        end;
    end;
    FreeAndNil(edgeList);
    result := canBevel;
end;

procedure TEdgeBeveller.GenerateVertexInfo(APrimitive: TBasePrimitive; AVertexInfoList: TObjectList);
var i,j: integer;
    vertex,otherVertex,dirVector: TVertice;
    vertexList: TVerticeList;
    edge: TEdge;
    vertexInfo: TVertexInfo;
begin
    // for each vertex of each selected edge, create a VertexInfo object
    // containing a new vertex for that vertex
    vertexList := TVerticeList.Create(false,true);

    // get a list of unique vertices used by the selected edges
    for i := 0 to APrimitive.SelectedEdges.Count-1 do begin
        edge := APrimitive.SelectedEdges[i];
        vertexList.Add(edge.StartVertex);
        vertexList.Add(edge.EndVertex);
    end;

    // get the adjacent edges for each vertex, excluding the selected edges
    for i := 0 to vertexList.Count-1 do begin
        vertex := vertexList[i];
        for j := 0 to vertex.EdgeList.Count-1 do begin
            edge := vertex.EdgeList[j];
            if not APrimitive.SelectedEdges.IsInList(edge) then begin
                vertexInfo := TVertexInfo.Create;

                // store the other vertex of this edge for direction purposes later
                otherVertex := edge.GetOtherVertex(vertex);

                // store a direction vector for scaling purposes later
                dirVector := TVertice.Create(0,0,0);
                dirVector.x := vertex.x-otherVertex.x;
                dirVector.y := vertex.y-otherVertex.y;
                dirVector.z := vertex.z-otherVertex.z;
                dirVector.Normalize;
                dirVector.MultiplyByScalar(-1);

                vertexInfo.Edge := edge;
                vertexInfo.EdgeVertex := vertex;
                vertexInfo.Vertex := vertex.Copy;
                // scale just a little to make sure polygons created with these vertices are not degenerate
                vertexInfo.Vertex.Add(dirVector);
                vertexInfo.Direction.Assign(dirVector);
                AVertexInfoList.Add(vertexInfo);

                FreeAndNil(dirVector);
            end;
        end;
    end;

    FreeAndNil(vertexList);
end;

procedure TEdgeBeveller.GetAdjacentPolygons(ASourcePrimitive: TBasePrimitive; APolygonList: TPolygonList);
var i: integer;
    polygonList: TPolygonList;
    selectedVertex: TVertice;
    vertexList: TVerticeList;
    edge: TEdge;
begin
    vertexList := TVerticeList.Create(false,true);

    // get a list of unique vertices used by the selected edges
    for i := 0 to ASourcePrimitive.SelectedEdges.Count-1 do begin
        edge := ASourcePrimitive.SelectedEdges[i];
        vertexList.Add(edge.StartVertex);
        vertexList.Add(edge.EndVertex);
    end;

    polygonList := TPolygonList.Create(false, true);
    for i := 0 to vertexList.Count-1 do begin
        selectedVertex := vertexList[i];
        selectedVertex.getAdjacent(PolygonList);
        APolygonList.AddList(polygonList);
    end;

    FreeAndNil(polygonList);
    FreeAndNil(vertexList);
end;

procedure TEdgeBeveller.RemoveOriginalGeometry(APolygonList: TPolygonList; ASourcePrimitive: TBasePrimitive);
var i: integer;
begin
    // remove polygons from the primitive
    for i := 0 to APolygonList.Count-1 do
        ASourcePrimitive.Polygon_Remove(APolygonList[i]);

    // clear selection
    ASourcePrimitive.SelectedPolygons.Clear;
end;

procedure TEdgeBeveller.GeneratePolygon(APolygon: TPolygon; AVertexInfoList: TObjectList; ADestPrimitive: TBasePrimitive);
var i,index: integer;
    polygon: TPolygon;
    vertex,otherVertex: TVertice;
    edge: TEdge;
    vertexInfo: TVertexInfo;
    isAdded: boolean;
begin
    polygon := TPolygon.Create;
    polygon.Material := fDefaultMaterial;

    for i := 0 to APolygon.NumVertices-1 do begin
        vertex := APolygon.VerticeList[i];

        isAdded := false;

        index := (i-1);
        if index<0 then index := APolygon.NumVertices-1;
        otherVertex := APolygon.VerticeList[index];
        edge := ADestPrimitive.Edge_Get(vertex,otherVertex);
        vertexInfo := GetVertexInfo(edge,vertex,AVertexInfoList);
        if Assigned(vertexInfo) then begin
            ADestPrimitive.Polygon_AddVertex(polygon,vertexInfo.Vertex);
            isAdded := true;
        end;

        index := (i+1) mod APolygon.NumVertices;
        otherVertex := APolygon.VerticeList[index];
        edge := ADestPrimitive.Edge_Get(vertex,otherVertex);
        vertexInfo := GetVertexInfo(edge,vertex,AVertexInfoList);
        if Assigned(vertexInfo) then begin
            ADestPrimitive.Polygon_AddVertex(polygon,vertexInfo.Vertex);
            isAdded := true;
        end;

        if not isAdded then
            ADestPrimitive.Polygon_AddVertex(polygon,vertex);
    end;
    ADestPrimitive.Polygon_Add(polygon);
end;

procedure TEdgeBeveller.GeneratePolygon(AEdge: TEdge; AVertexInfoList: TObjectList; ADestPrimitive: TBasePrimitive; ASelectionList: TEdgeList);
var polygonList: TPolygonList;
    polygon: TPolygon;
    vertexInfo: TVertexInfo;
    adjacentEdge: TEdge;
    vertexNormal: TVertice;
begin
    polygonList := TPolygonList.Create(false,true);

    polygon := TPolygon.Create;
    polygon.Material := fDefaultMaterial;

    AEdge.GetAdjacent(polygonList);

    adjacentEdge := GetEdgeFromPolygonByVertex(polygonList[0],AEdge,AEdge.StartVertex);
    vertexInfo := GetVertexInfo(adjacentEdge,AEdge.StartVertex,AVertexInfoList);
    ADestPrimitive.Polygon_AddVertex(polygon,vertexInfo.Vertex);

    adjacentEdge := GetEdgeFromPolygonByVertex(polygonList[1],AEdge,AEdge.StartVertex);
    vertexInfo := GetVertexInfo(adjacentEdge,AEdge.StartVertex,AVertexInfoList);
    ADestPrimitive.Polygon_AddVertex(polygon,vertexInfo.Vertex);

    // NOTE: the adjacent edge might be a selected one... account for that.

    adjacentEdge := GetEdgeFromPolygonByVertex(polygonList[1],AEdge,AEdge.EndVertex);
    vertexInfo := GetVertexInfo(adjacentEdge,AEdge.EndVertex,AVertexInfoList);
    ADestPrimitive.Polygon_AddVertex(polygon,vertexInfo.Vertex);

    adjacentEdge := GetEdgeFromPolygonByVertex(polygonList[0],AEdge,AEdge.EndVertex);
    vertexInfo := GetVertexInfo(adjacentEdge,AEdge.EndVertex,AVertexInfoList);
    ADestPrimitive.Polygon_AddVertex(polygon,vertexInfo.Vertex);

    ADestPrimitive.Polygon_Add(polygon);

    // flip polygon if needed
    vertexNormal := AEdge.StartVertex.GetNormal();
    if DotProduct(vertexNormal,polygon.Normal) <- fEPS then
        polygon.Flip;
    FreeAndNil(vertexNormal);

    if Assigned(polygon.EdgeList[1]) then
        ASelectionList.Add(polygon.EdgeList[1]);
    if Assigned(polygon.EdgeList[3]) then
        ASelectionList.Add(polygon.EdgeList[3]);

    FreeAndNil(polygonList);
end;

procedure TEdgeBeveller.GeneratePolygon(AEdge: TEdge; AVertex: TVertice; AVertexInfoList: TObjectList; ADestPrimitive: TBasePrimitive);
var i: integer;
    edgeList: TEdgeList;
    polygon: TPolygon;
    vertexInfo: TVertexInfo;
    vertexNormal: TVertice;
begin
    if AVertex.EdgeList.Count>3 then begin
        edgeList := TEdgeList.Create(false,true);

        AVertex.GetOrderedEdges(edgeList, false);
        edgeList.Remove(AEdge);

        polygon := TPolygon.Create;
        polygon.Material := fDefaultMaterial;

        for i := 0 to edgeList.Count-1 do begin
            vertexInfo := GetVertexInfo(edgeList[i],AVertex,AVertexInfoList);
            ADestPrimitive.Polygon_AddVertex(polygon,vertexInfo.Vertex);
        end;

        ADestPrimitive.Polygon_Add(polygon);

        // flip polygon if needed
        vertexNormal := AVertex.GetNormal();
        if DotProduct(vertexNormal,polygon.Normal) <- fEPS then
            polygon.Flip;
        FreeAndNil(vertexNormal);

        FreeAndNil(edgeList);
    end;
end;

procedure TEdgeBeveller.DoBevel(APrimitive: TBasePrimitive; AVertexInfoList: TObjectList);
var i: integer;
    polygonList: TPolygonList;
    newSelection: TEdgeList;
begin
    polygonList := TPolygonList.Create(false,true);
    newSelection := TEdgeList.Create(false,true);

    // determine which polygons are affected; this is done by traversing the
    // selected vertices (!) and getting each adjacent polygon of a selected vertex
    GetAdjacentPolygons(APrimitive, polygonList);

    GenerateVertexInfo(APrimitive, AVertexInfoList);

    for i := 0 to APrimitive.SelectedEdges.Count-1 do begin
        GeneratePolygon(APrimitive.SelectedEdges[i],APrimitive.SelectedEdges[i].StartVertex,AVertexInfoList,APrimitive);
        GeneratePolygon(APrimitive.SelectedEdges[i],APrimitive.SelectedEdges[i].EndVertex,AVertexInfoList,APrimitive);
    end;

    for i := 0 to APrimitive.SelectedEdges.Count-1 do
        GeneratePolygon(APrimitive.SelectedEdges[i],AVertexInfoList,APrimitive,newSelection);

    for i := 0 to polygonList.Count-1 do
        GeneratePolygon(polygonList[i],AVertexInfoList,APrimitive);

    // remove original geometry
    RemoveOriginalGeometry(polygonList, APrimitive);

    // assign the new selected edges
    APrimitive.SelectedEdges.Assign(newSelection);

    FreeAndNil(newSelection);
    FreeAndNil(polygonList);
end;

function TEdgeBeveller.GetVertexInfo(edge: TEdge; vertex: TVertice; vertexInfoList: TObjectList): TVertexInfo;
var i: integer;
    edgeVertex: TVertexInfo;
    isFound: boolean;
begin
    edgeVertex := nil;
    isFound := false;
    for i := 0 to vertexInfoList.Count-1 do begin
        edgeVertex := vertexInfoList[i] as TVertexInfo;
        if (edgeVertex.Edge = edge) and (edgeVertex.EdgeVertex = vertex) then begin
            isFound := true;
            break;
        end;
    end;
    if isFound then
        result := edgeVertex
    else
        result := nil;
end;

function TEdgeBeveller.GetVertexInfo(vertex: TVertice; vertexInfoList: TObjectList): TVertexInfo;
var i: integer;
    vertexInfo: TVertexInfo;
    isFound: boolean;
begin
    vertexInfo := nil;
    isFound := false;
    for i := 0 to vertexInfoList.Count-1 do begin
        vertexInfo := vertexInfoList[i] as TVertexInfo;
        if vertexInfo.EdgeVertex = vertex then begin
            isFound := true;
            break;
        end;
    end;
    if isFound then
        result := vertexInfo
    else
        result := nil;
end;

function TEdgeBeveller.GetEdgeFromPolygonByVertex(polygon: TPolygon; edge: TEdge; vertex: TVertice): TEdge;
var i: integer;
    currentEdge: TEdge;
    isFound: boolean;
begin
    currentEdge := nil;
    isFound := false;
    for i := 0 to polygon.EdgeList.Count-1 do begin
        currentEdge := polygon.EdgeList[i];
        if currentEdge <> edge then begin
            if (currentEdge.StartVertex = vertex) or (currentEdge.EndVertex = vertex) then begin
                isFound := true;
                break;
            end;
        end;
    end;
    if isFound then
        result := currentEdge
    else
        result := nil;
end;

{ TVertexInfo }

constructor TVertexInfo.Create;
begin
    fDirection := TVertice.Create(0,0,0);
end;

destructor TVertexInfo.Destroy;
begin
    FreeAndNil(fDirection);
    inherited Destroy;
end;

end.

