{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_EdgeConnectorTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TEdgeConnector = class(TPrimitiveTool)
  private
    procedure GetRelevantGeometry(APrimitive: TBasePrimitive; ARelevantPolygonsList: TPolygonList; ASplittableEdgesList: TEdgeList);
    procedure DoConnect(AParamList: TParamList);
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses unit_EdgeSplitterTool, unit_VertexConnectorTool;

// determine relevant geometry: a polygon is relevant if it contains 2 or more
// selected edges that do NOT form a straight line
procedure TEdgeConnector.GetRelevantGeometry(APrimitive: TBasePrimitive; ARelevantPolygonsList: TPolygonList; ASplittableEdgesList: TEdgeList);
var i,j: integer;
    relevantEdgesPerPolygon,selectedEdgesOfPolygon: TEdgeList;
    edge,nextEdge: TEdge;
    polygonList: TPolygonList;
    polygon: TPolygon;
    startVertex,middleVertex,endVertex: TVertice;
begin
    ARelevantPolygonsList.Clear;
    ASplittableEdgesList.Clear;

    polygonList := TPolygonList.Create(false,true);

    // based on the selected edges of the provided primitive,
    // get a list of polygons that use those selected edges;
    // at this point, such a polygon might contain only 1 edge
    // and if that is the case, that polygon will be filtered
    // out later in this routine
    APrimitive.SelectedEdges.GetAdjacent(polygonList);

    // walk through these polygons and per polygon, check if
    // the selected edges do NOT form a straight line; if they
    // don´t, they can be connected savely and are added to the
    // result edgelist
    selectedEdgesOfPolygon := TEdgeList.Create(false,true);
    relevantEdgesPerPolygon := TEdgeList.Create(false,true);
    for i := 0 to polygonList.Count-1 do begin
        polygon := polygonList[i];
        selectedEdgesOfPolygon.Clear;
        relevantEdgesPerPolygon.Clear;

        // how many edges of this polygon are selected? we need two or more
        // because we connect these edges by connecting their middle-vertices
        // which is done by a VertexConnector PER polygon
        if polygon.EdgeList.IsIn(APrimitive.SelectedEdges,selectedEdgesOfPolygon)>1 then begin
            // we´ve got at least 2 selected edges in this polygon;
            // let´s see if the selected edges are adjacent AND form a straight line
            for j := 0 to polygon.EdgeList.Count-1 do begin
                edge := polygon.EdgeList[j];
                if selectedEdgesOfPolygon.IsInList(edge) then begin
                    nextEdge := polygon.EdgeList[(j+1) mod polygon.EdgeList.Count];
                    if not selectedEdgesOfPolygon.IsInList(nextEdge) then
                        // the next edge is not selected, so only add current edge
                        relevantEdgesPerPolygon.Add(edge)
                    else begin
                        // both edges are selected and adjacent, let´s see if they form a straight line
                        middleVertex := edge.GetSharedVertex(nextEdge);
                        if Assigned(middleVertex) then begin        // somehow we need this check?
                            startVertex := edge.GetOtherVertex(middleVertex);
                            endVertex := nextEdge.GetOtherVertex(middleVertex);
                            if not middleVertex.OnEdge(startVertex,endVertex) then begin
                                relevantEdgesPerPolygon.Add(edge);
                                relevantEdgesPerPolygon.Add(nextEdge);
                            end;
                        end;
                    end;
                end;
            end;

            // store for later use
            if relevantEdgesPerPolygon.Count>1 then begin
                ASplittableEdgesList.AddList(relevantEdgesPerPolygon);
                ARelevantPolygonsList.Add(polygon);
            end;
        end;
    end;
    FreeAndNil(relevantEdgesPerPolygon);
    FreeAndNil(selectedEdgesOfPolygon);
    FreeAndNil(polygonList);
end;

procedure TEdgeConnector.DoConnect(AParamList: TParamList);
var i: integer;
    primitive: TBasePrimitive;
    relevantPolygons: TPolygonList;
    splitableEdges,newEdges,newSelection: TEdgeList;
    newVertices: TVerticeList;
begin
    newEdges := TEdgeList.Create(false,true);
    newSelection := TEdgeList.Create(false,true);
    newVertices := TVerticeList.Create(false,true);
    relevantPolygons := TPolygonList.Create(false,true);
    splitableEdges := TEdgeList.Create(false,true);

    for i := 0 to SelectedPrimitives.Count-1 do begin
        primitive := SelectedPrimitives[i];
        if primitive.ModifyGeometry then begin

            GetRelevantGeometry(primitive,relevantPolygons,splitableEdges);

            if splitableEdges.Count>1 then begin

                newVertices.Clear;
                newSelection.Clear;

                // split selected edges in half and set middle-vertex selected
                while splitableEdges.Count>0 do begin
                    newEdges.Clear;

                    // use an EdgeSplitter to spit the edge into two edges
                    TEdgeSpitFunction.Execute(primitive,splitableEdges[0],2,newEdges);

                    // add the middle-vertices to the new vertices list so we can connect them later
                    newVertices.Add(newEdges[0].GetSharedVertex(newEdges[1]));

                    newSelection.AddList(newEdges);

                    splitableEdges.Delete(0);
                end;

                // and connect the edges using the new vertices
                // NOTE: only affect relevant polygons here!
                TVertexConnectByPolygonsAndVerticesFunction.Execute(primitive,relevantPolygons,newVertices);

                // assign new edge selection
                primitive.SelectedEdges.Assign(newSelection);

                primitive.HasChanged := true;
            end;
        end;
    end;

    FreeAndNil(splitableEdges);
    FreeAndNil(relevantPolygons);
    FreeAndNil(newVertices);
    FreeAndNil(newSelection);
    FreeAndNil(newEdges);
end;

procedure TEdgeConnector.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Connect edges';
    Description := 'Connect edges';
end;

procedure TEdgeConnector.DoExecute(AParamList: TParamList);
begin
    DoConnect(AParamList);
end;

end.

