{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_EdgeSelector;

{$MODE Delphi}

interface

uses
  Classes,
  unit_Selector,
  unit_Types,
  unit_Primitives,
  unit_PrimitiveList;

type
  TEdgeSelector = class(TSelector)
  private
  protected
    function GetCenter: TVertice; override;
  public
    constructor Create(APrimitiveList: TPrimitiveList);

    procedure SelectByRange(AUpperLeftVertex, ALowerRightVertex: TVertice; AClearSelection: boolean = false);
    procedure SelectByVertex(AVertex: TVertice; ARange: double; AAddToSelection: boolean; AClearSelection: boolean = false); reintroduce; overload;
    procedure SelectByRay(ARay: TRay; AAddToSelection: boolean; AClearSelection: boolean = false);
    procedure SelectByRectangle(const ax1,ay1,ax2,ay2: integer;
                                const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                const aViewport: TViewport;
                                const AClearSelection: boolean);
    procedure SelectAll;
    procedure DeselectAll;
    procedure ReverseSelection;
    procedure GrowSelection;
    procedure ShrinkSelection;
    procedure SelectLoops;
    procedure SelectRings;

  end;

implementation

uses SysUtils,
     unit_Math;

// ### TEdgeSelector ###########################################################

constructor TEdgeSelector.Create(APrimitiveList: TPrimitiveList);
begin
    fPrimitives := APrimitiveList;
    inherited Create;
end;

function TEdgeSelector.GetCenter: TVertice;
var i,j,counter: integer;
    primitive: TBasePrimitive;
begin
    counter := 0;
    fCenter.Assign(0,0,0);
    for i := 0 to Primitives.Count-1 do begin
        primitive := Primitives[i];
        for j := 0 to primitive.SelectedEdges.Count-1 do begin
            fCenter.Add(primitive.SelectedEdges[j].StartVertex);
            fCenter.Add(primitive.SelectedEdges[j].EndVertex);
            inc(counter,2);
        end;
    end;
    if counter>0 then
        fCenter.DivideByScalar(counter)
    else
        fCenter.ResetToInfinite;
    Result := fCenter;
end;

procedure TEdgeSelector.SelectByRange(AUpperLeftVertex, ALowerRightVertex: TVertice; AClearSelection: boolean);
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do begin
        if AClearSelection then
            Primitives[i].SelectedEdges.Clear;
        Primitives[i].Edge_SelectMultiple(AUpperLeftVertex,ALowerRightVertex);
    end;
end;

procedure TEdgeSelector.SelectByVertex(AVertex: TVertice; ARange: double; AAddToSelection: boolean; AClearSelection: boolean);
var i: integer;
    primitive: TBasePrimitive;
    edge: TEdge;
begin
    for i := 0 to Primitives.Count-1 do begin
        primitive := Primitives[i];
        primitive.HighlightedEdges.Clear;

        if AClearSelection then
            primitive.SelectedEdges.Clear;
            
        edge := primitive.Edge_SelectByPoint(AVertex, ARange);
        if Assigned(edge) then begin
            if AAddToSelection then
                primitive.Edge_Select(edge)
            else
                primitive.HighlightedEdges.Add(edge);
        end;
    end;
end;

procedure TEdgeSelector.SelectByRay(ARay: TRay; AAddToSelection: boolean; AClearSelection: boolean);
var i: integer;
    primitive: TBasePrimitive;
    closestDistance, distance: single;
    closestPolygon, polygon: TPolygon;
begin
    closestDistance := -1;
    closestPolygon := nil;
    primitive := nil;

    // first determine if a polygon was hit
    for i := 0 to Primitives.Count-1 do begin
        primitive := Primitives[i];
        primitive.HighlightedEdges.Clear;
        if AClearSelection then
            primitive.SelectedEdges.Clear;
        if primitive.Polygon_Select(ARay, distance, polygon) and ((closestDistance = -1) or (distance < closestDistance)) then begin
            closestPolygon := polygon;
            closestDistance := distance;
        end;
    end;

    // if a polygon was hit, find out which edge is closest to
    // the intersection between the ray and the polygon
    if Assigned(closestPolygon) then
        primitive.Edge_Select(ARay,closestPolygon,closestDistance, AAddToSelection);
end;

procedure TEdgeSelector.SelectByRectangle(const ax1,ay1,ax2,ay2: integer;
                                          const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                          const aViewport: TViewport;
                                          const AClearSelection: boolean);
var
    primitive: TBasePrimitive;
    polygon: TPolygon;
    edge: TEdge;
    vertex1,vertex2: TVertice;
    point1,point2: TPoint2D;
    i,j,k: integer;
    Point1InRect,Point2InRect: boolean;
    PolygonPoints: TPoint2DArray;
begin
    if AClearSelection then
        DeselectAll;

    for i := 0 to Primitives.Count - 1 do begin
        primitive := Primitives[i];
        if AClearSelection then
            primitive.SelectedEdges.Clear;

        // for each polygon, find out if it faces the camera
        // if it does, check each edge to see if it is within our rectangle
        for j := 0 to primitive.NumPolygons - 1 do begin
            polygon := primitive.Polygons[j];
            polygon.ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewport,PolygonPoints);
            if Polygon2DFacesViewer(PolygonPoints) then begin
                for k := 0 to polygon.EdgeList.Count-1 do begin
                    edge := polygon.EdgeList[k];
                    vertex1 := edge.StartVertex;
                    vertex2 := edge.EndVertex;
                    point1 := vertex1.ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewPort);
                    point2 := vertex2.ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewPort);

                    Point1InRect := (point1.X >= ax1) and (point1.X <= ax2) and (point1.Y >= ay1) and (point1.Y <= ay2);
                    Point2InRect := (point2.X >= ax1) and (point2.X <= ax2) and (point2.Y >= ay1) and (point2.Y <= ay2);
                    if Point1InRect and Point2InRect then primitive.SelectedEdges.Add(edge);
                end;
            end;
        end;
    end;
end;

procedure TEdgeSelector.SelectAll;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Edge_SelectAll;
end;

procedure TEdgeSelector.DeselectAll;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Edge_DeselectAll;
end;

procedure TEdgeSelector.ReverseSelection;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Edge_ReverseSelection;
end;

procedure TEdgeSelector.GrowSelection;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Edge_GrowSelection;
end;

procedure TEdgeSelector.ShrinkSelection;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Edge_ShrinkSelection;
end;

procedure TEdgeSelector.SelectLoops;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Edge_SelectLoop;
end;

procedure TEdgeSelector.SelectRings;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Edge_SelectRing;
end;

end.

