{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_EdgeSplitterTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TEdgeSpitFunction = class(TObject)
  private
    procedure DoSplit(APrimitive: TBasePrimitive; AEdge: TEdge; ANumberOfSplits: integer; ANewEdgesList: TEdgeList);
  public
    class procedure Execute(APrimitive: TBasePrimitive; AEdge: TEdge; ANumberOfSplits: integer; ANewEdgesList: TEdgeList);
  end;

  TEdgeSplitter = class(TPrimitiveTool)
  private
    fNumberOfSplits: integer;
    procedure DoSplit();
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
    function CanExecute(AParamList: TParamList): boolean; override;
  end;

implementation

// ### TEdgeSpitFunction #######################################################

procedure TEdgeSpitFunction.DoSplit(APrimitive: TBasePrimitive; AEdge: TEdge; ANumberOfSplits: integer; ANewEdgesList: TEdgeList);
var i: integer;
    t: double;
    edgeToBeSplit: TEdge;
    vertex: TVertice;
    vertexList: TVerticeList;
    edgeList: TEdgeList;
begin
    if ANumberOfSplits>1 then begin

        vertexList := TVerticeList.Create(false,true);     // convenience templists
        edgeList := TEdgeList.Create(false);    // contains new edges while splitting the original edge: templist

        with APrimitive do begin

            // calculate new vertices which are to be inserted into the selected edges
            // and store them in a templist and the primitive verticeslist
            for i := 1 to ANumberOfSplits-1 do begin
                t := i / ANumberOfSplits;
                vertex := TVertice.Create(0,0,0);          // calculate new vertex
                vertex.x := AEdge.StartVertex.x + (AEdge.EndVertex.x-AEdge.StartVertex.x)*t;
                vertex.y := AEdge.StartVertex.y + (AEdge.EndVertex.y-AEdge.StartVertex.y)*t;
                vertex.z := AEdge.StartVertex.z + (AEdge.EndVertex.z-AEdge.StartVertex.z)*t;
                vertexList.Add(vertex);
                Vertices.Add(vertex);
            end;

            // now insert the new vertices into the current edge one by one
            edgeToBeSplit := AEdge;
            for i := 0 to vertexList.Count-1 do begin
                Edge_InsertVertex(edgeToBeSplit,vertexList[i],edgeList);
                // the first edge in the result is final, so store it in the final result set if wanted
                if Assigned(ANewEdgesList) then
                    ANewEdgesList.Add(edgeList[0]);
                // always use the second edge as next edge to be split, due to the way the new vertices are calculated
                edgeToBeSplit := edgeList[1];
            end;
            if Assigned(ANewEdgesList) then
                ANewEdgesList.Add(edgeList[1]);
        end;

        FreeAndNil(vertexList);
        FreeAndNil(edgeList);
    end;
end;

class procedure TEdgeSpitFunction.Execute(APrimitive: TBasePrimitive; AEdge: TEdge; ANumberOfSplits: integer; ANewEdgesList: TEdgeList);
begin
    with self.Create do
    try
        DoSplit(APrimitive, AEdge, ANumberOfSplits, ANewEdgesList);
    finally
        Free;
    end;
end;

// ### TEdgeSplitter ###########################################################

// any split business rules go here; from checking provided parameters for valid values
// to checking if geometry can or cannot be split for whatever reason
{$HINTS OFF}
function TEdgeSplitter.CanExecute(AParamList: TParamList): boolean;
var canDoSplit: boolean;
begin
    canDoSplit := false;
    if Assigned(AParamList) then begin
        if AParamList.Count=1 then begin

            canDoSplit := (AParamList[0] is TIntegerObject);
            if not canDoSplit then
                raise Exception.Create('Invalid parameter type provided for edge splitter tool.')
            else begin
                fNumberOfSplits := TIntegerObject(AParamList[0]).Value;
                if fNumberOfSplits<2 then begin
                    canDoSplit := false;
                    raise Exception.Create('Must split into at least 2 parts.');
                end;
            end;

        end else begin
            if AParamList.Count>1 then begin

                // check parameter types
                canDoSplit := (AParamList[0] is TBasePrimitive) and
                              (AParamList[1] is TEdge) and
                              (AParamList[2] is TIntegerObject) and
                              (AParamList[3] is TEdgeList);

                // the right types of parameters have been provided;
                // now check for valid contents of these parameters
                if not canDoSplit then
                    raise Exception.Create('Invalid parameter types provided for edge splitter tool.')
                else begin
                    fNumberOfSplits := TIntegerObject(AParamList[2]).Value;
                    if fNumberOfSplits<2 then begin
                        canDoSplit := false;
                        raise Exception.Create('Must split into at least 2 parts.');
                    end;
                end;
            end;
        end;
    end;

    Result := canDoSplit;
end;
{$HINTS ON}

procedure TEdgeSplitter.DoSplit();
var i: integer;
    edge: TEdge;
    newEdges: TEdgeList;
begin
    newEdges := TEdgeList.Create(false,true);

    for i := 0 to SelectedPrimitives.Count-1 do begin
        with SelectedPrimitives[i] do begin
            if ModifyGeometry then begin
                while SelectedEdges.Count>0 do begin
                    edge := SelectedEdges[0];
                    TEdgeSpitFunction.Execute(SelectedPrimitives[i],edge,fNumberOfSplits,newEdges);
                    // remove old edge from selection (it was already removed from the datastructure by SplitEdge)
                    SelectedEdges.Remove(edge);
                end;
                // store new edges as the new selection
                SelectedEdges.Assign(newEdges);
            end;
        end;
    end;
    FreeAndNil(newEdges);
end;

procedure TEdgeSplitter.DoPrepare(AParamList: TParamList);
begin
end;

procedure TEdgeSplitter.DoExecute(AParamList: TParamList);
begin
    if AParamList.Count=1 then
        DoSplit()
    else
        TEdgeSpitFunction.Execute(TBasePrimitive(AParamList[0]),TEdge(AParamList[1]),TIntegerObject(AParamList[2]).Value,TEdgeList(AParamList[3]));
end;


end.

