{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_FaceMaterialReplacerTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_Primitives,
  unit_ParamList,
  unit_PrimitiveTool,
  unit_PrimitiveList,
  unit_Materials,
  unit_MaterialList;

type
  TFaceEyeDropFunction = class(TObject)
  private
    function DoEyeDrop(APrimitiveList: TPrimitiveList; ARay: TRay):  TMaterial;
  public
    class function Execute(APrimitiveList: TPrimitiveList; ARay: TRay): TMaterial;
  end;

  TFaceGetUsedMaterialsFunction = class(TObject)
  private
    procedure GetUsedMaterials(APrimitiveList: TPrimitiveList; AGetMaterialOptions: TGetMaterialsOptions; var AMaterialList: TMaterialList);
  public
    class procedure Execute(APrimitiveList: TPrimitiveList; AGetMaterialOptions: TGetMaterialsOptions; var AMaterialList: TMaterialList);
  end;

  TFaceMaterialReplacerFunction = class(TObject)
  private
    procedure DoReplaceMaterial(APrimitive: TBasePrimitive; ACurrentMaterial, ANewMaterial: TMaterial; ASelectedPolygonsOnly: boolean);
  public
    class procedure Execute(APrimitive: TBasePrimitive; ACurrentMaterial, ANewMaterial: TMaterial; ASelectedPolygonsOnly: boolean);
  end;

  TFaceMaterialReplacer = class(TPrimitiveTool)
  private
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  Math,
  unit_Textures;

// ### TFaceEyeDropFunction ####################################################

function TFaceEyeDropFunction.DoEyeDrop(APrimitiveList: TPrimitiveList; ARay: TRay): TMaterial;
var i: integer;
    obj: TBasePrimitive;
    distance: single;
    polygon,selectedPolygon: TPolygon;
    aDistance: single;
begin
    Result := nil;
    aDistance := -1;
    selectedPolygon := nil;

    for i := 0 to APrimitiveList.Count-1 do begin
        obj := APrimitiveList[i];
        if obj.Polygon_Select(aRay, distance, polygon) and ((aDistance = -1) or (distance < aDistance)) then begin
            selectedPolygon := polygon;
            aDistance := distance;
        end;
    end;

    if Assigned(selectedPolygon) then begin
        if not selectedPolygon.Material.IsOriginal then
            Result := selectedPolygon.Material.OriginalMaterial
        else
            Result := selectedPolygon.Material;
    end;
end;

class function TFaceEyeDropFunction.Execute(APrimitiveList: TPrimitiveList; ARay: TRay): TMaterial;
begin
    with self.Create do
    try
        Result := DoEyeDrop(APrimitiveList, ARay);
    finally
        Free;
    end;
end;

// ### TFaceGetUsedMaterialsFunction ###########################################

procedure TFaceGetUsedMaterialsFunction.GetUsedMaterials(APrimitiveList: TPrimitiveList; AGetMaterialOptions: TGetMaterialsOptions; var AMaterialList: TMaterialList);
var i,j: integer;
    primitive: TBasePrimitive;
    primitiveMaterialList: TMaterialList;
begin
    if not Assigned(AMaterialList) then
      raise Exception.Create('TFaceGetUsedMaterialsFunction failed. AMaterialList must not be NIL.');
    
    for i := 0 to APrimitiveList.Count-1 do begin
        primitive := APrimitiveList[i];
        primitiveMaterialList := primitive.Face_GetUsedMaterials(AGetMaterialOptions);
        for j := 0 to primitiveMaterialList.Count-1 do begin
            if AMaterialList.IndexOfMaterial(primitiveMaterialList.Materials[j])=-1 then
                AMaterialList.Add(primitiveMaterialList.Materials[j]);
        end;
        FreeAndNil(primitiveMaterialList);
    end;
end;

class procedure TFaceGetUsedMaterialsFunction.Execute(APrimitiveList: TPrimitiveList; AGetMaterialOptions: TGetMaterialsOptions; var AMaterialList: TMaterialList);
begin
    with self.Create do
    try
        GetUsedMaterials(APrimitiveList, AGetMaterialOptions,AMaterialList);
    finally
        Free;
    end;
end;

// ### TFaceMaterialReplacerFunction ###########################################

procedure TFaceMaterialReplacerFunction.DoReplaceMaterial(APrimitive: TBasePrimitive; ACurrentMaterial, ANewMaterial: TMaterial; ASelectedPolygonsOnly: boolean);
var i: integer;
    polygon: TPolygon;
begin
    if ASelectedPolygonsOnly then begin
        for i := 0 to APrimitive.SelectedPolygons.Count - 1 do begin
            polygon := APrimitive.SelectedPolygons[i];
            if polygon.Material = ACurrentMaterial then
                polygon.Material := ANewMaterial;
        end;
    end else begin
        for i := 0 to APrimitive.Polygons.Count - 1 do begin
            polygon := APrimitive.Polygons[i];
            if polygon.Material = ACurrentMaterial then
                polygon.Material := ANewMaterial;
        end;
    end;
end;

class procedure TFaceMaterialReplacerFunction.Execute(APrimitive: TBasePrimitive; ACurrentMaterial, ANewMaterial: TMaterial; ASelectedPolygonsOnly: boolean);
begin
    with self.Create do
    try
        DoReplaceMaterial(APrimitive, ACurrentMaterial, ANewMaterial, ASelectedPolygonsOnly);
    finally
        Free;
    end;
end;

// ### TFaceMaterialReplacer ###################################################

procedure TFaceMaterialReplacer.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Replace material';
    Description := 'Replace the material of polygons';
end;

procedure TFaceMaterialReplacer.DoExecute(AParamList: TParamList);
var i,j: integer;
    primitive: TBasePrimitive;
    polygon: TPolygon;
    currentMaterials: TMaterialList;
    newMaterial: TMaterial;
    doSelectionOnly: boolean;
begin
    currentMaterials := TMaterialList(AParamList.NamedParameters['currentMaterials']);
    newMaterial := TMaterial(AParamList.NamedParameters['newMaterial']);
    doSelectionOnly := TBooleanObject(AParamList.NamedParameters['doSelectionOnly']).Value;

    if doSelectionOnly then begin
        for i := 0 to SelectedPrimitives.Count-1 do begin
            primitive := SelectedPrimitives[i];

            if primitive is TPrimitive then begin
                for j := 0 to primitive.SelectedPolygons.Count - 1 do begin
                    polygon := primitive.SelectedPolygons[j];
                    if currentMaterials.IndexOf(polygon.Material) <> - 1 then
                        polygon.Material := newMaterial;
                end;
            end;
        end;
    end else begin
        for i := 0 to ScenePrimitives.Count-1 do begin
            primitive := ScenePrimitives[i];

            if primitive is TPrimitive then begin
                for j := 0 to primitive.Polygons.Count - 1 do begin
                    polygon := primitive.Polygons[j];
                    if currentMaterials.IndexOf(polygon.Material) <> - 1 then
                        polygon.Material := newMaterial;
                end;
            end;
        end;
    end;
end;

end.

