{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_GenericColorDialog;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  Delgine's extended color dialog, with optional alpha support.

  Notes:
    -  When alpha support is enabled, the alpha is passed as the high byte in
       the color value (i.e. the usual 32-bit color convention). However,
       Delphi's VCL components use this high byte to mark system colors, so
       keep in mind those are not compatible.
    -  While to the user H, S and V are presented in the 0..360 (H), 0..100
       (S,V) range, we store them internally as bytes (0..255)
    -  Alpha support flag is passed through the constructor now, because we need
       to know about it during form initialisation. We should make this a little
       more elegant some day :)
    -  If you want to store the custom colors (in an .ini file, for example),
       you'll need to override the StoreCustomColors and RestoreCustomColors
       methods.
  ---------------------------------------------------------------------------- }


interface

uses
    Controls, StdCtrls, Windows, Graphics, Forms, ExtCtrls, Classes;

type
    TSimpleFrame = class;

    TGenericColorDialog = class( TObject )
    private
        fForm: TForm;
        fLeftArrow: TBitmap;
        fRightArrow: TBitmap;
        fPreviousColor: TColor;
        fColorBuffer: Array of TColor;
        fMarkerBuffer: TBitmap;
        fSaturationValueImage: TImage;
        fHueImage: TImage;
        fColorImage: TImage;
        fCustomColorImage: TImage;

        fCustomColors: array[0..31] of TColor;
        fSelectedCustomColorIndex: integer;
        fAssignToCustomButton: TButton;


        fREdit: TEdit;
        fGEdit: TEdit;
        fBEdit: TEdit;

        fHEdit: TEdit;
        fSEdit: TEdit;
        fVEdit: TEdit;

        fAEdit: TEdit;

        fHue: Byte;
        fSaturation: Byte;
        fValue: Byte;
        fAlpha: Byte;

        fIsHueSelected: Boolean;
        fIsSaturationValueSelected: Boolean;
        fIsUpdating: Boolean;

        fSupportsAlpha: Boolean;
        fAllowsZeroAlpha: Boolean;

        procedure InitializeForm( const aForm: TForm );
        procedure AddButton( const aCaption: String; const X, Y: Integer; const aModalResult: TModalResult; const anIsDefault, aShouldCancel: Boolean; const anOwner: TForm );
        procedure AddBevel( const Left, Top, Right, Bottom: Integer; const anOwner: TForm );
        function AddEditField(
            const aLabelCaption: String;
            const X, Y: Integer;
            const anOwner: TForm;
            const aChangeHandler: TNotifyEvent;
            const aShouldAllowZero: Boolean = TRUE ): TEdit;
        procedure DrawCustomColorSelection( const anIndex: integer; const aColor: TColor );

        procedure SetHueForPosition( Y: Integer );
        procedure SetSaturationValueForPosition( X, Y: Integer );
        procedure SetHue( aHue: Byte );
        procedure SetSaturationValue( aSaturation, aValue: Byte );

        procedure UpdateCustomColor( anIndex: Integer );
        procedure UpdateCustomColors;
        procedure UpdateColorImage;
        procedure UpdatePreviousColorImage;
        procedure UpdateSaturationValue;
        procedure UpdateRGBFields;
        procedure UpdateHSVFields;
        procedure UpdateAllFields;

        function HSV2RGBQuick( H, S, V: Byte ): TColor;
        function HSV2RGBPrecise( anH, anS, aV: Byte ): TColor;
        function RGB2BGR( const aColor: TColor ): TColor;
        procedure RGB2HSV( const aColor: TColor; out H, S, V: Byte );

        procedure HandleColorMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
        procedure HandleCustomColorMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
        procedure HandleAssignToCustomClick( aSender: TObject );
        procedure HandleHueMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
        procedure HandleHueMouseMove( aSender: TObject; aShift: TShiftState; X, Y: Integer);
        procedure HandleHueMouseUp( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
        procedure HandleSaturationValueMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
        procedure HandleSaturationValueMouseMove( aSender: TObject; aShift: TShiftState; X, Y: Integer);
        procedure HandleSaturationValueMouseUp( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
        procedure HandleRGBChange( aSender: TObject );
        procedure HandleHSVChange( aSender: TObject );
        procedure HandleIntegerKeyPress( aSender: TObject; var aKey: Char );
        procedure HandleNonZeroIntegerKeyPress( aSender: TObject; var aKey: Char );

    protected
        function GetForm: TForm;
        function GetLeftArrow: TBitmap;
        function GetRightArrow: TBitmap;

        function GetColor: TColor;
        procedure SetColor( const aColor: TColor );
        procedure SetColorEx( const aColor: TColor );

        function GetCustomColor( anIndex: integer ): TColor;
        procedure SetCustomColor( anIndex: integer; aColor: TColor );

        function GetR: Byte;
        function GetG: Byte;
        function GetB: Byte;
        function GetA: Byte;

        procedure RestoreCustomColors; virtual;
        procedure StoreCustomColors; virtual;

        property Form: TForm read GetForm;
        property LeftArrow: TBitmap read GetLeftArrow;
        property RightArrow: TBitmap read GetRightArrow;

    public
        constructor Create( const aSupportsAlpha: boolean = false );
        destructor Destroy; override;

        property R: Byte read GetR;
        property G: Byte read GetG;
        property B: Byte read GetB;
        property A: Byte Read GetA write fAlpha;

        function Execute: boolean;

        property Color: TColor read GetColor write SetColor;
        property SupportsAlpha: boolean read fSupportsAlpha;
        property AllowsZeroAlpha: boolean read fAllowsZeroAlpha;
        property CustomColors[anIndex: Integer]: TColor read GetCustomColor write SetCustomColor;

    end;

// ----------------------------------------------------------------------------

    // TSimpleFrame: A little helper component to draw a flat bevel
    TSimpleFrame = class( TGraphicControl )
    public
        procedure Paint; override;

        property Color;
    end;


implementation

uses
    Math, SysUtils;

const
    CUSTOM_COLORS_LEFT = 10;
    CUSTOM_COLORS_TOP = 10;

    SATURATION_VALUE_GRID_LEFT = 60;
    SATURATION_VALUE_GRID_TOP = 10;

    HUE_BAR_WIDTH = 24;
    HUE_BAR_OFFSET_X = 6;
    HUE_BAR_OFFSET_Y = 4;
    HUE_BAR_LEFT = 330;
    HUE_BAR_TOP = SATURATION_VALUE_GRID_TOP - HUE_BAR_OFFSET_Y;

    COLOR_IMAGE_TOP = SATURATION_VALUE_GRID_TOP;
    COLOR_IMAGE_LEFT = 380;
    COLOR_IMAGE_WIDTH = 34;
    COLOR_IMAGE_HEIGHT = 34;

    ASSIGN_BUTTON_LEFT = COLOR_IMAGE_LEFT + COLOR_IMAGE_WIDTH + 9;
    ASSIGN_BUTTON_WIDTH = 100;
    ASSIGN_BUTTON_TOP = COLOR_IMAGE_TOP;

    EDITS_LEFT = COLOR_IMAGE_LEFT + 12;
    EDITS_TOP = 91;

    BUTTONS_LEFT = 380;
    BUTTONS_TOP = 247;

    DEFAULT_CUSTOM_COLORS: array[0..31] of Cardinal = (
        $FF000000,
        $FF000080,
        $FF008000,
        $FF008080,
        $FF800000,
        $FF800080,
        $FF808000,
        $FFC0C0C0,
        $FFC0DCC0,
        $FFF0CAA6,
        $FFF0FBFF,
        $FFA4A0A0,
        $FF808080,
        $FF0000FF,
        $FF00FF00,
        $FF00FFFF,
        $FFFF0000,
        $FFFF00FF,
        $FFFFFF00,
        $FFFFFFFF,
        $FF101010,
        $FF202020,
        $FF303030,
        $FF404040,
        $FF505050,
        $FF606060,
        $FF707070,
        $FF808080,
        $FF909090,
        $FFA0A0A0,
        $FFB0B0B0,
        $FFC0C0C0
    );


procedure TGenericColorDialog.InitializeForm( const aForm: TForm );
var index, alphaOffset: integer;
    frame: TSimpleFrame;
    hue: byte;

begin
    aForm.Width := 543;
    aForm.Height := 304;
    aForm.BorderStyle := bsDialog;
    aForm.Caption := 'Color picker';
    aForm.Position := poScreenCenter;

    // custom colors
    fCustomColorImage := TImage.Create( aForm );
    fCustomColorImage.Parent := aForm;
    fCustomColorImage.Left := CUSTOM_COLORS_LEFT;
    fCustomColorImage.Top := CUSTOM_COLORS_TOP;
    fCustomColorImage.Width := 16 * 2 + 1;
    fCustomColorImage.Height := 16 * 16 + 1;
    fCustomColorImage.Canvas.Brush.Color := clBlack;
    fCustomColorImage.Canvas.FillRect(fCustomColorImage.ClientRect);
    fCustomColorImage.OnMouseDown := HandleCustomColorMouseDown;
    fCustomColorImage.Stretch := TRUE;

    for index := 0 to 31 do
    begin
        fCustomColors[index] := TColor(DEFAULT_CUSTOM_COLORS[index]);
    end;

    frame := TSimpleFrame.Create( aForm );
    frame.Parent := aForm;
    frame.Left := SATURATION_VALUE_GRID_LEFT;
    frame.Top := SATURATION_VALUE_GRID_TOP;
    frame.Width := 256 + 2;
    frame.Height := 256 + 2;
    frame.Color := clBlack;

    // Saturation value grid
    fSaturationValueImage := TImage.Create( aForm );
    fSaturationValueImage.Parent := aForm;
    fSaturationValueImage.Stretch := TRUE; // to prevent flicker
    fSaturationValueImage.Left := SATURATION_VALUE_GRID_LEFT + 1;
    fSaturationValueImage.Top := SATURATION_VALUE_GRID_TOP + 1;
    fSaturationValueImage.Width := 256;
    fSaturationValueImage.Height := 256;

    fSaturationValueImage.OnMouseDown := HandleSaturationValueMouseDown;
    fSaturationValueImage.OnMouseMove := HandleSaturationValueMouseMove;
    fSaturationValueImage.OnMouseUp := HandleSaturationValueMouseUp;

    // Hue bar
    fHueImage := TImage.Create( aForm );
    fHueImage.Parent := aForm;
    fHueImage.Transparent := false;
    fHueImage.Stretch := TRUE;
    fHueImage.Proportional := false;

    fHueImage.Left := HUE_BAR_LEFT;
    fHueImage.Top := HUE_BAR_TOP;
    fHueImage.Width := HUE_BAR_WIDTH + 2 + 2 * HUE_BAR_OFFSET_X;
    fHueImage.Height := 256 + 2 * HUE_BAR_OFFSET_Y + 2;
    fHueImage.Canvas.Brush.Color := clBtnFace;
    fHueImage.Canvas.FillRect( fHueImage.ClientRect );
    fHueImage.Canvas.Pen.Color := clBlack;
    fHueImage.Canvas.Rectangle( Rect( HUE_BAR_OFFSET_X, HUE_BAR_OFFSET_Y, HUE_BAR_OFFSET_X + HUE_BAR_WIDTH + 2, HUE_BAR_OFFSET_Y + 256 + 2 ));

    for index := 0 to 255 do
    begin
        fHueImage.Canvas.Brush.Color := HSV2RGBPrecise( 255 - index, 255, 255 );
        fHueImage.Canvas.FillRect( Rect( HUE_BAR_OFFSET_X + 1, HUE_BAR_OFFSET_Y + index + 1, HUE_BAR_OFFSET_X + HUE_BAR_WIDTH + 1, HUE_BAR_OFFSET_Y + index + 2 ));
    end;

    fHueImage.OnMouseDown := HandleHueMouseDown;
    fHueImage.OnMouseMove := HandleHueMouseMove;
    fHueImage.OnMouseUp := HandleHueMouseUp;

    // Color image
    fColorImage := TImage.Create( aForm );
    fColorImage.Parent := aForm;
    fColorImage.Top := COLOR_IMAGE_TOP;
    fColorImage.Left := COLOR_IMAGE_LEFT;
    fColorImage.Width := COLOR_IMAGE_WIDTH;
    fColorImage.Height := COLOR_IMAGE_HEIGHT;
    fColorImage.Canvas.Brush.Color := clBlack;
    fColorImage.Canvas.FrameRect( fColorImage.ClientRect );
    fColorImage.Stretch := TRUE;           // prevent flicker
    fColorImage.OnMouseDown := HandleColorMouseDown;

    // Assign to custom colors button
    fAssignToCustomButton := TButton.Create( aForm );
    fAssignToCustomButton.Parent := aForm;
    fAssignToCustomButton.Left := ASSIGN_BUTTON_LEFT;
    fAssignToCustomButton.Top := ASSIGN_BUTTON_TOP;
    fAssignToCustomButton.Caption := 'Assign to Custom';
    fAssignToCustomButton.Width := ASSIGN_BUTTON_WIDTH;
    fAssignToCustomButton.Height := 22;
    fAssignToCustomButton.Enabled := FALSE;
    fAssignToCustomButton.OnClick := HandleAssignToCustomClick;

    if fSupportsAlpha then begin
        alphaOffset := 24;
    end
    else
        alphaOffset := 0;
        
    // Fields and buttons
    fHEdit := AddEditField( 'H', EDITS_LEFT, EDITS_TOP - alphaOffset, aForm, HandleHSVChange );
    fSEdit := AddEditField( 'S', EDITS_LEFT, EDITS_TOP - alphaOffset + 22, aForm, HandleHSVChange );
    fVEdit := AddEditField( 'V', EDITS_LEFT, EDITS_TOP - alphaOffset + 44, aForm, HandleHSVChange );
    fREdit := AddEditField( 'R', EDITS_LEFT, EDITS_TOP - alphaOffset + 76, aForm, HandleRGBChange );
    fGEdit := AddEditField( 'G', EDITS_LEFT, EDITS_TOP - alphaOffset + 98, aForm, HandleRGBChange );
    fBEdit := AddEditField( 'B', EDITS_LEFT, EDITS_TOP - alphaOffset + 120, aForm, HandleRGBChange );

    AddBevel( EDITS_LEFT - 12, EDITS_TOP - alphaOffset - 6, aForm.Width - 21, EDITS_TOP - alphaOffset + 69, aForm );
    AddBevel( EDITS_LEFT - 12, EDITS_TOP - alphaOffset + 70, aForm.Width - 21, EDITS_TOP - alphaOffset + 145, aForm );

    if fSupportsAlpha then
    begin
        fAEdit := AddEditField( 'A', EDITS_LEFT, EDITS_TOP - alphaOffset + 152, aForm, nil, fAllowsZeroAlpha );
        AddBevel( EDITS_LEFT - 12, EDITS_TOP - alphaOffset + 146, aForm.Width - 21, EDITS_TOP - alphaOffset + 177, aForm );
    end;

    AddButton( 'OK', BUTTONS_LEFT, BUTTONS_TOP, mrOK, TRUE, FALSE, aForm );
    AddButton( 'Cancel', BUTTONS_LEFT + 75, BUTTONS_TOP, mrCancel, FALSE, TRUE, aForm );

    aForm.ActiveControl := fREdit;

    SetSaturationValue( fSaturation, fValue );

    // litle hack to enforce initial update without introducing extra state
    hue := fHue;
    SetHue( not hue );
    SetHue( hue );

    if fSupportsAlpha then
    begin
        fAEdit.Text := IntToStr( fAlpha );
    end;

    RestoreCustomColors;

    UpdatePreviousColorImage;
    UpdateCustomColors;
    UpdateAllFields;
    UpdateColorImage;
end;


procedure TGenericColorDialog.AddButton( const aCaption: String; const X, Y: Integer; const aModalResult: TModalResult; const anIsDefault, aShouldCancel: Boolean; const anOwner: TForm );
var button: TButton;
begin
    button := TButton.Create( anOwner );
    button.Parent := anOwner;
    button.Left := X;
    button.Top := Y;
    button.Caption := aCaption;
    button.ModalResult := aModalResult;
    button.Default := anIsDefault;
    button.Cancel := aShouldCancel;
    button.Height := 22;
    button.Width := 70;
end;

procedure TGenericColorDialog.AddBevel( const Left, Top, Right, Bottom: Integer; const anOwner: TForm );
var bevel: TBevel;
begin
    bevel := TBevel.Create( anOwner );
    bevel.Parent := anOwner;
    bevel.Left := Left;
    bevel.Top := Top;
    bevel.Width := Right - Left + 1;
    bevel.Height := Bottom - Top + 1;
    bevel.Shape := bsBox;
    bevel.Style := bsRaised;
end;

function TGenericColorDialog.AddEditField(
    const aLabelCaption: String;
    const X, Y: Integer;
    const anOwner: TForm;
    const aChangeHandler: TNotifyEvent;
    const aShouldAllowZero: Boolean = true ): TEdit;
var lbl: TLabel;
begin
    lbl := TLabel.Create( anOwner );
    lbl.Parent := anOwner;
    lbl.Left := X;
    lbl.Top := Y + 4;
    lbl.Caption := aLabelCaption;

    result := TEdit.Create( anOwner );
    result.Parent := anOwner;
    result.Left := X + 12;
    result.Top := Y;
//    result.Ctl3D := FALSE;
    result.Width := 100;

    result.OnChange := aChangeHandler;

    if aShouldAllowZero then
        result.OnKeyPress := handleIntegerKeyPress
    else
        result.OnKeyPress := handleNonZeroIntegerKeyPress; 
end;

procedure TGenericColorDialog.DrawCustomColorSelection( const anIndex: integer; const aColor: TColor );
var left, top: integer;
begin
    left := ( anIndex div 16 ) * 16;
    top := (anIndex mod 16 ) * 16;

    fCustomColorImage.Canvas.Brush.Color := aColor;
    fCustomColorImage.Canvas.FrameRect( Rect( left, top, left + 17, top + 17 ))
end;

procedure TGenericColorDialog.SetHue( aHue: Byte );
var huePosition: integer;
begin
    if aHue <> fHue then
    begin
        if Assigned(fForm) then
        begin
            huePosition := 255 - fHue;

            fHueImage.Canvas.Brush.Color := clBtnFace;
            fHueImage.Canvas.FillRect( Rect( 0, huePosition, HUE_BAR_OFFSET_X, huePosition + 11 ));
            fHueImage.Canvas.FillRect( Rect( HUE_BAR_OFFSET_X + HUE_BAR_WIDTH + 2, huePosition, 2 * HUE_BAR_OFFSET_X + HUE_BAR_WIDTH + 2, huePosition + 11 ));

            huePosition := 255 - aHue;
            fHueImage.Canvas.Draw( 0, huePosition, leftArrow );
            fHueImage.Canvas.Draw( HUE_BAR_OFFSET_X + HUE_BAR_WIDTH + 2, huePosition, rightArrow );

            fHueImage.Invalidate;
        end;

        fHue := aHue;

        if Assigned( fForm ) then
        begin
            UpdateSaturationValue;
        end;
    end;
end;

procedure TGenericColorDialog.SetSaturationValueForPosition( X, Y: Integer );
var saturation, value: integer;
begin
    saturation := X;
    if saturation < 0 then saturation := 0;
    if saturation > 255 then saturation := 255;

    value := 255- Y;
    if value < 0 then value := 0;
    if value > 255 then value := 255;

    SetSaturationValue( saturation, value );
    UpdateAllFields;
    UpdateColorImage;
end;

procedure TGenericColorDialog.SetSaturationValue( aSaturation, aValue: Byte );
var color: TColor;
begin
    if not Assigned( fMarkerBuffer ) then begin
        fMarkerBuffer := TBitmap.Create;
        fMarkerBuffer.Width := 7;
        fMarkerBuffer.Height := 7;
    end
    else begin
        fSaturationValueImage.Canvas.CopyRect(
            Rect( fSaturation-3, 252-fValue, fSaturation+4, 259-fValue ),
            fMarkerBuffer.Canvas,
            Rect(0,0,7,7));
    end;

    fMarkerBuffer.Canvas.CopyRect(
            Rect(0,0,7,7),
            fSaturationValueImage.Canvas,
            Rect( aSaturation-3, 252-aValue, aSaturation+4, 259-aValue ));

    if aValue > 127 then color := clBlack else color := clWhite;
    fSaturationValueImage.Canvas.Brush.Color := color;
    fSaturationValueImage.Canvas.FrameRect( Rect(aSaturation-3, 252-aValue, aSaturation+4, 259-aValue ));

    fSaturationValueImage.Invalidate;

    fSaturation := aSaturation;
    fValue := aValue;
end;

procedure TGenericColorDialog.UpdateCustomColor( anIndex: Integer );
var left, top: integer;
begin
    fCustomColorImage.Canvas.Brush.Color := CustomColors[anIndex] and $00FFFFFF;
    left := ( anIndex div 16 ) * 16 + 1;
    top := (anIndex mod 16 ) * 16 + 1;
    fCustomColorImage.Canvas.FillRect( Rect( left, top, left + 15, top + 15 ));
end;

procedure TGenericColorDialog.UpdateColorImage;
begin
    fColorImage.Canvas.Brush.Color := Color and $00FFFFFF;
    fColorImage.Canvas.FillRect( Rect( 1, 1, COLOR_IMAGE_WIDTH - 1, COLOR_IMAGE_HEIGHT div 2 ));
end;

procedure TGenericColorDialog.UpdatePreviousColorImage;
begin
    fColorImage.Canvas.Brush.Color := fPreviousColor and $00FFFFFF;
    fColorImage.Canvas.FillRect( Rect( 1, COLOR_IMAGE_HEIGHT div 2, COLOR_IMAGE_WIDTH - 1, COLOR_IMAGE_HEIGHT - 1 ));
end;

procedure TGenericColorDialog.UpdateSaturationValue;
var xIndex, yIndex, index: integer;
    bitmapInfo: TBitmapInfo;
begin
    if not Assigned( fColorBuffer ) then begin
        SetLength( fColorBuffer, 256 * 256 );
    end;

    index := 0;
    for xIndex := 0 to 255 do begin
        for yIndex := 0 to 255 do begin
            fColorBuffer[index] := RGB2BGR(HSV2RGBQuick( fHue, yIndex, (255 - xIndex )));
            index := index + 1;
        end;
    end;

    FillChar(bitmapInfo, SizeOf(bitmapInfo), 0);
    bitmapInfo.bmiHeader.biSize := SizeOf( TBitmapInfoHeader );
    bitmapInfo.bmiHeader.biWidth := 256;
    bitmapInfo.bmiHeader.biHeight := -256;
    bitmapInfo.bmiHeader.biPlanes := 1;
    bitmapInfo.bmiHeader.biBitCount := 32;
    bitmapInfo.bmiHeader.biCompression := BI_RGB;
    SetDIBits( fSaturationValueImage.Canvas.Handle, fSaturationValueImage.Picture.Bitmap.Handle, 0, 256, @fColorBuffer[0], bitmapInfo, DIB_RGB_COLORS );

    FreeAndNil( fMarkerBuffer );
    SetSaturationValue( fSaturation, fValue );

    fSaturationValueImage.Invalidate;
end;

procedure TGenericColorDialog.UpdateRGBFields;
var color: TColor;
begin
    fIsUpdating := TRUE;

    color := HSV2RGBPrecise( fHue, fSaturation, fValue );
    fREdit.Text := IntToStr( color and $000000FF );
    fGEdit.Text := IntToStr(( color and $0000FF00 ) shr 8);
    fBEdit.Text := IntToStr(( color and $00FF0000 ) shr 16);

    fIsUpdating := FALSE;
end;

procedure TGenericColorDialog.UpdateHSVFields;
begin
    fIsUpdating := TRUE;

    fHEdit.Text := IntToStr( fHue * 360 div 255 );
    fSEdit.Text := IntToStr( fSaturation * 100 div 255 );
    fVEdit.Text := IntToStr( fValue * 100 div 255 );

    fIsUpdating := FALSE;
end;

procedure TGenericColorDialog.UpdateCustomColors;
var index: integer;
begin
    for index := 0 to 31 do
    begin
        UpdateCustomColor( index );
    end;
end;

procedure TGenericColorDialog.UpdateAllFields;
begin
    UpdateColorImage;
    UpdateRGBFields;
    UpdateHSVFields;
end;

procedure TGenericColorDialog.RestoreCustomColors;
begin
    // Empty on purpose, override in project-specific class
end;

procedure TGenericColorDialog.StoreCustomColors; 
begin
    // Empty on purpose, override in project-specific class
end;

// Quick and dirty version (good for drawing grids)
function TGenericColorDialog.HSV2RGBQuick( H, S, V: Byte ): TColor;
var r, g, b: Integer;
    varH, varI: Integer;
    var1, var2, var3: Integer;
begin
    if S = 0 then
    begin
        r := V;
        g := V;
        b := V;
    end
    else
    begin
        varH := H * 6;
        varI := varH and $FFFFFF00;

        var1 := V * ( 255 - S ) div 256;
        var2 := V * ( 255 - S * ( varH - varI ) div 256 ) div 256;
        var3 := v * ( 255 - S * ( 255 - ( varH - varI )) div 256) div 256;

        case varI of
            0:  begin
                    r := V;
                    g := var3;
                    b := var1;
                end;
            256:  begin
                    r := var2;
                    g := V;
                    b := var1;
                end;
            512:  begin
                    r := var1;
                    g := V;
                    b := var3;
                end;
            768:  begin
                    r := var1;
                    g := var2;
                    b := V;
                end;
            1024:  begin
                    r := var3;
                    g := var1;
                    b := V;
                end;
            else
            begin
                r := V;
                g := var1;
                b := var2;
            end;
        end;
    end;

    result := RGB( r, g, b );
end;

// Precise version (good for doing final conversions)
function TGenericColorDialog.HSV2RGBPrecise( anH, anS, aV: Byte ): TColor;
var r, g, b: Single;
    h, s, v: Single;
    varH: Single;
    varI: Integer;
    var1, var2, var3: Single;
begin
    if anS = 0 then
    begin
        r := aV / 255;
        g := aV / 255;
        b := aV / 255;
    end
    else
    begin
        h := anH / 255;
        s := anS / 255;
        v := aV / 255;

        varH := h * 6;
        if varH >= 6 then varH := varH - 6;
        
        varI := Trunc(varH);

        var1 := v * ( 1 - s );
        var2 := v * ( 1 - s * ( varH - varI ));
        var3 := v * ( 1 - s * ( 1 - ( varH - varI )));

        case varI of
            0:  begin
                    r := v;
                    g := var3;
                    b := var1;
                end;
            1:  begin
                    r := var2;
                    g := v;
                    b := var1;
                end;
            2:  begin
                    r := var1;
                    g := v;
                    b := var3;
                end;
            3:  begin
                    r := var1;
                    g := var2;
                    b := v;
                end;
            4:  begin
                    r := var3;
                    g := var1;
                    b := v;
                end;
            else
            begin
                r := v;
                g := var1;
                b := var2;
            end;
        end;
    end;

    result := RGB( Round(r * 255), Round(g * 255), Round(b * 255));
end;

function TGenericColorDialog.RGB2BGR( const aColor: TColor ): TColor;
begin
    result :=
        ( aColor and $000000FF ) shl 16 +
        ( aColor and $0000FF00 ) +
        ( aColor and $00FF0000 ) shr 16;
end;

{$IFOPT R+}
 {$DEFINE IS_RANGE_CHECKING}
 {$R-}
{$ENDIF}
procedure TGenericColorDialog.RGB2HSV( const aColor: TColor; out H, S, V: Byte );
var r, g, b: integer; 
    minValue, maxValue, deltaValue: integer;
    deltaR, deltaG, deltaB: integer;
begin
    r := aColor and $000000FF;
    g := ( aColor and $0000FF00 ) shr 8;
    b := ( aColor and $00FF0000 ) shr 16;

    minValue := Min( r, Min( g, b ));
    maxValue := Max( r, Max( g, b ));
    deltaValue := maxValue - minValue;

    V := maxValue;
    if deltaValue = 0 then begin
        H := 0;
        S := 0;
    end
    else begin
        S := deltaValue * 255 div maxValue;
        deltaR := Round((( maxValue - r ) / 6 + deltaValue / 2 ) * 256 / deltaValue );
        deltaG := Round((( maxValue - g ) / 6 + deltaValue / 2 ) * 256 / deltaValue );
        deltaB := Round((( maxValue - b ) / 6 + deltaValue / 2 ) * 256 / deltaValue );

        if r = maxValue then
            H := deltaB - deltaG
        else if g = maxValue then
            H := 85 + deltaR - deltaB
        else // b = maxValue
            H := 170 + deltaG - deltaR;
    end;
end;
{$IFDEF IS_RANGE_CHECKING}
 {$R+}
{$ENDIF}

procedure TGenericColorDialog.HandleColorMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
begin
    if Y >= COLOR_IMAGE_HEIGHT div 2 then
        SetColorEx(fPreviousColor); 
end;

procedure TGenericColorDialog.HandleCustomColorMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
var index: integer;
begin
    index := (X div 16) * 16;
    if index > 16 then index := 16;
    index := index + (Y div 16);

    DrawCustomColorSelection( fSelectedCustomColorIndex, clBlack );

    if aButton = mbLeft then
    begin
        SetColorEx( fCustomColors[index] ); // Set Color, but do not affect previous color
    end;
    
    DrawCustomColorSelection( index, clWhite );

    fSelectedCustomColorIndex := index;
    fAssignToCustomButton.Enabled := TRUE;    
end;

procedure TGenericColorDialog.HandleAssignToCustomClick( aSender: TObject );
begin
    fCustomColors[fSelectedCustomColorIndex] := Color;
    UpdateCustomColor(fSelectedCustomColorIndex);
end;

procedure TGenericColorDialog.HandleHueMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
begin
    fIsHueSelected := true;

    SetHueForPosition( Y );
end;

procedure TGenericColorDialog.SetHueForPosition( Y: Integer );
var hue: Integer;
begin
    hue := 255 - y + HUE_BAR_OFFSET_Y;

    if hue < 0 then hue := 0;
    if hue > 255 then hue := 255;

    SetHue( hue );
    UpdateAllFields;
end;

procedure TGenericColorDialog.HandleHueMouseMove( aSender: TObject; aShift: TShiftState; X, Y: Integer);
begin
    if fIsHueSelected then begin
        SetHueForPosition( Y );
    end;
end;

procedure TGenericColorDialog.HandleHueMouseUp( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
begin
    fIsHueSelected := false;
end;

procedure TGenericColorDialog.HandleSaturationValueMouseDown( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
begin
    fIsSaturationValueSelected := true;

    SetSaturationValueForPosition( X, Y );
end;

procedure TGenericColorDialog.HandleSaturationValueMouseMove( aSender: TObject; aShift: TShiftState; X, Y: Integer);
begin
    if fIsSaturationValueSelected then begin
        SetSaturationValueForPosition( X, Y );
    end;
end;

procedure TGenericColorDialog.HandleSaturationValueMouseUp( aSender: TObject; aButton: TMouseButton; aShift: TShiftState; X, Y: Integer);
begin
    fIsSaturationValueSelected := FALSE;
end;

function TGenericColorDialog.GetForm: TForm;
begin
    if not Assigned( fForm ) then begin
        fForm := TForm.Create( nil );

        InitializeForm( fForm );
    end;

    result := fForm;
end;

procedure TGenericColorDialog.HandleRGBChange( aSender: TObject );
var hue, saturation, value: Byte;
begin
    if not fIsUpdating then begin
        RGB2HSV( Color, hue, saturation, value );

        SetHue( hue );
        SetSaturationValue( saturation, value );
        UpdateHSVFields;
        UpdateColorImage;
    end;
end;

procedure TGenericColorDialog.HandleHSVChange( aSender: TObject );
var hue, saturation, value: integer;
begin
    if not fIsUpdating then begin
        hue := StrToIntDef( fHEdit.Text, 0 ) * 255 div 360;
        if hue < 0 then hue := 0;
        if hue > 255 then hue := 255;

        saturation := StrToIntDef( fSEdit.Text, 0 ) * 255 div 100;
        if saturation < 0 then saturation := 0;
        if saturation > 255 then saturation := 255;

        value := StrToIntDef( fVEdit.Text, 0 ) * 255 div 100;
        if value < 0 then value := 0;
        if value > 255 then value := 255;

        SetHue( hue );
        SetSaturationValue( saturation, value );
        UpdateRGBFields;
        UpdateColorImage;
    end;
end;

procedure TGenericColorDialog.HandleIntegerKeyPress( aSender: TObject; var aKey: Char );
begin
    if not ( aKey in ['0'..'9', Chr(VK_BACK)] ) then begin
        aKey := #0;
    end;
end;

procedure TGenericColorDialog.HandleNonZeroIntegerKeyPress( aSender: TObject; var aKey: Char );
var edit: TEdit;
begin
    if not ( aKey in ['0'..'9', Chr(VK_BACK)] ) then begin
        aKey := #0;
    end;

    if ( aSender is TEdit ) then begin
        edit := aSender as TEdit;

        if (edit.SelStart = 0) and (aKey = '0') then
            aKey := #0;
    end;
end;

function TGenericColorDialog.GetLeftArrow: TBitmap;
begin
    if not Assigned( fLeftArrow ) then
    begin
        fLeftArrow := TBitmap.Create;
        fLeftArrow.Width := 6;
        fLeftArrow.Height := 11;
        fLeftArrow.Canvas.Brush.Color := clBtnFace;
        fLeftArrow.Canvas.FillRect( Rect( 0, 0, 6, 11 ));

        fLeftArrow.Canvas.Pen.Color := clBlack;
        fLeftArrow.Canvas.MoveTo( 0, 0 );
        fLeftArrow.Canvas.LineTo( 5, 5 );
        fLeftArrow.Canvas.LineTo( 0, 10 );
        fLeftArrow.Canvas.LineTo( 0, 0 );
    end;

    result := fLeftArrow;
end;

function TGenericColorDialog.GetRightArrow: TBitmap;
begin
    if not Assigned( fRightArrow ) then
    begin
        fRightArrow := TBitmap.Create;
        fRightArrow.Width := 6;
        fRightArrow.Height := 11;
        fRightArrow.Canvas.Brush.Color := clBtnFace;
        fRightArrow.Canvas.FillRect( Rect( 0, 0, 6, 11 ));

        fRightArrow.Canvas.Pen.Color := clBlack;
        fRightArrow.Canvas.MoveTo( 5, 0 );
        fRightArrow.Canvas.LineTo( 0, 5 );
        fRightArrow.Canvas.LineTo( 5, 10 );
        fRightArrow.Canvas.LineTo( 5, 0 );
    end;

    result := fRightArrow;
end;

function TGenericColorDialog.GetColor: TColor;
// R, G, B colors are slightly more precise than HSV conversion,
// so we read them directly
begin
    result := ( b shl 16 ) + ( g shl 8 ) + r;

    if fSupportsAlpha then
    begin
        result := result + ( a shl 24 );
    end;
end;

procedure TGenericColorDialog.SetColor( const aColor: TColor );
begin
    fPreviousColor := aColor;

    if Assigned( fForm ) then UpdatePreviousColorImage;

    SetColorEx( aColor );
end;

procedure TGenericColorDialog.SetColorEx( const aColor: TColor );
var hue, saturation, value: byte;
begin
    fAlpha := (aColor and $FF000000) shr 24;

    RGB2HSV( aColor, hue, saturation, value );

    if Assigned(fForm) then
    begin
        fIsUpdating := TRUE;

        fREdit.Text := IntToStr(( aColor and $000000FF ));
        fGEdit.Text := IntToStr(( aColor and $0000FF00 ) shr 8 );
        fBEdit.Text := IntToStr(( aColor and $00FF0000 ) shr 16 );
        if fSupportsAlpha then
        begin
            fAEdit.Text := IntToStr( fAlpha );
        end;

        fIsUpdating := FALSE;

        SetHue( hue );
        SetSaturationValue( saturation, value );

        UpdateHSVFields;
        UpdateColorImage;
    end
    else
    begin                                                         
        fHue := hue;
        fSaturation := saturation;
        fValue := value;
    end;
end;


function TGenericColorDialog.GetCustomColor( anIndex: integer ): TColor;
begin
    result := fCustomColors[anIndex];
end;

procedure TGenericColorDialog.SetCustomColor( anIndex: integer; aColor: TColor );
begin
    fCustomColors[anIndex] := aColor;
end;

function TGenericColorDialog.GetR: Byte;
var r: Integer;
begin
    r := StrToIntDef( fREdit.Text, 0 );
    if r < 0 then r := 0;
    if r > 255 then r := 255;

    result := r;
end;

function TGenericColorDialog.GetG: Byte;
var g: integer;
begin
    g := StrToIntDef( fGEdit.Text, 0 );
    if g < 0 then g := 0;
    if g > 255 then g := 255;

    result := g;
end;

function TGenericColorDialog.GetB: Byte;
var b: integer;
begin
    b := StrToIntDef( fBEdit.Text, 0 );
    if b < 0 then b := 0;
    if b > 255 then b := 255;

    result := b;
end;

function TGenericColorDialog.GetA: Byte;
var default: integer;
    a: integer;
begin
    if fAllowsZeroAlpha then
        default := 0
    else
        default := 255;

    a := StrToIntDef( fAEdit.Text, default );
    if a < 0 then a := 0;
    if a > 255 then a := 255;

    result := a;
end;

function TGenericColorDialog.Execute: boolean;
begin
    result := ( form.ShowModal = mrOK );

    if result then StoreCustomColors;
end;

constructor TGenericColorDialog.Create( const aSupportsAlpha: boolean );
begin
    Inherited Create;

    fSupportsAlpha := aSupportsAlpha;

    fHue := 128;
    fSaturation := 128;
    fValue := 128;
    fAlpha := 255;

    fPreviousColor := TColor($FFFFFFFF);
    fSelectedCustomColorIndex := -1;
end;

destructor TGenericColorDialog.Destroy;
begin
    FreeAndNil( fForm );
    FreeAndNil( fLeftArrow );
    FreeAndNil( fRightArrow );
    FreeAndNil( fMarkerBuffer );

    inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure TSimpleFrame.Paint;
begin
    Canvas.Brush.Color := Color;
    Canvas.FrameRect( ClientRect );
end;


end.
