{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Generics;

{$MODE Delphi}

 // generic datastructure for polygons, each having their own set of vertices

interface

uses Classes,Math,Contnrs,Dialogs,SysUtils,Types,Graphics,Windows,
     frmProgressForm,
     unit_Enums,
     unit_Types,
     unit_IntegerList,
     unit_HashedList,
     unit_Textures,
     unit_Materials;

type
  TGenericVertex = class;
  TGenericPolygonList = class;

  TGenericEdge = class(TObject)
  private
    fStartVertex: TGenericVertex;         // pointers to vertices; not owned by generic edge itself
    fEndVertex: TGenericVertex;
  public
    constructor Create(sVertex,eVertex: TGenericVertex);
    function IsUsed(vertex: TGenericVertex): boolean;
    function Equals(genericEdge: TGenericEdge): boolean;
    property StartVertex: TGenericVertex read fStartVertex write fStartVertex;
    property EndVertex: TGenericVertex read fEndVertex write fEndVertex;
  end;

  TGenericEdgeList = class(TObjectList) // should be a hashed list
  private
    function GetGenericEdge(anIndex: integer): TGenericEdge;
  public
    constructor Create(owns: boolean);
    destructor Destroy; override;
    function EqualEdges(genericEdgeList: TGenericEdgeList): boolean;
    function InList(genericEdge: TGenericEdge): boolean;
    property Edges[anIndex: integer]: TGenericEdge read GetGenericEdge; default;
  end;

  TGenericVertex = class(TObject)    // similar to a TVertice, but holds U,V info as well
  private
      fX: Double;
      fY: Double;
      fZ: Double;
      fUVCoordinates: TUVList;  // Only holds UV coordinates for the number of _assigned_ layers (no dummy coordinates for non-assigned layers) 
  public
      constructor Create; overload;
      constructor Create( const anX, aY, aZ: Double ); overload;
      constructor Create( const aVertex: TVertex ); overload;
      destructor Destroy; override;
      function Copy: TGenericVertex;
      procedure Assign( const aVertex: TGenericVertex );
      function Equals( const aVertex: TGenericVertex; const anEpsilon: Double = 0.001 ): Boolean;
      function OnEdge( const aP1, aP2: TGenericVertex ): Boolean;
      property X: Double read fX;
      property Y: Double read fY;
      property Z: Double read fZ;
      property UVCoordinates: TUVList read fUVCoordinates;
  end;

  TGenericVertexList = class(TObjectList) // holds a list of TGenericVertex
  private
      function GetVertex(const anIndex: integer): TGenericVertex;
  public
      function Contains(const aVertex: TGenericVertex): Boolean;
      procedure Sort(L,H: integer);
      property Items[const anIndex: integer]: TGenericVertex read GetVertex; default;
  end;

  TGenericPolygon = class(TObject)   // holds a list of vertices, thus resembling a polygon
  private
    fVertices: TGenericVertexList;
    fPlane: TPlane;
    fMaterial: TMaterial;
    function GetPlane: TPlane;
    function GetNumVertices: integer;
//    function PointInTriangle(Q,P1,P2,P3: TVertice): boolean;
//    function TestTriangle(p1,p2,p3: TVertice): boolean;
  public
    constructor Create; overload;
    constructor Create(AOwnsVertices: boolean); overload;
    destructor Destroy; override;
    procedure Assign(genericPolygon: TGenericPolygon);
    function Copy: TGenericPolygon;
    function GetSharedEdges(genericPolygon: TGenericPolygon): TGenericEdgeList;
    function RelativePosition(genericPolygon: TGenericPolygon): TRelativePosition;
    function IsUsed(genericEdge: TGenericEdge): boolean; overload;
    function IsUsed(vertex: TVertice): boolean; overload;
    function IsEqualNormal(genericPolygon: TGenericPolygon): boolean;
    function IsCoplanar(genericPolygon: TGenericPolygon): boolean; overload;
    function IsCoplanar(vertex: TVertice): boolean; overload;
    function IsStrictlyConvex: boolean;
    function IsConvex: boolean;
    procedure PolygonToGenericPolygon(APolygon: TPolygon);
    procedure Invert;
//    procedure Triangulate(var triangleList: TGenericPolygonList);
    property Vertices: TGenericVertexList read fVertices write fVertices;
    property Plane: TPlane read GetPlane;
    property Material: TMaterial read fMaterial write fMaterial;
    property NumVertices: integer read GetNumVertices;
  end;

  TGenericPolygonList = class(TObjectList)   // holds a list of TGenericPolygons, thus resembling a polygon soup  (one 3D object)
  private
    fName: string;
    function SubdivideStringToList(str,delimiter: string): TStringList;
    function GetGenericPolygon(anIndex: integer): TGenericPolygon;
  public
    function AddObject(AObject: TObject): integer;
    procedure InvertPolygons;
    function FixTJunctions(progressForm: TfrmProgress): TGenericPolygonList;
    procedure AddList(list: TGenericPolygonList);
    function Copy: TGenericPolygonList;
    procedure WriteToFile(fileName: string);
    procedure ReadFromFile(fileName: string);
    property Polygons[anIndex: integer]: TGenericPolygon read GetGenericPolygon; default;
    property Name: string read fName write fName;
  end;

  TGenericPolygonListList = class(TObjectList)  // holds a list of TGenericPolygonListList, thus resembling a list of 3D objects
  private
    function GetGenericPolygonList(anIndex: integer): TGenericPolygonList;
  public
    property GenericPolygonList[anIndex: integer]: TGenericPolygonList read GetGenericPolygonList; default;
  end;

implementation

uses unit_Math,
     unit_Constants;

// ### TGenericEdge ############################################################

function TGenericEdge.IsUsed(vertex: TGenericVertex): boolean;
begin
    Result := (startVertex.Equals(vertex)) or (endVertex.Equals(vertex));
end;

function TGenericEdge.Equals(genericEdge: TGenericEdge): boolean;
begin
    Result := ((startVertex.Equals(genericEdge.startVertex)) and (endVertex.Equals(genericEdge.endVertex))) or
              ((startVertex.Equals(genericEdge.endVertex)) and (endVertex.Equals(genericEdge.startVertex)));
end;

constructor TGenericEdge.Create(sVertex,eVertex: TGenericVertex);
begin
    startVertex := sVertex;
    endVertex := eVertex;
end;

// ### TGenericEdgeList ########################################################

function TGenericEdgeList.GetGenericEdge(anIndex: integer): TGenericEdge;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TGenericEdge(Items[anIndex])
    else
        Result := nil;
end;

function TGenericEdgeList.EqualEdges(genericEdgeList: TGenericEdgeList): boolean;
var i,j: integer;
begin
    Result := true;
    for i := 0 to genericEdgeList.Count-1 do begin
        for j := 0 to Count-1 do begin
            Result := genericEdgeList.Edges[i].Equals(Edges[j]);
            if not Result then break
        end;
        if not Result then break;
    end;
end;

function TGenericEdgeList.InList(genericEdge: TGenericEdge): boolean;
var counter: integer;
begin
    Result := false;
    counter := 0;
    while (counter<Count-1) and (not Result) do begin
        if Edges[counter].Equals(genericEdge) then
            Result := true
        else
            inc(counter);
    end;
end;

constructor TGenericEdgeList.Create(owns: boolean);
begin
    inherited Create(owns); // JuMa: call to inherited constructor was missing
end;

destructor TGenericEdgeList.Destroy;
begin
    inherited;
end;

// ### TGenericVertex ##########################################################

constructor TGenericVertex.Create;
begin
  inherited Create;

  fUVCoordinates := TUVList.Create;
end;      

constructor TGenericVertex.Create( const aVertex: TVertex );
begin
  Create;

  fX := aVertex.X;
  fY := aVertex.Y;
  fZ := aVertex.Z;
end;

constructor TGenericVertex.Create( const anX, aY, aZ: Double ); 
begin
  Create;

  fX := anX;
  fY := aY;
  fZ := aZ;
end;

destructor TGenericVertex.Destroy;
begin
  FreeAndNil( fUVCoordinates );

  inherited Destroy;
end;

function TGenericVertex.Copy: TGenericVertex;
begin
    result := TGenericVertex.Create;
    result.Assign( self );
end;

procedure TGenericVertex.Assign( const aVertex: TGenericVertex );
var i: Integer;
begin
    fX := aVertex.X;
    fY := aVertex.Y;
    fZ := aVertex.Z;

    for i := 0 to aVertex.UVCoordinates.Count - 1 do
        UVCoordinates.Add( aVertex.UVCoordinates[i].Copy );
end;

function TGenericVertex.Equals( const aVertex: TGenericVertex; const anEpsilon: Double = 0.001 ): Boolean;
begin
  result :=
      ( Abs( fX - aVertex.X ) < anEpsilon ) and
      ( Abs( fY - aVertex.Y ) < anEpsilon ) and
      ( Abs( fZ - aVertex.Z ) < anEpsilon );
end;

// returns wether the vertex lies on the edge p1->p2
function TGenericVertex.OnEdge( const aP1, aP2: TGenericVertex ): Boolean;
const EPS=0.001;
var p1p0, p2p1: TVertex;
    lengthP2P1, dp: Double;
    distance: Double;

begin
    Result := false;

    p1p0 := VertexSubtract( VerticeToVertex( aP1 ), VerticeToVertex( self ));
    p2p1 := VertexSubtract( VerticeToVertex( aP2 ), VerticeToVertex( aP1 ));

    // Squared distance formula as suggested by MathWorld,
    // http://mathworld.wolfram.com/Point-LineDistance3-Dimensional.html 
    distance :=
        VertexSquaredLength( VertexCrossProduct( p2p1, p1p0 )) /
        VertexSquaredLength( p2p1 );

    if distance < Sqr(EPS) then begin
        // je moet nog steeds bepalen of Q ook echt tussen P1 en P2 in ligt
        // Dat is het geval als de de lengte van de projectie van de lijn p1 -> Q op de edge
        // kleiner is dan de lengte van de edge zelf, dus
        // t = ( Q - P1 ) * ( P2 - P1 ) / || P2 - P1 ||   <    || P2 - P1 || - EPS

        lengthP2P1 := VertexLength( p2p1 );

        dp := - VertexDotProduct( p1p0, p2p1 ) / lengthP2P1;

        if dp>EPS then
            Result := dp < lengthP2P1-EPS
        else
            Result := false;
    end;
end;

// ### TGenericVertexList ######################################################

function TGenericVertexList.GetVertex(const anIndex: integer): TGenericVertex;
begin
    result := inherited Items[anIndex] as TGenericVertex;
end;

function TGenericVertexList.Contains(const aVertex: TGenericVertex ): Boolean;
var i: integer;
    vertex: TGenericVertex;
begin
    result := FALSE;

    // Only compare x, y, and z... UV coordinates are allowed to be different.
    for i := 0 to Count - 1 do begin
        vertex := Items[i];
        if ( vertex.X = aVertex.X ) and ( vertex.Y = aVertex.Y ) and ( vertex.Z = aVertex.Z ) then begin
            result := TRUE;
            break;
        end;
    end;
end;

// PJ: Looks like some kind of bubble sort to me.
//     Autosort while building would then be a lot quicker...
procedure TGenericVertexList.Sort(L,H: integer);
var t0, t1: TVertex;
begin
    t0 := VertexSubtract( VerticeToVertex( Items[L] ), VerticeToVertex( Items[0] ));
    t1 := VertexSubtract( VerticeToVertex( Items[H] ), VerticeToVertex( Items[0] ));

    if VertexSquaredLength( t1 ) < VertexSquaredLength( t0 ) then
        Exchange(L,H);

    // TODO, JC: decide what to do when length is 0

    inc(H);
    if H=Count then begin
        inc(L);
        H := L+1;
    end;

    if (H<Count) then
        Sort(L,H);
end;

{
procedure TVerticeList.Sort(L,H: integer);
var t0,t1: TVertice;
begin
    t0 := vectorsub(vertices[L],vertices[0]);
    t1 := vectorsub(vertices[H],vertices[0]);

    if t1.length<t0.length then
        Exchange(L,H);

    // TODO, JC: decide what to do when length is 0

    inc(H);
    if H=Count then begin
        inc(L);
        H := L+1;
    end;

    if (H<Count) then
        Sort(L,H);

    FreeAndNil(t0);
    FreeAndNil(t1);
end;   }

// ### TGenericPolygon #########################################################

function TGenericPolygon.GetPlane: TPlane;
var averageYZ, averageZX, averageXY: double;
    index: integer;
    vertex, previousVertice: TGenericVertex;
    length: double;
begin
    averageYZ := 0;
    averageZX := 0;
    averageXY := 0;

    previousVertice := Vertices[NumVertices - 1];
    for index := 0 to NumVertices - 1 do begin
        vertex := Vertices[index];
        averageYZ := averageYZ + (previousVertice.Z + vertex.Z) * (vertex.Y - previousVertice.Y);
        averageZX := averageZX + (previousVertice.X + vertex.X) * (vertex.Z - previousVertice.Z);
        averageXY := averageXY + (previousVertice.Y + vertex.Y) * (vertex.X - previousVertice.X);
        previousVertice := vertex;
    end;

    length := Sqrt(averageYZ * averageYZ + averageZX * averageZX + averageXY * averageXY);

    if length < EPSILON then length := EPSILON;
    fPlane.Normal.X := averageYZ / length;
    fPlane.Normal.Y := averageZX / length;
    fPlane.Normal.Z := averageXY / length;

    vertex := Vertices[0];
    fPlane.Distance := - ( vertex.X * fPlane.Normal.X + vertex.Y * fPlane.Normal.Y + vertex.Z * fPlane.Normal.Z );

    result := fPlane;
end;

function TGenericPolygon.GetNumVertices: integer;
begin
    Result := Vertices.Count;
end;

procedure TGenericPolygon.Assign(genericPolygon: TGenericPolygon);
var i: integer;
begin
    if Assigned(genericPolygon) then begin
        Vertices.Clear;
        for i := 0 to genericPolygon.Vertices.Count-1 do
            Vertices.Add(genericPolygon.Vertices[i].Copy);

        //Vertices.Assign(genericPolygon.Vertices);
        Material := genericPolygon.Material;
    end;
end;

function TGenericPolygon.Copy: TGenericPolygon;
begin
    Result := TGenericPolygon.Create;
    Result.Assign(self);
end;

// name: GetSharedEdges
// desc: returns all edges of current polygon that also occur in the provided polygon
function TGenericPolygon.GetSharedEdges(genericPolygon: TGenericPolygon): TGenericEdgeList;
var i: integer;
    v0,v1: TGenericVertex;
    edge1: TGenericEdge;
begin
    Result := nil;

    if IsCoplanar(genericPolygon) then begin

        Result := TGenericEdgeList.Create(true);

        // build a list of edges that are shared by both polygons
        for i := 0 to Vertices.Count-1 do begin

            // get edge vertices and create an generic edge object with them
            v0 := Vertices[i];
            v1 := Vertices[(i+1) mod Vertices.Count];
            edge1 := TGenericEdge.Create(v0,v1);

            if genericPolygon.IsUsed(edge1) then
                Result.Add(edge1)
            else
                FreeAndNil(edge1);

        end;
    end;
end;

function TGenericPolygon.RelativePosition(genericPolygon: TGenericPolygon): TRelativePosition;
const EPS = 0.001;
var i: integer;
    distance: double;
    front, back: boolean;
    plane: TPlane;
begin
    // reset variables
    front := false;
    back := false;
    plane := genericPolygon.Plane;

    // loop through the vertices of this (!) polygon and see
    // if they are in front, behind or on the provided plane
    for i := 0 to Vertices.Count-1 do begin
        distance := DistToPlane(Vertices[i],plane);
        if distance < -EPS then
            back := true;
        if distance > EPS then
            front := true;
    end;

    // return a result
    Result := rpCoincident;
    if front and (not back) then
        Result := rpFront;
    if back and (not front) then
        Result := rpBack;
    if (Back) and (Front) then
        result := rpSpanning;
end;

// determines if the provided edge is used by the polygon
function TGenericPolygon.IsUsed(genericEdge: TGenericEdge): boolean;
var counter: integer;
    v0, v1: TGenericVertex;
    edge1: TGenericEdge;
begin
    // reset variables
    Result := false;
    counter := 0;

    // now loop through other polygon edges and see if an equal edge exists
    while (not Result) and (counter<Vertices.Count) do begin
        v0 := Vertices[counter];
        v1 := Vertices[(counter+1) mod Vertices.Count];
        edge1 := TGenericEdge.Create(v0,v1);
        if edge1.Equals(genericEdge) then
            Result := true
        else
            inc(counter);
        FreeAndNil(edge1);
    end;
end;

// determines if the provided vertex is used by the polygon
function TGenericPolygon.IsUsed(vertex: TVertice): boolean;
begin
    Result := false;
end;

// NOTE: page 297 Lengyel
function TGenericPolygon.IsEqualNormal(genericPolygon: TGenericPolygon): boolean;
const EPS = 0.001;   // EPS because of inaccuracies in floating point calculations
begin
    Result := DotProduct(Plane.Normal,genericPolygon.Plane.Normal) > -EPS;
{    Result := abs(Plane.Normal.x-genericPolygon.Plane.Normal.x) +
              abs(Plane.Normal.y-genericPolygon.Plane.Normal.y) +
              abs(Plane.Normal.z-genericPolygon.Plane.Normal.z) < EPS;}
end;
(*
function TGenericPolygon.IsEqualNormal(genericPolygon: TGenericPolygon): boolean;
const EPS = 0.001;   // EPS because of inaccuracies in floating point calculations
var dp: double;
begin
    dp := DotProduct(Plane.Normal,genericPolygon.Plane.Normal);
    Result := (dp > -EPS) and (dp < EPS);
end;*)

function TGenericPolygon.IsCoplanar(genericPolygon: TGenericPolygon): boolean;
const EPS = 0.001;
var i: integer;
    distance: double;
begin
    Result := true;
    for i := 0 to Vertices.Count-1 do begin
        distance := abs(DistToPlane(Vertices[i],genericPolygon.Plane));
        if distance>EPS then begin
            Result := false;
            break;
        end;
    end;
end;

function TGenericPolygon.IsCoplanar(vertex: TVertice): boolean;
const EPS = 0.001;
begin
    Result := abs(DistToPlane(vertex,Plane))<EPS;
end;

// if the polygon has colinear edges, it isn't convex (hence the term 'StriclyConvex')
function TGenericPolygon.IsStrictlyConvex: boolean;
var previousEdge, edge, previousCrossProduct, crossProduct: TVertex;
    i: integer;
begin
    result := true;

    if NumVertices > 3 then begin
        previousEdge := VertexSubtract(VerticeToVertex(Vertices[1]), VerticeToVertex(Vertices[0]));
        edge := VertexSubtract(VerticeToVertex(Vertices[2]), VerticeToVertex(Vertices[1]));
        previousCrossProduct := vertexCrossProduct(previousEdge, edge);
        previousEdge := edge;

        for i := 3 to NumVertices - 1 do begin
            edge := VertexSubtract(VerticeToVertex(Vertices[i]), VerticeToVertex(Vertices[i-1]));

            crossProduct := vertexCrossProduct(previousEdge, edge); // plane direction, or 'not normalized plane normal'
            if vertexDotProduct(crossProduct, previousCrossProduct) <= 0 then begin
                result := false;
                break;
            end;
 
            previousCrossProduct := crossProduct;
            previousEdge := edge;
        end;
    end;
end;

function TGenericPolygon.IsConvex: boolean;
var previousEdge, edge, previousCrossProduct, crossProduct: TVertex;
    i: integer;
begin
    result := true;

    if NumVertices > 3 then begin
        previousEdge := VertexSubtract(VerticeToVertex(Vertices[1]), VerticeToVertex(Vertices[0]));
        edge := VertexSubtract(VerticeToVertex(Vertices[2]), VerticeToVertex(Vertices[1]));
        previousCrossProduct := vertexCrossProduct(previousEdge, edge);
        previousEdge := edge;

        for i := 3 to NumVertices - 1 do begin
            edge := VertexSubtract(VerticeToVertex(Vertices[i]), VerticeToVertex(Vertices[i-1]));

            crossProduct := vertexCrossProduct(previousEdge, edge); // plane direction, or 'not normalized plane normal'
            if vertexDotProduct(crossProduct, previousCrossProduct) < 0 then begin      // the < checks for convex; <= would be strictly convex
                result := false;
                break;
            end;

            previousCrossProduct := crossProduct;
            previousEdge := edge;
        end;
    end;
end;

procedure TGenericPolygon.PolygonToGenericPolygon(APolygon: TPolygon);
var i, j: integer;
    vertex: TGenericVertex;
begin
    Material := APolygon.Material;
    for i := 0 to APolygon.NumVertices-1 do begin
        vertex := TGenericVertex.Create( VerticeToVertex( APolygon.VerticeList[i] ));
        Vertices.Add(vertex);

        // Copy UV Coordinates
        for j := 0 to Material.NumTextureLayers - 1 do begin
            if Assigned( Material.TextureLayers[j] ) then
                vertex.UVCoordinates.Add( APolygon.UVCoordinates[j][i].Copy );
        end;
    end;
end;

procedure TGenericPolygon.Invert;
var i: integer;
    verticeList: TGenericVertexList;
begin
    verticeList := TGenericVertexList.Create(true);
    for i := 0 to Vertices.Count-1 do
        verticeList.Add(Vertices[i].Copy);

    Vertices.Clear;

    for i := verticeList.Count-1 downto 0 do
        Vertices.Add(verticeList[i].Copy);

    FreeAndNil(verticeList);
end;

{
// tests if the points forming a triangle are winding in correct order in relation to the polygon-plane
function TGenericPolygon.TestTriangle(p1,p2,p3: TVertice): boolean;
const EPS=0.001;
var n,P3P2,P1P2: TVertice;
begin
    P3P2 := vectorSub(p3,p2);
    P3P2.Normalize;
    n := crossproduct(plane.Normal,P3P2);

    P1P2 := vectorSub(p1,p2);

    Result := dotproduct(n,P1P2)>EPS;

    FreeAndNil(P3P2);
    FreeAndNil(P1P2);
    FreeAndNil(n);
end;
}
{
// should be integrated in Triangulate function instead of being used as a function (speed-optimization)
function TGenericPolygon.PointInTriangle(Q,P1,P2,P3: TVertice): boolean;
const EPS = 0.001;
var P2P1,P3P2,P1P3,QP1,QP2,QP3,N1,N2,N3: TVertice;
begin
    P2P1 := vectorSub(P2,P1);
    P2P1.Normalize;
    P3P2 := vectorSub(P3,P2);
    P3P2.Normalize;
    P1P3 := vectorSub(P1,P3);
    P1P3.Normalize;

    N1 := crossproduct(plane.normal,P2P1);
    N2 := crossproduct(plane.normal,P3P2);
    N3 := crossproduct(plane.normal,P1P3);

    QP1 := vectorSub(Q,P1);
    QP1.Normalize;
    QP2 := vectorSub(Q,P2);
    QP2.Normalize;
    QP3 := vectorSub(Q,P3);
    QP3.Normalize;

    Result := (dotproduct(N1,QP1)>-EPS) and
              (dotproduct(N2,QP2)>-EPS) and
              (dotproduct(N3,QP3)>-EPS);

    FreeAndNil(P2P1);
    FreeAndNil(P3P2);
    FreeAndNil(P1P3);
    FreeAndNil(N1);
    FreeAndNil(N2);
    FreeAndNil(N3);
    FreeAndNil(QP1);
    FreeAndNil(QP2);
    FreeAndNil(QP3);
end;
}
{
procedure TGenericPolygon.Triangulate(var triangleList: TGenericPolygonList);   // earclipping triangulation
var v1,v2,v3: TVertice;
    index,tempIndex,counter: integer;
    insideTriangle,continueSearch: boolean;
    triangle: TGenericPolygon;
begin
    if NumVertices=3 then begin
        triangle := TGenericPolygon.Create;
        triangle.Material := Material;
        triangle.Vertices.Add(Vertices[0].Copy);
        triangle.Vertices.Add(Vertices[1].Copy);
        triangle.Vertices.Add(Vertices[2].Copy);
        triangleList.Add(triangle);
    end else begin

        continueSearch := true;

        while (NumVertices>3) and (continueSearch) do begin

            index := 0;
            continueSearch := true;

            while (continueSearch) do begin

                tempIndex := index-1;
                if tempIndex<0 then tempIndex := NumVertices-1;

                v1 := Vertices[tempIndex];
                v2 := Vertices[index];
                v3 := Vertices[(index+1) mod NumVertices];

                // check if the 2 edges make a right turn;
                if not TestTriangle(v3,v2,v1) then begin

                    index := (index+1) mod NumVertices;
                    if index=0 then continueSearch := false;                  // looped through all Vertices but didn't find an ear (wrong polyline; e.g edge cross)

                end else begin       // right turn, this might be an ear

                    // check if no other Vertices are inside the triangle formed by v1,v2,v3
                    counter := 0;
                    insideTriangle := false;
                    while (counter<NumVertices) and (not insideTriangle) do begin

                        if (not Vertices[counter].Equals(v1)) and
                           (not Vertices[counter].Equals(v2)) and
                           (not Vertices[counter].Equals(v3)) then begin

                            if not PointInTriangle(Vertices[counter],v3,v2,v1) then
                                inc(counter)
                            else
                                insideTriangle := true;
                        end else
                            inc(counter);
                    end;

                    // if no Vertices are inside the triangle, we can add this triangle to the result
                    // and remove the middle-vertex v2 from the current set of Vertices
                    if insideTriangle then begin

                        index := (index+1) mod NumVertices;                     // a vertex is inside the triangle, continue search for another ear
                        if index=0 then continueSearch := false;                // looped through all Vertices but didn't find an ear (wrong polyline; e.g edge cross)

                    end else begin
                        triangle := TGenericPolygon.Create;
                        triangle.Material := Material;
                        triangle.Vertices.Add(v1.Copy);
                        triangle.Vertices.Add(v2.Copy);
                        triangle.Vertices.Add(v3.Copy);
                        triangleList.Add(triangle);
                        Vertices.Remove(v2);
                        index := 0;
                    end;
                end;
            end;
        end;
    end;
end;
}
constructor TGenericPolygon.Create;
begin
    Create(TRUE);
end;

constructor TGenericPolygon.Create(AOwnsVertices: boolean);
begin
    inherited Create;
    
    fPlane := TPlane.Create;
    fMaterial := nil;
    fVertices := TGenericVertexList.Create(AOwnsVertices);
end;

destructor TGenericPolygon.Destroy;
begin
    FreeAndNil(fPlane);
    FreeAndNil(fVertices);
    inherited;
end;

// ### TGenericPolygonList #####################################################

function TGenericPolygonList.SubdivideStringToList(str,delimiter: string): TStringList;
var resultstr: string;
    strings: TStringList;
    c,l: integer;
begin
    strings := TStringList.Create;
    c := 1;
    l := length(str);
    while c <= l do begin                                                       // process entire string
        resultstr := '';
        while (c<=l) and (str[c]<>delimiter) do begin                           // read characters until delimiter is encountered
            resultstr := resultstr + str[c];
            inc(c);
        end;
        strings.Add(Trim(resultstr));                                           // trims strings
        inc(c);                                                                 // skip divider
        while (c<=l) and (str[c]=' ') do inc(c);                                // skip any blacks
    end;
    Result := strings;
end;

function TGenericPolygonList.GetGenericPolygon(anIndex: integer): TGenericPolygon;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TGenericPolygon(Items[anIndex])
    else
        Result := nil;
end;

function TGenericPolygonList.AddObject(AObject: TObject): integer;
var index: integer;
begin
    index := IndexOf(AObject);
    if index = -1 then
        index := Add(AObject);
    Result := index;
end;

procedure TGenericPolygonList.InvertPolygons;
var i: integer;
begin
    for i := 0 to Count-1 do
        Polygons[i].Invert;
end;

function TGenericPolygonList.FixTJunctions(progressForm: TfrmProgress): TGenericPolygonList;
var i,j,k,l: integer;
    genericPolygon,testPolygon,newPolygon: TGenericPolygon;
    v0,v1,testVertex: TGenericVertex;
    edgeVertices: TGenericVertexList;
begin
    Result := TGenericPolygonList.Create(true);

    edgeVertices := TGenericVertexList.Create(false);

    progressForm.Title := 'Fixing T-junctions';
    progressForm.SubTitle := 'Analyzing Geometry and Fixing T-junctions...';
    progressForm.Description := 'Object: result from CSG operation';
    progressForm.Max := Count;

    // loop through all polygons
    for i := 0 to Count-1 do begin

        progressForm.Position := i;

        genericPolygon := Polygons[i];

        newPolygon := TGenericPolygon.Create;
        newPolygon.Material := genericPolygon.Material;

        // loop through all edges of current polygon (which should be in clockwise order)
        for j := 0 to genericPolygon.Vertices.Count-1 do begin

            // get edge vertices
            v0 := genericPolygon.Vertices[j];
            v1 := genericPolygon.Vertices[(j+1) mod genericPolygon.Vertices.Count];

            // add first vertex of this edge to the to be formed edge of the new polygon
            edgeVertices.Add(v0.Copy);

            // test all vertices of all polygons against this edge, adding vertices to the edge when needed
            for k := 0 to Count-1 do begin

                if k<>i then begin

                    testPolygon := Polygons[k];

                    for l := 0 to testPolygon.Vertices.Count-1 do begin
                        testVertex := testPolygon.Vertices[l];
                        if testVertex.OnEdge(v0,v1) and not edgeVertices.Contains(testVertex) then          // the vertex is (very) near or on the edge
                            edgeVertices.Add(testVertex.Copy);
                    end;
                end;
            end;

            // order the list of vertices based on distance of each vertex
            // to the first vertex in the list so vertices end up colinear
            if edgeVertices.Count>2 then
                edgeVertices.Sort(1,2);

            // now add them to the new polygon
            for k := 0 to edgeVertices.Count-1 do
                newPolygon.Vertices.Add(edgeVertices[k]);
            //newPolygon.Vertices.AddList(edgeVertices);

            // clear the edgelist
            edgeVertices.Clear;

        end;

        Result.Add(newPolygon);
    end;

    FreeAndNil(edgeVertices);
end;

// TODO, JC: make sure this function does not have to copy the contents of the provided list all the time
procedure TGenericPolygonList.AddList(list: TGenericPolygonList);
var i: integer;
begin
    for i := 0 to list.Count-1 do
        Add(list.Polygons[i].Copy);
end;

function TGenericPolygonList.Copy: TGenericPolygonList;
begin
    Result := TGenericPolygonList.Create(true);
    Result.AddList(self);
end;

// writes the polygon info to a textfile
procedure TGenericPolygonList.WriteToFile(fileName: string);
var f: TextFile;
    i,j: integer;
    genericPolygon: TGenericPolygon;
    vertex: TGenericVertex;
begin
    AssignFile(f, fileName);
    Rewrite(f);

    for i := 0 to Count-1 do begin
        genericPolygon := Polygons[i];

        write(f,inttostr(genericPolygon.Vertices.Count)+';');

        for j := 0 to genericPolygon.Vertices.Count-1 do begin
            vertex := genericPolygon.Vertices[j];
            write(F,floattostrf(round(vertex.x),ffFixed,4,0)+','+
                    floattostrf(round(vertex.y),ffFixed,4,0)+','+
                    floattostrf(round(vertex.z),ffFixed,4,0)+';');
        end;

        writeln(f);
    end;

    CloseFile(f);
end;

// reads a "bucket of polygons" from a textfile
procedure TGenericPolygonList.ReadFromFile(fileName: string);
var stringList,subItems,verticeString: TStringList;
    i,j: integer;
    genericPolygon: TGenericPolygon;
    vertex: TGenericVertex;
begin
    stringList := TStringList.Create;
    stringList.LoadFromFile(fileName);

    for i := 0 to stringList.Count-1 do begin           // no rangechecking, I/O checking, whatever checking at all
        genericPolygon := TGenericPolygon.Create;

        subItems := SubdivideStringToList(stringList[i],';');

        for j := 1 to strtoint(subItems[0]) do begin
            verticeString := SubdivideStringToList(subItems[j],',');
            vertex := TGenericVertex.Create(
              StrToFloat( verticeString[0] ),
              StrToFloat( verticeString[1] ),
              StrToFloat( verticeString[2] ));
            genericPolygon.Vertices.Add( vertex );
            FreeAndNil(verticeString);
        end;

        FreeAndNil(subItems);

        Add(genericPolygon);
    end;

    FreeAndNil(stringList);
end;

// ### TGenericPolygonListList #################################################

function TGenericPolygonListList.GetGenericPolygonList(anIndex: integer): TGenericPolygonList;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TGenericPolygonList(Items[anIndex])
    else
        Result := nil;
end;


end.
