{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_HashedList;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  TObjectList descendant that uses hashing to perform quick lookups.

  Most methods of TList are not

  There are two ways of using this class:
  1. Create a descendant, overriding both IsEqual and GetHashKey
  2. Assign your own methods to the OnGetHashKey and OnIsEqual events

  Author: Paul-Jan Pauptit
  Date:   01-10-2004
  ---------------------------------------------------------------------------- }

interface

uses
  Classes,
  Contnrs;

const
  HASH_TABLE_DEFAULT_SIZE = 2047;

type
  THOLGetHashKeyEvent = function( const anObject: TObject ): Integer of object;
  THOLIsEqualEvent = function( const anObject1, anObject2: TObject ): Boolean of object;

 // ----------------------------------------------------------------------------

  PHashTableEntry = ^THashTableEntry;
  THashTableEntry = record
      Index: Integer;
      Objct: TObject;
      NextEntry: PHashTableEntry;
      PrevEntry: PHashTableEntry;
  end;

// ----------------------------------------------------------------------------

  THashedObjectList = class( TObjectList )
  private
      fHashTableSize: Cardinal;                                                 
      fEntries: Array of PHashTableEntry;

      fIsHashingEnabled: Boolean;                                               

      fOnGetHashKey: THOLGetHashKeyEvent;
      fOnIsEqual: THOLIsEqualEvent;

  protected
      procedure Notify( aPointer: Pointer; anAction: TListNotification ); override;

      function GetHashKey( const anObject: TObject ): Integer; virtual;        
      function IsEqual( const anObject1, anObject2: TObject ): Boolean; virtual;

      function EntryForIndexAndObject( const anIndex: Integer; const anObject: TObject ): PHashTableEntry;
      procedure AddEntryForObject( const anObject: TObject );
      procedure RemoveEntryForObject( const anObject: TObject );

      procedure RemoveAllEntries;

      property HashTableSize: Cardinal read fHashTableSize;                     
      property IsHashingEnabled: Boolean read fIsHashingEnabled;

  public
      constructor Create( const aSize: Integer = HASH_TABLE_DEFAULT_SIZE );
      destructor Destroy; override;

      function IndexOf( anObject: TObject ): Integer;
      procedure Clear; Override;

      property OnGetHashKey: THOLGetHashKeyEvent read fOnGetHashKey write fOnGetHashKey;
      property OnIsEqual: THOLIsEqualEvent read fOnIsEqual write fOnIsEqual;
  end;


// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
    SysUtils;           

constructor THashedObjectList.Create( const aSize: Integer = HASH_TABLE_DEFAULT_SIZE );
begin
  Inherited Create;

  fIsHashingEnabled := true;                                                    

  fHashTableSize := aSize;
  SetLength( fEntries, aSize );
end;

// ----------------------------------------------------------------------------

destructor THashedObjectList.Destroy;
begin
  fIsHashingEnabled := false;                                                   

  RemoveAllEntries;

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure THashedObjectList.RemoveAllEntries;
var
  index: Integer;
  entry, tempEntry: PHashTableEntry;
begin
  for index := Low( fEntries ) to High( fEntries ) do
  begin
      entry := fEntries[index];

      while Assigned( entry ) do
      begin
          tempEntry := entry.NextEntry;

          Dispose( entry );

          entry := TempEntry;
      end;

      fEntries[index] := nil;                                                   
  end;
end;

// ----------------------------------------------------------------------------

procedure THashedObjectList.Notify( aPointer: Pointer; anAction: TListNotification );
// INFO, PJ: Is this the best way to extend a TList? Ugh!
begin
  if IsHashingEnabled then                                                      
  begin
      case anAction of
          lnAdded:
          begin
              AddEntryForObject( TObject( aPointer ));
          end;

          lnDeleted, lnExtracted:
          begin
              RemoveEntryForObject( TObject( aPointer ));
          end;
      end;
  end;

  Inherited Notify( aPointer, anAction );
end;

// ----------------------------------------------------------------------------

function THashedObjectList.EntryForIndexAndObject( const anIndex: Integer; const anObject: TObject ): PHashTableEntry;
var
  entry: PHashTableEntry;

begin
  result := nil;

  entry := fEntries[anIndex];

  while Assigned( entry ) do
  begin
      try
          if IsEqual( anObject, entry.Objct ) then
          begin
              result := entry;
              break;
          end;
      except
        result := nil
      end;

      entry := entry.NextEntry;
  end;
end;

// ----------------------------------------------------------------------------

procedure THashedObjectList.AddEntryForObject( const anObject: TObject );
var
  entry: PHashTableEntry;

begin
  New( entry );

  // INFO, PJ: We enter the object at the start of the objectChain at index 'hashkey'
  entry.Index := Count - 1;
  entry.Objct := anObject;
  entry.PrevEntry := nil;
  entry.NextEntry := fEntries[GetHashKey( anObject )];

  if Assigned( entry.NextEntry ) then
  begin
      entry.NextEntry.PrevEntry := entry;
  end;

  fEntries[GetHashKey( anObject )] := entry;
end;

// ----------------------------------------------------------------------------

procedure THashedObjectList.RemoveEntryForObject( const anObject: TObject );
var
  hashKey: Integer;
  entry: PHashTableEntry;

begin
  hashKey := GetHashKey( anObject );
  entry := EntryForIndexAndObject( hashKey, anObject );

  if not Assigned( entry ) then begin
      raise Exception.Create( 'No hash entry found for object.' );
  end;

  if Assigned( entry.NextEntry ) then begin
      entry.NextEntry.PrevEntry := entry.PrevEntry;
  end;

  if Assigned( entry.PrevEntry ) then begin
      entry.PrevEntry.NextEntry := entry.NextEntry;
  end
  else
      fEntries[hashKey] := entry.nextEntry;

  Dispose( entry );
end;

// ----------------------------------------------------------------------------

function THashedObjectList.IndexOf( anObject: TObject ): Integer;
var
  hashKey: Integer;
  entry: PHashTableEntry;

begin
  hashKey := GetHashKey( anObject );
  entry := EntryForIndexAndObject( hashKey, anObject );

  if Assigned( entry ) then
  begin
      result := entry.Index;
  end
  else
      result := -1;
end;

// ----------------------------------------------------------------------------

procedure THashedObjectList.Clear;
begin
  fIsHashingEnabled := false;
  try
      RemoveAllEntries;
      inherited Clear;
  finally
      fIsHashingEnabled := true;
  end;
end;

// ----------------------------------------------------------------------------

function THashedObjectList.GetHashKey( const anObject: TObject ): Integer;
begin
  if Assigned( fOnGetHashKey ) then
  begin
      result := fOnGetHashKey( anObject );
  end
  else
      result := Cardinal( anObject ) mod HashTableSize;                         
end;

// ----------------------------------------------------------------------------

function THashedObjectList.IsEqual( const anObject1, anObject2: TObject ): Boolean;
begin
  if Assigned( fOnIsEqual ) then
  begin
      result := fOnIsEqual( anObject1, anObject2 );
  end
  else
      result := ( anObject1 = anObject2 );
end;

end.
