{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Interpolation;

{$MODE Delphi}

interface

uses
  unit_Vectors;
  
procedure LinearInterpolation(p0,p1: TVector3f; t: single; var v: TVector3f);
procedure CatmullRomInterpolation(p0,p1,p2,p3: TVector3f; t: single; var v: TVector3f);
procedure CubicBezierInterpolation(p0,p1,p2,p3: TVector3f; t: single; var v: TVector3f);
procedure HermiteInterpolation(p0,p1,p2,p3: TVector3f; t: single; var v: TVector3f);

implementation

procedure LinearInterpolation(p0,p1: TVector3f; t: single; var v: TVector3f);
begin
{
  p0 and p1 are the end points
}
  v.x := p0.x + (p1.x - p0.x) * t;
  v.y := p0.y + (p1.y - p0.y) * t;
  v.z := p0.z + (p1.z - p0.z) * t;
end;

procedure CatmullRomInterpolation(p0,p1,p2,p3: TVector3f; t: single; var v: TVector3f);
{
  Catmull-Rom spline curve interpolation
  the two end points are p1 and p2
  p0 is the point prior to p1.
  p3 is the point after p2.
}
var
  tt : single;
  ttt: single;
  b0 : single;
  b1 : single;
  b2 : single;
  b3 : single;
begin
  tt   := t  * t;
  ttt  := tt * t;

  b0   := 0.5 * (    -ttt + 2 * tt - t    );
  b1   := 0.5 * ( 3 * ttt - 5 * tt     + 2);
  b2   := 0.5 * (-3 * ttt + 4 * tt + t    );
  b3   := 0.5 * (     ttt -     tt        );

  v.x := b0 * p0.x + b1 * p1.x + b2 * p2.x + b3 * p3.x;
  v.y := b0 * p0.y + b1 * p1.y + b2 * p2.y + b3 * p3.y;
  v.z := b0 * p0.z + b1 * p1.z + b2 * p2.z + b3 * p3.z;
end;

procedure HermiteInterpolation(p0,p1,p2,p3: TVector3f; t: single; var v: TVector3f);
{
  Hermite spline curve interpolation
  the two end points are p0 and p1
  p2 and p3 are the tension? points
}
var
  tt : single;
  ttt: single;
  b0 : single;
  b1 : single;
  b2 : single;
  b3 : single;
begin
  tt   := t  * t;
  ttt  := tt * t;

{
  b0   :=  2 * ttt - 3 * tt + 0 * t + 1 * 1;
  b1   := -2 * ttt + 3 * tt + 0 * t + 0 * 1;
  b2   :=  1 * ttt - 2 * tt + 1 * t + 0 * 1;
  b3   :=  1 * ttt - 1 * tt + 0 * t + 0 * 1;
}

  b0   :=  2 * ttt - 3 * tt +         1;  //p0 weight
  b1   := -2 * ttt + 3 * tt            ;  //p1 weight
  b2   :=      ttt - 2 * tt +     t    ;  //t0 weight
  b3   :=      ttt      -tt            ;  //t1 weight

  v.x := b0 * p0.x + b1 * p1.x + b2 * p2.x + b3 * p3.x;
  v.y := b0 * p0.y + b1 * p1.y + b2 * p2.y + b3 * p3.y;
  v.z := b0 * p0.z + b1 * p1.z + b2 * p2.z + b3 * p3.z;
end;

procedure CubicBezierInterpolation(p0,p1,p2,p3: TVector3f; t: single; var v: TVector3f);
{
   Piecewise cubic bezier curve as defined by Adobe in Postscript
   The two end points are p0 and p3
   Their associated control (or tangent) points are p1 and p2
}
var
  tt : single;
  ttt: single;
  pa : TVector3f;
  pb : TVector3f;
  pc : TVector3f;
begin
  tt   := t * t;
  ttt  := tt * t;

  pc.x := 3 * (p1.x - p0.x);
  pc.y := 3 * (p1.y - p0.y);
  pc.z := 3 * (p1.z - p0.z);
  pb.x := 3 * (p2.x - p1.x) - pc.x;
  pb.y := 3 * (p2.y - p1.y) - pc.y;
  pb.z := 3 * (p2.z - p1.z) - pc.z;
  pa.x := p3.x - p0.x - pc.x - pb.x;
  pa.y := p3.y - p0.y - pc.y - pb.y;
  pa.z := p3.z - p0.z - pc.z - pb.z;

  v.x := pa.x * ttt + pb.x * tt + pc.x * t + p0.x;
  v.y := pa.y * ttt + pb.y * tt + pc.y * t + p0.y;
  v.z := pa.z * ttt + pb.z * tt + pc.z * t + p0.z;
end;

end.
