{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Joint;

{$MODE Delphi}

interface

uses
  ComCtrls,
  unit_Types,
  unit_Primitives,
  unit_JointLocation,
  unit_BoneList;

type
  TJointList = class;

  // Joints are added to the scene using the TScene.AddJoint method.
  // This method determines if there is a joint selected by asking the
  // TJointSelector. If a joint is selected, the new joint is added to
  // the skeleton of the selected joint and it is connected to that same
  // selected joint. If no joint is selected, a new skeleton is created
  // using a TSkeletonBuilder object. Then a new joint is created and
  // added to that new skeleton.
  //
  // NOTES:
  //  - the number of childjoints and childbones is equal
  //  - AssignedVertices holds vertices of a primitive
  //  - each frame holds a JointLocation object for each Joint;
  //    a JointLocation object holds the rotation angles for that
  //    joint in that frame;

  TJoint = class(TBasePrimitive)
  private
    fID: integer;
    fName: string;
    fParentJoint: TJoint;
    fChildJoints: TJointList;
    fChildBones: TBoneList;
    fAssignedVertices: TVerticeList;
    fSkeleton: TObject;
    fTreeNode: TTreeNode;
    fIsDirty: Boolean;        // Rendering state, remove some day.
    fOriginalVertices: TVerticeList;
    fOriginalCenter: TVertice;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure Assign(AJoint: TJoint); reintroduce; overload;
    procedure SetupOriginalVertices(ACenter: TVertice);
    procedure Invalidate;
    property ID: integer read fID write fID;
    property Name: string read fName write fName;
    property ParentJoint: TJoint read fParentJoint write fParentJoint;
    property ChildJoints: TJointList read fChildJoints write fChildJoints;
    property ChildBones: TBoneList read fChildBones write fChildBones;
    property AssignedVertices: TVerticeList read fAssignedVertices write fAssignedVertices;
    property OriginalVertices: TVerticeList read fOriginalVertices write fOriginalVertices;
    property OriginalCenter: TVertice read fOriginalCenter write fOriginalCenter;
    property Skeleton: TObject read fSkeleton write fSkeleton;
    property TreeNode: TTreeNode read fTreeNode write fTreeNode;
    property IsDirty: Boolean read fIsDirty write fIsDirty;
  end;

  TJointList = class(TDeleDObjectList)
  private
    fJointIDValue: integer;
    function GetJoint(anIndex: integer): TJoint;
    function GetJointIDValue: integer;
    procedure AddToList(AJoint: TJoint);
    property JointIDValue: integer read GetJointIDValue write fJointIDValue;
  public
    constructor Create(AOwnsObjects, AUniquesOnly: boolean);
    procedure Assign(AJointList: TJointList);
    procedure AddJoint(AJoint: TJoint);
    procedure RemoveJointByID(AID: integer);
    procedure DeleteJointByID(AID: integer);
    function GetJointByID(AID: integer): TJoint;
    property Joints[anIndex: integer]: TJoint read GetJoint; default;
  end;

implementation

uses
  SysUtils,
  unit_Bone,
  unit_BoneBuilder,
  unit_Enums;

// ### TJoint ##################################################################

constructor TJoint.Create;
begin
    inherited Create;
    fChildJoints := TJointList.Create(false,true);
    fChildBones := TBoneList.Create(true,true);                                 // a joint owns its child bones
    fAssignedVertices := TVerticeList.Create(false,true);                       // references to vertices of primitives
    fOriginalVertices := TVerticeList.Create(true,true);                        // original vertices of the joint geometry
    fOriginalCenter := TVertice.Create(0,0,0);
    SnapMethod := smNone;
    fName := 'joint';
end;

destructor TJoint.Destroy;
begin
    FreeAndNil(fChildJoints);
    FreeAndNil(fChildBones);
    FreeAndNil(fAssignedVertices);
    FreeAndNil(fOriginalVertices);
    FreeAndNil(fOriginalCenter);
    inherited Destroy;
end;

procedure TJoint.Assign(AJoint: TJoint);
var i: integer;
    childJoint: TJoint;
    childBone: TBone;
begin
    fID := AJoint.ID;
    fName := AJoint.Name;
    for i := 0 to AJoint.ChildJoints.Count-1 do begin
        childJoint := TJoint.Create;
        childJoint.Assign(AJoint.ChildJoints[i]);
        childJoint.ParentJoint := self;
        ChildJoints.Add(childJoint);

        // the # of joints equals the # of bones
        childBone := TBone.Create;
        childBone.Assign(AJoint.ChildBones[i]);
        ChildBones.Add(childBone);
    end;
    inherited Assign(TBasePrimitive(AJoint));
end;

procedure TJoint.SetupOriginalVertices(ACenter: TVertice);
var i: integer;
begin
    fOriginalCenter.Assign(ACenter);
    fOriginalVertices.Clear;
    for i := 0 to Vertices.Count-1 do
        fOriginalVertices.Add(Vertices[i].Copy);
end;

procedure TJoint.Invalidate;
var i: integer;
begin
    IsDirty := true;

    for i := 0 to ChildBones.Count-1 do
        ChildBones[i].IsDirty := true;
end;

// ### TJointList ##############################################################

constructor TJointList.Create(AOwnsObjects, AUniquesOnly: boolean);
begin
    inherited Create(AOwnsObjects, AUniquesOnly);

    fJointIDValue := -1;
end;

procedure TJointList.AddToList(AJoint: TJoint);
var i: integer;
    joint: TJoint;
begin
    Add(AJoint);
    for i := 0 to AJoint.ChildJoints.Count-1 do begin
        joint := AJoint.ChildJoints[i];
        AddToList(joint);
    end;
end;

procedure TJointList.Assign(AJointList: TJointList);
var joint: TJoint;
begin
    // joints are linked to eachother and form a tree structure;
    // the first joint in a jointlist is the root of this
    // tree; if we create a copy of this rootjoint, we get
    // copies of all childjoints automatically too because
    // of the tree structure in which they are stored
    // (the Joint.Assign method also copies childjoints);
    // by traversing the tree formed by the copied rootjoint,
    // we can fill the jointlist

    joint := TJoint.Create;
    joint.Assign(AJointList[0]);
    AddToList(joint);
end;

function TJointList.GetJoint(anIndex: integer): TJoint;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TJoint(Items[anIndex])
    else
        Result := nil;
end;

function TJointList.GetJointIDValue: integer;
var i: integer;
    joint: TJoint;
begin
    if fJointIDValue = -1 then begin
        for i := 0 to Count-1 do begin
            joint := GetJoint(i);
            if fJointIDValue<joint.ID then
                fJointIDValue := joint.ID;
        end;
    end;
    Result := fJointIDValue;
end;

procedure TJointList.AddJoint(AJoint: TJoint);
begin
    JointIDValue := JointIDValue + 1;
    AJoint.ID := JointIDValue;
    Add(AJoint);
end;

procedure TJointList.RemoveJointByID(AID: integer);
var joint: TJoint;
begin
    joint := GetJointByID(AID);
    if Assigned(joint) then
        Remove(joint);
end;

procedure TJointList.DeleteJointByID(AID: integer);
var joint: TJoint;
begin
    joint := GetJointByID(AID);
    if Assigned(joint) then
        Delete(IndexOf(joint));
end;

function TJointList.GetJointByID(AID: integer): TJoint;
var counter: integer;
    isFound: boolean;
    joint: TJoint;
begin
    counter := 0;
    isFound := false;
    joint := nil;
    while (counter<Count) and (not isFound) do begin
        joint := Joints[counter];
        if joint.ID = AID then
            isFound := true
        else
            inc(counter);
    end;
    Result := joint;
end;

end.
