{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_JointCreatorTool;

{$MODE Delphi}

interface

uses
  unit_SkeletonTool,
  unit_ParamList,
  unit_Joint,
  unit_Skeleton,
  unit_Enums,
  unit_SkeletonUndoItem,
  unit_Types,
  unit_AnimationFrame,
  unit_JointLocation,
  unit_MaterialManager;

type
  {TSkeletonUndoInsertItem = class(TSkeletonUndoItem)
  private
  public
    constructor Create;
    destructor Destroy; override;
  end;}

  TJointInserterFunction = class(TObject)
  private
    procedure AddJointToSkeleton(ASkeleton: TSkeleton; AJoint, AParentJoint: TJoint);
    function CreateJointLocationForFrame(AAnimationFrame: TAnimationFrame; AParentID, AChildID: integer; ADeltaVector: TVertice): TJointLocation;
    function InsertJoint(ASkeleton: TSkeleton; AParamList: TParamList): TJoint;
  public
    class function Execute(ASkeleton: TSkeleton; AParamList: TParamList): TJoint;
  end;

  TJointCreatorTool = class(TSkeletonTool)
  private
    fSkeleton: TSkeleton;           // the skeleton that is created or to which a joint is added
    fJoint, fParentJoint: TJoint;   // the actual joint that is created and its parentjoint (if any)
    fPosition: TVertice;            // the position at which the joint is placed
    fMaterialManager: TMaterialManager;     // the materialmanager being passed to the tool
  protected
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  public
    constructor Create(AToolAction: TToolAction);
    destructor Destroy; override;
    function CanExecute(AParamList: TParamList): boolean; override;
    procedure Undo; override;
    procedure Redo; override;
  end;

implementation

uses
  SysUtils,
  Dialogs,
  unit_Animation,
  unit_BaseTool,
  unit_SkeletonBuilder,
  unit_Math,
  unit_JointBuilder,
  unit_Bone,
  unit_BoneBuilder;

// ### TJointInserterFunction ##################################################

procedure TJointInserterFunction.AddJointToSkeleton(ASkeleton: TSkeleton; AJoint, AParentJoint: TJoint);
begin
    // add the joint to the skeleton; this provides the ID value of the joint
    ASkeleton.Joints.AddJoint(AJoint);
    AJoint.Skeleton := ASkeleton;
    if Assigned(AParentJoint) then begin
        // link it to the selected joint (this creates a tree structure)
        AParentJoint.ChildJoints.Add(AJoint);
        AJoint.ParentJoint := AParentJoint;
    end;
end;

function TJointInserterFunction.CreateJointLocationForFrame(AAnimationFrame: TAnimationFrame; AParentID, AChildID: integer; ADeltaVector: TVertice): TJointLocation;
var jointLocation, parentJointLocation: TJointLocation;
begin
    parentJointLocation := AAnimationFrame.GetJointLocationByID(AParentID);
    jointLocation := TJointLocation.Create;
    jointLocation.ID := AChildID;
    jointLocation.Position.Assign(parentJointLocation.Position);
    jointLocation.Position.Add(ADeltaVector);
    AAnimationFrame.AddJointLocation(jointLocation, parentJointLocation);
    Result := jointLocation;
end;

function TJointInserterFunction.InsertJoint(ASkeleton: TSkeleton; AParamList: TParamList): TJoint;
var selectedJoint, joint: TJoint;
    selectedJointLocation: TJointLocation;
    i,j: integer;
    jointBuilder: TJointBuilder;
    paramList: TParamList;
    boneBuilder: TBoneBuilder;
    bone: TBone;
    activeFrame, animationFrame: TAnimationFrame;
    deltaVector,newPosition: TVertice;
    animation: TAnimation;
begin
    selectedJoint := ASkeleton.SelectedJoints[0];

    AParamList.Add(ASkeleton.JointSize, 'jointSize');

    // create a new joint
    jointBuilder := TJointBuilder.Create;
    joint := jointBuilder.Build(AParamList);
    FreeAndNil(jointBuilder);

    // add the joint to the skeleton and link it to the selected joint, creating a tree structure
    AddJointToSkeleton(ASkeleton, joint, selectedJoint);

    // auto-select the new joint in the selected skeleton (make it the only one selected)
    ASkeleton.SelectedJoints.Clear;
    ASkeleton.SelectedJoints.Add(joint);

    // create new jointlocations for the baseframe and each frame of all animations of the skeleton
    // NOTE: the jointlocation has the same ID as the corresponding joint

    activeFrame := ASkeleton.GetActiveFrame;
    // get the location of the selected joint in the active frame (!) of the skeleton
    // the selected joint is the parent joint of the joint that is being inserted
    selectedJointLocation := activeFrame.GetJointLocationByID(selectedJoint.ID);

    // calculate a deltavector going from the location of the selected joint
    // (which is the parent joint) to the provided vertex (which is the new jointlocation)
    newPosition := TVertice(AParamList[0]);
    deltaVector := vectorSub(newPosition, selectedJointLocation.Position);

    // create a new jointlocation for the baseframe
    CreateJointLocationForFrame(ASkeleton.BaseFrame, selectedJointLocation.ID, joint.ID, deltaVector);

    // create new jointlocations for all frames for all animations
    for i := 0 to ASkeleton.Animations.Count-1 do begin
        animation := ASkeleton.Animations[i];
        for j := 0 to animation.AnimationFrames.Count-1 do begin
            animationFrame := animation.AnimationFrames[j];
            CreateJointLocationForFrame(animationFrame, selectedJointLocation.ID, joint.ID, deltaVector);
        end;
    end;
    FreeAndNil(deltaVector);

    // create a bone based on the selected joint and the new joint
    // and add it to the childbones collection of the selected joint
    paramList := TParamList.Create(false);
    paramList.Add(activeFrame.GetJointLocationByID(selectedJoint.ID), 'jointLocation1');
    paramList.Add(activeFrame.GetJointLocationByID(joint.ID), 'jointLocation2');
    paramList.Add(AParamList[1], 'materialManager');                            // we pass the MaterialManager
    paramList.Add(ASkeleton.JointSize, 'boneSize');                             // NOTE, JCC: bonesize = jointsize
    boneBuilder := TBoneBuilder.Create;
    bone := boneBuilder.Build(paramList);
    FreeAndNil(boneBuilder);
    FreeAndNil(paramList);
    selectedJoint.ChildBones.Add(bone);

    // return
    Result := joint;
end;

class function TJointInserterFunction.Execute(ASkeleton: TSkeleton; AParamList: TParamList): TJoint;
begin
    with self.Create do
    try
        Result := InsertJoint(ASkeleton, AParamList);
    finally
        Free;
    end;
end;

// ### TJointCreatorTool #######################################################

procedure TJointCreatorTool.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Add joint';
    Description := 'Adds a joint to an existing skeleton or creates a new skeleton by creating a new joint.';
end;

// if there is no current selected joint, a new skeleton is created to which
// this new joint is added; if there is a selected joint, this new joint is
// added to the current skeleton and linked to the selected joint via a bone;
// then, it becomes the newly selected joint
// NOTE: this means that there can be zero or one selected joint at most when
// adding new joints to the scene
procedure TJointCreatorTool.DoExecute(AParamList: TParamList);
var skeletonBuilder: TSkeletonBuilder;
begin
    fSkeleton := nil;
    fParentJoint := nil;
    fJoint := nil;
    // fPosition is owned by class itself
    if Assigned(fPosition) then
        fPosition.Assign(TVertice(AParamList[0]))
    else
        fPosition := TVertice(AParamList[0]).Copy;
    fMaterialManager := TMaterialManager(AParamList[1]);

    if SelectedSkeletons.Count = 0 then begin
        // now create a new skeleton
        skeletonBuilder := TSkeletonBuilder.Create;
        fSkeleton := skeletonBuilder.Build(AParamList);

        // add it to the Skeleton collection
        Skeletons.AddSkeleton(fSkeleton);

        // add the new skeleton to the selected skeleton collection
        SelectedSkeletons.Add(fSkeleton);

        // free the builder
        FreeAndNil(skeletonBuilder);
    end else if SelectedSkeletons.Count = 1 then begin
        fSkeleton := SelectedSkeletons[0];
        // this if-statement has been done in CanExecute too but it can't hurt to do it again
        if fSkeleton.SelectedJoints.Count = 1 then begin
            fParentJoint := fSkeleton.SelectedJoints[0];
            fJoint := TJointInserterFunction.Execute(fSkeleton,AParamList);
            fSkeleton.SelectedJoints.Add(fJoint);
        end;
    end;
end;

function TJointCreatorTool.CanExecute(AParamList: TParamList): boolean;
var canExecuteTool: boolean;
    selectedSkeleton: TSkeleton;
begin
    canExecuteTool := true;

    if SelectedSkeletons.Count = 1 then begin
         selectedSkeleton := SelectedSkeletons[0];
         if selectedSkeleton.SelectedJoints.Count = 0 then begin
            canExecuteTool := false;
            ErrorMessage := 'Cannot add joint to selected skeleton if no joint has been selected first. Please select one joint.';
         end;

         if selectedSkeleton.SelectedJoints.Count > 1 then begin
            canExecuteTool := false;
            ErrorMessage := 'Cannot add joint to selected skeleton if multiple joint are selected. Please select one joint only.'
         end;

    end else if SelectedSkeletons.Count > 1 then begin
        canExecuteTool := false;
        ErrorMessage := 'Cannot add joint to multiple selected skeletons. Please select one skeleton only.'
    end;

    Result := canExecuteTool;
end;

procedure TJointCreatorTool.Undo;
var index,i,j: integer;
    animation: TAnimation;
begin
    if Assigned(fSkeleton) then begin
        if not Assigned(fParentJoint) then begin
            // a parentjoint does not exist so we're dealing with a skeleton
            // that has only one joint; we undo the creation of that joint
            // which means we actually need to remove the entire skeleton
            SelectedSkeletons.Remove(fSkeleton);
            Skeletons.Remove(fSkeleton);
        end else begin
            // TODO use a jointRemoverFunction here! See unit_JointRemoverTool!

            // a parentjoint does exist; this means the skeleton has more
            // than one joint; we remove everything related to the joint that has
            // been added by the instance of this JointCreatorTool
            index := fParentJoint.ChildJoints.IndexOf(fJoint);
            if index <> -1 then
                fParentJoint.ChildBones.Remove(fParentJoint.ChildBones[index]);

            fParentJoint.ChildJoints.Remove(fJoint);

            fskeleton.BaseFrame.RemoveJointLocationByID(fJoint.ID);
            // remove jointlocations from all frames of all animations
            for i := 0 to fskeleton.Animations.Count-1 do begin
                animation := fskeleton.Animations[i];
                for j := 0 to animation.AnimationFrames.Count-1 do
                    animation.AnimationFrames[j].RemoveJointLocationByID(fJoint.ID);
            end;

            fSkeleton.Joints.Remove(fJoint);

            fSkeleton.SelectedJoints.Clear;
            fSkeleton.SelectedJoints.Add(fParentJoint);
        end;
    end;
end;

procedure TJointCreatorTool.Redo;
var paramList: TParamList;
begin
    // prepare
    if Assigned(fSkeleton) then begin
        if Assigned(fParentJoint) then begin
            fSkeleton.SelectedJoints.Clear;
            fSkeleton.SelectedJoints.Add(fParentJoint);
        end;
    end;

    // execute
    paramList := TParamList.Create(false);
    paramList.Add(fPosition,'position');
    paramList.Add(fMaterialManager,'materialManager');
    DoExecute(paramList);
    FreeAndNil(paramList);
end;

constructor TJointCreatorTool.Create(AToolAction: TToolAction);
begin
    inherited Create(AToolAction);
end;

destructor TJointCreatorTool.Destroy;
begin
    if Assigned(fPosition) then
        FreeAndNil(fPosition);
    inherited Destroy;
end;

end.
