{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_JointFlipperTool;

{$MODE Delphi}

interface

uses
  unit_SkeletonTool,
  unit_ParamList,
  unit_Joint,
  unit_Skeleton,
  unit_Enums,
  unit_SkeletonUndoItem,
  unit_Types,
  unit_AnimationFrame,
  unit_JointLocation,
  unit_MaterialManager;

type
  TJointFlipperTool = class(TSkeletonTool)
  private
    fMaterialManager: TMaterialManager;
    fSkeleton: TSkeleton;
    fParentJoint: TJoint;
    fChildJoint: TJoint;

    procedure FlipJoints(AMaterialManager: TMaterialManager; ASkeleton: TSkeleton; AParentJoint, AChildJoint: TJoint);
  protected
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  public
    constructor Create(AToolAction: TToolAction);
    destructor Destroy; override;
    function CanExecute(AParamList: TParamList): boolean; override;
    procedure Undo; override;
    procedure Redo; override;
  end;

implementation

uses
  SysUtils,
  Dialogs,
  unit_Animation,
  unit_BaseTool,
  unit_Math,
  unit_Bone,
  unit_BoneBuilder;

// ### TJointFlipperTool #######################################################

procedure TJointFlipperTool.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Flip joints';
    Description := 'Flips the parent/child relation between two joints.';
end;

procedure TJointFlipperTool.FlipJoints(AMaterialManager: TMaterialManager; ASkeleton: TSkeleton; AParentJoint, AChildJoint: TJoint);
var childIndex, i, j: integer;
    bone: TObject;
    paramList: TParamList;
    boneBuilder: TBoneBuilder;
    parentJointLocation, childJointLocation: TJointLocation;
    animation: TAnimation;
    animationFrame: TAnimationFrame;
begin
    // remove the bone from the parent
    childIndex := AParentJoint.ChildJoints.IndexOf(AChildJoint);
    AParentJoint.ChildBones.Remove(AParentJoint.ChildBones[childIndex]);

    // flip the jointlocations of the baseframe
    parentJointLocation := ASkeleton.BaseFrame.GetJointLocationByID(AParentJoint.ID);
    childJointLocation := ASkeleton.BaseFrame.GetJointLocationByID(AChildJoint.ID);

    parentJointLocation.ChildLocations.DeleteByJointLocationID(AChildJoint.ID);
    parentJointLocation.ParentLocation := childJointLocation;
    childJointLocation.ChildLocations.Add(parentJointLocation);
    childJointLocation.ParentLocation := nil;

    ASkeleton.BaseFrame.JointLocations.Exchange(0,ASkeleton.BaseFrame.JointLocations.IndexOf(childJointLocation));

    // do the same for the animationframes of all animations of the skeleton
    for i := 0 to ASkeleton.Animations.Count-1 do begin
        animation := ASkeleton.Animations[i];
        for j := 0 to animation.AnimationFrames.Count-1 do begin
            animationFrame := animation.AnimationFrames[j];

            parentJointLocation := animationFrame.GetJointLocationByID(AParentJoint.ID);
            childJointLocation := animationFrame.GetJointLocationByID(AChildJoint.ID);

            parentJointLocation.ChildLocations.DeleteByJointLocationID(AChildJoint.ID);
            parentJointLocation.ParentLocation := childJointLocation;
            childJointLocation.ChildLocations.Add(parentJointLocation);
            childJointLocation.ParentLocation := nil;

            animationFrame.JointLocations.Exchange(0,animationFrame.JointLocations.IndexOf(childJointLocation));
        end;
    end;

    // flip joints
    AParentJoint.ChildJoints.DeleteJointByID(AChildJoint.ID);
    AParentJoint.ParentJoint := AChildJoint;
    AChildJoint.ChildJoints.Add(AParentJoint);
    AChildJoint.ParentJoint := nil;         // the child is now the new basejoint in the skeleton

    // the basejoint needs to be the first joint in the collection;
    // we could introduce a baseJoint property and use that instead of skeleton.Joints[0] everywhere
    ASkeleton.Joints.Exchange(0,ASkeleton.Joints.IndexOf(AChildJoint));

    // create a new bone for the child (which is now the parent)
    paramList := TParamList.Create(false);
    paramList.Add(ASkeleton.BaseFrame.GetJointLocationByID(AChildJoint.ID), 'jointLocation1');
    paramList.Add(ASkeleton.BaseFrame.GetJointLocationByID(AParentJoint.ID), 'jointLocation2');
    paramList.Add(AMaterialManager, 'materialManager');
    paramList.Add(ASkeleton.JointSize, 'boneSize');                             // NOTE, JCC: bonesize = jointsize
    boneBuilder := TBoneBuilder.Create;
    bone := boneBuilder.Build(paramList);
    FreeAndNil(boneBuilder);
    FreeAndNil(paramList);

    // and add the bone to the child
    AChildJoint.ChildBones.Add(bone);
end;

procedure TJointFlipperTool.DoExecute(AParamList: TParamList);
var firstSelectedJoint, secondSelectedJoint: TJoint;
begin
    fMaterialManager := TMaterialManager(AParamList[0]);
    fSkeleton := SelectedSkeletons[0];
    firstSelectedJoint := fSkeleton.SelectedJoints[0];
    secondSelectedJoint := fSkeleton.SelectedJoints[1];

    if firstSelectedJoint.ParentJoint = secondSelectedJoint then begin
        fParentJoint := secondSelectedJoint;
        fChildJoint := firstSelectedJoint;
    end else begin
        fParentJoint := firstSelectedJoint;
        fChildJoint := secondSelectedJoint;
    end;

    FlipJoints(fMaterialManager, fSkeleton, fParentJoint, fChildJoint);
end;

function TJointFlipperTool.CanExecute(AParamList: TParamList): boolean;
var canExecuteTool: boolean;
    selectedSkeleton: TSkeleton;
    firstSelectedJoint, secondSelectedJoint: TJoint;
begin
    canExecuteTool := TRUE;

    if SelectedSkeletons.Count = 1 then begin
         selectedSkeleton := SelectedSkeletons[0];
         if selectedSkeleton.SelectedJoints.Count <> 2 then begin
            canExecuteTool := FALSE;
            ErrorMessage := 'Please select the base joint and one of its childjoints.';
         end else begin
            // we can only flip the base joint (which has no parent) with one of its childjoints

            firstSelectedJoint := selectedSkeleton.SelectedJoints[0];
            secondSelectedJoint := selectedSkeleton.SelectedJoints[1];

            canExecuteTool := ((firstSelectedJoint.ParentJoint=nil) or (secondSelectedJoint.ParentJoint=nil));
            if not canExecuteTool then
                ErrorMessage := 'Please select the base joint and one of its childjoints.'
            else begin
                canExecuteTool := ((firstSelectedJoint.ParentJoint = secondSelectedJoint) or (secondSelectedJoint.ParentJoint = firstSelectedJoint));
                if not canExecuteTool then
                    ErrorMessage := 'Please select the base joint and one of its childjoints.';
            end;
         end;

    end else if SelectedSkeletons.Count > 1 then begin
        canExecuteTool := FALSE;
        ErrorMessage := 'Cannot flip joints for multiple selected skeletons. Please select one skeleton only.'
    end;

    Result := canExecuteTool;
end;

procedure TJointFlipperTool.Undo;
begin
    FlipJoints(fMaterialManager, fSkeleton, fChildJoint, fParentJoint);
end;

procedure TJointFlipperTool.Redo;
begin
    FlipJoints(fMaterialManager, fSkeleton, fParentJoint, fChildJoint);
end;

constructor TJointFlipperTool.Create(AToolAction: TToolAction);
begin
    inherited Create(AToolAction);
end;

destructor TJointFlipperTool.Destroy;
begin
    inherited Destroy;
end;

end.
