{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_JointGeometryAssigningTool;

{$MODE Delphi}

interface

uses
  unit_SkeletonTool,
  unit_ParamList,
  unit_Skeleton,
  unit_Joint,
  unit_AnimationFrame,
  unit_JointLocation,
  unit_Types,
  unit_JointGeometrySelectingTool,
  unit_PrimitiveList;

type
  // This class assigns primitive geometry to skeletons and joints
  // of skeletons. It assignes primitives to a skeleton and vertices
  // of those primitives to joints of the skeleton.
  //
  // NOTES:
  //  - a skeleton can have multiple primitives assigned to it
  //  - primitives have a SkeletonID property
  //  - vertices have a JointID property which is saved to the DXS
  //  - joints have a AssignedVertices collection property
  //  - we need un-assign methods as well, which resets properties
  //    of primitives, vertices and joints; probably a seperate class
  //  - a vertex can only be assigned to one joint
  //  - if a vertex was already assigned to a joint and the user tries
  //    to assign that same vertex to another joint, the software will
  //    automatically unassign the vertex from its current joint
  //  - vertices assigned to joints are drawn in a specific color
  //  - this class uses a TJointGeometrySelectingTool as input; that
  //    class exposes SelectedJoints, SelectedPrimitives and SelectedVertices
  //    collections which hold the geometry to assign. When done, the assigning
  //    tool clears those collections.
  //  - assigned geometry is drawn using a particular color

  TJointGeometryAssigningTool = class(TSkeletonTool)
  private
    procedure HandleAssignPrimitives(ASkeleton: TSkeleton; APrimitiveList,AScenePrimitiveList: TPrimitiveList);
    procedure HandleAssignVertices(ASkeleton: TSkeleton; AJoint: TJoint; AVertexList: TVerticeList);
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  SysUtils, Dialogs, Classes,
  unit_Enums,
  unit_Options,
  unit_Primitives;

// ### TJointGeometryAssigningTool #############################################

procedure TJointGeometryAssigningTool.HandleAssignPrimitives(ASkeleton: TSkeleton; APrimitiveList, AScenePrimitiveList: TPrimitiveList);
var i: integer;
    primitive: TBasePrimitive;
begin
    // add the selected primitives to the primitivelist of the skeleton
    // and remove them from the primitivelist of the scene
    for i := 0 to APrimitiveList.Count-1 do begin
        primitive := APrimitiveList[i];
        primitive.SkeletonID := ASkeleton.ID;
        primitive.SnapMethod := smNone;
        ASkeleton.AssignedPrimitives.Add(primitive);
        AScenePrimitiveList.Remove(primitive,true);

        // create copies of the vertices in a backup collection
        primitive.CopyToOriginalVertices;
    end;
end;

procedure TJointGeometryAssigningTool.HandleAssignVertices(ASkeleton: TSkeleton; AJoint: TJoint; AVertexList: TVerticeList);
var i: integer;
    vertex: TVertice;
    joint: TJoint;
begin
    for i := 0 to AVertexList.Count-1 do begin
        vertex := AVertexList[i];
        if vertex.JointID<>-1 then begin
            // automatic unassigning in case the vertex was already
            // assigned to another joint
            joint := ASkeleton.Joints.GetJointByID(vertex.JointID);
            if Assigned(joint) then begin
                joint.AssignedVertices.Remove(vertex);
                vertex.JointID := -1;
            end;
        end;
        vertex.JointID := AJoint.ID;
        AJoint.AssignedVertices.Add(vertex);
    end;
end;

procedure TJointGeometryAssigningTool.DoPrepare(AParamList: TParamList);
begin
//
end;

procedure TJointGeometryAssigningTool.DoExecute(AParamList: TParamList);
var skeleton: TSkeleton;
    joint: TJoint;
    jointGeometrySelector: TJointGeometrySelectingTool;
    assignOptions: TAssignOptions;
    scenePrimitives: TPrimitiveList;
    selectedSceneSkelectons: TSkeletonList;
begin
    assignOptions := TAssignOptions(AParamList.NamedParameters['assignOptions']);
    jointGeometrySelector := TJointGeometrySelectingTool(AParamList.NamedParameters['jointGeometrySelector']);
    scenePrimitives := TPrimitiveList(AParamList.NamedParameters['scenePrimitives']);
    selectedSceneSkelectons := TSkeletonList(AParamList.NamedParameters['selectedSkeletons']);

    if assignOptions.SelectionType = asPrimitive then begin
        if selectedSceneSkelectons.Count<>1 then
            MessageDlg('Please select one skeleton.',mtError,[mbOK],0)
        else begin
            skeleton := selectedSceneSkelectons[0];
            HandleAssignPrimitives(skeleton, jointGeometrySelector.SelectedPrimitives, scenePrimitives);
            // geometry has been (un)assigned, so clear the selections
            //jointGeometrySelector.SelectedJoints.Clear;
            jointGeometrySelector.SelectedPrimitives.Clear;
            jointGeometrySelector.SelectedVertices.Clear;
        end;
    end;

    if assignOptions.SelectionType = asVertex then begin
        if jointGeometrySelector.SelectedJoints.Count<>1 then
            MessageDlg('Please select one joint.',mtError,[mbOK],0)
        else begin
            joint := jointGeometrySelector.SelectedJoints[0];
            skeleton := TSkeleton(joint.Skeleton);
            HandleAssignVertices(skeleton, joint, jointGeometrySelector.SelectedVertices);
            // geometry has been (un)assigned, so clear the selections
            //jointGeometrySelector.SelectedJoints.Clear;
            jointGeometrySelector.SelectedPrimitives.Clear;
            jointGeometrySelector.SelectedVertices.Clear;
        end;
    end;
end;

end.
