{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_JointGeometrySelectingTool;

{$MODE Delphi}

interface

uses
  unit_SkeletonTool,
  unit_ParamList,
  unit_Enums,
  unit_Skeleton,
  unit_Joint,
  unit_PrimitiveSelector,
  unit_VertexSelector,
  unit_JointSelector,
  unit_Types,
  unit_PrimitiveList;

type
  // Assigning geometry to a joint:
  //  - first select geometry by using a GeometrySelectingTool
  //  - assign to the joint by using a GeometryAssigningTool;
  //    this tool uses the GeometrySelectingTool as input so
  //    that it knows what geometry to assign to which joint
  //
  // This tool is exposed by the Scene. The geometry
  // selected by this tool is drawn in a particular color so
  // that the user may easily recognize selected geometry.
  //
  // TODO: this tool should be able to receive selected skeletons
  //       and joints from the Scene Inspector too...
  // Additional note: this whole way of working seems overcomplicated...

  TJointGeometrySelectingTool = class(TSkeletonTool)
  private
    fSelectedJoints: TJointList;
    fSelectedPrimitives: TPrimitiveList;
    fSelectedVertices: TVerticeList;

    procedure HandleSelectJointsByRay(AParamList: TParamList);
    procedure HandleSelectPrimitivesByRay(APrimitiveList: TPrimitiveList; AParamList: TParamList);
    procedure HandleSelectVerticesByRay(ASkeleton: TSkeleton; AParamList: TParamList);

    procedure HandleSelectJointsByRange(AParamList: TParamList);
    procedure HandleSelectPrimitivesByRange(APrimitiveList: TPrimitiveList; AParamList: TParamList);
    procedure HandleSelectVerticesByRange(ASkeleton: TSkeleton; AParamList: TParamList);

    procedure HandleSelectJointsByVertex(AParamList: TParamList);
    procedure HandleSelectPrimitivesByVertex(APrimitiveList: TPrimitiveList; AParamList: TParamList);
    procedure HandleSelectVerticesByVertex(ASkeleton: TSkeleton; AParamList: TParamList);
  public
    constructor Create(AToolAction: TToolAction);
    destructor Destroy; override;
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
    procedure Clear;
    property SelectedJoints: TJointList read fSelectedJoints write fSelectedJoints;
    property SelectedPrimitives: TPrimitiveList read fSelectedPrimitives write fSelectedPrimitives;
    property SelectedVertices: TVerticeList read fSelectedVertices write fSelectedVertices;
  end;

implementation

uses
  SysUtils, Dialogs,
  unit_Options;

// ### TJointGeometrySelectingTool #############################################

constructor TJointGeometrySelectingTool.Create(AToolAction: TToolAction);
begin
    inherited Create(AToolAction);
    fSelectedJoints := TJointList.Create(false, true);
    fSelectedPrimitives := TPrimitiveList.Create(false, true);
    fSelectedVertices := TVerticeList.Create(false, true);
end;

destructor TJointGeometrySelectingTool.Destroy;
begin
    FreeAndNil(fSelectedJoints);
    FreeAndNil(fSelectedPrimitives);
    FreeAndNil(fSelectedVertices);
    inherited Destroy;
end;

// ### By Ray selecting routines ###############################################

procedure TJointGeometrySelectingTool.HandleSelectJointsByRay(AParamList: TParamList);
begin

end;

procedure TJointGeometrySelectingTool.HandleSelectPrimitivesByRay(APrimitiveList: TPrimitiveList; AParamList: TParamList);
begin
end;

procedure TJointGeometrySelectingTool.HandleSelectVerticesByRay(ASkeleton: TSkeleton; AParamList: TParamList);
begin
end;

// ### By Range selecting routines #############################################

procedure TJointGeometrySelectingTool.HandleSelectJointsByRange(AParamList: TParamList);
var upperLeftVertex, lowerRightVertex: TVertice;
    sceneSkeletons,selectedSkeletons: TSkeletonList;
    jointSelector: TJointSelector;
    skeleton: TSkeleton;
    joint: TJoint;
    i: integer;
begin
    sceneSkeletons := TSkeletonList(AparamList.NamedParameters['sceneSkeletons']);
    selectedSkeletons := TSkeletonList(AparamList.NamedParameters['selectedSkeletons']);

    upperLeftVertex := TVertice(AparamList.NamedParameters['point1']);
    lowerRightVertex := TVertice(AparamList.NamedParameters['point2']);

    jointSelector := TJointSelector.Create(sceneSkeletons,selectedSkeletons);
    jointSelector.SelectByRange(upperLeftVertex, lowerRightVertex, true);

    if jointSelector.SelectedSkeletons.Count<>1 then begin
        // we can only have a joint from one skeleton only
        selectedSkeletons.Clear;
        fSelectedJoints.Clear;
    end else begin
        skeleton := jointSelector.SelectedSkeletons[0];
        // add joints to the SelectedJoints collection of this class;
        for i := 0 to skeleton.SelectedJoints.Count-1 do begin
            joint := skeleton.SelectedJoints[i];
            fSelectedJoints.Clear;
            fSelectedJoints.Add(joint);
        end;
    end;

    FreeAndNil(jointSelector);
end;

procedure TJointGeometrySelectingTool.HandleSelectPrimitivesByRange(APrimitiveList: TPrimitiveList; AParamList: TParamList);
var upperLeftVertex, lowerRightVertex: TVertice;
    primitiveSelector: TPrimitiveSelector;
    clearSelection: boolean;
begin
    clearSelection := TBooleanObject(AParamList.NamedParameters['clearSelection']).Value;
    upperLeftVertex := TVertice(AparamList.NamedParameters['point1']);
    lowerRightVertex := TVertice(AparamList.NamedParameters['point2']);

    // select primitives from the provided list and put them in the fSelectedPrimitives collection
    primitiveSelector := TPrimitiveSelector.Create(APrimitiveList, fSelectedPrimitives);
    primitiveSelector.SelectByRange(upperLeftVertex, lowerRightVertex, clearSelection);
    FreeAndNil(primitiveSelector);
end;

procedure TJointGeometrySelectingTool.HandleSelectVerticesByRange(ASkeleton: TSkeleton; AParamList: TParamList);
var upperLeftVertex, lowerRightVertex: TVertice;
    vertexSelector: TVertexSelector;
    vertexList: TVerticeList;
    vertex: TVertice;
    clearSelection: boolean;
    i,counter: integer;
begin
    assignOptions := TAssignOptions(AParamList.NamedParameters['assignOptions']);

    clearSelection := TBooleanObject(AParamList.NamedParameters['clearSelection']).Value;
    upperLeftVertex := TVertice(AparamList.NamedParameters['point1']);
    lowerRightVertex := TVertice(AparamList.NamedParameters['point2']);

    // select vertices of primitives that have been assigned to the skeleton
    vertexList := TVerticeList.Create(false, true);
    vertexSelector := TVertexSelector.Create(ASkeleton.AssignedPrimitives, vertexList);
    vertexSelector.SelectByRange(upperLeftVertex, lowerRightVertex, true);

    if clearSelection then
        fSelectedVertices.Clear;

    // if they were already selected, deselect them
    // if they weren't selected, select them
    for i := 0 to vertexList.Count-1 do begin
        vertex := vertexList[i];
        if fSelectedVertices.IsInList(vertex) then
            fSelectedVertices.Remove(vertex)
        else
            fSelectedVertices.Add(vertex);
    end;

    // if we want to unassign vertices, don't select vertices
    // that haven't been assigned in the first place yet
    if not assignOptions.Assign then begin
        counter := 0;
        while counter<fSelectedVertices.Count do begin
            vertex := fSelectedVertices[counter];
            if vertex.JointID=-1 then
                fSelectedVertices.Remove(vertex)
            else
                inc(counter);
        end;
    end;

    // don't select vertices that are already assigned, if wanted;
    // this is useful when assigning vertices to joints -> in that
    // case, you usually don't want already assigned vertices being
    // selected
    if not assignOptions.SelectAssignedGeometry then begin
        counter := 0;
        while counter<fSelectedVertices.Count do begin
            vertex := fSelectedVertices[counter];
            if vertex.JointID<>-1 then
                fSelectedVertices.Remove(vertex)
            else
                inc(counter);
        end;
    end;

    // free stuff
    FreeAndNil(vertexSelector);
    FreeAndNil(vertexList);
end;

// ### By Vertex selecting routines ############################################

procedure TJointGeometrySelectingTool.HandleSelectJointsByVertex(AParamList: TParamList);
var selectorVertex: TVertice;
    sceneSkeletons,selectedSkeletons: TSkeletonList;
    jointSelector: TJointSelector;
    skeleton: TSkeleton;
    joint: TJoint;
    i: integer;
    range: double;
begin
    sceneSkeletons := TSkeletonList(AparamList.NamedParameters['sceneSkeletons']);
    selectedSkeletons := TSkeletonList(AparamList.NamedParameters['selectedSkeletons']);
    selectorVertex := TVertice(AParamList.NamedParameters['selectorVertex']);
    range := TDoubleObject(AParamList.NamedParameters['range']).Value;

    jointSelector := TJointSelector.Create(sceneSkeletons,selectedSkeletons);
    jointSelector.SelectByVertex(selectorVertex, range, true);

    if selectedSkeletons.Count<>1 then begin
        // we can only have a joint from one skeleton only
        selectedSkeletons.Clear;
        fSelectedJoints.Clear
    end else begin
        skeleton := selectedSkeletons[0];
        // add joints to the SelectedJoints collection of this class;
        fSelectedJoints.Clear;
        for i := 0 to skeleton.SelectedJoints.Count-1 do begin
            joint := skeleton.SelectedJoints[i];
            fSelectedJoints.Add(joint);
        end;
    end;

    FreeAndNil(jointSelector);
end;

procedure TJointGeometrySelectingTool.HandleSelectPrimitivesByVertex(APrimitiveList: TPrimitiveList; AParamList: TParamList);
var selectorVertex: TVertice;
    clearSelection: boolean;
    primitiveSelector: TPrimitiveSelector;
begin
    clearSelection := TBooleanObject(AParamList.NamedParameters['clearSelection']).Value;
    selectorVertex := TVertice(AParamList.NamedParameters['selectorVertex']);

    // select primitives from the provided list and put them in the fSelectedPrimitives collection
    primitiveSelector := TPrimitiveSelector.Create(APrimitiveList, fSelectedPrimitives);
    primitiveSelector.SelectByVertex(selectorVertex,-1,clearSelection);
    FreeAndNil(primitiveSelector);
end;

procedure TJointGeometrySelectingTool.HandleSelectVerticesByVertex(ASkeleton: TSkeleton; AParamList: TParamList);
var selectorVertex: TVertice;
    vertexSelector: TVertexSelector;
    assignOptions: TAssignOptions;
    vertex: TVertice;
    clearSelection,doSelectVertices: boolean;
    counter: integer;
    range: single;          // NOTE: has to be single, no double!
begin
    assignOptions := TAssignOptions(AParamList.NamedParameters['assignOptions']);
    clearSelection := TBooleanObject(AParamList.NamedParameters['clearSelection']).Value;
    selectorVertex := TVertice(AParamList.NamedParameters['selectorVertex']);
    range := TSingleObject(AParamList.NamedParameters['range']).Value;
    doSelectVertices := TBooleanObject(AParamList.NamedParameters['doSelectVertices']).Value;   // or hightlight them

    // select vertices from primitives that have been assigned to the skeleton
    vertexSelector := TVertexSelector.Create(ASkeleton.AssignedPrimitives, fSelectedVertices);
    vertexSelector.SelectByVertex(selectorVertex, range, doSelectVertices, clearSelection);

    //if clearSelection then
    //    fSelectedVertices.Clear;

    // if they were already selected, deselect them
    // if they weren't selected, select them
    {for i := 0 to vertexList.Count-1 do begin
        vertex := vertexList[i];
        if fSelectedVertices.IsInList(vertex) then
            fSelectedVertices.Remove(vertex)
        else
            fSelectedVertices.Add(vertex);
    end;   }


    // if we're actually selecting vertices (not highlighting them),
    // we need to perform a few more checks
    if doSelectVertices then begin

        // if we want to unassign vertices, don't select vertices
        // that haven't been assigned in the first place yet
        if not assignOptions.Assign then begin
            counter := 0;
            while counter<fSelectedVertices.Count do begin
                vertex := fSelectedVertices[counter];
                if vertex.JointID=-1 then
                    fSelectedVertices.Remove(vertex)
                else
                    inc(counter);
            end;
        end;

        // don't select vertices that are already assigned, if wanted;
        // this is useful when assigning vertices to joints -> in that
        // case, you usually don't want already assigned vertices being
        // selected
        if not assignOptions.SelectAssignedGeometry then begin
            counter := 0;
            while counter<fSelectedVertices.Count do begin
                vertex := fSelectedVertices[counter];
                if vertex.JointID<>-1 then
                    fSelectedVertices.Remove(vertex)
                else
                    inc(counter);
            end;
        end;

    end;

    // free stuff
    FreeAndNil(vertexSelector);
end;

// #############################################################################

procedure TJointGeometrySelectingTool.DoPrepare(AParamList: TParamList);
begin

end;

procedure TJointGeometrySelectingTool.DoExecute(AParamList: TParamList);
var assignOptions: TAssignOptions;
    primitives, primitiveList: TPrimitiveList;
    {skeletons, }selectedSkeletons: TSkeletonList;
    selectedJoints: TJointList;
    skeleton: TSkeleton;
    selectMethod: TSelectMethod;
    //i: integer;
    //clearSelection: boolean;
begin
    assignOptions := TAssignOptions(AParamList.NamedParameters['assignOptions']);
    primitives := TPrimitiveList(AParamList.NamedParameters['scenePrimitives']);
    //skeletons := TSkeletonList(AParamList.NamedParameters['sceneSkeletons']);
    selectedSkeletons := TSkeletonList(AParamList.NamedParameters['selectedSkeletons']);
    selectMethod := TSelectMethod(TIntegerObject(AParamList.NamedParameters['selectMethod']).Value);
    selectedJoints := TJointList(AParamList.NamedParameters['selectedJoints']);
    //clearSelection := TBooleanObject(AParamList.NamedParameters['clearSelection']).Value;

    fSelectedJoints.Clear;
    if selectedJoints.Count>0 then
        fSelectedJoints.Add(selectedJoints[0]);

(*
    if clearSelection then
        //fSelectedJoints.Clear
    else begin
        // take over any selected joints from the scene
        //for i := 0 to selectedJoints.Count-1 do
        //    fSelectedJoints.Add(selectedJoints[i]);
        // no more than 1 joint at any time
        {if fSelectedJoints.Count>1 then begin
            for i := fSelectedJoints.Count-1 downto 1 do
                fSelectedJoints.Delete(i);

        end;}
    end;   *)

    case assignOptions.SelectionType of
        asJoint:
            begin
                case selectMethod of
                    smByVertex:
                        HandleSelectJointsByVertex(AParamList);
                    smByRange:
                        HandleSelectJointsByRange(AParamList);
                    smByRay:
                        HandleSelectJointsByRay(AParamList);
                end;
            end;

        asPrimitive:
            begin
                if selectedSkeletons.Count<>1 then
                    // error: we can only assign primitives to/from 1 skeleton
                else begin
                    if assignOptions.Assign then
                        primitiveList := primitives
                    else
                        primitiveList := selectedSkeletons[0].AssignedPrimitives;
                    if Assigned(primitiveList) then begin
                        case selectMethod of
                            smByVertex:
                                HandleSelectPrimitivesByVertex(primitiveList, AParamList);
                            smByRange:
                                HandleSelectPrimitivesByRange(primitiveList, AParamList);
                            smByRay:
                                HandleSelectPrimitivesByRay(primitiveList, AParamList);
                        end;
                    end;
                end;
            end;

        asVertex:
            begin
                if selectedSkeletons.Count<>1 then
                    // error: we can only assign vertices to/from 1 skeleton
                else begin
                    skeleton := selectedSkeletons[0];

                    case selectMethod of
                        smByVertex:
                            HandleSelectVerticesByVertex(skeleton, AParamList);
                        smByRange:
                            HandleSelectVerticesByRange(skeleton, AParamList);
                        smByRay:
                            HandleSelectVerticesByRay(skeleton, AParamList);
                    end;
                end;
            end;
    end;
end;

procedure TJointGeometrySelectingTool.Clear;
begin
    fSelectedJoints.Clear;
    fSelectedPrimitives.Clear;
    fSelectedVertices.Clear;    
end;

end.
