{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_JointGeometryUpdatingTool;

{$MODE Delphi}

interface

uses
  unit_SkeletonTool,
  unit_ParamList,
  unit_Skeleton,
  unit_Joint,
  unit_AnimationFrame,
  unit_JointLocation,
  unit_Math,
  unit_Types;

type
  // This class alters the geometry of a joint based on its jointlocation;
  // A jointlocation contains the position and orientation of the joint in
  // a frame which can be the baseframe of the skeleton or a frame in an
  // animation of the skeleton
  TJointGeometryUpdatingTool = class(TSkeletonTool)
  private
    procedure InitializeJointLocation(AJointLocation, ABaseJointLocation: TJointLocation);
    procedure InitializeAssignedGeometry(AJoint: TJoint);
    procedure RotateJointLocation(AJointLocation: TJointLocation; ARotationMatrix: TMatrix); overload;
    procedure RotateAssignedGeometry(AJoint: TJoint; ARotationMatrix: TMatrix);
    procedure RotateJointLocation(AJoint: TJoint; AJointLocation: TJointLocation); overload;
    procedure MoveJointLocation(AJoint: TJoint; AJointLocation: TJointLocation);
    procedure MoveAssignedGeometry(AJoint: TJoint; ADeltaVector: TVertice);
    procedure UpdateGeometry(AJoint: TJoint; AAnimationFrame: TAnimationFrame);
    procedure UpdateJointGeometry(AJoint: TJoint; AAnimationFrame: TAnimationFrame);
    procedure UpdateBoneGeometry(AJoint: TJoint; AAnimationFrame: TAnimationFrame);
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  SysUtils,
  unit_Enums,
  unit_Primitives,
  unit_Bone,
  unit_BoneTransformerTool;

// ### TJointGeometryUpdatingTool ##############################################

procedure TJointGeometryUpdatingTool.InitializeJointLocation(AJointLocation, ABaseJointLocation: TJointLocation);
var i: integer;
begin
    for i := 0 to AJointLocation.ChildLocations.Count-1 do
        InitializeJointLocation(AJointLocation.ChildLocations[i],ABaseJointLocation.ChildLocations[i]);

    AJointLocation.Position.Assign(ABaseJointLocation.Position);
end;

procedure TJointGeometryUpdatingTool.InitializeAssignedGeometry(AJoint: TJoint);
var i: integer;
    vertex: TVertice;
begin
    for i := 0 to AJoint.ChildJoints.Count-1 do
        InitializeAssignedGeometry(AJoint.ChildJoints[i]);

    for i := 0 to AJoint.AssignedVertices.Count-1 do begin
        vertex := AJoint.AssignedVertices[i];
        vertex.x := vertex.ox;
        vertex.y := vertex.oy;
        vertex.z := vertex.oz;
    end;
end;

procedure TJointGeometryUpdatingTool.RotateJointLocation(AJointLocation: TJointLocation; ARotationMatrix: TMatrix);
var i: integer;
    vertex: TVertice;
    x,y,z: double;
begin
    for i := 0 to AJointLocation.ChildLocations.Count-1 do
        RotateJointLocation(AJointLocation.ChildLocations[i], ARotationMatrix);

    // rotate the jointlocation
    vertex := AJointLocation.Position;
    vertex.x := vertex.x-ARotationMatrix[3,0];
    vertex.y := vertex.y-ARotationMatrix[3,1];
    vertex.z := vertex.z-ARotationMatrix[3,2];
    x := (vertex.x*ARotationMatrix[0,0])+(vertex.y*ARotationMatrix[0,1])+(vertex.z*ARotationMatrix[0,2])+ARotationMatrix[3,0];
    y := (vertex.x*ARotationMatrix[1,0])+(vertex.y*ARotationMatrix[1,1])+(vertex.z*ARotationMatrix[1,2])+ARotationMatrix[3,1];
    z := (vertex.x*ARotationMatrix[2,0])+(vertex.y*ARotationMatrix[2,1])+(vertex.z*ARotationMatrix[2,2])+ARotationMatrix[3,2];
    vertex.x := x;
    vertex.y := y;
    vertex.z := z;
end;

procedure TJointGeometryUpdatingTool.RotateAssignedGeometry(AJoint: TJoint; ARotationMatrix: TMatrix);
var i: integer;
    vertex: TVertice;
    x,y,z: double;
begin
    for i := 0 to AJoint.ChildJoints.Count-1 do
        RotateAssignedGeometry(AJoint.ChildJoints[i], ARotationMatrix);

    // rotate assigned geometry
    for i := 0 to AJoint.AssignedVertices.Count-1 do begin
        vertex := AJoint.AssignedVertices[i];
        vertex.x := vertex.x-ARotationMatrix[3,0];
        vertex.y := vertex.y-ARotationMatrix[3,1];
        vertex.z := vertex.z-ARotationMatrix[3,2];
        x := (vertex.x*ARotationMatrix[0,0])+(vertex.y*ARotationMatrix[0,1])+(vertex.z*ARotationMatrix[0,2])+ARotationMatrix[3,0];
        y := (vertex.x*ARotationMatrix[1,0])+(vertex.y*ARotationMatrix[1,1])+(vertex.z*ARotationMatrix[1,2])+ARotationMatrix[3,1];
        z := (vertex.x*ARotationMatrix[2,0])+(vertex.y*ARotationMatrix[2,1])+(vertex.z*ARotationMatrix[2,2])+ARotationMatrix[3,2];
        vertex.x := x;
        vertex.y := y;
        vertex.z := z;
    end;
end;

procedure TJointGeometryUpdatingTool.RotateJointLocation(AJoint: TJoint; AJointLocation: TJointLocation);
var rotationMatrix: TMatrix;
    i: integer;
begin
    // process childs first; this means we rotate joints going from leafs to root
    for i := 0 to AJointLocation.ChildLocations.Count-1 do
        RotateJointLocation(AJoint.ChildJoints[i], AJointLocation.ChildLocations[i]);

    // create a rotation matrix
    rotationmatrix := MatrixEulerSetup(AJointLocation.Rotation.x,AJointLocation.Rotation.y,AJointLocation.Rotation.z,
                                       AJointLocation.Position.x,AJointLocation.Position.y,AJointLocation.Position.z);

    // rotate about the jointlocation; this recursively rotates all childs too
    RotateJointLocation(AJointLocation, rotationMatrix);

    // rotate assigned geometry
    RotateAssignedGeometry(AJoint, rotationMatrix);

    // free matrix
    FreeAndNil(rotationMatrix);
end;

procedure TJointGeometryUpdatingTool.MoveJointLocation(AJoint: TJoint; AJointLocation: TJointLocation);
var i: integer;
begin
    // move the position according to the location vector
    AJointLocation.Position.Subtract(AJointLocation.Location);

    // process childs
    for i := 0 to AJointLocation.ChildLocations.Count-1 do
        MoveJointLocation(AJoint.ChildJoints[i], AJointLocation.ChildLocations[i]);
end;

procedure TJointGeometryUpdatingTool.MoveAssignedGeometry(AJoint: TJoint; ADeltaVector: TVertice);
var i: integer;
    vertex: TVertice;
begin
    for i := 0 to AJoint.ChildJoints.Count-1 do
        MoveAssignedGeometry(AJoint.ChildJoints[i], ADeltaVector);

    // move assigned geometry
    for i := 0 to AJoint.AssignedVertices.Count-1 do begin
        vertex := AJoint.AssignedVertices[i];
        vertex.Subtract(ADeltaVector);
    end;
end;

procedure TJointGeometryUpdatingTool.UpdateGeometry(AJoint: TJoint; AAnimationFrame: TAnimationFrame);
var i: integer;
begin
    // update joint geometry
    UpdateJointGeometry(AJoint, AAnimationFrame);

    // update bones geometry
    UpdateBoneGeometry(AJoint, AAnimationFrame);

    // recursively update child joints
    for i := 0 to AJoint.ChildJoints.Count-1 do
        UpdateGeometry(AJoint.ChildJoints[i], AAnimationFrame);
end;

procedure TJointGeometryUpdatingTool.UpdateJointGeometry(AJoint: TJoint; AAnimationFrame: TAnimationFrame);
var i: integer;
    vertex,scaleVector,originalVertex: TVertice;
    jointLocation: TJointLocation;
    skeleton: TSkeleton;
begin
    // update the joint geometry
    jointLocation := AAnimationFrame.GetJointLocationByID(AJoint.ID);
    //center := AJoint.Center;
    skeleton := TSkeleton(AJoint.Skeleton);
    scaleVector := TVertice.Create(skeleton.JointSize/64,skeleton.JointSize/64,skeleton.JointSize/64);
    for i := 0 to AJoint.Vertices.Count-1 do begin
        //vertex := AJoint.Vertices[i];
        // move to the origin first, then back to location
        //vertex.Subtract(center);

        originalVertex := AJoint.OriginalVertices[i];
        vertex := AJoint.Vertices[i];

        vertex.x := ((originalVertex.x-AJoint.OriginalCenter.x) * scaleVector.x);
        vertex.y := ((originalVertex.y-AJoint.OriginalCenter.y) * scaleVector.y);
        vertex.z := ((originalVertex.z-AJoint.OriginalCenter.z) * scaleVector.z);

        vertex.Add(jointLocation.Position);
    end;
    FreeAndNil(scaleVector);

    // update the AABB of the joint primitive
    AJoint.AABB.Update;
    AJoint.HasCenterChanged := true;
end;

procedure TJointGeometryUpdatingTool.UpdateBoneGeometry(AJoint: TJoint; AAnimationFrame: TAnimationFrame);
var i: integer;
    jointLocation: TJointLocation;
    childJoint: TJoint;
    childJointLocation: TJointLocation;
    boneTransformer: TBoneTransformerTool;
    paramList: TParamList;
    skeleton: TSkeleton;
begin
    // update the bones going from this joint to all childjoints
    if AJoint.ChildJoints.Count>0 then begin
        paramList := TParamList.Create(false);
        boneTransformer := TBoneTransformerTool.Create;
        jointLocation := AAnimationFrame.GetJointLocationByID(AJoint.ID);
        skeleton := TSkeleton(AJoint.Skeleton);

        // we have as many bones as we have joints
        for i := 0 to AJoint.ChildJoints.Count-1 do begin
            childJoint := AJoint.ChildJoints[i];
            childJointLocation := AAnimationFrame.GetJointLocationByID(childJoint.ID);

            paramList.Add(AJoint.ChildBones[i], 'childBone', true, false);
            paramList.Add(jointLocation, 'jointLocation');
            paramList.Add(childJointLocation, 'childJointLocation');
            paramList.Add(skeleton.JointSize, 'boneSize');
            boneTransformer.Execute(paramList);
        end;
        FreeAndNil(boneTransformer);
        FreeAndNil(paramList);
    end;
end;

procedure TJointGeometryUpdatingTool.DoPrepare(AParamList: TParamList);
begin
end;

procedure TJointGeometryUpdatingTool.DoExecute(AParamList: TParamList);
var i: integer;
    skeleton: TSkeleton;
    baseFrame,frame: TAnimationFrame;
    joint: TJoint;
    baseJointLocation,jointLocation: TJointLocation;
begin
    for i := 0 to SelectedSkeletons.Count-1 do begin
        // get some data to work with
        skeleton := SelectedSkeletons[i];
        baseFrame := skeleton.BaseFrame;
        frame := skeleton.GetActiveFrame;   // either baseframe or a frame of an animation
        joint := skeleton.Joints[0];        // get the first joint of the skeleton and recursively update all joints
        baseJointLocation := baseFrame.GetJointLocationByID(joint.ID);
        jointLocation := frame.GetJointLocationByID(joint.ID);

        // first give all jointlocations their original position from the baseframe
        InitializeJointLocation(jointLocation, baseJointLocation);

        // reset all assigned vertices of all joints of the skeleton to their original positions;
        // we need to do this once before rotating all vertices into position in the next step
        InitializeAssignedGeometry(joint);

        // rotate jointlocations and assigned geometry into position
        RotateJointLocation(joint, jointLocation);

        // apply movement vector
        MoveJointLocation(joint, jointLocation);

        // move assigned geometry by the same amount
        MoveAssignedGeometry(joint, jointLocation.Location);

        // update the geometry of the joint using the new position
        UpdateGeometry(joint,frame);
    end;
end;


end.

