{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
// TODO rename to JointMetaData

unit unit_JointLocation;

{$MODE Delphi}

interface

uses
  unit_Types,
  unit_Primitives;

type
  TJointLocationList = class;

  TJointLocation = class(TObject)
  private
    fID: integer;
    fLocation: TVertice;
    fRotation: TVertice;
    fPosition: TVertice;
    fParentLocation: TJointLocation;
    fChildLocations: TJointLocationList;
    procedure ClearChildLocations(AJointLocation: TJointLocation);
  public
    constructor Create;
    destructor Destroy; override;
    function Copy(AJointLocationList: TJointLocationList): TJointLocation;
    procedure Assign(AJointLocation: TJointLocation);
    property ID: integer read fID write fID;
    property Location: TVertice read fLocation write fLocation;
    property Rotation: TVertice read fRotation write fRotation;
    property Position: TVertice read fPosition write fPosition;
    property ParentLocation: TJointLocation read fParentLocation write fParentLocation;
    property ChildLocations: TJointLocationList read fChildLocations write fChildLocations;
  end;

  // jointlocations can be stored in a jointlocationlist but also form
  // a tree structure by nature; the first jointlocation in the list is
  // the root of the tree
  TJointLocationList = class(TDeleDObjectList)
  private
    function GetJointLocation(anIndex: integer): TJointLocation;
    procedure AddToList(AJointLocation: TJointLocation; AJointLocationList: TJointLocationList);
  public
    //procedure Copy(AJointLocationList: TJointLocationList);
    function GetJointLocationByID(AID: integer): TJointLocation;
    procedure DeleteByJointLocationID(AID: integer);
    procedure Assign(AJointLocationList: TJointLocationList);
    property Locations[anIndex: integer]: TJointLocation read GetJointLocation; default;
  end;

implementation

uses SysUtils;

// ### TJointLocation ##########################################################

constructor TJointLocation.Create;
begin
    fLocation := TVertice.Create(0,0,0);
    fRotation := TVertice.Create(0,0,0);
    fPosition := TVertice.Create(0,0,0);
    fChildLocations := TJointLocationList.Create(false,true);                   // a frame owns the jointlocations
end;

destructor TJointLocation.Destroy;
begin
    FreeAndNil(fChildLocations);
    FreeAndNil(fRotation);
    FreeAndNil(fLocation);
    FreeAndNil(fPosition);
    inherited Destroy;
end;

function TJointLocation.Copy(AJointLocationList: TJointLocationList): TJointLocation;
var jointLocation, childJointLocation: TJointLocation;
    i: integer;
begin
    jointLocation := TJointLocation.Create;
    jointLocation.ID := ID;
    jointLocation.Location.Assign(Location);
    jointLocation.Rotation.Assign(Rotation);

    AJointLocationList.Add(jointLocation);

    for i := 0 to ChildLocations.Count-1 do begin
        childJointLocation := ChildLocations[i].Copy(AJointLocationList);
        childJointLocation.ParentLocation := jointLocation;
        jointLocation.ChildLocations.Add(childJointLocation);
    end;

    Result := jointLocation;
end;

procedure TJointLocation.ClearChildLocations(AJointLocation: TJointLocation);
var i: integer;
begin
    for i := 0 to AJointLocation.ChildLocations.Count-1 do begin
        ClearChildLocations(AJointLocation.ChildLocations[i]);
        AJointLocation.ChildLocations.Clear;
    end;
end;

procedure TJointLocation.Assign(AJointLocation: TJointLocation);
var i: integer;
    childLocation: TJointLocation;
begin
    ID := AJointLocation.ID;
    Location.Assign(AJointLocation.Location);
    Rotation.Assign(AJointLocation.Rotation);
    Position.Assign(AJointLocation.Position);
    ClearChildLocations(self);
    for i := 0 to AJointLocation.ChildLocations.Count-1 do begin
        childLocation := TJointLocation.Create;
        childLocation.Assign(AJointLocation.ChildLocations[i]);
        childLocation.ParentLocation := self;
        ChildLocations.Add(childLocation);
    end;
end;

// ### TJointLocationList ######################################################

function TJointLocationList.GetJointLocation(anIndex: integer): TJointLocation;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TJointLocation(Items[anIndex])
    else
        Result := nil;
end;
{
procedure TJointLocationList.Copy(AJointLocationList: TJointLocationList);
var originalJointLocation: TJointLocation;
begin
    // traverse the original tree, copying jointlocations and
    // build a new tree while adding the nodes to the new list too
    originalJointLocation := Locations[0];
    originalJointLocation.Copy(AJointLocationList);
end; }

procedure TJointLocationList.AddToList(AJointLocation: TJointLocation; AJointLocationList: TJointLocationList);
var i: integer;
    childLocation: TJointLocation;
begin
    AJointLocationList.Add(AJointLocation);
    for i := 0 to AJointLocation.ChildLocations.Count-1 do begin
        childLocation := AJointLocation.ChildLocations[i];
        AddToList(childLocation,AJointLocationList);
        //AJointLocationList.Add(childLocation);
    end;
end;

procedure TJointLocationList.Assign(AJointLocationList: TJointLocationList);
var jointLocation: TJointLocation;
begin
    Clear;
    if AJointLocationList.Count>0 then begin
        jointLocation := TJointLocation.Create;
        jointLocation.Assign(AJointLocationList[0]);
        AddToList(jointLocation,self);
    end;
end;

procedure TJointLocationList.DeleteByJointLocationID(AID: integer);
var jointLocation: TJointLocation;
begin
    jointLocation := GetJointLocationByID(AID);
    if Assigned(jointLocation) then
        Delete(IndexOf(jointLocation));
end;

function TJointLocationList.GetJointLocationByID(AID: integer): TJointLocation;
var counter: integer;
    isFound: boolean;
    jointLocation: TJointLocation;
begin
    counter := 0;
    isFound := false;
    jointLocation := nil;
    while (counter<Count) and (not isFound) do begin
        jointLocation := Locations[counter];
        if jointLocation.ID = AID then
            isFound := true
        else
            inc(counter);
    end;
    Result := jointLocation;
end;

end.




