{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_JointLocationMoverTool;

{$MODE Delphi}

interface

uses
  unit_SkeletonTool,
  unit_JointLocation,
  unit_Types,
  unit_IntegerList,
  unit_ParamList,
  unit_Math,
  unit_Enums,
  unit_Skeleton,
  unit_AnimationFrame,
  unit_SkeletonUndoItem,
  unit_Joint;

type
  TSkeletonUndoMoveItem = class(TSkeletonUndoItem)
  private
    fDeltaVector: TVertice;
  public
    constructor Create;
    destructor Destroy; override;
    property DeltaVector: TVertice read fDeltaVector write fDeltaVector;
  end;

  // This class moves the jointlocations of the selected joints in a given
  // frame of a given animation of the skeleton (or the baseframe).
  // It can optionally move the jointlocations of the childjoints of a
  // selected joint as well.
  // Also, when used on an Animation frame, the class makes sure the length of
  // bones isn't changed. This means joints can only move in a circular motion.
  //
  // NOTE: this class does NOT alter the geometry of the joints and bones.
  // It only moves the jointlocations by a certain amount.
  TJointLocationMoverTool = class(TSkeletonTool)
  private
    fSkeletonUndoList: TSkeletonUndoList;
    procedure PerformUndoRedo(APerformUndo: boolean);
    procedure UpdateJointGeometry;
    //function GetJointsFromIDs(AJointIDList: TIntegerList): TJointList;
    function CreateJointList(ASkeleton: TSkeleton): TJointList;
    procedure MoveAssignedGeometry(AJoint: TJoint; ADeltaVector: TVertice);
    procedure MoveJointLocationForBaseFrame(AJoint: TJoint; AJointLocation: TJointLocation; ADeltaVector: TVertice);
    procedure MoveJointLocationForAnimationFrame(AJointLocation: TJointLocation; ADeltaVector: TVertice);
  public
    constructor Create(AToolAction: TToolAction);
    destructor Destroy; override;
    procedure Undo; override;
    procedure Redo; override;
    function CanExecute(AParamList: TParamList): boolean; override;
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  SysUtils, Math, Dialogs,
  unit_Variables,
  unit_Animation,
  unit_JointGeometryUpdatingTool;

// ### TJointLocationMoverTool #################################################

procedure TJointLocationMoverTool.PerformUndoRedo(APerformUndo: boolean);
var i,j: integer;
    skeletonUndoItem: TSkeletonUndoMoveItem;
    skeleton: TSkeleton;
    joint: TJoint;
    jointLocation: TJointLocation;
    frame: TAnimationFrame;
    deltaVector: TVertice;
begin
    SelectedSkeletons.Clear;

    // one Undo for each skeleton
    for i := 0 to fSkeletonUndoList.Count-1 do begin
        skeletonUndoItem := TSkeletonUndoMoveItem(fSkeletonUndoList[i]);
        skeleton := skeletonUndoItem.Skeleton;
        skeleton.SelectedJoints.Clear;
        frame := skeletonUndoItem.Frame;

        if APerformUndo then
            deltaVector := skeletonUndoItem.DeltaVector.Inverse
        else
            deltaVector := skeletonUndoItem.DeltaVector.Copy;

        for j := 0 to skeletonUndoItem.Joints.Count-1 do begin
            joint := skeletonUndoItem.Joints[j];
            jointLocation := frame.GetJointLocationByID(joint.ID);
            if frame.IsBaseFrame then
                MoveJointLocationForBaseFrame(joint, jointLocation, deltaVector)
            else begin
                // NOTE: we move the entire skeleton and its assigned geometry
                MoveJointLocationForAnimationFrame(frame.JointLocations[0], deltaVector);

                // it might not be a keyframe anymore
                frame.IsKeyFrame := skeletonUndoItem.WasKeyFrameBeforeExecute;

                // animation has changed
                if Assigned(frame.Animation) then begin
                    TAnimation(frame.Animation).HasChanged := true;
                    skeleton.SelectedAnimation := TAnimation(frame.Animation);
                end;
            end;
            skeleton.SelectedJoints.Add(joint);
        end;
        FreeAndNil(deltaVector);

        SelectedSkeletons.Add(skeleton);
    end;

    UpdateJointGeometry;
end;

procedure TJointLocationMoverTool.Undo;
begin
    PerformUndoRedo(true);
end;

procedure TJointLocationMoverTool.Redo;
begin
    PerformUndoRedo(false);
end;

procedure TJointLocationMoverTool.UpdateJointGeometry;
var jointGeometryUpdater: TJointGeometryUpdatingTool;
begin
    jointGeometryUpdater := TJointGeometryUpdatingTool.Create(taNotApplicable);
    jointGeometryUpdater.Skeletons := Skeletons;
    jointGeometryUpdater.SelectedSkeletons := SelectedSkeletons;
    try
        jointGeometryUpdater.Execute(nil);
    finally
        FreeAndNil(jointGeometryUpdater);
    end;
end;

function TJointLocationMoverTool.CreateJointList(ASkeleton: TSkeleton): TJointList;
var i: integer;
    jointList: TJointList;
    selectedJoint,parentJoint: TJoint;
    isSelected: boolean;
begin
    jointList := TJointList.Create(false,true);

    for i := 0 to ASkeleton.SelectedJoints.Count-1 do begin
        selectedJoint := ASkeleton.SelectedJoints[i];
        parentJoint := selectedJoint.ParentJoint;
        isSelected := false;
        // walk upwards within the tree, searching for selected parent joints
        while Assigned(parentJoint) and (not isSelected) do begin
            isSelected := ASkeleton.SelectedJoints.IsInList(parentJoint);
            parentJoint := parentJoint.ParentJoint;
        end;
        // if no parent joint was selected, add the current selected joint to the list
        if not isSelected then
            jointList.Add(selectedJoint);
    end;

    Result := jointList;
end;

procedure TJointLocationMoverTool.MoveAssignedGeometry(AJoint: TJoint; ADeltaVector: TVertice);
var i: integer;
    vertex: TVertice;
begin
    for i := 0 to AJoint.AssignedVertices.Count-1 do begin
        vertex := AJoint.AssignedVertices[i];
        vertex.ox := vertex.ox - ADeltaVector.x;
        vertex.oy := vertex.oy - ADeltaVector.y;
        vertex.oz := vertex.oz - ADeltaVector.z;
    end;
end;

procedure TJointLocationMoverTool.MoveJointLocationForBaseFrame(AJoint: TJoint; AJointLocation: TJointLocation; ADeltaVector: TVertice);
var i: integer;
begin
    for i := 0 to AJointLocation.ChildLocations.Count-1 do
        MoveJointLocationForBaseFrame(AJoint.ChildJoints[i], AJointLocation.ChildLocations[i], ADeltaVector);

    AJointLocation.Position.Subtract(ADeltaVector);

    MoveAssignedGeometry(AJoint, ADeltaVector);
end;

procedure TJointLocationMoverTool.MoveJointLocationForAnimationFrame(AJointLocation: TJointLocation; ADeltaVector: TVertice);
var i: integer;
begin
    // TODO: insert PJ code here which moves individual joints while preserving bone lengths
    //
    // NOTE: for now, we move the entire skeleton and its assigned geometry; the provided AJointLocation
    // is the first JointLocation of the current animation (see the Execute method of this class)
    for i := 0 to AJointLocation.ChildLocations.Count-1 do
        MoveJointLocationForAnimationFrame(AJointLocation.ChildLocations[i], ADeltaVector);

    AJointLocation.Location.Add(ADeltaVector);
end;

function TJointLocationMoverTool.CanExecute(AParamList: TParamList): boolean;
var counter: integer;
    skeleton: TSkeleton;
    canExecuteMove: boolean;
begin
    canExecuteMove := false;
    counter := 0;
    while (counter<SelectedSkeletons.Count) and (not canExecuteMove) do begin
        skeleton := SelectedSkeletons[counter];
        if skeleton.SelectedJoints.Count>0 then
            canExecuteMove := true
        else
            inc(counter);
    end;
    Result := canExecuteMove;
end;

procedure TJointLocationMoverTool.DoPrepare(AParamList: TParamList);
var i,j: integer;
    joint: TJoint;
    jointList: TJointList;
    skeleton: TSkeleton;
    frame: TAnimationFrame;
    skeletonUndoItem: TSkeletonUndoMoveItem;
begin
    Hint := 'Move joint';
    Description := 'Moves a joint. Works on the skeleton or within a given frame of an animation.';

    // we store information for Undo purposes up front
    for i := 0 to SelectedSkeletons.Count-1 do begin
        skeleton := SelectedSkeletons[i];
        frame := skeleton.GetActiveFrame;

        skeletonUndoItem := TSkeletonUndoMoveItem.Create;
        skeletonUndoItem.Skeleton := skeleton;
        skeletonUndoItem.Frame := frame;
        skeletonUndoItem.WasKeyFrameBeforeExecute := frame.IsKeyFrame;

        jointList := CreateJointList(skeleton);
        for j := 0 to jointList.Count-1 do begin
            joint := jointList[j];
            skeletonUndoItem.Joints.Add(joint);
        end;
        FreeAndNil(jointList);

        fSkeletonUndoList.Add(skeletonUndoItem);
    end;
end;

procedure TJointLocationMoverTool.DoExecute(AParamList: TParamList);
var i,j: integer;
    deltaVector: TVertice;
    skeleton: TSkeleton;
    joint: TJoint;
    jointList: TJointList;
    frame: TAnimationFrame;
    jointLocation: TJointLocation;
    skeletonUndoItem: TSkeletonUndoMoveItem;
begin
    deltaVector := TVertice(AParamList[0]);

    for i := 0 to SelectedSkeletons.Count-1 do begin
        skeleton := SelectedSkeletons[i];
        frame := skeleton.GetActiveFrame;

        skeletonUndoItem := TSkeletonUndoMoveItem(fSkeletonUndoList[i]);
        skeletonUndoItem.DeltaVector.Add(deltaVector);

        // When moving a joint (its jointlocation actually), we also
        // move the child joints.
        // If we have more than 1 joint selected, we need to determine
        // if any of the childs of the selected joints are selected as well.
        // If so, those childs need to be removed from the selection to
        // avoid moving them (and their children) multiple times.
        // NOTE: we use the same principle when deleting joints.
        jointList := CreateJointList(skeleton);
        for j := 0 to jointList.Count-1 do begin
            joint := jointList[j];

            jointLocation := frame.GetJointLocationByID(joint.ID);
            if frame.IsBaseFrame then
                MoveJointLocationForBaseFrame(joint, jointLocation, deltaVector)
            else begin
                // NOTE: we move the entire skeleton and its assigned geometry
                MoveJointLocationForAnimationFrame(frame.JointLocations[0],deltaVector);

                // jointlocation changed by user; this frame is now a keyframe
                frame.IsKeyFrame := true;

                // animation has changed
                if Assigned(frame.Animation) then
                    TAnimation(frame.Animation).HasChanged := true;
            end;
        end;
        FreeAndNil(jointList);
    end;

    UpdateJointGeometry;
end;

constructor TJointLocationMoverTool.Create(AToolAction: TToolAction);
begin
    inherited Create(AToolAction);
    fSkeletonUndoList := TSkeletonUndoList.Create(true,true);
end;

destructor TJointLocationMoverTool.Destroy;
begin
    inherited Destroy;
    FreeAndNil(fSkeletonUndoList);
end;

// ### TSkeletonUndoMoveItem ###################################################

constructor TSkeletonUndoMoveItem.Create;
begin
    fDeltaVector := TVertice.Create(0,0,0);
    inherited Create;
end;

destructor TSkeletonUndoMoveItem.Destroy;
begin
    FreeAndNil(fDeltaVector);
    inherited Destroy;
end;

end.
 
