{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_JointLocationRotatorTool;

{$MODE Delphi}

interface

uses
  unit_SkeletonTool,
  unit_JointLocation,
  unit_Types,
  unit_ParamList,
  unit_Math,
  unit_SkeletonUndoItem,
  unit_Enums,
  unit_Joint;

type
  TSkeletonUndoRotateItem = class(TSkeletonUndoItem)
  private
    fDeltaVector: TVertice;
  public
    constructor Create;
    destructor Destroy; override;
    property DeltaVector: TVertice read fDeltaVector write fDeltaVector;
  end;

  TJointLocationRotatorTool = class(TSkeletonTool)
  private
    fSkeletonUndoList: TSkeletonUndoList;
    procedure PerformUndoRedo(APerformUndo: boolean);
    procedure UpdateJointGeometry;
    procedure RotateJointLocation(AJointLocation: TJointLocation; ARotationMatrix: TMatrix);
    procedure RotateAssignedGeometry(AJoint: TJoint; ARotationMatrix: TMatrix);
    procedure RotateJointLocationForBaseFrame(AJoint: TJoint; AJointLocation: TJointLocation; ADeltaVector: TVertice);
    procedure RotateJointLocationForAnimationFrame(AJointLocation: TJointLocation; ADeltaVector: TVertice);
  public
    constructor Create(AToolAction: TToolAction);
    destructor Destroy; override;
    procedure Undo; override;
    procedure Redo; override;
    function CanExecute(AParamList: TParamList): boolean; override;
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  SysUtils,
  unit_Variables,
  unit_Skeleton,
  unit_Animation,
  unit_AnimationFrame,
  unit_JointGeometryUpdatingTool;

// ### TJointLocationRotatorTool ###############################################

constructor TJointLocationRotatorTool.Create(AToolAction: TToolAction);
begin
    inherited Create(AToolAction);
    fSkeletonUndoList := TSkeletonUndoList.Create(true,true);
end;

destructor TJointLocationRotatorTool.Destroy;
begin
    inherited Destroy;
    FreeAndNil(fSkeletonUndoList);
end;

procedure TJointLocationRotatorTool.PerformUndoRedo(APerformUndo: boolean);
var i,j: integer;
    skeletonUndoItem: TSkeletonUndoRotateItem;
    skeleton: TSkeleton;
    joint: TJoint;
    jointLocation: TJointLocation;
    frame: TAnimationFrame;
    deltaVector: TVertice;
begin
    SelectedSkeletons.Clear;

    for i := 0 to fSkeletonUndoList.Count-1 do begin
        skeletonUndoItem := TSkeletonUndoRotateItem(fSkeletonUndoList[i]);
        skeleton := skeletonUndoItem.Skeleton;
        skeleton.SelectedJoints.Clear;
        frame := skeletonUndoItem.Frame;

        if APerformUndo then
            deltaVector := skeletonUndoItem.DeltaVector.Inverse
        else
            deltaVector := skeletonUndoItem.DeltaVector.Copy;

        for j := 0 to skeletonUndoItem.Joints.Count-1 do begin
            joint := skeletonUndoItem.Joints[j];
            jointLocation := frame.GetJointLocationByID(joint.ID);
            if frame.IsBaseFrame then
                RotateJointLocationForBaseFrame(joint, jointLocation, deltaVector)
            else begin
                RotateJointLocationForAnimationFrame(jointLocation, deltaVector);

                // it might not be a keyframe anymore
                frame.IsKeyFrame := skeletonUndoItem.WasKeyFrameBeforeExecute;

                // animation has changed
                if Assigned(frame.Animation) then begin
                    TAnimation(frame.Animation).HasChanged := true;
                    skeleton.SelectedAnimation := TAnimation(frame.Animation);
                end;
            end;
            skeleton.SelectedJoints.Add(joint);
        end;
        FreeAndNil(deltaVector);

        SelectedSkeletons.Add(skeleton);
    end;

    UpdateJointGeometry;
end;

procedure TJointLocationRotatorTool.Undo;
begin
    PerformUndoRedo(true);
end;

procedure TJointLocationRotatorTool.Redo;
begin
    PerformUndoRedo(false);
end;

procedure TJointLocationRotatorTool.UpdateJointGeometry;
var jointGeometryUpdater: TJointGeometryUpdatingTool;
begin
    jointGeometryUpdater := TJointGeometryUpdatingTool.Create(taNotApplicable);
    jointGeometryUpdater.Skeletons := Skeletons;
    jointGeometryUpdater.SelectedSkeletons := SelectedSkeletons;
    try
        jointGeometryUpdater.Execute(nil);
    finally
        FreeAndNil(jointGeometryUpdater);
    end;
end;

procedure TJointLocationRotatorTool.RotateJointLocation(AJointLocation: TJointLocation; ARotationMatrix: TMatrix);
var i: integer;
    vertex: TVertice;
    x,y,z: double;
begin
    for i := 0 to AJointLocation.ChildLocations.Count-1 do
        RotateJointLocation(AJointLocation.ChildLocations[i],ARotationMatrix);

    vertex := AJointLocation.Position;
    vertex.x := vertex.x-ARotationMatrix[3,0];
    vertex.y := vertex.y-ARotationMatrix[3,1];
    vertex.z := vertex.z-ARotationMatrix[3,2];
    x := (vertex.x*ARotationMatrix[0,0])+(vertex.y*ARotationMatrix[0,1])+(vertex.z*ARotationMatrix[0,2])+ARotationMatrix[3,0];
    y := (vertex.x*ARotationMatrix[1,0])+(vertex.y*ARotationMatrix[1,1])+(vertex.z*ARotationMatrix[1,2])+ARotationMatrix[3,1];
    z := (vertex.x*ARotationMatrix[2,0])+(vertex.y*ARotationMatrix[2,1])+(vertex.z*ARotationMatrix[2,2])+ARotationMatrix[3,2];
    vertex.x := x;
    vertex.y := y;
    vertex.z := z;
end;

procedure TJointLocationRotatorTool.RotateAssignedGeometry(AJoint: TJoint; ARotationMatrix: TMatrix);
var i: integer;
    vertex: TVertice;
    x,y,z: double;
begin
    for i := 0 to AJoint.ChildJoints.Count-1 do
        RotateAssignedGeometry(AJoint.ChildJoints[i], ARotationMatrix);

    // rotate assigned geometry
    for i := 0 to AJoint.AssignedVertices.Count-1 do begin
        vertex := AJoint.AssignedVertices[i];
        vertex.x := vertex.ox-ARotationMatrix[3,0];
        vertex.y := vertex.oy-ARotationMatrix[3,1];
        vertex.z := vertex.oz-ARotationMatrix[3,2];
        x := (vertex.x*ARotationMatrix[0,0])+(vertex.y*ARotationMatrix[0,1])+(vertex.z*ARotationMatrix[0,2])+ARotationMatrix[3,0];
        y := (vertex.x*ARotationMatrix[1,0])+(vertex.y*ARotationMatrix[1,1])+(vertex.z*ARotationMatrix[1,2])+ARotationMatrix[3,1];
        z := (vertex.x*ARotationMatrix[2,0])+(vertex.y*ARotationMatrix[2,1])+(vertex.z*ARotationMatrix[2,2])+ARotationMatrix[3,2];
        // save to the ox,oy and oz values -> these are the original x,y,z values of the vertex and are used
        // by the InitializeAssignedGeometry of the JointGeometryUpdatingTool to reset assigned vertices
        // to their default location
        vertex.ox := x;
        vertex.oy := y;
        vertex.oz := z;
    end;
end;

// jointlocations of the baseframe are rotated immediately
procedure TJointLocationRotatorTool.RotateJointLocationForBaseFrame(AJoint: TJoint; AJointLocation: TJointLocation; ADeltaVector: TVertice);
var rotationMatrix: TMatrix;
begin
    // create a rotation matrix
    rotationmatrix := MatrixEulerSetup(ADeltaVector.x,ADeltaVector.y,ADeltaVector.z,
                                       AJointLocation.Position.x,AJointLocation.Position.y,AJointLocation.Position.z);

    // rotate about the jointlocation; this recursively rotates all childs too
    RotateJointLocation(AJointLocation,rotationMatrix);

    // rotate assigned vertices
    RotateAssignedGeometry(AJoint, rotationMatrix);

    // free matrix
    FreeAndNil(rotationMatrix);
end;

procedure TJointLocationRotatorTool.RotateJointLocationForAnimationFrame(AJointLocation: TJointLocation; ADeltaVector: TVertice);
begin
    // simply add to the rotation vector; this vector represents rotation
    // angles and is used by the TJointGeometryUpdating tool which takes
    // the original position of a joint from the baseframe and then rotates
    // it into position using its rotation vector;
    //
    // the same goes for any assigned vertex ->
    // it's original position is taken from the baseframe after which it is
    // rotated into its new position using the rotation vector of the
    // corresponding jointlocation
    AJointLocation.Rotation.Add(ADeltaVector);
end;

function TJointLocationRotatorTool.CanExecute(AParamList: TParamList): boolean;
var counter: integer;
    skeleton: TSkeleton;
    canExecuteRotate: boolean;
begin
    canExecuteRotate := false;
    counter := 0;
    while (counter<SelectedSkeletons.Count) and (not canExecuteRotate) do begin
        skeleton := SelectedSkeletons[counter];
        if skeleton.SelectedJoints.Count=1 then
            canExecuteRotate := true
        else
            inc(counter);
    end;
    Result := canExecuteRotate;
end;

procedure TJointLocationRotatorTool.DoPrepare(AParamList: TParamList);
var i,j: integer;
    skeleton: TSkeleton;
    frame: TAnimationFrame;
    skeletonUndoItem: TSkeletonUndoRotateItem;
begin
    Hint := 'Rotate joint';
    Description := 'Rotates a joint. Works on the skeleton or within a given frame of an animation.';

    // we store information for Undo purposes up front
    for i := 0 to SelectedSkeletons.Count-1 do begin
        skeleton := SelectedSkeletons[i];
        frame := skeleton.GetActiveFrame;

        skeletonUndoItem := TSkeletonUndoRotateItem.Create;
        skeletonUndoItem.Skeleton := skeleton;
        skeletonUndoItem.Frame := frame;
        skeletonUndoItem.WasKeyFrameBeforeExecute := frame.IsKeyFrame;

        for j := 0 to skeleton.SelectedJoints.Count-1 do begin
            skeletonUndoItem.Joints.Add(skeleton.SelectedJoints[j]);
        end;

        fSkeletonUndoList.Add(skeletonUndoItem);
    end;
end;

procedure TJointLocationRotatorTool.DoExecute(AParamList: TParamList);
var i: integer;
    deltaVector: TVertice;
    skeleton: TSkeleton;
    frame: TAnimationFrame;
    joint: TJoint;
    jointLocation: TJointLocation;
    skeletonUndoItem: TSkeletonUndoRotateItem;
begin
    deltaVector := TVertice(AParamList[0]);

    for i := 0 to SelectedSkeletons.Count-1 do begin
        skeleton := SelectedSkeletons[i];

        if skeleton.SelectedJoints.Count=1 then begin
            frame := skeleton.GetActiveFrame;
            joint := skeleton.SelectedJoints[0];
            jointLocation := frame.GetJointLocationByID(joint.ID);

            skeletonUndoItem := TSkeletonUndoRotateItem(fSkeletonUndoList[i]);
            skeletonUndoItem.DeltaVector.Add(deltaVector);

            // rotate differently based on whether the frame
            // is a baseframe or an animationframe
            if frame.IsBaseFrame then
                RotateJointLocationForBaseFrame(joint, jointLocation, deltaVector)
            else begin
                RotateJointLocationForAnimationFrame(jointLocation, deltaVector);

                // jointlocation changed by user; this frame is now a keyframe
                frame.IsKeyFrame := true;

                // animation has changed
                if Assigned(frame.Animation) then
                    TAnimation(frame.Animation).HasChanged := true;
            end;
        end;
    end;

    UpdateJointGeometry;
end;

// ### TSkeletonUndoRotateItem #################################################

constructor TSkeletonUndoRotateItem.Create;
begin
    fDeltaVector := TVertice.Create(0,0,0);
    inherited Create;
end;

destructor TSkeletonUndoRotateItem.Destroy;
begin
    FreeAndNil(fDeltaVector);
    inherited Destroy;
end;

end.
 
