{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_JointRemoverTool;

{$MODE Delphi}

interface

uses
  unit_SkeletonTool,
  unit_ParamList,
  unit_Joint,
  unit_Skeleton;

type
  TJointRemoverTool = class(TSkeletonTool)
  private
    function CreateRemovalList(ASkeleton: TSkeleton): TJointList;
    procedure RemoveJoint(ASkeleton: TSkeleton; AJoint: TJoint);
    procedure UnassignVertices(AJoint: TJoint);
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  SysUtils,
  unit_Types,
  unit_Enums,
  unit_PrimitiveList,
  unit_Animation,
  unit_SkeletonRemoverTool, unit_BaseTool;

// ### TJointRemoverTool #######################################################

// Joints are stored in a tree structure. A child of a selected joint might be
// selected too. But when removing joints, we remove all childs as well.
// Therefor, to avoid removing childs that were already removed when we removed
// their parents, we need to construct a list of joints to remove which only
// contains "parent" joints. In other words: if no parents of a joint are selected,
// we can savely add it to the list of joints to be removed
function TJointRemoverTool.CreateRemovalList(ASkeleton: TSkeleton): TJointList;
var i: integer;
    jointList: TJointList;
    selectedJoint,parentJoint: TJoint;
    isSelected: boolean;
begin
    jointList := TJointList.Create(false,true);

    for i := 0 to ASkeleton.SelectedJoints.Count-1 do begin
        selectedJoint := ASkeleton.SelectedJoints[i];
        parentJoint := selectedJoint.ParentJoint;
        isSelected := false;
        // walk upwards within the tree, searching for selected parent joints
        while Assigned(parentJoint) and (not isSelected) do begin
            isSelected := ASkeleton.SelectedJoints.IsInList(parentJoint);
            parentJoint := parentJoint.ParentJoint;
        end;
        // if no parent joint was selected, add the current selected joint to the list
        if not isSelected then
            jointList.Add(selectedJoint);
    end;

    Result := jointList;
end;

procedure TJointRemoverTool.RemoveJoint(ASkeleton: TSkeleton; AJoint: TJoint);
var i: integer;
begin
    // recursively remove joints and bones from the skeleton
    // also remove jointlocations of the baseframe (and animations?)
    // unassign any geometry that was assigned to the joint
    for i := 0 to AJoint.ChildJoints.Count-1 do
        RemoveJoint(ASkeleton,AJoint.ChildJoints[i]);

    AJoint.ChildBones.Clear;
    AJoint.ChildJoints.Clear;
    ASkeleton.Joints.Remove(AJoint);
    ASkeleton.SelectedJoints.Remove(AJoint);
end;

procedure TJointRemoverTool.UnassignVertices(AJoint: TJoint);
var i: integer;
begin
    for i := 0 to AJoint.AssignedVertices.Count-1 do
        AJoint.AssignedVertices[i].JointID := -1;
end;

procedure TJointRemoverTool.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Remove joint';
    Description := 'Removes a joint and its childjoints from a skeleton.';
end;

procedure TJointRemoverTool.DoExecute(AParamList: TParamList);
var i,j,k,l,jointID,boneIndex: integer;
    skeleton: TSkeleton;
    jointList: TJointList;
    animation: TAnimation;
    parentJoint: TJoint;
    skeletonsToRemove: TSkeletonList;
begin
    skeletonsToRemove := TSkeletonList.Create(false,true);
    for i := 0 to SelectedSkeletons.Count-1 do begin
        skeleton := SelectedSkeletons[i];
        jointList := CreateRemovalList(skeleton);
        for j := 0 to jointList.Count-1 do begin
            jointID := jointList[j].ID;
            // if no parent joint exists, the current joint is the rootjoint of
            // the skeleton; this means we will remove the skeleton as a whole
            parentJoint := jointList[j].ParentJoint;
            if not Assigned(parentJoint) then
                // we will remove the rootjoint; thus, we remove the entire skeleton
                skeletonsToRemove.Add(skeleton)
            else begin
                // if a parentjoint exists, remove bone from parentjoint to the current joint
                // and remove the childjoint from that parentjoint because the child
                // will not exist anymore; NOTE:  indices of joints and bones are the same
                boneIndex := parentJoint.ChildJoints.IndexOf(jointList[j]);
                parentJoint.ChildJoints.Remove(jointList[j]);
                parentJoint.ChildBones.Remove(parentJoint.ChildBones[boneIndex]);

                // unassign vertices from the joint
                UnassignVertices(jointList[j]);
                // recursively remove joints and bones from the skeleton
                RemoveJoint(skeleton,jointList[j]);
                // remove corresponding jointlocations from the baseframe
                skeleton.BaseFrame.RemoveJointLocationByID(jointID);
                // remove jointlocations from all frames of all animations
                for k := 0 to skeleton.Animations.Count-1 do begin
                    animation := skeleton.Animations[k];
                    for l := 0 to animation.AnimationFrames.Count-1 do
                        animation.AnimationFrames[l].RemoveJointLocationByID(jointID);
                    // TODO remove frames that don't have any jointlocations anymore?
                    // Also remove animations that don't have any frames anymore?
                    // We might not need to do this because if we delete all joints
                    // (by deleting the rootjoint), the entire skeleton will be
                    // removed from the scene anyway.
                end;
                // make the parent joint the selected joint
                skeleton.SelectedJoints.Clear;
                skeleton.SelectedJoints.Add(parentJoint);
            end;
        end;
        FreeAndNil(jointList);
    end;

    // use a TSkeletonRemoverFunction to remove entire skeletons from the scene;

    // NOTE: we need to store removed skeletons in a buffer local to this tool
    // It might be a good idea to just copy the ENTIRE skeleton to a buffer and
    // copy it back to the scene on Undo because removing a joint seriously affects
    // various parts of the skeleton (geometry, joints, frames...).

    while skeletonsToRemove.Count>0 do begin
        TSkeletonRemoverFunction.Execute(Skeletons,skeletonsToRemove[0],TPrimitiveList(AParamList.NamedParameters['primitives']));
        SelectedSkeletons.Remove(skeletonsToRemove[0]);
        skeletonsToRemove.Delete(0);
    end;

    FreeAndNil(skeletonsToRemove);
end;


end.
