{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_JointSelector;

{$MODE Delphi}

interface

uses
  Classes,
  unit_Selector,
  unit_Types,
  unit_Primitives,
  unit_Skeleton,
  unit_Joint;

type
  TJointSelector = class(TSelector)
  private
    fSkeletons: TSkeletonList;                  // usually fed with the Skeletons of the scene
    fSelectedSkeletons: TSkeletonList;          // usually fed with the SelectedSkeletons of the scene; contains the selected skeletons based on selected joints
    fSkeletonSelectionIndex: integer;
    procedure Clear;
    procedure ClearSkeletonsWithNoSelectedJoints;
    function VertexInJoint(AVertex: TVertice): TJoint;
    function SelectJointByRay(const ARay: TRay; out ASkeleton: TSkeleton; out ADistance: single): TJoint;
  public
    constructor Create(ASkeletonList,ASelectedSkeletonList: TSkeletonList);
    destructor Destroy; override;

    procedure SelectByRange(AUpperLeftVertex, ALowerRightVertex: TVertice; AClearSelection: boolean = false);
    procedure SelectByVertex(AVertex: TVertice; ARange: double; AClearSelection: boolean = false);
    procedure SelectByRay(ARay: TRay; AClearSelection: boolean = false);
    procedure SelectByRectangle(const ax1,ay1,ax2,ay2: integer;
                                const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                const aViewport: TViewport;
                                const AClearSelection: boolean);
    procedure SelectAll;
    procedure DeselectAll;
    procedure ReverseSelection;
    procedure GrowSelection;
    procedure ShrinkSelection;

    property Skeletons: TSkeletonList read fSkeletons write fSkeletons;
    property SelectedSkeletons: TSkeletonList read fSelectedSkeletons write fSelectedSkeletons;
  end;

implementation

uses SysUtils;

// ### TJointSelector ##########################################################

constructor TJointSelector.Create(ASkeletonList,ASelectedSkeletonList: TSkeletonList);
begin
    fSkeletons := ASkeletonList;
    fSelectedSkeletons := ASelectedSkeletonList;
end;

destructor TJointSelector.Destroy;
begin
//
end;

procedure TJointSelector.Clear;
var i: integer;
begin
    for i := 0 to SelectedSkeletons.Count-1 do
        SelectedSkeletons[i].SelectedJoints.Clear;
    // Note JCC: we cannot clear the SelectedSkeletons as the SkeletonAnimationEditor
    // uses the same collection. If we clear it here, the SkeletonAnimationEditor will
    // loose its animation with which it works when we try to select a joint from any
    // given animationframe...
    // The SkeletonAnimationEditor should probably just work with an Animation property...
    // As a "temporary" solution, we use the ClearSkeletonsWithNoSelectedJoints method
    // to remove skeletons from the SelectedSkeletons collection that don't have any
    // joints selected. This makes sure the selection process is intuitive while also
    // keeping the selected animation active in the SkeletonAnimationEditor when
    // selecting joints. Not pretty, but it works.
    //SelectedSkeletons.Clear;
end;

procedure TJointSelector.ClearSkeletonsWithNoSelectedJoints;
var counter: integer;
    skeleton: TSkeleton;
begin
    counter := 0;

    while (counter < SelectedSkeletons.Count) do begin
        skeleton := SelectedSkeletons[counter];
        if skeleton.SelectedJoints.Count=0 then
            SelectedSkeletons.Remove(skeleton)
        else
            inc(counter);
    end;
end;

procedure TJointSelector.SelectByRange(AUpperLeftVertex, ALowerRightVertex: TVertice; AClearSelection: boolean);
var i,j: integer;
    skeleton: TSkeleton;
    joint: TJoint;
begin
    if AClearSelection then
        Clear;

    for i := 0 to Skeletons.Count-1 do begin
        skeleton := Skeletons[i];
        for j := 0 to skeleton.Joints.Count-1 do begin
            joint := skeleton.Joints[j];
            if joint.Object_SelectByRange(AUpperLeftVertex,ALowerRightVertex) then begin
                if skeleton.SelectedJoints.IsInList(joint) then
                    skeleton.SelectedJoints.Remove(joint)
                else
                    skeleton.SelectedJoints.Add(joint);

                if skeleton.SelectedJoints.Count=0 then
                    SelectedSkeletons.Remove(skeleton)
                else
                    SelectedSkeletons.Add(skeleton);
            end;
        end;
    end;

    ClearSkeletonsWithNoSelectedJoints;

    if Assigned(OnSelect) then
        OnSelect(SelectedSkeletons);
end;

function TJointSelector.VertexInJoint(AVertex: TVertice): TJoint;
var skeletonCounter, jointCounter: integer;
    skeleton: TSkeleton;
    joint: TJoint;
    polygon: TPolygon;
    isFound: boolean;
begin
    joint := nil;
    isFound := false;

    if Skeletons.Count>0 then begin
        skeletonCounter := 0;

        // loop through the available skeletons, searching for a joint that
        // has been clicked on by the user
        // this means we have two loops: one for the available skeletons
        // and one for the joints within a skeleton
        while (not isFound) and (skeletonCounter < Skeletons.Count) do begin

            fSkeletonSelectionIndex := (fSkeletonSelectionIndex+1) mod Skeletons.Count;
            skeleton := Skeletons[fSkeletonSelectionIndex];

            jointCounter := 0;

            while (not isFound) and (jointCounter<skeleton.Joints.Count) do begin
                skeleton.JointSelectionIndex := (skeleton.JointSelectionIndex+1) mod skeleton.Joints.Count;
                joint := skeleton.Joints[skeleton.JointSelectionIndex];
                isFound := joint.IsPointInObject(AVertex,polygon);
                inc(jointCounter);
            end;

            inc(skeletonCounter);
        end;
    end;

    if isFound then
        Result := joint
    else
        Result := nil;
end;

procedure TJointSelector.SelectByVertex(AVertex: TVertice; ARange: double; AClearSelection: boolean);
var joint: TJoint;
    skeleton: TSkeleton;
begin
    if AClearSelection then
        Clear;

    joint := VertexInJoint(AVertex);
    if not Assigned(joint) then
        SelectedSkeletons.Clear
    else begin
        skeleton := TSkeleton(joint.Skeleton);
        if skeleton.SelectedJoints.IsInList(joint) then
            skeleton.SelectedJoints.Remove(joint)
        else
            skeleton.SelectedJoints.Add(joint);
        SelectedSkeletons.Add(skeleton);
    end;

    ClearSkeletonsWithNoSelectedJoints;

    if Assigned(OnSelect) then
        OnSelect(SelectedSkeletons);
end;

function TJointSelector.SelectJointByRay(const ARay: TRay; out ASkeleton: TSkeleton; out ADistance: single): TJoint;
var i,j: integer;
    skeleton: TSkeleton;
    joint: TJoint;
    distance: single;
    polygon: TPolygon;
begin
    ADistance := -1;
    ASkeleton := nil;
    Result := nil;

    for i := 0 to Skeletons.Count-1 do begin
        skeleton := Skeletons[i];

        for j := 0 to skeleton.Joints.Count-1 do begin
            joint := skeleton.Joints[j];
            if joint.Polygon_Select(ARay, distance, polygon) and  ((ADistance = -1) or (distance < ADistance)) then begin
                Result := joint;
                ADistance := distance;
                ASkeleton := skeleton;
            end;
        end;
    end;
end;

procedure TJointSelector.SelectByRay(ARay: TRay; AClearSelection: boolean);
var joint: TJoint;
    skeleton: TSkeleton;
    distance: single;
begin
    if AClearSelection then
        Clear;

    joint := SelectJointByRay(ARay, skeleton, distance);
    if not Assigned(joint) then
        SelectedSkeletons.Clear
    else begin
        if skeleton.SelectedJoints.IsInList(joint) then
            skeleton.SelectedJoints.Remove(joint)
        else begin
            skeleton.SelectedJoints.Add(joint);
            SelectedSkeletons.Add(skeleton);
        end;
    end;

    ClearSkeletonsWithNoSelectedJoints;
end;

procedure TJointSelector.SelectByRectangle(const ax1,ay1,ax2,ay2: integer;
                                           const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                           const aViewport: TViewport;
                                           const AClearSelection: boolean);
var
    iSkeleton  : integer;
    iJoint     : integer;
    iVertice   : integer;
    vertice    : TVertice;
    point      : TPoint2D;
    skeleton   : TSkeleton;
    joint      : TJoint;
    JointInRect: boolean;
begin
    if AClearSelection then
        Clear;

    for iSkeleton := 0 to Skeletons.Count-1 do begin
        skeleton := Skeletons[iSkeleton];
        for iJoint := 0 to skeleton.Joints.Count-1 do begin
            joint := skeleton.Joints[iJoint];
            JointInRect := true;

            for iVertice := 0 to joint.NumVertices - 1 do begin
                vertice := joint.Vertices[iVertice];
                point := vertice.ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewPort);
                if (point.X < ax1) or (point.X > ax2) or (point.Y < ay1) or (point.Y > ay2) then begin
                    JointInRect := false;
                    Break;
                end;
            end;

            if JointInRect then begin
                if skeleton.SelectedJoints.IsInList(joint) then
                    skeleton.SelectedJoints.Remove(joint)
                else
                    skeleton.SelectedJoints.Add(joint);

                if skeleton.SelectedJoints.Count = 0 then
                    SelectedSkeletons.Remove(skeleton)
                else
                    SelectedSkeletons.Add(skeleton);
            end;
        end;
    end;

    ClearSkeletonsWithNoSelectedJoints;

    if Assigned(OnSelect) then
        OnSelect(SelectedSkeletons);
end;

procedure TJointSelector.SelectAll;
begin
    raise Exception.Create('TJointSelector.SelectAll not implemented.');
end;

procedure TJointSelector.DeselectAll;
begin
    // deselect all joints from selected skeletons;
    // then deselect all selected skeletons; this will
    // deselect all selected animations too
    Clear;
    SelectedSkeletons.Clear;
end;

procedure TJointSelector.ReverseSelection;
begin
    raise Exception.Create('TJointSelector.ReverseSelection not implemented.');
end;

procedure TJointSelector.GrowSelection;
begin
    raise Exception.Create('TJointSelector.GrowSelection not implemented.');
end;

procedure TJointSelector.ShrinkSelection;
begin
    raise Exception.Create('TJointSelector.ShrinkSelection not implemented.');
end;

end.
