{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_JointSplitterTool;

{$MODE Delphi}

interface

uses
  unit_SkeletonTool,
  unit_ParamList,
  unit_Joint,
  unit_Skeleton,
  unit_Enums,
  unit_SkeletonUndoItem,
  unit_Types,
  unit_AnimationFrame,
  unit_JointLocation,
  unit_MaterialManager;

type
  TJointSplitterTool = class(TSkeletonTool)
  private
    fMaterialManager: TMaterialManager;
    fSkeleton: TSkeleton;
    fParentJoint: TJoint;
    fChildJoint: TJoint;
    fNewJoint: TJoint;

    procedure SplitJoints(AMaterialManager: TMaterialManager; ASkeleton: TSkeleton; AParentJoint, AChildJoint: TJoint);
    procedure UndoSplitJoints;
  protected
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  public
    constructor Create(AToolAction: TToolAction);
    destructor Destroy; override;
    function CanExecute(AParamList: TParamList): boolean; override;
    procedure Undo; override;
    procedure Redo; override;
  end;

implementation

uses
  SysUtils,
  Dialogs,
  unit_Animation,
  unit_BaseTool,
  unit_Math,
  unit_Bone,
  unit_JointBuilder,
  unit_BoneBuilder;

// ### TJointSplitterTool #######################################################

procedure TJointSplitterTool.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Split joints';
    Description := 'Creates a new joint between two selected joints that have a parent/child relation.';
end;

procedure TJointSplitterTool.SplitJoints(AMaterialManager: TMaterialManager; ASkeleton: TSkeleton; AParentJoint, AChildJoint: TJoint);
var childIndex, i, j: integer;
    parentJointLocation, childJointLocation, newJointLocation: TJointLocation;
    position: TVertice;
    jointBuilder: TJointBuilder;
    paramList: TParamList;
    boneBuilder: TBoneBuilder;
    newBone: TBone;
    animation: TAnimation;
    animationFrame: TAnimationFrame;
begin
    // get jointlocations of the baseframe
    parentJointLocation := ASkeleton.BaseFrame.GetJointLocationByID(AParentJoint.ID);
    childJointLocation := ASkeleton.BaseFrame.GetJointLocationByID(AChildJoint.ID);

    // calculate the position of the new joint
    position := TVertice.Create;
    position.x := (parentJointLocation.Position.x+childJointLocation.Position.x) / 2;
    position.y := (parentJointLocation.Position.y+childJointLocation.Position.y) / 2;
    position.z := (parentJointLocation.Position.z+childJointLocation.Position.z) / 2;

    // remove the bone between parentJoint and childJoint
    childIndex := AParentJoint.ChildJoints.IndexOf(AChildJoint);
    AParentJoint.ChildBones.Remove(AParentJoint.ChildBones[childIndex]);

    // remove the parent/child relation between the joints
    AParentJoint.ChildJoints.DeleteJointByID(AChildJoint.ID);   // note: Delete does not free the object (remove does)
    AChildJoint.ParentJoint := nil;

    // create a new joint, add it to the skeleton and restore parent/child relations
    jointBuilder := TJointBuilder.Create;
    paramList := TParamList.Create(false);
    paramList.Add(position, 'position');
    paramList.Add(AMaterialManager, 'materialManager');
    paramList.Add(ASkeleton.JointSize, 'jointSize');
    fNewJoint := jointBuilder.Build(paramlist);
    FreeAndNil(paramlist);
    FreeAndNil(jointBuilder);

    ASkeleton.Joints.AddJoint(fNewJoint);    // fNewJoint.ID is set here
    fNewJoint.Skeleton := ASkeleton;

    fNewJoint.ParentJoint := AParentJoint;
    AParentJoint.ChildJoints.Add(fNewJoint);
    AChildJoint.ParentJoint := fNewJoint;
    fNewJoint.ChildJoints.Add(AChildJoint);

    // remove the parent/child relation between the jointlocations of the baseframe
    parentJointLocation.ChildLocations.DeleteByJointLocationID(AChildJoint.ID);
    childJointLocation.ParentLocation := nil;

    // create a new jointlocation and restore parent/child relations
    newJointLocation := TJointLocation.Create;
    newJointLocation.ID := fNewJoint.ID;
    newJointLocation.Position.Assign(position);

    ASkeleton.BaseFrame.AddJointLocation(newJointLocation,parentJointLocation);
    newJointLocation.ChildLocations.Add(childJointLocation);
    childJointLocation.ParentLocation := newJointLocation;

    // do the same for the animationframes of all animations of the skeleton
    for i := 0 to ASkeleton.Animations.Count-1 do begin
        animation := ASkeleton.Animations[i];
        for j := 0 to animation.AnimationFrames.Count-1 do begin
            animationFrame := animation.AnimationFrames[j];

            parentJointLocation := animationFrame.GetJointLocationByID(AParentJoint.ID);
            childJointLocation := animationFrame.GetJointLocationByID(AChildJoint.ID);

            position.x := (parentJointLocation.Position.x+childJointLocation.Position.x) / 2;
            position.y := (parentJointLocation.Position.y+childJointLocation.Position.y) / 2;
            position.z := (parentJointLocation.Position.z+childJointLocation.Position.z) / 2;

            // remove the parent/child relation between the jointlocations of the animationframe
            parentJointLocation.ChildLocations.DeleteByJointLocationID(AChildJoint.ID);
            childJointLocation.ParentLocation := nil;

            // create a new jointlocation and restore parent/child relations
            newJointLocation := TJointLocation.Create;
            newJointLocation.ID := fNewJoint.ID;
            newJointLocation.Position.Assign(position);

            animationFrame.AddJointLocation(newJointLocation,parentJointLocation);
            newJointLocation.ChildLocations.Add(childJointLocation);
            childJointLocation.ParentLocation := newJointLocation;
        end;
    end;

    // create a bone from parentJoint to new joint
    paramList := TParamList.Create(false);
    paramList.Add(ASkeleton.BaseFrame.GetJointLocationByID(AParentJoint.ID), 'jointLocation1');
    paramList.Add(ASkeleton.BaseFrame.GetJointLocationByID(fNewJoint.ID), 'jointLocation2');
    paramList.Add(AMaterialManager, 'materialManager');
    paramList.Add(ASkeleton.JointSize, 'boneSize');                             // NOTE, JCC: bonesize = jointsize
    boneBuilder := TBoneBuilder.Create;
    newBone := boneBuilder.Build(paramList);
    FreeAndNil(boneBuilder);
    FreeAndNil(paramList);
    AParentJoint.ChildBones.Add(newBone);

    // create a bone from new joint to childJoint
    paramList := TParamList.Create(false);
    paramList.Add(ASkeleton.BaseFrame.GetJointLocationByID(fNewJoint.ID), 'jointLocation1');
    paramList.Add(ASkeleton.BaseFrame.GetJointLocationByID(AChildJoint.ID), 'jointLocation2');
    paramList.Add(AMaterialManager, 'materialManager');
    paramList.Add(ASkeleton.JointSize, 'boneSize');                             // NOTE, JCC: bonesize = jointsize
    boneBuilder := TBoneBuilder.Create;
    newBone := boneBuilder.Build(paramList);
    FreeAndNil(boneBuilder);
    FreeAndNil(paramList);
    fNewJoint.ChildBones.Add(newBone);

    // free stuff
    FreeAndNil(position);
end;

procedure TJointSplitterTool.UndoSplitJoints;
var childIndex, boneIndex, i, j: integer;
    parentJointLocation, childJointLocation, newJointLocation: TJointLocation;
    animation: TAnimation;
    animationFrame: TAnimationFrame;
    paramList: TParamList;
    boneBuilder: TBoneBuilder;
    newBone: TBone;
begin
    // remove bone from parent to new joint
    childIndex := fParentJoint.ChildJoints.IndexOf(fNewJoint);
    fParentJoint.ChildBones.Remove(fParentJoint.ChildBones[childIndex]);

    // remove bone from new joint to child (not really needed as fNewJoint is freed later anyway)
    childIndex := fNewJoint.ChildJoints.IndexOf(fChildJoint);
    fNewJoint.ChildBones.Remove(fNewJoint.ChildBones[childIndex]);

    // remove the jointlocation of the new joint from the baseframe
    parentJointLocation := fSkeleton.BaseFrame.GetJointLocationByID(fParentJoint.ID);
    newJointLocation := fSkeleton.BaseFrame.GetJointLocationByID(fNewJoint.ID);
    childJointLocation := fSkeleton.BaseFrame.GetJointLocationByID(fChildJoint.ID);

    parentJointLocation.ChildLocations.DeleteByJointLocationID(newJointLocation.ID);

    // restore the parent/child relation of jointlocations in the baseframe
    parentJointLocation.ChildLocations.Add(childJointLocation);
    childJointLocation.ParentLocation := parentJointLocation;

    // do the same for all animationframes
    for i := 0 to fSkeleton.Animations.Count-1 do begin
        animation := fSkeleton.Animations[i];
        for j := 0 to animation.AnimationFrames.Count-1 do begin
            animationFrame := animation.AnimationFrames[j];

            parentJointLocation := animationFrame.GetJointLocationByID(fParentJoint.ID);
            newJointLocation := animationFrame.GetJointLocationByID(fNewJoint.ID);
            childJointLocation := animationFrame.GetJointLocationByID(fChildJoint.ID);

            // remove the jointlocation of the new joint from the animationframe
            parentJointLocation.ChildLocations.DeleteByJointLocationID(newJointLocation.ID);

            // restore the parent/child relation of jointlocations in the animationframe
            parentJointLocation.ChildLocations.Add(childJointLocation);
            childJointLocation.ParentLocation := parentJointLocation;
        end;
    end;

    // remove the joint from the childJoints of the parentjoint
    fParentJoint.ChildJoints.RemoveJointByID(fNewJoint.ID);

    // now remove the joint from the skeleton
    fSkeleton.SelectedJoints.DeleteJointByID(fNewJoint.ID);
    fSkeleton.Joints.RemoveJointByID(fNewJoint.ID);

    // restore the parent/child relation between the joints
    fParentJoint.ChildJoints.Add(fChildJoint);  // use Add instead of AddJoint to preserve the original childJoint.ID value
    fChildJoint.ParentJoint := fParentJoint;

    // create bone from parent to child
    paramList := TParamList.Create(false);
    paramList.Add(fSkeleton.BaseFrame.GetJointLocationByID(fParentJoint.ID), 'jointLocation1');
    paramList.Add(fSkeleton.BaseFrame.GetJointLocationByID(fChildJoint.ID), 'jointLocation2');
    paramList.Add(fMaterialManager, 'materialManager');
    paramList.Add(fSkeleton.JointSize, 'boneSize');                             // NOTE, JCC: bonesize = jointsize
    boneBuilder := TBoneBuilder.Create;
    newBone := boneBuilder.Build(paramList);
    FreeAndNil(boneBuilder);
    FreeAndNil(paramList);
    fParentJoint.ChildBones.Add(newBone);

    // make sure the bone has the same index as the childjoint
    childIndex := fParentJoint.ChildJoints.IndexOf(fChildJoint);
    boneIndex := fParentJoint.ChildBones.IndexOf(newBone);
    fParentJoint.ChildBones.Move(boneIndex,childIndex);

    // set to nil
    fNewJoint := nil;
end;

procedure TJointSplitterTool.DoExecute(AParamList: TParamList);
var firstSelectedJoint, secondSelectedJoint: TJoint;
begin
    // IMPORTANT NOTE: this tool stores the items worked with in private variables
    // for Undo purposes later. This assumes these objects still exist when Undo
    // is called.
    //
    // If in a chain of Undo calls one of these items is destroyed, the Undo for
    // this tool won't work. We might want to check on item existance before calling
    // Undo and/or implement a methology that guarantees item existance in any
    // Undo chain.  
    fMaterialManager := TMaterialManager(AParamList[0]);
    fSkeleton := SelectedSkeletons[0];
    firstSelectedJoint := fSkeleton.SelectedJoints[0];
    secondSelectedJoint := fSkeleton.SelectedJoints[1];

    if firstSelectedJoint.ParentJoint = secondSelectedJoint then begin
        fParentJoint := secondSelectedJoint;
        fChildJoint := firstSelectedJoint;
    end else begin
        fParentJoint := firstSelectedJoint;
        fChildJoint := secondSelectedJoint;
    end;

    SplitJoints(fMaterialManager, fSkeleton, fParentJoint, fChildJoint);
end;

function TJointSplitterTool.CanExecute(AParamList: TParamList): boolean;
var canExecuteTool: boolean;
    selectedSkeleton: TSkeleton;
    firstSelectedJoint, secondSelectedJoint: TJoint;
begin
    canExecuteTool := TRUE;

    if SelectedSkeletons.Count = 1 then begin
         selectedSkeleton := SelectedSkeletons[0];
         if selectedSkeleton.SelectedJoints.Count <> 2 then begin
            canExecuteTool := FALSE;
            ErrorMessage := 'Please select two joints with a parent/child relation.';
         end else begin
            firstSelectedJoint := selectedSkeleton.SelectedJoints[0];
            secondSelectedJoint := selectedSkeleton.SelectedJoints[1];

            canExecuteTool := ((firstSelectedJoint.ParentJoint = secondSelectedJoint) or (secondSelectedJoint.ParentJoint = firstSelectedJoint));
            if not canExecuteTool then
                ErrorMessage := 'Please select two joints with a parent/child relation.';
         end;

    end else if SelectedSkeletons.Count > 1 then begin
        canExecuteTool := FALSE;
        ErrorMessage := 'Cannot split joints for multiple selected skeletons. Please select one skeleton only.'
    end;

    Result := canExecuteTool;
end;

procedure TJointSplitterTool.Undo;
begin
    UndoSplitJoints;
end;

procedure TJointSplitterTool.Redo;
begin
    SplitJoints(fMaterialManager, fSkeleton, fParentJoint, fChildJoint);
end;

constructor TJointSplitterTool.Create(AToolAction: TToolAction);
begin
    inherited Create(AToolAction);
end;

destructor TJointSplitterTool.Destroy;
begin
    inherited Destroy;
end;

end.
