{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_LightMapRenderSystem;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  The actual lightmap rendersystem. Requires vertex normals (polygon.Normals) to
  be calculated, and all polygons with AllowLightmap on to be non-degenerate.

  NOTE, PJ: The rendersystem publishes the 'faceinfo' metadata structure,
            so an outside process can assign the new materials to the polygons
            in a single pass.

  TODO, PJ:
  - Intersessant, als je in de normal smoother de 'triangle misses' op wit zet
    bij het renderen van een sphere, dan wordt er _erg veel_ wit. Je zou bijna
    zeggen dat sommige faces niet op het juiste plane worden gemapt.
  - Gile[s] has a much much much better mapping scheme... Figure out how. Probably enough to skip the whole 'adjacent polygon' checking thing...
  - Gile[s] has a much much much better packing scheme... Figure out how.
  ---------------------------------------------------------------------------- }


interface

uses
  Windows, ContNrs, Classes,
  unit_Types,
  unit_Enums,
  unit_Textures,
  unit_Octree,
  unit_Primitives,
  unit_Lights,
  unit_TextureLoader,
  unit_Scene,
  unit_PrimitiveList,
  unit_Materials,
  unit_TextureTypes;

type
  TDWordArray = array[0..0] of DWORD;

  TTriplet = Array[0..2] of Byte;
  PTriplet = ^TTriplet;

  TTripletArray = Array[0..MAXINT div 4] of TTriplet;
  PTripletArray = ^TTripletArray;

  TTripleIndices = array[0..2] of Integer;

  TProgressHandler = procedure( const aPercentageDone: integer ) of object;

  TFaceInfo = record
      Vertices: array[0..2] of TVertex;
      UVCoordinates: array[0..2] of TUVCoordinate;
      AdjacentPolygons: TPolygonList;
      UVVertex, Edge1, Edge2: TVertex;
      XResolution, YResolution: integer;
      Polygon: TPolygon;
      CastShadows: Boolean;
      ReceiveShadows: Boolean;
      LightMap: TTextureData;
      LightMapIndex: integer; // For use outside the rendersystem, quick TTextureData->TTexture lookups.
      TransparentTextureIndex: integer; // Only used for transparent surface, for quick lookup in texture cache
      XOffset, YOffset: integer;
      LastRayIndex: integer;
      Alignment: TPlaneAlignment;
  end;
  PFaceInfo = ^TFaceInfo;
  TFaceInfoArray = array of TFaceInfo;

// ============================================================================

  TPlanarMappedTextureInfo = record
      MinU, MaxU, DeltaU: single;
      MinV, MaxV, DeltaV: single;
      Alignment: TPlaneAlignment;
  end;

// ============================================================================

  TClippingCode = ( coClipRight, coClipLeft, coClipTop, coClipBottom, coClipFront, coClipBack );
  TClippingOutcode = Set of TClippingCode;

// ============================================================================

  // Convenient structure used for lightmap packing
  T2DBitmapTreeNode = class( TObject )
  private
      fChildNodes: Array[0..1] of T2DBitmapTreeNode;
      fRectangle: TRect;
      fUsed: boolean;

  public
      constructor Create( const aLeft, aTop, aRight, aBottom: integer );
      destructor Destroy; override;

      function Insert( const aWidth, aHeight: integer ): T2DBitmapTreeNode;
      function HasChildren: boolean;

      property Rectangle: TRect read fRectangle;
  end;

// ============================================================================

  // Record and class for maintaining a list of triangle intersections
  TTriangleIntersection = record
      Face: TFaceInfo;
      Distance: single;
      U, V: single;
  end;
  PTriangleIntersection = ^TTriangleIntersection;

  TTriangleIntersections = class(TList)
  protected
      procedure Notify(aPointer: Pointer; anAction: TListNotification); override;
      function GetTriangleIntersection(const anIndex: integer): TTriangleIntersection;

  public
      property Items[const anIndex: integer]: TTriangleIntersection read GetTriangleIntersection; default;

      procedure Add(const aTriangleIntersection: TTriangleIntersection); overload;
      procedure Add(const aFace: TFaceInfo; const aDistance, aU, aV: single); overload;
      procedure Sort;

      function ContainsTransparentTriangles: boolean;
  end;

// ============================================================================

  TLightmapRenderSystem = class( TObject )
  private
      fScene: TScene;
      fSelectedPrimitivesOnly: boolean;
      fSelectedPolygonsOnly: boolean;
      fConsiderSelectedOnly: boolean;

      fLightmaps: Array of TTextureData;
      fTransparentTextures: Array of TTextureData;
      fBitmapTreeNode: T2DBitmapTreeNode;

      fLights: TObjectList;

      fMinimumTextureSize: integer;
      fMaximumTextureSize: integer;
      fLightMapSize: integer;
      fNrBlurFilterPasses: integer;

      fProcessingFaceIndex: integer;

      fFaces: TFaceInfoArray;
      fShadowCasters: TFaceInfoArray;
      fShadowCasterIndex: integer;
      fOctree: TOctree;
      fRayIndex: integer;
      fFaceIndex: integer;

      fNrLightmappedFaces: integer;
      fProgressHandler: TProgressHandler;

      fContainsTransparentShadowCasters: boolean;

      fShouldGenerateShadows: boolean;
      fShouldUseAmbientColour: boolean;
      fShadowOpacity: single;

      fEpsilon: single;

      fStopped: boolean;
      fInstructionCounter: integer;

      procedure RegisterFaces;
      procedure RegisterFacesForPrimitive( const aPrimitive: TPrimitive );
      procedure RegisterLights;
      procedure RegisterLight( const aLight: TLight );
      procedure RegisterShadowCasters;
      procedure RegisterShadowCastersForPolygon(const aPolygon: TPolygon);
      procedure UpdateShadowCasterTransparency;
      procedure LoadTransparentTextures(const aTransparentTextureNames: TStringList);
      function SignificantIndicesForPolygon(const aPolygon: TPolygon): TTripleIndices;

      procedure InitializeFaces(const aPrimitiveList: TPrimitiveList);
      procedure UpdateFaces;

      procedure RenderScene;
      procedure RenderFaces;
      procedure RenderFace( const aFace: TFaceInfo );

      function InfluenceForLight( const aPosition, aNormal: TVertex; const aLight: TLight; const aFace: TFaceInfo ): TRGBColor;

      function PlaneAlignmentForNormal( const aNormal: TVertex ): TPlaneAlignment;
      function GeneratePlanarMappedTexture( const aPolygon: TPolygon ): TPlanarMappedTextureInfo;

      function IntersectsTriangle(
          const anOrigin, aDirection: TVertex;
          const aVertex0, aVertex1, aVertex2: TVertex;
          var aT, aU, aV: single ): Boolean;

      procedure FindIntersectedTrianglesInOctreeNode(
          const anOrigin, aDestination, aDirection: TVertex;
          const aNode: TOctreeNode;
          const aShouldStopOnFirstHit: boolean;
          const aTriangleIntersections: TTriangleIntersections );

      procedure FilterColorThroughMaterial( var aColor: TRGBColor; const aFace: TFaceInfo; const aU, aV: single );

      function CalculateSmoothNormal( const aPosition: TVertex; const aFace: TFaceInfo ): TVertex;
//      function CalculateAmbientOcclusion( const aPosition, aNormal: TVertex; const aFace: TFaceInfo ): single;

      function MakeLightMap: TTextureData;
      procedure RenderLightMapTexture( const aFace: TFaceInfo );
      procedure RenderTextureBorder( const anX, aY, aWidth, aHeight: integer; const aTextureData: TTextureData );
      procedure BlurTexture( const anX, aY, aWidth, aHeight: integer; const aTextureData: TTextureData; const aNrBlurs: integer = 1 );

      procedure ResolveOctree;

      procedure HandleTriangleRequest( const anIndex: integer; var aTriangle: TOctreeTriangle );
      procedure HandleProgress( const aPercentageDone: integer );

      procedure DestroyFaces;

      procedure KeepGUIAlive;
  protected
      function GetLightMapForIndex(anIndex: integer): TTextureData;
      function GetFaceForIndex(anIndex: integer): TFaceInfo;

      property Scene: TScene read fScene;
      property SelectedPrimitivesOnly: boolean read fSelectedPrimitivesOnly;
      property SelectedPolygonsOnly: boolean read fSelectedPolygonsOnly;
      property ConsiderSelectedOnly: boolean read fConsiderSelectedOnly;

  public
      constructor Create( const aScene: TScene;
          const aSelectedPrimitivesOnly, aSelectedPolygonsOnly, aConsiderSelectedOnly: boolean );
      destructor Destroy; override;

      // INFO, PJ: Minimum distance for objects to cast a shadow
      property Epsilon: single read fEpsilon write fEpsilon;

      procedure Start;
      procedure Stop;

      function NumLightMaps: Integer;
      function NumFaces: Integer;

      property OnProgress: TProgressHandler read fProgressHandler write fProgressHandler;

      property LightMapSize: integer read fLightMapSize write fLightMapSize;
      property MinimumTextureSize: integer read fMinimumTextureSize write fMinimumTextureSize;
      property MaximumTextureSize: integer read fMaximumTextureSize write fMaximumTextureSize;

      property ShouldGenerateShadows: boolean read fShouldGenerateShadows write fShouldGenerateShadows;
      property ShouldUseAmbientColour: boolean read fShouldUseAmbientColour write fShouldUseAmbientColour;
      property ShadowOpacity: single read fShadowOpacity write fShadowOpacity;
      property NrBlurFilterPasses: integer read fNrBlurFilterPasses write fNrBlurFilterPasses;

      property LightMaps[anIndex: integer]: TTextureData read GetLightMapForIndex;
      property Faces[anIndex: integer]: TFaceInfo read GetFaceForIndex;
  end;

implementation

uses
    SysUtils, Math, Forms, Dialogs,
    unit_Constants,
    unit_Functions,
    unit_Math;

const
    LIGHTMAP_BORDER_SIZE = 1;  // INFO, PJ: Border size to prevent bilinear filtering artefacts

constructor TLightmapRenderSystem.Create( const aScene: TScene; const aSelectedPrimitivesOnly, aSelectedPolygonsOnly, aConsiderSelectedOnly: boolean );
begin
  inherited create;

  fScene := aScene;
  fSelectedPrimitivesOnly := aSelectedPrimitivesOnly;
  fSelectedPolygonsOnly := aSelectedPolygonsOnly;
  fConsiderSelectedOnly := aConsiderSelectedOnly;

  fLights := TObjectList.Create(false);

  SetLength(fFaces, 0);
  SetLength(fShadowCasters, 0);
  SetLength(fLightMaps, 0);

  fMinimumTextureSize := 16;
  fMaximumTextureSize := 32;
  fLightMapSize := 512;
  fShouldGenerateShadows := true;
  fNrBlurFilterPasses := 1;
  fEpsilon := 0.001;
end;

// ============================================================================

destructor TLightmapRenderSystem.Destroy;
begin
  FreeAndNil( fOctree );
  FreeAndNil( fLights );
  FreeAndNil( fBitmapTreeNode );

  fLightMaps := nil;
  DestroyFaces;
  fShadowCasters := nil;

  inherited Destroy;
end;

// ============================================================================

procedure TLightmapRenderSystem.DestroyFaces;
var i: integer;
begin
    if Assigned(fFaces) then begin
        for i := Low(fFaces) to High(fFaces) do
            FreeAndNil(fFaces[i].AdjacentPolygons);
    end;

    fFaces := nil;
end;

// ============================================================================

procedure TLightmapRenderSystem.Start;
begin
  fStopped := false;

  RenderScene;
end;

// ============================================================================

procedure TLightmapRenderSystem.Stop;
begin
  fStopped := true;
end;

// ============================================================================

function TLightmapRenderSystem.NumLightMaps: Integer;
begin
    result := Length( fLightMaps );
end;

// ============================================================================

function TLightmapRenderSystem.NumFaces: Integer;
begin
    result := Length( fFaces );
end;

// ============================================================================

procedure TLightmapRenderSystem.ResolveOctree;
begin
  if not Assigned(fOctree) then
  begin
      fOctree := TOctree.Create( Length(fShadowCasters), 1, 1, 4, handleTriangleRequest );
  end;
end;

// ============================================================================

procedure TLightmapRenderSystem.RenderScene;
begin
  fNrLightmappedFaces := 0;

  RegisterLights;
  RegisterFaces;
  RegisterShadowCasters;

  if fNrLightmappedFaces > 0 then
  begin
      UpdateFaces;

      ResolveOctree;

      fProcessingFaceIndex := 0;
      RenderFaces;
  end;
end;

// ============================================================================

procedure TLightmapRenderSystem.RenderFaces;
var index: integer;
begin
    for index := Low( fFaces ) to high( fFaces ) do begin
        Renderface( fFaces[index] );

        if fStopped then
            exit;
    end;
end;

// ============================================================================

procedure TLightmapRenderSystem.RenderFace( const aFace: TFaceInfo );
var polygon: TPolygon;
    vertexIndex: integer;
    xOffset, yOffset: integer;
    xResolution, yResolution: integer;
    uvCoords: TUVList;

begin
    polygon := aFace.Polygon;
    RenderLightMapTexture( aFace );

    xOffset := fFaces[fProcessingFaceIndex].XOffset + LIGHTMAP_BORDER_SIZE;
    yOffset := fFaces[fProcessingFaceIndex].YOffset + LIGHTMAP_BORDER_SIZE;
    xResolution := fFaces[fProcessingFaceIndex].XResolution - 2 * LIGHTMAP_BORDER_SIZE;
    yResolution := fFaces[fProcessingFaceIndex].YResolution - 2 * LIGHTMAP_BORDER_SIZE;;

    for vertexIndex := 0 to polygon.NumVertices - 1 do
    begin
        uvCoords := polygon.UVCoordinates[LIGHTMAP_TEXTURE_LAYER_INDEX];

        uvCoords[vertexIndex].U := ( uvCoords[vertexIndex].U * xResolution + xOffset ) / LightMapSize;
        uvCoords[vertexIndex].V := ( uvCoords[vertexIndex].V * yResolution + yOffset ) / LightMapSize;
    end;

    if fStopped then
    begin
        exit;
    end;

    HandleProgress( Round( fProcessingFaceIndex * 100 / Length( fFaces )));
    fProcessingFaceIndex := fProcessingFaceIndex + 1;
end;

// ============================================================================

// Separate pass to initializes the faces array to the correct length (efficiency!)
procedure TLightmapRenderSystem.InitializeFaces(const aPrimitiveList: TPrimitiveList);
var primitiveIndex, polygonIndex, numFaces: Integer;
    primitive: TBasePrimitive;
    polygon: TPolygon;
begin
    numFaces := 0;
    
    for primitiveIndex := 0 to aPrimitiveList.Count - 1 do begin
        primitive := aPrimitiveList.Get(primitiveIndex);

        //Make sure we have a real primitive (not a path or light etc..)
        if (primitive is TPrimitive) then
        begin
            for polygonIndex := 0 to primitive.NumPolygons - 1 do begin
                polygon := primitive.Polygons[polygonIndex];

                if polygon.AllowLightmap and (not SelectedPolygonsOnly or primitive.SelectedPolygons.IsInList(polygon)) then
                    numFaces := numFaces + 1;
            end;
        end;
    end;

    SetLength(fFaces, numFaces);
end;

// ============================================================================

procedure TLightmapRenderSystem.RegisterFaces;
var
  list: TPrimitiveList;
  index: integer;
  primitive: TBasePrimitive;

begin
  if SelectedPrimitivesOnly then begin
      list := Scene.SelectedPrimitives;
  end
  else
      list := Scene.Primitives;

  InitializeFaces(list);

  fFaceIndex := 0;
  for index := 0 to list.Count - 1 do begin
      primitive := list.Get( index );

      if (primitive is TPrimitive) then
          RegisterFacesForPrimitive( primitive as TPrimitive );
  end;
end;

// ============================================================================

procedure TLightmapRenderSystem.RegisterLights;
var
  list: TPrimitiveList;
  index: integer;
  primitive: TBasePrimitive;

begin
  if ConsiderSelectedOnly then begin
      list := Scene.SelectedPrimitives;
  end
  else
      list := Scene.Primitives;

  for index := 0 to list.Count - 1 do begin
      primitive := list.Get( index );

      if primitive is TLight then
        RegisterLight( primitive as TLight );
  end;
end;

// ============================================================================

procedure TLightmapRenderSystem.RegisterShadowCasters;
var primitiveIndex, polygonIndex, faceIndex: integer;
    basePrimitive: TBasePrimitive;
    primitive: TPrimitive;
    polygon: TPolygon;
    castShadows: boolean;
begin
    fShadowCasterIndex := 0;

    if SelectedPrimitivesOnly and not ConsiderSelectedOnly then begin
        // While there is a limited amount of lightmapped faces, all polygons in the scene may cast shadows

        for primitiveIndex := 0 to Scene.Primitives.Count - 1 do begin
            basePrimitive := Scene.Primitives[primitiveIndex];
            if basePrimitive is TPrimitive then begin
                primitive := basePrimitive as TPrimitive;

                for polygonIndex := 0 to primitive.NumPolygons - 1 do begin
                    polygon := primitive.Polygons[polygonIndex];

                    if primitive.OverrideMaterialLighting then
                        castShadows := primitive.CastShadows
                    else
                        castShadows := polygon.Material.CastShadows;

                    if castShadows then
                        RegisterShadowCastersForPolygon(polygon);
                end;
            end;
        end;
    end
    else for faceIndex := Low(fFaces) to High(fFaces) do begin
        // faces represent the shadowcasters
        if ffaces[faceIndex].CastShadows then
            RegisterShadowCastersForPolygon(fFaces[faceIndex].polygon);
    end;

    // Shadowcasters might very well have grown too large. We now know the exact size.
    SetLength(fShadowCasters, fShadowCasterIndex);

    UpdateShadowCasterTransparency;
end;

// ============================================================================

procedure TLightmapRenderSystem.RegisterShadowCastersForPolygon(const aPolygon: TPolygon);
var i: integer;
    shadowCaster: PFaceInfo;
begin
    for i := 0 to aPolygon.NumVertices - 3 do begin
        // Dynamic (and exponential) growth of shadowcasters array
        while Length(fShadowCasters) <= fShadowCasterIndex do begin
            if fShadowCasterIndex < 8 then
                SetLength(fShadowCasters, fShadowCasterIndex + 1)
            else
                SetLength(fShadowCasters, Length(fShadowCasters) * 2);
        end;

        shadowCaster := @fShadowCasters[fShadowCasterIndex];

        shadowCaster.Polygon := aPolygon;
        shadowCaster.Vertices[0] := verticeToVertex(aPolygon.VerticeList[0]);
        shadowCaster.Vertices[1] := verticeToVertex(aPolygon.VerticeList[i + 1]);
        shadowCaster.Vertices[2] := verticeToVertex(aPolygon.VerticeList[i + 2]);

        if  (aPolygon.Material.NumTextureLayers > 0) and
            (aPolygon.Material.TextureLayers[0].LayerType = mlTexture) then begin
            shadowCaster.UVCoordinates[0] := UVCoordToUVCoordinate(aPolygon.UVCoordinates[0][0]);
            shadowCaster.UVCoordinates[1] := UVCoordToUVCoordinate(aPolygon.UVCoordinates[0][i + 1]);
            shadowCaster.UVCoordinates[2] := UVCoordToUVCoordinate(aPolygon.UVCoordinates[0][i + 2]);
        end;

        fShadowCasterIndex := fShadowCasterIndex + 1;
    end;
end;

// ============================================================================

// By preparsing for shadowcaster transparency, we can optimize for scenes
// that don't contain any transparent shadow casters;
procedure TLightmapRenderSystem.UpdateShadowCasterTransparency;
var index: integer;
    shadowCaster: PFaceInfo;
    textureNames: TStringList;
    layerIndex, textureIndex: integer;
    material: TMaterial;
    textureLayer: TTextureLayer;
    fileName: string;
begin
    textureNames := TStringList.Create;

    fContainsTransparentShadowCasters := false;
    for index := Low(fShadowCasters) to High(fShadowCasters) do begin
        shadowCaster := @fShadowCasters[index];

        shadowCaster.TransparentTextureIndex := -1;
                
        material := shadowCaster.Polygon.Material;

        for layerIndex := 0 to material.NumTextureLayers do begin
            textureLayer := material.TextureLayers[layerIndex];

            if Assigned(textureLayer) and textureLayer.IsTransparent then begin
                fContainsTransparentShadowCasters := true;

                // For now, the lightmapper only supports transparency in the first layer
                if (textureLayer.LayerType = mlTexture) and (layerIndex = 0) then begin
                    fileName := textureLayer.Texture.FileName;
                    textureIndex := textureNames.IndexOf(fileName);
                    if textureIndex = -1 then
                        textureIndex := textureNames.Add(fileName);

                    shadowCaster.TransparentTextureIndex := textureIndex;
                end;
            end;
        end;
    end;

    LoadTransparentTextures(textureNames);

    FreeAndNil( textureNames );
end;

// ============================================================================

procedure TLightmapRenderSystem.LoadTransparentTextures(const aTransparentTextureNames: TStringList);
var index: integer;
    textureLoader: TTextureLoader;
begin
    SetLength(fTransparentTextures, aTransparentTextureNames.Count);
    textureLoader := TTextureLoader.Create;
    try
        for index := 0 to aTransparentTextureNames.Count - 1 do begin
            fTransparentTextures[index] := textureLoader.loadTextureData(aTransparentTextureNames[index]);
        end;
    finally
        FreeAndNil(textureLoader);
    end;
end;

// ============================================================================

// Retrieves significant indices for a polygon... These are the indices of any
// vertices that form a non-degenerate triangle. For any stricly convex polygon,
// the first three indices would do. For any non-convex polygon, they might be
// any three vertices.
function TLightmapRenderSystem.SignificantIndicesForPolygon(const aPolygon: TPolygon): TTripleIndices;
const EPSILON = 0.001;
var i, j: integer;
    v1, v2, v3: TVertex;
    edge1, edge2: TVertex;
begin
    result[0] := 0;
    result[1] := 1;
    result[2] := 2;

    if aPolygon.NumVertices = 3 then exit;

    v1 := VerticeToVertex(aPolygon.VerticeList[0]);
    for i := 1 to aPolygon.NumVertices - 1 do begin
        v2 := VerticeToVertex(aPolygon.VerticeList[i]);
        edge1 := VertexSubtract(v2, v1);
        for j := i + 1 to aPolygon.NumVertices - 1 do begin
            v3 := VerticeToVertex(aPolygon.VerticeList[j]);
            edge2 := VertexSubtract(v3, v2);

            if  (abs(edge1.Y * edge2.Z - edge1.Z * edge2.Y) > EPSILON) or
                (abs(edge1.Z * edge2.X - edge1.X * edge2.Z) > EPSILON) or
                (abs(edge1.X * edge2.Y - edge1.Y * edge2.X) > EPSILON) then
            begin
                result[1] := i;
                result[2] := j;
                exit;
            end
            else
                result[0] := 0;
        end;
    end;
end;

// ============================================================================

procedure TLightmapRenderSystem.RegisterFacesForPrimitive( const aPrimitive: TPrimitive );
var
  index, i: integer;
  info: TPlanarMappedTextureInfo;
  p0: TVertex;
  normal: TVertex;
  vertex1, vertex2: TVertex;
  distanceToPlane: single;
  polygon: TPolygon;
  indices: TTripleIndices;

begin
  for index := 0 to aPrimitive.NumPolygons - 1 do begin
      polygon := aPrimitive.Polygons[index];

      if polygon.AllowLightmap and (not SelectedPolygonsOnly or aPrimitive.SelectedPolygons.IsInList(polygon)) then begin
          fFaces[fFaceIndex].Polygon := polygon;

          if aPrimitive.OverrideMaterialLighting then
          begin
              fFaces[fFaceIndex].CastShadows := aPrimitive.CastShadows;
              fFaces[fFaceIndex].ReceiveShadows := aPrimitive.ReceiveShadows;
          end
          else begin
              fFaces[fFaceIndex].CastShadows := polygon.Material.CastShadows;
              fFaces[fFaceIndex].ReceiveShadows := polygon.Material.ReceiveShadows;
          end;

          indices := SignificantIndicesForPolygon(polygon);

          fFaces[fFaceIndex].Vertices[0] := verticeToVertex(polygon.VerticeList[indices[0]]);
          fFaces[fFaceIndex].Vertices[1] := verticeToVertex(polygon.VerticeList[indices[1]]);
          fFaces[fFaceIndex].Vertices[2] := verticeToVertex(polygon.VerticeList[indices[2]]);

          fNrLightmappedFaces := fNrLightmappedFaces + 1;

          info := GeneratePlanarMappedTexture( polygon );
          fFaces[fFaceIndex].Alignment := info.Alignment;
          p0 := fFaces[fFaceIndex].Vertices[0];
          normal := verticeToVertex(polygon.Normal);

          // INFO, PJ: We register polygons that are connected to the face ->
          //           If those face the same way, we don't cast shadows
          fFaces[fFaceIndex].AdjacentPolygons := TPolygonList.Create( FALSE );
          polygon.VerticeList.GetAdjacent(fFaces[fFaceIndex].AdjacentPolygons);
          for i := fFaces[fFaceIndex].AdjacentPolygons.Count - 1 downto 0 do begin
              if VertexDotProduct(normal, VerticeToVertex(fFaces[fFaceIndex].AdjacentPolygons[i].Normal)) < 0 then
                  fFaces[fFaceIndex].AdjacentPolygons.Delete(i);
          end;

          distanceToPlane := -vertexDotProduct( normal, p0 );

          case info.Alignment of
             paXZ:
              begin
                  fFaces[fFaceIndex].UVVertex.X := info.MinU;
                  fFaces[fFaceIndex].UVVertex.Y := - ( normal.X * info.MinU + normal.Z * info.MinV + distanceToPlane ) / normal.Y;
                  fFaces[fFaceIndex].UVVertex.Z := info.MinV;

                  vertex1.X := info.MaxU;
                  vertex1.Y := - ( normal.X * info.MaxU + normal.Z * info.MinV + distanceToPlane ) / normal.Y;
                  vertex1.Z := info.MinV;

                  vertex2.X := info.MinU;
                  vertex2.Y := - ( normal.X * info.MinU + normal.Z * info.MaxV + distanceToPlane ) / normal.Y;
                  vertex2.Z := info.MaxV;
              end;

              paYZ:
              begin
                  fFaces[fFaceIndex].UVVertex.X := -( normal.Z * info.MinU + normal.Y * info.MinV + distanceToPlane ) / normal.X;
                  fFaces[fFaceIndex].UVVertex.Y := info.MinV;
                  fFaces[fFaceIndex].UVVertex.Z := info.MinU;

                  vertex1.X := -( normal.Z * info.MaxU + normal.Y * info.MinV + distanceToPlane ) / normal.X;
                  vertex1.Y := info.MinV;
                  vertex1.Z := info.MaxU;

                  vertex2.X := -( normal.Z * info.MinU + normal.Y * info.MaxV + distanceToPlane ) / normal.X;
                  vertex2.Y := info.MaxV;
                  vertex2.Z := info.MinU;
              end;

              paXY:
              begin
                  fFaces[fFaceIndex].UVVertex.X := info.MinU;
                  fFaces[fFaceIndex].UVVertex.Y := info.MinV;
                  fFaces[fFaceIndex].UVVertex.Z := -( normal.X * info.MinU + normal.Y * info.MinV + distanceToPlane ) / normal.Z;

                  vertex1.X := info.MaxU;
                  vertex1.Y := info.MinV;
                  vertex1.Z := -( normal.X * info.MaxU + normal.Y * info.MinV + distanceToPlane ) / normal.Z;

                  vertex2.X := info.MinU;
                  vertex2.Y := info.MaxV;
                  vertex2.Z := -( normal.X * info.MinU + normal.Y * info.MaxV + distanceToPlane ) / normal.Z;
              end;
          end;

          fFaces[fFaceIndex].Edge1 := vertexSubtract( vertex1, fFaces[fFaceIndex].UVVertex );
          fFaces[fFaceIndex].Edge2 := vertexSubtract( vertex2, fFaces[fFaceIndex].UVVertex );

          fFaceIndex := fFaceIndex + 1;
      end;
  end;
end;

// ============================================================================

procedure TLightmapRenderSystem.RegisterLight( const aLight: TLight );
begin
    if aLight.active then
        fLights.Add( aLight );
end;

// ============================================================================

procedure TLightmapRenderSystem.RenderLightMapTexture( const aFace: TFaceInfo );
var
  xResolution, yResolution: integer;
  uvVertex: TVertex;
  edge1, edge2: TVertex;
  position: TVertex;
  lightIndex: integer;
  light: TLight;
  xOffset, yOffset: integer;
  xIndex, yIndex: integer;
  xStep, yStep: single;
  colour: TRGBColor;
  normal: TVertex;
  pixels: TPackedByteArray;
  targetPixel: PTexturePixel;
  targetIndex: integer;
  lightColour: TRGBColor;
  shouldSmoothNormals: boolean;
//  ambientOcclusion: single;

begin
  pixels := aFace.Lightmap.Pixels;

  shouldSmoothNormals := not aFace.polygon.HasIdenticalVertexNormals;

  xResolution := aFace.XResolution - 2 * LIGHTMAP_BORDER_SIZE;
  YResolution := aFace.YResolution - 2 * LIGHTMAP_BORDER_SIZE;

  uvVertex := aFace.UVVertex;
  edge1 := aFace.Edge1;
  edge2 := aFace.Edge2;
  xOffset := aFace.XOffset + LIGHTMAP_BORDER_SIZE;
  yOffset := aFace.YOffset + LIGHTMAP_BORDER_SIZE;

  xStep := 1 / xResolution;
  yStep := 1 / yResolution;

  for yIndex := 0 to yResolution - 1 do begin
      for xIndex := 0 to xResolution - 1 do begin
          colour.R := 1 - fShadowOpacity;
          colour.G := 1 - fShadowOpacity;
          colour.B := 1 - fShadowOpacity;

          position.X := uvVertex.X + yIndex * yStep * edge2.X + xIndex * xStep * edge1.X;
          position.Y := uvVertex.Y + yIndex * yStep * edge2.Y + xIndex * xStep * edge1.Y;
          position.Z := uvVertex.Z + yIndex * yStep * edge2.Z + xIndex * xStep * edge1.Z;

          if shouldSmoothNormals then
              normal := CalculateSmoothNormal( position, aFace )
          else
              normal := VerticeToVertex( aFace.Polygon.VertexNormals[0] );

//
          // HACK, PJ:
//          colour.R := 1;
//          colour.G := 1;
//          colour.B := 1;

//          ambientOcclusion := CalculateAmbientOcclusion( position, normal, aFace );
//          colour.R := colour.R * ambientOcclusion;
//          colour.G := colour.G * ambientOcclusion;
//          colour.B := colour.B * ambientOcclusion;

          for lightIndex := 0 to fLights.Count - 1 do begin
              light := fLights[lightIndex] as TLight;

              lightColour := InfluenceForLight( position, normal, light, aFace );
              {$IFDEF DELED_LIGHTMAPPER_RENDER_NORMALS}
              colour.R := 0.5 + normal.X / 2;
              colour.G := 0.5 + normal.Y / 2;
              colour.B := 0.5 + normal.Z / 2;
              {$ELSE}
              colour.R := colour.R + lightColour.R;
              colour.G := colour.G + lightColour.G;
              colour.B := colour.B + lightColour.B;
              {$ENDIF}
          end;

          if fShouldUseAmbientColour then begin
              colour.R := Scene.AmbientColor.R * colour.R;
              colour.G := Scene.AmbientColor.G * colour.G;
              colour.B := Scene.AmbientColor.B * colour.B;
          end;

          // INFO, PJ: Clip.
          if colour.R > 1 then colour.R := 1;
          if colour.G > 1 then colour.G := 1;
          if colour.B > 1 then colour.B := 1;

          targetIndex := ( yOffset + yIndex ) * LightMapSize + xOffset + xIndex;

          targetPixel := @Pixels[targetIndex * 3];
          targetPixel[0] := round(colour.B * 255);
          targetPixel[1] := round(colour.G * 255);
          targetPixel[2] := round(colour.R * 255);

          if fStopped then
              exit;
      end;
  end;

  with aFace do begin
      RenderTextureBorder( XOffset, YOffset, XResolution, YResolution, LightMap );

      if fNrBlurFilterPasses > 0 then begin
          BlurTexture( XOffset, YOffset, XResolution, YResolution, LightMap, fNrBlurFilterPasses );

          // INFO, PJ: Rerender border after blurring!
          RenderTextureBorder( XOffset, YOffset, XResolution, YResolution, LightMap );
      end;
  end;
end;

// ============================================================================

procedure TLightmapRenderSystem.RenderTextureBorder( const anX, aY, aWidth, aHeight: integer; const aTextureData: TTextureData );
var
  index: integer;
  offset: integer;
  buffer: PTripletArray;

begin
    buffer := @aTextureData.Pixels[0];
    
    for index := anX + 1 to anX + aWidth - 2 do
    begin
        offset := aY * LightMapSize + index;
        buffer[offset] := buffer[offset + LightmapSize];

        offset := offset + LightmapSize * ( aHeight - 1 );
        buffer[offset] := buffer[offset - LightmapSize];
    end;

    offset := ( aY + 1 ) * LightMapSize + anX;
    for index := 0 to aHeight - 2 do
    begin
        buffer[offset] := buffer[offset + 1];
        buffer[offset + aWidth - 1] := buffer[offset + aWidth - 2];

        offset := offset + LightMapSize;
    end;

    buffer[aY * LightmapSize + anX ] := buffer[( aY + 1 ) * LightmapSize + ( anX + 1 ) ];
    buffer[( aY + aHeight - 1 ) * LightmapSize + anX ] := buffer[( aY + aHeight - 2 ) * LightmapSize + ( anX + 1 )];
    buffer[aY * LightmapSize + ( anX + aWidth - 1 )] := buffer[(aY + 1 ) * LightmapSize + ( anX + aWidth - 1 )];
    buffer[( aY + aHeight - 1 ) * LightmapSize + ( anX + aWidth - 1 )] := buffer[( aY + aHeight - 2 ) * LightmapSize + ( anX + aWidth - 1 )];

end;

// ============================================================================

procedure TLightmapRenderSystem.BlurTexture( const anX, aY, aWidth, aHeight: integer; const aTextureData: TTextureData; const aNrBlurs: integer = 1 );
// INFO, PJ: Conveniently uses the fact that there is a border around the texture

  procedure AddSample( const anX, aY: integer; var aBlend: array of integer );
  // INFO, PJ: SubProcedures are evil, but can be convenient :)
  var
      offset: integer;
      triplet: PTriplet;

  begin
      offset := aY * LightmapSize + anX;
      triplet := @aTextureData.Pixels[offset * 3];

      aBlend[0] := aBlend[0] + triplet[0];
      aBlend[1] := aBlend[1] + triplet[1];
      aBlend[2] := aBlend[2] + triplet[2];
  end;

  // --------------------------------------------------------------------------

var
  buffer, tempBuffer: PTripletArray;
  x, y: integer;
  blend: Array[0..2] of integer;
  index: integer;
  offset: integer;
  triplet: PTriplet;
  targetTriplet: PTriplet;

begin
  buffer := @aTextureData.Pixels[0];
  getMem( tempBuffer, aHeight * aWidth * 3 );

  // INFO, PJ: We copy the buffer into the temp buffer to store the border
  for y := aY  to aY + aHeight - 1 do
  begin
      move( buffer[y * LightmapSize + anX], tempBuffer[(y - aY ) * aWidth], aWidth * 3 );
  end;

  for index := 0 to aNrBlurs - 1 do
  begin
      // Apply 9x9 blur and store in temporary blur buffer
      for y := aY + 1 to aY + aHeight - 2 do
      begin
          for x := anX + 1 to anX + aWidth - 2 do
          begin
              offset := y * LightmapSize + x;
              triplet := @buffer[offset];
              targetTriplet := @tempBuffer[(y - aY ) * aWidth + x - anX];

              blend[0] := triplet[0];
              blend[1] := triplet[1];
              blend[2] := triplet[2];

              AddSample( x - 1, y - 1, blend );
              AddSample( x    , y - 1, blend );
              AddSample( x + 1, y - 1, blend );
              AddSample( x - 1, y    , blend );
              AddSample( x + 1, y    , blend );
              AddSample( x - 1, y + 1, blend );
              AddSample( x    , y + 1, blend );
              AddSample( x + 1, y + 1, blend );

              targetTriplet[0] := blend[0] div 9;
              targetTriplet[1] := blend[1] div 9;
              targetTriplet[2] := blend[2] div 9;
          end;
      end;

      // INFO, PJ: We copy the blurred buffer back into the texture
      for y := aY  to aY + aHeight - 1 do
      begin
          move( tempBuffer[(y - aY ) * aWidth], buffer[y * LightmapSize + anX], aWidth * 3 );
      end;
  end;

  FreeMem( tempBuffer );
end;

// ============================================================================

function TLightmapRenderSystem.InfluenceForLight( const aPosition, aNormal: TVertex; const aLight: TLight; const aFace: TFaceInfo ): TRGBColor;
var
  pointLight: TPointLight;
  directionalLight: TDirectionalLight;
  spotLight: TSpotLight;
  direction: TVertex;
  destination: TVertex;
  multiplier: TVertex;
  lightDirection: TVertex;
  distance: single;
  cosAngle: single;
  cosInner: single;
  cosOuter: single;
  d: single;                  // cos(angle of incidence)
  triangleIntersections: TTriangleIntersections;
  intersection: TTriangleIntersection;
  intensity: single;
  index: integer;


begin
  KeepGUIAlive;

  if ( aLight is TPointLight ) then begin
      pointLight := aLight as TPointLight;

      direction := vertexSubtract( verticeToVertex( aLight.Position ), aPosition );
      distance := vertexLength( direction );
      direction := vertexDivide( direction, distance );

      d := vertexDotProduct( aNormal, direction );
      if d < 0 then begin
          result := ZERO_RGBCOLOR;
          exit;
      end
      else
          intensity := d;

      if ( distance > pointLight.Range ) then begin
          result := ZERO_RGBCOLOR;
          exit;
      end;

      intensity := intensity / ( pointLight.Constant + pointLight.Linear * distance +
          pointLight.Quadratic * distance * distance );

      if intensity > 1 then intensity := 1;
  end
  else if ( aLight is TDirectionalLight ) then begin
      directionalLight := aLight as TDirectionalLight;
      direction.X := -directionalLight.Direction.X;
      direction.Y := -directionalLight.Direction.Y;
      direction.Z := -directionalLight.Direction.Z;

      vertexNormalize( direction );

      d := vertexDotProduct( aNormal, direction );
      if d < 0 then begin
          result := ZERO_RGBCOLOR;
          exit;
      end
      else
          intensity := d;

      if intensity > 1 then intensity := 1;
  end
  else if ( aLight is TSpotLight ) then begin
      spotLight := aLight as TSpotLight;

      direction := vertexSubtract( verticeToVertex( spotLight.Position ), aPosition );
      distance := vertexLength( direction );
      direction := vertexDivide( direction, distance );

      if ( distance > spotLight.Range ) then begin
          result := ZERO_RGBCOLOR;
          exit;
      end;

      lightDirection.X := -spotLight.Direction.X;
      lightDirection.Y := -spotLight.Direction.Y;
      lightDirection.Z := -spotLight.Direction.Z;
      vertexNormalize( lightDirection );

      cosAngle := vertexDotProduct( lightDirection, direction );
      if cosAngle < 0 then begin
           result := ZERO_RGBCOLOR;
           exit;
      end;

      cosInner := Cos( DegToRad( spotLight.InnerAngle ));
      cosOuter := Cos( DegToRad( spotLight.OuterAngle ));

      d := vertexDotProduct( aNormal, direction );
      if d < 0 then begin
          result := ZERO_RGBCOLOR;
          exit;
      end;

      if cosAngle > cosInner then begin
          intensity := d;
      end
      else if cosAngle > cosOuter then begin
          intensity := d * Power( ( cosAngle - cosOuter ) / ( cosInner - cosOuter ), spotLight.FallOff );
      end
      else
          intensity := 0;

      intensity := intensity / ( spotLight.Constant + spotLight.Linear * distance +
          spotLight.Quadratic * distance * distance );

      if intensity > 1 then intensity := 1;
  end
  else
      raise Exception.Create(Format('Unsupported light class "%s".', [aLight.ClassName]));

  intensity := intensity * aLight.intensity / 100;
  if intensity > 1 then intensity := 1;

  result := RGBColor( aLight.Diffuse.R * intensity, aLight.Diffuse.G * intensity, aLight.Diffuse.B * intensity );

  fRayIndex := fRayIndex + 1;

  // INFO, PJ: Intersection testing (shadows)
  if fShouldGenerateShadows and aFace.ReceiveShadows and aLight.CastShadows then begin
      triangleIntersections := TTriangleIntersections.Create;
      try
          if aLight.objectType = otDirectionalLight then begin
              // INFO, PJ: Little trick: we 'fake' light position outside the AABB
              // INFO, PJ: To be sure, we take 2 times the AABB dimension)
              multiplier := vertexScaleNormalVertex( direction, 2 );

              destination.X := aPosition.X + fOctree.RootNode.Size * multiplier.X;
              destination.Y := aPosition.Y + fOctree.RootNode.Size * multiplier.Y;
              destination.Z := aPosition.Z + fOctree.RootNode.Size * multiplier.Z;
          end
          else
              destination := verticeToVertex( aLight.Position );

          FindIntersectedTrianglesInOctreeNode(aPosition, destination, direction, fOctree.RootNode, not fContainsTransparentShadowCasters, triangleIntersections);

          if fContainsTransparentShadowCasters and triangleIntersections.ContainsTransparentTriangles then begin
              triangleIntersections.Sort;

              for index := triangleIntersections.Count - 1 downto 0 do begin
                  intersection := triangleIntersections[index];

                  // adjacent polygons that face the same way towards the light don't cast shadows.
                  if  ( not aFace.AdjacentPolygons.IsInList(intersection.Face.Polygon)) or
                      ( vertexDotProduct( verticeToVertex(intersection.Face.Polygon.Normal), direction ) < 0 ) then begin
                      if not intersection.Face.Polygon.Material.IsTransparent then begin
                          result := ZERO_RGBCOLOR;
                          break;
                      end;

                      FilterColorThroughMaterial( result, intersection.Face, intersection.U, intersection.V );
                  end;
              end;

          end
          else begin
              for index := triangleIntersections.Count - 1 downto 0 do begin
                  intersection := triangleIntersections[index];

                  if  ( not aFace.AdjacentPolygons.IsInList(intersection.Face.Polygon)) or
                      ( vertexDotProduct( verticeToVertex(intersection.Face.Polygon.Normal), direction ) < 0 ) then begin

                      result := ZERO_RGBCOLOR;
                      break;
                  end;
              end;
          end;
      finally
          FreeAndNil(triangleIntersections);
      end;
  end;
end;

// ============================================================================

function TLightmapRenderSystem.PlaneAlignmentForNormal( const aNormal: TVertex ): TPlaneAlignment;
var
  x, y, z: single;

begin
  x := Abs( aNormal.X );
  y := Abs( aNormal.Y );
  z := Abs( aNormal.Z );

  if ( x >= y ) and ( x >= z ) then
  begin
      result := paYZ;
  end
  else if ( y >= x ) and ( y >= z ) then
  begin
      result := paXZ;
  end
  else
      result := paXY;
end;

// ============================================================================

// INFO, PJ: Planar mapping as suggested by Joachim de Vries and others
function TLightmapRenderSystem.GeneratePlanarMappedTexture( const aPolygon: TPolygon ): TPlanarMappedTextureInfo;
var
  vertexIndex: integer;
  vertex: TVertex;
  index: integer;
  u, v: single;

begin
  result.Alignment := PlaneAlignmentForNormal( verticeToVertex( aPolygon.Normal ));
  result.MinU :=  9999999;
  result.MaxU := -9999999;
  result.MinV :=  9999999;
  result.MaxV := -9999999;

  u := 0;
  v := 0;

  index := 0;
  for vertexIndex := 0 to aPolygon.NumVertices - 1 do begin
      vertex := verticeToVertex( aPolygon.VerticeList[vertexIndex] );

      case result.Alignment of
          paYZ: begin
              u := vertex.Z;
              v := vertex.Y;
          end;

          paXZ: begin
              u := vertex.X;
              v := vertex.Z;
          end;

          paXY: begin
              u := vertex.X;
              v := vertex.Y;
          end;
      end;

      aPolygon.UVCoordinates[LIGHTMAP_TEXTURE_LAYER_INDEX][index].U := u;
      aPolygon.UVCoordinates[LIGHTMAP_TEXTURE_LAYER_INDEX][index].V := v;

      result.MinU := Min( result.MinU, u );
      result.MinV := Min( result.MinV, v );
      result.MaxU := Max( result.MaxU, u );
      result.MaxV := Max( result.MaxV, v );

      index := index + 1;
  end;

  result.DeltaU := result.MaxU - result.MinU;
  result.DeltaV := result.MaxV - result.MinV;

  // INFO, PJ: Degenerate polygon (all coordinates in one line).
  if result.DeltaU = 0 then result.DeltaU := 1;
  if result.DeltaV = 0 then result.DeltaV := 1;

  index := 0;
  for vertexIndex := 0 to aPolygon.NumVertices - 1 do
  begin
      u := ( aPolygon.UVCoordinates[LIGHTMAP_TEXTURE_LAYER_INDEX][index].U - result.MinU ) / result.DeltaU;
      v := ( aPolygon.UVCoordinates[LIGHTMAP_TEXTURE_LAYER_INDEX][index].V - result.MinV ) / result.DeltaV;

      aPolygon.UVCoordinates[LIGHTMAP_TEXTURE_LAYER_INDEX][index].U := u;
      aPolygon.UVCoordinates[LIGHTMAP_TEXTURE_LAYER_INDEX][index].V := v;

      index := index + 1;
  end;
end;

// ============================================================================

// fills aTriangleIntersections with all triangles intersected with
procedure TLightmapRenderSystem.FindIntersectedTrianglesInOctreeNode(
  const anOrigin, aDestination, aDirection: TVertex;
  const aNode: TOctreeNode;
  const aShouldStopOnFirstHit: boolean;
  const aTriangleIntersections: TTriangleIntersections );
var
  index: integer;
  faceIndex: integer;
  face: PFaceInfo;
  t, u, v: single;
  squaredDistance: single;

begin
  KeepGUIAlive;

  if aNode.IntersectsLineSegment( anOrigin, aDestination ) then
  begin
      for index := Low( aNode.Children ) to High( aNode.Children ) do
      begin
          if Assigned( aNode.Children[index] ) then
          begin
              FindIntersectedTrianglesInOctreeNode( anOrigin, aDestination, aDirection, aNode.Children[index], aShouldStopOnFirstHit, aTriangleIntersections );

              if aShouldStopOnFirstHit and ( aTriangleIntersections.Count > 0 ) then
                  exit;
          end;
      end;

      // INFO, PJ: Only terminal nodes actually have items
      for index := Low( aNode.Items ) to High( aNode.Items ) do
      begin
          faceIndex := aNode.Items[index].TriangleIndex;

          face := @fShadowCasters[faceIndex];
          if face.Polygon <> fFaces[fProcessingFaceIndex].Polygon then
          begin // INFO, PJ: Avoid the triangle we are currently working on

              if face.LastRayIndex <> fRayIndex then
              begin // INFO, PJ: Triangles might be in multiple nodes. Avoid processing the same triangle twice.
                  face.LastRayIndex := fRayIndex;
              end;

              KeepGUIAlive;

              if IntersectsTriangle( anOrigin, aDirection, face.Vertices[0], face.Vertices[1], face.Vertices[2], t, u, v ) then begin
                  squaredDistance := vertexSquaredLength( vertexSubtract(aDestination, anOrigin));

                  if ( t > 1 ) and ( t * t < squaredDistance ) then begin
                      aTriangleIntersections.Add(face^, t, u, v);

                      if aShouldStopOnFirstHit then exit;
                  end;
              end;
          end;
      end;
  end;
end;

// ============================================================================

// INFO, PJ: MollerTrumbore97 algorithm for Fast, Minimum Storage Ray-Triangle Intersection.
function TLightmapRenderSystem.IntersectsTriangle(
  const anOrigin, aDirection: TVertex;
  const aVertex0, aVertex1, aVertex2: TVertex;
  var aT, aU, aV: single ): Boolean;

var
  edge1, edge2: TVertex;
  tVec, pVec, qVec: TVertex;
  det, invDet: single;

begin
  // find vectors for two edges sharing vert0
  edge1 := vertexSubtract( aVertex1, aVertex0 );
  edge2 := vertexSubtract( aVertex2, aVertex0 );

  // begin calculating determinant - also used to calculate U parameter
  pVec := vertexCrossProduct( aDirection, edge2 );

  // if determinant is near zero, ray lies in plane of triangle
  det := vertexDotProduct( edge1, pVec );

  // To make culling, remove either half of the condition
  if (det > -epsilon) and (det < epsilon) then begin
      result := false;
      exit;
  end;

  invDet := 1 / det;

  //calculate distance from vert0 to ray origin
  tVec := vertexSubtract( anOrigin, aVertex0 );

  // calculate U parameter and test bounds
  aU := vertexDotProduct( tVec, pVec ) * invDet;

  if ( aU < 0 ) or ( aU > 1 ) then begin
      result := false;
      exit;
  end;

  // prepare to test V parameter
  qVec := vertexCrossProduct( tVec, edge1 );

  // calculate V parameter and test bounds */
  aV := vertexDotProduct( aDirection, qVec ) * invDet;
  if ( aV < 0 ) or ( aU + aV > 1 ) then begin
      result := false;
      exit;
  end;

  // calculate t, scale parameters, ray intersects triangle
  aT := vertexDotProduct( edge2, qVec ) * invDet;

  result := true;
end;

// ============================================================================

function TLightmapRenderSystem.CalculateSmoothNormal( const aPosition: TVertex; const aFace: TFaceInfo ): TVertex;
var
  i, minIndex: integer;
  v0, v1, v2, n0, n1, n2: TVertex;
  u, v, w: single;
  polygon: TPolygon;
  found: boolean;
  minDistance, distance, minU, minV: single;


begin
  polygon := aFace.Polygon;
  found := FALSE;
  minDistance := 9999999;
  minU := 0;
  minV := 0;
  minIndex := 0;

  for i := 0 to polygon.NumVertices - 3 do begin
      v0 := VerticeToVertex(polygon.VerticeList[i + 2]);
      v1 := VerticeToVertex(polygon.VerticeList[i + 1]);
      v2 := VerticeToVertex(polygon.VerticeList[0]);
      n0 := VerticeToVertex(polygon.VertexNormals[i + 2]);
      n1 := VerticeToVertex(polygon.VertexNormals[i + 1]);
      n2 := VerticeToVertex(polygon.VertexNormals[0]);
      if IsPointInTriangle(aPosition, v0, v1, v2, u, v) then begin
          w := 1 - u - v;

          result.X := u * n1.X + v * n2.X + w * n0.X;
          result.Y := u * n1.Y + v * n2.Y + w * n0.Y;
          result.Z := u * n1.Z + v * n2.Z + w * n0.Z;

          found := TRUE;
          break;
      end
      else begin
          // INFO, PJ: If we miss the triangle, we store stuff.
          //           Perhaps we should re-loop if the triangle was missed?
          w := 1 - u - v;

          distance := 0;
          if u < 0 then distance := distance + u * u;
          if u > 1 then distance := distance + ( u - 1 ) * ( u - 1 );
          if v < 0 then distance := distance + v * v;
          if v > 1 then distance := distance + ( v - 1 ) * ( v - 1 );
          if w < 0 then distance := distance + w * w;
          if w > 1 then distance := distance + ( w - 1 ) * ( w - 1 );

          if distance < minDistance then
          begin
              minDistance := distance;
              minIndex := i;
              minU := u;
              minV := v;
          end;
      end;
  end;

  // INFO, PJ: We might have missed the triangle.
  //           We use the coordinates from the closest triangle

  if not found then begin
      w := 1 - minU - minV;
      n0 := VerticeToVertex(polygon.VertexNormals[minIndex + 2]);
      n1 := VerticeToVertex(polygon.VertexNormals[minIndex + 1]);
      n2 := VerticeToVertex(polygon.VertexNormals[0]);
      result.X := minU * n1.X + minV * n2.X + w * n0.X;
      result.Y := minU * n1.Y + minV * n2.Y + w * n0.Y;
      result.Z := minU * n1.Z + minV * n2.Z + w * n0.Z;
  end;
end;

// ============================================================================
{
function TLightmapRenderSystem.CalculateAmbientOcclusion( const aPosition, aNormal: TVertex; const aFace: TFaceInfo ): single;
const
  NUM_SAMPLES = 5;
  MAX_ANGLE = 0.9 * PI;
  MAX_DISTANCE = 5000;

var
  i, j, numHits: integer;
  r1, r2, x, y, z, distance: single;
  tangent, VX, VY, VZ: TVertex;


begin
  numHits := 0;

  if VertexEquals( aPosition, aFace.Vertices[0] ) then
  begin
      tangent := VertexSubtract( aPosition, aFace.Vertices[1] );
  end
  else
      tangent := VertexSubtract( aPosition, aFace.Vertices[0] );

  for i := 0 to NUM_SAMPLES - 1 do
  begin
      for j := 0 to NUM_SAMPLES - 1 do
      begin
            r1 := i / NUM_SAMPLES + ( random - 0.5 ) / NUM_SAMPLES;
            r2 := j / NUM_SAMPLES + ( random - 0.5 ) / NUM_SAMPLES;

            x := cos( 2 * pi * r1 ) * Sqrt( 1 - Sqr( 1 - r2 * ( 1 - cos( MAX_ANGLE ))));
            y := sin( 2 * pi * r1 ) * Sqrt( 1 - Sqr( 1 - r2 * ( 1 - cos( MAX_ANGLE ))));
            z := 1 - r2 * ( 1 - cos( MAX_ANGLE ));

            // Convert local hemisphere coordinates to global coordinates
            VZ := aNormal;
            VY := VertexNormalize( VertexCrossProduct( tangent, aNormal ));
            VX := VertexNormalize( VertexCrossProduct( VY, VZ ));

            origin := position;
            direction := VertexAdd( VertexAdd( VertexMultiply( VX, x ), VertexMultiply( VY, y )), VertexMultiply( Vz, Z ));
            destination :=

            // INFO, PJ: Be ware: distance is modified (var argument) at every call!
            distance := MAX_DISTANCE;
//            if Assigned( FindIntersectedPrimitiveInOctreeNode(ray, fOctree.RootNode, distance, TRUE)) then
//            begin
//                numHits := numHits + 1;
//            end;
      end;
  end;

//  totalSamples := NUM_SAMPLES * NUM_SAMPLES;
//  result.R := ( totalSamples - numHits ) / totalSamples;
end;
}
// ============================================================================

procedure TLightmapRenderSystem.FilterColorThroughMaterial( var aColor: TRGBColor; const aFace: TFaceInfo; const aU, aV: single );
var
  material: TMaterial;
  textureLayer: TTextureLayer;
  alpha: single;
  materialColor: TRGBColor;
  textureData: TTextureData;
  u, v: single;
  x, y: integer;
  pixel: PTexturePixel;
  maxColorValue: single;


begin
  if aFace.Polygon.Material.IsOriginal then
      material := aFace.Polygon.Material
  else
      material := aFace.Polygon.Material.OriginalMaterial;

  textureLayer := material.TextureLayers[0];

  if textureLayer.LayerType = mlColor then
  begin
      alpha := ((textureLayer.Color and $FF000000) shr 24 ) / 255;
      materialColor := ColorToRGB( textureLayer.Color );
  end
  else begin

     textureData := fTransparentTextures[aFace.TransparentTextureIndex];

      // Interpolate uv coordinates (we got relative coordinates aU, aV)
      u := (1 - aU - aV) * aFace.UVCoordinates[0].U + aU * aFace.UVCoordinates[1].U + aV * aFace.UVCoordinates[2].U;
      v := (1 - aU - aV) * aFace.UVCoordinates[0].V + aU * aFace.UVCoordinates[1].V + aV * aFace.UVCoordinates[2].V;  

      u := Frac(u);
      v := Frac(v);
      if u < 0 then u := u + 1;
      if v < 0 then v := v + 1;

      x := Trunc(u * textureData.Width);
      y := Trunc(v * textureData.Height);
      pixel := @textureData.Pixels[(y * textureData.Width + x) * (textureData.bitsPerPixel div 8)];

      materialColor.R := pixel[2] / 255;
      materialColor.G := pixel[1] / 255;
      materialColor.B := pixel[0] / 255;

      if TextureData.BitsPerPixel = 32 then
          alpha := pixel[3] / 255
      else
          alpha := 1;
  end;

  // Alpha and color are treated seperately:
  // 1. alpha determines amount of light, and brightens the color
  // 2. color determines hue
  case textureLayer.BlendOperation of
      boAlphaBlend: begin
          if (alpha > 0) then begin
              // We change the color to values representing the relative 'hue'.
              maxColorValue := Max( materialColor.R, materialColor.G );
              maxColorValue := Max( maxColorValue, materialColor.B );

              if maxColorValue <> 0 then begin
                  materialColor.R := materialColor.R / maxColorValue;
                  materialColor.G := materialColor.G / maxColorValue;
                  materialColor.B := materialColor.B / maxColorValue;
              end;

              // Now we brighten the color using the alpha.
              materialColor.R := alpha * materialColor.R + (1 - alpha);
              materialColor.G := alpha * materialColor.G + (1 - alpha);
              materialColor.B := alpha * materialColor.B + (1 - alpha);

              // And filter
              alpha := 1 - alpha * alpha;

              aColor.R := aColor.R * materialColor.R * alpha;
              aColor.G := aColor.G * materialColor.G * alpha;
              aColor.B := aColor.B * materialColor.B * alpha;
          end;
      end;

      boAdd, boModulate: begin
          aColor.R := aColor.R * materialColor.R;
          aColor.G := aColor.G * materialColor.G;
          aColor.B := aColor.B * materialColor.B;
      end;
  end;
end;

// ============================================================================

procedure TLightmapRenderSystem.HandleProgress( const aPercentageDone: integer );
begin
  if Assigned( fProgressHandler ) then
  begin
      fProgressHandler( aPercentageDone );
  end;
end;

// ============================================================================

procedure TLightmapRenderSystem.HandleTriangleRequest( const anIndex: integer; var aTriangle: TOctreeTriangle );
begin
  aTriangle[0] := fShadowCasters[anIndex].Vertices[0];
  aTriangle[1] := fShadowCasters[anIndex].Vertices[1];
  aTriangle[2] := fShadowCasters[anIndex].Vertices[2];
end;

// ============================================================================

function TLightmapRenderSystem.MakeLightMap: TTextureData;
var
  index: integer;

begin
  result.Width := LightMapSize;
  result.Height := LightMapSize;
  result.BitsPerPixel := 24;
  SetLength( result.Pixels, result.Width * result.Height * (result.BitsPerPixel div 8));

  index := Length( fLightMaps );
  SetLength( fLightMaps, index + 1 );
  fLightMaps[index] := result;

  FreeAndNil( fBitmapTreeNode );
  fBitmapTreeNode := T2DBitmapTreeNode.Create( 0, 0, result.Width - 1, result.Height - 1 );
end;

// ============================================================================

procedure TLightmapRenderSystem.UpdateFaces;
// INFO, PJ: Scaled texture densities... perhaps I should change this to
//           constant density.
var
  index: integer;
  sizeX, sizeY: single;
  xResolution, yResolution: integer;
  minSizeX, maxSizeX: single;
  minSizeY, maxSizeY: single;
  lightMap: TTextureData;
  lightMapIndex: integer;
  insertedNode: T2DBitmapTreeNode;

begin
  minSizeX :=  9999;
  maxSizeX := -9999;
  minSizeY :=  9999;
  maxSizeY := -9999;

  for index := Low( fFaces ) to high( fFaces ) do begin
      sizeX := vertexLength( fFaces[index].Edge1 );
      sizeY := vertexLength( fFaces[index].Edge2 );

      if sizeX < minSizeX then minSizeX := sizeX;
      if sizeX > maxSizeX then maxSizeX := sizeX;
      if sizeY < minSizeY then minSizeY := sizeY;
      if sizeY > maxSizeY then maxSizeY := sizeY;
  end;

  lightMap := MakeLightMap;
  lightMapIndex := NumLightMaps - 1;

  for index := Low( fFaces ) to high( fFaces ) do begin
      sizeX := vertexLength( fFaces[index].Edge1 );
      sizeY := vertexLength( fFaces[index].Edge2 );

      if maxSizeX > minSizeX then begin
          xResolution := fMinimumTextureSize +
              Round((( sizeX - minSizeX ) / ( maxSizeX - minSizeX )) *
              ( fMaximumTextureSize - fMinimumTextureSize ));
      end
      else
          xResolution := ( fMinimumTextureSize + fMaximumTextureSize ) div 2;

      if maxSizeY > minSizeY then begin
          yResolution := fMinimumTextureSize +
              Round((( sizeY - minSizeY ) / ( maxSizeY - minSizeY )) *
              ( fMaximumTextureSize - fMinimumTextureSize ));
      end
      else
          yResolution := ( fMinimumTextureSize + fMaximumTextureSize ) div 2;

      insertedNode := fBitmapTreeNode.Insert( xResolution, yResolution );
      if not Assigned( insertedNode ) then begin
          lightMap := MakeLightMap;
          lightMapIndex := NumLightMaps - 1;

          insertedNode := fBitmapTreeNode.Insert( xResolution, yResolution );
          if not Assigned( insertedNode ) then
          begin
              raise Exception.Create( 'Failed to insert light texture into lightmap.' );
          end;
      end;

      fFaces[index].XResolution := xResolution;
      fFaces[index].YResolution := yResolution;
      fFaces[index].LightMap := lightMap;
      fFaces[index].LightMapIndex := lightMapIndex;
      fFaces[index].XOffset := insertedNode.Rectangle.Left;
      fFaces[index].YOffset := insertedNode.Rectangle.Top;
  end;
end;

// ============================================================================

procedure TLightmapRenderSystem.KeepGUIAlive;
// INFO, PJ: Call this from somewhere in the inner loops
begin
  fInstructionCounter := fInstructionCounter + 1;

  if fInstructionCounter >= 1000 then
  begin
      Application.ProcessMessages;
      fInstructionCounter := 0;
  end;
end;

// ============================================================================

function TLightmapRenderSystem.GetLightMapForIndex(anIndex: Integer): TTextureData;
begin
    result := fLightMaps[anIndex];
end;

// ============================================================================

function TLightmapRenderSystem.GetFaceForIndex(anIndex: integer): TFaceInfo;
begin
    result := fFaces[anIndex];
end;

// ============================================================================
// ============================================================================

constructor T2DBitmapTreeNode.Create( const aLeft, aTop, aRight, aBottom: integer );
begin
  inherited Create;

  fRectangle.Left := aLeft;
  fRectangle.Top := aTop;
  fRectangle.Right := aRight;
  fRectangle.Bottom := aBottom;
end;

// ============================================================================

destructor T2DBitmapTreeNode.Destroy;
begin
  FreeAndNil( fChildNodes[0] );
  FreeAndNil( fChildNodes[1] );

  inherited Destroy;
end;

// ============================================================================

function T2DBitmapTreeNode.HasChildren: boolean;
begin
  result :=
      Assigned( fChildNodes[0] );
end;

// ============================================================================

function T2DBitmapTreeNode.Insert( const aWidth, aHeight: integer ): T2DBitmapTreeNode;
var
  width, height: integer;
  deltaX, deltaY: integer;

begin
  if HasChildren then begin
      // try inserting into first child
      result := fChildNodes[0].Insert( aWidth, aHeight );

      if not Assigned( result ) then // no room, insert into second
          result := fChildNodes[1].Insert( aWidth, aHeight );
  end
  else begin
      // if there's already a lightmap here, return)
      if fUsed then begin
          result := nil;
          exit;
      end;

      // if we're too small, return
      width := fRectangle.right - fRectangle.left + 1;
      height := fRectangle.bottom - fRectangle.top + 1;

      if  ( aWidth > width ) or
          ( aHeight > height ) then begin
          result := nil;
          exit;
      end;

      // if we're just right, accept without splitting any further)
      if ( aWidth = width ) and
          ( aHeight = height ) then begin
          fUsed := true;

          result := self;
          exit;
      end;

      // otherwise, gotta split this node (create some kids)
      // decide which way to split
      deltaX := width - aWidth;
      deltaY := height - aHeight;

      if deltaX > deltaY then begin
          fChildNodes[0] := T2DBitmapTreeNode.Create( fRectangle.Left, fRectangle.Top,
              fRectangle.Left + aWidth - 1, fRectangle.Bottom );

          fChildNodes[1] := T2DBitmapTreeNode.Create( fRectangle.Left + aWidth, fRectangle.Top,
              fRectangle.Right, fRectangle.Bottom );
      end
      else begin
          fChildNodes[0] := T2DBitmapTreeNode.Create( fRectangle.Left, fRectangle.Top,
              fRectangle.Right, fRectangle.Top + aHeight - 1 );

          fChildNodes[1] := T2DBitmapTreeNode.Create( fRectangle.Left, fRectangle.Top + aHeight,
              fRectangle.Right, fRectangle.Bottom );
      end;

      // insert into first child we created
      result := fChildNodes[0].Insert( aWidth, aHeight );
  end;
end;

// ============================================================================
// ============================================================================

function CompareTriangleIntersections(anItem1, anItem2: Pointer): Integer;
var distance1, distance2: single;
begin
    distance1 := PTriangleIntersection(anItem1).distance;
    distance2 := PTriangleIntersection(anItem2).distance;

    if distance1 > distance2 then
        result := 1
    else if distance1 < distance2 then
        result := -1
    else
        result := 0;
end;

// ============================================================================

procedure TTriangleIntersections.Notify(aPointer: Pointer; anAction: TListNotification);
begin
  if anAction in [lnDeleted, lnExtracted] then Dispose(aPointer);
end;

// ============================================================================

function TTriangleIntersections.GetTriangleIntersection(const anIndex: integer): TTriangleIntersection;
var
  intersection: PTriangleIntersection;
  
begin
  intersection := PTriangleIntersection(Inherited Items[anIndex]);

  result := intersection^;
end;

// ============================================================================

procedure TTriangleIntersections.Add(const aTriangleIntersection: TTriangleIntersection);
var
  triangleIntersection: PTriangleIntersection;

begin
  New(triangleIntersection);
  
  triangleIntersection^ := aTriangleIntersection;

  Inherited Add(triangleIntersection);
end;

// ============================================================================

procedure TTriangleIntersections.Add(const aFace: TFaceInfo; const aDistance, aU, aV: single);
var
  intersection: TTriangleIntersection;
  
begin
  intersection.Face := aFace;
  intersection.Distance := aDistance;
  intersection.U := aU;
  intersection.V := aV;

  Add(intersection);
end;

// ============================================================================

// Note, PJ: Also removes duplicates (afer sorting)
procedure TTriangleIntersections.Sort;
var
  index: integer;

begin
  Inherited Sort(CompareTriangleIntersections);

  // Remove duplicates... because of sort, they will be subsequent in the list
  for index := Count - 1 downto 1 do
      if Items[index-1].Face.Polygon = Items[index].Face.Polygon then
          Delete(index);
end;

// ============================================================================

function TTriangleIntersections.ContainsTransparentTriangles: boolean;
var
  index: integer;
  intersection: TTriangleIntersection;

begin
  result := false;

  for index := 0 to Count - 1 do begin
      intersection := Items[index];

      if intersection.Face.Polygon.Material.IsTransparent then begin
          result := true;
          break;
      end;
  end;
end;

// ============================================================================

end.

