{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_LightMapper;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  The TLightMapper class is a wrapper for the actual LightMapRenderSystem.
  It triangulates the scene, makes sure the normals are generated, passes the
  globals into properties, saves the generated lightmaps to disk... all the
  stuff the actual renderer shouldn't be bothered with.

  The vertex normals are generated semi-intelligent (crease-angle based). At
  some point in time this should be changed into concepts like smoothing
  groups and explicit editing of normals, integrated in DeleD.


  ---------------------------------------------------------------------------- }

interface

uses
    Contnrs,
    Classes, FileUtil,                    // Delphi
    frmProgressForm,
    unit_Scene,
    unit_Types,
    unit_Lights,
    unit_Textures,
    unit_PrimitiveList,
    unit_LightMapRenderSystem,
    unit_TextureTypes,                                                          
    unit_Primitives;            // DeleD

type
    TLightMapper = class(TObject)
    private
        fScene: TScene;
        fLightMapRenderSystem: TLightmapRenderSystem;
        fProgressForm: TfrmProgress;

        fShouldLightmapSelectedOnly: Boolean;

        // State
        fTextures: TStringList; // filled in two stages: first the filenames, then the actual ttextures as well
        fIsCanceled: Boolean;

        procedure RenderLightMaps;
        procedure SaveLightMaps;
        procedure UpdateMaterials;
        procedure UpdateScene;

        procedure HandleProgress( const aPercentageDone: Integer );
        {$IFNDEF CONSOLE_TEST}
        procedure HandleCancel( aSender: TObject );
        {$ENDIF}

        procedure InitializePrimitives;
        procedure InitializePrimitive(const aPrimitive: TBasePrimitive);

        function LightMapDirectory: String;

    protected
        function GetProgressForm: TfrmProgress;
        function GetTextures: TStringList;
        function GetShouldConsiderSelectedOnly: Boolean;

        property Scene: TScene read fScene write fScene;
        property ProgressForm: TfrmProgress read GetProgressForm;
        property Textures: TStringList read GetTextures;
        property ShouldConsiderSelectedOnly: Boolean read GetShouldConsiderSelectedOnly;
        property ShouldLightmapSelectedOnly: Boolean read fShouldLightmapSelectedOnly;

    public
        procedure Process;
        procedure RemoveLightmaps;

        constructor Create(const aScene: TScene; const anIsSelectedOnly: Boolean);
        destructor Destroy; override; 
    end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
    Forms,
    StrUtils,
    Math,
    Dialogs,
    SysUtils,                       // Delphi
    unit_Enums,
    unit_Variables,
    unit_Materials,
    unit_Math,
    unit_Constants,
    unit_TextureWriter,
    unit_TextureLoader,
    unit_MaterialList,
    unit_ParamList;

constructor TLightMapper.Create(const aScene: TScene; const anIsSelectedOnly: Boolean);
var selectedPolygonsOnly: boolean;
begin
    inherited Create;

    fScene := aScene;
    fShouldLightmapSelectedOnly := anIsSelectedOnly;

    // Initialize LightmapRenderSystem and pass global settings.
    selectedPolygonsOnly := anIsSelectedOnly and ((editMode = emPolygon) or (editMode = emFace));
    fLightMapRenderSystem := TLightmapRenderSystem.Create(aScene, anIsSelectedOnly, selectedPolygonsOnly, ShouldConsiderSelectedOnly);
    fLightMapRenderSystem.LightMapSize := lightmapSize;
    fLightMapRenderSystem.MinimumTextureSize := lightmapMinTextureSize;
    fLightMapRenderSystem.MaximumTextureSize := lightmapMaxTextureSize;
    fLightMapRenderSystem.ShouldGenerateShadows := toggleLightmapGenerateShadows;
    fLightMapRenderSystem.ShouldUseAmbientColour := toggleLightmapUseAmbientColor;
    fLightmapRenderSystem.ShadowOpacity := Scene.ShadowOpacity / 100;
    fLightMapRenderSystem.NrBlurFilterPasses := lightmapNrBlurFilterPasses;
    fLightMapRenderSystem.OnProgress := HandleProgress;
end;

// ----------------------------------------------------------------------------

destructor TLightMapper.Destroy;
begin
    FreeAndNil( fLightMapRenderSystem );
    FreeAndNil( fProgressForm );
    FreeAndNil( fTextures );

    inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure TLightMapper.RenderLightmaps;
begin
    {$IFNDEF CONSOLE_TEST}
    progressForm.OnCancel := HandleCancel;
    progressForm.Description := 'Rendering lightmaps...';
    progressForm.Max := 100;
    {$ENDIF}

    fLightmapRenderSystem.Start;

    {$IFNDEF CONSOLE_TEST}  
    // You can only cancel during the rendering phase
    progressForm.OnCancel := nil;
    {$ENDIF}
end;

// ----------------------------------------------------------------------------

procedure TLightMapper.SaveLightMaps;
var i, lightMapIndex, index: integer;
    texture: TTexture;
    textureData: TTextureData;
    fileName, directory: string;
    textureWriter: TTextureWriter;
    lightmapBaseName: string;
begin
    // figure out the next lightmap index (highest existing index parsed from filenames + 1)
    lightmapBaseName := ChangeFileExt(ExtractFileName(Scene.FileName), '') + '_' + LIGHTMAP_FILENAME_SUFFIX;
    lightmapIndex := 0;
    for i := 0 to Scene.MaterialManager.Lightmaps.Count - 1 do begin
        texture := Scene.MaterialManager.Lightmaps[i];
        fileName := ChangeFileExt(ExtractFileName(texture.FileName), '');
        if texture.isValid then begin
            if LeftStr(fileName, Length(lightmapBaseName)) = lightmapBaseName then begin
                index := StrToIntDef(RightStr(fileName, Length(fileName)-Length(lightmapBaseName)),0);
                lightmapIndex := Max(lightmapIndex, index + 1);
            end;
        end;
    end;

    // Create lightmap folder ($DELED_TEXTURE_DIR)\LightMaps\($DMF_NAME)
    directory := dir_texture + 'LightMaps';
    CreateDirUTF8(directory );
    directory := LightMapDirectory;
    CreateDirUTF8(directory );

    // Store the lightmap bitmaps.
    textureWriter := TTextureWriter.Create;
    try
        textureWriter.CompressionQuality := lightmapCompressionQuality;
        for i := 0 to fLightmapRenderSystem.NumLightMaps - 1 do
        begin
            textureData := fLightMapRenderSystem.LightMaps[i];

            fileName := directory + '\' + lightmapBaseName + IntToStr(lightMapIndex) + '.' + lightmapFileFormat;
            textureWriter.SaveTextureData(textureData, fileName);

            Textures.Add(fileName); // Store filename
            lightMapIndex := lightMapIndex + 1;
        end;
    finally
        FreeAndNil(textureWriter);
    end;
end;

// ----------------------------------------------------------------------------

procedure TLightMapper.UpdateMaterials;
var index: integer;
    texture: TTexture;
    fileName: string;
begin
    for index := 0 to Textures.Count - 1 do
    begin
        fileName := Textures[index];
        {$IFNDEF CONSOLE_TEST}
        texture := TTexture.create(fileName, Scene.MaterialManager.Lightmaps.OGLContext.H_DC, Scene.MaterialManager.Lightmaps.OGLContext.H_RC);
        {$ELSE}
        texture := TTexture.create(fileName, 0, 0);
        {$ENDIF}
        texture.MipMap := false;

        Textures.Objects[index] := texture; // register TTexture, for efficient lookup when updating scene
        Scene.MaterialManager.LightMaps.AddTexture(texture);
    end;
end;

// ----------------------------------------------------------------------------

procedure TLightMapper.UpdateScene;
var index: integer;
    face: TFaceInfo;
    polygon: TPolygon;
    lightMap: TTexture;
begin
    for index := 0 to fLightMapRenderSystem.NumFaces - 1 do
    begin
        face := fLightMapRenderSystem.Faces[index];
        polygon := face.Polygon;

        if ( face.LightMapIndex >= 0 ) and ( face.LightMapIndex <= Textures.Count ) then begin
            lightMap := Textures.Objects[face.LightMapIndex] as TTexture;
            polygon.Material := Scene.MaterialManager.MaterialForMaterialAndLightmap(polygon.Material, lightMap);
        end
        else
            raise Exception.Create( Format( 'Invalid lightmap index %d encountered.', [face.LightmapIndex] ));
    end;
end;

// ----------------------------------------------------------------------------

procedure TLightMapper.Process;
begin
    InitializePrimitives;
    RenderLightmaps;

    if not fIsCanceled then
    begin
        if not ShouldLightmapSelectedOnly then
            Scene.MaterialManager.Lightmaps.InvalidateTextures;

        SaveLightMaps;
        UpdateMaterials;
        UpdateScene;
    end;

    // mark all lightmaps that are no longer used as invalid
    if ShouldLightmapSelectedOnly then
        Scene.MaterialManager.InvalidateDanglingLightmaps;

    Scene.MaterialManager.DeleteInvalidMaterials;
    Scene.MaterialManager.DeleteInvalidLightmaps;
end;

// ----------------------------------------------------------------------------

// Removes all lightmapping information from the scene, and all lightmap files
// from disk.
procedure TLightMapper.RemoveLightmaps;
var originalMaterials, similarMaterials: TMaterialList;
    originalMaterial: TMaterial;
    i: integer;
    fParamList: TParamList;
begin
    originalMaterials := Scene.MaterialManager.GetOriginalMaterials;
    for i := 0 to originalMaterials.Count - 1 do begin
        originalMaterial := originalMaterials[i];
        similarMaterials := Scene.MaterialManager.SimilarMaterialsForMaterial(originalMaterial);
        fParamList := TParamList.Create(false);
        fParamList.Add(similarMaterials,'currentMaterials',true);
        fParamList.Add(originalMaterial,'newMaterial');
        fParamList.Add(false,'doSelectionOnly');
        Scene.ToolExecuter.Execute(ptFaceMaterialReplacer,taNotApplicable,fParamList);
        FreeAndNil(fParamList);
        FreeAndNil(similarMaterials);
    end;
    FreeAndNil(originalMaterials);

    Scene.MaterialManager.DeleteNonOriginalMaterials;
    Scene.MaterialManager.DeleteLightmaps;
end;

// ----------------------------------------------------------------------------

procedure TLightMapper.HandleProgress( const aPercentageDone: Integer );
begin
    {$IFNDEF CONSOLE_TEST}
    progressForm.SubTitle := 'Generating Lightmaps...';
    ProgressForm.Position := aPercentageDone;
    {$ENDIF}
end;

// ----------------------------------------------------------------------------
{$IFNDEF CONSOLE_TEST}
procedure TLightMapper.HandleCancel( aSender: TObject );
begin
    if not fIsCanceled then begin
        fIsCanceled := TRUE;
        fLightMapRenderSystem.Stop;
    end;
end;
{$ENDIF}
// ----------------------------------------------------------------------------

procedure TLightMapper.InitializePrimitives;
var index: integer;
    list: TPrimitiveList;
    primitive: TBasePrimitive;
begin
    //I removed the old Scene.InitializeNormals method.
    //The loop will be executed here.
    for index := 0 to Scene.Primitives.Count - 1 do
    begin
        primitive := Scene.Primitives[index];
        primitive.InitializeNormals(lightmapCreaseAngle);
    end;

    if ShouldConsiderSelectedOnly then
    begin
        list := Scene.SelectedPrimitives.Copy;
    end
    else
        list := Scene.Primitives.GetItems(TPrimitive);  // Getitems call probably not needed

    for index := 0 to list.Count-1 do begin
        primitive := list[index];

        InitializePrimitive(primitive);
    end;
    FreeAndNil(list);
end;

// ----------------------------------------------------------------------------

// NOTE, PJ: Sets the AllowLightmap state... (degenerate polygons, lightmap-disabled materials) 
procedure TLightMapper.InitializePrimitive(const aPrimitive: TBasePrimitive);
var polygonIndex: integer;
    polygon: TPolygon;
    normal: TVertex;

begin
    for polygonIndex := 0 to aPrimitive.NumPolygons-1 do begin
        polygon := aPrimitive.Polygons[polygonIndex];

        polygon.AllowLightMap := true;

        normal := verticeToVertex(polygon.Normal);

        if  ( polygon.NumVertices < 3 ) or
            (normal.X = 0) and (normal.Y = 0) and (normal.Z = 0) then begin
            // Degenerate polygon

            polygon.AllowLightmap := false;
        end
        else
            polygon.AllowLightMap := polygon.Material.EnableLightmap;

        // degenerate faces might store old materials -> same for materials that just got lightmap-disabled.
        if (not polygon.AllowLightMap) and (not polygon.Material.IsOriginal) then
            polygon.Material := polygon.Material.OriginalMaterial;
    end;
end;

// ----------------------------------------------------------------------------

function TLightMapper.LightMapDirectory: String;
var fileName: String;
begin
    fileName := ChangeFileExt( ExtractFileName( Scene.fileName ), '' );

    result :=  dir_texture + 'LightMaps\' + fileName;
end;

// ----------------------------------------------------------------------------

function TLightMapper.GetProgressForm: TfrmProgress;
// INFO, PJ: Resolves the progress form if it doesn't exist yet;
begin
    if not Assigned( fProgressForm ) then
    begin
        fProgressForm := TfrmProgress.Create(nil);
        fProgressForm.Title := 'Lightmapping...';
        fProgressForm.Show;
    end;

    result := fProgressForm;
end;

// ----------------------------------------------------------------------------

function TLightMapper.GetTextures: TStringList;
begin
    if not Assigned( fTextures ) then
    begin
        fTextures := TStringList.Create;
    end;

    result := fTextures;
end;

// ----------------------------------------------------------------------------

function TLightMapper.GetShouldConsiderSelectedOnly: Boolean;
begin
    result := ( fShouldLightmapSelectedOnly and not toggleLightmapUseNonSelectedObjects );
end;

// ----------------------------------------------------------------------------

end.
