{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Lights;

{$MODE Delphi}

interface

uses SysUtils,Graphics,Classes,Dialogs,
     unit_Enums,
     unit_Types,
     unit_Math,
     unit_Primitives;

type
  TLightClass = class of TLight;

  TLight = class(TBasePrimitive)                // base class for light objects (a light is a 3D object in DeleD)
  private
    fPosition: TVertice;                        // position in 3D space
    fDiffuse: TRGBColor;                        // diffuse color of the light (defaul=white)
    fSpecular: TRGBColor;                       // specular color of the light (default=white; this is the "highlight" color)
    fFlash: integer;                            // light is flashing with a specific timeinterfall (seconds)
    fActive: boolean;                           // is light active or not
    fCastShadows: boolean;
    fIntensity: integer;                        // 0=no intensity; 100=full intensity (default=full intensity)
    fDirection: TVertice;
    fPreviousDirection: TVertice;
    fRange: longint;
    fConstant: single;                          // attenuation variables
    fLinear: single;
    fQuadratic: single;
    fSnap: boolean;
  protected
    procedure Obj_RotateInternal(matrix: TMatrix);

    property PreviousDirection: TVertice read fPreviousDirection write fPreviousDirection;
  public
    constructor Create; override;
    destructor Destroy; override;
    procedure Assign(Source: TBasePrimitive); override;
    procedure FaceDirection;
    procedure UpdateDirection;
    property Position: TVertice read fPosition write fPosition;
    property Diffuse: TRGBColor read fDiffuse write fDiffuse;
    property Specular: TRGBColor read fSpecular write fSpecular;
    property Flash: integer read fFlash write fFlash;       // TODO: use or remove
    property Active: boolean read fActive write fActive;
    property CastShadows: boolean read fCastShadows write fCastShadows;
    property Intensity: integer read fIntensity write fIntensity;
    property Direction: TVertice read fDirection write fDirection;
    property Range: longint read fRange write fRange;
    property Constant: single read fConstant write fConstant;
    property Linear: single read fLinear write fLinear;
    property Quadratic: single read fQuadratic write fQuadratic;
    property Snap: boolean read fSnap write fSnap;
  end;

  TPointLight = class(TLight)                   // lightbeams in all directions from one defined point in space
  private
  public
    constructor Create; override;
    procedure Assign(Source: TBasePrimitive); override;
  end;

  TDirectionalLight = class(TLight)             // parralel lightbeams coming from one direction from an infinity lightsource
  private
  protected
  public
    constructor Create; override;
  end;

  TSpotLight = class(TLight)                    // lightbeams going in a certain direction from one defined point in space
  private
    fInnerangle: integer;
    fOuterangle: integer;
    fFalloff: integer;
  protected
  public
    constructor Create; override;
    procedure Assign(Source: TBasePrimitive); override;
    property InnerAngle: integer read fInnerangle write fInnerangle;
    property OuterAngle: integer read fOuterangle write fOuterangle;
    property FallOff: integer read fFalloff write fFalloff;
  end;

implementation

uses unit_Functions;

{ ### TLight ################################################################# }

procedure TLight.Assign(Source: TBasePrimitive);
begin
    if Source is TLight then begin
        Position.Assign(TLight(Source).Position);
        Diffuse := TLight(Source).Diffuse;
        Specular := TLight(Source).Specular;
        Flash := TLight(Source).Flash;
        Active := TLight(Source).Active;
        Snap := TLight(Source).Snap;
        Intensity := TLight(Source).Intensity;
        CastShadows := TLight(Source).CastShadows;
        Tag := TLight(Source).Tag;
        Direction.Assign(TLight(Source).Direction);
        PreviousDirection.Assign(TLight(Source).PreviousDirection);
    end;
    inherited;
end;

constructor TLight.Create;
begin
    inherited Create;
    ModifyGeometry := false;
    Name := 'light';
    fPosition := TVertice.Create(0,0,0);
    fDiffuse := RGBColor(1, 1, 1);
    fSpecular := RGBColor(1, 1, 1);
    fCastShadows := TRUE;
    Active := true;
    Flash := 0;         // not used
    Snap := false;
    Intensity := 100;
    SnapMethod := smNone;

    // Prefabs point up, actual lights point down.
    // JCC 26-04-2009: changed to 0,1,0 because directional
    // and spotlights seemed to have an inverted direction...
    fDirection := TVertice.Create(0,1,0);
    fPreviousDirection := TVertice.Create(0,1,0);
end;

destructor TLight.Destroy;
begin
    inherited;
    FreeAndNil(fPosition);
    FreeAndNil(fDirection);
    FreeAndNil(fPreviousDirection);
end;

procedure TLight.FaceDirection;
var axis: TVertice;
    angle: Double;
    rotationMatrix: TMatrix;
    absDirection: TVertex;
begin
    angle := vectorAngle( fPreviousDirection, direction );
    axis := crossProduct( fPreviousDirection, direction );

    // INFO, PJ: Parrallel or anti-parrallel vertices, special case. Experimental handling,
    //           works fine for current (0,1,0) -> (0,-1,0) handling.
    if axis.Length < 0.005 then begin
        FreeAndNil( axis );
        absDirection := Vertex( Abs( direction.X ), Abs( direction.Y ), Abs( Direction.Z ));
        if (absDirection.X <= absDirection.Y) and (absDirection.X <= absDirection.Z) then begin
            axis := TVertice.Create( 1, 0, 0 );
        end
        else if (absDirection.Y <= absDirection.X) and (absDirection.Y <= absDirection.Z) then begin
            axis := TVertice.Create( 0, 1, 0 );
        end
        else
            axis := TVertice.Create( 0, 0, 1 );
    end;

    axis.Normalize;
    rotationMatrix := MatrixArbAxisSetup( axis, Position, angle );

    // INFO, PJ: We call the internal version, that doesn't rotate the direction
    Obj_RotateInternal( rotationMatrix );

    fPreviousDirection.Assign( direction );

    FreeAndNil( axis );
    FreeAndNil( rotationMatrix );
end;

// Call updateDirection whenever the direction is changed through anything but a FaceDirection
// (for example, after manually rotating the light)
procedure TLight.UpdateDirection;
begin
    fPreviousDirection.Assign( direction );
end;

procedure TLight.Obj_RotateInternal(matrix: TMatrix);
var i: integer;
    x,y,z: double;
    vertex: TVertice;
begin
    for i := 0 to Vertices.Count-1 do begin
        vertex := Vertices[i];
        vertex.x := vertex.x-matrix[3,0];
        vertex.y := vertex.y-matrix[3,1];
        vertex.z := vertex.z-matrix[3,2];
        x := (vertex.x*matrix[0,0])+(vertex.y*matrix[0,1])+(vertex.z*matrix[0,2])+matrix[3,0];
        y := (vertex.x*matrix[1,0])+(vertex.y*matrix[1,1])+(vertex.z*matrix[1,2])+matrix[3,1];
        z := (vertex.x*matrix[2,0])+(vertex.y*matrix[2,1])+(vertex.z*matrix[2,2])+matrix[3,2];
        vertex.x := x;
        vertex.y := y;
        vertex.z := z;
    end;
end;


{ ### TPointLight ############################################################ }

constructor TPointLight.Create;
begin
    inherited Create;
    Name := 'pointlight';
    ObjectType := otPointlight;
    Constant := 1.0;
    Linear := 0;
    Quadratic := 0.000003;
    Range := 1000;
end;

procedure TPointLight.Assign(Source: TBasePrimitive);
begin
    if Source is TPointlight then begin
        Constant := TPointlight(Source).Constant;
        Linear := TPointlight(Source).Linear;
        Quadratic := TPointlight(Source).Quadratic;
        Range := TPointlight(Source).Range;
    end;
    inherited;
end;

{ ### TDirectionalLight ###################################################### }

constructor TDirectionalLight.Create;
begin
    inherited Create;
    name := 'directionallight';
    objecttype := otDirectionalLight;
end;

{ ### TSpotLight ############################################################# }

procedure TSpotLight.Assign(Source: TBasePrimitive);
begin
    if Source is TSpotLight then begin
        constant := TSpotLight(Source).constant;
        linear := TSpotLight(Source).linear;
        quadratic := TSpotLight(Source).quadratic;
        range := TSpotLight(Source).range;
        innerangle := TSpotLight(Source).innerangle;
        outerangle := TSpotLight(Source).outerangle;
        falloff := TSpotLight(Source).falloff;
    end;
    inherited;
end;

constructor TSpotLight.Create;
begin
    inherited Create;
    name := 'spotlight';
    objecttype := otSpotlight;
    constant := 1.0;
    linear := 0;
    quadratic := 0;
    range := 1000;
    innerangle := 30;
    outerangle := 40;
    falloff := 1;
end;


end.
