{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_MaterialList;

{$MODE Delphi}

interface

uses Windows,dglOpenGL,Classes,Contnrs,SysUtils,Dialogs,
     frmOGLContextForm,
     unit_Enums,
     unit_Functions,
     unit_Textures,
     unit_Materials,
     unit_TextureTypes;

type
    // TODO, JC: refactor this class; it is a list and should not have to
    // worry about concepts like categories and things like that...
    TMaterialList = class(TObjectList)
    private
      fTextures: TTextureList;
      fLightMaps: TTexturelist;
      fOnSelectMaterial: TNotifyEvent;
      fSelectedMaterial: TMaterial;
      function GetHighestID: integer;
      function GetNumMaterials: integer;
      function GetSelectedMaterial: TMaterial;
      procedure SetSelectedMaterial(AMaterial: TMaterial);
    protected
      function GetMaterial(const anIndex: integer): TMaterial;
    public
      constructor Create(aOwnsObjects: boolean);
      destructor Destroy; override;
      function AddMaterial(const aMaterial: TMaterial): TMaterial; overload;
      function AddMaterial(const aTexture: TTexture; aCategory: string; aBlendOperation: TBlendOperation): TMaterial; overload;
      function GetMaterialByID(const anID: integer): TMaterial;
      function GetMaterialByName(const aName: string): TMaterial;
      procedure RemoveLayer(const aMaterial: TMaterial; aLayer: TTextureLayer);
      procedure RemoveTexture(const aTexture: TTexture);
      procedure RemoveUnusedTextures;
      function GetOriginalMaterials: TMaterialList;
      function TextureIsUsed(aTexture: TTexture): integer;
      function SimilarMaterialsForMaterial(const aMaterial: TMaterial): TMaterialList;
      function NameIsUnique(aName: string): boolean;
      function CreateUniqueName(aName: string = 'material'): string;
      function IndexOfMaterial(const aMaterial: TMaterial): integer;
      function ExtractCategories: TStringList;
      function CategoryIsUsed(aCategoryName: string): boolean;
      function ExtractMaterialsForCategory(aCategoryName: string): TMaterialList;
      procedure Assign(const aMaterialList: TMaterialList);
      procedure Clear; override;
      property OnSelectMaterial: TNotifyEvent read fOnSelectMaterial write fOnSelectMaterial;
      property SelectedMaterial: TMaterial read GetSelectedMaterial write SetSelectedMaterial;
      property HighestID: integer read GetHighestID;
      property NumMaterials: integer read GetNumMaterials;
      property Textures: TTextureList read fTextures write fTextures;           // unique textures as used by materials
      property Lightmaps: TTextureList read fLightMaps write fLightMaps;        // lightmaps (separately stored)
      property Materials[const anIndex: integer]: TMaterial read GetMaterial; default;
    end;

implementation

// ### TMaterialList ###########################################################

// To make sure the answer returned is valid, this function also resolves ID's
// for any material that don't have an ID assigned yet.
function TMaterialList.GetHighestID: integer;
var i: integer;
    material: TMaterial;
begin
    result := 0;
    
    for i := 0 to Count-1 do begin
        material := Materials[i];
        if material.ID > result then result := material.ID;
    end;

    // assign ID's to materials that don't have one yet
    for i := 0 to Count-1 do begin
        material := Materials[i];
        if material.ID = -1 then begin
           result := result + 1;
           material.ID := result;
        end;
    end;
end;

function TMaterialList.GetNumMaterials: integer;
begin
    Result := Count;
end;

// if no material is selected, return the system material
function TMaterialList.GetSelectedMaterial: TMaterial;
begin
    if not Assigned(fSelectedMaterial) then
        fSelectedMaterial := Materials[0];
    Result := fSelectedMaterial;
end;

procedure TMaterialList.SetSelectedMaterial(AMaterial: TMaterial);
begin
    if Assigned(AMaterial) then begin
        fSelectedMaterial := AMaterial;
        if Assigned(fOnSelectMaterial) then
            fOnSelectMaterial(fSelectedMaterial);    // notify the world a material was selected
    end;
end;

function TMaterialList.GetMaterial(const anIndex: integer): TMaterial;
begin
    if anIndex < Count then
        Result := GetItem(anIndex) as TMaterial
    else
        raise Exception.Create(Format('No material found for index %d', [anIndex]));
end;

// if a similar material is already in the list, that material is returned instead
function TMaterialList.AddMaterial(const aMaterial: TMaterial): TMaterial;
var material: TMaterial;
    i: integer;
begin
    Result := nil;
    for i := 0 to Count-1 do begin
        material := GetMaterial(i);

        if material.Equals(aMaterial) then begin
            Result := material;
            break;
        end;
    end;

    if not Assigned(Result) then begin
        Add(aMaterial);
        Result := aMaterial;
    end;
end;

function TMaterialList.AddMaterial(const aTexture: TTexture; aCategory: string; aBlendOperation: TBlendOperation): TMaterial;
begin
    Result := nil;
end;

function TMaterialList.GetMaterialByID(const anID: integer): TMaterial;
var counter: integer;
    found: boolean;
    material: TMaterial;
begin
    Result := nil;
    counter := 0;
    found := false;
    while (counter<NumMaterials) and (not found) do begin
        material := Materials[counter];
        if material.ID=anID then begin
            found := true;
            Result := material
        end else
            inc(counter);
    end;
end;

function TMaterialList.GetMaterialByName(const aName: string): TMaterial;
var i: integer;
    material: TMaterial;
begin
    Result := nil;

    for i := 0 to NumMaterials - 1 do begin
        material := Materials[i];

        if material.Name = aName then begin
            result := material;
            break;
        end;
    end;
end;

// removes a layer from a specified material and also removes the texture from the
// texturelist if this texture is not used by other materials anymore
// note: should be here because materiallist is responsible for maintaining a list of unique textures
procedure TMaterialList.RemoveLayer(const aMaterial: TMaterial; aLayer: TTextureLayer);
var index: integer;
begin
    if Assigned(aMaterial) then begin
        index := aMaterial.IndexOfLayer(aLayer);            // index should equal ID
        if index<>-1 then begin     // layer exists
            if aLayer.LayerType=mlTexture then begin        // is it a texture-layer?
                if TextureIsUsed(alayer.Texture)=1 then     // texture is only used by this particular layer
                    Textures.Remove(aLayer.Texture);        // so remove it from the texturelist

            end;
            aMaterial.RemoveTextureLayer(index);
        end;
    end;
end;

// removes a texture from the texturelist if it is NOT used by any material-layer
procedure TMaterialList.RemoveTexture(const aTexture: TTexture);
begin
    if Assigned(aTexture) then begin
        if TextureIsUsed(aTexture)=0 then
            Textures.Remove(aTexture);
    end;
end;

// removes all textures from the texturelist that are not used by any material-layer anymore
procedure TMaterialList.RemoveUnusedTextures;
begin
    showmessage('not implemented yet');
end;

function TMaterialList.GetOriginalMaterials: TMaterialList;
var i: integer;
    material: TMaterial;
begin
    Result := TMaterialList.Create(False);
    for i := 0 to Count - 1 do begin
        material := GetMaterial(i);
        if material.IsOriginal then Result.Add(material);
    end;
end;

// returns the # of times a particular texture is used by the listed materials (scans material-layers)
function TMaterialList.TextureIsUsed(aTexture: TTexture): integer;
var i,count: integer;
    material: TMaterial;
begin
    count := 0;
    for i := 0 to numMaterials-1 do begin
        material := Materials[i];
        count := count + material.TextureIsUsed(aTexture);
    end;
    Result := count;
end;

function TMaterialList.SimilarMaterialsForMaterial(const aMaterial: TMaterial): TMaterialList;
var i: integer;
    originalMaterial, material: TMaterial;
begin
    Result := TMaterialList.Create(false);
    if Assigned(aMaterial.OriginalMaterial) then
        originalMaterial := aMaterial.OriginalMaterial
    else
        originalMaterial := aMaterial;
    for i := 0 to Count - 1 do begin
        material := Get(i);
        if (material = aMaterial) or                                            // note, don't use .equal here! Identical test.
           (material.OriginalMaterial = originalMaterial) then begin
            Result.Add(material);
        end;
    end;
end;

function TMaterialList.NameIsUnique(aName: string): boolean;
var i: integer;
    material: TMaterial;
    unique: boolean;
begin
    unique := true;
    for i := 0 to numMaterials-1 do begin
        material := Materials[i];
        if material.Name=aName then begin
            unique := false;
            break;
        end;
    end;
    Result := unique;
end;

function TMaterialList.CreateUniqueName(aName: string = 'material'): string;
var value: integer;
    name: string;
begin
    value := Count;
    name := aName+IntToStr(value);
    while not NameIsUnique(name) do begin
        inc(value);
        name := aName+IntToStr(value);
    end;
    Result := name;
end;

function TMaterialList.IndexOfMaterial(const aMaterial: TMaterial): integer;
begin
    Result := IndexOf(aMaterial);
end;

// NOTE JC: we could rewrite the classmodel so it supports categories by default
function TMaterialList.ExtractCategories: TStringList;
var i: integer;
    categories: TStringList;
begin
    categories := TStringList.Create;
    for i := 0 to nummaterials-1 do begin
        if categories.IndexOf(Materials[i].Category)=-1 then
            categories.Add(Materials[i].Category);
    end;
    Result := categories;
end;

function TMaterialList.CategoryIsUsed(aCategoryName: string): boolean;
var i: integer;
    found: boolean;
begin
    found := false;
    for i := 0 to nummaterials-1 do begin
        if Materials[i].Category=aCategoryName then begin
            found := true;
            break;
        end;
    end;
    Result := found;
end;

// retrieves all materials from the list that have a particular category name
function TMaterialList.ExtractMaterialsForCategory(aCategoryName: string): TMaterialList;
var i: integer;
    materialList: TMaterialList;
begin
    materialList := TMaterialList.Create(false);
    for i := 0 to Count-1 do begin
        if Materials[i].Category=aCategoryName then
            materialList.Add(Materials[i]);
    end;
    Result := materialList;
end;

// only assigns the materials, does nothing with ownership of textures and lightmaps
procedure TMaterialList.Assign(const aMaterialList: TMaterialList);
var i: integer;
    material: TMaterial;
begin
    inherited Clear;

    for i := 0 to aMaterialList.Count - 1 do begin
        material := aMaterialList.GetMaterial(i);
        Add(material.Copy);
    end;
end;

procedure TMaterialList.Clear;
begin
    if Assigned(Textures) then
        Textures.Clear;

    if Assigned(Lightmaps) then
        Lightmaps.Clear;

    fSelectedMaterial := nil;

    inherited Clear;
end;

constructor TMaterialList.Create(aOwnsObjects: boolean);
begin
    inherited Create(AOwnsObjects); // JuMa: call to inherited constructor was missing

    fTextures := TTextureList.Create(true);
    fTextures.MipMap := true;
    fTextures.AddressingMode := tamWrap;
    fTextures.ShouldFilter := true;

    fLightMaps := TTextureList.Create(true);
    fLightMaps.MipMap := false;
    fLightMaps.AddressingMode := tamClamp;
    fLightMaps.ShouldFilter := true;
end;

destructor TMaterialList.Destroy;
begin
    FreeAndNil(fTextures);
    FreeAndNil(fLightMaps);
    inherited;
end;

end.
