{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_MaterialManager;

{$MODE Delphi}

interface

uses Windows,dglOpenGL,Classes,Contnrs,SysUtils,StrUtils, FileUtil,
     unit_Functions,
     unit_PrimitiveList,
     unit_IntegerList,
     unit_Textures,
     unit_Materials,
     unit_MaterialList;

type
    TMaterialManager = class(TMaterialList)
      fTextureDirectory: string;
      fScenePrimitives: TPrimitiveList;
    public
      constructor Create(owns: boolean; AScenePrimitiveList: TPrimitiveList); overload;
      procedure InvalidateDanglingLightmaps;
      procedure DeleteInvalidLightmaps;
      procedure DeleteDanglingLightmappedMaterials;
      procedure DeleteLightmaps;
      procedure DeleteNonOriginalMaterials;
      procedure DeleteInvalidMaterials;
      function CreateUniqueTextures: TTextureList;
      function ExtractUniqueLightmaps: TTextureList;
      function ReadTexturepack(f: string): boolean;
      function MaterialForMaterialAndLightmap(const aMaterial: TMaterial; const aLightmap: TTexture): TMaterial;
      procedure UpdateMaterialUsage;
      property TextureDirectory: string read fTextureDirectory write fTextureDirectory;
      property ScenePrimitives: TPrimitiveList read fScenePrimitives write fScenePrimitives;
    end;

implementation

uses unit_Types,
     unit_Primitives;

// ### TMaterialManager ########################################################

constructor TMaterialManager.Create(owns: boolean; AScenePrimitiveList: TPrimitiveList);
begin
    inherited Create(owns);
    fScenePrimitives := AScenePrimitiveList;
end;

// mark all lightmaps that are no longer used as invalid
procedure TMaterialManager.InvalidateDanglingLightmaps;
var isLightmapUsed: array of Boolean;
    i,j,index: integer;
    list: TPrimitiveList;
    obj: TBasePrimitive;
    poly: TPolygon;
    texture: TTexture;
begin
    SetLength(isLightmapUsed,Lightmaps.Count);
    for i := low(isLightmapUsed) to high(isLightmapUsed) do
        isLightmapUsed[i] := false;

    // mark all used lightmaps
    list := ScenePrimitives.GetItems(TPrimitive);
    for i := 0 to list.Count-1 do begin
        obj := list[i];
        for j := 0 to obj.numpolygons-1 do begin
            poly := obj.Polygons[j];

            if poly.Material.ContainsLightmap then begin
                texture := poly.Material.Lightmap;

                // this should probably be a hashed lookup: change some day.
                index := Lightmaps.IndexOf(texture);
                if index <> -1 then isLightmapUsed[index] := true;
            end;
        end;
    end;
    FreeAndNil(list);

    // invalidate all lightmaps not marked as used
    for i := 0 to LightMaps.Count - 1 do begin
        if not isLightmapUsed[i] then begin
            texture := Lightmaps.Textures[i];
            texture.Invalidate;
        end;
    end;
end;

// removes all invalid lightmaps form the list
// if there is no valid counterpart, it removes the file from disk as well
procedure TMaterialManager.DeleteInvalidLightmaps;
var i, index: integer;
    lightmap: TTexture;
    fileName: string;
    invalidFileNames: TStringList;
begin
    // we remove all invalid lightmaps form the list and build a list of filenames
    invalidFileNames := TStringList.Create;
    for i := LightMaps.Count - 1 downto 0 do begin
        lightmap := Lightmaps[i];
        if not Lightmap.IsValid then begin
            invalidFileNames.Add(lightmap.FileName);
            Lightmaps.Delete(i);
        end;
    end;

    // we remove each valid filename from the list of names
    for i := 0 to Lightmaps.Count - 1 do begin
        filename := Lightmaps[i].FileName;
        index := invalidFileNames.IndexOf(fileName);
        if index <> -1 then begin
            invalidFileNames.Delete(index);
        end;
    end;

    // now we delete all remaining files
    for i := 0 to invalidFileNames.Count - 1 do begin
        DeleteFileUTF8(invalidFileNames[i]);
    end;

    FreeAndNil(invalidFileNames);
end;

// deletes all lightmapping-materials that are not used anymore
procedure TMaterialManager.DeleteDanglingLightmappedMaterials;
var i,j,k: integer;
    obj: TBasePrimitive;
    poly: TPolygon;
    list: TPrimitiveList;
    isMaterialUsed: array of Boolean;
    material: TMaterial;
begin
    SetLength(isMaterialUsed,Count);
    for i := low(isMaterialUsed) to high(isMaterialUsed) do begin
        isMaterialUsed[i] := false;
    end;

    // mark all used lightmapped materials
    list := ScenePrimitives.GetItems(TPrimitive);
    for i := 0 to list.Count-1 do begin
        obj := list[i];
        for j := 0 to obj.numpolygons-1 do begin
            poly := obj.Polygons[j];

            if poly.Material.ContainsLightmap then begin
                for k := 0 to Count - 1 do begin
                    if materials[k].Equals(poly.Material) then begin
                        isMaterialUsed[k] := true;
                        break;
                    end;
                end;
            end;
        end;
    end;
    FreeAndNil(list);

    for i := Count - 1 downto 0 do
    begin
        material := GetMaterial(i);

        if material.containsLightmap and not isMaterialUsed[i] then
            Delete(i);
    end;
end;

// Deletes all lightmaps
procedure TMaterialManager.DeleteLightmaps;
var i: integer;
    lightmap: TTexture;
begin
    for i := 0 to Lightmaps.Count - 1 do begin
        lightmap := Lightmaps[i];
        DeleteFileUTF8(lightmap.FileName);
    end;

    Lightmaps.Clear;
end;

// Deletes all material that is not original. Use with care, make sure
// no polygon references these non-original (lightmapped) materials first.
procedure TMaterialManager.DeleteNonOriginalMaterials;
var i: integer;
    material: TMaterial;
begin
    for i := Count - 1 downto 0 do begin
        material := Materials[i];
        if not material.isOriginal then Delete(i);
    end;
end;

// deletes all materials that refer to invalidated textures
// deletes all materials that refer to invalidated textures
procedure TMaterialManager.DeleteInvalidMaterials;
var i: integer;
    material: TMaterial;
begin
    for i := Count-1 downto 0 do begin
        material := GetMaterial(i);

        if not material.isValid then Delete(i);
    end;
end;

// walks through the material list and creates a list with unique textures
// as used by the materials; this texturelist can be used, for example, to
// upload the individual textures to the OpenGL windows that need them
// note: this list also includes lightmaps + is obsolete because local texturelist is implemented also
function TMaterialManager.CreateUniqueTextures: TTextureList;
var i,j,k: integer;
    list: TTextureList;
    material: TMaterial;
    layer: TTextureLayer;
    IsEqual: boolean;
begin
    list := TTextureList.Create(false);

    for i := 0 to nummaterials-1 do begin

        material := GetMaterial(i);

        for j := 0 to material.NumTextureLayers-1 do begin  // all layers within this material

            layer := material.TextureLayers[j];

            if assigned(layer) then begin
                IsEqual := false;

                for k := 0 to list.Count-1 do begin     // check if the texture of the current layer is in the final list
                    if list.Textures[k].Equals(layer.Texture) then begin
                        IsEqual := true;
                        break;
                    end;
                end;

                if not IsEqual then
                    list.AddTexture(layer.Texture);
            end;
        end;

    end;
    Result := list;
end;

// this function creates a texturelist containing unique lightmaps as used by materials
function TMaterialManager.ExtractUniqueLightmaps: TTextureList;
var i,k: integer;
    list: TTextureList;
    material: TMaterial;
    texture: TTexture;
    IsEqual: boolean;
begin
    list := TTextureList.Create(false);

    for i := 0 to nummaterials-1 do begin

        material := GetMaterial(i);

        if material.ContainsLightmap then begin 
            texture := material.lightmap;

            IsEqual := false;

            for k := 0 to list.Count-1 do begin     // check if the texture of the current layer is in the final list
                if list.Textures[k].Equals(texture) then begin
                    IsEqual := true;
                    break;
                end;
            end;

            if not IsEqual then
                list.AddTexture(texture);
        end;
    end;
    Result := list;
end;

function TMaterialManager.ReadTexturepack(f: string): boolean;
var i,numcategories,numtextures,offset: integer;
    tempMaterial, material: TMaterial;
    texture: TTexture;
    tpack,categories,values: TStringList; filename: string;
begin
    if FileExistsUTF8(f) then
    begin
        Clear;

        tpack := TStringList.Create;
        tpack.LoadFromFile(f);

        offset := 1;
        numcategories := strtoint(tpack[offset]);
        inc(offset);
        categories := TStringList.Create;
        for i := 0 to numcategories-1 do begin
            values := SubdivideStringToList(tpack[offset],' ');
            categories.Add(values[1]);
            values.Free;
            inc(offset);
        end;

        numtextures := strtoint(tpack[offset]);
        inc(offset);
        for i := 0 to numtextures-1 do begin
            values := SubdivideStringToList(tpack[offset],' ');

            tempMaterial := TMaterial.Create;
            tempMaterial.ID := i;
            tempMaterial.Category := categories[strtoint(values[1])];

            filename := texturedirectory+categories[strtoint(values[1])]+'\'+values[2];
            texture := Textures.AddTexture(filename);       // store the texture in the texturelist
            tempMaterial.Name := LeftStr(values[2],length(values[2])-length(ExtractFileExt(values[2])));
            tempMaterial.AddLayerForTexture(texture);

            material := addMaterial(tempMaterial);

            if material <> tempMaterial then
                FreeAndNil(tempMaterial);

            values.Free;
            inc(offset);
        end;

        FreeAndNil(categories);
        FreeAndNil(tpack);

        Result := true;
    end
    else
        Result := false;
end;

function TMaterialManager.MaterialForMaterialAndLightmap(const aMaterial: TMaterial; const aLightmap: TTexture): TMaterial;
var i: integer;
    newMaterial: TMaterial;
    material: TMaterial;
begin
    newMaterial := aMaterial.Copy;
    newMaterial.OriginalMaterial := aMaterial.OriginalMaterial;
    if not Assigned(aMaterial.OriginalMaterial) then
        newMaterial.OriginalMaterial := aMaterial;
    newMaterial.Lightmap := aLightmap;
    newMaterial.IsVisible := false;

    // check if we already have a similar material in our collection
    material := nil;
    for i := 0 to Count - 1 do begin
        if materials[i].Equals(newMaterial) then begin
            material := materials[i];
            break;
        end;
    end;

    // if so, return it and destroy the newly created material
    if Assigned( material ) then begin
        FreeAndNil( newMaterial );
        result := material;
    end else begin // otherwise, the new material is added to the collection and returned
        Add( newMaterial );
        result := newMaterial;
    end;
end;

// Updates the 'isUsed' flag on all material to reflect the current situation.
procedure TMaterialManager.UpdateMaterialUsage;
var i, j: integer;
    primitives: TPrimitiveList;
    primitive: TBasePrimitive;
    polygon: TPolygon;
    material: TMaterial;
begin
    for i := 0 to Count - 1 do
        Materials[i].IsUsed := false;

    primitives := ScenePrimitives.GetItems(TPrimitive);
    for i := 0 to primitives.Count - 1 do begin
        primitive := primitives[i];
        for j := 0 to primitive.NumPolygons - 1  do begin
            polygon := primitive.Polygons[j];

            material := polygon.Material;
            material.IsUsed := true;

            // Make sure to mark the original material as well (for lightmapped materials)
            if Assigned( material.OriginalMaterial ) then
                material.OriginalMaterial.IsUsed := true;
        end;
    end;
    FreeAndNil(primitives);
end;

end.
