{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Materials;

{$MODE Delphi}

interface

uses Windows,dglOpenGL,Classes,Graphics,Contnrs,
     unit_Enums,
     unit_Textures;

type
    TMaterialRaytracingProperties = class(TObject)
    private
        fAmbientReflection: Single;
        fDiffuseReflection: Single;
        fSpecularReflection: Single;
        fSpecularExponent: Single;
        fSpecularColor: TColor;
        fReflection: Single;
        fReflectiveExponent: Single;
        fReflectiveColor: TColor;
        fTransparency: Single;
        fRefraction: Single;

    public
        constructor Create;
        procedure Assign(const aProperties: TMaterialRaytracingProperties);
        function Equals(const aProperties: TMaterialRaytracingProperties): Boolean;
        function IsEmpty: Boolean;
        function HasAmbientReflection: Boolean;
        function HasDiffuseReflection: Boolean;
        function HasSpecularReflection: Boolean;
        function HasSpecularExponent: Boolean;
        function HasSpecularColor: Boolean;
        function HasReflection: Boolean;
        function HasReflectiveExponent: Boolean;
        function HasReflectiveColor: Boolean;
        function HasTransparency: Boolean;
        function HasRefraction: Boolean;
        property AmbientReflection: Single read fAmbientReflection write fAmbientReflection;
        property DiffuseReflection: Single read fDiffuseReflection write fDiffuseReflection;
        property SpecularReflection: Single read fSpecularReflection write fSpecularReflection;
        property SpecularExponent: Single read fSpecularExponent write fSpecularExponent;
        property SpecularColor: TColor read fSpecularColor write fSpecularColor;
        property Reflection: Single read fReflection write fReflection;
        property ReflectiveExponent: Single read fReflectiveExponent write fReflectiveExponent;
        property ReflectiveColor: TColor read fReflectiveColor write fReflectiveColor;
        property Transparency: Single read fTransparency write fTransparency;
        property Refraction: Single read fRefraction write fRefraction;
    end;

    TTextureLayer = class(TObject)
    private
      fColor: TColor;
      fLayerType: TMaterialLayerType;
      fTexture: TTexture;
      fOperation: TBlendOperation;
      fInvalidTextureFileName: string; // INFO, PJ: Only set for invalid texture files encountered (replaced by color)
      procedure SetLayerType(value: TMaterialLayerType);
    public
      constructor Create(const aTexture: TTexture; const anOperation: TBlendOperation); overload;
      constructor Create; overload;
      destructor Destroy; override;
      function Equals(const aTextureLayer: TTextureLayer): boolean;
      function IsTransparent: boolean;
      property Color: TColor read fColor write fColor;
      property LayerType: TMaterialLayerType read fLayerType write SetLayerType;
      property Texture: TTexture read fTexture write fTexture;
      property BlendOperation: TBlendOperation read fOperation write fOperation;
      property InvalidTextureFileName: string read fInvalidTextureFileName write fInvalidTextureFileName;
    end;

    TMaterial = class(TObject)
    private
      fTextureLayers: TObjectList;      // should be a TTextureLayerList
      fOriginalMaterial: TMaterial;
      fName: string;
      fTag: string;
      fComments: string;
      fCategory: string;
      fIsVisible: boolean;          // visible in GUI (e.g. when replacing materials)
      fIsUsed: boolean;
      fEnableLightmap: boolean;
      fCastShadows: boolean;
      fReceiveShadows: boolean;
      fRaytracingProperties: TMaterialRaytracingProperties;
      fID: integer;
    protected
      function GetTextureLayerForIndex(const anIndex: Integer): TTextureLayer;
      function GetIsVisible: boolean;
      function GetLightmap: TTexture;
      procedure SetLightmap(const aLightmap: TTexture);
      function GetEnableLightmap: boolean;
      function GetCastShadows: boolean;
      function GetReceiveShadows: boolean;
    public
      constructor Create;
      destructor Destroy; override;

      procedure AddLayer(const aLayer: TTextureLayer);
      function AddLayerForTexture(const aTexture: TTexture; const anOperation: TBlendOperation = boReplace; const anIndex: integer = -1): TTextureLayer;
      procedure AddEmptyLayer;
      procedure Assign(const aMaterial: TMaterial);
      procedure SynchronizeWithOriginal;
      function IsTransparent: boolean;
      function IsOriginal: boolean;
      function IsValid: boolean;
      function IsSystem: boolean;
      function HasRaytracingProperties: boolean;
      function NumTextureLayers: Integer;
      function NumUsedTextureLayers: Integer;
      function IndexOfLayer(aLayer: TTextureLayer): integer;
      function TextureIsUsed(aTexture: TTexture): integer;                      // change name some day
      function LayerIsUsed(index: integer): boolean;
      function ClearTextureLayers: boolean;
      function RemoveTextureLayer(aLayer: integer): boolean;
      function Equals(const material: TMaterial): boolean;
      function ContainsLightmap: boolean;
      function Copy: TMaterial;

      property Lightmap: TTexture read getLightmap write setLightmap;
      property OriginalMaterial: TMaterial read fOriginalMaterial write fOriginalMaterial;
      property EnableLightmap: boolean read GetEnableLightmap write fEnableLightmap;
      property CastShadows: Boolean read GetCastShadows write fCastShadows;
      property ReceiveShadows: Boolean read GetReceiveShadows write fReceiveShadows;
      property Name: string read fName write fName;
      property Tag: string read fTag write fTag;
      property Comments: string read fComments write fComments;
      property TextureLayers[const anIndex: Integer]: TTextureLayer read GetTextureLayerForIndex;
      property Category: string read fCategory write fCategory;
      property IsVisible: boolean read GetIsVisible write fIsVisible;
      property IsUsed: boolean read fIsUsed write fIsUsed;
      property RaytracingProperties: TMaterialRaytracingProperties read fRaytracingProperties write fRaytracingProperties;
      property ID: integer read fID write fID;  // INFO, PJ: ID currently doesn't have much status. Changes some day!
    end;

const
    //All blend modes will "use" the alpha value of the texture, except "replace"

    GL_SOURCE_BLEND_FACTORS: Array[TBlendOperation] of GLEnum = (
        GL_ZERO,       // ignore
        GL_ONE,        // replace IGNORES ALPHA
        GL_SRC_ALPHA,  // add
        GL_DST_COLOR,  // modulate
        GL_SRC_ALPHA,  // alphaBlend
        GL_SRC_COLOR   // normalMap
    );

    GL_DESTINATION_BLEND_FACTORS: Array[TBlendOperation] of GLEnum = (
        GL_ZERO,                  // ignore
        GL_ZERO,                  // replace
        GL_ONE,                   // add
        GL_ONE_MINUS_SRC_ALPHA,   // modulate
        GL_ONE_MINUS_SRC_ALPHA,   // alphaBlend
        GL_SRC_COLOR              // normalMap
    ); 

    LIGHTMAP_TEXTURE_LAYER_INDEX = 3;

implementation

uses SysUtils,
     Dialogs,
     unit_Functions;

// ### TMaterialRaytracingProperties ###########################################

const
    DEFAULT_AMBIENT_REFLECTION = 0.1;
    DEFAULT_DIFFUSE_REFLECTION = 0.9;
    DEFAULT_SPECULAR_REFLECTION = 0;
    DEFAULT_SPECULAR_EXPONENT = 150;
    DEFAULT_SPECULAR_COLOR = $FFFFFF;
    DEFAULT_REFLECTION = 0;
    DEFAULT_REFLECTIVE_EXPONENT = 0;
    DEFAULT_REFLECTIVE_COLOR = $FFFFFF;
    DEFAULT_TRANSPARENCY = 0;
    DEFAULT_REFRACTION = 1;


constructor TMaterialRaytracingProperties.Create;
begin
    inherited Create;

    fAmbientReflection := DEFAULT_AMBIENT_REFLECTION;
    fDiffuseReflection := DEFAULT_DIFFUSE_REFLECTION;
    fSpecularReflection := DEFAULT_SPECULAR_REFLECTION;
    fSpecularExponent := DEFAULT_SPECULAR_EXPONENT;
    fSpecularColor := DEFAULT_SPECULAR_COLOR;
    fReflection := DEFAULT_REFLECTION;
    fReflectiveExponent := DEFAULT_REFLECTIVE_EXPONENT;
    fReflectiveColor := DEFAULT_REFLECTIVE_COLOR;
    fTransparency := DEFAULT_TRANSPARENCY;
    fRefraction := DEFAULT_REFRACTION;
end;

procedure TMaterialRaytracingProperties.Assign(const aProperties: TMaterialRaytracingProperties);
begin
    fAmbientReflection := aProperties.AmbientReflection;
    fDiffuseReflection := aProperties.DiffuseReflection;
    fSpecularReflection := aProperties.SpecularReflection;
    fSpecularExponent := aProperties.SpecularExponent;
    fSpecularColor := aProperties.SpecularColor;
    fReflection := aProperties.Reflection;
    fReflectiveExponent := aProperties.ReflectiveExponent;
    fReflectiveColor := aProperties.ReflectiveColor;
    fTransparency := aProperties.Transparency;
    fRefraction := aProperties.Refraction;
end;

function TMaterialRaytracingProperties.Equals(const aProperties: TMaterialRaytracingProperties): Boolean;
begin
    result :=
        ( fAmbientReflection = aProperties.AmbientReflection ) and
        ( fDiffuseReflection = aProperties.DiffuseReflection ) and
        ( fSpecularReflection = aProperties.SpecularReflection ) and
        ( fSpecularExponent = aProperties.SpecularExponent ) and
        ( fSpecularColor = aProperties.SpecularColor ) and
        ( fReflection = aProperties.Reflection ) and
        ( fReflectiveExponent = aProperties.ReflectiveExponent ) and
        ( fReflectiveColor = aProperties.ReflectiveColor ) and
        ( fTransparency = aProperties.Transparency ) and
        ( fRefraction = aProperties.Refraction );
end;

function TMaterialRaytracingProperties.IsEmpty: Boolean;
begin
    result := not (
        HasAmbientReflection or
        HasDiffuseReflection or
        HasSpecularReflection or
        HasSpecularExponent or
        HasSpecularColor or
        HasReflection or
        HasReflectiveExponent or
        HasReflectiveColor or
        HasTransparency or
        HasRefraction
    );
end;

function TMaterialRaytracingProperties.HasAmbientReflection: Boolean;
begin
    Result := fAmbientReflection <> DEFAULT_AMBIENT_REFLECTION;
end;

function TMaterialRaytracingProperties.HasDiffuseReflection: Boolean;
begin
    Result := fDiffuseReflection <> DEFAULT_DIFFUSE_REFLECTION;
end;

function TMaterialRaytracingProperties.HasSpecularReflection: Boolean;
begin
    Result := fSpecularReflection <> DEFAULT_SPECULAR_REFLECTION;
end;

function TMaterialRaytracingProperties.HasSpecularExponent: Boolean;
begin
    Result := fSpecularExponent <> DEFAULT_SPECULAR_EXPONENT;
end;

function TMaterialRaytracingProperties.HasSpecularColor: Boolean;
begin
    Result := fSpecularColor <> DEFAULT_SPECULAR_COLOR;
end;

function TMaterialRaytracingProperties.HasReflection: Boolean;
begin
    Result := fReflection <> DEFAULT_REFLECTION;
end;

function TMaterialRaytracingProperties.HasReflectiveExponent: Boolean;
begin
    Result := fReflectiveExponent <> DEFAULT_REFLECTIVE_EXPONENT;
end;

function TMaterialRaytracingProperties.HasReflectiveColor: Boolean;
begin
    Result := fReflectiveColor <> DEFAULT_REFLECTIVE_COLOR;
end;

function TMaterialRaytracingProperties.HasTransparency: Boolean;
begin
    Result := fTransparency <> DEFAULT_TRANSPARENCY;
end;

function TMaterialRaytracingProperties.HasRefraction: Boolean;
begin
    Result := fRefraction <> DEFAULT_REFRACTION;
end;

// ### TTextureLayer ###########################################################

procedure TTextureLayer.SetLayerType(value: TMaterialLayerType);
begin
    fLayerType := value;
    if fLayerType=mlColor then Texture := nil;  // does NOT free texture from MaterialList in which this material is registered!
end;

function TTextureLayer.Equals(const aTextureLayer: TTextureLayer): boolean;
begin
    Result := false;
    if Assigned(aTextureLayer) then
        Result :=
            Texture.Equals(aTextureLayer.Texture) or
            ( // Only set in case of errors (texture not supported / not found)
                (InvalidTextureFileName <> '') and
                (InvalidTextureFileName = aTextureLayer.InvalidTextureFileName)
            );
end;

function TTextureLayer.IsTransparent: boolean;
begin
    Result := (blendOperation in [boModulate, boAdd, boAlphaBlend]);
end;

constructor TTextureLayer.Create(const aTexture: TTexture; const anOperation: TBlendOperation);
begin
    Create;
    fTexture := aTexture;
    if Assigned(fTexture) then
        fLayerType := mlTexture
    else
        fLayerType := mlColor;
    fOperation := anOperation;
end;

constructor TTextureLayer.Create;
begin
    inherited Create;
    fColor := Integer($FFABA6A0);
end;

destructor TTextureLayer.Destroy;
begin
    inherited Destroy;
end;

// ### TMaterial ###############################################################

function TMaterial.GetTextureLayerForIndex(const anIndex: Integer): TTextureLayer;
begin
    if LayerIsUsed(anIndex) then
        Result := fTextureLayers[anIndex] as TTextureLayer
    else
        Result := nil;
end;

procedure TMaterial.AddLayer(const aLayer: TTextureLayer);
begin
    fTextureLayers.Add(aLayer);
end;

function TMaterial.AddLayerForTexture(const aTexture: TTexture; const anOperation: TBlendOperation = boReplace; const anIndex: integer = -1): TTextureLayer;
var textureLayer: TTextureLayer;
    i: integer;
begin
    textureLayer := TTextureLayer.Create(aTexture, anOperation);
    textureLayer.LayerType := mlTexture;

    if anIndex=-1 then
        fTextureLayers.Add(textureLayer)
    else begin

        if (anIndex<NumTextureLayers) or (ContainsLightmap) then    // insert
            fTextureLayers.Items[anIndex] := textureLayer       // overwrites the current entry which should (!) be nil
        else begin
            for i := NumUsedTextureLayers to anIndex-1 do
                AddEmptyLayer;
            fTextureLayers.Add(textureLayer);
        end;
    end;
    Result := textureLayer;
end;

procedure TMaterial.AddEmptyLayer;
begin
    fTextureLayers.Add(nil);
end;

procedure TMaterial.Assign(const aMaterial: TMaterial);
var index: Integer;
    textureLayer: TTextureLayer;
    newTextureLayer: TTextureLayer;
begin
    Name := aMaterial.Name;

    fTextureLayers.Clear;

    for index := 0 to aMaterial.NumTextureLayers - 1 do begin
        textureLayer := aMaterial.TextureLayers[index];

        if Assigned(textureLayer) then begin

            newTextureLayer := TTextureLayer.Create;
            with newTextureLayer do begin
                LayerType := textureLayer.LayerType;
                Texture := textureLayer.Texture;
                Color := textureLayer.Color;
                BlendOperation := textureLayer.BlendOperation;
            end;
            fTextureLayers.Add(newTextureLayer);

        end else
            fTextureLayers.add(nil)
    end;

    Category := aMaterial.Category;
    RaytracingProperties.Assign(aMaterial.RaytracingProperties);
end;

procedure TMaterial.SynchronizeWithOriginal;
// Copies all characteristics (minus any possible lightmap) from the original.
// Only works on lightmapped (non-original) materials.
var index: Integer;
    sourceLayer, targetLayer: TTextureLayer;
begin
    if not Assigned(OriginalMaterial) then
        raise Exception.Create( 'Cannot synchronize original material.' );

    // Synch all layers from the original material
    for index := 0 to OriginalMaterial.NumTextureLayers - 1 do begin
        if index <> LIGHTMAP_TEXTURE_LAYER_INDEX then begin
            sourceLayer := OriginalMaterial.TextureLayers[index];

            if Assigned(sourceLayer) then begin
                targetLayer := TTextureLayer.Create;
                targetLayer.LayerType := sourceLayer.LayerType;
                targetLayer.Texture := sourceLayer.Texture;
                targetLayer.Color := sourceLayer.Color;
                targetLayer.BlendOperation := sourceLayer.BlendOperation;
                fTextureLayers[index] := targetLayer;
            end;
        end;
    end;

    // Delete any layers not present in the original material
    for index := 0 to NumTextureLayers - 1 do begin
        if  ( index <> LIGHTMAP_TEXTURE_LAYER_INDEX ) and
            Assigned( fTextureLayers[index] ) and
            not Assigned( OriginalMaterial.TextureLayers[index] ) then begin

            fTextureLayers[index] := nil;
        end;
    end;

    Category := OriginalMaterial.Category;
    RaytracingProperties.Assign(OriginalMaterial.RaytracingProperties);
end;

function TMaterial.IsTransparent: boolean;
begin
    Result := false;

    if NumTextureLayers > 0 then
        Result := TextureLayers[0].IsTransparent;
end;

function TMaterial.IsOriginal: boolean;
begin
    Result := not Assigned(OriginalMaterial);
end;

function TMaterial.Copy: TMaterial;
begin
    Result := TMaterial.create;
    Result.Assign(self);
end;

// A material is invalid when any one of it's layers refers to an invalid texture
function TMaterial.IsValid: boolean;
var i: integer;
    textureLayer: TTextureLayer;
begin
    Result := true;

    for i := NumTextureLayers - 1 downto 0 do begin                             // last layers are more likely to be invalid
        textureLayer := TextureLayers[i];

        if Assigned(textureLayer) and not textureLayer.Texture.IsValid then begin
            Result := false;
            exit;
        end;
    end;
end;

function TMaterial.IsSystem: boolean;
begin
    result := Uppercase(fName) = 'SYSTEM';    
end;

function TMaterial.HasRaytracingProperties: boolean;
begin
    result := not fRaytracingProperties.IsEmpty;
end;

function TMaterial.NumTextureLayers: Integer;
begin
    Result := fTextureLayers.Count;
end;

function TMaterial.numUsedTextureLayers: Integer;
var i: integer;
begin
    Result := 0;

    for i := 0 to fTextureLayers.count - 1 do begin
        if Assigned(fTexturelayers[i]) then
            Result := Result + 1;
    end;
end;

function TMaterial.IndexOfLayer(aLayer: TTextureLayer): integer;
begin
    Result := fTextureLayers.IndexOf(aLayer);       // -1 if layer does not exist in this material
end;

function TMaterial.TextureIsUsed(aTexture: TTexture): integer;
var i: integer;
    layer: TTextureLayer;
begin
    Result := 0;
    for i := 0 to NumTextureLayers-1 do begin
        layer := TextureLayers[i];
        if Assigned(layer) then begin
            if Assigned(layer.Texture) then begin
                if layer.Texture.Equals(aTexture) then begin
                    Result := Result + 1;
                end;
            end;
        end;
    end;
end;

function TMaterial.LayerIsUsed(index: integer): boolean;
begin
    if index<NumTextureLayers then
        Result := Assigned(fTexturelayers[index])
    else
        Result := false;
end;

function TMaterial.ClearTextureLayers: boolean;
begin
    fTextureLayers.Clear;
    Result := true;
end;

function TMaterial.RemoveTextureLayer(aLayer: integer): boolean;
var layer: TTextureLayer;
    i: integer;
    canRemove: boolean;
begin
    layer := TextureLayers[aLayer];

    // Only remove if there isn't any other layer above this one
    canRemove := true;
    for i := aLayer + 1 to NumTextureLayers - 1 do begin
        if Assigned(fTextureLayers[i]) then begin
            canRemove := false;
            break;
        end;
    end;

    if canRemove then begin
        fTextureLayers.Remove(layer);           // free the layer and remove it from the list

        // See if any other underlying non-assigned layers are now freed up
        for i := aLayer - 1 downto 0 do begin
            if not Assigned(fTextureLayers[i]) then
                fTextureLayers.Delete(i)
            else
                break;
        end;
    end
    else
        fTextureLayers.Items[aLayer] := nil;    // free layer and set its place in the list to nil (setting to nil frees the layer automatically)

    Result := true;
end;

function TMaterial.Equals(const material: TMaterial): boolean;
var i: integer;
begin
    Result :=
        (Name = material.Name) and
        (EnableLightmap = material.EnableLightMap) and
        (CastShadows = material.CastShadows) and
        (ReceiveShadows = material.ReceiveShadows) and
        RaytracingProperties.Equals(material.RaytracingProperties);

    Result := Result and (NumTextureLayers=material.NumTextureLayers);

    if Result then begin
        for i := 0 to NumTextureLayers-1 do begin
            if Assigned(fTextureLayers[i]) then begin
                if not TextureLayers[i].Equals(material.TextureLayers[i]) then begin
                    Result := false;
                    break;
                end;
            end
            else if Assigned(material.TextureLayers[i]) then begin
                Result := false;
                break;
            end;
        end;
    end;
end;

function TMaterial.ContainsLightmap: Boolean;
begin
    Result := (fTextureLayers.count > LIGHTMAP_TEXTURE_LAYER_INDEX ) and Assigned(fTextureLayers[LIGHTMAP_TEXTURE_LAYER_INDEX]);
end;

function TMaterial.GetLightmap: TTexture;
begin
    if containsLightmap then
        Result := TextureLayers[LIGHTMAP_TEXTURE_LAYER_INDEX].Texture
    else
        Result := nil;
end;

procedure TMaterial.SetLightmap(const aLightmap: TTexture);
var textureLayer: TTextureLayer;
    i: integer;
begin
    if containsLightmap then begin
        textureLayer := TextureLayers[LIGHTMAP_TEXTURE_LAYER_INDEX];
        textureLayer.Texture := aLightmap;
        textureLayer.LayerType := mlTexture;
    end else begin
        for i := fTextureLayers.count to LIGHTMAP_TEXTURE_LAYER_INDEX - 1 do begin
            fTextureLayers.add(nil);
        end;

        textureLayer := TTextureLayer.create(aLightmap,boModulate);

        if fTextureLayers.count <= LIGHTMAP_TEXTURE_LAYER_INDEX then begin
            fTextureLayers.add(textureLayer);
        end
        else // Unlikely case, where fTextureLayers[index] existed but wasn't assigned
            fTextureLayers[LIGHTMAP_TEXTURE_LAYER_INDEX] := textureLayer;
    end;
end;


function TMaterial.GetEnableLightmap: boolean;
begin
    if Assigned( OriginalMaterial ) then
        result := OriginalMaterial.EnableLightmap
    else
        result := fEnableLightmap;
end;

// Lightmapped materials are always in synch with original
function TMaterial.GetCastShadows: boolean;
begin
    if Assigned( OriginalMaterial ) then
        result := OriginalMaterial.CastShadows
    else
        result := fCastShadows;
end;

// Lightmapped materials are always in synch with original
function TMaterial.GetReceiveShadows: boolean;
begin
    if Assigned( OriginalMaterial ) then
        result := Originalmaterial.ReceiveShadows
    else
        result := fReceiveShadows;
end;

function TMaterial.GetIsVisible: Boolean;
begin
    // materials with a lightmap should never show up in the GUI
    Result := fIsVisible and (not containsLightmap);
end;

constructor TMaterial.Create;
begin
    inherited Create;
    fTextureLayers := TObjectList.Create(true);
    Category := 'default';
    IsVisible := true;
    IsUsed := false;
    fCastShadows := true;
    fReceiveShadows := true;
    ID := -1;
    fEnableLightmap := true;
    fRaytracingProperties := TMaterialRaytracingProperties.Create;
end;

destructor TMaterial.Destroy;
begin
    FreeAndNil(fRaytracingProperties);
    FreeAndNil(fTextureLayers);
    inherited Destroy;
end;

end.
