{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Math;

{$MODE Delphi}

interface

uses types, unit_Types, Dialogs, Math, dglOpenGL, unit_Enums, unit_Generics;

const
  rad = pi/180;
  infinite = 9999999;
  SINGLE_EPSILON = 1.19209290E-7;

type
  TMatrix = class(TObject)
  private
    matrix: array of array of double;
    function GetValue(ARow,AColumn: integer): double;
    procedure SetValue(ARow,AColumn: integer; value: double);
  public
    numRows: integer;
    numColumns: integer;
    constructor Create(NRows,NColumns: integer);
    destructor Destroy; override;
    function Add(a,b: TMatrix): TMatrix;
    function Sub(a,b: TMatrix): TMatrix;
    function Mul(a,b: TMatrix): TMatrix;
    property Cell[ARow,AColumn: integer]: double read GetValue write SetValue; default;
  end;

  TIntegerMatrix = class(TObject)
  private
    matrix: array of array of integer;
    function GetValue(ARow,AColumn: integer): integer;
    procedure SetValue(ARow,AColumn, Value: integer);
  public
    numRows: integer;
    numColumns: integer;
    constructor Create(NRows,NColumns: integer);
    destructor Destroy; override;
    property Cell[ARow,AColumn: integer]: integer read GetValue write SetValue;
  end;

  // matrix functions (expecting 4x4 matrices)
function MatrixEulerSetup(xa,ya,za,xm,ym,zm: double): TMatrix;
function MatrixArbAxisSetup(axis, position: TVertice; angle: Double): TMatrix;
{
function matrixmultiply(a,b: TMatrix): TMatrix;
function matrixadd(a,b: TMatrix): TMatrix;
function matrixsub(a,b: TMatrix): TMatrix;
}

// vector functions
//function vectormultmatrix(m: TMatrix; v: TVertice): TVertice;
function vectorRotate(v: TVertice; xa,ya,za: integer): TVertice;
function vectorLength(v: TVertice): double;
function vectorNormalize(v: TVertice): TVertice;
function vectorSub(a,b: TVertice): TVertice;
function vectorAdd(a,b: TVertice): TVertice;
function vectorMultiply(a,b: TVertice): TVertice;
function vectorAngle(a,b: TVertice): double;
function vectorMultiplyScalar(a: TVertice; b: double): TVertice;

// vertex functions (light-weight TVertice alternative)
function vertex(const anX, aY, aZ: single): TVertex; overload;
function vertex(const anX, aY, aZ, aW: single): TVertex4; overload;
function VertexToVertex4(const aVertex: TVertex; const aW: Single = 1): TVertex4;
function vertexEquals( const aV1, aV2: TVertex ): boolean;
function vertexAdd(const aVertex1, aVertex2: TVertex): TVertex; overload;
function vertexAdd(const aVertex1, aVertex2, aVertex3: TVertex): TVertex; overload;
function vertexSubtract(const aVertex1, aVertex2: TVertex): TVertex;
function vertexAddScalar(const aVertex: TVertex; const aScalar: single): TVertex;
function VertexMultiply( const aVector: TVertex; const aScalar: single ): TVertex;
function vertexDivide( const aV1: TVertex; const aScalar: single ): TVertex;
function vertexLength(const aVertex: TVertex): single;
function vertexSquaredLength( const aV1: TVertex ): single;
function vertexDistance( const aV1, aV2: TVertex ): single;
function vertexNormalize(const aVertex: TVertex): TVertex;
function vertexCrossProduct(const aV1, aV2: TVertex): TVertex;
function vertexDotProduct(const aV1, aV2: TVertex): Single; overload;
function vertexDotProduct(const aV1, aV2: TVertex4): Single; overload;
function vertexMaxElements( const aV1, aV2: TVertex ): TVertex;
function vertexMinElements( const aV1, aV2: TVertex ): TVertex;
function vertexFaceNormal( const aV1, aV2, aV3: TVertex ): TVertex;
function vertexScaleNormalVertex( const aVertex: TVertex; const aScaleValue: single = 1 ): TVertex;
function vertexToVertice(const aVertex: TVertex): TVertice;
function VerticeToVertex(const aVertice: TVertice): TVertex; overload;
function VerticeToVertex(const aVertice: TGenericVertex): TVertex; overload;
function Vertex4ToVertex(const aVertex4: TVertex4): TVertex;
function Vertex4FromOpenGL(const anOpenGLVertex: TGLVectori4 ): TVertex4;

function VerticeToVertex4(const aVertice: TVertice; const aW: Single = 1): TVertex4;
function VerticesWeightedCovariance( const aVertices: TVertexArray; const aWeights: TSingleDynArray ): TSymMatrix3x3;

// Matrix functions (lightweight TMatrix alternative)
function Matrix4x4EulerSetup(xa,ya,za,xm,ym,zm: single): TMatrix4x4;
function IdentityMatrix: TMatrix4x4;
function Matrix4x4Multiply(const aMatrix: TMatrix4x4; const aVertex: TVertex4): TVertex4; overload;
function Matrix4x4Multiply(const aMatrix1, aMatrix2: TMatrix4x4 ): TMatrix4x4; overload;
function Matrix4x4Transpose(const aMatrix: TMatrix4x4): TMatrix4x4;
function Matrix4x4Invert(const aMatrix: TMatrix4x4): TMatrix4x4;
function Matrix4x4FromOpenGL( const anOpenGLMatrix: TGLMatrixd4 ): TMatrix4x4; overload;
function Matrix4x4FromOpenGL( const anOpenGLMatrix: TGLMatrixf4 ): TMatrix4x4; overload;

function SymMatrix3x3( const anM11, anM12, anM13, anM21, anM22, anM31: Single ): TSymMatrix3x3;
function SymMatrix3x3PrincipalComponents( const aMatrix: TSymMatrix3x3 ): TVertex;


// UV Coordinate functions
function UVCoordToUVCoordinate(const aUVCoord: TUVCoord): TUVCoordinate;
function UVCoordinate(const aU, aV: Single): TUVCoordinate;
function ApplyEditDimension(const aUVCoordinate: TUVCoordinate; const aDimension: TEditUVDimension; const aRestrictedValue: Integer = 0): TUVCoordinate;

// plane functions
function PolygonToPlane(const aPolygon: TPolygon): TVertex4;
function plane_normal(v1,v2,v3: TVertice): TVertice;
function planeD(v1,normal: TVertice): double;
function DistToPlane(v: TVertice; plane: TPlane): double; overload;
function DistToPlane(v: TGenericVertex; plane: TPlane): double; overload;

// misc functions
function dotproduct(a,b: TVertice): double;
function dotproduct2d(a,b: TPoint2D): double;
function crossproduct(a,b: TVertice): TVertice;
function signed(number: double): boolean;
function IntersectsRayTriangle(const aRay: TRay; const aVertex0, aVertex1, aVertex2: TVertex; var aT, aU, aV: Single): boolean;
function IsPointInTriangle(const aPoint: TVertex; const aVertex0, aVertex1, aVertex2: TVertex; var aU, aV: Single): boolean;
function Intersect(const aRay: TRay; const aPlane: TVertex4): TVertex; // overload for more intersection tests.
function PointInFrustum(aPoint: TVertex; aFrustumPlanes,aFrustumPoints: array of TVertex): boolean;
function Polygon2DFacesViewer(const aPoints: TPoint2DArray): boolean;

implementation

uses
    SysUtils;

// matrix functions

function MatrixEulerSetup(xa,ya,za,xm,ym,zm: double): TMatrix;                  // generates a rotation matrix
var a,b,c,d,e,f,ad,bd: double;                                                  // based on euler coordinates
begin
    a := cos(xa*rad);
    b := sin(xa*rad);
    c := cos(ya*rad);
    d := sin(ya*rad);
    e := cos(za*rad);
    f := sin(za*rad);

    ad := a * d;
    bd := b * d;

    Result := TMatrix.Create(4,4);
    Result.matrix[0,0] :=  C * E;
    Result.matrix[0,1] := -C * F;
    Result.matrix[0,2] :=  D;
    Result.matrix[0,3] :=  0;
    Result.matrix[1,0] :=  BD * E + A * F;
    Result.matrix[1,1] := -BD * F + A * E;
    Result.matrix[1,2] := -B * C;
    Result.matrix[1,3] :=  0;
    Result.matrix[2,0] := -AD * E + B * F;
    Result.matrix[2,1] :=  AD * F + B * E;
    Result.matrix[2,2] :=  A * C;
    Result.matrix[2,3] :=  0;
    Result.matrix[3,0] :=  xm;    // movement vector here?
    Result.matrix[3,1] :=  ym;
    Result.matrix[3,2] :=  zm;
    Result.matrix[3,3] :=  1;
end;

function MatrixArbAxisSetup(axis, position: TVertice; angle: Double): TMatrix;
var s,c,t: extended;
begin
    s := sin(angle*(pi/180));
    c := cos(angle*(pi/180));
    t := 1 - c;

{    result[0, 0] := t*axis.x*axis.x+c;
    result[1, 0] := t*axis.x*axis.y+s*axis.z;
    result[2, 0] := t*axis.x*axis.z-s*axis.y;
    result[0, 1] := t*axis.x*axis.y-s*axis.z;
    result[1, 1] := t*axis.y*axis.y+c;
    result[2, 1] := t*axis.y*axis.z+s*axis.x;
    result[0, 2] := t*axis.x*axis.z+s*axis.y;
    result[1, 2] := t*axis.y*axis.z-s*axis.x ;
    result[2, 2] := t*axis.z*axis.z+c;  }

    result := TMatrix.Create(4,4);
    result.matrix[0, 0] := t*axis.x*axis.x+c;
    result.matrix[0, 1] := t*axis.x*axis.y-s*axis.z;
    result.matrix[0, 2] := t*axis.x*axis.z+s*axis.y;
    result.matrix[0, 3] := 0;
    result.matrix[1, 0] := t*axis.x*axis.y+s*axis.z;
    result.matrix[1, 1] := t*axis.y*axis.y+c;
    result.matrix[1, 2] := t*axis.y*axis.z-s*axis.x;
    result.matrix[1, 3] := 0;
    result.matrix[2, 0] := t*axis.x*axis.z-s*axis.y;
    result.matrix[2, 1] := t*axis.y*axis.z+s*axis.x;
    result.matrix[2, 2] := t*axis.z*axis.z+c;
    result.matrix[2, 3] := 0;
    result.matrix[3, 0] := Position.X;
    result.matrix[3, 1] := Position.Y;
    result.matrix[3, 2] := Position.Z;
    result.matrix[3, 3] := 1;
end;
(*
function matrixmultiply(a,b: TMatrix): TMatrix;
var r: TMatrix; i,j: integer;
begin
    for i := 0 to 3 do
        for j := 0 to 3 do
            r[i,j] := a[i,j] * b[j,i];
    Result := r;
end;

function matrixadd(a,b: TMatrix): TMatrix;
var r: TMatrix; i,j: integer;
begin
    for i := 0 to 3 do
        for j := 0 to 3 do
            r[i,j] := a[i,j] + b[i,j];
    Result := r;
end;

function matrixsub(a,b: TMatrix): TMatrix;
var r: TMatrix; i,j: integer;
begin
    for i := 0 to 3 do
        for j := 0 to 3 do
            r[i,j] := a[i,j] - b[i,j];
    Result := r;
end;   *)

// vector functions

function vectorMultiplyMatrix(m: TMatrix; v: TVertice): TVertice;
var r: TVertice;
begin
    r := TVertice.Create(v.x*m.matrix[0,0]+v.y*m.matrix[0,1]+v.z*m.matrix[0,2],
                         v.x*m.matrix[1,0]+v.y*m.matrix[1,1]+v.z*m.matrix[1,2],
                         v.x*m.matrix[2,0]+v.y*m.matrix[2,1]+v.z*m.matrix[2,2]);
    Result := r;
end;

function vectorRotate(v: TVertice; xa,ya,za: integer): TVertice;
var r: TVertice; x,y,z,value: double;
begin
    r := v.Copy;

    value := ya*rad;
    x:=r.x*cos(value) - r.z*sin(value);                                         // rotate around Y
    z:=r.x*sin(value) + r.z*cos(value);
    r.x:=x;
    r.z:=z;

    value := xa*rad;
    y:=r.y*cos(value) - r.z*sin(value);                                         // rotate around X
    z:=r.y*sin(value) + r.z*cos(value);
    r.y:=y;
    r.z:=z;

    value := za*rad;
    x:=r.x*cos(value) - r.y*sin(value);                                         // rotate around Z
    y:=r.x*sin(value) + r.y*cos(value);
    r.x:=x;
    r.y:=y;

    Result := r;
end;

function vectorLength(v: TVertice): double;
begin
    Result := sqrt(sqr(v.x) + sqr(v.y) + sqr(v.z));
end;

function vectorNormalize(v: TVertice): TVertice;
var length: double;
begin
    length := vectorlength(v);
    if length=0 then length := 0.001;
    Result := TVertice.Create(v.x/length,v.y/length,v.z/length);
end;

function vectorSub(a,b: TVertice): TVertice;
begin
    Result := TVertice.Create(a.x-b.x,a.y-b.y,a.z-b.z);
end;

function vectorAdd(a,b: TVertice): TVertice;
begin
    Result := TVertice.Create(a.x+b.x,a.y+b.y,a.z+b.z);
end;

function vectorMultiply(a,b: TVertice): TVertice;
begin
    Result := TVertice.Create(a.x*b.x,a.y*b.y,a.z*b.z);
end;

//The dot product of two vectors is the product of the cosine of the angle
//between them and the magnitude (length) of the two vectors: Dot(A,B) = cos(ang) * len(A) * len(B)
//Thus if A and B were both normalized vectors (length = 1.0), the result
//would be simply cos(ang). So: angle := arccos(dotproduct(e1,e2)) * (180/pi);
function vectorAngle(a,b: TVertice): double;
var x: double;
begin
    a.Normalize;
    b.Normalize;
    x := dotproduct(a,b);
    Result := arccos(x)*(180/pi);
end;

function vectorMultiplyScalar(a: TVertice; b: double): TVertice;
begin
    Result := TVertice.Create(a.x*b,a.y*b,a.z*b);
end;

// ### Vertex functions ########################################################

function vertex(const anX, aY, aZ: single): TVertex;
begin
  result.X := anX;
  result.Y := aY;
  result.Z := aZ;
end;

function vertex(const anX, aY, aZ, aW: single): TVertex4;
begin
    result[0] := anX;
    result[1] := aY;
    result[2] := aZ;
    result[3] := aW;
end;

function VertexToVertex4(const aVertex: TVertex; const aW: Single = 1): TVertex4;
begin
    result[0] := aVertex.X;
    result[1] := aVertex.Y;
    result[2] := aVertex.Z;
    result[3] := aW;
end;

function vertexEquals(const aV1, aV2: TVertex): boolean;
begin
  result :=
      ( aV1.X = aV2.X ) and
      ( aV1.Y = aV2.Y ) and
      ( aV1.Z = aV2.Z );
end;

function vertexAdd(const aVertex1, aVertex2: TVertex): TVertex;
begin
    result.X := aVertex1.X + aVertex2.X;
    result.Y := aVertex1.Y + aVertex2.Y;
    result.Z := aVertex1.Z + aVertex2.Z;
end;

function vertexAdd(const aVertex1, aVertex2, aVertex3: TVertex): TVertex; overload;
begin
    result.X := aVertex1.X + aVertex2.X + aVertex3.X;
    result.Y := aVertex1.Y + aVertex2.Y + aVertex3.Y;
    result.Z := aVertex1.Z + aVertex2.Z + aVertex3.Z;
end;

function vertexSubtract(const aVertex1, aVertex2: TVertex): TVertex;
begin
    result.X := aVertex1.X - aVertex2.X;
    result.Y := aVertex1.Y - aVertex2.Y;
    result.Z := aVertex1.Z - aVertex2.Z;
end;

function vertexAddScalar(const aVertex: TVertex; const aScalar: single): TVertex;
begin
    result.X := aVertex.X + aScalar;
    result.Y := aVertex.Y + aScalar;
    result.Z := aVertex.Z + aScalar;
end;

function VertexMultiply( const aVector: TVertex; const aScalar: single ): TVertex;
begin
    result.X := aVector.X * aScalar;
    result.Y := aVector.Y * aScalar;
    result.Z := aVector.Z * aScalar;
end;

function vertexDivide( const aV1: TVertex; const aScalar: single ): TVertex;
begin
    result.X := aV1.X / aScalar;
    result.Y := aV1.Y / aScalar;
    result.Z := aV1.Z / aScalar;
end;

function vertexLength(const aVertex: TVertex): single;
begin
    result := Sqrt( aVertex.X * aVertex.X + aVertex.Y * aVertex.Y + aVertex.Z * aVertex.Z );
end;

function vertexSquaredLength( const aV1: TVertex ): single;
begin
  result := aV1.X * aV1.X + aV1.Y * aV1.Y + aV1.Z * aV1.Z;
end;

function vertexDistance( const aV1, aV2: TVertex ): single;
begin
  result := vertexLength( vertexSubtract( aV1, aV2 ));
end;

function vertexNormalize(const aVertex: TVertex): TVertex;
var length, invLength: single;
begin
    length := vertexLength(aVertex);

    if length > 0 then
        invLength := 1 / length
    else
        invLength := 999999;

    result.X := aVertex.X * invLength;
    result.Y := aVertex.Y * invLength;
    result.Z := aVertex.Z * invLength;
end;

function vertexCrossProduct(const aV1, aV2: TVertex): TVertex;
begin
    Result.X := aV1.Y * aV2.Z - aV1.Z * aV2.Y;
    Result.Y := aV1.Z * aV2.X - aV1.X * aV2.Z;
    Result.Z := aV1.X * aV2.Y - aV1.Y * aV2.X;
end;

function vertexDotProduct(const aV1, aV2: TVertex): Single;
begin
    result := aV1.X * aV2.X + aV1.Y * aV2.Y + aV1.Z * aV2.Z;
end;

function vertexDotProduct(const aV1, aV2: TVertex4): Single;
begin
    result := aV1[0]*aV2[0] + aV1[1]*aV2[1] + aV1[2]*aV2[2] + aV1[3]*aV2[3]; 
end;

function vertexMaxElements( const aV1, aV2: TVertex ): TVertex;
begin
  if aV1.X > aV2.X then Result.X := aV1.X else Result.X := aV2.X;
  if aV1.Y > aV2.Y then Result.Y := aV1.Y else Result.Y := aV2.Y;
  if aV1.Z > aV2.Z then Result.Z := aV1.Z else Result.Z := aV2.Z;
end;

function vertexMinElements( const aV1, aV2: TVertex ): TVertex;
begin
  if aV1.X < aV2.X then Result.X := aV1.X else Result.X := aV2.X;
  if aV1.Y < aV2.Y then Result.Y := aV1.Y else Result.Y := aV2.Y;
  if aV1.Z < aV2.Z then Result.Z := aV1.Z else Result.Z := aV2.Z;
end;

function vertexFaceNormal( const aV1, aV2, aV3: TVertex ): TVertex;
var edge1, edge2: TVertex;
begin
  edge1 := vertexSubtract( aV2, aV1 );
  edge2 := vertexSubtract( aV2, aV3 );
  result := vertexCrossProduct( edge1, edge2 );
  vertexNormalize( result );
end;

function vertexScaleNormalVertex( const aVertex: TVertex; const aScaleValue: single = 1 ): TVertex;
// INFO, PJ: Scales the vertex such that the minimum non-zero element becomes the given value
var
  index: integer;
  value: single;
  minElement: single;
  multiplier: single;

begin
  minElement := 999999;  // INFO, PJ: Safe, because the vertex should be normal

  for index := 0 to 2 do begin
      value := abs( aVertex.Elements[index] );

      if ( value < minElement ) and ( value > 0 ) then
          minElement := value;
  end;

  multiplier := aScaleValue / minElement;

  result.X := aVertex.X * multiplier;
  result.Y := aVertex.Y * multiplier;
  result.Z := aVertex.Z * multiplier;
end;


function VertexToVertice(const aVertex: TVertex): TVertice;
begin
    result := TVertice.Create(aVertex.X, aVertex.Y, aVertex.Z);
end;

function VerticeToVertex(const aVertice: TVertice): TVertex;
begin
    result.X := aVertice.X;
    result.Y := aVertice.y;
    result.Z := aVertice.Z;
end;

function VerticeToVertex(const aVertice: TGenericVertex): TVertex;
begin
    result.X := aVertice.X;
    result.Y := aVertice.y;
    result.Z := aVertice.Z;
end;

function Vertex4ToVertex(const aVertex4: TVertex4): TVertex;
begin
    result.X := aVertex4[0];
    result.Y := aVertex4[1];
    result.Z := aVertex4[2];
end;

function VerticeToVertex4(const aVertice: TVertice; const aW: Single = 1): TVertex4;
begin
    result[0] := aVertice.X;
    result[1] := aVertice.Y;
    result[2] := aVertice.Z;
    result[3] := aW;
end;

function Vertex4FromOpenGL(const anOpenGLVertex: TGLVectori4 ): TVertex4;
begin
    result[0] := anOpenGLVertex[0];
    result[1] := anOpenGLVertex[1];
    result[2] := anOpenGLVertex[2];
    result[3] := anOpenGLVertex[3];
end;

// ### Matrix functions ########################################################

// Used to reproduce the OpenGL Modelview matrix.
function Matrix4x4EulerSetup(xa,ya,za,xm,ym,zm: single): TMatrix4x4;
var a,b,c,d,e,f,ad,bd: double;
begin
    a := cos(xa*rad);
    b := sin(xa*rad);
    c := cos(ya*rad);
    d := sin(ya*rad);
    e := cos(za*rad);
    f := sin(za*rad);

    ad := a * d;
    bd := b * d;

    result[0,0] :=  C * E;
    result[0,1] := -C * F;
    result[0,2] :=  D;
    result[0,3] :=  ( result[0, 0] * xm + result[0, 1] * ym + result[0, 2] * zm );
    result[1,0] :=  BD * E + A * F;
    result[1,1] := -BD * F + A * E;
    result[1,2] := -B * C;
    result[1,3] :=  ( result[1, 0] * xm + result[1, 1] * ym + result[1, 2] * zm );
    result[2,0] := -AD * E + B * F;
    result[2,1] :=  AD * F + B * E;
    result[2,2] :=  A * C;
    result[2,3] :=  ( result[2, 0] * xm + result[2, 1] * ym + result[2, 2] * zm );
    result[3,0] :=  0;
    result[3,1] :=  0;
    result[3,2] :=  0;
    result[3,3] :=  1;
end;

function Matrix4x4Multiply(const aMatrix: TMatrix4x4; const aVertex: TVertex4): TVertex4;
begin
   result[0] := aMatrix[0, 0] * aVertex[0] + aMatrix[0, 1] * aVertex[1] + aMatrix[0, 2] * aVertex[2] + aMatrix[0, 3] * aVertex[3];
   result[1] := aMatrix[1, 0] * aVertex[0] + aMatrix[1, 1] * aVertex[1] + aMatrix[1, 2] * aVertex[2] + aMatrix[1, 3] * aVertex[3];
   result[2] := aMatrix[2, 0] * aVertex[0] + aMatrix[2, 1] * aVertex[1] + aMatrix[2, 2] * aVertex[2] + aMatrix[2, 3] * aVertex[3];
   result[3] := aMatrix[3, 0] * aVertex[0] + aMatrix[3, 1] * aVertex[1] + aMatrix[3, 2] * aVertex[2] + aMatrix[3, 3] * aVertex[3];
end;

function Matrix4x4Multiply(const aMatrix1, aMatrix2: TMatrix4x4 ): TMatrix4x4;
var x, y, i: integer;
    value: single;
begin
    for y := 0 to 3 do begin
        for x := 0 to 3 do begin
            value := 0;
            for i := 0 to 3 do begin
                value := value + aMatrix1[y, i] * aMatrix2[i, x];
            end;

            result[y, x] := value;
        end;
    end;
end;

function Matrix4x4Transpose(const aMatrix: TMatrix4x4): TMatrix4x4;
var i, j: integer;
begin
    for i := 0 to 3 do
        for j := 0 to 3 do
            result[i, j] := aMatrix[j, i];
end;

function IdentityMatrix: TMatrix4x4;
var i, j: integer;
begin
    for i := 0 to 3 do begin
        for j := 0 to 3 do begin
            if i = j then
                result[i, j] := 1
            else
                result[i, j] := 0;
        end;
    end;
end;

function Matrix4x4FromOpenGL( const anOpenGLMatrix: TGLMatrixd4 ): TMatrix4x4;
var
  x, y: integer;

begin
  for y := 0 to 3 do begin
      for x := 0 to 3 do begin
          result[y, x] := anOpenGLMatrix[x, y];
      end;
  end;
end;

function Matrix4x4FromOpenGL( const anOpenGLMatrix: TGLMatrixf4 ): TMatrix4x4;
var
  x, y: integer;

begin
  for y := 0 to 3 do begin
      for x := 0 to 3 do begin
          result[y, x] := anOpenGLMatrix[x, y];
      end;
  end;
end;

function VerticesWeightedCovariance( const aVertices: TVertexArray; const aWeights: TSingleDynArray ): TSymMatrix3x3;
var i: Integer;
    totalWeight: Single;
    centroid, a, b: TVertex;
begin
    if Length( aVertices ) <> Length( aWeights ) then
        raise Exception.Create( 'Number of vertices does not match number of weights.' );

    // compute the centroid
    totalWeight := 0;
    centroid := Vertex( 0, 0, 0 );
    for i := Low( aVertices ) to High( aVertices ) do begin
        totalWeight := totalWeight + aWeights[i];
        centroid := VertexAdd( centroid, VertexMultiply( aVertices[i], aWeights[i] ));
    end;
    centroid := VertexDivide( centroid, totalWeight );

    // Accumulate the covariance matrix
    result := SymMatrix3x3( 0, 0, 0, 0, 0, 0 );
    for i := Low( aVertices ) to High( aVertices ) do begin
        a := VertexSubtract( aVertices[i], centroid );
        b := VertexMultiply( a, aWeights[i] );

        result[0] := result[0] + a.X * b.Y;
		result[1] := result[1] + a.X * b.Y;
		result[2] := result[2] + a.X * b.Z;
		result[3] := result[3] + a.Y * b.Y;
		result[4] := result[4] + a.Y * b.Z;
		result[5] := result[5] + a.Z * b.Z;
    end;
end;

function SymMatrix3x3Multiplicity1Eigenvector( const aMatrix: TSymMatrix3x3; const anEigenValue: Single ): TVertex;
var m, u: TSymMatrix3x3;
    i, maxIndex: Integer;
    value, maxValue: Single;
    
begin
    // construct the matrix
	m[0] := aMatrix[0] - anEigenValue;
	m[1] := aMatrix[1];
	m[2] := aMatrix[2];
	m[3] := aMatrix[3] - anEigenValue;
	m[4] := aMatrix[4];
	m[5] := aMatrix[5] - anEigenValue;

    // compute U
	u[0] := m[3] * m[5] - m[4] * m[4];
	u[1] := m[2] * m[4] - m[1] * m[5];
	u[2] := m[1] * m[4] - m[2] * m[3];
	u[3] := m[0] * m[5] - m[2] * m[2];
	u[4] := m[1] * m[2] - m[4] * m[0];
	u[5] := m[0] * m[3] - m[1] * m[1];

    // Find the index of the largest component
    maxIndex := 0;
    maxValue := Abs( m[0] );

    for i := Low( m ) to High( m ) do begin
        value := Abs( m[i] );
        if value > maxValue then begin
            maxValue := value;
            maxIndex := i;
        end;
    end;

    // Based on this index, return a Column
    case maxIndex of
        0:      result := Vertex( u[0], u[1], u[2] );
        1, 3:   result := Vertex( u[1], u[3], u[4] );
        else    result := Vertex( u[2], u[4], u[5] );
    end;
end;

function SymMatrix3x3Multiplicity2Eigenvector( const aMatrix: TSymMatrix3x3; const anEigenValue: Single ): TVertex;
var m: TSymMatrix3x3;
    i, maxIndex: Integer;
    value, maxValue: Single;
    
begin
    // Construct the matrix
	m[0] := aMatrix[0] - anEigenValue;
	m[1] := aMatrix[1];
	m[2] := aMatrix[2];
	m[3] := aMatrix[3] - anEigenValue;
	m[4] := aMatrix[4];
	m[5] := aMatrix[5] - anEigenValue;

    // Find the index of the largest component
    maxIndex := 0;
    maxValue := Abs( m[0] );

    for i := Low( m ) to High( m ) do begin
        value := Abs( m[i] );
        if value > maxValue then begin
            maxValue := value;
            maxIndex := i;
        end;
    end;

    // Based on this index, return an eigenVector
    case maxIndex of
        0, 1:   result := Vertex( -m[1], m[0], 0 );
        2:      result := Vertex( m[2], 0, -m[0] );
        3, 4:   result := Vertex( 0, -m[4], m[3] );
        else    result := Vertex( 0, -m[5], m[4] );
    end;
end;

function SymMatrix3x3( const anM11, anM12, anM13, anM21, anM22, anM31: Single ): TSymMatrix3x3;
begin
    result[0] := anM11;
    result[1] := anM12;
    result[2] := anM13;
    result[3] := anM21;
    result[4] := anM22;
    result[5] := anM31;
end;

function SymMatrix3x3PrincipalComponents( const aMatrix: TSymMatrix3x3 ): TVertex;
var c0, c1, c2: Single;
    a, b, q: Single;
    theta, rho: Single;
    rt, ct, st: Single;
    l1, l2, l3: Single;

begin
	// compute the cubic coefficients
    c0 := aMatrix[0] * aMatrix[3] * aMatrix[5]
          + 2 * aMatrix[1] * aMatrix[2] * aMatrix[4]
          - aMatrix[0] * aMatrix[4] * aMatrix[4]
          - aMatrix[3] * aMatrix[2] * aMatrix[2]
          - aMatrix[5] * aMatrix[1] * aMatrix[1];
    c1 := aMatrix[0] * aMatrix[3] + aMatrix[0] * aMatrix[5] + aMatrix[3] * aMatrix[5]
          - aMatrix[1] * aMatrix[1] - aMatrix[2] * aMatrix[2] - aMatrix[4] * aMatrix[4];
	c2 := aMatrix[0] + aMatrix[3] + aMatrix[5];

	// compute the quadratic coefficients
	a := c1 - ( 1/3 ) * c2 * c2;
	b := ( -2 / 27 ) * c2 * c2 * c2 + ( 1/ 3 ) * c1 * c2 - c0;

	// compute the root count check
	q := 0.25 * b * b + ( 1 / 27 ) * a * a * a;

	// test the multiplicity
	if( q > SINGLE_EPSILON  ) then begin
        result := Vertex( 1, 1, 1 );
    end
    else if q < -SINGLE_EPSILON then begin
		// three distinct roots
		theta := ArcTan2( sqrt( -Q ), -0.5 * b );
		rho := Sqrt( 0.25 * b * b - Q );

		rt := Power( rho, 1/3 );
		ct := Cos( theta / 3 );
		st := Sin( theta / 3 );

		l1 := ( 1 / 3 ) * c2 + 2.0 * rt * ct;
		l2 := ( 1 / 3 ) * c2 - rt * ( ct + Sqrt( 3 ) * st );
		l3 := ( 1 / 3 ) * c2 - rt * ( ct - Sqrt( 3 ) * st );

		// pick the larger
		if abs( l2 ) > abs( l1 ) then
			l1 := l2;
		if abs( l3 ) > abs( l1 ) then
			l1 := l3;

		// get the eigenvector
		result := SymMatrix3x3Multiplicity1EigenVector( aMatrix, l1 );
	end
	else begin // -SINGLE_EPSILON <= Q <= SINGLE_EPSILON
		// two roots
		if b < 0 then
			rt := -Power( -0.5 * b, 1/3 )
		else
			rt := Power( 0.5 * b, 1/3 );

		l1 := ( 1/3 ) * c2 + rt;        // repeated
		l2 := ( 1/3 ) * c2 - 2.0 * rt;

		// get the eigenvector
		if abs( l1 ) > abs( l2 ) then
			result := SymMatrix3x3Multiplicity2EigenVector( aMatrix, l1 )
		else
			result := SymMatrix3x3Multiplicity1EigenVector( aMatrix, l2 )
    end;
end;

function Matrix4x4Invert(const aMatrix: TMatrix4x4): TMatrix4x4;
const
  EPSILON = 0.0001;

  function SignificantRowForColumn( const aMatrix: TMatrix4x4; const aColumn: integer ): integer;
  var
    i: integer;
  begin
      result := -1;

      // REFA, PJ: This should probably be a < epsilon check
      for i := aColumn to 3 do begin
          if Abs( aMatrix[i, aColumn] ) > EPSILON then begin
              result := i;
              exit;
          end;
      end;

//      if result = -1 then begin
//          raise Exception.Create( 'Cannot invert singular matrix.' );
//      end;
  end;

  procedure ExchangeRows( var aMatrix, anInverse: TMatrix4x4; const aSourceRow, aTargetRow: integer );
  var tempValue: single;
      i: integer;
  begin
      for i := 0 to 3 do begin
          tempValue := aMatrix[aSourceRow, i];
          aMatrix[aSourceRow, i] := aMatrix[aTargetRow, i];
          aMatrix[aTargetRow, i] := tempValue;

          tempValue := anInverse[aSourceRow, i];
          anInverse[aSourceRow, i] := anInverse[aTargetRow, i];
          anInverse[aTargetRow, i] := tempValue;
      end;
  end;

  procedure SweepRow( var aMatrix, anInverse: TMatrix4x4; const aSourceRow, aTargetRow: integer );
  var factor: single;
      i: integer;
  begin
      factor := aMatrix[aSourceRow, aSourceRow];

      if factor = 0 then exit;

      if factor <> 1 then begin
          factor := 1 / factor;
          for i := 0 to 3 do begin
              aMatrix[aSourceRow, i] := aMatrix[aSourceRow, i] * factor;
              anInverse[aSourceRow, i] := anInverse[aSourceRow, i] * factor;
          end;
      end;

      factor := aMatrix[aTargetRow, aSourceRow];
      if factor = 0 then exit;

      for i := 0 to 3 do begin
          aMatrix[aTargetRow, i] := aMatrix[aTargetRow, i] - factor * aMatrix[aSourceRow, i];
          anInverse[aTargetRow, i] := anInverse[aTargetRow, i] - factor * anInverse[aSourceRow, i];
      end;
  end;

var
  i, j: integer;
  matrix: TMatrix4x4;

begin
  matrix := aMatrix;
  result := IdentityMatrix;

  for j := 0 to 3 do begin
      if Abs( matrix[j, j] ) < EPSILON then begin
          i := SignificantRowForColumn( matrix, j );
          if i <> -1 then
              ExchangeRows( matrix, result, j, i );
      end;

      for i := 0 to 3 do begin
          if i <> j then
              SweepRow( matrix, result, j, i );
      end;
  end;
end;


// ### UV Coord functions ######################################################

function UVCoordToUVCoordinate(const aUVCoord: TUVCoord): TUVCoordinate;
begin
    result.U := aUVCoord.U;
    result.V := aUVCoord.V;
end;

function UVCoordinate(const aU, aV: Single): TUVCoordinate;
begin
    result.U := aU;
    result.V := aV;
end;

function ApplyEditDimension(const aUVCoordinate: TUVCoordinate; const aDimension: TEditUVDimension; const aRestrictedValue: Integer = 0): TUVCoordinate;
begin
    result := aUVCoordinate;
    case aDimension of
        edU:    result.V := aRestrictedValue;
        edV:    result.U := aRestrictedValue;
    end;
end;

// ### Plane functions #########################################################

function PolygonToPlane( const aPolygon: TPolygon ): TVertex4;
var
  normal, p: TVertex;
  d: single;

begin
  normal := VerticeToVertex( aPolygon.Normal );
  p := VerticeToVertex( aPolygon.VerticeList[0] );
  d := - VertexDotProduct( normal, p );

  result := Vertex( normal.X, normal.Y, normal.Z, d );
end;

function plane_normal(v1,v2,v3:TVertice): TVertice;
var a,b,n: TVertice;
begin
    a := vectorsub(v2,v1);                                                      // create vectors a and b to construct a plane
    b := vectorsub(v2,v3);
    n := crossproduct(a,b);                                                     // normal of polygon-plane
    Result := vectornormalize(n);                                               // unit normal of polygon-plane
    a.Free;
    b.Free;
    n.Free;
end;

function planeD(v1,normal: TVertice): double;
begin
    Result := -dotproduct(v1,normal);                                           // plane equation value
end;

function DistToPlane(v: TVertice; plane: TPlane): double;
begin
    Result := plane.Normal.x*v.x + plane.Normal.y*v.y + plane.Normal.z*v.z + plane.Distance;
end;

function DistToPlane(v: TGenericVertex; plane: TPlane): double;
begin
    Result := plane.Normal.x*v.x + plane.Normal.y*v.y + plane.Normal.z*v.z + plane.Distance;
end;


// misc functions

function DotProduct(a,b: TVertice): double;
begin
    Result := a.x*b.x + a.y*b.y + a.z*b.z;
end;

function DotProduct2d(a,b: TPoint2D): double;                                   
begin
    Result := a.X*b.X + a.Y*b.Y;
end;

function CrossProduct(a,b: TVertice): TVertice;
begin
    Result := TVertice.Create(a.y*b.z - a.z*b.y,a.z*b.x - a.x*b.z,a.x*b.y - a.y*b.x);
end;

function signed(number: double): boolean;
begin
    if number<0 then
        Result := true
    else
        Result := false;
end;

// MollerTrumbore97 algorithm for Fast, Minimum Storage Ray-Triangle Intersection.
function IntersectsRayTriangle(const aRay: TRay; const aVertex0, aVertex1, aVertex2: TVertex; var aT, aU, aV: Single): boolean;
var edge1, edge2: TVertex;
    tVec, pVec, qVec: TVertex;
    det, invDet: Single;
begin
    // find vectors for two edges sharing vert0
    edge1 := vertexSubtract(aVertex1, aVertex0);
    edge2 := vertexSubtract(aVertex2, aVertex0);

    // begin calculating determinant - also used to calculate U parameter
    pVec := vertexCrossProduct(aRay.Direction, edge2);

    // if determinant is near zero, ray lies in plane of triangle
    det := vertexDotProduct(edge1, pVec);

    if det < 0 then begin
      result := false;
      exit;
    end;

    //calculate distance from vert0 to ray origin
    tVec := vertexSubtract(aRay.Origin, aVertex0);

    // calculate U parameter and test bounds
    aU := vertexDotProduct(tVec, pVec);

    if ( aU < 0 ) or ( aU > det ) then begin
      result := false;
      exit;
    end;

    // prepare to test V parameter
    qVec := vertexCrossProduct(tVec, edge1);

    // calculate V parameter and test bounds
    aV := vertexDotProduct(aRay.Direction, qVec);
    if (aV < 0) or (aU + aV > det) then begin
      result := false;
      exit;
    end;

    // calculate t, scale parameters, ray intersects triangle
    aT := vertexDotProduct(edge2, qVec);
    invDet := 1 / det;
    aT := aT * invDet;
    aU := aU * invDet;
    aV := aV * invDet;

    result := true;
end;

// Point in triangle test returning barycentric coordinates (for easy normal interpolation)
function IsPointInTriangle(const aPoint: TVertex; const aVertex0, aVertex1, aVertex2: TVertex; var aU, aV: Single): boolean;
const EPSILON = 0.0001;
var r, q1, q2: TVertex;
    q1Sqr, q2Sqr, q1q2, rq1, rq2, det, invDet: single;
begin
    result := FALSE;

    r := VertexSubtract( aPoint, aVertex0 );
    q1 := VertexSubtract( aVertex1, aVertex0 );
    q2 := VertexSubtract( aVertex2, aVertex0 );
    q1Sqr := VertexDotProduct( q1, q1 );
    q2Sqr := VertexDotProduct( q2, q2 );
    q1q2 := VertexDotProduct( q1, q2 );
    rq1 := VertexDotProduct( r, q1 );
    rq2 := VertexDotProduct( r, q2 );
    det := q1Sqr * q2Sqr - q1q2 * q1q2;
    if det = 0 then exit;
    invDet := 1 / det;
    aU := invDet * ( q2Sqr * rq1 - q1q2 * rq2 );
    aV := invDet * ( q1Sqr * rq2 - q1q2 * rq1 );

    result := (aU >= 0 - EPSILON) and (aV >= 0 - EPSILON) and (aU + aV <= 1 + EPSILON);
end;

function PointInFrustum(aPoint: TVertex; aFrustumPlanes,aFrustumPoints: array of TVertex): boolean;
// based on rectangular selection in 3d OpenGL view
// http://www.3dkingdoms.com/selection.html
var
    Dot: single;
    x: integer;
begin
    Result := false;
    // Check if point is within the Frustum formed by the selected rectangle.

    for x := 0 to High(aFrustumPlanes) do begin
        Dot := vertexDotProduct(vertexSubtract(aPoint,aFrustumPoints[x*2]), aFrustumPlanes[x] );
        if Dot > 0 then Exit;
    end;
    Result := true;
end;

function Polygon2DFacesViewer(const aPoints: TPoint2DArray): boolean;
{
p0, p1, p2 are the first 3 points of the 2d polygon (should be anti-clockwise to face viewer).
n is the normal formed from the cross-product of the 2 vectors (p1 - p0) and (p2 - p0).

n.Z will be > 0 if the polygon normal faces viewer.

  n
  |   p2
  |  /
  | /
  |/
p0\
   \
    \
     p1
}
var
    p0,p1,p2: TVertex;
    a,b,n: TVertex;
begin
    Result := false;
    if Length(aPoints) < 3 then exit;
    p0 := Vertex(aPoints[0].X, aPoints[0].Y, 0);
    p1 := Vertex(aPoints[1].X, aPoints[1].Y, 0);
    p2 := Vertex(aPoints[2].X, aPoints[2].Y, 0);
    a := vertexSubtract(p1,p0);
    b := vertexSubtract(p2,p0);
    n := vertexCrossProduct(a,b);
    Result := n.Z > 0;
end;

function Intersect(const aRay: TRay; const aPlane: TVertex4): TVertex;
const EPSILON = 0.0001;
var nominator, denominator: single;
begin
    denominator :=
        aPlane[0] * ( aRay.Destination.X - aRay.Origin.X ) +
        aPlane[1] * ( aRay.Destination.Y - aRay.Origin.Y ) +
        aPlane[2] * ( aRay.Destination.Z - aRay.Origin.Z );

    if abs(denominator) < EPSILON then
    begin
        result := Vertex( 0, 0, 0 );
        exit;
    end;

    nominator := aPlane[3] - aPlane[0] * aRay.Origin.X - aPlane[1] * aRay.Origin.Y - aPlane[2] * aRay.Origin.Z;

    result := VertexAdd( aRay.Origin, VertexMultiply( VertexSubtract( aRay.Destination, aRay.Origin ), nominator / denominator ));
end;

// ### TMatrix #################################################################

constructor TMatrix.Create(NRows,NColumns: integer);
var i: integer;
begin
    inherited Create;
    if (NRows>0) and (NColumns>0) then begin
        numRows := NRows;
        numColumns := NColumns;
    end else begin                                                              // standard matrix dimensions
        numRows := 4;
        numColumns := 4;
    end;
    setlength(matrix,numRows);                                                  // create the actual matrix
    for i := Low(matrix) to High(matrix) do
        setlength(matrix[i],numColumns);
end;

destructor TMatrix.Destroy;
begin
    finalize(matrix);
    inherited;
end;

function TMatrix.GetValue(ARow,AColumn: integer): double;
begin
    if (ARow>=0) and (ARow<numRows) and (AColumn>=0) and (AColumn<numColumns) then
        Result := matrix[ARow,AColumn]
    else
        Result := -1;
end;

procedure TMatrix.SetValue(ARow,AColumn: integer; value: double);
begin
    if (ARow>=0) and (ARow<numRows) and (AColumn>=0) and (AColumn<numColumns) then
        matrix[ARow,AColumn] := Value;
end;

function TMatrix.Add(a, b: TMatrix): TMatrix;
var r: TMatrix; i,j: integer;
begin
    Result := nil;
    if (a.numRows<>b.numRows) or (a.numColumns<>b.numColumns) then exit;
    r := TMatrix.Create(a.numRows,a.numColumns);
    for i := 0 to numRows-1 do
        for j := 0 to numColumns-1 do
            matrix[i,j] := a.GetValue(i,j) + b.GetValue(i,j);
    Result := r;
end;

function TMatrix.Sub(a, b: TMatrix): TMatrix;
var r: TMatrix; i,j: integer;
begin
    Result := nil;
    if (a.numRows<>b.numRows) or (a.numColumns<>b.numColumns) then exit;
    r := TMatrix.Create(a.numRows,a.numColumns);
    for i := 0 to numRows-1 do
        for j := 0 to numColumns-1 do
            r.matrix[i,j] := a.GetValue(i,j) - b.GetValue(i,j);
    Result := r;
end;

function TMatrix.Mul(a,b: TMatrix): TMatrix;
var r: TMatrix; i,j: integer;
begin
    r := TMatrix.Create(a.numRows,b.numColumns);
    for i := 0 to 3 do
        for j := 0 to 3 do
            r.matrix[i,j] := a.matrix[i,j] * b.matrix[j,i];
    Result := r;
end;

// ### TIntegerMatrix ##########################################################

function TIntegerMatrix.GetValue(ARow, AColumn: integer): integer;
begin
    Result := matrix[ARow,AColumn];
end;

procedure TIntegerMatrix.SetValue(ARow, AColumn, Value: integer);
begin
    matrix[ARow,AColumn] := Value;
end;

constructor TIntegerMatrix.Create(NRows, NColumns: integer);
var i: integer;
begin
    inherited Create;
    numRows := NRows;
    numColumns := NColumns;
    setlength(matrix,numRows);
    for i := Low(matrix) to High(matrix) do
        setlength(matrix[i],NColumns);
end;

destructor TIntegerMatrix.Destroy;
begin
    finalize(matrix);
    inherited;
end;

end.
