{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_NoiseTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Options,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool,
  unit_PrimitiveList;

type
  TNoiseTool = class(TPrimitiveTool)
  private
    procedure DoNoise(AParamList: TParamList);
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  unit_PerlinNoise;
  
procedure TNoiseTool.DoNoise(AParamList: TParamList);
const
    cYPos       = 0;
    cYNeg       = 1;
    cXNeg       = 2;
    cXPos       = 3;
    cZPos       = 4;
    cZNeg       = 5;
    cOutwards   = 6;

var
    usePerlin: Boolean;
    noiseSeed: integer;
    minNoise: single;
    maxNoise: single;
    persistence: single;
    frequency: single;
    octaves: integer;
    perlinNoise: TPerlinNoise;
    vertex: TVertice;
    scale: single;
    offset: single;
    normal: TVertice;
    center: TVertice;
    primitive: TBasePrimitive;
    p: integer;
    i: integer;
begin
    usePerlin   := TBooleanObject(AParamList.NamedParameters['usePerlin']).Value;
    noiseSeed   := TIntegerObject(AParamList.NamedParameters['noiseSeed']).Value;
    minNoise    := TSingleObject(AParamList.NamedParameters['minNoise']).Value;
    maxNoise    := TSingleObject(AParamList.NamedParameters['maxNoise']).Value;
    persistence := TSingleObject(AParamList.NamedParameters['persistence']).Value;
    frequency   := TSingleObject(AParamList.NamedParameters['frequency']).Value;
    octaves     := TIntegerObject(AParamList.NamedParameters['octaves']).Value;

    // scale vertices using noise (perlin, or fbm)
    perlinNoise := TPerlinNoise.Create(noiseSeed);
    try
        for p := 0 to SelectedPrimitives.Count - 1 do begin
            primitive := SelectedPrimitives.Primitives[p];
            if primitive.ModifyGeometry then begin

                // calculate vertex normals
                primitive.InitializeNormals(66);
            
                center    := primitive.Center;
            
                for i := 0 to primitive.Vertices.Count - 1 do begin
                    vertex := primitive.Vertices.Vertices[i];
                    normal := vertex.GetNormal();

                    // use vertex vector for noise
                    if usePerlin then
                        scale := perlinNoise.PerlinNoise3d(vertex.x - center.x,vertex.y - center.y,vertex.z - center.z,persistence,frequency,octaves)
                    else
                        scale := perlinNoise.fbm(vertex.x - center.x,vertex.y - center.y,vertex.z - center.z,octaves);

                    // transform theoretical ramdom noise values from [-1,+1] to [0..1]
                    scale := (scale + 1) / 2;

                    // calculate offset to match min/max noise using scale
                    offset := minNoise + (maxNoise - minNoise) * scale;

                    // now move vertex to new position using normal and offset
                    vertex.x := vertex.x + normal.x * offset;
                    vertex.y := vertex.y + normal.y * offset;
                    vertex.z := vertex.z + normal.z * offset;
                    normal.Free;
                end;
            end;
        end;
    finally
        perlinNoise.Free;
    end;
end;

procedure TNoiseTool.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Move primitive vertices using noise';
    Description := 'Moves primitive vertices using noise';
end;

procedure TNoiseTool.DoExecute(AParamList: TParamList);
begin
    try
        if not Assigned(AParamList) then
            raise Exception.Create('No parameters for Noise tool provided.')
        else begin
            if AParamList.Count <> 7 then
                raise Exception.Create('Wrong number of parameters for Noise tool provided. Received "'+IntToStr(AParamList.Count)+
                                       '" parameters, expected 7.')
            else begin
                if SelectedPrimitives.Count<1 then
                    raise Exception.Create('Select 1 or more objects to warp.')
                else begin
                    DoNoise(AParamList);
                end;
            end;
        end;
    except
        on E: Exception do MessageDlg(E.Message,mtError,[mbOK],0);
    end;
end;

end.
