{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Octree;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  Simple octree.
  ---------------------------------------------------------------------------- }

interface

uses
  unit_Types;

type
  TOctreeTriangle = Array[0..2] of TVertex;
  TTriangleProviderMethod = procedure( const anIndex: integer; var aTriangle: TOctreeTriangle ) of object;

// ----------------------------------------------------------------------------

  TOctreeAABB = record
      Min, Max: TVertex;
  end;

// ----------------------------------------------------------------------------

  TOctreeItem = record
      TriangleIndex: integer;
      IsFullyContained: boolean;
  end;

  TOctreeItemArray = Array of TOctreeItem;

// ----------------------------------------------------------------------------

  TOctreeBuildInfo = record
      NrTriangles: integer;
      MinTrianglesPerNode: integer;
      MinNodeSize: single;
      MaxNrLevels: integer;
      TriangleProvider: TTriangleProviderMethod;
  end;

// ----------------------------------------------------------------------------

  TOctreeNode = class;
  TOctreeChildrenArray = Array[0..7] of TOctreeNode;
  TOctreeNode = class( TObject )
  private
      fParent: TOctreeNode;
      fChildren: TOctreeChildrenArray;

      fIndex: integer;
      fAABB: TOctreeAABB;
      fCenter: TVertex;
      fSize: single;

      fItems: TOctreeItemArray;

      function CalculateOutCode( const aPoint, aMin, aMax: TVertex ): TClippingOutcode;

      function IsPlaneOverlappingAABB( const aNormal: TVertex; const d: single; const aMaxBox: TVertex ): boolean;
      function IsTriangleOverlappingAABB( const aTriangle: TOctreeTriangle; const aBoxCenter, aBoxHalfSize: TVertex ): boolean;
      function NrTriangleVerticesInsideAABB( const aTriangle: TOctreeTriangle; const anAABB: TOctreeAABB ): integer;

  public
      constructor Create( const aParent: TOctreeNode; const anIndex: integer );
      destructor Destroy; override;

      procedure Initialize;
      procedure ResolveItems( const anInfo: TOctreeBuildInfo );
      procedure ResolveChildren( const anInfo: TOctreeBuildInfo; const aLevel: integer );

      function IntersectsLineSegment(const aP1, aP2: TVertex): boolean;

      property Parent: TOctreeNode read fParent;
      property Children: TOctreeChildrenArray read fChildren;

      property Items: TOctreeItemArray read fItems;
      property AABB: TOctreeAABB read fAABB write fAABB;
      property Size: single read fSize write fSize;
      property Center: TVertex read fCenter write fCenter;
  end;

// ----------------------------------------------------------------------------

  TOctree = class( TObject )
  private
      fNrNodes: integer;
      fNrTerminalNodes: integer;

      fRootNode: TOctreeNode;

      procedure InitializeRootNode( const anInfo: TOctreeBuildInfo );

      procedure CountNode( const aNode: TOctreeNode; var aNrNodes, aNrTerminalNodes: integer );

  public
      constructor Create( const aNrTriangles: integer;
          const aMinTrianglesPerNode: integer;
          const aMinNodeSize: single;
          const aMaxNrLevels: integer;
          const aTriangleProvider: TTriangleProviderMethod );

      destructor Destroy; override;

      property RootNode: TOctreeNode read fRootNode;
      property NrTerminalNodes: Integer read fNrTerminalNodes;
  end;

// ============================================================================
// ============================================================================

implementation

uses
  SysUtils,
  unit_Math;

const
    EPSILON = 0.001;

constructor TOctreeNode.Create( const aParent: TOctreeNode; const anIndex: integer );
begin
  inherited Create;

  fParent := aParent;
  fIndex := anIndex;

  Initialize;
end;

// ----------------------------------------------------------------------------

destructor TOctreeNode.Destroy;
var
  index: integer;

begin
  for index := Low( fChildren ) to High( fChildren ) do begin
      FreeAndNil( fChildren[index] );
  end;

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure TOctreeNode.Initialize;
var
  radius: single;

begin
  if Assigned( Parent ) then begin
      // INFO, PJ: The rootnode is resolved from outside (directly from the octree)
      radius := fParent.Size / 4;

      with fParent.Center do begin
          case fIndex of
              0: fCenter := vertex( X + radius, Y + radius, Z + radius );
              1: fCenter := vertex( X - radius, Y + radius, Z + radius );
              2: fCenter := vertex( X + radius, Y + radius, Z - radius );
              3: fCenter := vertex( X - radius, Y + radius, Z - radius );
              4: fCenter := vertex( X + radius, Y - radius, Z + radius );
              5: fCenter := vertex( X - radius, Y - radius, Z + radius );
              6: fCenter := vertex( X + radius, Y - radius, Z - radius );
              7: fCenter := vertex( X - radius, Y - radius, Z - radius );
          end;
      end;

      fAABB.Min := vertexAddScalar( fCenter, -radius );
      fAABB.Max := vertexAddScalar( fCenter, radius );

      fSize := fParent.Size / 2;
  end;
end;

// ----------------------------------------------------------------------------

procedure TOctreeNode.ResolveItems( const anInfo: TOctreeBuildInfo );
var
  boxHalfSize: TVertex;
  index: integer;
  usedIndex: integer;
  triangle: TOctreeTriangle;
  nrVerticesInside: integer;

begin
  if Assigned( fParent ) then begin
      boxHalfSize := vertex( fSize / 2, fSize / 2, fSize / 2 );
      SetLength( fItems, Length( fParent.Items ));

      usedIndex := 0;
      for index := Low( fParent.Items ) to High( fParent.Items ) do begin
          anInfo.TriangleProvider( fParent.Items[index].TriangleIndex, triangle );

          nrVerticesInside := nrTriangleVerticesInsideAABB( triangle, fAABB );

          if  ( nrVerticesInside > 0 ) or
              isTriangleOverlappingAABB(  triangle, fCenter, boxHalfSize ) then begin
              fItems[usedIndex].TriangleIndex := fParent.Items[index].TriangleIndex;
              fItems[usedIndex].IsFullyContained := ( nrVerticesInside = 3 );

              usedIndex := usedIndex + 1;
          end;
      end;

      // INFO, PJ: Adjust buffer to size actually used
      SetLength( fItems, usedIndex );
  end
  else begin // INFO, PJ: Root node!
      SetLength( fItems, anInfo.NrTriangles );

      for index := 0 to anInfo.NrTriangles - 1 do begin
          fItems[index].TriangleIndex := index;
          fItems[index].IsFullyContained := TRUE;
      end;
  end;
end;

// ----------------------------------------------------------------------------

procedure TOctreeNode.ResolveChildren( const anInfo: TOctreeBuildInfo; const aLevel: integer );
var
  index: integer;
  childNode: TOctreeNode;

begin
  // INFO, PJ: Check if any of the stop conditions is reached.
  if  (( anInfo.MinTrianglesPerNode > 0 ) and ( Length( fItems ) < anInfo.MinTrianglesPerNode )) or
      (( anInfo.MinNodeSize > 0 ) and ( fSize < anInfo.MinNodeSize )) or
      (( anInfo.MaxNrLevels > 0 ) and ( aLevel > anInfo.MaxNrLevels )) then begin
      Exit;
  end;

  for index := Low( fChildren ) to High( fChildren ) do begin
      // INFO, PJ: Create the child node
      childNode := TOctreeNode.Create( self, index );
      childNode.ResolveItems( anInfo );

      if Length( childNode.Items ) = 0 then begin
          // INFO, PJ: No items in this node, nillify it
          FreeAndNil( childNode );
      end;

      fChildren[index] := childNode;
  end;

  // INFO, PJ: Now is the time to free our own items (items at terminal nodes only)
  fItems := nil;

  // INFO, PJ: Recurse down
  for index := Low( fChildren ) to High( fChildren ) do begin
      if Assigned( fChildren[index] ) then
      begin
          fChildren[index].ResolveChildren( anInfo, aLevel + 1 );
      end;
  end;
end;

// ----------------------------------------------------------------------------

function TOctreeNode.CalculateOutCode( const aPoint, aMin, aMax: TVertex ): TClippingOutcode;
// INFO, PJ: calculates the cohen-sutherland outcode for a point and a bounding box.
begin
  Result := [];

  if aPoint.X > aMax.X then
  begin
      Result := Result + [coClipRight];
  end
  else if aPoint.X < aMin.X then
  begin
      Result := Result + [coClipLeft];
  end;

  if aPoint.Y > aMax.Y then
  begin
      Result := Result + [coClipTop];
  end
  else if aPoint.Y < aMin.Y then
  begin
      Result := Result + [coClipBottom];
  end;

  if aPoint.Z > aMax.Z then
  begin
      Result := Result + [coClipBack];
  end
  else if aPoint.Z < aMin.Z then
  begin
      Result := Result + [coClipFront];
  end;
end;

// ----------------------------------------------------------------------------

// INFO, PJ: Implementation of the cohen-sutherland line-clipping algorithm.
function TOctreeNode.IntersectsLineSegment(const aP1, aP2: TVertex): boolean;
var
  outcode1, outcode2: TClippingOutcode;
  min, max: TVertex;
  v1, v2: single;
  intersection: TVertex;

begin
  Result := false;

//  min := vertexAddScalar( AABB.Min, -EPSILON );
//  max := vertexAddScalar( AABB.Max, EPSILON );
// !!!

    min := AABB.Min;
    max := AABB.Max;

  // point 1 inside bounding box
  outcode1 := calculateOutCode( aP1, min, max );
  if ( outcode1 = [] ) then
  begin
    Result := true;
    exit;
  end;

  // point 2 inside bounding box
  outcode2 := calculateOutCode( aP2, min, max );
  if ( outcode2 = [] ) then
  begin
    Result := true;
    exit;
  end;

  // both points on same side of box
  if ( outcode1 * outcode2 ) <> [] then
  begin
      exit;
  end;

  // check intersections
  if ( coClipLeft in outcode1 ) or ( coClipRight in outCode1 ) then
  begin
      if coClipRight in outcode1 then
      begin
          intersection.X := max.X;
      end
      else
          intersection.X := min.X;

      v1 := aP2.X - aP1.X;
      v2 := intersection.X - aP1.X;
      intersection.Y := aP1.Y + v2 * ( aP2.Y - aP1.Y ) / v1;
      intersection.Z := aP1.Z + v2 * ( aP2.Z - aP1.Z ) / v1;
      if ( intersection.Y <= max.Y ) and ( intersection.Y >= min.Y ) and
         ( intersection.Z <= max.Z ) and ( intersection.Z >= min.Z ) then
      begin
          Result := true;
          exit;
      end;
  end;

  if ( coClipTop in outcode1 ) or ( coClipBottom in outcode1 ) then
  begin
      if coClipTop in outcode1 then
      begin
          intersection.Y := max.Y;
      end
      else
          intersection.Y := min.Y;

      v1 := aP2.Y - aP1.Y;
      v2 := intersection.Y - aP1.Y;
      intersection.X := aP1.X + v2 * ( aP2.X - aP1.X ) / v1;
      intersection.Z := aP1.Z + v2 * ( aP2.Z - aP1.Z ) / v1;
      if ( intersection.X <= max.X ) and ( intersection.X >= min.X ) and
         ( intersection.Z <= max.Z ) and ( intersection.Z >= min.Z ) then
      begin
          result := true;
          exit;
      end;
  end;

  if ( coClipFront in outcode1 ) or ( coClipBack in outcode1 )  then
  begin
      if coClipBack in outCode1 then
      begin
          intersection.Z := max.Z;
      end
      else
          intersection.Z := min.Z;

      v1 := aP2.Z - aP1.Z;
      v2 := intersection.Z - aP1.Z;
      intersection.X := aP1.X + v2 * ( aP2.X - aP1.X ) / v1;
      intersection.Y := aP1.Y + v2 * ( aP2.Y - aP1.Y ) / v1;
      if ( intersection.X <= max.X ) and ( intersection.X >= min.X ) and
         ( intersection.Y <= max.Y ) and ( intersection.Y >= min.Y ) then
      begin
          Result := true;
          exit;
      end;
  end;
end;

// ----------------------------------------------------------------------------

function TOctreeNode.isPlaneOverlappingAABB( const aNormal: TVertex; const d: single; const aMaxBox: TVertex ): boolean;
var
  index: integer;
  vMin, vMax: TVertex;

begin
  for index := 0 to 2 do begin
      if aNormal.Elements[index] > 0 then begin
          vMin.Elements[index] := -aMaxBox.Elements[index];
          vMax.Elements[index] := aMaxBox.Elements[index];
      end
      else begin
          vMin.Elements[index] := aMaxBox.Elements[index];
          vMax.Elements[index] := -aMaxBox.Elements[index];
      end;
  end;

  if vertexDotProduct( aNormal, vMin ) + d > 0 then begin
      result := FALSE;
  end
  else if vertexDotProduct( aNormal, vMax ) + d >= 0 then begin
      result := TRUE;
  end
  else
      result := FALSE;
end;

// ----------------------------------------------------------------------------

function TOctreeNode.nrTriangleVerticesInsideAABB( const aTriangle: TOctreeTriangle; const anAABB: TOctreeAABB ): integer;
var
  index: integer;

begin
  Result := 0;

  for index := Low( aTriangle ) to High( aTriangle ) do begin
      with aTriangle[index] do begin
          if  ( X >= anAABB.Min.X ) and ( X <= anAABB.Max.X ) and
              ( Y >= anAABB.Min.Y ) and ( Y <= anAABB.Max.Y ) and
              ( Z >= anAABB.Min.Z ) and ( Z <= anAABB.Max.Z ) then begin

              Result := Result + 1;
          end;
      end;
  end;
end;

// ----------------------------------------------------------------------------

function TOctreeNode.isTriangleOverlappingAABB( const aTriangle: TOctreeTriangle; const aBoxCenter, aBoxHalfSize: TVertex ): boolean;
// INFO, PJ: Original Delphi translation by Lord Trancos. Algorithm by Tomas Akenine-Möller.
var
  v0, v1, v2: TVertex;
  e0, e1, e2: TVertex;
  normal: TVertex;
  d: single;
  min, max, fex, fey, fez: single;
  p0, p1, p2, rad: single;

  // ------------------------------------------------------------

  function axisTestX01(const a, b, fa, fb: single): boolean;
  begin
      Result := FALSE;

      p0 := a * v0.Y - b * v0.Z;
      p2 := a * v2.Y - b * v2.Z;
      if p0 < p2 then begin
          min := p0;
          max := p2;
      end
      else begin
          min := p2;
          max := p0;
      end;
      rad := fa * aBoxHalfSize.Y + fb * aBoxHalfSize.Z;
      if (min > rad) or (max < -rad) then exit;

      Result := TRUE;
  end;

  // ------------------------------------------------------------

  function axisTestX2(const a, b, fa, fb: single): boolean;
  begin
      Result := FALSE;

      p0 := a * v0.Y - b * v0.Z;
      p1 := a * v1.Y - b * v1.Z;
      if p0 < p1 then begin
          min := p0;
          max := p1;
      end
      else begin
          min := p1;
          max := p0;
      end;
      rad := fa * aBoxHalfSize.Y + fb * aBoxHalfSize.Z;
      if (min > rad) or (max < -rad) then exit;

      Result := TRUE;
  end;

  // ------------------------------------------------------------

  function axisTestY02(const a, b, fa, fb: single): boolean;
  begin
      Result := FALSE;

      p0 := -a * v0.X + b * v0.Z;
      p2 := -a * v2.X + b * v2.Z;
      if p0 < p2 then begin
          min := p0;
          max := p2;
      end
      else begin
          min := p2;
          max := p0;
      end;
      rad := fa * aBoxHalfSize.X + fb * aBoxHalfSize.Z;
      if (min > rad) or (max < -rad) then exit;

      Result := TRUE;
  end;

  // ------------------------------------------------------------

  function axisTestY1(const a, b, fa, fb: single): boolean;
  begin
      Result := FALSE;

      p0 := -a * v0.X + b * v0.Z;
      p1 := -a * v1.X + b * v1.Z;
      if p0 < p1 then begin
          min := p0;
          max := p1;
      end
      else begin
          min := p1;
          max := p0;
      end;
      rad := fa * aBoxHalfSize.X + fb * aBoxHalfSize.Z;
      if (min > rad) or (max < -rad) then exit;

      Result := TRUE;
  end;

  // ------------------------------------------------------------

  function axisTestZ12(const a, b, fa, fb: single): boolean;
  begin
      Result := FALSE;

      p1 := a * v1.X - b * v1.Y;
      p2 := a * v2.X - b * v2.Y;
      if p2 < p1 then begin
          min := p2;
          max := p1;
      end
      else begin
          min := p1;
          max := p2;
      end;
      rad := fa * aBoxHalfSize.X + fb * aBoxHalfSize.Y;
      if (min > rad) or (max < -rad) then exit;

      Result := TRUE;
  end;

  // ------------------------------------------------------------

  function axisTestZ0(const a, b, fa, fb: single): boolean;
  begin
      Result := FALSE;

      p0 := a * v0.X - b * v0.Y;
      p1 := a * v1.X - b * v1.Y;
      if p0 < p1 then begin
          min := p0;
          max := p1;
      end
      else begin
          min := p1;
          max := p0;
      end;
      rad := fa * aBoxHalfSize.X + fb * aBoxHalfSize.Y;
      if (min > rad) or (max < -rad) then exit;

      Result := TRUE;
  end;

  // ------------------------------------------------------------

  procedure findMinMax(const anX0, anX1, anX2: single; var aMin, aMax: single);
  begin
      aMin := anX0;
      aMax := anX0;
      if (anX1 < aMin) then aMin := anX1;
      if (anX1 > aMax) then aMax := anX1;
      if (anX2 < aMin) then aMin := anX2;
      if (anX2 > aMax) then aMax := anX2;
  end;


begin
  Result := FALSE;

  // move everything so that the boxcenter is in ( 0,0,0 )
  v0 := vertexSubtract( aTriangle[0], aBoxCenter );
  v1 := vertexSubtract( aTriangle[1], aBoxCenter );
  v2 := vertexSubtract( aTriangle[2], aBoxCenter );

  // compute triangle edges
  e0 := vertexSubtract( v1, v0 ); // edge 0
  e1 := vertexSubtract( v2, v1 ); // edge 1
  e2 := vertexSubtract( v0, v2 ); // edge 2

  // test the 9 tests first ( this was faster )
  fex := abs( e0.X );
  fey := abs( e0.Y );
  fez := abs( e0.Z );
  if not axisTestX01( e0.Z, e0.Y, fez, fey ) then exit;
  if not axisTestY02( e0.Z, e0.X, fez, fex ) then exit;
  if not axisTestZ12( e0.Y, e0.X, fey, fex ) then exit;

  fex := abs( e1.X );
  fey := abs( e1.Y );
  fez := abs( e1.Z );
  if not axisTestX01( e1.Z, e1.Y, fez, fey ) then exit;
  if not axisTestY02( e1.Z, e1.X, fez, fex ) then exit;
  if not axisTestZ0( e1.Y, e1.X, fey, fex ) then exit;

  fex := abs( e2.X );
  fey := abs( e2.Y );
  fez := abs( e2.Z );
  if not axisTestX2( e2.Z, e2.Y, fez, fey ) then exit;
  if not axisTestY1( e2.Z, e2.X, fez, fex ) then exit;
  if not axisTestZ12( e2.Y, e2.X, fey, fex ) then exit;

  // test overlap in the {x,y,z}-directions
  // find min, max of the triangle for each direction, and test for overlap in
  // that direction -- this is equivalent to testing a minimal AABB around
  // the triangle against the AABB

  // test in X-direction
  findMinMax( v0.X, v1.X, v2.X, min, max );
  if ( min > aBoxHalfSize.X ) or ( max < -aBoxHalfSize.X ) then exit;

  // test in Y-direction
  findMinMax( v0.Y, v1.Y, v2.Y, min, max );
  if ( min > aBoxHalfSize.Y ) or ( max < -aBoxHalfSize.Y ) then exit;

  // test in Z-direction
  findMinMax( v0.Z, v1.Z, v2.Z, min, max );
  if ( min > aBoxHalfSize.Z ) or ( max < -aBoxHalfSize.Z ) then exit;

  // test if the box intersects the plane of the triangle
  // compute plane equation of triangle: normal*x+d=0
  normal := vertexCrossProduct( e0, e1 );
  d := - vertexDotProduct( normal, v0 ); // plane eq: normal.x+d=0
  if not isPlaneOverlappingAABB( normal, d, aBoxHalfSize ) then exit;

  Result := TRUE; // box and triangle overlaps
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TOctree.create( const aNrTriangles: integer;
  const aMinTrianglesPerNode: integer;
  const aMinNodeSize: single;
  const aMaxNrLevels: integer;
  const aTriangleProvider: TTriangleProviderMethod );

var
  info: TOctreeBuildInfo;

begin
  inherited create;

  if  ( aMinTrianglesPerNode = 0 ) and
      ( aMinNodeSize = 0 ) and
      ( aMaxNrLevels = 0 ) then begin
      raise Exception.create( 'Cannot build octree. ' +
          ' Specify either a minimum number of triangles per node, a minimum node AABB size, ' +
          'or a maximum level of splitting.' );
  end;

  info.NrTriangles := aNrTriangles;
  info.MinTrianglesPerNode := aMinTrianglesPerNode;
  info.MinNodeSize := aMinNodeSize;
  info.MaxNrLevels := aMaxNrLevels;
  info.TriangleProvider := aTriangleProvider;

  InitializeRootNode( info );
  fRootNode.ResolveItems( info );
  fRootNode.ResolveChildren( info, 1 );

  CountNode( fRootNode, fNrNodes, fNrTerminalNodes );
end;

// ----------------------------------------------------------------------------

destructor TOctree.Destroy;
begin
  FreeAndNil( fRootNode );

  Inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure TOctree.InitializeRootNode( const anInfo: TOctreeBuildInfo );
var
  triangle: TOctreeTriangle;
  triangleIndex: integer;
  vertexIndex: integer;
  xSize, ySize, zSize: single;
  maxSize: single;
  aabb: TOctreeAABB;

begin
  fRootNode := TOctreeNode.Create( nil, 0 );

  // INFO, PJ: First we calculate the AABB
  for triangleIndex := 0 to anInfo.NrTriangles - 1 do begin
      anInfo.TriangleProvider( triangleIndex, triangle );

      for vertexIndex := 0 to 2 do begin
          if  ( vertexIndex = 0 ) and
              ( triangleIndex = 0 ) then begin
              aabb.Min := triangle[0];
              aabb.Max := triangle[0];
          end
          else begin
              aabb.Min := vertexMinElements( aabb.Min, triangle[vertexIndex] );
              aabb.Max := vertexMaxElements( aabb.Max, triangle[vertexIndex] );
          end;
      end;
  end;

  // INFO, PJ: Change AABB into a cube
  xSize := aabb.Max.X - aabb.Min.X;
  ySize := aabb.Max.Y - aabb.Min.Y;
  zSize := aabb.Max.Z - aabb.Min.Z;

  maxSize := xSize;
  if maxSize < ySize then maxSize := ySize;
  if maxSize < zSize then maxSize := zSize;

  // INFO, PJ: Delta sizes needed to make a cube
  xSize := ( maxSize - xSize  ) / 2;
  ySize := ( maxSize - ySize ) / 2;
  zSize := ( maxSize - zSize ) / 2;

  aabb.Min.X := aabb.Min.X - xSize;
  aabb.Max.X := aabb.Max.X + xSize;

  aabb.Min.Y := aabb.Min.Y - ySize;
  aabb.Max.Y := aabb.Max.Y + ySize;

  aabb.Min.Z := aabb.Min.Z - zSize;
  aabb.Max.Z := aabb.Max.Z + zSize;

  fRootNode.AABB := aabb;

  // INFO, PJ: Calculate center point
  fRootNode.Size := maxSize;
  fRootNode.Center := vertexAddScalar( fRootNode.AABB.Min, maxSize / 2 );
end;

// ----------------------------------------------------------------------------

procedure TOctree.CountNode( const aNode: TOctreeNode; var aNrNodes, aNrTerminalNodes: integer );
var
  index: integer;

begin
  aNrNodes := aNrNodes + 1;
  if Length( aNode.Items ) = 0 then begin
      aNrTerminalNodes := aNrTerminalNodes + 1;
  end;

  for index := Low( aNode.Children ) to High( aNode.Children ) do begin
      if Assigned( aNode.Children[index] ) then begin
          CountNode( aNode.Children[index], aNrNodes, aNrTerminalNodes );
      end;
  end;
end;


end.

