{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PNGSupport;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  Based on the specifications in Wikipedia and the official W3C PNG specs.

  PJ: PNGWriter is far from complete yet.
      - check to see if CRC calculations are correct (probably not)
      - Deflate algorithm doesn't seem to work properly (barely any compression)
      - Filters need to be applied on the raw data.
  ---------------------------------------------------------------------------- }

interface

uses
    Classes,
    MZLib,
    unit_TextureTypes;

type
    TChunkID = packed array[0..3] of char;
    TChunkHeader = packed record
        Length: cardinal;
        ChunkID: TChunkID;
    end;

    TBuffer = record
        Size: integer;
        Data: pointer;
    end;

    TPaletteEntry = packed record
        R, G, B: byte;
    end;

// ----------------------------------------------------------------------------

    TPNGProcessor = class(TObject)
    protected
        // utility function
        function SwapLong( const aValue: cardinal ): cardinal;

    end;

// ----------------------------------------------------------------------------

    TPNGLoader = class(TPNGProcessor)
    private
        fStream: TStream;

        // Various properties cached from the header
        fIndexed: boolean;
        fInterlaced: boolean;
        f16BitChannels: boolean;
        fGreyScale: boolean;
        fPalette: array of TPaletteEntry;
        fAlphaPalette: array of byte;

        procedure Decode(const aSource, aDestination: Pointer; const aPackedSize, anUnpackedSize: Integer);
        function ContainsAlphaPalette: boolean;

        procedure ProcessCRC;
        function ReadImageData: TBuffer;
        function ReadChunkHeader: TChunkHeader;
        procedure ReadPaletteChunk( const aChunkHeader: TChunkHeader );
        procedure ReadTransparencyChunk( const aChunkHeader: TChunkHeader );

        function UnpackData(const aBuffer: TBuffer; const aTargetSize: integer): TBuffer;
        procedure ConvertImageData(const aBuffer: TBuffer; const aTextureData: TTextureData);
        procedure ConvertColorData( const aSource, aTarget: PByte; const aWidth, aNumBytesPerPixel, anOffset, aStride: integer );
        procedure ApplyFilter(aFilter: Byte; aLineBuffer, aPreviousLineBuffer, aTargetBuffer: PByte; aBytesPerPixel, aBytesPerRow: Integer);

    public
        constructor Create(aStream: TStream);

        function LoadHeader: TTextureData;
        function Load: TTextureData;
    end;

// ----------------------------------------------------------------------------

    TPNGWriter = class(TPNGProcessor)
    private
        fStream: TStream;

        procedure WriteHeader(const aTextureData: TTextureData);
        procedure WriteImageData(const aTextureData: TTextureData);
        procedure WriteChunk(const aChunk; const aChunkID: TChunkID; const aSize: integer);

        function ConvertImageData(const aTextureData: TTextureData): TBuffer;
        procedure Encode(const aSource, aDestination: Pointer; const anUnpackedSize: Integer; var aPackedSize: Integer);
        function PackData(const aBuffer: TBuffer; const aTargetSize: integer): TBuffer;

    public
        constructor Create(aStream: TStream);

        procedure Write(const aTextureData: TTextureData);
    end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
    SysUtils, Math;

type
    TPNGHeaderMagic = packed array[0..7] of char;

const
    PNG_HEADER_MAGIC: TPNGHeaderMagic = #$89#$50#$4E#$47#$0D#$0A#$1A#$0A;
    CHUNK_ID_IHDR = 'IHDR';
    CHUNK_ID_IDAT = 'IDAT';
    CHUNK_ID_IEND = 'IEND';
    CHUNK_ID_PLTE = 'PLTE';
    CHUNK_ID_TRNS = 'tRNS';

    COLOR_TYPE_GREYSCALE        =       0;
    COLOR_TYPE_TRUECOLOR        =       2;
    COLOR_TYPE_INDEXED          =       3;
    COLOR_TYPE_GREYSCALE_ALPHA  =       4;
    COLOR_TYPE_TRUECOLOR_ALPHA  =       6;

    // Adam7 interlacing
    INTERLACE_ROW_START: array[0..6] of Integer = (0, 0, 4, 0, 2, 0, 1);
    INTERLACE_COLUMN_START: array[0..6] of Integer = (0, 4, 0, 2, 0, 1, 0);
    INTERLACE_ROW_INCREMENT: array[0..6] of Integer = (8, 8, 8, 4, 4, 2, 2);
    INTERLACE_COLUMN_INCREMENT: array[0..6] of Integer = (8, 8, 4, 4, 2, 2, 1);

type
  TIHDRChunk = packed record
    Width: cardinal;
    Height: cardinal;
    BitDepth: Byte;
    ColorType: Byte;
    CompressionMethod: Byte;
    FilterMethod: Byte;
    InterlaceMethod: Byte;
  end;
  PIHDRChunk = ^TIHDRChunk;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------
  
function TPNGProcessor.SwapLong( const aValue: Cardinal ): Cardinal; //assembler;
begin
  Result := BeToN(aValue);
{asm
    mov eax, aValue
    bswap eax;  }
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TPNGLoader.Create(aStream: TStream);
begin
    inherited Create;

    fStream := aStream;
end;

// ----------------------------------------------------------------------------

function TPNGLoader.LoadHeader: TTextureData;
var headerValue: TPNGHeaderMagic;
    chunkHeader: TChunkHeader;
    headerChunk: TIHDRChunk;
    isUnsupportedCombination: boolean;
begin
    fStream.Read(headerValue, SizeOf(headerValue));
    if headerValue <> PNG_HEADER_MAGIC then
        raise Exception.Create( 'Invalid PNG header.' );
    fStream.Read(chunkHeader, SizeOf(chunkHeader));
    if chunkHeader.ChunkID <> CHUNK_ID_IHDR then
        raise Exception.Create( Format( 'IDHR chunk missing (found "%s" instead).', [chunkHeader.ChunkID] ));
    fStream.Read(headerChunk, SizeOf(headerChunk));

    result.Width := SwapLong(headerChunk.Width);
    result.Height := SwapLong(headerChunk.Height);

    isUnsupportedCombination := FALSE;
    if headerChunk.ColorType = COLOR_TYPE_GREYSCALE then
    begin
        if headerChunk.BitDepth <> 8 then
            isUnsupportedCombination := TRUE;

        result.BitsPerPixel := 24;
        fGreyScale := TRUE;
    end
    else if headerChunk.ColorType = COLOR_TYPE_TRUECOLOR then begin
        if (headerChunk.BitDepth <> 8) and (headerChunk.BitDepth <> 16) then
            isUnsupportedCombination := TRUE;

        result.BitsPerPixel := 24;
    end
    else if headerChunk.ColorType = COLOR_TYPE_INDEXED then begin
        if headerChunk.BitDepth <> 8 then
            isUnsupportedCombination := TRUE;

        fIndexed := TRUE;
        result.BitsPerPixel := 24;
    end
    else if headerChunk.ColorType = COLOR_TYPE_GREYSCALE_ALPHA then begin
        if headerChunk.BitDepth <> 8 then
            isUnsupportedCombination := TRUE;

        result.BitsPerPixel := 32;
        fGreyScale := TRUE;
    end
    else if headerChunk.ColorType = COLOR_TYPE_TRUECOLOR_ALPHA then begin
        if (headerChunk.BitDepth <> 8) and (headerChunk.BitDepth <> 16) then
            isUnsupportedCombination := TRUE;

        result.BitsPerPixel := 32;
    end
    else
        raise Exception.Create( Format( 'Unsupported color type (%d) encountered.', [headerChunk.ColorType] ));

    f16BitChannels := ( headerChunk.BitDepth = 16 );

    if isUnsupportedCombination then
        raise Exception.Create( Format( 'Unsupported combination of color type (%d) and bit depth (%d)',
            [headerChunk.ColorType, headerChunk.BitDepth] ));

    if headerChunk.CompressionMethod <> 0 then
        raise Exception.Create( Format( 'Unsupported compression method (%d) encountered.', [headerChunk.CompressionMethod] ));

    if headerChunk.FilterMethod <> 0 then
        raise Exception.Create( Format( 'Unsupported filter method (%d) encountered.', [headerChunk.FilterMethod] ));

    fInterlaced := (headerChunk.InterlaceMethod = 1);

    if headerChunk.InterlaceMethod > 1 then
        raise Exception.Create( Format( 'Unsupported interlace method (%d) encountered.', [headerChunk.InterlaceMethod] ));

    ProcessCRC;
end;

function TPNGLoader.Load: TTextureData;
var chunkHeader: TChunkHeader;
    buffer, decodedBuffer: TBuffer;
    size, rawSize: integer;
begin
    result := LoadHeader;

    repeat
        chunkHeader := ReadChunkHeader;

        if chunkHeader.ChunkID = CHUNK_ID_PLTE then begin
            ReadPaletteChunk( chunkHeader );
        end
        else if chunkHeader.ChunkID = CHUNK_ID_TRNS then begin
            ReadTransparencyChunk( chunkHeader );
        end
        else if chunkHeader.ChunkID <> CHUNK_ID_IDAT then begin
            fStream.Seek(chunkHeader.Length + 4, soCurrent)
        end;
    until (chunkHeader.ChunkID = CHUNK_ID_IDAT) or (fStream.Position > fStream.Size);

    if chunkHeader.ChunkID <> CHUNK_ID_IDAT then
        raise Exception.Create( 'No image data chunk found.' );

    // INFO, PJ: In case of an indexed 8-bits with a tRNS chunk, BitsPerPixel equals 32 after all
    if ContainsAlphaPalette then
        result.BitsPerPixel := 32;

    size := result.Width * result.Height * ( result.BitsPerPixel div 8 );
    SetLength(result.Pixels, size);

    // rawSize calculation ... every row contains 1 extra byte
    rawSize := result.Width;
    if fIndexed then
        rawSize := rawSize * 1
    else if fGreyScale then
        rawSize := rawSize * ( result.BitsPerPixel div 8 - 2 )
    else
        rawSize := rawSize * ( result.BitsPerPixel div 8 );

    if f16BitChannels then
    begin
        rawSize := rawSize * 2;
    end;
    rawSize := ( rawSize + 1 ) * result.height;

    fStream.Position := fStream.Position - SizeOf(TChunkHeader);
    buffer := ReadImageData;
    try
        decodedBuffer := UnpackData( buffer, rawSize );
        try
            ConvertImageData( decodedBuffer, result );
        finally
            FreeMem( decodedBuffer.Data );
        end;
    finally
        FreeMem( buffer.Data );
    end;

end;

// INFO, PJ: We are not really interested in CRC checks, so we simply skip it.
procedure TPNGLoader.ProcessCRC;
var crc: Cardinal;
begin
    fStream.Read(crc, SizeOf(crc));
end;

function TPNGLoader.ContainsAlphaPalette: boolean;
begin
    result := Length( fAlphaPalette ) <> 0 ;
end;

function TPNGLoader.ReadImageData: TBuffer;
var size: cardinal;
    position: int64;
    chunkHeader: TChunkHeader;
    nrChunks, i: integer;
    pBuffer: ^Byte;
begin
    // INFO, PJ: First we figure out how many IDAT blocks there are
    //           thus calculating total size.
    position := fStream.Position;
    size := 0;
    nrChunks := 0;
    repeat
        chunkHeader := ReadChunkHeader;

        if chunkHeader.ChunkID = CHUNK_ID_IDAT then
        begin
            nrChunks := nrChunks + 1;
            size := size + chunkHeader.Length;

            fStream.Seek(chunkHeader.Length + 4, soCurrent);
        end;
    until chunkHeader.ChunkID <> CHUNK_ID_IDAT;

    result.Size := size;
    GetMem(result.Data, size);
    pBuffer := result.Data;
    fStream.Position := position;

    // INFO, PJ: And then we read the actual data from the IDAT blocks
    for i := 0 to nrChunks -  1 do
    begin
        chunkHeader := ReadChunkHeader;
        fStream.Read(pBuffer^, chunkHeader.Length);

        ProcessCRC;

        inc(pBuffer, chunkHeader.Length);
    end;
end;

procedure TPNGLoader.ReadPaletteChunk( const aChunkHeader: TChunkHeader );
var i, numEntries: integer;
    tempByte: Byte;
begin
    if (aChunkHeader.Length mod 3) <> 0 then
        raise Exception.Create( 'Invalid chunk header encountered.' );

    numEntries := aChunkHeader.Length div 3;

    SetLength( fPalette, numEntries );
    for i := 0 to numEntries - 1 do begin
        fStream.Read( fPalette[i], 3 );

        tempByte := fPalette[i].R;
        fPalette[i].R := fPalette[i].B;
        fPalette[i].B := tempByte;
    end;

    ProcessCRC;
end;

procedure TPNGLoader.ReadTransparencyChunk( const aChunkHeader: TChunkHeader );
var i, numEntries: integer;
begin
    numEntries := aChunkHeader.Length;

    SetLength( fAlphaPalette, numEntries );
    for i := 0 to numEntries - 1 do begin
        fStream.Read( fAlphaPalette[i], 1 );
    end;

    ProcessCRC;
end;


function TPNGLoader.ReadChunkHeader: TChunkHeader;
begin
    fStream.Read(result, SizeOf(result));

    result.Length := SwapLong( result.Length );
end;

function TPNGLoader.UnpackData(const aBuffer: TBuffer; const aTargetSize: integer): TBuffer;
var source, target: Pointer;
begin
    result.Size := aTargetSize;
    result.Data := nil;
    GetMem( result.Data, aTargetSize );

    source := aBuffer.Data;
    target := result.Data;

    Decode( source, target, aBuffer.Size, result.Size );
end;

procedure TPNGLoader.Decode(const aSource, aDestination: Pointer; const aPackedSize, anUnpackedSize: Integer);
var state: TZState;
begin
    if InflateInit(state) < 0 then
        raise Exception.Create( 'Failed to initialize LZ77 decompression.' );

    state.NextInput := aSource;
    state.AvailableInput := aPackedSize;
    state.NextOutput := aDestination;
    state.AvailableOutput := anUnpackedSize;
    Inflate(state, Z_PARTIAL_FLUSH);

    if InflateEnd(state) < 0 then
        raise Exception.Create( 'Failed to deinitialize LZ77 decompression.' );
end;


procedure TPNGLoader.ConvertImageData(const aBuffer: TBuffer; const aTextureData: TTextureData);
var i, width, y, passIndex, rowSize, columnSize, columnStart: integer;
    numBytesPerPixel, numBytesPerRow, numBytesPerTargetPixel, numBytesPerTargetRow: integer;
    targetData, rowData, lastRowData, filteredRowData: PByte;
    filter: byte;
begin
    if fIndexed then
        numBytesPerPixel := 1
    else if fGreyScale then
        numBytesPerPixel := ( aTextureData.BitsPerPixel div 8 ) - 2
    else
        numBytesPerPixel := ( aTextureData.BitsPerPixel div 8 );

    if f16BitChannels then
        numBytesPerPixel := numBytesPerPixel * 2;

    if fIndexed then begin
        if ContainsAlphaPalette then
            numBytesPerTargetPixel := 4
        else
            numBytesPerTargetPixel := 3;
    end
    else
        numBytesPerTargetPixel := aTextureData.BitsPerPixel div 8;

    numBytesPerRow := aTextureData.Width * numBytesPerPixel;
    numBytesPerTargetRow := aTextureData.Width * numBytesPerTargetPixel;

    rowData := aBuffer.Data;
    targetData := @aTextureData.Pixels[0];

    GetMem( filteredRowData, numBytesPerRow );
    GetMem( lastRowData, numBytesPerRow );

    filter := rowData^;
    inc( rowData );

    try
        if fInterlaced then begin
            for passIndex := 0 to 6 do begin
                if aTextureData.Width < INTERLACE_COLUMN_START[passIndex] then continue;

                width := ( aTextureData.Width + INTERLACE_COLUMN_INCREMENT[passIndex] - 1 - INTERLACE_COLUMN_START[passIndex] ) div INTERLACE_COLUMN_INCREMENT[passIndex];
                rowSize := numBytesPerPixel * width;
                columnSize := INTERLACE_COLUMN_INCREMENT[passIndex] * numBytesPerTargetPixel;
                columnStart := INTERLACE_COLUMN_START[passIndex] * numBytesPerTargetPixel;

                y := INTERLACE_ROW_START[passIndex];
                while y < aTextureData.Height do begin
                    ApplyFilter( filter, rowData, pointer( lastRowData ), filteredRowData, numBytesPerPixel, rowSize );

                    Move( filteredRowData^, lastRowData^, rowSize );

                    ConvertColorData( filteredRowData, targetData, width, numBytesPerPixel, y * numBytesPerTargetRow + columnStart, columnSize );

                    Inc( rowData, rowSize );
                    filter := rowData^;
                    Inc( rowData );

                    y := y + INTERLACE_ROW_INCREMENT[passIndex];
                end;
            end;
        end
        else begin // Non-interlaced.
            // INFO, PJ: First byte of each row contains filter, and bytes need r, g, b swap.
            for i := 0 to aTextureData.Height - 1  do begin
                ApplyFilter( filter, rowData, pointer( lastRowData ), filteredRowData, numBytesPerPixel, numBytesPerRow );

                ConvertColorData( filteredRowData, targetData, aTextureData.Width, numBytesPerPixel, i * numBytesPerTargetRow, numBytesPerTargetPixel );

                // INFO, PJ: This could also be done using a simple swap
                Move( filteredRowData^, lastRowData^, numBytesPerRow );

                Inc( rowData, numBytesPerRow );
                filter := rowData^;
                Inc( rowData );
            end;
        end;

    finally
        FreeMem( filteredRowData );
        FreeMem( lastRowData );
    end;
end;

// ----------------------------------------------------------------------------

procedure TPNGLoader.ConvertColorData( const aSource, aTarget: PByte; const aWidth, aNumBytesPerPixel, anOffset, aStride: integer );
var
  source, target: PByte;
  source16: PWord;
  i: integer;
begin
  source := aSource;
  source16 := Pointer(aSource);
  target := aTarget;
  inc( target, anOffset );

  if fIndexed then begin
      if ContainsAlphaPalette then begin
          for i := 0 to aWidth - 1 do begin
              Move( fPalette[source^], target^, 3 );
              inc(target, 3);
              target^ := fAlphaPalette[source^];
              inc(target);
              inc(source);
              inc(target, aStride - 4);
          end;
      end
      else begin
          for i := 0 to aWidth - 1 do begin
              Move( fPalette[source^], target^, 3 );
              inc(target, 3);
              inc(source);
              inc(target, aStride - 3);
          end;
      end;
  end
  else if fGreyScale then begin
      for i := 0 to aWidth - 1 do begin
          target^ := source^;
          inc(target);
          target^ := source^;
          inc(target);
          target^ := source^;
          inc(target);
          inc(source);

          if aNumBytesPerPixel > 1 then begin
              target^ := source^;
              inc(source);
          end;

          inc(target, aStride - 3);
      end;
  end
  else if f16BitChannels then begin
      for i := 0 to aWidth - 1 do begin
          if source16^ <> 65535 then
          begin
             source16^:= source16^;
          end;

          PByteArray(target)[2] := source16^ and $FF;
          inc( source16 );

          if aNumBytesPerPixel > 1 then begin
              PByteArray(target)[1] := source16^ and $FF;
              inc( source16 );

              PByteArray(target)[0] := source16^ and $FF;
              inc( source16 );
          end;

          if aNumBytesPerPixel > 3 then begin
              PByteArray(target)[3] := source16^ and $FF;
              inc( source16 );
          end;

          inc( target, aStride );
      end;
  end
  else begin
      for i := 0 to aWidth - 1 do begin
          PByteArray(target)[2] := source^;
          inc( source );

          if aNumBytesPerPixel > 1 then begin
              PByteArray(target)[1] := source^;
              inc( source );

              PByteArray(target)[0] := source^;
              inc( source );
          end;

          if aNumBytesPerPixel > 3 then begin
              PByteArray(target)[3] := source^;
              inc( source );
          end;

          inc( target, aStride );
      end;
  end;
end;

// ----------------------------------------------------------------------------

function PaethPredictor(a, b, c: Byte): Byte;
var
    p, pa, pb, pc: Integer;

begin
    // a = left, b = above, c = upper left
    p := a + b - c;        // initial estimate
    pa := Abs(p - a);      // distances to a, b, c
    pb := Abs(p - b);
    pc := Abs(p - c);
    // return nearest of a, b, c, breaking ties in order a, b, c
    if (pa <= pb) and (pa <= pc) then
        Result := a
    else if pb <= pc then
        Result := b
    else
        Result := c;
end;

procedure TPNGLoader.ApplyFilter(aFilter: Byte; aLineBuffer, aPreviousLineBuffer, aTargetBuffer: PByte; aBytesPerPixel, aBytesPerRow: Integer);
var i: integer;
    raw, decoded, prior, priorDecoded, target: PByte;

begin
    case aFilter of
        0: // no filter, just copy data
            Move(aLineBuffer^, aTargetBuffer^, aBytesPerRow);

        1: // subtraction filter
        begin
            raw := aLineBuffer;
            target := aTargetBuffer;
            // Transfer BPP bytes without filtering. This mimics the effect of bytes left to the
            // scanline being zero.
            Move(raw^, target^, aBytesPerPixel);

            // now do rest of the line
            decoded := target;
            Inc(raw, aBytesPerPixel);
            Inc(target, aBytesPerPixel);
            Dec(aBytesPerRow, aBytesPerPixel);
            while aBytesPerRow > 0 do begin
              target^ := Byte(raw^ + decoded^);
              Inc(raw);
              Inc(decoded);
              Inc(target);
              Dec(aBytesPerRow);
            end;
        end;

        2: // Up filter
        begin
            raw := aLineBuffer;
            prior := aPreviousLineBuffer;
            target := aTargetBuffer;
            while aBytesPerRow > 0 do begin
                target^ := Byte(raw^ + prior^);
                Inc(raw);
                Inc(prior);
                Inc(target);
                Dec(aBytesPerRow);
            end;
        end;

        3: // average filter
        begin
            // first handle BPP virtual pixels to the left
            raw := aLineBuffer;
            decoded := aTargetBuffer;
            prior := aPreviousLineBuffer;
            target := aTargetBuffer;
            for i := 0 to aBytesPerPixel - 1 do begin
                target^ := Byte(raw^ + Floor(prior^ / 2));
                Inc(raw);
                Inc(prior);
                Inc(target);
            end;
            Dec(aBytesPerRow, aBytesPerPixel);

            // now do rest of line
            while aBytesPerRow > 0 do begin
                target^ := Byte(raw^ + Floor((decoded^ + prior^) / 2));
                Inc(raw);
                Inc(decoded);
                Inc(prior);
                Inc(target);
                Dec(aBytesPerRow);
            end;
        end;

        4: // paeth prediction
        begin
            // again, start with first BPP pixel which would refer to non-existing pixels to the left
            raw := aLineBuffer;
            decoded := aTargetBuffer;
            prior := aPreviousLineBuffer;
            priorDecoded := aPreviousLineBuffer;
            target := aTargetBuffer;
            for i := 0 to aBytesPerPixel - 1 do begin
                target^ := Byte(raw^ + PaethPredictor(0, prior^, 0));
                Inc(raw);
                Inc(prior);
                Inc(target);
            end;
            Dec(aBytesPerRow, aBytesPerPixel);

            // finally do rest of line
            while aBytesPerRow > 0 do begin
                target^ := Byte(raw^ + PaethPredictor(decoded^, prior^, priorDecoded^));
                Inc(raw);
                Inc(decoded);
                Inc(prior);
                Inc(priorDecoded);
                Inc(target);
                Dec(aBytesPerRow);
            end;
        end;
    else
        Move(aLineBuffer^, aTargetBuffer^, aBytesPerRow);
    end;
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TPNGWriter.Create(aStream: TStream);
begin
    inherited Create;

    fStream := aStream;
end;

// ----------------------------------------------------------------------------

procedure TPNGWriter.WriteChunk(const aChunk; const aChunkID: TChunkID; const aSize: integer);
var size: integer;
    crc: Cardinal;
begin
    size := SwapLong(aSize);
    fStream.Write( size, SizeOf( size ));
    fStream.Write( aChunkID, SizeOf( aChunkID ));
    fStream.Write( aChunk, aSize );

    crc := CRC32( 0, @aChunkID, SizeOf( aChunkID ));
    crc := CRC32( crc, @aChunk, aSize );
    crc := SwapLong(crc);
    fStream.Write( crc, SizeOf( crc ));
end;

// ----------------------------------------------------------------------------

procedure TPNGWriter.WriteHeader(const aTextureData: TTextureData);
var headerChunk: TIHDRChunk;
begin
    fStream.Write(PNG_HEADER_MAGIC, SizeOf(PNG_HEADER_MAGIC));

    headerChunk.Width := SwapLong(aTextureData.Width);
    headerChunk.Height := SwapLong(aTextureData.Height);
    headerChunk.BitDepth := 8;
    if aTextureData.BitsPerPixel = 24 then
        headerChunk.ColorType := COLOR_TYPE_TRUECOLOR
    else
        headerChunk.ColorType := COLOR_TYPE_TRUECOLOR_ALPHA;
    headerChunk.CompressionMethod := 0;
    headerChunk.FilterMethod := 0;
    headerChunk.InterlaceMethod := 0;

    WriteChunk(headerChunk, CHUNK_ID_IHDR, SizeOf(headerChunk));
end;

// ----------------------------------------------------------------------------

procedure TPNGWriter.WriteImageData(const aTextureData: TTextureData);
var imageBuffer, packedBuffer: TBuffer;
    maxSize: integer;
begin
    imageBuffer := ConvertImageData(aTextureData);
    try
        // INFO, PJ: buffer size is size + 10% + 12 bytes, rounded to next multiple of 256
        maxSize := (imageBuffer.Size + 12 + (imageBuffer.Size div 10) + 255) and not 255;
        packedBuffer := PackData(imageBuffer, maxSize);
        try
            WriteChunk(packedBuffer.Data^, CHUNK_ID_IDAT, packedBuffer.Size);
        finally
            FreeMem(packedBuffer.Data);
        end;
    finally
        FreeMem(imageBuffer.Data);
    end;
end;

// ----------------------------------------------------------------------------

procedure TPNGWriter.Write(const aTextureData: TTextureData);
var endChunk: Pointer;
begin
    WriteHeader(aTextureData);
    WriteImageData(aTextureData);

    endChunk := nil;
    WriteChunk(endChunk, CHUNK_ID_IEND, 0);
end;

// ----------------------------------------------------------------------------

function TPNGWriter.PackData(const aBuffer: TBuffer; const aTargetSize: integer): TBuffer;
var source, target: Pointer;
begin
    result.Size := aTargetSize;
    result.Data := nil;
    GetMem( result.Data, aTargetSize );

    source := aBuffer.Data;
    target := result.Data;

    Encode( source, target, aBuffer.Size, result.Size );
end;

// ----------------------------------------------------------------------------

procedure TPNGWriter.Encode(const aSource, aDestination: Pointer; const anUnpackedSize: Integer; var aPackedSize: Integer);
var state: TZState;
begin
    if DeflateInit(state, Z_DEFAULT_COMPRESSION) < 0 then
        raise Exception.Create( 'Failed to initialize LZ77 decompression.' );

    state.NextInput := aSource;
    state.AvailableInput := anUnpackedSize;
    state.NextOutput := aDestination;
    state.AvailableOutput := aPackedSize;
    Deflate(state, Z_FINISH);

    if DeflateEnd(state) < 0 then
        raise Exception.Create( 'Failed to deinitialize LZ77 compression.' );

    aPackedSize := state.TotalOutput;
end;

// ----------------------------------------------------------------------------

// Converts raw texture data in filtered scanlines prefixed with a filter type
function TPNGWriter.ConvertImageData(const aTextureData: TTextureData): TBuffer;
var bytesPerPixel, sourceLineSize, targetLineSize, y: integer;
    source, target: PByte;
begin
    bytesPerPixel := aTextureData.BitsPerPixel div 8;
    sourceLineSize := bytesPerPixel * aTextureData.Width;
    targetLineSize := sourceLineSize + 1;
    result.Size := aTextureData.Height * targetLineSize;
    GetMem(result.Data, result.Size);

    source := @aTextureData.Pixels[0];
    target := result.Data;
    for y := 0 to aTextureData.Height - 1 do
    begin
        target^ := 0;
        Inc(target);

        // TODO, PJ: Implement actual (and smart) filtering
        Move(source^, target^, sourceLineSize);

        Inc(source, sourceLineSize);
        Inc(target, sourceLineSize);
    end; 
end;

// ----------------------------------------------------------------------------

end.
