{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PathCreationRoutines;

{$MODE Delphi}

interface

uses
  Classes,
  Controls,
  frmViewPortForm,
  unit_Paths,
  unit_Vectors,
  unit_Types,
  unit_ParamList,
  unit_BaseTool;

// routines to facilitate the merging of top, side, front viewport code into one place
procedure HandlePathMouseDownCreation(aViewPort: TfrmViewPort;
                                      aVertice: TVertice;
                                      aParamList: TParamList;
                                      aBaseTool: TBaseTool;
                                      aShift: TShiftState;
                                      aToggleSnap: Boolean; aGridSize: Integer;
                                      var aPath: TPath;
                                      var aControlPoint: TPathControlPoint);
                                      
procedure HandlePathMouseMoveCreation(aViewPort: TfrmViewPort;
                                      aShift: TShiftState; X, Y: Integer;
                                      aToggleSnap: Boolean; aGridSize: Integer;
                                      aDragOrigin: TVector3f;
                                      aPath: TPath;
                                      var aControlPoint: TPathControlPoint;
                                      var aIsDraggingTangent: Boolean);
                                      
procedure HandlePathMouseUpCreation(aViewPort: TfrmViewPort;
                                    aButton: TMouseButton; aShift: TShiftState; X, Y: Integer;
                                    aToggleSnap: Boolean; aGridSize: Integer;
                                    var aPath: TPath;
                                    var aControlPoint: TPathControlPoint;
                                    var aIsDraggingTangent: Boolean);

implementation

uses
  frmTopView,
  frmSideView,
  frmFrontView,
  unit_Variables,
  unit_Enums;

function GetViewPortPosition(aViewPort: TfrmViewPort; x,y: Integer): TVector3f;
var
  TopView: TfrmTop;
  SideView: TfrmSide;
  FrontView: TfrmFront;
begin
  Result := Vector3f(0,0,0);
  if aViewPort is TfrmTop then begin
    TopView := TfrmTop(aViewPort);
    Result := Vector3f(TopView.Base.x+(x*TopView.uppx),
                       0,
                       TopView.Base.z-TopView.upsy+(y*TopView.uppy));
  end
  else if aViewPort is TfrmSide then begin
    SideView := TfrmSide(aViewPort);
    Result := Vector3f(0,
                       -SideView.Base.y+SideView.upsy-(Y*SideView.uppy),
                       SideView.Base.z+(X*SideView.uppx));
  end
  else if aViewPort is TfrmFront then begin
    FrontView := TfrmFront(aViewPort);
    Result := Vector3f(FrontView.Base.x+(X*FrontView.uppx),
                       FrontView.Base.y+FrontView.upsy-(Y*FrontView.uppy),
                       0);
  end;
end;

procedure SetPathControlPointUpVector(const aViewMode: TViewMode;
                                      const aControlPoint: TPathControlPoint);
begin
  if      aViewMode = vmFront then aControlPoint.UpVector := Vector3f(0,0,1)
  else if aViewMode = vmSide  then aControlPoint.UpVector := Vector3f(1,0,0);
end;

procedure HandlePathMouseDownCreation(aViewPort: TfrmViewPort;
                                      aVertice: TVertice;
                                      aParamList: TParamList;
                                      aBaseTool: TBaseTool;
                                      aShift: TShiftState;
                                      aToggleSnap: Boolean; aGridSize: Integer;
                                      var aPath: TPath;
                                      var aControlPoint: TPathControlPoint);
begin
  if not Assigned(aPath) then begin
      aParamList.Add(aVertice,'startVertex',false,false);
      aParamList.Add(Ord(defaultPathType),'PathType',false);
      aParamList.Add(defaultPathIsClosed,'PathIsClosed',false);
      aBaseTool := aViewPort.Scene.ToolFactory.CreateTool(ptPrimitivePathCreator);
      aViewPort.Scene.ToolExecuter.Execute(aBaseTool,aParamList);
      aPath := TPath(aViewPort.Scene.Primitives.Last);
      // store last duplicate control point for moving around with mouse
      aControlPoint := aPath.GetControlPoint(aPath.GetControlPointCount - 1);

      // setup correct up-vectors depending on the view
      SetPathControlPointUpVector(aViewPort.ViewMode,aPath.ControlPoints[0]);
      SetPathControlPointUpVector(aViewPort.ViewMode,aControlPoint);
  end else begin
      if (ssLeft in aShift) then begin
          aControlPoint := aPath.AddControlPoint(Vector3f(aVertice.x,aVertice.y,aVertice.z),aToggleSnap,aGridSize);

          // setup correct up-vectors depending on the view
          SetPathControlPointUpVector(aViewPort.ViewMode,aControlPoint);
      end;
  end;
  aViewPort.ViewportState := vsPathCreation;
end;

procedure HandlePathMouseMoveCreation(aViewPort: TfrmViewPort;
                                      aShift: TShiftState; X, Y: Integer;
                                      aToggleSnap: Boolean; aGridSize: Integer;
                                      aDragOrigin: TVector3f;
                                      aPath: TPath;
                                      var aControlPoint: TPathControlPoint;
                                      var aIsDraggingTangent: Boolean);
var
  position: TVector3f;
  tangent: TVector3f;
begin
  if Assigned(aPath) and Assigned(aControlPoint) then
  begin
    // left drag moves tangents for control point
    // left drag + Alt moves 1 tangent for control point
    // no drag just moves control point

    //position calculated from view port
    position := GetViewPortPosition(aViewPort,X,Y);

    // tangent vector calculated from control point to position
    tangent  := VecNegate(Vector3f(aDragOrigin.x - position.x,aDragOrigin.y - position.y,aDragOrigin.z - position.z));

    //Are we dragging a tangent (left or right mousebutton)?
    if (aPath.PathType in [ptCubicBezier,ptHermite]) and
       ((ssLeft in aShift) or (ssRight in aShift)) and
       (VecMagnitude(tangent) > 5) then
    begin
      // must be moving a tangent instead of the control point itself
      if not aIsDraggingTangent then
      begin
        aIsDraggingTangent := true;

        // When dragging a tangent we temporarily select our previous controlpoint
        // That's the one we just positioned.
        // We DON'T do this when the right mouse is used to finish the path
        // In that case, no new CP has been made, so no reason to jump back
        if (ssLeft in aShift) then
          aControlPoint := aControlPoint.Previous;
      end;

      if ssAlt in aShift then begin
        // when Alt is held during tangent drag, only one tangent is moved
        // independently of the other - handy for sharp corners
        if aPath.PathType = ptCubicBezier then begin
          // move bezier tangent
          aControlPoint.TangentVector1 := tangent;
          aControlPoint.UserDefinedTangent1 := true;
          aControlPoint.UserDefinedTangent2 := false;
        end else begin
          // hermite splines only use tangent vector 2
          aControlPoint.TangentVector2 := tangent;
          aControlPoint.UserDefinedTangent2 := true;
          aControlPoint.UserDefinedTangent1 := false;
        end;
      end else begin
        // no alt key, so move both tangents
        aControlPoint.TangentVector1 := tangent;
        aControlPoint.TangentVector2 := VecNegate(tangent);
        aControlPoint.UserDefinedTangent1 := true;
        aControlPoint.UserDefinedTangent2 := true;
      end;
    end else begin
      // move the last added vertex around (thus moving the last edge of the path)
      aControlPoint.Position := position;
    end;
    aPath.CalculateTangentVectors;
    aViewPort.glDraw;
  end;
end;

procedure HandlePathMouseUpCreation(aViewPort: TfrmViewPort;
                                    aButton: TMouseButton; aShift: TShiftState; X, Y: Integer;
                                    aToggleSnap: Boolean; aGridSize: Integer;
                                    var aPath: TPath;
                                    var aControlPoint: TPathControlPoint;
                                    var aIsDraggingTangent: Boolean);
begin
  if Assigned(aPath) and (aButton = mbLeft) and aIsDraggingTangent then
  begin
    //We are done dragging the tangent of our controlpoint
    //We will move on with positioning the next controlpoint that
    //was created in the mouseDown event
    aControlPoint      := aControlPoint.Next;
    aIsDraggingTangent := false;
  end;

  if aButton = mbRight then
  begin
      //This code will quit "Path creation" state
      //When we position our last controlpoint with a right-click
      if aPath.ControlPointCount >= 2 then
      begin
          if aToggleSnap then
            aControlPoint.Position := VecSnap(aControlPoint.Position,aGridSize);

          aPath.CalculateTangentVectors;
          aPath.HasChanged := True;
          aPath := nil;
          aViewPort.ViewportState := vsNone;
      end;
  end;
end;

end.
