{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PathRenderer;

{$MODE Delphi}

interface

uses
  unit_Paths;

type
  TRenderView = (
    rvXY, //front view
    rvXZ, //top view
    rvZY, //side view
    rv3d
  );

procedure RenderPath(aPath: TPath; aNumberOfSegments: Integer; aRenderView: TRenderView);
procedure RenderPathRecursive(aPath: TPath; aMinimumSegmentLength: Integer; aRenderView: TRenderView);
procedure RenderPathControlPoints(aPath: TPath; aRenderView: TRenderView);
procedure RenderPathTangents(aPath: TPath; aRenderView: TRenderView);
procedure RenderPathUpVectors(aPath: TPath; aNormalLength: Single; aRenderView: TRenderView);

implementation

uses
  dglOpenGL,
  unit_Vectors,
  unit_Variables;

procedure DrawViewDependentVertex(aVertex: TVector3f; aRenderView: TRenderView);
begin
  case aRenderView of
    rvXY: glVertex2f(aVertex.x,aVertex.y);
    rvXZ: glVertex2f(aVertex.x,aVertex.z);
    rvZY: glVertex2f(aVertex.z,aVertex.y);
    rv3d: glVertex3f(aVertex.x,aVertex.y,aVertex.z);
  end;
end;

procedure RenderPathTangents(aPath: TPath; aRenderView: TRenderView);
var
  oldPointSize: Single;
  i: Integer;
  cp: TPathControlPoint;
  t1,t2: TVector3f;
begin
  if aPath.PathType in[ptLinear,ptCatmullRom] then Exit;

  //Retrieve old point size, and overwrite with new one
  glGetFloatv(GL_POINT_SIZE, @oldPointSize);
  glPointSize(defaultPointSize);

  for i := 0 to aPath.ControlPointCount - 1 do begin
    cp := aPath.ControlPoints[i];
    // calculate absolute tangent positions relative to the control point location
    t1 := VecAdd(cp.Position,cp.TangentVector1);
    t2 := VecAdd(cp.Position,cp.TangentVector2);
    // draw line strip
    // t1----cp----t2
    glBegin(GL_LINE_STRIP);
      if aPath.PathType = ptCubicBezier then
        // cubic bezier uses both t1 and t2, hermite uses t2 only
        DrawViewDependentVertex(t1,aRenderView);
      DrawViewDependentVertex(cp.Position,aRenderView);
      DrawViewDependentVertex(t2,aRenderView);
    glEnd;

    // draw tangent points
      glBegin(GL_POINTS);
        if aPath.PathType = ptCubicBezier then
          // cubic bezier uses both t1 and t2, hermite uses t2 only
          DrawViewDependentVertex(t1,aRenderView);
        DrawViewDependentVertex(t2,aRenderView);
      glEnd;
  end;

  //Reset old point size
  glPointSize(oldPointSize);
end;

procedure RenderPathUpVectors(aPath: TPath; aNormalLength: Single; aRenderView: TRenderView);
var
  i: Integer;
  cp: TPathControlPoint;
  v: TVector3f;
begin
  glPushAttrib(GL_ALL_ATTRIB_BITS);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    glColor3f(1,0.5,0.5);               // y=green
//    glColor3f(0.5,1,0.5);               // y=green
    for i := 0 to aPath.ControlPointCount - 1 do begin
      cp := aPath.ControlPoints[i];
      // calculate absolute up-vector position relative to the control point location
      v := VecAdd(cp.Position,VecScale(cp.UpVector,aNormalLength));
      // draw up-vector line
      glBegin(GL_LINES);
        DrawViewDependentVertex(cp.Position,aRenderView);
        DrawViewDependentVertex(v,aRenderView);
      glEnd;
    end;
  glPopAttrib;
end;

procedure RenderPathControlPoints(aPath: TPath; aRenderView: TRenderView);
var
  i: Integer;
  cp: TPathControlPoint;
  oldPointSize: Single;
begin
  //Retrieve old point size, and overwrite with new one
  glGetFloatv(GL_POINT_SIZE, @oldPointSize);
  glPointSize(defaultPointSize);

  glBegin(GL_POINTS);
    for i := 0 to aPath.ControlPointCount - 1 do begin
      cp := aPath.ControlPoints[i];
      DrawViewDependentVertex(cp.Position,aRenderView);
    end;
  glEnd;

  //Reset old point size
  glPointSize(oldPointSize);
end;

procedure RenderPath(aPath: TPath; aNumberOfSegments: Integer; aRenderView: TRenderView);
var
  c: Single;
  t: Single;
  vertex: TVector3f;
begin
  if aNumberOfSegments <= 0 then Exit;

  glPushAttrib(GL_ALL_ATTRIB_BITS);
    // render path
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    t := 0.0;
    c := 1.0 / aNumberOfSegments;
    glBegin(GL_LINE_STRIP);
      while t <= 1.0 do begin
        vertex := aPath.GetPosition(t);
        DrawViewDependentVertex(vertex,aRenderView);
        t := t + c;
      end;
      // draw last point on path
      vertex := aPath.GetPosition(1.0);
      DrawViewDependentVertex(vertex,aRenderView);
    glEnd;
  glPopAttrib;
end;

procedure RenderPathSegmentLength(aPath: TPath; t0,t1: Single; aMinimumSegmentLengthSqr: Integer; aRenderView: TRenderView);
var
  m0,m1: TVector3f;
  t,d: Single;
begin
  m0 := aPath.GetPosition(t0);
  m1 := aPath.GetPosition(t1);
  d := Sqr(m0.x - m1.x) + Sqr(m0.y - m1.y) + Sqr(m0.z - m1.z);
  if d > aMinimumSegmentLengthSqr then begin
    // segment length is larger than the minimum segment length so split again
    t := (t0 + t1) / 2;
    RenderPathSegmentLength(aPath,t0,t,aMinimumSegmentLengthSqr,aRenderView);
    RenderPathSegmentLength(aPath,t,t1,aMinimumSegmentLengthSqr,aRenderView);
  end else begin
    glBegin(GL_LINES);
      DrawViewDependentVertex(m0,aRenderView);
      DrawViewDependentVertex(m1,aRenderView);
    glEnd;
  end;
end;

procedure RenderPathRecursive(aPath: TPath; aMinimumSegmentLength: Integer; aRenderView: TRenderView);
begin
  if aMinimumSegmentLength <= 0 then Exit;

  glPushAttrib(GL_ALL_ATTRIB_BITS);
    // render path
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    RenderPathSegmentLength(aPath,0.0,0.5,aMinimumSegmentLength * aMinimumSegmentLength,aRenderView);
    RenderPathSegmentLength(aPath,0.5,1.0,aMinimumSegmentLength * aMinimumSegmentLength,aRenderView);
  glPopAttrib;
end;

end.
