{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Paths;

{$MODE Delphi}

interface

uses
  windows,
  classes,
  SysUtils,
  unit_Primitives,
  unit_Vectors,
  unit_Types;
  
type
  TPathType = (
    ptLinear,
    ptCatmullRom,
    ptCubicBezier,
    ptHermite
  );

  TPath = class;

  TPathControlPoint = class
  private
    FPath                : TPath;
    FPosition            : TVector3f;
    FUpVector            : TVector3f;
    FTangentVector1      : TVector3f;
    FTangentVector2      : TVector3f;
    FUserDefinedTangent1 : Boolean;
    FUserDefinedTangent2 : Boolean;
    FTag                 : String;
    procedure SetPosition(v: TVector3f);
    procedure SetUpVector(v: TVector3f);
    procedure SetTangentVector1(v: TVector3f);
    procedure SetTangentVector2(v: TVector3f);
    function GetPreviousControlPoint : TPathControlPoint;
    function GetNextControlPoint : TPathControlPoint;
    function GetID: Integer;
  public
    constructor Create(aPath: TPath);
    procedure CalculateTangentVectors;
    function  Clone: TPathControlPoint;
    procedure Assign(cp: TPathControlPoint);

    property ID: Integer read GetID;
    property Path: TPath read FPath;
    property Tag: String read FTag write FTag;
    property Position: TVector3f read FPosition write SetPosition;
    property UpVector: TVector3f read FUpVector write SetUpVector;
    property TangentVector1: TVector3f read FTangentVector1 write SetTangentVector1;
    property TangentVector2: TVector3f read FTangentVector2 write SetTangentVector2;
    property Previous: TPathControlPoint read GetPreviousControlPoint;
    property Next: TPathControlPoint read GetNextControlPoint;
    property UserDefinedTangent1 : Boolean read FUserDefinedTangent1 write FUserDefinedTangent1;
    property UserDefinedTangent2 : Boolean read FUserDefinedTangent2 write FUserDefinedTangent2;
  end;

  TPathInterpolationProcedure = procedure(CP1, CP2: TPathControlPoint; t: Single; var v: TVector3f);

  TPath = class(TBasePrimitive)
  private
    FPathType : TPathType;
    FPathInterpolationProcedure: TPathInterpolationProcedure;
    FIsClosed : Boolean;
    FTangentInfluence: Single;
    FControlPoints : TList;
    FSelectedControlPoints : TList;
    procedure SetPathType(const aValue: TPathType);
    procedure SetIsClosed(const aValue: Boolean);
    procedure SetTangentInfluence(const aValue: Single);
    function  CalculatePathSegmentLength(t0,t1: Single): Single;
    function  GetSegment(var t: single; out cp1,cp2: TPathControlPoint): Boolean;
    function DY(t: single; interval: single = 1e-3): TVector3f;
    function GetSpeedAt(t: single): single;
    function GetFirstSelectedControlPoint: TPathControlPoint;
  protected
    function GetCenter: TVertice;                 override;
  public
    constructor Create;  override;
    destructor  Destroy; override;

    procedure Clear;
    procedure Assign(Source: TBasePrimitive); override;
    procedure CalculateTangentVectors;
    // delete and free path control point
    procedure DeleteControlPoint(aPoint: TPathControlPoint); overload;
    procedure DeleteControlPoint(aIndex: Integer);           overload;
    procedure AddControlPoint(aPoint: TPathControlPoint); overload;
    function  AddControlPoint(Vertex: TVector3f; doSnap: boolean; gridSize: integer): TPathControlPoint; overload;
    function  GetPosition(t: single): TVector3f;
    function  GetDirection(t: single): TVector3f;
    function  GetUp(t: single): TVector3f;
    function  GetControlPointCount : Integer;
    function  GetControlPoint(aIndex: Integer): TPathControlPoint;
    function  GetControlPointIndex(aControlPoint: TPathControlPoint): Integer;
    function  CalculatePathLength: Single;
    function  GetCurveParameter(s: single): single;
    function GetControlPointsCenter: TVector3f;                 

    // Stuff Used by the primitive-selector
    //Derived from TBasePrimitive. Is used by the selector
    function Object_SelectByRange(v1,v2: TVertice): boolean; override;
    function IsPointOnPath(p: TVertice): boolean;
    function GetNextPathSegmentForSelection(var aIndex: Integer; out v1, v2: TVector3f): boolean;

    //Stuff used by the vertex-selector (a.k.a controlpoint selector)
    procedure SelectControlPoints(const v1, v2: TVertice; AClearSelection: Boolean);   overload;
    procedure SelectControlPoint(AVertex: TVertice; ARange: double; const AClearSelection: Boolean);
    procedure SelectControlPoints(const ax1,ay1,ax2,ay2: integer;
                                  const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                  const aViewport: TViewport;
                                  const AClearSelection: boolean); overload;

    property ControlPoints[I: Integer]: TPathControlPoint read GetControlPoint;
    property ControlPointCount: Integer read GetControlPointCount;
    property PathType: TPathType read FPathType write SetPathType;
    property IsClosed: Boolean   read FIsClosed write SetIsClosed;
    property TangentInfluence: Single read FTangentInfluence write SetTangentInfluence;
    property SelectedControlPoints: TList read FSelectedControlPoints;
    property FirstSelectedControlPoint: TPathControlPoint read GetFirstSelectedControlPoint;
  end;

  //Path-type conversion routines
  function PathTypeToString(aType: TPathType): string;
  function StringToPathType(aTypeStr: string): TPathType;

  //Used for click selection
  function IsPointOnSegment(point, startVec, endVec: TVector3f; Tolerance: Single): Boolean;

const
  //The tolerance value used to click-select paths
  //Added because the spline line is too thin to click
  PATH_SELECT_TOLERANCE = 10;

  //The number segments per controlpoint used for
  //the "testing spline" that is used when selecting splines
  //The segments will not be uniformly distributed according to the distance
  //between the controlpoints, because this is CPU-intensive
  PATH_SELECT_SEGMENTS_PER_CONTROLPOINT = 5;
implementation

uses         
  unit_Enums,
  unit_Interpolation;

//
// interpolation routines
//
procedure LinearInterpolationProcedure(CP1, CP2: TPathControlPoint; t: Single; var v: TVector3f);
var
  p0,p1: TVector3f;
begin
  p0 := CP1.Position;
  p1 := CP2.Position;

  LinearInterpolation(p0,p1,t,v);
end;

procedure CatmullRomInterpolationProcedure(CP1, CP2: TPathControlPoint; t: Single; var v: TVector3f);
var
  p0,p1,p2,p3: TVector3f;
  cp0,cp3: TPathControlPoint;
begin
  cp0 := CP1.Previous;
  if cp0 = nil then
    p0 := VecAdd(CP1.Position,VecSub(CP1.Position,CP2.Position))
  else
    p0 := cp0.position;

  cp3 := CP2.Next;
  if cp3 = nil then
    p3 := VecAdd(CP2.Position,VecSub(CP2.Position,CP1.Position))
  else
    p3 := cp3.position;

  // curve end points
  p1 := CP1.Position;
  p2 := CP2.Position;

  CatmullRomInterpolation(p0,p1,p2,p3,t,v);
end;

procedure CubicBezierInterpolationProcedure(CP1, CP2: TPathControlPoint; t: Single; var v: TVector3f);
var
  p0,p1,p2,p3: TVector3f;
begin
  // curve control (or tangent points)
  p1 := vecAdd(CP1.Position, VecScale(CP1.TangentVector2,CP1.Path.TangentInfluence));
  p2 := vecAdd(CP2.Position, VecScale(CP2.TangentVector1,CP1.Path.TangentInfluence));

  // curve end points
  p0 := CP1.Position;
  p3 := CP2.Position;

  CubicBezierInterpolation(p0,p1,p2,p3,t,v);
end;

procedure HermiteInterpolationProcedure(CP1, CP2: TPathControlPoint; t: Single; var v: TVector3f);
var
  p0,p1,p2,p3: TVector3f;
begin
  // tangent points
  p2 := VecScale(CP1.TangentVector2,CP1.Path.TangentInfluence);
  p3 := VecScale(CP2.TangentVector2,CP1.Path.TangentInfluence);

  // curve end points
  p0 := CP1.Position;
  p1 := CP2.Position;

  HermiteInterpolation(p0,p1,p2,p3,t,v);
end;

//
// TControlPoint
//
constructor TPathControlPoint.Create(aPath: TPath);
begin
  inherited Create;
  FPath                := aPath;
  FUserDefinedTangent1 := False;
  FUserDefinedTangent2 := False;
end;

procedure TPathControlPoint.SetPosition(v: TVector3f);
begin
  FPosition := v;
end;

procedure TPathControlPoint.SetUpVector(v: TVector3f);
begin
  FUpVector := v;
end;

procedure TPathControlPoint.SetTangentVector1(v: TVector3f);
begin
  FTangentVector1 := v;
end;

procedure TPathControlPoint.SetTangentVector2(v: TVector3f);
begin
  FTangentVector2 := v;
end;

function TPathControlPoint.GetPreviousControlPoint : TPathControlPoint;
begin
  Result := FPath.GetControlPoint(FPath.GetControlPointIndex(Self) - 1);
  if Result = nil then
    if FPath.IsClosed then begin
      // wrap-around to end of path
      Result := FPath.GetControlPoint(FPath.GetControlPointCount - 1);
    end;
end;

//The ID should reflect the position of this controlpoint in the list in it's path
function TPathControlPoint.GetID: Integer;
begin
  Result := FPath.GetControlPointIndex(Self);
end;

function TPathControlPoint.GetNextControlPoint : TPathControlPoint;
begin
  Result := FPath.GetControlPoint(FPath.GetControlPointIndex(Self) + 1);
  if Result = nil then
    if FPath.IsClosed then begin
      // wrap-around to beginning of path
      Result := FPath.GetControlPoint(0);
    end;
end;

function  TPathControlPoint.Clone: TPathControlPoint;
begin
  Result := TPathControlPoint.Create(FPath);
  Result.Position := FPosition;
  Result.UpVector := FUpVector;
  Result.TangentVector1 := FTangentVector1;
  Result.TangentVector2 := FTangentVector2;
  Result.Tag := FTag;
  Result.UserDefinedTangent1 := FUserDefinedTangent1;
  Result.UserDefinedTangent2 := FUserDefinedTangent2;
end;

procedure TPathControlPoint.Assign(cp: TPathControlPoint);
begin
  FPosition := cp.Position;
  FUpVector := cp.UpVector;
  FTangentVector1 := cp.TangentVector1;
  FTangentVector2 := cp.TangentVector2;
  FTag := cp.Tag;
  FUserDefinedTangent1 := cp.UserDefinedTangent1;
  FUserDefinedTangent2 := cp.UserDefinedTangent2;
end;

procedure TPathControlPoint.CalculateTangentVectors;
{
  cp  = this control point
  cp1 = previous control point
  cp2 = next control point
  m1  = magnitude of a;
  m2  = magnitude of b;
  d   = maximum tangent length (1/2 of minimum of both neighbouring segment lengths [m1,m2])

  a  = vector(cp1 - cp)
  b  = vector(cp2 - cp)
  c  = a X b (cross product)
  n  = -((a + b)/2)

  t1 = c X n (cross product)
  t2 = -t1

      cp
      /\
     /  \
 cp1/    \cp2
}
var
  cp,cp1,cp2: TPathControlPoint;
  a,b,c,n,t1,t2: TVector3f;
  d,m1,m2: single;
begin
  cp  := Self;
  cp1 := GetPreviousControlPoint;
  cp2 := GetNextControlPoint;

  //Initialize the tangentvectors (quick & dirty)
  if Assigned(cp2) then
  begin
    if not cp.UserDefinedTangent1 then
      cp.FTangentVector1 := VecScale(VecSub(cp.Position,cp2.Position),1/3);

    if not Assigned(cp1) and not cp.UserDefinedTangent2 then
      cp.FTangentVector2 := VecNegate(cp.FTangentVector1);
  end;

  if Assigned(cp1) then
  begin
    if not cp.UserDefinedTangent2 then
      cp.FTangentVector2 := VecScale(VecSub(cp.Position,cp1.Position),1/3);

    if not Assigned(cp2) and not cp.UserDefinedTangent1 then
      cp.FTangentVector1 := VecNegate(cp.FTangentVector2);
  end;

  //If both neighbouring controlpoints exist, do a more sophisticated computation
  if Assigned(cp1) and Assigned(cp2) then
  begin
    a := VecSub(cp1.Position,cp.Position);
    b := VecSub(cp2.Position,cp.Position);
    c := VecCrossProduct(a,b);

    //We are actually checking the angle here.
    //If the angle is too close to 0 or 180, we skip the following calculation
    //Explanation: The length of A x B = cos(angle)
    if VecMagnitude(c) > 0.1 then
    begin
      //normalize c for future calculations
      c := VecNormalize(c);

      //Compute normal vector
      //This vector is perpendicular to the tangents
      //We compute it by averaging the a and b vectors
      //The result is then scaled to the average length of a and b
      n := VecNegate(VecScale(VecAdd(VecNormalize(a),VecNormalize(b)),0.5));

      //Compute tangents (will have unit length)
      t1 := VecNormalize(VecCrossProduct(c,n));
      t2 := VecNegate(t1);

      //Compute tangent length
      //We will use the minimum length of a and b, divided by 2
      m1 := VecMagnitude(a);
      m2 := VecMagnitude(b);
      d  := Minf(m1,m2)/2;

      //Scale tangents
      //don't need to normalize because they are unit length allready
      t1 := VecScale(t1,d);
      t2 := VecScale(t2,d);

      if not cp.UserDefinedTangent1 then cp.FTangentVector1  := t1;
      if not cp.UserDefinedTangent2 then cp.FTangentVector2  := t2;
    end;
  end;
end;

//
// TPath
//
constructor TPath.Create;
begin
  inherited Create;

  FControlPoints := TList.Create;
  FSelectedControlPoints := TList.Create;

  SetPathType(ptLinear);
  SetIsClosed(False);
  FTangentInfluence := 1.0;
  ObjectType := otPath;
  Name       := 'Path';
end;

destructor  TPath.Destroy;
begin
  Clear;
  FControlPoints.Free;
  FSelectedControlPoints.Free;
  inherited Destroy;
end;

procedure TPath.Assign(Source: TBasePrimitive);
var
  i: Integer;
  cp: TPathControlPoint;
  path: TPath;
begin
  if Source is TPath then begin
    path := TPath(Source);
    Clear;
    for i := 0 to path.ControlPointCount - 1 do begin
      cp := AddControlPoint(Vector3f(0,0,0),False,0);
      cp.Assign(path.ControlPoints[i]);
    end;
    IsClosed := path.IsClosed;
    PathType := path.PathType;
    TangentInfluence := path.TangentInfluence;
  end else
    inherited Assign(Source);
end;

procedure TPath.CalculateTangentVectors;
var
  i: Integer;
begin
  //Calculate tangents for every controlpoint
  for i := 0 to GetControlPointCount - 1 do
    GetControlPoint(i).CalculateTangentVectors();
end;

procedure TPath.DeleteControlPoint(aPoint: TPathControlPoint);
var
  Index: Integer;
begin
  Index := GetControlPointIndex(aPoint);
  if Index <> -1 then DeleteControlPoint(Index);
end;

procedure TPath.DeleteControlPoint(aIndex: Integer);
begin
  if (aIndex < 0) or (aIndex >= FControlpoints.Count) then Exit;
  TPathControlPoint(FControlPoints.Items[aIndex]).Free;
  FControlPoints.Delete(aIndex);
end;

procedure TPath.SetPathType(const aValue: TPathType);
begin
  FPathType := aValue;
  case FPathType of
    ptLinear      : FPathInterpolationProcedure := LinearInterpolationProcedure;
    ptCatmullRom  : FPathInterpolationProcedure := CatmullRomInterpolationProcedure;
    ptCubicBezier : FPathInterpolationProcedure := CubicBezierInterpolationProcedure;
    ptHermite     : FPathInterpolationProcedure := HermiteInterpolationProcedure;
  end;
end;

procedure TPath.SetIsClosed(const aValue: Boolean);
begin
  if aValue <> FIsClosed then begin
    FIsClosed := aValue;
    CalculateTangentVectors;
  end;
end;

procedure TPath.SetTangentInfluence(const aValue: Single);
begin
  if aValue <> FTangentInfluence then begin
    FTangentInfluence := aValue;
    CalculateTangentVectors;
  end;
end;

function  TPath.GetControlPointCount : Integer;
begin
  Result := FControlPoints.Count;
end;

function  TPath.GetControlPoint(aIndex: Integer): TPathControlPoint;
begin
  Result := nil;
  if (aIndex < 0) Or (aIndex >= FControlPoints.Count) then exit;
  Result := TPathControlPoint(FControlPoints.Items[aIndex]);
end;

function  TPath.GetControlPointIndex(aControlPoint: TPathControlPoint): Integer;
begin
  Result := FControlPoints.IndexOf(aControlPoint);
end;

function  TPath.CalculatePathSegmentLength(t0,t1: Single): Single;
const
  cMinSegmentLength    = 1; // 1 units in length
  cMinSegmentLengthSqr = Sqr(cMinSegmentLength);

var
  m0,m1: TVector3f;
  t,d: Single;
begin
  m0 := GetPosition(t0);
  m1 := GetPosition(t1);
  d := Sqr(m0.x - m1.x) + Sqr(m0.y - m1.y) + Sqr(m0.z - m1.z);
  if d > cMinSegmentLengthSqr then begin
    // segment length is larger than the minimum segment length so split again
    t := (t0 + t1) / 2;
    Result := CalculatePathSegmentLength(t0,t) +
              CalculatePathSegmentLength(t,t1);
  end else begin
    Result := Sqrt(d);
  end;
end;

function  TPath.CalculatePathLength: Single;
begin
  Result := CalculatePathSegmentLength(0,0.5) +
            CalculatePathSegmentLength(0.5,1);
end;

function  TPath.GetSegment(var t: single; out cp1,cp2: TPathControlPoint): Boolean;
var
  fIndex: Single;
  iIndex1: Integer;
  iIndex2: Integer;
begin
  Result := False;

  CP1 := nil;
  CP2 := nil;

  if GetControlPointCount <= 1 then Exit;

  if t <= 0.0 then begin
    CP1 := GetControlPoint(0);
    CP2 := GetControlPoint(1);
    t   := 0.0;
  end else begin
    if FIsClosed then
      fIndex := t * (GetControlPointCount)
    else
      fIndex := t * (GetControlPointCount - 1);

    iIndex1 := Trunc(fIndex);
    iIndex2 := iIndex1 + 1;
    fIndex  := fIndex - iIndex1;

    iIndex1 := iIndex1 mod GetControlPointCount;
    iIndex2 := iIndex2 mod GetControlPointCount;

    CP1 := GetControlPoint(iIndex1);
    CP2 := GetControlPoint(iIndex2);
    if fIndex >= 1 then fIndex := 1;
    t := fIndex; 
  end;

  Result := True;
end;

function  TPath.GetPosition(t: single): TVector3f;
var
  cp1 : TPathControlPoint;
  cp2 : TPathControlPoint;
begin
  Result := Vector3f(0,0,0);

  if not GetSegment(t,cp1,cp2) then Exit;

  FPathInterpolationProcedure(cp1,cp2,t,Result);
end;

function  TPath.GetDirection(t: single): TVector3f;
const
  cDirThreshold = 0.0001;
var
  pa,pb: TVector3f;
  cp1    : TPathControlPoint;
  cp2    : TPathControlPoint;
  t1     : single;
  t2     : single;
begin
  Result := Vector3f(0,0,0);

  if not GetSegment(t,cp1,cp2) then Exit;

  t1 := t - cDirThreshold;
  t2 := t + cDirThreshold;

  FPathInterpolationProcedure(cp1,cp2,t1,pa);
  FPathInterpolationProcedure(cp1,cp2,t2,pb);
  Result := VecNormalize(VecSub(pb,pa));
end;

function  TPath.GetUp(t: single): TVector3f;
var
  cp1    : TPathControlPoint;
  cp2    : TPathControlPoint;
begin
  Result := Vector3f(0,0,0);

  if not GetSegment(t,cp1,cp2) then Exit;

  Result.x := cp1.Position.x + (cp2.Position.x - cp1.Position.x) * t;
  Result.y := cp1.Position.y + (cp2.Position.y - cp1.Position.y) * t;
  Result.z := cp1.Position.z + (cp2.Position.z - cp1.Position.z) * t;

  Result := VecNormalize(Result);
end;

//Computes the tangent vector at a position along the curve denoted by
//the curve parameter t (in range 0...1)
function TPath.DY(t: single; interval: single = 1e-3): TVector3f;
begin
  Result := VecScale(VecSub(GetPosition(t + interval), GetPosition(t)), 1/interval);
end;

//Returns the "speed" at a position t on the curve.
//The speed is "how fast the position changes, over a change of parameter t".
function TPath.GetSpeedAt(t: single): single;
begin
  Result := VecMagnitude(DY(t));
end;

//This is the most important function here. It offers a better
//so called "parameterization" for our curve.
//s will be a value between 0 and the length of the curve.
//The function will return a curve parameter (0 <= t <= 1).
function TPath.GetCurveParameter(s: single): single;
var
  h              : Single;
  i              : Integer;
  k1, k2, k3, k4 : Single;
const
  n:  Integer = 100;
begin
  Result := 0;
  h := s / n;

  //We use the Runge-Kutta method for solving ordinary differential
  //equations here. I just copied this part and it worked YAY. :)
  for i := 1 to n do
  begin
    k1 := h/GetSpeedAt(Result);
    k2 := h/GetSpeedAt(Result + k1/2);
    k3 := h/GetSpeedAt(Result + k2/2);
    k4 := h/GetSpeedAt(Result + k3);
    Result := Result + (k1 + 2*(k2 + k3) + k4)/6;
  end;
end;

function TPath.GetCenter: TVertice;
var
  c: TVector3f;
begin
  if HasCenterChanged then begin
    c := TPath(Self).GetControlPointsCenter;
    fCenter.x := c.x;
    fCenter.y := c.y;
    fCenter.z := c.z;
    HasCenterChanged := false;
  end;
  Result := fCenter;
end;

function TPath.GetControlPointsCenter: TVector3f;
var
  i: Integer;
begin
  Result := Vector3f(0,0,0);
  if ControlPointCount = 0 then Exit;
  Result := ControlPoints[0].Position;
  for i := 1 to ControlPointCount - 1 do
    Result := VecAdd(Result,ControlPoints[i].Position);
  Result := VecScale(Result,1/ControlPointCount);
end;

function TPath.Object_SelectByRange(v1,v2: TVertice): boolean;
var i: integer;
    sv, ev: TVector3f;  //Start and end vectors of segment
begin
    result := false;
    if Visible then
    begin
        result := true;
        I := 0;
        while GetNextPathSegmentForSelection(I, sv, ev) do
        begin
            if (v1.x=0) and (v2.x=0) then begin                                 // side view
                if (sv.y>v1.y) or (sv.y<v2.y) or
                   (sv.z<v1.z) or (sv.z>v2.z) then result := false;
            end;
            if (v1.y=0) and (v2.y=0) then begin                                 // top view
                if (sv.x<v1.x) or (sv.x>v2.x) or
                   (sv.z<v1.z) or (sv.z>v2.z) then result := false;
            end;
            if (v1.z=0) and (v2.z=0) then begin                                 // front view
                if (sv.x<v1.x) or (sv.x>v2.x) or
                   (sv.y>v1.y) or (sv.y<v2.y) then result := false;
            end;
        end;
    end;
end;

//the function returns true if the point is
//very close to one of the paths (This is used for selection).
//Because this is used in the 2D views, one dimention will be reduced to zero
function TPath.IsPointOnPath(p: TVertice): boolean;
var I: integer;
    sv, ev: TVector3f;
begin
   Result := false;
   if Visible and (not Frozen) then
   begin
       //Split the path into segments and check for each segment whether
       // The "click" position is close enough
       I := 0;
       while GetNextPathSegmentForSelection(I, sv, ev) do
         if IsPointOnSegment( Vector3f(p.x, p.y, p.z), sv, ev, PATH_SELECT_TOLERANCE) then
         begin
           Result := True;
           Exit;
         end;
   end;
end;

//Used to iterate over the segments of a subdivided spline
//This is used for the selection routines as these need a lot of points and segments of our spline
//to do the calculations with
//  @aIndex: The index of the segment (will automaticly increase)
//  @v1, v2: The vertices of the segment that is returned
//  False is returned, when all segments have been iterated
function TPath.GetNextPathSegmentForSelection(var aIndex: Integer; out v1,v2: TVector3f): boolean;
var
  MaxNumberOfSegs: Integer;
  SegmentOffset: Single;
begin
  MaxNumberOfSegs := ControlPointCount * PATH_SELECT_SEGMENTS_PER_CONTROLPOINT;
  SegmentOffset := 1 / MaxNumberOfSegs;

  if (aIndex < MaxNumberOfSegs) then
  begin
    v1 := GetPosition( aIndex       * SegmentOffset );
    v2 := GetPosition( (aIndex + 1) * SegmentOffset );
    Result := True;
    Inc(aIndex);
  end
  else
    Result := False;    //Done
end;

//The user clicked at "AVertex". We will select a controlpoint that lies
//within the given range of that point
procedure TPath.SelectControlPoint(AVertex: TVertice; ARange: double; const AClearSelection: Boolean);
var counter: integer;
    CP: TPathControlPoint;
    vec: TVector3f;
begin
    if AClearSelection then
      FSelectedControlPoints.Clear();

    counter := 0;
    if ARange<1 then ARange := 1;

    while (counter < FControlPoints.Count) do
    begin
        CP := TPathControlPoint(FControlPoints[Counter]);
        vec := TPathControlPoint(FControlPoints[Counter]).Position;
        if AVertex.x=0 then vec.x := 0;
        if AVertex.y=0 then vec.y := 0;
        if AVertex.z=0 then vec.z := 0;
        if (AVertex.x>=vec.x-ARange) and (AVertex.x<=vec.x+ARange) and
           (AVertex.y>=vec.y-ARange) and (AVertex.y<=vec.y+ARange) and
           (AVertex.z>=vec.z-ARange) and (AVertex.z<=vec.z+ARange) then
        begin
            //Add this controlpoint to the list of selected ones..
            //Remove it if it's allready there
            if FSelectedControlPoints.IndexOf(CP) <> -1 then
              FSelectedControlPoints.Remove(CP)
            else
              FSelectedControlPoints.Add(CP);

            Exit;
        end;
        inc(counter);
    end;
end;


//We will select the controlpoints that lie between the two given coordinates
procedure TPath.SelectControlPoints(const v1, v2: TVertice; AClearSelection: Boolean);
var i: integer;
    CPSelected: boolean;
    CP: TPathControlPoint;
begin
    if Visible then
    begin
        if AClearSelection then
          FSelectedControlPoints.Clear();

        for i := 0 to FControlPoints.Count-1 do
        begin
            CPSelected := false;
            CP := FControlPoints[i];
            if (v1.x=0) and (v2.x=0) then begin                                 // side view
                if (CP.Position.y<v1.y) and (CP.Position.y>v2.y) and
                   (CP.Position.z>v1.z) and (CP.Position.z<v2.z) then CPSelected := true;
            end;
            if (v1.y=0) and (v2.y=0) then begin                                 // top view
                if (CP.Position.x>v1.x) and (CP.Position.x<v2.x) and
                   (CP.Position.z>v1.z) and (CP.Position.z<v2.z) then CPSelected := true;
            end;
            if (v1.z=0) and (v2.z=0) then begin                                 // front view
                if (CP.Position.x>v1.x) and (CP.Position.x<v2.x) and
                   (CP.Position.y<v1.y) and (CP.Position.y>v2.y) then CPSelected := true;
            end;

            if CPSelected then
              if FSelectedControlPoints.IndexOf(CP) <> -1 then
                FSelectedControlPoints.Remove(CP)
              else
                FSelectedControlPoints.Add(CP);
        end;
    end;
end;

//Used in 3d view to select controlpoints by dragging
procedure TPath.SelectControlPoints(const ax1, ay1, ax2, ay2: integer;
                                    const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                    const aViewport: TViewport; const AClearSelection: boolean);
var
  j, k: Integer;
  CP: TPathControlPoint;
  vertex: TVertice;
  point: TPoint2D;
  CPList: TList;
begin
  if AClearSelection then
    FSelectedControlPoints.Clear();

  CPList := TList.Create();
  try
    for k := 0 to FControlpoints.Count-1 do
    begin
      CP := Controlpoints[k];
      vertex := TVertice.Create(CP.Position.x, CP.Position.y, CP.Position.z);
      point := vertex.ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewPort);
      if (point.X >= ax1) and (point.X <= ax2) and (point.Y >= ay1) and (point.Y <= ay2) then
          CPList.Add(CP);
    end;

    for j := 0 to CPList.Count-1 do
    begin
        CP := TPathControlPoint(CPList[j]);
        if FSelectedControlPoints.IndexOf(CP) = -1 then
          FSelectedControlPoints.Add(CP)
        else
          FSelectedControlPoints.Remove(CP);
    end;
  finally
    CPList.Free();
  end;
end;

procedure TPath.Clear;
var
  i: Integer;
begin
  for i := FControlPoints.Count - 1 downto 0 do
    TPathControlPoint(FControlPoints.Items[i]).Free;
  FControlPoints.Clear;
end;

procedure TPath.AddControlPoint(aPoint: TPathControlPoint);
begin
  FControlPoints.Add(aPoint);
end;

function  TPath.AddControlPoint(vertex: TVector3f; doSnap: boolean; gridSize: integer): TPathControlPoint;
begin
  Result := TPathControlPoint.Create(Self);

  if doSnap then vertex := VecSnap(vertex,gridSize);

  Result.Position := vertex;
  Result.TangentVector1 := Vector3f(0,0,0);
  Result.TangentVector2 := Vector3f(0,0,0);
  Result.UpVector := Vector3f(0,1,0);

  FControlPoints.Add(Result);
  CalculateTangentVectors;
end;

function TPath.GetFirstSelectedControlPoint: TPathControlPoint;
begin
  if FSelectedControlPoints.Count > 0 then
    Result := TPathControlPoint(FSelectedControlPoints[0])
  else
    Result := nil;
end;


{   Helper routines  }

function PathTypeToString(aType: TPathType): string;
begin
  case aType of
    ptLinear:       Result := 'Linear';
    ptCatmullRom:   Result := 'CatmullRom';
    ptCubicBezier:  Result := 'CubicBezier';
    ptHermite:      Result := 'Hermite';
  end;
end;

function StringToPathType(aTypeStr: string): TPathType;
begin
  aTypeStr := LowerCase(aTypeStr); //We don't care about the case

  if aTypeStr = 'linear' then
    Result := ptLinear
  else if aTypeStr = 'catmullrom' then
    Result := ptCatmullRom
  else if aTypeStr = 'cubicbezier' then
    Result := ptCubicBezier
  else if aTypeStr = 'hermite' then
    Result := ptHermite
  else
    raise Exception.CreateFmt('"%s" is not a valid path-type',[aTypeStr]);
end;

function IsPointOnSegment(point, startVec, endVec: TVector3f; Tolerance: Single): Boolean;
var
  Pt:     TVector3f; //point relative to starting point of segment
  SegDir: TVector3f; //segment direction
//  Dist:   Single;    //Distance between point and line
  SphereCenter: TVector3f;
  SphereSqrRadius: Single; //The squared radius of the sphere
begin
  Result := False;

  //Make everything relative to startVec. StartVec is now our new origin
  Pt := VecSub( point, startVec );
  SegDir := VecSub( endVec, startVec );

  //Find the sphere that exactly encloses both segment points
  //The computed sphere center will be relative to "startVec"
  SphereCenter := VecScale(SegDir, 0.5);
  SphereSqrRadius := VecSqrMagnitude( SphereCenter );

  //If the point lies inside the sphere, then the point is on the segment
  //(just a very rough check but it works)
  //We also add a bit of tolerance to make clicking the path easier
  if VecSqrMagnitude( VecSub(Pt,SphereCenter) ) < SphereSqrRadius + Tolerance then
    Result := True;

  {//Project point onto segment-line by using dot-product
  Dist := VecDotProduct( Pt, VecNormalize(SegDir) );

  Result := (Dist < Tolerance); }
end;

end.
