{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
// Unit:    unit_pluginmanager
// Purpose: manages plugin objects (execute specific plugin etc)
// Note:    works closely with unit_plugins and unit_scene

unit unit_PluginManager;

{$MODE Delphi}

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Dialogs, Menus, Contnrs,
  StrUtils, Forms, FileUtil,
  unit_Log,
  unit_Functions,
  unit_Scene,
  unit_PluginVars,
  unit_Plugins,
  unit_PrimitiveList;

type
  TPluginManager = class(TObjectList)
  private
    fScene: TScene;
    fVersion: currency;
    fCallBack: pCallBack;
    fDirectory: string;
    fOnBeforeExecute: TNotifyEvent;
    fOnAfterExecute: TNotifyEvent;
    fLog: TLogWriter;
    // generic functions used by plugin-system
    function ValidatePlugin(APlugin: TPlugin): integer;
    function GetPlugin(AnIndex: integer): TPlugin;
    procedure AddPlugin(APlugin: TPlugin);
    procedure InitializePlugin(APlugin: TPlugin);
  public
    // constructor/destructor
    constructor Create(AScene: TScene; AVersionInfo: currency; ACallBackFunction: pCallBack; ADirectory: string);
    destructor Destroy; override;
    // methods
    function LoadDLLFiles: boolean;
    procedure UpdateMenu(AParentMenuItem: TMenuItem);
    procedure PluginExecute(Sender: TObject);
    // methods used by callback proc; they process data send from/to plugins
    function GetMem(var ACallBackRecord: TCallBackRecord): Cardinal;
    function GetData(var ACallBackRecord: TCallBackRecord): Cardinal;
    function SetData(ACallBackRecord: TCallBackRecord): Cardinal;
    // properties
    property Plugins[AnIndex: integer]: TPlugin read GetPlugin;
    property Log: TLogWriter read fLog write fLog;
    property OnBeforeExecute: TNotifyEvent read fOnBeforeExecute write fOnBeforeExecute;
    property OnAfterExecute: TNotifyEvent read fOnAfterExecute write fOnAfterExecute;
  end;

implementation

uses //frmProgressForm,
     unit_XMLPluginReader,
     unit_XMLPluginWriter;

// ### TPluginManager ##########################################################

constructor TPluginManager.Create(AScene: TScene; AVersionInfo: currency; ACallBackFunction: pCallBack; ADirectory: string);
begin
    inherited Create;
    OwnsObjects := true;
    fScene := AScene;
    fVersion := AVersionInfo;
    fCallBack := ACallBackFunction;
    fDirectory := ADirectory;
    fLog := TLogWriter.Create(fDirectory+'plugins.log',DateTimeToStr(Now)+': Plugin logfile created.');
end;

destructor TPluginManager.Destroy;
begin
    inherited;
    if Assigned(fLog) then begin
//        log.WriteLog(DateTimeToStr(Now)+': Plugins unloaded.');
        FreeAndNil(fLog);
    end;
    fCallBack := nil;
    fScene := nil;
end;

procedure TPluginManager.AddPlugin(APlugin: TPlugin);
begin
    Add(APlugin);
end;

procedure TPluginManager.InitializePlugin(APlugin: TPlugin);
begin
    if Assigned(APlugin.Initialize) then
        APlugin.Initialize({Application.Handle});  //!!!

    // INFO, PJ: When we want more initialization, we add different methods with a serial number (initialize2, initialize3, etc)        
end;

function TPluginManager.GetPlugin(AnIndex: integer): TPlugin;
begin
    if (AnIndex>=0) and (AnIndex<Count) then
        Result := TPlugin(Items[AnIndex])
    else
        Result := nil;
end;

// basic validation on APlugin: TODO - expand these validations
function TPluginManager.ValidatePlugin(APlugin: TPlugin): integer;
var pversion: currency;
begin
    if not Assigned(APlugin) then
        Result := 0
    else begin
        Result := -1;
        if length(string(APlugin.name))=0 then Result := 1;
        if length(string(APlugin.description))=0 then Result := 2;
        if length(string(APlugin.author))=0 then Result := 3;
        if length(string(APlugin.email))=0 then Result := 4;
        if length(string(APlugin.versionPlugin))=0 then Result := 5;
        if IsNumeric(string(APlugin.versionDeleD)) then begin
            pversion := strtocurr(string(APlugin.versionDeleD));
            if (pversion<1.6) or (pversion>fVersion) then
                Result := 6;
        end;
    end;
end;

// loads all DeleD DLL's in a specific directory and creates a plugin object for each DLL
function TPluginManager.LoadDLLFiles: boolean;
var srDLL: TSearchRec;
    hLib: THandle;
    plugin: TPlugin;
    validation: integer;
begin
    if FindFirstUTF8(fDirectory+'*.DLL',0,srDLL) = 0 then repeat

        hLib := LoadLibrary(PChar(fDirectory+srDLL.Name));
        if hLib <> 0 then begin

            plugin := TPlugin.Create(hlib,srDLL.Name,fCallBack);
            validation := ValidatePlugin(plugin);
            if validation=-1 then begin
                InitializePlugin(plugin);
                AddPlugin(plugin);
            end else begin
                case validation of
                    0:  log.WriteLog(DateTimeToStr(Now)+': '+srDLL.Name+' not loaded.');
                    1:  log.WriteLog(DateTimeToStr(Now)+': '+srDLL.Name+': no Name property.');
                    2:  log.WriteLog(DateTimeToStr(Now)+': '+srDLL.Name+': no Description property.');
                    3:  log.WriteLog(DateTimeToStr(Now)+': '+srDLL.Name+': no Author property.');
                    4:  log.WriteLog(DateTimeToStr(Now)+': '+srDLL.Name+': no Email property.');
                    5:  log.WriteLog(DateTimeToStr(Now)+': '+srDLL.Name+': no Plugin Version property.');
                    6:  log.WriteLog(DateTimeToStr(Now)+': '+srDLL.Name+': wrong DeleD Version property.');
                end;
            end;

        end;

    until FindNextUTF8(srDLL) <> 0;
    FindCloseUTF8(srDLL);

    Result := Count > 0;
end;

// creates a menu structure for plugin objects
procedure TPluginManager.UpdateMenu(AParentMenuItem: TMenuItem);
var i: integer;
    plugin: TPlugin;
    menuItem: TMenuItem;
begin
    if Assigned(AParentMenuItem) then begin      // if menuitem to which plugins are to be added exists...
        for i := 0 to Count-1 do begin           // create new menuitems under that specific menuitem
            plugin := GetPlugin(i);
            menuItem := TMenuItem.Create(AParentMenuItem.GetParentComponent);
            menuItem.Caption := plugin.name;
            menuItem.OnClick := PluginExecute;
            menuItem.Tag := i;
            AParentMenuItem.Insert(2+i,menuItem);
            //AParentMenuItem.Add(menuItem);
        end;
    end;
end;

// executes a plugin identified by the tag-property of the menu-item under which the plugin is placed
procedure TPluginManager.PluginExecute(Sender: TObject);
var plugin: TPlugin;
begin
    if Sender is TMenuItem then begin
        plugin := GetPlugin(TMenuItem(Sender).Tag);
        log.WriteLog(DateTimeToStr(Now)+': Executing: '+plugin.name);
        try
            if Assigned(fOnBeforeExecute) then fOnBeforeExecute(plugin);
            plugin.Execute();
            if Assigned(fOnAfterExecute) then fOnAfterExecute(plugin);
        except
            on E: Exception do begin
                MessageDlg(E.Message, mtError, [mbOk], 0);
                log.WriteLog(DateTimeToStr(Now)+': '+E.Message);
            end;
        end;
        log.WriteLog(DateTimeToStr(Now)+': Finished: '+plugin.name);
    end;
end;

// a plugin wants to know how much memory is needed to store a specific set of
// XML data; DeleD creates the XML data and specifies the size of that data in
// the ResponseSize property of the CallBackRecord; using that size, the plugin
// can now allocate the amount of memory needed for the XML data and issue a
// GetData command to get the actual data, which is then returned in the ResponseXML
// property of the CallBackRecord
{$HINTS OFF}
function TPluginManager.GetMem(var ACallBackRecord: TCallBackRecord): Cardinal;
var xmlPluginWriter: TXMLPluginWriter;
    xmlResponse: string;
begin
    try
        Result := PE_NOERROR;
        // create an XML string based on what is requested by the plugin; a plugin
        // can request various items from DeleD (primitives, materials etc); this
        // request is submitted in XML-format; here, we read that XML request and
        // create a response in XML-format too; the size of that response is returned
        // to the plugin so it can allocate memory for the final transfer of data
        // using a GetData command; this GetData command sends the same XML request
        // to DeleD after which DeleD returns its XML response to the plugin
        xmlPluginWriter := TXMLPluginWriter.Create;
        try
            xmlPluginWriter.WriteToString(fScene,string(ACallBackRecord.RequestXML),xmlResponse);
        finally
            FreeAndNil(xmlPluginWriter);
        end;

        // set # of bytes needed to store the string in memory
        ACallBackRecord.ResponseSize := Length(xmlResponse)+1;   // # of bytes needed (+1 for savety)
    except
        on E: Exception do begin
            Result := PE_INVALIDMEM;
            raise Exception.Create('TPluginManager.GetMem: '+E.Message);
        end;
    end;
end;

// a plugin has requested data from DeleD; DeleD creates a response in XML format
// and copies that data into the ResponseXML property of the CallBackRecord
function TPluginManager.GetData(var ACallBackRecord: TCallBackRecord): Cardinal;
var xmlPluginWriter: TXMLPluginWriter;
    xmlResponse: string;
begin
    try
        Result := PE_NOERROR;
        // create an XML string based on what is requested by the plugin
        xmlPluginWriter := TXMLPluginWriter.Create;
        try
            xmlPluginWriter.WriteToString(fScene,string(ACallBackRecord.RequestXML),xmlResponse);
        finally
            FreeAndNil(xmlPluginWriter);
        end;
        // copy the XML string to the ResponseXML property of the callback record
        // or signal an error if not enough memory was allocated
        if Length(xmlResponse) <= Integer(ACallBackRecord.ResponseSize) then
            StrCopy(ACallBackRecord.ResponseXML,PChar(xmlResponse))
        else begin
            ACallBackRecord.ResponseSize := Length(xmlResponse);
            Result := PE_INSUFFICIENT_DATASIZE;
        end;
    except
        on E: Exception do begin
            Result := PE_INVALIDGET;
            raise Exception.Create('TPluginManager.GetData: '+E.Message);
        end;
    end;
end;

// a plugin has send a request to set data within DeleD; the plugin sends data
// in XML format; DeleD reads this XML data and updates its own objects accordingly
function TPluginManager.SetData(ACallBackRecord: TCallBackRecord): Cardinal;
var xmlPluginReader: TXMLPluginReader;
    xmlRequest: string;
begin
    try
        Result := PE_NOERROR;
        // get the XML string from the plugin
        xmlRequest := string(ACallBackRecord.RequestXML);

        // and process the XML using a TXMLPluginReader object; this object
        // uses a TXMLStringReader object to read and process the XML string
        xmlPluginReader := TXMLPluginReader.Create;
        try
            xmlPluginReader.ReadFromString(xmlRequest,fScene);
        finally
            FreeAndNil(xmlPluginReader);
        end;
    except
        on E: Exception do begin
            Result := PE_INVALIDSET;
            raise Exception.Create('TPluginManager.SetData: '+E.Message);
        end;
    end;
end;
{$HINTS ON}

end.
