{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PolygonAlongEdgeScalerTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_Primitives,
  unit_ParamList,
  unit_Lights,
  unit_PrimitiveTool,
  unit_PrimitiveList;

type
  TPolygonAlongEdgeScalerFunction = class(TObject)
  private
    procedure DoPolygonAlongEdgeScale(APrimitive: TBasePrimitive;
                                      APolygon: TPolygon;
                                      ANumberOfUnits: integer;
                                      var AContinueScale: boolean;
                                      APerformEdgeCheck: boolean = true);
  public
    class procedure Execute(APrimitive: TBasePrimitive;
                            APolygon: TPolygon;
                            ANumberOfUnits: integer;
                            var AContinueScale: boolean;
                            APerformEdgeCheck: boolean = true);
  end;

  TPolygonAlongEdgeScaler = class(TPrimitiveTool)
  private
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  unit_PrimitiveScalerTool;

// ### TPolygonAlongEdgeScalerFunction #########################################

// scales selected polygons by moving their vertices along adjacent edges;
// if a vertex has more than 1 adjacent edge that are all NOT part of the
// current polygon, it isn't moved. NOTE: what if 2 adjacent polygons are selected?
procedure TPolygonAlongEdgeScalerFunction.DoPolygonAlongEdgeScale(
  APrimitive: TBasePrimitive; APolygon: TPolygon; ANumberOfUnits: integer;
  var AContinueScale: boolean; APerformEdgeCheck: boolean);
const EPS = 0.01;
var i: integer;
    currentVertex,edgeVertex,dirVector: TVertice;
    deltaVectors,edgeVectors: TVerticeList;
    edgeList,adjacentEdges: TEdgeList;
    rollBack: boolean;
begin
    with APrimitive do begin

        edgeList := TEdgeList.Create(false,true);
        adjacentEdges := TEdgeList.Create(false,true);
        deltaVectors := TVerticeList.Create(true,true);
        edgeVectors := TVerticeList.Create(true,true);

        try

            // get adjacent edges in a list + make a copy of the edge directionvectors in a seperate list
            Polygon_GetAdjacentEdges(APolygon,adjacentEdges);
            for i := 0 to adjacentEdges.Count-1 do
                edgeVectors.Add(adjacentEdges[i].DirectionVector.Copy);

            // we use the edge that is NOT part of the current polygon as direction vector
            for i := 0 to APolygon.NumVertices-1 do begin

                currentVertex := APolygon.VerticeList[i];

                // find out which adjacent edge(s) of the current vertex are NOT
                // part of the polygon; we use that one as direction vector
                currentVertex.EdgeList.IsNotIn(APolygon.EdgeList,edgeList);

                // we found exactly one adjacent edge that isn't part of the current polygon
                if edgeList.Count=1 then begin

                    // get the other vertex of the edge in question
                    edgeVertex := edgeList[0].GetOtherVertex(currentVertex);

                    // create a normalized direction vector
                    dirVector := TVertice.Create(0,0,0);
                    dirVector.x := currentVertex.x-edgeVertex.x;
                    dirVector.y := currentVertex.y-edgeVertex.y;
                    dirVector.z := currentVertex.z-edgeVertex.z;
                    dirVector.Normalize;

                    // multiply with # of units to move to get the final direction vector
                    dirVector.MultiplyByScalar(ANumberOfUnits);

                    // and save it for later use
                    deltaVectors.Add(dirVector);
                end;
            end;

            // perform scale
            for i := 0 to APolygon.NumVertices-1 do begin
                currentVertex := APolygon.VerticeList[i];
                currentVertex.Add(deltaVectors[i]);
            end;

            // perform edge-check if wanted: checks to see if any edge flipped direction
            // and rolls back the scale operation if that is the case
            if APerformEdgeCheck then begin
                rollBack := false;
                for i := 0 to adjacentEdges.Count-1 do begin
                    if DotProduct(edgeVectors[i],adjacentEdges[i].DirectionVector) < EPS then begin // dot of original and changed edge
                        rollBack := true;
                        break;
                    end;
                end;

                // rollback if needed
                if rollBack then begin
                    AContinueScale := false;
                    for i := 0 to APolygon.NumVertices-1 do begin
                        currentVertex := APolygon.VerticeList[i];
                        currentVertex.Subtract(deltaVectors[i]);
                    end;
                end;
            end;


        finally
            FreeAndNil(deltaVectors);
            FreeAndNil(edgeVectors);
            FreeAndNil(adjacentEdges);
            FreeAndNil(edgeList);
        end;

        HasChanged := true;
    end;
end;

class procedure TPolygonAlongEdgeScalerFunction.Execute(
  APrimitive: TBasePrimitive; APolygon: TPolygon; ANumberOfUnits: integer;
  var AContinueScale: boolean; APerformEdgeCheck: boolean);
begin
    with self.Create do
    try
        DoPolygonAlongEdgeScale(APrimitive,APolygon, ANumberOfUnits, AContinueScale, APerformEdgeCheck);
    finally
        Free;
    end;
end;


// ### TPolygonAlongEdgeScaler #################################################

procedure TPolygonAlongEdgeScaler.DoPrepare(AParamList: TParamList);
begin
end;

procedure TPolygonAlongEdgeScaler.DoExecute(AParamList: TParamList);
var i,j: integer;
    primitive: TBasePrimitive;
    numberOfUnits: integer;
    continueScale: boolean;
begin
    numberOfUnits := TIntegerObject(AParamList.NamedParameters['numberOfUnits']).Value;
    continueScale := true;

    for i := 0 to SelectedPrimitives.Count-1 do begin
        primitive := SelectedPrimitives[i];
        if primitive.ModifyGeometry then begin
            for j := 0 to primitive.SelectedPolygons.Count-1 do
                TPolygonAlongEdgeScalerFunction.Execute(primitive,primitive.SelectedPolygons[j],numberOfUnits,continueScale);
            primitive.HasChanged := true;
        end;
    end;
end;


end.

