{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
// a Bridge method

unit unit_PolygonConnectorTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_PrimitiveList,
  unit_ParamList,
  unit_PrimitiveTool;

type
  TLightWeightPolygon = class(TObject)
  private
    fVertices: TVerticeList;
    fNormal: TVertice;
    fCenter: TVertice;
    function GetNormal: TVertice;
    function GetCenter: TVertice;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(APolygon: TPolygon);
    property Vertices: TVerticeList read fVertices write fVertices;
    property Normal: TVertice read GetNormal;
    property Center: TVertice read GetCenter;
  end;

  TPolygonConnector = class(TPrimitiveTool)
  private
    function FindIndexOfClosestVertex(AVertex: TVertice; APolygon: TLightWeightPolygon): integer;
    procedure DoConnectPolygons(APrimitive: TBasePrimitive);
  protected
  public
    function CanExecute(AParamList: TParamList): boolean; override;
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses unit_BaseTool;

// ### TLightWeightPolygon #####################################################

constructor TLightWeightPolygon.Create;
begin
    fVertices := TVerticeList.Create(true,true);
    fNormal := TVertice.Create(0,0,0);
    fCenter := TVertice.Create(0,0,0);
end;

destructor TLightWeightPolygon.Destroy;
begin
    FreeAndNil(fNormal);
    FreeAndNil(fCenter);
    FreeAndNil(fVertices);
    inherited Destroy;
end;

procedure TLightWeightPolygon.Assign(APolygon: TPolygon);
var i: integer;
begin
    fVertices.Clear;
    for i := 0 to APolygon.VerticeList.Count-1 do
        fVertices.Add(APolygon.VerticeList[i].Copy)
end;

function TLightWeightPolygon.GetCenter: TVertice;
var i: integer;
    c,v: TVertice;
begin
    c := TVertice.Create(0,0,0);
    for i := 0 to Vertices.Count-1 do begin
        v := Vertices[i];
        c.x := c.x + v.x;
        c.y := c.y + v.y;
        c.z := c.z + v.z;
    end;
    fCenter.x := c.x / Vertices.Count;
    fCenter.y := c.y / Vertices.Count;
    fCenter.z := c.z / Vertices.Count;
    FreeAndNil(c);
    Result := fCenter;
end;

function TLightWeightPolygon.GetNormal: TVertice;
const EPS=0.001;
var averageYZ, averageZX, averageXY: double;
    index: integer;
    vertex, previousVertice: TVertice;
    length: double;
begin
    averageYZ := 0;
    averageZX := 0;
    averageXY := 0;

    previousVertice := Vertices[Vertices.Count - 1];
    for index := 0 to Vertices.Count - 1 do begin
        vertex := Vertices[index];
        averageYZ := averageYZ + (previousVertice.Z + vertex.Z) * (vertex.Y - previousVertice.Y);
        averageZX := averageZX + (previousVertice.X + vertex.X) * (vertex.Z - previousVertice.Z);
        averageXY := averageXY + (previousVertice.Y + vertex.Y) * (vertex.X - previousVertice.X);
        previousVertice := vertex;
    end;

    length := Sqrt(averageYZ * averageYZ + averageZX * averageZX + averageXY * averageXY);

    if length < EPS then length := EPS;
    fNormal.X := averageYZ / length;
    fNormal.Y := averageZX / length;
    fNormal.Z := averageXY / length;

    result := fNormal;
end;

// ### TPolygonConnector #######################################################

function TPolygonConnector.FindIndexOfClosestVertex(AVertex: TVertice; APolygon: TLightWeightPolygon): integer;
var vertex,vector: TVertice;
    i,index: integer;
    length: double;
begin
    index := 0;
    length := 99999999;

    for i := 0 to APolygon.Vertices.Count-1 do begin
        vertex := APolygon.Vertices[i];
        vector := vectorSub(vertex,AVertex);
        if vector.Length<=length then begin
            index := i;
            length := vector.Length;
        end;
        FreeAndNil(vector);
    end;

    Result := index;
end;

procedure TPolygonConnector.DoConnectPolygons(APrimitive: TBasePrimitive);
var selectedPolygon1,selectedPolygon2,newPolygon: TPolygon;
    polygon1,polygon2: TLightWeightPolygon;
    axis,vertex,polygonInverseNormal,center: TVertice;
    angle,x,y,z: double;
    matrix: TMatrix;
    i,indexOfClosestVertex: integer;
begin
    selectedPolygon1 := APrimitive.SelectedPolygons[0];
    selectedPolygon2 := APrimitive.SelectedPolygons[1];

    // make local copies of the polygons for rotational purposes;
    // we do not want to rotate the original mesh
    polygon1 := TLightWeightPolygon.Create;
    polygon1.Assign(selectedPolygon1);
    polygon2 := TLightWeightPolygon.Create;
    polygon2.Assign(selectedPolygon2);

    center := polygon1.Center.Copy;
    for i := 0 to polygon1.Vertices.Count-1 do
        polygon1.Vertices[i].Subtract(center);
    FreeAndNil(center);

    center := polygon2.Center.Copy;
    for i := 0 to polygon2.Vertices.Count-1 do
        polygon2.Vertices[i].Subtract(center);
    FreeAndNil(center);
    
    // rotate one polygon to the normal of the other one so they
    // each face in the same direction
    polygonInverseNormal := polygon1.Normal.Inverse;

    angle := vectorAngle(polygon2.Normal,polygonInverseNormal);
    while (angle>1) and (angle<179) do begin
        axis := crossproduct(polygonInverseNormal,polygon2.Normal);
        matrix := MatrixArbAxisSetup(axis,polygon2.Center,angle);

        for i := 0 to polygon1.Vertices.Count-1 do begin
            vertex := polygon1.Vertices[i];
            ///vertex.x := vertex.x-center1.x;
            //vertex.y := vertex.y-center1.y;
            //vertex.z := vertex.z-center1.z;
            x := (vertex.x*matrix[0,0])+(vertex.y*matrix[0,1])+(vertex.z*matrix[0,2]);//+matrix[3,0];
            y := (vertex.x*matrix[1,0])+(vertex.y*matrix[1,1])+(vertex.z*matrix[1,2]);//+matrix[3,1];
            z := (vertex.x*matrix[2,0])+(vertex.y*matrix[2,1])+(vertex.z*matrix[2,2]);//+matrix[3,2];
            vertex.x := x;
            vertex.y := y;
            vertex.z := z;
        end;
        FreeAndNil(axis);
        FreeAndNil(matrix);
        FreeAndNil(polygonInverseNormal);
        polygonInverseNormal := polygon1.Normal.Inverse;
        angle := vectorAngle(polygon2.Normal,polygonInverseNormal);
    end;
    FreeAndNil(polygonInverseNormal);

    // now loop through the vertex lists of the original (!) polygons,
    // create new polygons along the way and add them to the primitive
    for i := 0 to selectedPolygon1.VerticeList.Count-1 do begin
        newPolygon := TPolygon.Create;
        newPolygon.Material := selectedPolygon1.Material;

        APrimitive.Polygon_AddVertex(newPolygon,selectedPolygon1.VerticeList[i]);
        APrimitive.Polygon_AddVertex(newPolygon,selectedPolygon1.VerticeList[(i+1) mod selectedPolygon1.NumVertices]);
        indexOfClosestVertex := FindIndexOfClosestVertex(polygon1.Vertices[(i+1) mod polygon1.Vertices.Count],polygon2);
        APrimitive.Polygon_AddVertex(newPolygon,selectedPolygon2.VerticeList[indexOfClosestVertex]);
        indexOfClosestVertex := FindIndexOfClosestVertex(polygon1.Vertices[i],polygon2);
        APrimitive.Polygon_AddVertex(newPolygon,selectedPolygon2.VerticeList[indexOfClosestVertex]);

        APrimitive.Polygon_Add(newPolygon);
    end;

    // remove selected polygons from object
    while APrimitive.SelectedPolygons.Count>0  do begin
        APrimitive.Polygon_Remove(APrimitive.SelectedPolygons[0]);
        APrimitive.SelectedPolygons.Delete(0);
    end;

    // free stuff
    FreeAndNil(polygon1);
    FreeAndNil(polygon2);
end;

procedure TPolygonConnector.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Polygon connect';
    Description := 'Connects two selected polygons if they have an equal number of vertices';
end;

procedure TPolygonConnector.DoExecute(AParamList: TParamList);
begin
    try
        DoConnectPolygons(SelectedPrimitives[0]);
    except
        on E: Exception do MessageDlg('TPolygonConnector.Execute: '+E.Message,mtError,[mbOK],0);
    end;
end;

function TPolygonConnector.CanExecute(AParamList: TParamList): boolean;
var canExecuteTool: boolean;
    primitive: TBasePrimitive;
    polygon1,polygon2: TPolygon;
    adjacentPolygons: TPolygonList;
begin
    // we need exactly one selected primitive
    canExecuteTool := SelectedPrimitives.Count=1;
    if not canExecuteTool then
        ErrorMessage := 'Multiple primitives selected. Please select one primitive only.'
    else begin
        primitive := SelectedPrimitives[0];
        // we need exactly two selected polygons in the primitive
        canExecuteTool := primitive.SelectedPolygons.Count=2;
        if not canExecuteTool then
            ErrorMessage := 'Please select two polygons with an equal number of vertices.'
        else begin
            polygon1 := primitive.SelectedPolygons[0];
            polygon2 := primitive.SelectedPolygons[1];
            canExecuteTool := (polygon1.NumVertices=polygon2.NumVertices);
            if not canExecuteTool then
                ErrorMessage := 'Selected polygons must have an equal number of vertices.'
            else begin
                adjacentPolygons := TPolygonList.Create(false,true);
                primitive.Polygon_GetAdjacentPolygons(polygon1,adjacentPolygons,true);
                canExecuteTool := not adjacentPolygons.IsInList(polygon2);
                if not canExecuteTool then
                    ErrorMessage := 'Selected polygons must not be adjacent to eachother.';
                FreeAndNil(adjacentPolygons);
            end;
        end;
    end;

    Result := canExecuteTool;
end;

end.


