{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PolygonConvexionaterTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TPolygonConvexionateFunction = class(TObject)
  private
    fNumOfBadRuns: integer;
    fAllowedRuns: integer;
    fSuccesful: boolean;
    function FindFirstCounterPolygonWindingIndex(APolygonVertexList: TVerticeList; APolygon: TPolygon; AStartIndex: integer): integer;
    function FindFirstPolygonWindingIndex(APolygonVertexList: TVerticeList; APolygon: TPolygon; AStartIndex: integer): integer;
    function FindFirstIndex(APolygonVertexList: TVerticeList; APolygon: TPolygon; AStartIndex: integer; APolygonWinding: boolean) : integer;
    function AreEdgesPolygonWinded(APolygonVertexList: TVerticeList; APolygon: TPolygon; AStartIndex: integer): boolean;
    function CreatePolygon(APolygonVertexList: TVerticeList; APolygon: TPolygon; var AStartIndex: integer): TPolygon;
    procedure Convexionate(APrimitive: TBasePrimitive; APolygonVertexList: TVerticeList; APolygon: TPolygon; AStartIndex: integer; APolygonWinding: boolean);
    procedure DoConvexionatePolygons(APrimitive: TBasePrimitive);
  public
    class procedure Execute(APrimitive: TBasePrimitive);
  end;

  TPolygonConvexionator = class(TPrimitiveTool)
  private
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

// ### TPolygonConvexionateFunction ############################################

// we use a modified earclipping algorithm for convex decomposition;
// we will be searching for 2 consecutive edges making turns in the
// direction of the winding of the polygon vertices, e.g left-turns;
// when found, store the vertices of those edges in a new polygon,
// then consider the next consecutive edges; for as long as those
// edges make left-turns, store the vertices of those edges in the
// new polygon; after adding new vertices, check if the new polygon
// is still convex and if no other vertices of the provided vertex-
// list are inside the new polygon;
// we stop when two new consecutive edges make a right-turn (this
// means the new polygon will not be convex) or if any of the other
// vertices are inside the new convex polygon;
// when we found a convex polygon, we remove the vertices of that
// polygon from the provided vertexlist, except (!) the first and
// the last vertex; then we recursively call the routine again in
// search of the next convex polygon with the provided vertexlist
// of course, the new polygon is added to the provided primitive :)

// TODO: think of better names for these methods

// terminology:
//   - polygon winding: the direction in which vertices of the polygon
//   are ordered; either clockwise or counterclockwise

// determines if the edges at the specified vertex index are ordered in
// the same direction as the polygon itself
function TPolygonConvexionateFunction.AreEdgesPolygonWinded(APolygonVertexList: TVerticeList; APolygon: TPolygon; AStartIndex: integer): boolean;
const EPS=0.001;
var index,tempIndex: integer;
    v0,v1,v2,n: TVertice;
begin
    if (AStartIndex<0) or (AStartIndex>APolygonVertexList.Count) then
        Result := false
    else begin
        index := AStartIndex;
        tempIndex := index-1;
        if tempIndex < 0 then tempIndex := APolygonVertexList.Count-1;
        v0 := APolygonVertexList[tempIndex];
        v1 := APolygonVertexList[index];
        v2 := APolygonVertexList[(index+1) mod APolygonVertexList.Count];
        n := plane_normal(v0,v1,v2);      // normalized vector
        // kind of hack: if 3 vertices are in line, the triagle formed by these
        // vertices has a normal with all components set to (minus) zero; such
        // normal has a length of 0 and we classify this direction as false
        if (n.Length>-EPS) and (n.Length<EPS) then
            Result := true
        else
            Result := dotproduct(n,APolygon.Normal)>EPS;
        FreeAndNil(n);
    end;
end;

// searches for the polygon winded turn in the vertexlist
function TPolygonConvexionateFunction.FindFirstPolygonWindingIndex(APolygonVertexList: TVerticeList; APolygon: TPolygon; AStartIndex: integer): integer;
var index,counter: integer;
    edgesArePolygonWinded: boolean;
begin
    if (AStartIndex<0) or (AStartIndex>APolygonVertexList.Count) then
        Result := -1
    else begin
        index := AStartIndex;
        counter := 0;
        repeat
            edgesArePolygonWinded := AreEdgesPolygonWinded(APolygonVertexList,APolygon,index);
            if not edgesArePolygonWinded then
                index := (index+1) mod APolygonVertexList.Count;
            inc(counter);
        until (edgesArePolygonWinded) or (counter>APolygonVertexList.Count);

        if counter>APolygonVertexList.Count then // if counter>APolygonVertexList.Count, we looped through all vertices
            Result := -1                         // and there is polygon winded turn -> all vertices represent one convex polygon
        else
            Result := index;                     // return index of the middle vertex
    end;
end;

// searches for the counter polygon winded (!) turn in the vertexlist
function TPolygonConvexionateFunction.FindFirstCounterPolygonWindingIndex(APolygonVertexList: TVerticeList; APolygon: TPolygon; AStartIndex: integer): integer;
var index,counter: integer;
    edgesArePolygonWinded: boolean;
begin
    if (AStartIndex<0) or (AStartIndex>APolygonVertexList.Count) then
        Result := -1
    else begin
        index := AStartIndex;
        counter := 0;
        repeat
            edgesArePolygonWinded := AreEdgesPolygonWinded(APolygonVertexList,APolygon,index);
            if edgesArePolygonWinded then
                index := (index+1) mod APolygonVertexList.Count;
            inc(counter);
        until (not edgesArePolygonWinded) or (counter>APolygonVertexList.Count);

        if counter>APolygonVertexList.Count then // if counter>APolygonVertexList.Count, we looped through all vertices
            Result := -1                         // and there is no counter polygon winded turn -> all vertices represent one convex polygon
        else
            Result := index;                     // return index of the middle vertex
    end;
end;

function TPolygonConvexionateFunction.FindFirstIndex(APolygonVertexList: TVerticeList; APolygon: TPolygon; AStartIndex: integer; APolygonWinding: boolean) : integer;
var index,counter: integer;
    edgesArePolygonWinded: boolean;
begin
    if APolygonWinding then
        Result := FindFirstPolygonWindingIndex(APolygonVertexList,APolygon,AStartIndex)
    else begin
        counter := 0;
        index := FindFirstCounterPolygonWindingIndex(APolygonVertexList,APolygon,AStartIndex);
        repeat
            index := (index+1) mod APolygonVertexList.Count;
            edgesArePolygonWinded := AreEdgesPolygonWinded(APolygonVertexList,APolygon,index);
            inc(counter);
        until (edgesArePolygonWinded) or (counter>APolygonVertexList.Count);
        if counter<APolygonVertexList.Count then
            Result := index
        else
            Result := -1;
    end;
end;

function TPolygonConvexionateFunction.CreatePolygon(APolygonVertexList: TVerticeList; APolygon: TPolygon; var AStartIndex: integer): TPolygon;
var tempIndex,i,counter: integer;
    newPolygon: TPolygon;
    edgesArePolygonWinded,polygonHasVerticesInside,doContinue: boolean;
begin
    newPolygon := TPolygon.Create;
    polygonHasVerticesInside := false;
    doContinue := true;
    edgesArePolygonWinded := true;

    counter := 0;
    while (edgesArePolygonWinded) and (not polygonHasVerticesInside) and (doContinue) do begin
        edgesArePolygonWinded := AreEdgesPolygonWinded(APolygonVertexList,APolygon,AStartIndex);
        if edgesArePolygonWinded then begin
            // add vertices to the new polygon; at the first run, all three
            // vertices are added; on all consecutive runs, only the last
            // vertex is added to the polygon
            tempIndex := AStartIndex-1;
            if tempIndex<0 then tempIndex := APolygonVertexList.Count-1;    // we add to the polygon itself here because
            newPolygon.VerticeList.Add(APolygonVertexList[tempIndex]);      // we need the IsPointInPolygon method later
            newPolygon.VerticeList.Add(APolygonVertexList[AStartIndex]);
            newPolygon.VerticeList.Add(APolygonVertexList[(AStartIndex+1) mod APolygonVertexList.Count]);

            // if the last and the first edges of the new polygon make a right
            // turn, the polygon is not convex anymore; therefor we will remove
            // the last vertex so the polygon stays convex
            edgesArePolygonWinded := AreEdgesPolygonWinded(newPolygon.VerticeList,APolygon,0);
            if not edgesArePolygonWinded then
                newPolygon.VerticeList.Delete(newPolygon.VerticeList.Count-1);

            // although the polygon is convex, it might contain one or more
            // of the remaining vertices; we check to see if this is the case
            // and if so, we disqualify this polygon, clear its verticelist
            // and continue searching for a valid, convex polygon
            for i := 0 to APolygonVertexList.Count-1 do begin
                if newPolygon.VerticeList.IndexOf(APolygonVertexList[i])=-1 then begin
                    if newPolygon.IsPointInPolygon(APolygonVertexList[i],0.001,false) then begin
                        polygonHasVerticesInside := true;
                        FreeAndNil(newPolygon);
                        break;
                    end;
                end;
            end;
            AStartIndex := (AStartIndex+1) mod APolygonVertexList.Count;
        end;

        inc(counter);
        doContinue := counter<APolygonVertexList.Count;   // looped through all vertices but didn't find an ear (wrong polyline; e.g edge cross)
    end;
    Result := newPolygon;
end;

procedure TPolygonConvexionateFunction.Convexionate(APrimitive: TBasePrimitive; APolygonVertexList: TVerticeList; APolygon: TPolygon; AStartIndex: integer; APolygonWinding: boolean);
var newPolygon: TPolygon;
    vertexList: TVerticeList;
    index,i: integer;
begin
    index := FindFirstIndex(APolygonVertexList,APolygon,AStartIndex,APolygonWinding);
    if index<>-1 then begin
        newPolygon := CreatePolygon(APolygonVertexList,APolygon,index);

        // if the polygon was convex and none of the remaining vertices are within it...
        if not Assigned(newPolygon) then begin
            // DIRTY TRICK: in a number of cases, the concave polygon can't be
            // decomposed using the algorithm implemented; due to the way the first index
            // is found (using FindFirstIndex), the algorithm might find only (!) polygons
            // that have other vertices inside them, which makes them invalid polygons;
            // this results in an infinite loop and, eventually, a stack overflow because
            // of the recursive nature of the implementation; here, we stop trying when
            // a userdefined number of runs have been executed
            inc(fNumOfBadRuns);
            if fNumOfBadRuns>fAllowedRuns then begin        // APolygonVertexList.Count
                fSuccesful := false;
                fNumOfBadRuns := 0;
                Exit;
            end;
        end else begin
            // feed found vertices to new polygon - this creates new edges as well!
            vertexList := TVerticeList.Create(false,true);
            try
                fNumOfBadRuns := 0;
                vertexList.Assign(newPolygon.VerticeList);  // we need to copy the vertices from the polygon to a temp list
                newPolygon.VerticeList.Clear;               // and clear the polygon itself; after that, vertices are added
                for i := 0 to vertexList.Count-1 do         // using Polygon_AddVertex, which updates the entire datastructure
                    APrimitive.Polygon_AddVertex(newPolygon,vertexList[i]);     // including edges etc
                newPolygon.Material := APolygon.Material;
                APrimitive.Polygon_Add(newPolygon);
            finally
                FreeAndNil(vertexList);
            end;

            // remove the inner-vertices of the new, convex polygon from the
            // provided vertexlist; note: we do not remove the first+last vertices!
            for i := 1 to newPolygon.VerticeList.Count-2 do begin
                APolygonVertexList.Remove(newPolygon.VerticeList[i]);
                dec(index);
                if index<0 then
                    index := APolygonVertexList.Count-1;
            end;
        end;

        // recursive solution; why? because we like recursiveness! :)
        if APolygonVertexList.Count>=3 then
            Convexionate(APrimitive, APolygonVertexList, APolygon, index, APolygonWinding);
    end;
end;

procedure TPolygonConvexionateFunction.DoConvexionatePolygons(APrimitive: TBasePrimitive);
var i,counter: integer;
    polygon: TPolygon;
begin
    with APrimitive do begin
        counter := Polygons.Count-1;
        for i := counter downto 0 do begin
            // TODO: check on polygon.NumVertices < 3

            polygon := APrimitive.Polygons[i];
            fNumOfBadRuns := 0;
            fAllowedRuns := polygon.NumVertices div 3;     // arbitrary number for now
            fSuccesful := true;
            // try first method (finding indexes via counter clockwise order)
            Convexionate(APrimitive,polygon.VerticeList,polygon,0,false);
            if fSuccesful then
                APrimitive.Polygon_Remove(polygon)
            else begin
                // try second method (finding indexes via clockwise order)
                fSuccesful := true;
                Convexionate(APrimitive,polygon.VerticeList,polygon,0,true);
                if fSuccesful then
                    APrimitive.Polygon_Remove(polygon)
            end;
        end;
    end;
end;

class procedure TPolygonConvexionateFunction.Execute(APrimitive: TBasePrimitive);
begin
    with self.Create do
    try
        DoConvexionatePolygons(APrimitive);
    finally
        Free;
    end;
end;

// ### TPolygonConvexionator ###################################################

procedure TPolygonConvexionator.DoPrepare(AParamList: TParamList);
begin
    
end;

procedure TPolygonConvexionator.DoExecute(AParamList: TParamList);
var i: integer;
begin
    try
        for i := 0 to SelectedPrimitives.Count-1 do
            TPolygonConvexionateFunction.Execute(SelectedPrimitives[i]);
    except
        on E: Exception do MessageDlg('TPolygonConvexionator.Execute: '+E.Message,mtError,[mbOK],0);
    end;
end;

end.
