{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PolygonExtruderTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TPolygonExtrudeFunction = class(TObject)
  private
    procedure DoExtrude(APrimitive: TBasePrimitive;
                        AExtrudeNumberOfUnits: integer;
                        AExtrudeByRegion: boolean;
                        ACreateNewPolygonSelection: boolean;
                        AKeepOriginal,
                        AFlipOriginal: boolean);
  public
    class procedure Execute(APrimitive: TBasePrimitive;
                            AExtrudeNumberOfUnits: integer;
                            AExtrudeByRegion: boolean;
                            ACreateNewPolygonSelection: boolean;
                            AKeepOriginal,
                            AFlipOriginal: boolean);
  end;

  TPolygonExtruder = class(TPrimitiveTool)
  private
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  unit_PolygonMoverTool;

// ### TPolygonExtrudeFunction #################################################

procedure TPolygonExtrudeFunction.DoExtrude(APrimitive: TBasePrimitive;
                                            AExtrudeNumberOfUnits: integer;
                                            AExtrudeByRegion: boolean;
                                            ACreateNewPolygonSelection: boolean;
                                            AKeepOriginal,
                                            AFlipOriginal: boolean);
var i,j,index: integer;
    newVertex: TVertice;
    newVertices,processedVertices: TVerticeList;
    newPolygonSelection,adjacentPolygons: TPolygonList;
    polygon,newPolygon,inbetweenPolygon: TPolygon;
    edge: TEdge;
    doCreatePolygon: boolean;
begin
    newPolygonSelection := TPolygonList.Create(false,true);
    adjacentPolygons := TPolygonList.Create(false,true);
    newVertices := TVerticeList.Create(false,true);
    processedVertices := TVerticeList.Create(false,true);

    with APrimitive do begin

        for i := 0 to SelectedPolygons.Count-1 do begin

            polygon := SelectedPolygons[i];

            // get adjacent selected polygons
            Polygon_GetAdjacentPolygons(polygon,adjacentPolygons,true);

            // duplicate the current polygon and add the new polygon to the primitive polygonlist;
            // vertices that are shared between adjacent selected polygons are duplicated only once
            newPolygon := TPolygon.Create;
            newPolygon.Material := polygon.Material;
            for j := 0 to polygon.VerticeList.Count-1 do begin
                if not AExtrudeByRegion then begin
                    // create a new vertex
                    newVertex := polygon.VerticeList[j].Copy;
                    newVertices.Add(newVertex);
                end else begin
                    // check if the current vertex was already processed (when
                    // we processed a previous selected polygon); if so, we use
                    // its equivalent vertex which we stored in the newVertices
                    // list
                    index := processedVertices.IndexOf(polygon.VerticeList[j]);
                    if index<>-1 then
                        newVertex := newVertices[index]
                    else begin
                        newVertex := polygon.VerticeList[j].Copy;
                        newVertices.Add(newVertex);
                        processedVertices.Add(polygon.VerticeList[j]);
                    end;
                end;
                Polygon_AddVertex(newPolygon,newVertex);
                newPolygon.AddCopiedUVCoordinates(polygon,j);
            end;

            // move new polygon by given number of units along its normal
            TPolygonMoveByNormalForSinglePolygon.Execute(polygon, AExtrudeNumberOfUnits);

            // and add new polygon to primitive polygon collection + new selection if wanted
            APrimitive.Polygon_Add(newPolygon);
            if ACreateNewPolygonSelection then
                newPolygonSelection.Add(newPolygon);

            // create new polygons in between the original and duplicate polygons
            for j := 0 to polygon.VerticeList.Count-1 do begin

                edge := Edge_Get(polygon.VerticeList[j],polygon.VerticeList[(j+1) mod polygon.VerticeList.Count]);

                // if performing a region extrude, determine if we need to create a polygon;
                // we do this by checking if the polygonlist of the current edge contains exactly
                // one polygon (which would be the current polygon); if so, create a new polygon;
                // if not, the edge is used by multiple polygons that are being extruded and
                // as we want a Region extrude, we will not transform this edge into a polygon
                doCreatePolygon := false;
                if not AExtrudeByRegion then
                    doCreatePolygon := true
                else begin
                    if edge.PolygonList.IsIn(SelectedPolygons)=1 then
                        doCreatePolygon := true
                end;

                if doCreatePolygon then begin
                    inbetweenPolygon := TPolygon.Create;
                    inbetweenPolygon.Material := polygon.Material;
                    Polygon_AddVertex(inbetweenPolygon,polygon.VerticeList[j]);
                    Polygon_AddVertex(inbetweenPolygon,polygon.VerticeList[(j+1) mod polygon.VerticeList.Count]);
                    Polygon_AddVertex(inbetweenPolygon,newPolygon.VerticeList[(j+1) mod polygon.VerticeList.Count]);
                    Polygon_AddVertex(inbetweenPolygon,newPolygon.VerticeList[j]);
                    APrimitive.Polygon_Add(inbetweenPolygon);
                end;
            end;
        end;

        // handle original polygon as wanted
        for i := 0 to SelectedPolygons.Count-1 do begin

            polygon := SelectedPolygons[i];
            if not AKeepOriginal then
                Polygon_Remove(polygon)
            else begin
                if AFlipOriginal then
                    polygon.Flip;
            end;
        end;

        // create a new polygon selection if wanted
        if ACreateNewPolygonSelection then
            SelectedPolygons.Assign(newPolygonSelection);
    end;

    FreeAndNil(processedVertices);
    FreeAndNil(newVertices);
    FreeAndNil(adjacentPolygons);
    FreeAndNil(newPolygonSelection);
end;

class procedure TPolygonExtrudeFunction.Execute(APrimitive: TBasePrimitive;
                                                AExtrudeNumberOfUnits: integer;
                                                AExtrudeByRegion: boolean;
                                                ACreateNewPolygonSelection: boolean;
                                                AKeepOriginal,
                                                AFlipOriginal: boolean);
begin
    with self.Create do
    try
        DoExtrude(APrimitive, AExtrudeNumberOfUnits, AExtrudeByRegion, ACreateNewPolygonSelection, AKeepOriginal, AFlipOriginal);
    finally
        Free;
    end;
end;

// ### TPolygonExtruder ########################################################

// NOTE: the extrude tool differs from other tools. The actual Extrude is done
// in the Prepare method. The Execute method moves the extruded polygon along
// its normal. 

procedure TPolygonExtruder.DoPrepare(AParamList: TParamList);
var i: integer;
    primitive: TBasePrimitive;
    extrudeNumberOfUnits: integer;
    extrudeByRegion,createNewPolygonSelection,keepOriginal,flipOriginal: boolean;
begin
    Hint := 'Extrude polygons';
    Description := 'Extrudes polygons';

    try
        extrudeNumberOfUnits := TIntegerObject(AParamList.NamedParameters['extrudeNumberOfUnits']).Value;
        extrudeByRegion := TBooleanObject(AParamList.NamedParameters['extrudeByRegion']).Value;
        createNewPolygonSelection := TBooleanObject(AParamList.NamedParameters['createNewPolygonSelection']).Value;
        keepOriginal := TBooleanObject(AParamList.NamedParameters['keepOriginal']).Value;
        flipOriginal := TBooleanObject(AParamList.NamedParameters['flipOriginal']).Value;
        for i := 0 to SelectedPrimitives.Count-1 do begin
            primitive := SelectedPrimitives[i];
            TPolygonExtrudeFunction.Execute(primitive, extrudeNumberOfUnits, extrudeByRegion, createNewPolygonSelection, keepOriginal, flipOriginal);
        end;
    except
        on E: Exception do MessageDlg('TPolygonExtruder.Prepare: '+E.Message,mtError,[mbOK],0);
    end;
end;

procedure TPolygonExtruder.DoExecute(AParamList: TParamList);
var i,numberOfUnits: integer;
    primitive: TBasePrimitive;
    deltaVector: TVertice;
    doMoveByNormal: boolean;
begin
    try
        deltaVector := TVertice(AParamList.NamedParameters['deltaVector']);
        doMoveByNormal := TBooleanObject(AParamList.NamedParameters['moveByNormal']).Value;
        numberOfUnits := TIntegerObject(AParamList.NamedParameters['numberOfUnits']).Value;
        for i := 0 to SelectedPrimitives.Count-1 do begin
            primitive := SelectedPrimitives[i];
            if not doMoveByNormal then
                TPolygonMoveByDeltaVectorForPrimitive.Execute(primitive,deltaVector)
            else begin
                //if primitive.ObjectType=otCube then
                //    TPolygonMoveByNormalForCube.Execute(primitive,deltaVector)
                //else
                    TPolygonMoveByNormalForPrimitive.Execute(primitive,numberOfUnits);

                //TPolygonMoveByNormalForPrimitive.Execute(primitive,numberOfUnits);
            end;
        end;
    except
        on E: Exception do MessageDlg('TPolygonExtruder.Execute: '+E.Message,mtError,[mbOK],0);
    end;
end;


end.

