{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PolygonFlattenerTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TPolygonFlattener = class(TPrimitiveTool)
  private
    procedure DoFlatten;
  protected
  public
    function CanExecute(AParamList: TParamList): boolean; override;
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

// NOTE: this flattener should work with sets of adjacent polygons, much like the polygon merger does
procedure TPolygonFlattener.DoFlatten;
var i,j: integer;
    vertexWorkLoad: TVerticeList;
    planeNormal, averageVertex, deltaVertex: TVertice;
    plane: TPlane;
    moveNumOfUnits: double;
begin
    for i := 0 to SelectedPrimitives.Count-1 do begin

        with SelectedPrimitives[i] do begin

            if ModifyGeometry then begin

                // acquire all vertices to work with
                vertexWorkLoad := TVerticeList.Create(false,true);
                for j := 0 to SelectedPolygons.Count-1 do
                    vertexWorkLoad.AddList(SelectedPolygons[j].VerticeList);

                // calculate the average vertex-normal of all vertices in the selection
                // we need a plane to project the selected vertices upon and this average
                // vertex-normal will be the normal of that plane
                planeNormal := TVertice.Create(0,0,0);
                for j := 0 to SelectedPolygons.Count-1 do
                    planeNormal.Add(SelectedPolygons[j].Normal);
                planeNormal.DivideByScalar(SelectedPolygons.Count);     // NOTE: it is important to this first and normalize afterwards!
                planeNormal.Normalize;

                // calculate the average vertex of the selection; this will be a vertex
                // that lies on the plane calculated earlier; this can be done in the
                // previous loop also, but for clarification, it is done seperately
                averageVertex := TVertice.Create(0,0,0);
                for j := 0 to vertexWorkLoad.Count-1 do
                    averageVertex.Add(vertexWorkLoad[j]);
                averageVertex.DivideByScalar(vertexWorkLoad.Count);

                // create a plane
                plane := TPlane.Create(averageVertex, planeNormal);

                // move each vertex a # of units so that the vertex lies on the plane when done
                for j := 0 to vertexWorkLoad.Count-1 do begin
                    moveNumOfUnits := DistToPlane(vertexWorkLoad[j],plane);
                    deltaVertex := TVertice.Create(planeNormal.x*moveNumOfUnits,planeNormal.y*moveNumOfUnits,planeNormal.z*moveNumOfUnits);
                    vertexWorkLoad[j].Subtract(deltaVertex);
                    FreeAndNil(deltaVertex);
                end;

                // free stuff
                FreeAndNil(plane);
                FreeAndNil(planeNormal);
                FreeAndNil(averageVertex);
                FreeAndNil(vertexWorkLoad);
            end;
        end;
    end;
end;

procedure TPolygonFlattener.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Flatten polygons';
    Description := 'Flattens polygons by their shared normal';
end;

procedure TPolygonFlattener.DoExecute(AParamList: TParamList);
begin
    try
        DoFlatten;
    except
        on E: Exception do MessageDlg('TPolygonFlattener.Execute: '+E.Message,mtError,[mbOK],0);
    end;
end;

function TPolygonFlattener.CanExecute(AParamList: TParamList): boolean;
var canExecuteTool: boolean;
    i: integer;
    primitive: TBasePrimitive;
begin
    canExecuteTool := SelectedPrimitives.Count>0;
    if not canExecuteTool then
        ErrorMessage := 'Please select multiple polygons of one or more primitives.'
    else begin
        for i := 0 to SelectedPrimitives.Count-1 do begin
            primitive := SelectedPrimitives[i];
            canExecuteTool := primitive.SelectedPolygons.Count>1;
            if canExecuteTool then Break;
        end;
        if not canExecuteTool then
            ErrorMessage := 'Please select multiple polygons of one or more primitives.';
    end;
    Result := canExecuteTool;
end;

end.

