{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PolygonMergerTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TPolygonMergeFunction = class(TObject)
  private
    procedure GetNonSharedVertices(primitive: TBasePrimitive; polygon: TPolygon; sharedEdges: TEdgeList; vertexList: TVerticeList);
    procedure DoMergePolygons(APrimitive: TBasePrimitive; ANewPolygons: TPolygonList; ASelectedPolygonsOnly: boolean);
    procedure RemoveRedundantVertices(APrimitive: TBasePrimitive; P1, P2: TPolygon; ANewVertices: TVerticeList);
    function IsVertexRedundant(P1, P2: TPolygon; AVertexList: TVerticeList; AVertexIndex: Integer): Boolean;
    procedure RemoveVertexFromPolygon(APrimitive: TBasePrimitive; APolygon: TPolygon; AVertex: TVertice);
  public
    class procedure Execute(APrimitive: TBasePrimitive; ANewPolygons: TPolygonList; ASelectedPolygonsOnly: boolean);
  end;

  TPolygonMerger = class(TPrimitiveTool)
  private
  protected
  public
    function CanExecute(AParamList: TParamList): boolean; override;
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses unit_BaseTool;

// ### TPolygonMergeFunction ###################################################

{  === GetNonSharedVertices ===

  Returns the vertices of a polygon that are not shared with the other polygon

  @primitive: The primitive we are working with
  @polygon: The polygon who's non-shared vertices we want
  @sharededges: The edges that are shared between the polygons
  @vertexList: The resulting vertices will be added to this list

  The two polygons are connected by a continuous list of edges.
  This routine will put all non-shared vertices on the "vertexList" list, and will
  include either the start-vertex or the end-vertex of the line of shared edges.

  If you call this method twice (for each polygon once) and combine the results,
  you will get all the vertices for the new merged polygon.
}
procedure TPolygonMergeFunction.GetNonSharedVertices(primitive: TBasePrimitive; polygon: TPolygon; sharedEdges: TEdgeList; vertexList: TVerticeList);
var i: integer;
    currentVertex,endVertex: TVertice;
    startEndVertices,edgeVertices: TVerticeList;
begin
    // get all vertices used by the shared edges
    edgeVertices := TVerticeList.Create(false,true);
    for i := 0 to sharedEdges.Count-1 do
    begin
        edgeVertices.Add(sharedEdges[i].StartVertex);
        edgeVertices.Add(sharedEdges[i].EndVertex);
    end;

    // get the start- and end vertices of the shared edges
    startEndVertices := TVerticeList.Create(false,true);
    sharedEdges.GetStartEndVertices(startEndVertices);

    if edgeVertices.Count=polygon.VerticeList.Count then
    begin

        // if all vertices of the current polygon will become part of the newly formed polygon,
        // only add the starting vertex of the start-end vertices in relation to the verticelist
        // of the current polygon

        currentVertex := polygon.VerticeList.GetNext(startEndVertices[0]);
        if currentVertex = startEndVertices[1] then
           vertexList.Add(startEndVertices[0])
        else
           vertexList.Add(startEndVertices[1]);
    end
    else
    begin

        // find out which vertex to use as starting vertex
        currentVertex := polygon.VerticeList.GetNext(startEndVertices[0]); //+1) mod polygon.NumVertices;
        if edgeVertices.IndexOf(currentVertex)<>-1 then
        begin
            currentVertex := startEndVertices[1];
            endVertex     := startEndVertices[0];
        end
        else
        begin
            currentVertex := startEndVertices[0];
            endVertex     := startEndVertices[1];
        end;

        // add vertices to new polygon
        while currentVertex <> endVertex do
        begin
            vertexList.Add(currentVertex);
            currentVertex := polygon.VerticeList.GetNext(currentVertex);
        end;
    end;

    FreeAndNil(startEndVertices);
    FreeAndNil(edgeVertices);
end;

{ === RemoveRedundantVertices ===

This method removes redundant vertices during a merge operation

@APrimitive: Current primitive
@P1, P2: The polygons we want to merge
@ANewVertices: The list of vertices that will make up our new polygon

This routine will remove redundant vertices from ANewVertices
and from the surrounding polygons  }
procedure TPolygonMergeFunction.RemoveRedundantVertices(APrimitive: TBasePrimitive; P1, P2: TPolygon; ANewVertices: TVerticeList);
var
  I, J: Integer;
  adjacentPolygons: TPolygonList;
begin
  I := 0;
  while I < ANewVertices.Count do
    if IsVertexRedundant(P1, P2, ANewVertices, I) then
    begin
      //If the vertex is also part of another polygon,
      //remove it from that polygon aswell
      adjacentPolygons := TPolygonList.Create(false, true);
      try
        ANewVertices[I].GetAdjacent(adjacentPolygons);

        for J := 0 to adjacentPolygons.Count - 1 do
          if (adjacentPolygons[J] <> P1) and
             (adjacentPolygons[J] <> P2) then
             RemoveVertexFromPolygon(APrimitive, adjacentPolygons[J], ANewVertices[I]);
      finally
        adjacentPolygons.Free();
      end;

      //We don't want the redundant vertex in our new polygon
      ANewVertices.Delete(I);
    end
    else
      Inc(I);
end;

{  ====  IsVertexRedundant ====

  This routine checks whether a vertex is redundant in the newly created
  polygon during the merging process.
  A redundant vertex is a vertex that is shared between exactly two edges who are
  co-linear.

  @P1, P2:      The polygons that are to be merged
  @AVertexList: The list of vertices that form our new merged polygon.
                NOTE: The vertices should be ordered in the way the edges connect them
  @AVertexIndex: The index of the vertex in "AVertexList" that we want to check

  If the function returns false, the vertex is not redundant and should be
  added to the polygon
}
function TPolygonMergeFunction.IsVertexRedundant(P1, P2: TPolygon; AVertexList: TVerticeList;
                                                 AVertexIndex: Integer): Boolean;
const
  EPSILON = 1e-2; //A threshold for checking the angle
var
  PolygonEdges:TEdgeList;
  Vtx:     TVertice;
  NextVtx: TVertice;
  PrevVtx: TVertice;
  angle:   Single;
begin
  Result := False;
  NextVtx := nil;
  PrevVtx := nil;

  //Retrieve the vertex we will work with
  Vtx := AVertexList[AVertexIndex];

  PolygonEdges := TEdgeList.Create(false, true);
  try
    //Add all edges of the two polygons to the list
    PolygonEdges.AddList(P1.EdgeList);
    PolygonEdges.AddList(P2.EdgeList);

    //If the vertex is used by an edge that is not part of P1 or P2
    //the edge cannot be removed
    if Vtx.UsedByNumberOfEdges(PolygonEdges) < Vtx.EdgeList.Count then
      Exit;

    //Retrieve vertices adjacent to vtx in the new polygon
    NextVtx := AVertexList.GetNext(Vtx).Copy;
    PrevVtx := AVertexList.GetPrev(Vtx).Copy;

    //Checks if vertex is colinear with it's neighbours
    PrevVtx.Subtract( Vtx );
    NextVtx.Subtract( Vtx );
    Angle := VectorAngle( PrevVtx, NextVtx );
    if Abs(Angle - 180) < EPSILON then
      Result := True;

  finally
    FreeAndNIL(PolygonEdges);
    FreeAndNIL(NextVtx);
    FreeAndNIL(PrevVtx);
  end;
end;

{ === RemoveVertexFromPolygon ===

This routine removes a vertex and it's adjacent edges from a polygon
It will be replaced by a single edge

@APrimitive: Current primitive
@APolygon: The polygon we want to remove the vertex from
@AVertex: The vertex we want to remove}
procedure TPolygonMergeFunction.RemoveVertexFromPolygon(APrimitive: TBasePrimitive; APolygon: TPolygon; AVertex: TVertice);
var
  Neighbours: TVerticeList;
begin
  Neighbours := TVerticeList.Create(false, true);
  try
    with APrimitive do
    begin
      //Find the two neighbouring vertices of AVertex in APolygon
      AVertex.GetAdjacent(Neighbours, APolygon.VerticeList);
      if (Neighbours.Count <> 2) then Exit; //Should never happen!!

      //Remove old edges
      Polygon_RemoveEdge(APolygon, Edge_get(AVertex, Neighbours[0]));
      Polygon_RemoveEdge(APolygon, Edge_get(AVertex, Neighbours[1]));

      //Create new edge
      Polygon_AddEdge(APolygon,     Edge_CreateEdge(Neighbours[0], Neighbours[1]));

      //Remove vertex
      Polygon_RemoveVertex(APolygon, AVertex);
    end;
  finally
    Neighbours.Free();
  end;
end;

procedure TPolygonMergeFunction.DoMergePolygons(APrimitive: TBasePrimitive; ANewPolygons: TPolygonList; ASelectedPolygonsOnly: boolean);
var counter,
    I:                  Integer;
    adjacentPolygon,
    newPolygon:         TPolygon;
    polygonWorkload,
    adjacentPolygons:   TPolygonList;
    sharedEdges:        TEdgeList;
    sharedVertices,
    vertexList:         TVerticeList;
    canMerge,
    edgesAreAdjacent:   Boolean;
begin
    with APrimitive do
    begin

        polygonWorkload := TPolygonList.Create(false,true);
        if ASelectedPolygonsOnly then
            polygonWorkload.Assign(SelectedPolygons)
        else
            polygonWorkload.Assign(Polygons);

        if polygonWorkload.Count > 1 then
        begin
            adjacentPolygons := TPolygonList.Create(false,true);
            sharedEdges := TEdgeList.Create(false,true);

            while (polygonWorkload.Count > 1) do
            begin

                // get any polygons adjacent to the first polygon in our selected polygon list
                Polygon_GetAdjacentPolygons(polygonWorkload[0],adjacentPolygons,ASelectedPolygonsOnly);

                // Only consider polygons that meet the basic requirements
                // That is: coplanar, same material, same normal
                counter := 0;
                while counter<adjacentPolygons.Count do
                begin
                    adjacentPolygon := adjacentPolygons[counter];
                    if (not (polygonWorkload[0].Material = adjacentPolygon.Material)) or
                       (not polygonWorkload[0].IsCoplanar(adjacentPolygon)) or
                       (not polygonWorkload[0].IsEqualNormal(adjacentPolygon)) then
                        adjacentPolygons.Remove(adjacentPolygon)
                    else
                        inc(counter);
                end;

                // if there are adjacent polygons, merge the first one of those adjacent polygons
                // with the current selected polygon
                if adjacentPolygons.Count > 0 then
                begin

                    // find a polygon we can merge with the current selected one; we can merge a polygon
                    // with another polygon if they share a continuous set of edges; also, any vertex that
                    // is shared between the polygons, must be part of the continues set of edges too;
                    // here we find a polygon in our set of adjacent polygons that meets that criteria
                    canMerge := false;
                    counter := -1;
                    repeat
                        inc(counter);
                        if counter<adjacentPolygons.Count then
                        begin
                            Polygon_GetSharedEdges(polygonWorkload[0],adjacentPolygons[counter],sharedEdges,true);
                            edgesAreAdjacent := sharedEdges.isContinuousSet();
                        end else
                            edgesAreAdjacent := false;

                        // all shared vertices between the polygons should be part of the shared edges; we cannot
                        // merge the polygons if they share an individual vertex (a vertex that isn't used by the shared edges)
                        if edgesAreAdjacent then
                        begin
                            sharedVertices := TVerticeList.Create(false,true);
                            try
                                Polygon_GetSharedVertices(polygonWorkload[0],adjacentPolygons[counter],sharedVertices,true);
                                for i := 0 to sharedVertices.Count-1 do
                                begin
                                    canMerge := sharedEdges.IsVertexUsed(sharedVertices[i]);
                                    if not canMerge then   // canMerge is false if a shared vertex isn't used by the shared edges
                                        break;
                                end;
                            finally
                                FreeAndNil(sharedVertices);
                            end;
                        end;

                    until (canMerge) or (counter > adjacentPolygons.Count-1);


                    // if one of the adjacent polygons is adjacent due to a continuous set of adjacent edges
                    // we can savely use that polygon to merge it with the current selected one
                    if canMerge then
                    begin
                        vertexList := TVerticeList.Create(false,true);

                        GetNonSharedVertices(APrimitive,polygonWorkload[0],sharedEdges,vertexList);
                        GetNonSharedVertices(APrimitive,adjacentPolygons[counter],sharedEdges,vertexList);

                        if vertexList.Count>=3 then
                        begin
                            // create a new polygon
                            newPolygon := TPolygon.Create;
                            newPolygon.Material := polygonWorkload[0].Material;

                            // remove redundant vertices (vertices that lie in a straight line)
                            RemoveRedundantVertices(APrimitive, polygonWorkload[0], adjacentPolygons[counter], vertexList);

                            // add vertices to the polygon, updating the internal structure of the primitive
                            for i := 0 to vertexList.Count-1 do
                            begin
                                Polygon_AddVertex(newPolygon,vertexList[i]);

                                //>>>>> Copy vertex UV's from old to new polygon
                                if polygonWorkload[0].VerticeList.IndexOf( vertexList[i] ) <> -1 then
                                begin
                                  //Copy UV coordinate from polygonWorkload[0] to newPolygon
                                  newPolygon.AddCopiedUVCoordinates(polygonWorkload[0],
                                                                    polygonWorkload[0].VerticeList.IndexOf(vertexList[i]) )
                                end
                                else
                                begin
                                  //Copy UV coordinate from adjacentPolygons[counter] to newPolygon
                                  newPolygon.AddCopiedUVCoordinates(adjacentPolygons[counter],
                                                                    adjacentPolygons[counter].VerticeList.IndexOf(vertexList[i]) )
                                end;
                            end;

                            // add our new polygon to the selection because it might be adjacent to
                            // other polygons in our selection; if so, it needs to be processed "again"
                            polygonWorkload.Insert(1,newPolygon);   // insert it at the first position, so it gets processed right away again!

                            // and of course, add it to the polygon collection of this primitive
                            Polygon_Add(newPolygon);

                            // delete original polygons from the primitive
                            Polygon_Remove(polygonWorkload[0]);
                            Polygon_Remove(adjacentPolygons[counter]);

                            if Assigned(ANewPolygons) then
                            begin
                                ANewPolygons.Remove(polygonWorkload[0]);
                                ANewPolygons.Add(newPolygon);
                            end;

                            // remove the adjacent polygon from the selection; it has been replaced
                            // by our newly merged polygon
                            polygonWorkload.Remove(adjacentPolygons[counter]);
                        end;

                        FreeAndNil(vertexList);
                    end;
                end;

                // always remove the polygon currently being processed
                polygonWorkload.Remove(polygonWorkload[0]);
            end;

            // make the merged polygon the selected one
            SelectedPolygons.Clear;
            SelectedPolygons.Add(polygonWorkload[0]);

            // free stuff
            FreeAndNil(sharedEdges);
            FreeAndNil(adjacentPolygons);
        end;
        FreeAndNil(polygonWorkload);
    end;
end;

class procedure TPolygonMergeFunction.Execute(APrimitive: TBasePrimitive; ANewPolygons: TPolygonList; ASelectedPolygonsOnly: boolean);
begin
    with self.Create do
    try
        DoMergePolygons(APrimitive, ANewPolygons, ASelectedPolygonsOnly);
    finally
        Free;
    end;
end;

// ### TPolygonMerger ##########################################################

procedure TPolygonMerger.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Merge polygon';
    Description := 'Merges polygons';
end;

procedure TPolygonMerger.DoExecute(AParamList: TParamList);
var i: integer;
begin
    try
        for i := 0 to SelectedPrimitives.Count-1 do begin
            if SelectedPrimitives[i].SelectedPolygons.Count>1 then
                TPolygonMergeFunction.Execute(SelectedPrimitives[i],nil,true);
        end;
    except
        on E: Exception do MessageDlg('TPolygonMerger.Execute: '+E.Message,mtError,[mbOK],0);
    end;
end;

function TPolygonMerger.CanExecute(AParamList: TParamList): boolean;
var canExecuteTool: boolean;
    i: integer;
    primitive: TBasePrimitive;
begin
    canExecuteTool := SelectedPrimitives.Count>0;
    if not canExecuteTool then
        ErrorMessage := 'Please select multiple polygons of one or more primitives.'
    else begin
        for i := 0 to SelectedPrimitives.Count-1 do begin
            primitive := SelectedPrimitives[i];
            canExecuteTool := primitive.SelectedPolygons.Count>1;
            if canExecuteTool then Break;
        end;
        if not canExecuteTool then
            ErrorMessage := 'Please select multiple polygons of one or more primitives.';
    end;
    Result := canExecuteTool;
end;

end.
