{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PolygonRemoverTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_Primitives,
  unit_ParamList,
  unit_Lights,
  unit_PrimitiveTool,
  unit_PrimitiveList;

type
  TRemoveInvalidPolygonFunction = class(TObject)
  private
    procedure DoRemoveInvalidPolygons(APrimitive: TBasePrimitive);
  public
    class procedure Execute(APrimitive: TBasePrimitive);
  end;

  TPolygonRemover = class(TPrimitiveTool)
  private
    procedure RemoveSelectedPolygons;
    procedure RemoveSystemPolygons;
    procedure RemoveInvalidPolygons;
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses unit_BaseTool;

// ### TRemoveInvalidPolygonFunction ###########################################

procedure TRemoveInvalidPolygonFunction.DoRemoveInvalidPolygons(APrimitive: TBasePrimitive);
var polygonCounter: integer;
    polygon: TPolygon;
    normal: TVertice;
begin
    with APrimitive do begin
        polygonCounter := 0;
        while polygonCounter < NumPolygons do begin
            polygon := Polygons[polygonCounter];
            if polygon.NumVertices<3 then
                // polygon does not have enough vertices - remove it
                Polygon_Remove(polygon)
            else begin
                // polygon has enough vertices - see if it is a degenerate polygon
                normal := Polygons[polygonCounter].Normal;
                if (normal.X = 0) and (normal.Y = 0) and (normal.Z = 0) then
                    Polygon_Remove(polygon)
                else
                    inc(polygonCounter);
            end;
        end;
    end;
end;

class procedure TRemoveInvalidPolygonFunction.Execute(APrimitive: TBasePrimitive);
begin
    with self.Create do
    try
        DoRemoveInvalidPolygons(APrimitive);
    finally
        Free;
    end;
end;

// ### TPolygonRemover #########################################################

procedure TPolygonRemover.RemoveSelectedPolygons;
var counter: integer;
begin
    counter := 0;
    while counter<SelectedPrimitives.Count do begin

        with SelectedPrimitives[counter] do begin

            if ModifyGeometry then begin

                // remove complete object if needed...
                if numPolygons=numSelectedPolygons then begin
                    ScenePrimitives.Remove(SelectedPrimitives[counter]);
                    SelectedPrimitives.Delete(counter);
                end else begin

                    // remove selected polygons from object
                    while SelectedPolygons.Count>0  do begin
                        Polygon_Remove(SelectedPolygons[0]);
                        SelectedPolygons.Delete(0);
                    end;
                    inc(counter);
                end;
            end;
        end;
    end;
end;

procedure TPolygonRemover.RemoveSystemPolygons;
var counter,polygonCounter: integer;
begin
    counter := 0;
    while counter<SelectedPrimitives.Count do begin

        with SelectedPrimitives[counter] do begin

            if (ModifyGeometry) and (not (SelectedPrimitives[counter] is TLight)) then begin

                polygonCounter := 0;

                while polygonCounter<NumPolygons do begin

                    // if system material was assigned, remove the polygon
                    if Polygons[polygonCounter].Material.ID = 0 then
                        Polygon_Remove(Polygons[polygonCounter])
                    else
                        inc(polygonCounter);
                end;

                // if no more polygons are left, remove the object completely
                if NumPolygons=0 then begin
                    ScenePrimitives.Remove(SelectedPrimitives[counter]);
                    SelectedPrimitives.Delete(counter);
                end else
                    inc(counter);
            end else
                inc(counter);
        end;
    end;
end;

// a polygon is invalid when:
//      - it contains less than 3 vertices
//      - its vertices are all lined up colinear (degenerate)
//      - its normal is 0,0,0 (which is the case when vertices are lined up colinear)
procedure TPolygonRemover.RemoveInvalidPolygons;
var primitiveCounter: integer;
    primitive: TBasePrimitive;
begin
    primitiveCounter := 0;
    while primitiveCounter < ScenePrimitives.Count do begin
        primitive := ScenePrimitives[primitiveCounter];

        TRemoveInvalidPolygonFunction.Execute(primitive);

        if primitive.NumPolygons>0 then
            inc(primitiveCounter)
        else begin
            ScenePrimitives.Remove(primitive);
            SelectedPrimitives.Remove(primitive);
        end;
    end;
end;

procedure TPolygonRemover.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Remove polygons';
    Description := 'Removes polygons from primitives';    
end;

procedure TPolygonRemover.DoExecute(AParamList: TParamList);
var polygonRemoveMethod: TPolygonRemoveMethod;
begin
    polygonRemoveMethod := TPolygonRemoveMethod(TIntegerObject(AParamList.NamedParameters['polygonRemoveMethod']).Value);

    case polygonRemoveMethod of
        prmSelected:
            RemoveSelectedPolygons;
        prmSystem:
            RemoveSystemPolygons;
        prmInvalid:
            RemoveInvalidPolygons;
    end;
end;

end.

