{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PolygonSelector;

{$MODE Delphi}

interface

uses
  Classes,
  unit_Selector,
  unit_Types,
  unit_Materials,
  unit_Primitives,
  unit_PrimitiveList;

type
  TPolygonSelector = class(TSelector)
  private
  protected
    function GetCenter: TVertice; override;
  public
    constructor Create(APrimitiveList: TPrimitiveList);

    procedure SelectByRange(AUpperLeftVertex, ALowerRightVertex: TVertice; AClearSelection: boolean = false); overload;
    procedure SelectByRange(AUpperLeftVertex, ALowerRightVertex: TUVCoordinate; AMaterial: TMaterial; AClearSelection: boolean = false); reintroduce; overload;
    procedure SelectByVertex(AVertex: TVertice; ARange: double; AClearSelection: boolean = false);
    procedure SelectByMouse(AVertex: TVertice);
    procedure SelectByUVCoordinate(AUVCoordinate: TUVCoordinate; AMaterial: TMaterial; AClearSelection: boolean = false);
    procedure SelectByRay(ARay: TRay; AAddToSelection: boolean; AClearSelection: boolean = false);
    procedure SelectByRectangle(const ax1,ay1,ax2,ay2: integer;
                                const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                const aViewport: TViewport;
                                const AClearSelection: boolean);
    procedure SelectAll; overload;
    procedure SelectAll(AMaterial: TMaterial); overload;
    procedure SelectByMaterial;
    procedure DeselectAll;
    procedure ReverseSelection;
    procedure GrowSelection;
    procedure ShrinkSelection;
  end;

implementation

uses
  SysUtils, Dialogs,
  unit_Math;

// ### TPolygonSelector ########################################################

constructor TPolygonSelector.Create(APrimitiveList: TPrimitiveList);
begin
    fPrimitives := APrimitiveList;
    inherited Create;
end;

function TPolygonSelector.GetCenter: TVertice;
var i,j,counter: integer;
    primitive: TBasePrimitive;
begin
    counter := 0;
    fCenter.Assign(0,0,0);
    for i := 0 to Primitives.Count-1 do begin
        primitive := Primitives[i];
        for j := 0 to primitive.SelectedPolygons.Count-1 do begin
            fCenter.Add(primitive.SelectedPolygons[j].Center);
            inc(counter);
        end;
    end;
    if counter>0 then
        fCenter.DivideByScalar(counter)
    else
        fCenter.ResetToInfinite;
    Result := fCenter;
end;

procedure TPolygonSelector.SelectByRange(AUpperLeftVertex, ALowerRightVertex: TVertice; AClearSelection: boolean);
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do begin
        if AClearSelection then
            Primitives[i].SelectedPolygons.Clear;
        Primitives[i].Polygon_SelectMultiple(AUpperLeftVertex,ALowerRightVertex);
    end;
end;

procedure TPolygonSelector.SelectByRange(AUpperLeftVertex, ALowerRightVertex: TUVCoordinate; AMaterial: TMaterial; AClearSelection: boolean = false);
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do begin
        if AClearSelection then
            Primitives[i].SelectedPolygons.Clear;
        Primitives[i].Polygon_SelectMultiple(AUpperLeftVertex,ALowerRightVertex,AMaterial);
    end;
end;

procedure TPolygonSelector.SelectByVertex(AVertex: TVertice; ARange: double; AClearSelection: boolean);
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do begin
        if AClearSelection then
            Primitives[i].SelectedPolygons.Clear;
        Primitives[i].Polygon_Select(AVertex);
    end;
end;

procedure TPolygonSelector.SelectByMouse(AVertex: TVertice);
var i,j: integer;
    primitive: TBasePrimitive;
    polygon: TPolygon;
begin
    for i := 0 to Primitives.Count-1 do begin
        primitive := Primitives[i];
        primitive.SelectedPolygons.Clear;
        for j := 0 to primitive.numpolygons-1 do begin    // select polygons based on mousecoords
            polygon := primitive.Polygons[j];
            if polygon.IsPointInFront(AVertex) then begin
                primitive.SelectedPolygons.Add(polygon);  // polygon selected; remember its plane
            end;
        end;
    end;
end;

procedure TPolygonSelector.SelectByUVCoordinate(AUVCoordinate: TUVCoordinate; AMaterial: TMaterial; AClearSelection: boolean);
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do begin
        if AClearSelection then
            Primitives[i].SelectedPolygons.Clear;
        Primitives[i].Polygon_Select(AUVCoordinate, AMaterial);
    end;
end;

procedure TPolygonSelector.SelectByRay(ARay: TRay; AAddToSelection: boolean; AClearSelection: boolean);
var primitive: TBasePrimitive;
    distance: single;
begin
    SelectPolygonByRay(ARay, primitive, distance, AAddToSelection, AClearSelection);
end;

procedure TPolygonSelector.SelectByRectangle(const ax1,ay1,ax2,ay2: integer;
                                             const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                             const aViewport: TViewport;
                                             const AClearSelection: boolean);
var
    primitive    : TBasePrimitive;
    polygon      : TPolygon;
    point        : TPoint2D;
    i,j,k   : integer;
    PolygonInRect: boolean;
    PolygonPoints: TPoint2DArray;
begin
    if AClearSelection then
        DeselectAll;

    for i := 0 to Primitives.Count - 1 do begin
        primitive := Primitives[i];

        if AClearSelection then
            primitive.SelectedPolygons.Clear;

        for j := 0 to primitive.NumPolygons - 1 do begin
            polygon := primitive.Polygons[j];
            polygon.ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewport,PolygonPoints);
            if Polygon2DFacesViewer(PolygonPoints) then begin
                PolygonInRect := true;
                for k := 0 to High(PolygonPoints) do begin
                    point := PolygonPoints[k];
                    if (point.X < ax1) or (point.X > ax2) or (point.Y < ay1) or (point.Y > ay2) then begin
                        PolygonInRect := false;
                        Break;
                    end;
                end;
                if PolygonInRect then primitive.SelectedPolygons.Add(polygon);
            end;
        end;
    end;
end;

procedure TPolygonSelector.SelectAll;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Polygon_SelectAll;
end;

procedure TPolygonSelector.SelectAll(AMaterial: TMaterial);
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Polygon_SelectAll(AMaterial);
end;

procedure TPolygonSelector.SelectByMaterial;
var i: integer;
    selectedPolygons: TPolygonList;
begin
    selectedPolygons := TPolygonList.Create(false,true);
    for i := 0 to Primitives.Count-1 do
        selectedPolygons.AddList(Primitives[i].SelectedPolygons);

    if selectedPolygons.Count<>1 then
        MessageDlg('Please select one polygon.',mtWarning,[mbOK],0)
    else
        SelectAll(selectedPolygons[0].Material);
    FreeAndNil(selectedPolygons);
end;

procedure TPolygonSelector.DeselectAll;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Polygon_DeselectAll;
end;

procedure TPolygonSelector.ReverseSelection;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Polygon_ReverseSelection;
end;

procedure TPolygonSelector.GrowSelection;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Polygon_GrowSelection;
end;

procedure TPolygonSelector.ShrinkSelection;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Polygon_ShrinkSelection;
end;





end.

