{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PolygonTriangulatorTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TPolygonCentriodTriangulateFunction = class(TObject)
  private
    procedure DoCentriodTriangulation(APrimitive: TBasePrimitive; APolygon: TPolygon; var AResultingPolygons: TPolygonList);
  public
    class procedure Execute(APrimitive: TBasePrimitive; APolygon: TPolygon; var AResultingPolygons: TPolygonList);
  end;

  TPolygonEarTriangulateFunction = class(TObject)
  private
    function SameDirection(v1,v2,v3,APolygonNormal: TVertice): boolean;
    procedure DoEarTriangulation(APrimitive: TBasePrimitive; APolygon: TPolygon; var AResultingPolygons: TPolygonList);
  public
    class procedure Execute(APrimitive: TBasePrimitive; APolygon: TPolygon; var AResultingPolygons: TPolygonList);
  end;

  TPolygonTriangulator = class(TPrimitiveTool)
  private
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

// ### TPolygonCentriodTriangulateFunction #####################################

procedure TPolygonCentriodTriangulateFunction.DoCentriodTriangulation(APrimitive: TBasePrimitive; APolygon: TPolygon; var AResultingPolygons: TPolygonList);
var i, uvIndex: integer;
    midVertex: TVertice;
    newPolygon: TPolygon;
begin
    with APrimitive do begin

        if (APolygon.NumVertices>3) then begin

            // calculate middle vertex
            midVertex := TVertice.Create(0,0,0);
            for i := 0 to APolygon.NumVertices-1 do
                midVertex.Add(APolygon.VerticeList[i]);
            midVertex.DivideByScalar(APolygon.NumVertices);
            midVertex := Vertices.AddVertex(midVertex,true);    // add to primitive verticelist
            APolygon.AddUVCoordinatesForVertex(midVertex); // bit of a hack: easiest way to calculate uv coordinates based on original poly
            uvIndex := APolygon.NumVertices;

            // form triangles with consecutive vertices of each edge combined with the middle vertex
            for i := 0 to APolygon.NumVertices-1 do begin
                newPolygon := TPolygon.Create;
                newPolygon.Material := APolygon.Material;
                Polygon_AddVertex(newPolygon,APolygon.VerticeList[i]);
                Polygon_AddVertex(newPolygon,APolygon.VerticeList[(i+1) mod APolygon.NumVertices]);
                Polygon_AddVertex(newPolygon,midVertex);
                newPolygon.AddCopiedUVCoordinates(APolygon, i);
                newPolygon.AddCopiedUVCoordinates(APolygon, (i+1) mod APolygon.NumVertices);
                newPolygon.AddCopiedUVCoordinates(APolygon, uvIndex);
                Polygon_Add(newPolygon);
                if Assigned(AResultingPolygons) then
                    AResultingPolygons.Add(newPolygon);
            end;

            HasChanged := true;
        end;
    end;
end;

class procedure TPolygonCentriodTriangulateFunction.Execute(APrimitive: TBasePrimitive; APolygon: TPolygon; var AResultingPolygons: TPolygonList);
begin
    with self.Create do
    try
        DoCentriodTriangulation(APrimitive, APolygon, AResultingPolygons);
    finally
        Free;
    end;
end;

// ### TPolygonEarTriangulateFunction ##########################################

//The dot product of two vectors is the product of the cosine of the angle
//between them and the magnitude (length) of the two vectors:
//  Dot(A,B) = cos(ang) * len(A) * len(B)
//Thus if A and B were both normalized vectors (length = 1.0), the result
//would be simply cos(ang).

//    angle := arccos(dotproduct(e1,e2)) * (180/pi);

//    Result := angle>90;

function TPolygonEarTriangulateFunction.SameDirection(v1,v2,v3,APolygonNormal: TVertice): boolean;
const EPS=0.001;
var n: TVertice;
begin
    n := plane_normal(v1,v2,v3);
    Result := dotproduct(n,APolygonNormal)>EPS;
    n.Free;
end;

procedure TPolygonEarTriangulateFunction.DoEarTriangulation(APrimitive: TBasePrimitive; APolygon: TPolygon; var AResultingPolygons: TPolygonList);
var vertexIndex,tempIndex,counter, i1,i2,i3: integer;
    newPolygon: TPolygon;
    isEar,continue: boolean;
    p,v1,v2,v3: TVertice;
begin
    if APolygon.NumVertices<4 then exit;

    with APrimitive do begin

        continue := true;

        while (APolygon.NumVertices>3) and (continue) do begin

            vertexIndex := 0;
            continue := true;
            isEar := false;

            while not (isEar) and (continue) do begin

                // get vertices to form a triangle with
                tempIndex := vertexIndex-1;
                if tempIndex<0 then tempIndex := APolygon.numvertices-1;

                i1 := tempIndex;
                i2 := vertexIndex;
                i3 := (vertexIndex+1) mod APolygon.VerticeList.Count;

                v1 := APolygon.VerticeList[i1];
                v2 := APolygon.VerticeList[i2];
                v3 := APolygon.VerticeList[i3];

                // find out if angle is 180 degrees or less (added triangle angles <= 180)
                if not SameDirection(v1,v2,v3,APolygon.Normal) then begin

                    vertexIndex := (vertexIndex+1) mod APolygon.numvertices;
                    continue := vertexIndex>0;                                  // looped through all vertices but didn't find an ear (wrong polyline; e.g edge cross)

                end else begin

                    // check if no other vertex from the original polygon is in this triangle
                    counter := ((vertexIndex+1) mod APolygon.numvertices)+1;
                    isEar := true;
                    while (counter<APolygon.numvertices) and (isEar) do begin
                        p := APolygon.VerticeList[counter];
                        if APolygon.IsPointInTriangle(p,v1,v2,v3) then
                            isEar := false
                        else
                            inc(counter);
                    end;

                    if not isEar then begin
                        vertexIndex := (vertexIndex+1) mod APolygon.NumVertices;
                        continue := vertexIndex>0;                              // looped through all vertices but didn't find an ear (wrong polyline; e.g edge cross)
                    end else begin
                        // we can now create a new APolygon using the triangle vertices

                        newPolygon := TPolygon.Create;
                        newPolygon.Material := APolygon.Material;
                        Polygon_AddVertex(newPolygon,v1);
                        Polygon_AddVertex(newPolygon,v2);
                        Polygon_AddVertex(newPolygon,v3);
                        newPolygon.AddCopiedUVCoordinates(APolygon, i1);
                        newPolygon.AddCopiedUVCoordinates(APolygon, i2);
                        newPolygon.AddCopiedUVCoordinates(APolygon, i3);
                        Polygon_Add(newPolygon);
                        if Assigned(AResultingPolygons) then
                            AResultingPolygons.Add(newPolygon);
                    end;
                end;
            end;

            if continue then begin
                APolygon.VerticeList.Delete(vertexIndex);
                APolygon.DeleteUVCoordinates(vertexIndex);
            end;

        end;

        // final polygon
        newPolygon := TPolygon.Create;
        newPolygon.Material := APolygon.Material;
        Polygon_AddVertex(newPolygon,APolygon.VerticeList[0]);
        Polygon_AddVertex(newPolygon,APolygon.VerticeList[1]);
        Polygon_AddVertex(newPolygon,APolygon.VerticeList[2]);
        newPolygon.AddCopiedUVCoordinates(APolygon, 0);
        newPolygon.AddCopiedUVCoordinates(APolygon, 1);
        newPolygon.AddCopiedUVCoordinates(APolygon, 2);
        Polygon_Add(newPolygon);
        if Assigned(AResultingPolygons) then
            AResultingPolygons.Add(newPolygon);

        HasChanged := true;
    end;
end;

class procedure TPolygonEarTriangulateFunction.Execute(APrimitive: TBasePrimitive; APolygon: TPolygon; var AResultingPolygons: TPolygonList);
begin
    with self.Create do
    try
        DoEarTriangulation(APrimitive, APolygon, AResultingPolygons);
    finally
        Free;
    end;
end;

// ### TPolygonTriangulator ####################################################

procedure TPolygonTriangulator.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Triangulate polygons';
    Description := 'Triangulates polygons';
end;

procedure TPolygonTriangulator.DoExecute(AParamList: TParamList);
var i,j,numberOfPolygonsToProcess: integer;
    primitive: TBasePrimitive;
    polygon: TPolygon;
    doCentriodTriangulation, doSelectionOnly: boolean;
    workloadPolygons, handledPolygons, triangulatedPolygons: TPolygonList;
begin
    try
        doCentriodTriangulation := TBooleanObject(AParamList.NamedParameters['doCentriodTriangulation']).Value;
        doSelectionOnly := TBooleanObject(AParamList.NamedParameters['doSelectionOnly']).Value;
        triangulatedPolygons := TPolygonList.Create(false, true);
        handledPolygons := TPolygonList.Create(false, true);
        for i := 0 to SelectedPrimitives.Count-1 do begin
            primitive := SelectedPrimitives[i];

            if primitive.ModifyGeometry then begin

                if doSelectionOnly then
                    workloadPolygons := primitive.SelectedPolygons
                else
                    workloadPolygons := primitive.Polygons;

                numberOfPolygonsToProcess := workloadPolygons.Count;
                for j := 0 to numberOfPolygonsToProcess-1 do begin
                    polygon := workloadPolygons[j];
                    // NOTE: the triangulation functions directly add the new triangles to the primitive
                    if doCentriodTriangulation then
                        TPolygonCentriodTriangulateFunction.Execute(primitive,polygon,triangulatedPolygons)
                    else
                        TPolygonEarTriangulateFunction.Execute(primitive,polygon,triangulatedPolygons);

                    // mark polygon for removal if we've got resulting polygons
                    if triangulatedPolygons.Count>0then
                        handledPolygons.Add(polygon);
                end;

                // remove the processed polygons from the primitive
                for j := 0 to handledPolygons.Count-1 do
                    primitive.Polygon_Remove(handledPolygons[j]);

                // set the new selection
                primitive.SelectedPolygons.Assign(triangulatedPolygons);

                triangulatedPolygons.Clear;
                handledPolygons.Clear;
            end;
        end;

        FreeAndNil(triangulatedPolygons);
        FreeAndNil(handledPolygons);

    except
        on E: Exception do MessageDlg('TPolygonTriangulator.Execute: '+E.Message,mtError,[mbOK],0);
    end;
    
end;




end.

