{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PrimitiveCatmullClarkSubdividerTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool,
  unit_PrimitiveList;

type
  TPrimitiveCatmullClarkSubdivider = class(TPrimitiveTool)
  private
    function GenerateVertex(AVertex: TVertice; ABasePrimitive: TBasePrimitive): TVertice; overload;
    function GenerateVertex(AEdge: TEdge; ABasePrimitive: TBasePrimitive): TVertice; overload;
    function GenerateVertex(APolygon: TPolygon; ABasePrimitive: TBasePrimitive): TVertice; overload;
    function AddVertexToPrimitive(AVertex: TVertice; APrimitive: TPrimitive): TVertice;
    procedure DoSubdivide(APrimitive: TBasePrimitive; ANewPrimitiveSelection: TPrimitiveList);
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

// ### TPrimitiveCatMullClarkSubdivider ########################################

function TPrimitiveCatMullClarkSubdivider.AddVertexToPrimitive(AVertex: TVertice; APrimitive: TPrimitive): TVertice;
var index: integer;
begin
    AVertex.SimpleRoundValue;
    index := APrimitive.Vertices.InList(AVertex);
    if index<>-1 then begin
        FreeAndNil(AVertex);
        Result := APrimitive.Vertices[index];
    end else begin
        APrimitive.Vertex_Add(AVertex);
        Result := AVertex;
    end;
end;

//The old vertices are repositioned according to the equation:
//
//     Q     2R     S(n-3)
//    --- + ---- + --------
//     n      n        n
//
//where:
//
//    * Q is the average of the new face points surrounding the old vertex,
//    * R is the average of the midpoints of the edges that share the old vertex
//    * S is the old vertex point
//    * n is the number of edges that share the old vertex.
function TPrimitiveCatMullClarkSubdivider.GenerateVertex(AVertex: TVertice; ABasePrimitive: TBasePrimitive): TVertice;
var i: integer;
    adjacentPolygons: TPolygonList;
    vertex,Q,R,S: TVertice;
    n: integer;
begin
    n := AVertex.EdgeList.Count;

    adjacentPolygons := TPolygonList.Create(false);
    AVertex.GetAdjacent(adjacentPolygons);
    Q := TVertice.Create;
    for i := 0 to adjacentPolygons.Count-1 do
        Q.Add(adjacentPolygons[i].Center);
    Q.DivideByScalar(adjacentPolygons.Count);
    Q.DivideByScalar(n);
    FreeAndNil(adjacentPolygons);

    R := TVertice.Create;
    for i := 0 to AVertex.EdgeList.Count-1 do
        R.Add(AVertex.EdgeList[i].CenterVertex);
    R.DivideByScalar(n);
    R.MultiplyByScalar(2);
    R.DivideByScalar(n);

    S := AVertex.Copy;
    S.MultiplyByScalar(n-3,false);
    S.DivideByScalar(n);

    vertex := TVertice.Create;
    vertex.Add(Q);
    vertex.Add(R);
    vertex.Add(S);
    FreeAndNil(Q);
    FreeAndNil(R);
    FreeAndNil(S);
    Result := vertex;
end;

// edge point locations are calculated as the average of the center point of
// the original edge and the average of the center points of the polygons using this edge
function TPrimitiveCatMullClarkSubdivider.GenerateVertex(AEdge: TEdge; ABasePrimitive: TBasePrimitive): TVertice;
var i: integer;
    vertex: TVertice;
    adjacentPolygons: TPolygonList;
begin
    vertex := TVertice.Create;
    adjacentPolygons := TPolygonList.Create(false);
    AEdge.GetAdjacent(adjacentPolygons);
    for i := 0 to adjacentPolygons.Count-1 do
        vertex.Add(adjacentPolygons[i].Center);
    vertex.DivideByScalar(adjacentPolygons.Count);
    vertex.Add(AEdge.CenterVertex);
    vertex.DivideByScalar(2);
    FreeAndNil(adjacentPolygons);
    Result := vertex;
end;

function TPrimitiveCatMullClarkSubdivider.GenerateVertex(APolygon: TPolygon; ABasePrimitive: TBasePrimitive): TVertice;
begin
    Result := APolygon.Center.Copy;
end;

// Cattmull-Clark subdivision scheme implementation; clearly not optimized because
// vertices are calculated more than once, but it is the result that matters for now ;)
procedure TPrimitiveCatMullClarkSubdivider.DoSubdivide(APrimitive: TBasePrimitive; ANewPrimitiveSelection: TPrimitiveList);
var i,j,k,index: integer;
    primitive: TPrimitive;
    originalPolygon,polygon: TPolygon;
    originalVertex,vertex: TVertice;
    originalEdge,previousEdge: TEdge;
    vertexList: TVerticeList;
begin
    with APrimitive do begin
        TBasePrimitive(primitive) := APrimitive.Copy;
        primitive.ID := APrimitive.ID;
        primitive.SnapMethod := smNone;
        primitive.Clear;
        vertexList := TVerticeList.Create(false,true);
        for i := 0 to APrimitive.NumPolygons-1 do begin
            originalPolygon := APrimitive.Polygons[i];
            for j := 0 to originalPolygon.NumVertices-1 do begin

                // get original geometry
                originalVertex := originalPolygon.VerticeList[j];
                originalEdge := Edge_Get(originalVertex,originalPolygon.VerticeList[(j+1) mod originalPolygon.NumVertices]);
                index := j-1;
                if index<0 then index := originalPolygon.NumVertices-1;
                previousEdge := Edge_Get(originalVertex,originalPolygon.VerticeList[index]);

                // create new vertices in the order to which they will be fed
                // to a new polygon; check if a newly created vertex already
                // exists in the new primitive and if so, use that vertex instead
                vertexList.Clear;
                vertex := GenerateVertex(originalVertex,APrimitive);
                vertexList.Add(AddVertexToPrimitive(vertex,primitive));
                vertex := GenerateVertex(originalEdge,APrimitive);
                vertexList.Add(AddVertexToPrimitive(vertex,primitive));
                vertex := GenerateVertex(originalPolygon,APrimitive);
                vertexList.Add(AddVertexToPrimitive(vertex,primitive));
                vertex := GenerateVertex(previousEdge,APrimitive);
                vertexList.Add(AddVertexToPrimitive(vertex,primitive));

                // create a new polygon, providing the newly created vertices to it
                polygon := TPolygon.Create;
                polygon.Material := originalPolygon.Material;
                for k := 0 to vertexList.Count-1 do
                    primitive.Polygon_AddVertex(polygon,vertexList[k]);
                primitive.Polygon_Add(polygon);
            end;
        end;
        FreeAndNil(vertexList);

        // update scene
        ScenePrimitives.AddPrimitive(primitive);
        ANewPrimitiveSelection.Add(primitive);
    end;
end;

procedure TPrimitiveCatmullClarkSubdivider.DoPrepare(AParamList: TParamList);
begin
    
end;

procedure TPrimitiveCatmullClarkSubdivider.DoExecute(AParamList: TParamList);
var i: integer;
    newPrimitiveSelection  : TPrimitiveList;
begin
    newPrimitiveSelection  := TPrimitiveList.Create(false);
    for i := 0 to SelectedPrimitives.Count-1 do begin
        {if SelectedPrimitives[i].ObjectType = otPath then
            newPrimitiveSelection.AddPrimitive(SelectedPrimitives[i])
        else}
            with SelectedPrimitives[i] do begin
                if ModifyGeometry then
                    DoSubdivide(SelectedPrimitives[i],newPrimitiveSelection);
            end;
    end;
    ScenePrimitives.DeletePrimitives(SelectedPrimitives);
    //SelectedPrimitives.Assign(newPrimitiveSelection,false,true,true);
    SelectedPrimitives.Assign(newPrimitiveSelection);
    FreeAndNil(newPrimitiveSelection);
end;


end.
