{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PrimitiveDooSabinSubdividerTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool,
  unit_PrimitiveList;

type
  TPrimitiveDooSabinSubdivider = class(TPrimitiveTool)
  private
    function CalculateOffset(APolygon: TPolygon; ASourcePrimitive: TBasePrimitive): integer;
    procedure GetAdjacentPolygons(ASourcePrimitive: TBasePrimitive; APolygonList: TPolygonList);
    procedure ExtractOriginalGeometry(APolygonList: TPolygonList; ASourcePrimitive, ADestPrimitive: TBasePrimitive);
    procedure RemoveOriginalGeometry(APolygonList: TPolygonList; ASourcePrimitive: TBasePrimitive);
    procedure GenerateVertices(APolygonList: TPolygonList; ASourcePrimitive: TBasePrimitive; AVertexList, ASelectedVerticesList: TVerticeList; ADeltaValue: double);
    procedure GeneratePolygon(AVertex: TVertice; AVertexList: TVerticeList; ASourcePrimitive, ADestPrimitive: TBasePrimitive); overload;
    procedure GeneratePolygon(AEdge: TEdge; AVertexList: TVerticeList; ASourcePrimitive, ADestPrimitive: TBasePrimitive); overload;
    procedure GeneratePolygon(APolygon: TPolygon; AVertexList: TVerticeList; ASourcePrimitive, ADestPrimitive: TBasePrimitive); overload;
    procedure CreatePolygonSelection(APrimitive: TBasePrimitive; ASelectedVerticesList: TVerticeList);
    procedure CreateEdgeSelection(APrimitive: TBasePrimitive; ASelectedVerticesList: TVerticeList);
    procedure DoSubdivide(APrimitive: TBasePrimitive; ARoundingPercentage: double; AEditMode: TEditMode);
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

const EPS = 0.001;

// ### TPrimitiveDooSabinSubdivider ###############################################

// steps:
//   - create a temp primitive containing selected polygons AND their adjacent polygons from the current primitive
//   - create a seperate vertex list based on the geometry of the current primitive
//   - remove the selected polygons and their adjancent polygons from the current primitive
//   - add new polygons to the current primitive using the seperate vertex list and the geometry info of the temp primitive

function TPrimitiveDooSabinSubdivider.CalculateOffset(APolygon: TPolygon; ASourcePrimitive: TBasePrimitive): integer;
var i,indexOfPolygon: integer;
begin
    Result := 0;
    indexOfPolygon := ASourcePrimitive.Polygons.IndexOf(APolygon);
    for i := 0 to indexOfPolygon-1 do
        inc(Result, ASourcePrimitive.Polygons[i].NumVertices);
end;

// gets adjacent polygons of all selected polygons in a primitive by determining
// the polygons adjacent to each vertex (!) of a selected polygon; all vertices
// of all selected polygons have been put into the SelectedVertices collection already
procedure TPrimitiveDooSabinSubdivider.GetAdjacentPolygons(ASourcePrimitive: TBasePrimitive; APolygonList: TPolygonList);
var i: integer;
    polygonList: TPolygonList;
    selectedVertex: TVertice;
begin
    polygonList := TPolygonList.Create(false, true);
    for i := 0 to ASourcePrimitive.SelectedVertices.Count-1 do begin
        selectedVertex := ASourcePrimitive.SelectedVertices[i];
        selectedVertex.GetAdjacent(polygonList);
        APolygonList.AddList(polygonList);
    end;
    FreeAndNil(polygonList);
end;

// extracts each selected polygon AND its adjacent polygons to a temp primitive
// this temp primitive is used to create new subdivided geometry which is added to the current primitive
// we need a temp primitive so we can still obtain adjacency information of geometry even after the real
// geometry (from the current primitive) has already been deleted
procedure TPrimitiveDooSabinSubdivider.ExtractOriginalGeometry(APolygonList: TPolygonList; ASourcePrimitive, ADestPrimitive: TBasePrimitive);
var i,j,index: integer;
    selectedPolygon, polygon: TPolygon;
begin
    // now construct a primitive using these polygons
    ADestPrimitive.Clear;
    for i := 0 to APolygonList.Count-1 do begin
        selectedPolygon := APolygonList[i];
        polygon := TPolygon.Create;
        polygon.Material := selectedPolygon.Material;
        for j := 0 to selectedPolygon.NumVertices-1 do begin
            // if a vertex with the same coordinates already exists, use that one
            // if not, create a copy of the current vertex and add it to the polygon
            index := ADestPrimitive.Vertices.InList(selectedPolygon.VerticeList[j]);
            if index=-1 then
                ADestPrimitive.Polygon_AddVertex(polygon, selectedPolygon.VerticeList[j].Copy)
            else
                ADestPrimitive.Polygon_AddVertex(polygon, ADestPrimitive.Vertices[index]);
        end;
        ADestPrimitive.Polygon_Add(polygon);
    end;
    // add selected polygons of source primitive to destination primitive also
    //ADestPrimitive.SelectedPolygons.Assign(ASourcePrimitive.SelectedPolygons);
end;

procedure TPrimitiveDooSabinSubdivider.RemoveOriginalGeometry(APolygonList: TPolygonList; ASourcePrimitive: TBasePrimitive);
var i: integer;
begin
    // remove polygons from the primitive
    for i := 0 to APolygonList.Count-1 do
        ASourcePrimitive.Polygon_Remove(APolygonList[i]);

    // clear selection
    ASourcePrimitive.SelectedPolygons.Clear;

    for i := 0 to ASourcePrimitive.Polygons.Count-1 do begin
//        if ASourcePrimitive.Polygons[i].NumVertices = 0 then
    end;
end;

// vertices are generated based on the current geometry; to connect new geometry with the existing geometry,
// references to existing vertices are copied into the list as well as newly created vertices
procedure TPrimitiveDooSabinSubdivider.GenerateVertices(APolygonList: TPolygonList; ASourcePrimitive: TBasePrimitive; AVertexList, ASelectedVerticesList: TVerticeList; ADeltaValue: double);
var i,j,index: integer;
    selectedPolygon: TPolygon;
    currentVertex, previousVertex, nextVertex, newVertex, edgeCenter1, edgeCenter2, averageVertex: TVertice;
begin
    // TODO:
    //  - first get all adjacent polygons of selected vertices using procedure defined earlier
    //  - then loop through these adjacent polygons, creating new vertices if needed
    //    - use the selected polygons list to determine if a new vertex should be created
    //      or if an original one should be used

    for i := 0 to APolygonList.Count-1 do begin
        selectedPolygon := APolygonList[i];
        for j := 0 to selectedPolygon.NumVertices-1 do begin
            currentVertex := selectedPolygon.VerticeList[j];
            if not ASourcePrimitive.SelectedVertices.IsInList(currentVertex) then
                // copy reference (!) to original vertex to the list; as this is a vertex
                // from the current primitive, we connect new geometry with the current
                // geometry automatically
                AVertexList.Add(currentVertex)
            else begin
                // collect geometry
                nextVertex := selectedPolygon.VerticeList[(j+1) mod selectedPolygon.NumVertices];
                index := j-1;
                if index<0 then index := selectedPolygon.VerticeList.Count-1;
                previousVertex := selectedPolygon.VerticeList[index];
                edgeCenter1 := ASourcePrimitive.Edge_Get(currentVertex,previousVertex).CenterVertex;
                edgeCenter2 := ASourcePrimitive.Edge_Get(currentVertex,nextVertex).CenterVertex;

                // calculate an average vertex
                averageVertex := TVertice.Create;
                averageVertex.Add(edgeCenter1);
                averageVertex.Add(edgeCenter2);
                averageVertex.Add(selectedPolygon.Center);
                averageVertex.Add(currentVertex);
                averageVertex.DivideByScalar(4);

                if ADeltaValue=1 then begin  // small optimalization
                    AVertexList.Add(averageVertex);
                    ASelectedVerticesList.Add(averageVertex);
                end else begin
                    newVertex := currentVertex.Interpolate(averageVertex,ADeltaValue);
                    FreeAndNil(averageVertex);
                    AVertexList.Add(newVertex);
                    ASelectedVerticesList.Add(newVertex);
                end;
            end;
        end;
    end;
end;

procedure TPrimitiveDooSabinSubdivider.GeneratePolygon(AVertex: TVertice; AVertexList: TVerticeList; ASourcePrimitive, ADestPrimitive: TBasePrimitive);
var i,indexOfVertex: integer;
    adjacentPolygons: TPolygonList;
    orderedEdges: TEdgeList;
    edge: TEdge;
    originalPolygon, previousPolygon, polygon: TPolygon;
    vertexNormal: TVertice;
    polygonVertexList: TVerticeList;
begin
    adjacentPolygons := TPolygonList.Create(false,true);
    polygonVertexList := TVerticeList.Create(false,true);
    AVertex.GetAdjacent(adjacentPolygons);

    // # of adjacent polygons should be 3 or more
    if adjacentPolygons.Count>2 then begin
        orderedEdges := TEdgeList.Create(false,true);
        AVertex.GetOrderedEdges(orderedEdges, true);

        // each edge should be shared by only 2 polygons; if this is not
        // the case, the orderedEdges list will be empty; also, we need
        // at least 3 edges to work with, so check these conditions here
        if orderedEdges.Count<>2 then begin
            previousPolygon := nil;
            originalPolygon := nil;

            for i := 0 to orderedEdges.Count-1 do begin
                edge := orderedEdges[i];

                if edge.PolygonList.Count=1 then
                    originalPolygon := edge.PolygonList[0]
                else begin
                    if edge.PolygonList[0]<>previousPolygon then
                        originalPolygon := edge.PolygonList[0]
                    else
                        originalPolygon := edge.PolygonList[1];
                end;

                indexOfVertex := CalculateOffset(originalPolygon,ASourcePrimitive)+originalPolygon.VerticeList.IndexOf(AVertex);
                polygonVertexList.Add(AVertexList[indexOfVertex]);
                previousPolygon := originalPolygon;
            end;

            if polygonVertexList.Count>2 then begin
                polygon := TPolygon.Create;
                polygon.Material := originalPolygon.Material;   // material of last known polygon is used
                for i := 0 to polygonVertexList.Count-1 do
                    ADestPrimitive.Polygon_AddVertex(polygon,polygonVertexList[i]);
                ADestPrimitive.Polygon_Add(polygon);

                // get the normal of this vertex
                vertexNormal := AVertex.GetNormal();

                // new polygon should point in the same direction as the original vertex;
                // if this is not the case, flip the polygon
                if DotProduct(vertexNormal,polygon.Normal) < -EPS then
                    polygon.Flip;
                FreeAndNil(vertexNormal);
            end;
        end;
        FreeAndNil(orderedEdges);
    end;

    FreeAndNil(polygonVertexList);
    FreeAndNil(adjacentPolygons);
end;

procedure TPrimitiveDooSabinSubdivider.GeneratePolygon(AEdge: TEdge; AVertexList: TVerticeList; ASourcePrimitive, ADestPrimitive: TBasePrimitive);
var i: integer;
    adjacentPolygons: TPolygonList;
    originalPolygon, polygon: TPolygon;
    indexOfVertex,indexOfFirstVertexInPolygon: integer;
    polygonVertexList: TVerticeList;
begin
    adjacentPolygons := TPolygonList.Create(false,true);
    polygonVertexList := TVerticeList.Create(false,true);

    // adjacent polygons from temp primitive
    AEdge.GetAdjacent(adjacentPolygons);

    // # of adjacent polygons should be 2
    if adjacentPolygons.Count=2 then begin

        // add vertices from the first adjacent polygon
        originalPolygon := adjacentPolygons[0];
        indexOfFirstVertexInPolygon := CalculateOffset(originalPolygon,ASourcePrimitive);
        indexOfVertex := indexOfFirstVertexInPolygon+originalPolygon.VerticeList.IndexOf(AEdge.EndVertex);  // edge from temp primitive
        polygonVertexList.Add(AVertexList[indexOfVertex]);
        indexOfVertex := indexOfFirstVertexInPolygon+originalPolygon.VerticeList.IndexOf(AEdge.StartVertex);
        polygonVertexList.Add(AVertexList[indexOfVertex]);

        // add vertices from the second adjacent polygon
        originalPolygon := adjacentPolygons[1];
        indexOfFirstVertexInPolygon := CalculateOffset(originalPolygon,ASourcePrimitive);
        indexOfVertex := indexOfFirstVertexInPolygon+originalPolygon.VerticeList.IndexOf(AEdge.StartVertex);
        polygonVertexList.Add(AVertexList[indexOfVertex]);
        indexOfVertex := indexOfFirstVertexInPolygon+originalPolygon.VerticeList.IndexOf(AEdge.EndVertex);
        polygonVertexList.Add(AVertexList[indexOfVertex]);

        if polygonVertexList.Count>2 then begin
            polygon := TPolygon.Create;
            polygon.Material := originalPolygon.Material;   // material of last known polygon is used
            for i := 0 to polygonVertexList.Count-1 do
                ADestPrimitive.Polygon_AddVertex(polygon,polygonVertexList[i]);
            ADestPrimitive.Polygon_Add(polygon);
        end;
    end;

    FreeAndNil(polygonVertexList);
    FreeAndNil(adjacentPolygons);
end;

procedure TPrimitiveDooSabinSubdivider.GeneratePolygon(APolygon: TPolygon; AVertexList: TVerticeList; ASourcePrimitive, ADestPrimitive: TBasePrimitive);
var i,indexOfVertex: integer;
    polygon: TPolygon;
begin
    indexOfVertex := CalculateOffset(APolygon,ASourcePrimitive);
    polygon := TPolygon.Create;
    polygon.Material := APolygon.Material;
    for i := 0 to APolygon.NumVertices-1 do
        ADestPrimitive.Polygon_AddVertex(polygon,AVertexList[indexOfVertex+i]);
    ADestPrimitive.Polygon_Add(polygon);
end;

procedure TPrimitiveDooSabinSubdivider.CreatePolygonSelection(APrimitive: TBasePrimitive; ASelectedVerticesList: TVerticeList);
var i,counter: integer;
    selectedPolygonsList: TPolygonList;
    polygon: TPolygon;
    isInList: boolean;
begin
    selectedPolygonsList := TPolygonList.Create(false,true);

    ASelectedVerticesList.GetAdjacent(selectedPolygonsList);
    // filter out all polygons that do not have all their vertices
    // present in the selectedVerticesList
    counter := 0;
    while counter < selectedPolygonsList.Count do begin
        polygon := selectedPolygonsList[counter];
        isInList := true;
        for i := 0 to polygon.NumVertices-1 do begin
            if not ASelectedVerticesList.IsInList(polygon.VerticeList[i]) then
                isInList := false;      // we could stop traversing the vertices right now...
        end;
        if not isInList then
            selectedPolygonsList.Remove(polygon)
        else
            inc(counter);
    end;
    APrimitive.SelectedPolygons.Assign(selectedPolygonsList);
    FreeAndNil(selectedPolygonsList);
end;

procedure TPrimitiveDooSabinSubdivider.CreateEdgeSelection(APrimitive: TBasePrimitive; ASelectedVerticesList: TVerticeList);
var counter: integer;
    selectedEdgesList: TEdgeList;
    edge: TEdge;
begin
    selectedEdgesList := TEdgeList.Create(false,true);
    ASelectedVerticesList.GetAdjacent(selectedEdgesList);
    // filter out all edges that do not have all their vertices
    // present in the selectedVerticesList
    counter := 0;
    while counter < selectedEdgesList.Count do begin
        edge := selectedEdgesList[counter];
        if (not ASelectedVerticesList.IsInList(edge.StartVertex)) or (not ASelectedVerticesList.IsInList(edge.EndVertex)) then
            selectedEdgesList.Remove(edge)
        else
            inc(counter);
    end;
    APrimitive.SelectedEdges.Assign(selectedEdgesList);
    FreeAndNil(selectedEdgesList);
end;

procedure TPrimitiveDooSabinSubdivider.DoSubdivide(APrimitive: TBasePrimitive; ARoundingPercentage: double; AEditMode: TEditMode);
var i: integer;
    vertexList, selectedVerticesList: TVerticeList;
    tempPrimitive: TBasePrimitive;
    polygonList: TPolygonList;
begin
    // this vertex list doesn't have to be the owner because its vertices
    // (references to vertices) will be copied to the primitive anyway and
    // the primitive is the actual owner of the vertices (both new and old)
    vertexList := TVerticeList.Create(false,false);
    selectedVerticesList := TVerticeList.Create(false,true);
    polygonList := TPolygonList.Create(false,true);

    tempPrimitive := TBasePrimitive.Create;

    // determine which polygons are affected; this is done by traversing the
    // selected vertices (!) and getting each adjacent polygon of a selected vertex
    GetAdjacentPolygons(APrimitive, polygonList);

    // generate vertices based on the current (unaltered) geometry of the
    // primitive; the list will contain both newly created vertices and
    // references to vertices of the current primitive (for connectivity);
    // the selectedVerticesList contains the newly created vertices so we
    // can use them to make a new selection of items (polygons, edges or
    // vertices) when we're done with smoothing
    GenerateVertices(polygonList, APrimitive, vertexList, selectedVerticesList, ARoundingPercentage/100);

    // clear the selected vertices list
    APrimitive.SelectedVertices.Clear;

    // extract all selected polygons and their adjacent polygons to a
    // temp primitive; we use the adjacency information stored in this
    // primitive to construct and add new polygons to the current primitive
    // using the vertex list created earlier in this procedure
    ExtractOriginalGeometry(polygonList, APrimitive, tempPrimitive);

    // generate new geometry and add it to the current primitive
    for i := 0 to tempPrimitive.NumPolygons-1 do
        GeneratePolygon(tempPrimitive.Polygons[i],vertexList,tempPrimitive,APrimitive);

    for i := 0 to tempPrimitive.NumEdges-1 do
        GeneratePolygon(tempPrimitive.Edges[i],vertexList,tempPrimitive,APrimitive);

    for i := 0 to tempPrimitive.NumVertices-1 do
        GeneratePolygon(tempPrimitive.Vertices[i],vertexList,tempPrimitive,APrimitive);

    // remove original geometry
    RemoveOriginalGeometry(polygonList, APrimitive);

    // create a new selection of items (polygons, edges or vertices)
    // based on the newly created vertices
    
    case AEditMode of
        emPolygon:
            CreatePolygonSelection(APrimitive,selectedVerticesList);
        emEdge:
            CreateEdgeSelection(APrimitive,selectedVerticesList);
        emVertex:
            APrimitive.SelectedVertices.Assign(selectedVerticesList);
    end;

    // free stuff and exit
    FreeAndNil(tempPrimitive);
    FreeAndNil(polygonList);
    FreeAndNil(selectedVerticesList);
    FreeAndNil(vertexList);
end;

procedure TPrimitiveDooSabinSubdivider.DoPrepare(AParamList: TParamList);
begin
    
end;

procedure TPrimitiveDooSabinSubdivider.DoExecute(AParamList: TParamList);
var i,roundingPercentage: integer;
    editMode: TEditMode;
begin
    roundingPercentage := TIntegerObject(AParamList.NamedParameters['dooSabinRounding']).Value;
    editMode := TEditMode(TIntegerObject(AParamList.NamedParameters['editMode']).Value);

    for i := 0 to SelectedPrimitives.Count-1 do begin
        //if SelectedPrimitives[i].ObjectType <> otPath then
            with SelectedPrimitives[i] do begin
                if ModifyGeometry then begin
                    // setup selected vertices collection based on editmode
                    case editMode of
                        emObject:
                            Vertex_SelectAll;
                        emPolygon:
                            Vertex_SelectedPolygonsToSelectedVertices;
                        emEdge:
                            Vertex_SelectedEdgesToSelectedVertices;
                    end;

                    // perform subdivision on selected vertices
                    DoSubDivide(SelectedPrimitives[i],roundingPercentage,editMode);

                    HasChanged := true;
                end;
            end;
    end;
end;

end.
