{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PrimitiveHollowerTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_Primitives,
  unit_ParamList,
  unit_Lights,
  unit_PrimitiveTool,
  unit_PrimitiveList;

type
  TPrimitiveHollower = class(TPrimitiveTool)
  private
    procedure HollowByUnit(ANumberOfUnits: integer);
    procedure HollowByPercentage(APercentage: integer);
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  unit_PrimitiveMergerTool,
  unit_PrimitiveScalerTool,
  unit_PolygonExtruderTool;

// ### TPrimitiveHollower ######################################################

procedure TPrimitiveHollower.HollowByUnit(ANumberOfUnits: integer);
var i,j,k: integer;
    currentPrimitive,primitive: TBasePrimitive;
    polygon: TPolygon;
    newSelection: TPrimitiveList;
begin
    newSelection := TPrimitiveList.Create(false);

    for i := 0 to SelectedPrimitives.Count-1 do begin

        currentPrimitive := SelectedPrimitives[i];

        // transform each polygon into a new primitive
        for j := 0 to currentPrimitive.Polygons.Count-1 do begin

            primitive := TUserDefined.CreateUserDefined;

            // create verticelist for new primitive
            for k := 0 to currentPrimitive.Polygons[j].NumVertices-1 do
                primitive.Vertices.Add(currentPrimitive.Polygons[j].VerticeList[k].Copy);

            // add one polygon to the new primitive
            polygon := TPolygon.Create;
            polygon.Material := currentPrimitive.Polygons[j].Material;
            for k := 0 to primitive.Vertices.Count-1 do
                primitive.Polygon_AddVertex(polygon,primitive.Vertices[k]);
            primitive.Polygon_Add(polygon);

            // extrude the polygon
            primitive.SelectedPolygons.Add(polygon);
            TPolygonExtrudeFunction.Execute(primitive,ANumberOfUnits,false,false,true,true);
            primitive.SelectedPolygons.Clear;

            // should not be needed...
            for k := 0 to primitive.Polygons.Count-1 do
                primitive.Polygons[k].Flip;

            // and add the primitive to the scene
            ScenePrimitives.AddPrimitive(primitive,true,true);

            primitive.HasChanged := true;

            // add to new selection of objects
            newSelection.Add(primitive);
        end;
    end;

    // remove selected items from the scene
    ScenePrimitives.RemoveList(SelectedPrimitives);

    // set the new selection
    SelectedPrimitives.Assign(newSelection);

    // add ID's for Undo purposes
    PrimitiveIDList.Clear;
    for i := 0 to newSelection.Count-1 do
        PrimitiveIDList.Add(newSelection[i].ID);

    FreeAndNil(newSelection);
end;

procedure TPrimitiveHollower.HollowByPercentage(APercentage: integer);
var i,j: integer;
    deltaVector: TVertice;
    currentPrimitive,clonedPrimitive,hollowedPrimitive: TBasePrimitive;
    list: TPrimitiveList;
    newSelection: TPrimitiveList;
    //primitiveScaler: TPrimitiveScaler;
begin
    newSelection := TPrimitiveList.Create(false);

    for i := 0 to SelectedPrimitives.Count-1 do begin

        currentPrimitive := SelectedPrimitives[i];

        if (currentPrimitive.ModifyGeometry) and not (currentPrimitive is TLight) then begin
            // clone object
            clonedPrimitive := currentPrimitive.Copy;

            // scale clone by given percentage (percentage should be <= 100)
            deltaVector := TVertice.Create(APercentage/100,APercentage/100,APercentage/100);
            TPrimitiveScaleByDeltaVectorFunction.Execute(clonedPrimitive,deltaVector);
            FreeAndNil(deltaVector);

            // flip polygons of cloned object
            for j := 0 to clonedPrimitive.Polygons.Count-1 do
                clonedPrimitive.Polygons[j].Flip;

            // merge clone with original
            list := TPrimitiveList.Create(false);
            list.Add(currentPrimitive);
            list.Add(clonedPrimitive);
            TPrimitiveMergerByList.Execute(list,hollowedPrimitive);
            FreeAndNil(list);

            // add the primitive to the scene
            ScenePrimitives.AddPrimitive(hollowedPrimitive,true,true);

            // and save it for our new selection later
            newSelection.Add(hollowedPrimitive);

            FreeAndNil(clonedPrimitive);
        end;
    end;

    // remove selected items from the scene
    ScenePrimitives.RemoveList(SelectedPrimitives);

    // set the new selection
    SelectedPrimitives.Assign(newSelection);

    // add ID's for Undo purposes
    PrimitiveIDList.Clear;
    for i := 0 to newSelection.Count-1 do
        PrimitiveIDList.Add(newSelection[i].ID);

    FreeAndNil(newSelection);
end;

procedure TPrimitiveHollower.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Hollow primitive';
    Description := 'Hollows primitives';
end;

procedure TPrimitiveHollower.DoExecute(AParamList: TParamList);
var doHollowByUnits: boolean;
    integerValue: integer;
begin
    doHollowByUnits := TBooleanObject(AParamList.GetObjectByName('doHollowByUnits')).Value;
    integerValue := TIntegerObject(AParamList.GetObjectByName('integerValue')).Value;

    if doHollowByUnits then
        HollowByUnit(integerValue)
    else
        HollowByPercentage(integerValue);
end;


end.

