{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PrimitiveLatherTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_Primitives,
  unit_ParamList,
  unit_Lights,
  unit_PrimitiveTool,
  unit_PrimitiveList;

type
  TPrimitiveLather = class(TPrimitiveTool)
  private
    procedure DoLathe(APrimitive: TBasePrimitive; ANumberOfSegments: integer; AAmountOfDegrees: double);
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

// ### TPrimitiveLather ########################################################

procedure TPrimitiveLather.DoLathe(APrimitive: TBasePrimitive; ANumberOfSegments: integer; AAmountOfDegrees: double);
var i,j,numOriginalVertices: integer;
    edge: TEdge;
    axis,vertex,newVertex: TVertice;
    originalVertices,tempVertices: TVerticeList;
    polygon: TPolygon;
    rotationMatrix: TMatrix;
    rotationStep,currentRotation,nx,ny,nz,ox,oy,oz: double;
//    Vertex1: TVertice;
//    Vertex2: TVertice;
    DeleteEdge: Boolean;
begin
    with APrimitive do begin
        // only continue if object consists of a single polygon
        if NumPolygons = 1 then begin

            DeleteEdge := false;

            // use last/selected edge as axis
            if NumSelectedEdges=0 then begin
                {if APrimitive.ObjectType = otPath then begin
                    Vertex1 := Polygons.Polygons[0].VerticeList[Polygons.Polygons[0].VerticeList.Count - 1];
                    Vertex2 := Polygons.Polygons[0].VerticeList[0];
                    edge := TEdge.Create(Vertex1,Vertex2);
                    DeleteEdge := true;
                end
                else}
                    edge := TEdge(Edges.Last)
            end
            else begin
                edge := SelectedEdges[0];
                SelectedEdges.Clear;
            end;
            axis := edge.DirectionVector;

            // create temporary lists for vertices and initialize them
            tempVertices := TVerticeList.Create(false,true);
            originalVertices := TVerticeList.Create(false,true);
            originalVertices.Assign(Polygons[0].VerticeList);

            // calculate rotation variables and set other variables
            rotationStep := AAmountOfDegrees / ANumberOfSegments;
            currentRotation := rotationStep;
            numOriginalVertices := NumVertices;

            for i := 0 to ANumberOfSegments-1 do begin

                rotationMatrix := MatrixArbAxisSetup(axis, edge.CenterVertex, currentRotation);
                tempVertices.Clear;

                if currentRotation=360 then begin
                    // when rotating 360 degrees, we can use the original vertices again so we create a closed object
                    for j := 0 to numOriginalVertices-1 do
                        tempVertices.Add(Polygons[0].VerticeList[j]);
                end else begin
                    // rotate original vertices + create and add new ones + do not duplicate vertices of axis-edge
                    for j := 0 to numOriginalVertices-1 do begin
                        vertex := Polygons[0].VerticeList[j];           // IMPORTANT: use the verticelist of the polygon; that's the only correct vertex-order (instead of using the verticelist of the object)!
                        if (vertex=edge.StartVertex) or (vertex=edge.EndVertex) then
                            tempVertices.Add(vertex)
                        else begin
                            ox := vertex.x-rotationMatrix[3,0];
                            oy := vertex.y-rotationMatrix[3,1];
                            oz := vertex.z-rotationMatrix[3,2];
                            nx := (ox*rotationMatrix[0,0])+(oy*rotationMatrix[0,1])+(oz*rotationMatrix[0,2])+rotationMatrix[3,0];
                            ny := (ox*rotationMatrix[1,0])+(oy*rotationMatrix[1,1])+(oz*rotationMatrix[1,2])+rotationMatrix[3,1];
                            nz := (ox*rotationMatrix[2,0])+(oy*rotationMatrix[2,1])+(oz*rotationMatrix[2,2])+rotationMatrix[3,2];
                            newVertex := TVertice.Create(nx,ny,nz);
                            Vertex_Add(newVertex);
                            tempVertices.Add(newVertex);
                        end;
                    end;
                end;

                // create polygons between the vertices
                for j := 0 to numOriginalVertices-1 do begin
                    if j<>Polygons[0].EdgeList.IndexOf(edge) then begin
                        polygon := TPolygon.Create;
                        polygon.Material := Polygons[0].Material;
                        Polygon_AddVertex(polygon,originalVertices[j]);
                        Polygon_AddVertex(polygon,originalVertices[(j+1) mod numOriginalVertices]);
                        Polygon_AddVertex(polygon,tempVertices[(j+1) mod numOriginalVertices]);
                        Polygon_AddVertex(polygon,tempVertices[j]);
                        Polygon_Add(polygon);
                    end;
                end;

                // update variables for next run
                originalVertices.Assign(tempVertices);
                FreeAndNil(rotationMatrix);
                currentRotation := currentRotation + rotationStep;
            end;

            if DeleteEdge then edge.Free;

            FreeAndNil(tempVertices);
            FreeAndNil(originalVertices);

            // delete original (first) polygon
            Polygon_Remove(Polygons[0]);

            HasChanged := true;
        end;
    end;
end;

procedure TPrimitiveLather.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Lathe path or polyline';
    Description := 'Lathes path or polyline around selected or last edge';
end;

procedure TPrimitiveLather.DoExecute(AParamList: TParamList);
var i: integer;
    primitive: TBasePrimitive;
    numberOfSegments,amountOfDegrees: integer;
    hasMultiplePolygons: boolean;
begin
    numberOfSegments := TIntegerObject(AParamList.NamedParameters['numberOfSegments']).Value;
    amountOfDegrees := TIntegerObject(AParamList.NamedParameters['amountOfDegrees']).Value;

    // first do a quick check to see if Lathe is used on 1-polygon primitives only
    hasMultiplePolygons := false;
    for i := 0 to SelectedPrimitives.Count-1 do begin
        primitive := SelectedPrimitives[i];
        if primitive.NumPolygons<>1 then
            hasMultiplePolygons := true;
    end;

    if hasMultiplePolygons then
        MessageDlg('Use Lathe on objects that have only one polygon.', mtWarning, [mbOK],0)
    else begin
        // now do the actual Lathe
        for i := 0 to SelectedPrimitives.Count-1 do begin
            primitive := SelectedPrimitives[i];
            if primitive.ModifyGeometry then
                DoLathe(primitive,numberOfSegments,amountOfDegrees);
        end;
    end;
end;

end.

