{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PrimitiveMoverTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_Primitives,
  unit_ParamList,
  unit_Lights,
  unit_PrimitiveTool,
  unit_PrimitiveList;

type
  TPrimitiveMoveFunction = class(TObject)
  private
    procedure DoMove(APrimitive: TBasePrimitive; ADeltaVector: TVertice);
  public
    class procedure Execute(APrimitive: TBasePrimitive; ADeltaVector: TVertice); overload;
    class procedure Execute(APrimitive: TBasePrimitive; x,y,z: double); overload;
  end;

  TPrimitiveMoveToFunction = class(TObject)
  private
    procedure DoMoveTo(APrimitive: TBasePrimitive; x,y,z: double);
  public
    class procedure Execute(APrimitive: TBasePrimitive; x,y,z: double);
  end;

  TPrimitiveMover = class(TPrimitiveTool)
  private
  protected
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses unit_BaseTool,
     unit_Paths,
     unit_Vectors;

// ### TPrimitiveMoveFunction ##################################################

procedure TPrimitiveMoveFunction.DoMove(APrimitive: TBasePrimitive; ADeltaVector: TVertice);
var i: integer;
begin
    with APrimitive do begin
        for i := 0 to NumVertices-1 do begin
            Vertices[i].Subtract(ADeltaVector);
        end;

        if APrimitive is TLight then
            TLight(APrimitive).Position.Subtract(ADeltaVector);

        HasChanged := true;
    end;
end;

class procedure TPrimitiveMoveFunction.Execute(APrimitive: TBasePrimitive; ADeltaVector: TVertice);
begin
    with self.Create do
    try
        DoMove(APrimitive, ADeltaVector);
    finally
        Free;
    end;
end;

class procedure TPrimitiveMoveFunction.Execute(APrimitive: TBasePrimitive; x,y,z: double);
var deltaVector: TVertice;
begin
    with self.Create do
    try
        deltaVector := TVertice.Create(x,y,z);
        DoMove(APrimitive, deltaVector);
    finally
        FreeAndNil(deltaVector);
        Free;
    end;
end;

// ### TPrimitiveMoveToFunction ################################################

procedure TPrimitiveMoveToFunction.DoMoveTo(APrimitive: TBasePrimitive; x,y,z: double);
var i: integer;
    deltaVector: TVertice;
begin
    with APrimitive do begin
        deltaVector := TVertice.Create(Center.x-x,Center.y-y,Center.z-z);

        for i := 0 to NumVertices-1 do
            Vertices[i].Subtract(deltaVector);

        if APrimitive is TLight then
            TLight(APrimitive).Position.Subtract(deltaVector);

        FreeAndNil(deltaVector);

        HasChanged := true;
    end;
end;

class procedure TPrimitiveMoveToFunction.Execute(APrimitive: TBasePrimitive; x,y,z: double);
begin
    with self.Create do
    try
        DoMoveTo(APrimitive, x, y, z);
    finally
        Free;
    end;
end;

// ### TPrimitiveMover #########################################################

procedure TPrimitiveMover.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Move primitives';
    Description := 'Moves primitives in 3D space';
end;

procedure TPrimitiveMover.DoExecute(AParamList: TParamList);
var i,j: integer;
    primitive: TBasePrimitive;
    deltaVector,usedVector: TVertice;
    doMoveRelative: boolean;
    path: TPath;
    cp: TPathControlPoint;
begin
    deltaVector := TVertice(AParamList.GetObjectByName('deltaVector'));
    if AParamList.HasObject('moveRelative') then
        doMoveRelative := TBooleanObject(AParamList.GetObjectByName('moveRelative')).Value
    else
        doMoveRelative := false;

    for i := 0 to SelectedPrimitives.Count-1 do begin

        primitive := SelectedPrimitives[i];
        if primitive.ModifyGeometry or (primitive.ObjectType in [otPointLight,otDirectionalLight,otSpotLight]) then begin

            if doMoveRelative then
                usedVector := vectorSub(primitive.Center,deltaVector)
            else
                usedVector := deltaVector.Copy;

            if primitive.ObjectType = otPath then begin
              path := TPath(primitive);

              for j := 0 to path.ControlPointCount - 1 do begin
                cp := path.ControlPoints[j];
                cp.Position := VecSub(cp.Position,Vector3f(usedVector.x,usedVector.y,usedVector.z));
              end;
            end
            else
              for j := 0 to primitive.NumVertices-1 do
                  primitive.Vertices[j].Subtract(usedVector);

            if primitive is TLight then
                (primitive as TLight).Position.Subtract(usedVector);

            FreeAndNil(usedVector);
            primitive.HasChanged := true;
        end;
    end;
end;


end.

