{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PrimitiveSelector;

{$MODE Delphi}

interface

uses
  unit_Selector,
  unit_Primitives,
  unit_Paths,
  unit_PrimitiveList,
  unit_Types,
  unit_Vectors,
  unit_Enums;

type
  TPrimitiveSelector = class(TSelector)
  private
    fSelectedPrimitives: TPrimitiveList;
    fSelectIndex: integer;
    function CalculateOutCode(const aPoint, aMin, aMax: TVertex): TClippingOutcode;
    function IsLineSegmentOverlappingAABB(const aP1, aP2, aMin, aMax: TVertex): boolean;
    procedure DeselectVertices;
    procedure DeselectEdges;
    procedure DeselectPolygons;
    procedure SelectForGroups;

    function IsPathInFrustum(aPath: TPath; aFrustumPlanes,aFrustumPoints: array of TVertex): Boolean;
    function IsPrimitiveInFrustum(aPrimitive: TBasePrimitive; aFrustumPlanes,aFrustumPoints: array of TVertex): Boolean;
    function IsPathInRectangle(aPath: TPath; const ax1,ay1,ax2,ay2: integer;
                                             const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                             const aViewport: TViewport): boolean;
    function IsPrimitiveInRectangle(aPrimitive: TBasePrimitive; const ax1,ay1,ax2,ay2: integer;
                                                                const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                                                const aViewport: TViewport): boolean;
  protected
    function GetCenter: TVertice; override;
  public
    constructor Create(APrimitiveList, ASelectedPrimitiveList: TPrimitiveList);

    // NOTE: when clearing the SelectedPrimitives list, all subcollections (vertices, polygons, edges)
    // of the primitives in that list should be cleared too.

    procedure SelectByEditMode(ACurrentEditMode,ANewEditMode: TEditMode; ASelectAdjacentItems: boolean);
    procedure SelectByRange(AUpperLeftVertex, ALowerRightVertex: TVertice; AClearSelection: boolean);
    procedure SelectByVertex(AVertex: TVertice; ARange: double; AClearSelection: boolean);
    procedure SelectByRay(ARay: TRay; AClearSelection: boolean);
    procedure SelectByFrustum(aFrustumPlanes,aFrustumPoints: array of TVertex; AClearSelection: boolean);
    procedure SelectByRectangle(const ax1,ay1,ax2,ay2: integer;
                                const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                const aViewport: TViewport;
                                const AClearSelection: boolean);
    procedure SelectAll;
    procedure DeselectAll;
    procedure ReverseSelection;
    procedure GrowSelection;
    procedure ShrinkSelection; 
    property SelectedPrimitives: TPrimitiveList read fSelectedPrimitives write fSelectedPrimitives;
  end;

implementation

uses
  SysUtils,
  unit_Constants,
  unit_Math,
  unit_IntegerList;

// ### TPrimitiveSelector ######################################################

constructor TPrimitiveSelector.Create(APrimitiveList, ASelectedPrimitiveList: TPrimitiveList);
begin
    fPrimitives := APrimitiveList;
    fSelectedPrimitives := ASelectedPrimitiveList;
    fSelectIndex := 0;
    inherited Create;
end;

// calculates the cohen-sutherland outcode for a point and a bounding box; move to a specialized class some day?
function TPrimitiveSelector.CalculateOutCode(const aPoint, aMin, aMax: TVertex): TClippingOutcode;
begin
    Result := [];

    if aPoint.X > aMax.X then
    begin
        Result := Result + [coClipRight];
    end
    else if aPoint.X < aMin.X then
    begin
        Result := Result + [coClipLeft];
    end;

    if aPoint.Y > aMax.Y then
    begin
        Result := Result + [coClipTop];
    end
    else if aPoint.Y < aMin.Y then
    begin
        Result := Result + [coClipBottom];
    end;

    if aPoint.Z > aMax.Z then
    begin
        Result := Result + [coClipBack];
    end
    else if aPoint.Z < aMin.Z then
    begin
        Result := Result + [coClipFront];
    end;
end;

// implementation of the cohen-sutherland line-clipping algorithm; move to a specialized class some day?
function TPrimitiveSelector.IsLineSegmentOverlappingAABB(const aP1, aP2, aMin, aMax: TVertex): boolean;
var outcode1, outcode2: TClippingOutcode;
    min, max: TVertex;
    v1, v2: single;
    intersection: TVertex;
begin
    result := false;

    min := vertexAddScalar(aMin, -EPSILON);
    max := vertexAddScalar(aMax, EPSILON);

    // point 1 inside bounding box
    outcode1 := calculateOutCode( aP1, min, max );
    if ( outcode1 = [] ) then begin
        Result := TRUE;
        exit;
    end;

    // point 2 inside bounding box
    outcode2 := calculateOutCode( aP2, min, max );
    if ( outcode2 = [] ) then begin
        Result := TRUE;
        exit;
    end;

    // both points on same side of box
    if ( outcode1 * outcode2 ) <> [] then begin
        exit;
    end;

    // check intersections
    if ( coClipLeft in outcode1 ) or ( coClipRight in outCode1 ) then begin
        if coClipRight in outcode1 then
        begin
          intersection.X := max.X;
        end
        else
          intersection.X := min.X;

        v1 := aP2.X - aP1.X;
        v2 := intersection.X - aP1.X;
        intersection.Y := aP1.Y + v2 * ( aP2.Y - aP1.Y ) / v1;
        intersection.Z := aP1.Z + v2 * ( aP2.Z - aP1.Z ) / v1;
        if ( intersection.Y <= max.Y ) and ( intersection.Y >= min.Y ) and
         ( intersection.Z <= max.Z ) and ( intersection.Z >= min.Z ) then begin
          Result := true;
          exit;
        end;
    end;

    if ( coClipTop in outcode1 ) or ( coClipBottom in outcode1 ) then begin
        if coClipTop in outcode1 then begin
            intersection.Y := max.Y;
        end
        else
            intersection.Y := min.Y;

        v1 := aP2.Y - aP1.Y;
        v2 := intersection.Y - aP1.Y;
        intersection.X := aP1.X + v2 * ( aP2.X - aP1.X ) / v1;
        intersection.Z := aP1.Z + v2 * ( aP2.Z - aP1.Z ) / v1;
        if ( intersection.X <= max.X ) and ( intersection.X >= min.X ) and
           ( intersection.Z <= max.Z ) and ( intersection.Z >= min.Z ) then begin
            result := true;
            exit;
        end;
    end;

    if ( coClipFront in outcode1 ) or ( coClipBack in outcode1 )  then begin
        if coClipBack in outCode1 then begin
            intersection.Z := max.Z;
        end
        else
            intersection.Z := min.Z;

        v1 := aP2.Z - aP1.Z;
        v2 := intersection.Z - aP1.Z;
        intersection.X := aP1.X + v2 * ( aP2.X - aP1.X ) / v1;
        intersection.Y := aP1.Y + v2 * ( aP2.Y - aP1.Y ) / v1;
        if ( intersection.X <= max.X ) and ( intersection.X >= min.X ) and
           ( intersection.Y <= max.Y ) and ( intersection.Y >= min.Y ) then begin
            Result := true;
            exit;
        end;
    end;
end;

procedure TPrimitiveSelector.DeselectVertices;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Vertex_DeselectAll
end;

procedure TPrimitiveSelector.DeselectEdges;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Edge_DeselectAll
end;

procedure TPrimitiveSelector.DeselectPolygons;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Polygon_DeselectAll
end;

procedure TPrimitiveSelector.SelectForGroups;
var i,j,groupID: integer;
    groupIDs: TIntegerList;
begin
    if SelectedPrimitives.Count>0 then begin
        groupIDs := TIntegerList.Create;

        // first gather all group ID's in a list
        for i := 0 to SelectedPrimitives.Count-1 do begin
            if SelectedPrimitives[i].GroupID<>-1 then
                groupIDs.Add(SelectedPrimitives[i].GroupID);
        end;

        // now select all primitives with those ID's that weren't selected yet
        for i := 0 to groupIDs.Count-1 do begin
            groupID := groupIDs[i];
            for j := 0 to Primitives.Count-1 do begin
                if Primitives[j].GroupID=groupID then
                    SelectedPrimitives.Add(Primitives[j]);
            end;
        end;

        // free stuff
        FreeAndNil(groupIDs);
    end;
end;

function TPrimitiveSelector.GetCenter: TVertice;
var i: integer;
    primitive: TBasePrimitive;
begin
    fCenter.Assign(0,0,0);
    for i := 0 to SelectedPrimitives.Count-1 do begin
        primitive := SelectedPrimitives[i];
        fCenter.Add(primitive.Center);
    end;
    if SelectedPrimitives.Count>0 then
        fCenter.DivideByScalar(SelectedPrimitives.Count)
    else
        fCenter.ResetToInfinite;
    Result := fCenter;
end;

procedure TPrimitiveSelector.SelectByEditMode(ACurrentEditMode, ANewEditMode: TEditMode; ASelectAdjacentItems: boolean);
var i: integer;
    primitive: TBasePrimitive;
begin
    if not ASelectAdjacentItems then begin

        DeselectVertices;                                                       // deselect everything except selected objects
        DeselectEdges;

        if ((ACurrentEditMode=emPolygon) and (ANewEditMode<>emFace)) or         // only deselect polygons when switching to vertex/edge/object modes
           ((ACurrentEditMode=emFace) and (ANewEditMode<>emPolygon)) then
            DeselectPolygons;

    end else begin

        // based on the current editmode, use the currently selected items
        // to select other items in the new editmode; for example, if we are in
        // vertexmode + we have 1 vertex selected + we switch to polygonmode, we
        // determine which polygons use that particular vertex and select them
        // automatically; the vertex itself is deselected

        for i := 0 to SelectedPrimitives.Count-1 do begin
            primitive := SelectedPrimitives[i];

            primitive.HighlightedVertices.Clear;
            primitive.HighlightedEdges.Clear;
            primitive.HighlightedPolygons.Clear;

            case ACurrentEditMode of
                emVertex:
                    begin
                        primitive.SelectedEdges.Clear;
                        primitive.SelectedPolygons.Clear;

                        case ANewEditMode of
                            emEdge:
                                begin
                                    primitive.Vertex_SelectOnSwitchToEdgeMode;
                                end;
                            emPolygon,emFace:
                                begin
                                    primitive.Vertex_SelectOnSwitchToPolygonMode;
                                end;
                        end;
                        primitive.SelectedVertices.Clear;
                    end;

                emEdge:
                    begin
                        primitive.SelectedVertices.Clear;
                        primitive.SelectedPolygons.Clear;

                        case ANewEditMode of
                            emVertex:
                                begin
                                    primitive.Edge_SelectOnSwitchToVertexMode;
                                end;
                            emPolygon,emFace:
                                begin
                                    primitive.Edge_SelectOnSwitchToPolygonMode;
                                end;
                        end;
                        primitive.SelectedEdges.Clear;
                    end;

                emPolygon,emFace:
                    begin
                        primitive.SelectedVertices.Clear;
                        primitive.SelectedEdges.Clear;

                        case ANewEditMode of
                            emVertex:
                                begin
                                    primitive.Polygon_SelectOnSwitchToVertexMode;
                                end;
                            emEdge:
                                begin
                                    primitive.Polygon_SelectOnSwitchToEdgeMode;
                                end;
                        end;
                        if (ANewEditMode<>emPolygon) and (ANewEditMode<>emFace) then
                            primitive.SelectedPolygons.Clear;
                    end;

                emObject:
                    begin
                        primitive.SelectedVertices.Clear;
                        primitive.SelectedEdges.Clear;
                        primitive.SelectedPolygons.Clear;
                    end;
            end;
        end;
    end;
end;

procedure TPrimitiveSelector.SelectByRange(AUpperLeftVertex, ALowerRightVertex: TVertice; AClearSelection: boolean);
var i: integer;
begin
    if AClearSelection then
        SelectedPrimitives.Clear;

    //This also works for TPath because that class has an
    //overridden version of Object_SelectByRange
    for i := 0 to Primitives.Count-1 do begin
        if not Primitives[i].Frozen and Primitives[i].Object_SelectByRange(AUpperLeftVertex, ALowerRightVertex) then
            SelectedPrimitives.Add(Primitives[i]);
    end;
    SelectForGroups;
end;

procedure TPrimitiveSelector.SelectByVertex(AVertex: TVertice; ARange: double; AClearSelection: boolean);
var counter: integer;
    primitive: TBasePrimitive;
    polygon: TPolygon;
    isFound: boolean;
begin
    if AClearSelection then
        SelectedPrimitives.Clear;

    if Primitives.Count>0 then begin
        fSelectIndex := (fSelectIndex+1) mod Primitives.Count;

        counter := 0;
        primitive := Primitives.Primitives[fSelectIndex];
        //Take care of paths too
        if primitive is TPath then
          isFound := TPath(primitive).IsPointOnPath(AVertex)
        else
          isFound := primitive.IsPointInObject(AVertex,polygon);

        while (not isFound) and (counter<Primitives.Count) do
        begin
            inc(counter);
            fSelectIndex := (fSelectIndex+1) mod Primitives.Count;
            primitive := Primitives[fSelectIndex];
            // IsPointInObject checks for Frozen objects
            if primitive is TPath then
              isFound := TPath(primitive).IsPointOnPath(AVertex)
            else
              isFound := primitive.IsPointInObject(AVertex,polygon);
        end;

        if isFound then
            SelectedPrimitives.Add(primitive);

        SelectForGroups;
    end;
end;

//NOTE: This method does not yet support selection of paths. we need to add this
procedure TPrimitiveSelector.SelectByRay(ARay: TRay; AClearSelection: boolean);
var i, j: integer;
    primitive,closestPrimitive: TBasePrimitive;
    aabbMin, aabbMax: TVertex;
    polygon: TPolygon;
    distance,closestDistance: single;
begin
    if AClearSelection then
        SelectedPrimitives.Clear;

    closestPrimitive := nil;
    closestDistance := -1;

    for i := 0 to Primitives.Count - 1 do begin
        primitive := Primitives[i];

        if primitive.Visible and not primitive.Frozen then begin
            // Don't pick lights if lights are off
            //and (RenderOptions.RenderLights or not (primitive is TLight)) then begin
            aabbMin := verticeToVertex(primitive.AABB.MinVertice);
            aabbMax := verticeToVertex(primitive.AABB.MaxVertice);

            if IsLineSegmentOverlappingAABB(aRay.Origin, aRay.Destination, aabbMin, aabbMax) then begin
                j := 0;
                while (j < primitive.NumPolygons) do begin
                    polygon := primitive.Polygons[j];

                    if polygon.IsIntersectedByRay(aRay, distance) and
                       ((closestDistance = -1) or (distance < closestDistance)) then begin
                        closestDistance := distance;
                        closestPrimitive := primitive;
                    end;

                    j := j + 1;
                end;

            end;
        end;
    end;

    if Assigned(closestPrimitive) then begin
        if SelectedPrimitives.IndexOf(closestPrimitive)=-1 then
            SelectedPrimitives.Add(closestPrimitive)
        else
            SelectedPrimitives.Remove(closestPrimitive);
    end;
    SelectForGroups;
end;

procedure TPrimitiveSelector.SelectByFrustum(aFrustumPlanes,aFrustumPoints: array of TVertex; AClearSelection: boolean);
var primitive: TBasePrimitive;
    i: integer;
    InFrustum: boolean;
begin
    if AClearSelection then
        SelectedPrimitives.Clear;

    for i := 0 to Primitives.Count - 1 do begin
        primitive := Primitives[i];
        if not Primitives[i].Frozen then
        begin

            //The path is a special case here (it has no vertices, but controlpoints instead)
            if Primitives[i] is TPath then
              InFrustum := IsPathInFrustum( TPath(Primitives[i]), aFrustumPlanes, aFrustumPoints)
            else
              InFrustum := IsPrimitiveInFrustum(Primitives[i], aFrustumPlanes, aFrustumPoints);

            if InFrustum then SelectedPrimitives.Add(primitive);
        end;
    end;
end;

function TPrimitiveSelector.IsPathInFrustum(aPath: TPath; aFrustumPlanes,
                                            aFrustumPoints: array of TVertex): Boolean;
var
  i: Integer;
  vertex: TVertex;
  v1, v2: TVector3f;
begin
  Result := true;
  I := 0;
  while aPath.GetNextPathSegmentForSelection(I, v1, v2) do
  begin
      Vertex.X := v1.x;
      Vertex.Y := v1.y;
      Vertex.Z := v1.z;
      if not PointInFrustum(vertex,aFrustumPlanes,aFrustumPoints) then
      begin
          Result := false;
          Exit;
      end;
  end;
end;

function TPrimitiveSelector.IsPrimitiveInFrustum(aPrimitive: TBasePrimitive; aFrustumPlanes,
                                                 aFrustumPoints: array of TVertex): Boolean;
var
  j, k: Integer;
  Polygon: TPolygon;
  vertex: TVertex;
begin
  Result := true;
  for j := 0 to aPrimitive.NumPolygons - 1 do
  begin
      polygon := aPrimitive.Polygons[j];
      for k := 0 to polygon.NumVertices - 1 do
      begin
          vertex := verticeToVertex(polygon.VerticeList[k]);
          if not PointInFrustum(vertex,aFrustumPlanes,aFrustumPoints) then
          begin
              Result := false;
              Exit;
          end;
      end;
  end;
end;


procedure TPrimitiveSelector.SelectByRectangle(const ax1,ay1,ax2,ay2: integer;
                                               const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                               const aViewport: TViewport;
                                               const AClearSelection: boolean);
var
    primitive       : TBasePrimitive;
    i               : integer;
    InRect : boolean;
begin
    if AClearSelection then
        SelectedPrimitives.Clear;

    for i := 0 to Primitives.Count - 1 do begin
        primitive := Primitives[i];
        if not (Primitives[i].Frozen) and (Primitives[i].Visible) then
        begin
//            InRect := true;

            //The path is a special case here (it has no vertices, but controlpoints instead)
            if primitive is TPath then
              InRect := IsPathInRectangle( TPath(primitive), ax1, ay1, ax2, ay2,
                                           aModelViewMatrix, aProjectionMatrix,
                                           aViewport)
            else
              InRect := IsPrimitiveInRectangle( TPath(primitive), ax1, ay1, ax2, ay2,
                                                aModelViewMatrix, aProjectionMatrix,
                                                aViewport);

            if InRect then
              SelectedPrimitives.Add(primitive);
        end;
    end;
end;

function TPrimitiveSelector.IsPathInRectangle(aPath: TPath;
                                              const ax1, ay1, ax2,ay2: integer;
                                              const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                              const aViewport: TViewport): boolean;
var
  i: Integer;
  v1, v2: TVector3f;
  vertice: TVertice;
  point: TPoint2D;
begin
  Result := True;
  I := 0;
  vertice := TVertice.Create();
  try
    while aPath.GetNextPathSegmentForSelection(I, v1, v2) do
    begin
        vertice.x := v1.x;
        vertice.y := v1.y;
        vertice.z := v1.z;

        point := vertice.ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewPort);
        if (point.X < ax1) or
           (point.X > ax2) or
           (point.Y < ay1) or
           (point.Y > ay2) then
        begin
            Result := false;
            Exit;
        end;
    end;
  finally
    Vertice.Free();
  end;
end;

function TPrimitiveSelector.IsPrimitiveInRectangle(aPrimitive: TBasePrimitive;
                                                  const ax1, ay1, ax2, ay2: integer;
                                                  const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                                  const aViewport: TViewport): boolean;
var
  j: Integer;
  vertice: TVertice;
  point: TPoint2D;
begin
  Result := True;
  for j := 0 to aPrimitive.Vertices.Count - 1 do
  begin
      vertice := aPrimitive.Vertices[j];
      point := vertice.ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewPort);
      if (point.X < ax1) or
         (point.X > ax2) or
         (point.Y < ay1) or
         (point.Y > ay2) then
      begin
          Result := false;
          Exit;
      end;
  end;
end;

procedure TPrimitiveSelector.SelectAll;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do begin
        if not (Primitives[i].Frozen) and (Primitives[i].Visible) then
            SelectedPrimitives.Add(Primitives[i]);
    end;
end;

procedure TPrimitiveSelector.DeselectAll;
var i: integer;
    primitive: TBasePrimitive;
begin
    // clear individual selections per primitive
    for i := 0 to SelectedPrimitives.Count-1 do
    begin
        primitive := SelectedPrimitives[i];
        primitive.Selected := false;
        primitive.SelectedVertices.Clear;
        primitive.SelectedEdges.Clear;
        primitive.SelectedPolygons.Clear;

        //Also deselect controlpoints that a path may have
        if primitive is TPath then
          TPath(Primitive).SelectedControlPoints.Clear();
    end;
    // now clear the selection of primitives
    SelectedPrimitives.Clear;
end;

procedure TPrimitiveSelector.ReverseSelection;
var i: integer;
    primitive: TBasePrimitive;
    newSelection: TPrimitiveList;
begin
    newSelection := TPrimitiveList.Create(false);
    for i := 0 to Primitives.Count-1 do begin
        primitive := Primitives[i];
        if SelectedPrimitives.IndexOf(primitive)=-1 then
            newSelection.Add(primitive);
    end;
    SelectedPrimitives.Clear;
    SelectedPrimitives.AddList(newSelection);
    FreeAndNil(newSelection);
end;

procedure TPrimitiveSelector.GrowSelection;
begin

end;

procedure TPrimitiveSelector.ShrinkSelection;
begin

end;


end.
