{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
// the slice tool is actually a simpified CSG tool (it doesn´t use BSP trees);
// it uses a primitive containing a single polygon;
// the plane of this polygon is used as cutting plane;
// the first primitive of the SelectedPrimitives collection
// should be that single polygon primitive; this way, the
// Slice tool works quite simular to the CSG routines,
// which is what is to be expected

// TODO: rename to unit_CSGSliceTool and TCSGSliceTool someday

unit unit_PrimitiveSliceTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Options,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool,
  unit_PrimitiveList;

type
  TPrimitiveSlicer = class(TPrimitiveTool)
  private
    fCSGOptions: TCSGOptions;
    function IsValidOperator(AObject: TObject): boolean;
    procedure AddPolygonToPrimitive(APolygon: TPolygon; APrimitive: TBasePrimitive);
    procedure AddPlanePolygonToPrimitive(AVertexList: TVerticeList; APrimitive: TBasePrimitive);
    procedure SlicePrimitive(APrimitive: TBasePrimitive; APlane: TPlane; var ABackPrimitive, AFrontPrimitive: TBasePrimitive);
    procedure DoSlice(AResultingPrimitives: TPrimitiveList);
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses unit_Lights,
     unit_PolygonByPlaneSplitterTool,
     unit_VertexFillerTool, unit_BaseTool;

function TPrimitiveSlicer.IsValidOperator(AObject: TObject): boolean;
var primitive: TBasePrimitive;
begin
    primitive := TBasePrimitive(AObject);
    Result := primitive.NumPolygons=1;
end;

procedure TPrimitiveSlicer.AddPolygonToPrimitive(APolygon: TPolygon; APrimitive: TBasePrimitive);
var i,index: integer;
    vertex: TVertice;
    polygon: TPolygon;
    verticesToBeFreed: TVerticeList;
begin
    // the verticesToBeFreed list owns objects; it is used specifically to free
    // vertices that aren´t added to a polygon because another vertex with the
    // same coordinates already exists in the primitive; in that case, that vertex
    // is used instead (reuse vertices to avoid double vertices in the mesh);
    // the vertex provided by APolygon needs to be freed and is therefore stored
    // in a seperate list which is then freed at the end of this procedure
    verticesToBeFreed := TVerticeList.Create(true,false);

    polygon := TPolygon.Create;
    polygon.Material := APolygon.Material;
    for i := 0 to APolygon.NumVertices-1 do begin
        index := APrimitive.Vertices.InList(APolygon.VerticeList[i]);
        if index =-1 then
            APrimitive.Polygon_AddVertex(polygon,APolygon.VerticeList[i])
        else begin
            // use an existing vertex with the same coordinates instead
            vertex := APrimitive.Vertices[index];
            APrimitive.Polygon_AddVertex(polygon,vertex);
            // register the provided vertex for deletion (done later)
            verticesToBeFreed.Add(APolygon.VerticeList[i]);
        end;
    end;
    APrimitive.Polygon_Add(polygon);

    // free vertices of the provided APolygon that weren´t used in the final polygon
    FreeAndNil(verticesToBeFreed);
end;

procedure TPrimitiveSlicer.AddPlanePolygonToPrimitive(AVertexList: TVerticeList; APrimitive: TBasePrimitive);
var i,index: integer;
begin
    // at this stage in the slicing process, each vertex that is on the splitting
    // plane (provided in AVertexList), should already be a part of the primitive;
    // we search for those shared vertices here and put them in the SelectedVertices list
    for i := 0 to AVertexList.Count-1 do begin
        index := APrimitive.Vertices.InList(AVertexList[i]);
        if index<>-1 then
            APrimitive.SelectedVertices.Add(APrimitive.Vertices[index]);
    end;

    // we now fill the region of vertices with a Vertex Filler
    TVertexFillerFunction.Execute(APrimitive,APrimitive.Polygons[0].Material);
end;

procedure TPrimitiveSlicer.SlicePrimitive(APrimitive: TBasePrimitive; APlane: TPlane; var ABackPrimitive, AFrontPrimitive: TBasePrimitive);
var i: integer;
    newPolygons: TPolygonList;
    verticesOnPlane: TVerticeList;
    paramList: TParamList;
begin
    paramList := TParamList.Create(false);
    newPolygons := TPolygonList.Create(true,true);
    verticesOnPlane := TVerticeList.Create(true,true);

    for i := 0 to APrimitive.NumPolygons-1 do begin
        newPolygons.Clear;

        // feed each polygon to a polygon-by-plane splitter function
        paramList.Add(APrimitive.Polygons[i], 'polygon', true);
        paramList.Add(APlane, 'plane');
        paramList.Add(newPolygons, 'newPolygons');
        paramList.Add(verticesOnPlane, 'verticesOnPlane');         // this list keeps track of vertices that are ON the splitting plane
        TPolygonByPlaneSplitterFunction.Execute(paramList);

        // add each polygon to its appropiate primitive
        if newPolygons[0].NumVertices>0 then
            AddPolygonToPrimitive(newPolygons[0],ABackPrimitive);

        if newPolygons[1].NumVertices>0 then
            AddPolygonToPrimitive(newPolygons[1],AFrontPrimitive);
    end;

    // for each primitive, use a Vertex Filler to create a new polygon
    // using the vertices that are ON the cutting plane (the intersections)
    if verticesOnPlane.Count>0 then begin
        AddPlanePolygonToPrimitive(verticesOnPlane,ABackPrimitive);
        AddPlanePolygonToPrimitive(verticesOnPlane,AFrontPrimitive);
    end;

    // free stuff
    FreeAndNil(verticesOnPlane);
    FreeAndNil(newPolygons);
    FreeAndNil(paramList);
end;

procedure TPrimitiveSlicer.DoSlice(AResultingPrimitives: TPrimitiveList);
var i,counter,firstIndex: integer;
    plane: TPlane;
    backPrimitive, frontPrimitive: TBasePrimitive;
begin
    SelectedPrimitives.RemoveItems(TLight);

    // construct plane using first primitive
    plane := TPlane.Create(SelectedPrimitives[0].Polygons[0].VerticeList[0],SelectedPrimitives[0].Polygons[0].Normal);

    // slice each primitive and save the results in a seperate list
    for i := 1 to SelectedPrimitives.Count-1 do begin
        backPrimitive := TUserDefined.CreateUserDefined;
        frontPrimitive := TUserDefined.CreateUserDefined;
        SlicePrimitive(SelectedPrimitives[i],plane,backPrimitive,frontPrimitive);
        if backPrimitive.NumPolygons>0 then
            AResultingPrimitives.Add(backPrimitive)
        else
            FreeAndNil(backPrimitive);
        if frontPrimitive.NumPolygons>0 then
            AResultingPrimitives.Add(frontPrimitive)
        else
            FreeAndNil(frontPrimitive);
    end;

    // delete selected primitives from scene and clear the selection list;
    // optionally delete the operator primitive too
    counter := SelectedPrimitives.Count;
    if fCSGOptions.KeepOperator then
        firstIndex := 1
    else
        firstIndex := 0;
    while counter>firstIndex do begin
        ScenePrimitives.Remove(SelectedPrimitives[counter-1]);
        SelectedPrimitives.Delete(counter-1);
        dec(counter);
    end;

    // copy objects to scene
    for i := 0 to AResultingPrimitives.Count-1 do
        ScenePrimitives.AddPrimitive(TBasePrimitive(AResultingPrimitives[i]),true);

    // free stuff
    FreeAndNil(plane);
end;

procedure TPrimitiveSlicer.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Slice primitive';
    Description := 'Slices primitives';
end;

procedure TPrimitiveSlicer.DoExecute(AParamList: TParamList);
var primitiveList: TPrimitiveList;
    i: integer;
begin
    try
        if not Assigned(AParamList) then
            raise Exception.Create('No parameters for Slice provided.')
        else begin
            if AParamList.Count<>1 then
                raise Exception.Create('Wrong number of parameters for Slice provided. CSGOptions expected.')
            else begin
                fCSGOptions := TCSGOptions(AParamList.NamedParameters['csgOptions']);

                if SelectedPrimitives.Count<2 then
                    raise Exception.Create('Select two or more objects. The first object is used as cutting tool and should contain one polygon only.')
                else begin
                    if not IsValidOperator(SelectedPrimitives[0]) then
                        raise Exception.Create('The operator object should contain one polygon only.')
                    else begin
                        // if we want to keep the operator, we don't want it stored
                        // in the Undo buffer; if we do, it'll get copied into the
                        // scene on Undo again while it is still there because it
                        // was never deleted in the first place
                        if fCSGOptions.KeepOperator then begin
                            UndoPrimitiveList.DeleteByID(SelectedPrimitives[0].ID);
                            PrimitiveIDList.Remove(SelectedPrimitives[0].ID,true);
                        end;

                        primitiveList := TPrimitiveList.Create(false,true);
                        DoSlice(primitiveList);
                        // add the ID´s of the resulting primitives to the Undo buffer
                        // so we can remove them when this Undo is executed
                        for i := 0 to primitiveList.Count-1 do
                            PrimitiveIDList.Add(primitiveList[i].ID);

                        FreeAndNil(primitiveList);
                    end;
                end;
            end;
        end;
    except
        on E: Exception do MessageDlg(E.Message,mtError,[mbOK],0);
    end;
end;

end.
