{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_PrimitiveTool;

{$MODE Delphi}

interface

uses
  unit_PrimitiveList,
  unit_BaseTool,
  unit_Enums,
  unit_IntegerList;

type
  TPrimitiveTool = class(TBaseTool)
  private
    fPrimitiveIDList: TIntegerList;
    fUndoPrimitiveList: TPrimitiveList;
    fRedoPrimitiveList: TPrimitiveList;
    fPrimitives: TPrimitiveList;
    fSelectedPrimitives: TPrimitiveList;
  protected
    procedure PrepareUndoForModification; override;
    property PrimitiveIDList: TIntegerList read fPrimitiveIDList write fPrimitiveIDList;
    property UndoPrimitiveList: TPrimitiveList read fUndoPrimitiveList write fUndoPrimitiveList;
    property RedoPrimitiveList: TPrimitiveList read fRedoPrimitiveList write fRedoPrimitiveList;
  public
    constructor Create(AToolAction: TToolAction);
    destructor Destroy; override;
    procedure Undo; override;
    procedure Redo; override;
    property ScenePrimitives: TPrimitiveList read fPrimitives write fPrimitives;
    property SelectedPrimitives: TPrimitiveList read fSelectedPrimitives write fSelectedPrimitives;
  end;

implementation

uses
  SysUtils,
  unit_Primitives;

// ### TPrimitiveTool ##########################################################

constructor TPrimitiveTool.Create(AToolAction: TToolAction);
begin
    fPrimitiveIDList := TIntegerList.Create;
    fUndoPrimitiveList := TPrimitiveList.Create(true,true);
    fRedoPrimitiveList := TPrimitiveList.Create(true,true);
    inherited Create(AToolAction);
end;

destructor TPrimitiveTool.Destroy;
begin
    FreeAndNil(fPrimitiveIDList);
    FreeAndNil(fUndoPrimitiveList);
    FreeAndNil(fRedoPrimitiveList);
end;

procedure TPrimitiveTool.PrepareUndoForModification;
var i: integer;
    primitive: TBasePrimitive;
begin
    // copy selected primitives to buffer and store ID's of selected primitives
    // in a seperate list;
    // on Undo, we use these ID's to remove primitives from the scene and place
    // copies of the primitives in the buffer in the scene
    for i := 0 to SelectedPrimitives.Count-1 do begin
        primitive := SelectedPrimitives[i].Copy;
        primitive.ID := SelectedPrimitives[i].ID;
        fUndoPrimitiveList.Add(primitive);

        fPrimitiveIDList.Add(SelectedPrimitives[i].ID);
    end;
end;

procedure TPrimitiveTool.Undo;
var i,id: integer;
    primitive,primitiveCopy: TBasePrimitive;
begin
    // use the ID list to remove primitives from the scene
    // and to save copies of those primitives for Redo purposes
    for i := 0 to fPrimitiveIDList.Count-1 do begin
        id := fPrimitiveIDList[i];
        primitive := ScenePrimitives.GetPrimitiveByID(id);
        if Assigned(primitive) then begin
            primitiveCopy := primitive.Copy;
            primitiveCopy.ID := primitive.ID;
            fRedoPrimitiveList.Add(primitiveCopy);
            ScenePrimitives.Remove(primitive);
        end;
    end;

    fPrimitiveIDList.Clear;

    // now place copies of the saved primitives back in the scene
    for i := 0 to fUndoPrimitiveList.Count-1 do begin
        primitiveCopy := fUndoPrimitiveList[i].Copy;
        primitiveCopy.ID := fUndoPrimitiveList[i].ID;
        ScenePrimitives.Add(primitiveCopy);
        SelectedPrimitives.Add(primitiveCopy);
        fPrimitiveIDList.Add(primitiveCopy.ID);
    end;

    // clear Undo buffer
    fUndoPrimitiveList.Clear;
end;

procedure TPrimitiveTool.Redo;
var i,id: integer;
    primitive,primitiveCopy: TBasePrimitive;
begin
    // use the ID list to remove primitives from the scene
    // and to save copies of those primitives for Redo purposes
    for i := 0 to fPrimitiveIDList.Count-1 do begin
        id := fPrimitiveIDList[i];
        primitive := ScenePrimitives.GetPrimitiveByID(id);
        if Assigned(primitive) then begin
            primitiveCopy := primitive.Copy;
            primitiveCopy.ID := primitive.ID;
            fUndoPrimitiveList.Add(primitiveCopy);
            ScenePrimitives.Remove(primitive);
        end;
    end;

    fPrimitiveIDList.Clear;

    // now place copies of the saved primitives in the scene
    for i := 0 to fRedoPrimitiveList.Count-1 do begin
        primitiveCopy := fRedoPrimitiveList[i].Copy;
        primitiveCopy.ID := fRedoPrimitiveList[i].ID;
        ScenePrimitives.Add(primitiveCopy);
        SelectedPrimitives.Add(primitiveCopy);
        fPrimitiveIDList.Add(primitiveCopy.ID);
    end;

    // clear Redo buffer
    fRedoPrimitiveList.Clear;
end;




end.
 
