{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_RaytraceManager;

{$MODE Delphi}

// INFO, PJ: Intermediate layer between raytracer and raytracing thread.

interface

uses
    Graphics, Contnrs,
    GR32, GR32_Transforms, GR32_Resamplers,
    frmRaytracingProgressForm,
    frmRaytracingOptionsForm,
    unit_Cameras,
    unit_Scene,
    unit_TextureTypes,
    unit_Textures,
    unit_Types,
    unit_Primitives,
    unit_PrimitiveList,
    unit_Math,
    unit_Lights,
    unit_Enums,
    unit_Materials,
    RTRenderWorldCl, RTSphereCl, RTMathFn, RTTextureCl, RTBVHCl,
    RTWorldCl, RTCameraCl, RTMeshCl, RTMaterialCl, RTMiscTp;

type
    TRaytracingItemStatus = ( risNew, risProcessing, risPaused, risDone );
    TRaytracingItem = class( TObject )
    private
        fCamera: TCamera3D;
        fSceneName: string;
        fSceneFileName: string;
        fThumbFilename: string;
        fOutputFileName: string;
        fOptions: TRaytracingOptions;
        fStatus: TRaytracingItemStatus;
        fPosition: integer;
        fRenderingTime: integer;

        // state
        fScene: TScene;
        fThumbNail: TBitmap32;
        fOutput: TFloatTextureData;

        function GetPercentageDone: Integer;
        function GetThumbnail: TBitmap32;
        function GetOutput: TFloatTextureData;
        function GetScene: TScene;

    public
        constructor Create;
        destructor Destroy; override;

        procedure Invalidate;

        property Camera: TCamera3D read fCamera;
        property SceneName: string read fSceneName write fSceneName;
        property SceneFilename: string read fSceneFileName write fSceneFileName;
        property ThumbFilename: string read fThumbFilename write fThumbFilename;
        property OutputFilename: string read fOutputFilename write fOutputFilename;
        property RenderingTime: integer read fRenderingTime write fRenderingTime;
        property Options: TRaytracingOptions read fOptions write fOptions;
        property Status: TRaytracingItemStatus read fStatus write fStatus;
        property Position: integer read fPosition write fPosition;

        property PercentageDone: Integer read GetPercentageDone;

        property Thumbnail: TBitmap32 read GetThumbnail;
        property Output: TFloatTextureData read GetOutput;
        property Scene: TScene read GetScene;
    end;

// ----------------------------------------------------------------------------

    TRaytraceManager = class(TObject)
    private
        fModelViewMatrixInverse: TMatrix4x4;
        fItem: TRaytracingItem;
        fRenderingStartTime: Cardinal;

        fProgressHandler: TRTProgressHandler;

        fLastPosition: integer;
        fCamera: TRTCamera;

        fTextures: TObjectList;
        fMaterials: TObjectList;
        fMeshes: TObjectList;

        procedure InitializeWorld( const aWorld: TRTWorld );

        procedure AddLightToWorld(const aLight: TLight; const aWorld: TRTWorld);
        procedure AddTrianglesForPolygon(const aPolygon: TPolygon; const aPrimitive: TPrimitive; const aMesh: TRTMesh);
        function ConvertMaterial(const aMaterial: TMaterial): TRTMaterial;
        function ConvertCamera(const aCamera: TCamera3D): TRTCamera;
        function ConvertVector(const aVector: TVertice ): TRTVector3D;
        function ConvertUVCoordinate(const aUVCoordinate: TUVCoord): TRTUVCoordinate;
        function ConvertSphere(const aPrimitive: TPrimitive): TRTSphere;
        function ConvertColor(const aColor: TColor): TRGBColor;
        function ConvertTexture(const aTexture: TTexture): TRTTexture;
        function MakeMaterial(const aMaterial: TMaterial): TRTMaterial;
        function MakeTexture(const aTexture: TTexture): TRTTexture;

        function IsTransparentMaterial(const aMaterial: TMaterial): Boolean;

        procedure HandleProgress( const aPosition: integer; const aTexture: TFloatTextureData );
        procedure UpdateThumbNail( const aStartPosition, anEndPosition: integer; const anImage: TFloatTextureData );

        procedure StoreTextureDataForItem( const anItem: TRaytracingItem );

    public
        constructor Create(const anItem: TRaytracingItem);
        destructor Destroy; override;

        procedure Raytrace;
        procedure Stop;

        property OnProgress: TRTProgressHandler read fProgressHandler write fProgressHandler;
    end;

// ----------------------------------------------------------------------------

    TRaytraceTexture = class( TRTTexture )
    private
        fTextureData: TTextureData;

    public
        constructor Create( const aData: TTextureData );

        function SampleColor( const aShadeRec: TRTShadeRec ): TRGBColor; override;
    end;

// ----------------------------------------------------------------------------

    procedure TestRaytracer;

{$IFDEF RAYTRACER_DEBUG}
var
  _bvh: TRTBVH;
{$ENDIF}


implementation

uses
    Windows, Controls, Forms, SysUtils, Dialogs, Math,
    unit_TextureWriter,
    unit_TextureManipulator,
    unit_TextureLoader,
    unit_FloatTextureLoader,
    unit_FloatTextureWriter,
    unit_Functions,
    unit_Variables,
    unit_XMLSceneReader,
    unit_Constants,
    frm3dView, frmRaytracingPreviewForm,
    RTLightCl, RTPointLightCl, RTDirectionalLightCl, RTSpotLightCl,
    RTMatteMaterialCl, RTPhongMaterialCl, RTReflectiveMaterialCl,
    RTTransparentMaterialCl, RTMiscCn,
    RTPinholeCameraCl, RTRayCastTracerCl, RTAmbientLightCl,
    RTGridCl, RTTriangleCl, RTCompoundObjectCl, unit_XMLDataReader;

constructor TRaytracingItem.Create;
begin
  inherited Create;

  fCamera := TCamera3D.create( 0, 0, 0 );
end;

destructor TRaytracingItem.Destroy;
begin
  FreeAndNil( fCamera );
  FreeAndNil( fScene );
  FreeAndNil( fThumbNail );

  inherited Destroy;
end;

procedure TRaytracingItem.Invalidate;
begin
    // INFO, PJ: To keep this method safe to call at all times, it only
    //           flushes the output buffer when status <> processing
    if fStatus <> risProcessing then
        FinalizeFloatTextureData( fOutput );
end;

function TRaytracingItem.GetPercentageDone: Integer;
begin
    result := ( 100 * ( fPosition + 1 )) div ( Options.Width * Options.Height );
end;

function TRaytracingItem.GetThumbnail: TBitmap32;
var filename: string;
begin
    if fThumbNail = nil then begin
        filename := dir_raytracing + ThumbFilename;

        fThumbNail := TBitmap32.Create;
        fThumbNail.Resampler := TKernelResampler.Create( nil );
        TKernelResampler(fThumbNail.Resampler).Kernel := TCubicKernel.Create;

        try
            fThumbNail.LoadFromFile(filename);

        except
            fThumbNail.Width := RAYTRACER_THUMBNAIL_SIZE * 2;
            fThumbNail.Height := RAYTRACER_THUMBNAIL_SIZE * 2;
        end;
    end;

    result := fThumbNail;
end;

function TRaytracingItem.GetOutput: TFloatTextureData;
var isLoaded: Boolean;
    filename: String;
    loader: TFloatTextureLoader;
begin
    if fOutput.Width = 0 then begin
        isLoaded := false;

        // INFO, PJ: Only try to load if marked as saved (filename <> '')
        if fOutputFilename <> '' then begin
            filename := dir_raytracing + fOutputFileName;

            loader := TFloatTextureLoader.Create;
            try
                try
                    fOutput := loader.LoadTextureData( filename );
                    isLoaded := true;
                finally
                    FreeAndNil( loader );
                end;
            except
                // INFO, PJ: Suppress any errors that might occur
            end;
        end;

        if not isLoaded then begin
            // INFO, PJ: Create a brand new
            fOutput := FloatTextureData(Options.Width, Options.Height);
            Status := risNew;
            Position := 0;
        end;
    end;

    result := fOutput;
end;

function TRaytracingItem.GetScene: TScene;
var filename: string;
    reader: TXMLSceneReader;
begin
    if fScene = nil then
    begin
        filename := dir_raytracing + SceneFileName;

        fScene := TScene.Create;

        // reader := TXMLSceneReader.Create; // chrisu: Quickfix for the raytracer-hang bug.
        reader := TXMLSceneReader.CreateWithoutProgressform; // chrisu: Quickfix for the raytracer-hang bug.
        try
            reader.ReadFromFile( filename, fScene );
        finally
            FreeAndNil( reader );
        end;
    end;

    result := fScene;
end;

// ----------------------------------------------------------------------------
    
function Vertex4ToVertex(const aVertex4: TVertex4): TVertex;
begin
    result.X := aVertex4[0];
    result.Y := aVertex4[1];
    result.Z := aVertex4[2];
end;

constructor TRaytraceManager.Create(const anItem: TRaytracingItem);
begin
    inherited Create;

    fItem := anItem;
    fModelViewMatrixInverse := Matrix4x4Transpose(Matrix4x4Invert(fItem.Camera.Matrix));

    fMeshes := TObjectList.Create( TRUE );
    fMaterials := TObjectList.Create( TRUE );
    fMaterials.Count := fItem.Scene.MaterialManager.NumMaterials;
    fTextures := TObjectList.Create( TRUE );
    fTextures.Count := fItem.Scene.MaterialManager.Textures.Count;
end;

destructor TRaytraceManager.Destroy;
begin
    FreeAndNil( fMeshes );
    FreeAndNil( fMaterials );
    FreeAndNil( fTextures );

    inherited Destroy;
end;

procedure TRaytraceManager.InitializeWorld( const aWorld: TRTWorld );
var primitive: TBasePrimitive;
    primitives: TPrimitiveList;
    primitiveIndex, polygonIndex: integer;
    mesh: TRTMesh;
    light: TRTAmbientLight;
    compoundObject: TRTCompoundObject;

begin
    // INFO, PJ: The compoundObject is our space partitioning scheme.

    compoundObject := TRTBVH.Create;

    {$IFDEF RAYTRACER_DEBUG}
    _bvh := TRTBVH(compoundObject);
    {$ENDIF}

    mesh := TRTMesh.Create( compoundObject );
    fMeshes.Add( mesh );

    primitives := fItem.Scene.Primitives;
    for primitiveIndex := 0 to primitives.Count - 1 do begin
        primitive := primitives[primitiveIndex];

        if primitive.Visible then
        begin
            if primitive is TLight then begin
                AddLightToWorld(primitive as TLight, aWorld);
            end
            else if (primitive as TPrimitive).IsTrueSphere then begin
                compoundObject.AddObject( ConvertSphere( primitive as TPrimitive ));
            end
            else begin
                for polygonIndex := 0 to primitive.NumPolygons - 1 do
                    AddTrianglesForPolygon(primitive.Polygons[polygonIndex], primitive as TPrimitive, mesh);
            end;
        end;
    end;

    if aWorld.NumLights = 0 then begin
        light := TRTAmbientLight.Create;
        light.RadianceScale := 1;
        aWorld.AddLight( light );
    end;

//    for i := 0 to mesh.CompoundObject.NumObjects - 1 do
//    begin
//        aWorld.AddObject( mesh.CompoundObject.Objects[i] );
//    end;

    aWorld.AddObject( mesh.CompoundObject );

    TRTRenderWorld(aWorld).OnProgress := HandleProgress;
end;

procedure TRaytraceManager.AddLightToWorld(const aLight: TLight; const aWorld: TRTWorld);
var light: TRTLight;
    directionalLight: TRTDirectionalLight;
    spotLight: TRTSpotLight;
//    colorScale: single;
    pointLight: TRTPointLight;


begin
    if not aLight.Active then exit;

    light := nil;

    case aLight.ObjectType of
        otPointLight: begin
            pointLight := TRTPointLight.Create;
            pointLight.Location := ConvertVector( aLight.Position );
            pointLight.RadianceScale := aLight.Intensity / 100;
            light := pointLight;
        end;

        otDirectionalLight: begin
            // INFO, PJ:
            //   Raytracing directional lights are "specified by the direction from which the light is coming".
            //   DeleD's directional lights contain the direction the light is traveling (the exact opposite).
            directionalLight := TRTDirectionalLight.Create;
            directionalLight.Direction := VectorMultiply( ConvertVector( (aLight as TDirectionalLight).Direction ), -1 );
            directionalLight.RadianceScale := aLight.Intensity / 100;
            light := directionalLight;
        end;

        otSpotLight: begin
            // INFO, PJ:
            //   Raytracing spotlights are "specified by the direction from which the light is coming".
            //   DeleD's directional lights contain the direction the light is traveling (the exact opposite).
            spotLight := TRTSpotLight.Create;
            spotLight.Direction := VectorMultiply( ConvertVector( (aLight as TSpotLight).Direction ), -1 );
            spotLight.RadianceScale := aLight.Intensity / 100;
            spotLight.Location := ConvertVector( aLight.Position );
            spotLight.InnerAngle := (aLight as TSpotLight).InnerAngle;
            spotLight.OuterAngle := (aLight as TSpotLight).OuterAngle;
            spotLight.FallOff := (aLight as TSpotLight).FallOff;
            light := spotLight;
        end;
    end;

    light.CastsShadows := aLight.CastShadows;
    light.Color := RTMiscTp.RGBColor(
        aLight.Diffuse.R,
        aLight.Diffuse.G,
        aLight.Diffuse.B );


    aWorld.AddLight(light);
end;

// Triangulate (fan, only works for convex polygons) and add to scene
procedure TRaytraceManager.AddTrianglesForPolygon(const aPolygon: TPolygon; const aPrimitive: TPrimitive; const aMesh: TRTMesh);
var index: integer;

begin
    for index := 0 to aPolygon.NumVertices - 3 do begin
        aMesh.AddTriangle(
            ConvertVector( aPolygon.VerticeList[0] ),
            ConvertVector( aPolygon.VerticeList[index + 1] ),
            ConvertVector( aPolygon.VerticeList[index + 2] ),
            ConvertVector( aPolygon.VertexNormals[0] ),
            ConvertVector( aPolygon.VertexNormals[index + 1] ),
            ConvertVector( aPolygon.VertexNormals[index + 2] ),
            ConvertUVCoordinate( aPolygon.UVCoordinates[0][0] ),
            ConvertUVCoordinate( aPolygon.UVCoordinates[0][index + 1] ),
            ConvertUVCoordinate( aPolygon.UVCoordinates[0][index + 2] ),
            ConvertMaterial( aPolygon.Material ));
    end;
end;

function TRaytraceManager.ConvertColor(const aColor: TColor): TRGBColor;
begin
    result.R := aColor and $FF / 255;
    result.G := (aColor and $0000FF00) shr 8 / 255;
    result.B := (aColor and $00FF0000) shr 16 / 255;
end;

function TRaytraceManager.ConvertTexture(const aTexture: TTexture): TRTTexture;
var index: integer;
begin
    index := fItem.Scene.MaterialManager.Textures.IndexOf( aTexture );

    if fTextures[index] = nil then
        fTextures[index] := MakeTexture( aTexture );

    result := fTextures[index] as TRTTexture;
end;

function TRaytraceManager.MakeTexture(const aTexture: TTexture): TRTTexture;
var loader: TTextureLoader;
    textureData: TTextureData;
begin
    loader := TTextureLoader.create;
    try
        textureData := loader.LoadTextureData( aTexture.FileName );

    finally
        FreeAndNil( loader );
    end;

    result := TRaytraceTexture.Create( textureData );
end;

function TRaytraceManager.ConvertMaterial(const aMaterial: TMaterial): TRTMaterial;
var index: integer;
begin
    index := fItem.Scene.MaterialManager.IndexOfMaterial(aMaterial);

    if fMaterials[index] = nil then
        fMaterials[index] := MakeMaterial( aMaterial );

    result := fMaterials[index] as TRTMaterial;
end;

function TRaytraceManager.IsTransparentMaterial(const aMaterial: TMaterial): Boolean;
begin
    // TODO, PJ: Implement detection of alpha-blended layers.
    result := ( aMaterial.RaytracingProperties.Transparency > 0 );
end;

function TRaytraceManager.MakeMaterial(const aMaterial: TMaterial): TRTMaterial;
var material: TRTPhongMaterial;
    reflectiveMaterial: TRTReflectiveMaterial;
    transparentMaterial: TRTTransparentMaterial;
    layer: TTextureLayer;

begin
    if IsTransparentMaterial( aMaterial ) then
    begin
        material := TRTTransparentMaterial.Create;
    end
    else if aMaterial.RaytracingProperties.Reflection > 0 then
    begin
        material := TRTReflectiveMaterial.Create;
    end
    else
        material := TRTPhongMaterial.create;

    if IsTransparentMaterial( aMaterial ) then begin
        transparentMaterial := TRTTransparentMaterial(material);
        transparentMaterial.Transparency := aMaterial.RaytracingProperties.Transparency;
        transparentMaterial.RefractiveIndex := aMaterial.RaytracingProperties.Refraction;
    end;

    if aMaterial.RaytracingProperties.Reflection > 0 then begin
        reflectiveMaterial := TRTReflectiveMaterial(material);
        reflectiveMaterial.Reflection := aMaterial.RaytracingProperties.Reflection;
        reflectiveMaterial.ReflectiveExponent := aMaterial.RaytracingProperties.ReflectiveExponent;
        reflectiveMaterial.ReflectiveColor := ConvertColor( aMaterial.RaytracingProperties.ReflectiveColor );

        if aMaterial.RaytracingProperties.ReflectiveExponent > 0 then
            reflectiveMaterial.NumSamples := fItem.Options.NumSamples;
    end;

    material.AmbientReflection := aMaterial.RaytracingProperties.AmbientReflection;
    material.DiffuseReflection := aMaterial.RaytracingProperties.DiffuseReflection;
    material.SpecularReflection := aMaterial.RaytracingProperties.SpecularReflection;
    material.SpecularExponent := aMaterial.RaytracingProperties.SpecularExponent;

    material.Color := COLOR_WHITE;

    if (aMaterial.NumTextureLayers > 0) then begin
        layer := aMaterial.TextureLayers[0];

        case layer.LayerType of
            mlColor:    material.Color := ConvertColor( layer.Color );
            mlTexture:  material.Texture := ConvertTexture( layer.Texture );
        else
            raise Exception.Create( 'Unsupported layer type.' );
        end;


        if layer.LayerType = mlColor then
            material.Color := ConvertColor( layer.Color );

    end;

    result := material;

    // rgbColor := ColorToRGB(aColor);
    // result.R := rgbColor.R;
    // result.G := rgbColor.G;
    // result.B := rgbColor.B;
    // result.Transparency (o.i.d.) := ((aColor and $FF000000) shr 24) / 255;
end;

function TRaytraceManager.ConvertCamera(const aCamera: TCamera3D): TRTCamera;
var camera: TRTPinholeCamera;
begin
    camera := TRTPinHoleCamera.Create;
    camera.Eye := ConvertVector( aCamera.Position );
    camera.ViewDirection := Vector( 0, 0, -1 );
    camera.Pitch := -aCamera.Rotation.X;
    camera.Yaw := -aCamera.Rotation.Y;
    camera.Roll := aCamera.Rotation.Z;
    camera.FOV := aCamera.FOV;

    result := camera;
end;

function TRaytraceManager.ConvertUVCoordinate(const aUVCoordinate: TUVCoord): TRTUVCoordinate;
begin
    result.U := aUVCoordinate.U;
    result.V := aUVCoordinate.V;
end;

function TRaytraceManager.ConvertVector(const aVector: TVertice): TRTVector3D;
begin
    result.X := aVector.X;
    result.Y := aVector.Y;
    result.Z := aVector.Z;
end;

function TRaytraceManager.ConvertSphere(const aPrimitive: TPrimitive): TRTSphere;
var center, vertex: TRTPoint3D;
    radius: Single;
begin
    result := TRTSphere.Create;

    center := ConvertVector( aPrimitive.Center );

    radius := 0;
    if aPrimitive.NumVertices > 0 then begin
        vertex := ConvertVector( aPrimitive.Vertices[0] );
        radius := VectorDistance( vertex, center );
    end;

    result.Center := center;
    result.Radius := radius;

    if aPrimitive.NumPolygons > 0 then begin
        result.Material := ConvertMaterial( aPrimitive.Polygons[0].Material );
        result.OwnsMaterial := FALSE;
    end;
end;

procedure TRaytraceManager.UpdateThumbNail( const aStartPosition, anEndPosition: integer; const anImage: TFloatTextureData );
var thumbnailStart, thumbnailEnd, i: integer;
    targetX, targetY: integer;
    sourceX, sourceY: integer;
    positionX, positionY: integer;
    sourcePosition: integer;
    target: PColor32Entry;
    pixel: PFloatPixel;

begin
    // Translate the progress position in the image to the start/end position within the thumbnail.

    positionX := Trunc(( aStartPosition mod anImage.Width ) * fItem.ThumbNail.Width / anImage.Width );
    positionY := Trunc(( aStartPosition div anImage.Width ) * fItem.ThumbNail.Height / anImage.Height );
    thumbnailStart := positionY * fItem.ThumbNail.Width + positionX;

    positionX := Trunc(( anEndPosition mod anImage.Width ) * fItem.ThumbNail.Width / anImage.Width );
    positionY := Trunc(( anEndPosition div anImage.Width ) * fItem.ThumbNail.Height / anImage.Height );
    thumbnailEnd := positionY * fItem.ThumbNail.Width + positionX;

    // Last batch is special: normally, we never want to run ahead of the source, but the last batch
    // also has to run to the exact end of the thumbnail.
    if anEndPosition >= anImage.Width * anImage.Height - 1 then
        thumbNailEnd := fItem.Thumbnail.Width * fItem.Thumbnail.Height - 1;

    // Simple loop, recalculating source/target positions every iteration.
    // We might be able to make this more efficient by writing a more complex (y,x) loop
    // with start/end conditions.

    target := @fItem.Thumbnail.Bits[thumbnailStart];
    for i := thumbnailStart to thumbnailEnd do
    begin
        targetX := i mod fItem.Thumbnail.Width;
        targetY := i div fItem.Thumbnail.Width;
        sourceX := targetX * anImage.Width div fItem.Thumbnail.Width;
        sourceY := targetY * anImage.Height div fItem.Thumbnail.Height;
        sourcePosition := sourceY * anImage.Width + sourceX;

        pixel := @anImage.Pixels[sourcePosition];
        target.R := Trunc( Min( Pixel.R, 1 ) * 255 );
        target.G := Trunc( Min( Pixel.G, 1 ) * 255 );
        target.B := Trunc( Min( Pixel.B, 1 ) * 255 );

        Inc( target );
    end;
end;

procedure TRaytraceManager.HandleProgress( const aPosition: integer; const aTexture: TFloatTextureData );
begin
    UpdateThumbNail( fLastPosition, aPosition, aTexture );

    fItem.Position := aPosition;
    fLastPosition := aPosition;

    if Assigned( fProgressHandler ) then begin
        fProgressHandler( aPosition, aTexture );
    end;
end;

procedure TRaytraceManager.StoreTextureDataForItem( const anItem: TRaytracingItem );
var filename: string;
    floatWriter: TFloatTextureWriter;
    image: TJPEGImage;
    bitmap: Graphics.TBitmap;

begin
    // We store the raw data
    if anItem.OutputFilename = '' then begin
        anItem.OutputFilename := ChangeFileExt( anItem.SceneFilename, '.drf' );
    end;

    filename := dir_raytracing + anItem.OutputFileName;

    floatWriter := TFloatTextureWriter.Create;
    try
        floatWriter.SaveTextureData( anItem.Output, filename );
    finally
        freeAndNil( floatWriter );
    end;

    // and the thumbnail
    image := TJPEGImage.create;
    bitmap := Graphics.TBitmap.Create;
    try
        image.CompressionQuality := 100;
        bitmap.Assign( anItem.ThumbNail );
        image.Assign( bitmap );
        filename := dir_raytracing + anItem.ThumbFileName;
        image.SaveToFile(filename);
    finally
        FreeAndNil( image );
        FreeAndNil( bitmap );
    end;
end;

procedure TRaytraceManager.Raytrace;
var
  index: Integer;
  world: TRTRenderWorld;
  primitive: TBasePrimitive;
begin
    fRenderingStartTime := GetTickCount;

    //I removed the old Scene.InitializeNormals method.
    //The loop will be executed here.
    with fItem.Scene do
      for index := 0 to Primitives.Count - 1 do
      begin
          primitive := Primitives[index];
          primitive.InitializeNormals(fItem.Options.CreaseAngle);
      end;

    world := TRTRenderWorld.Create( fItem.Output );

    // INFO, PJ: Order is important, fItem.Output might change status to new (change some day).
    fItem.Status := risProcessing;

    InitializeWorld( world );

    world.ViewPlane.HRes := fItem.Options.Width;
    world.ViewPlane.VRes := fItem.Options.Height;
    world.ViewPlane.MaxDepth := fItem.Options.MaxDepth;
    world.ViewPlane.NumSamples := fItem.Options.NumSamples;
    world.ViewPlane.ToneMapping := tmNone;
    world.BackgroundColor := ConvertColor( fItem.Options.BackgroundColor );
    world.Tracer := TRTRayCastTracer.Create( world );

//    fEngine.UseAmbientOcclusion := fItem.Options.UseAmbientOcclusion;

    fCamera := ConvertCamera( fItem.Camera );

    fCamera.RenderPosition := fItem.Position;
    fCamera.Render( world );

    StoreTextureDataForItem( fItem );

    // INFO, PJ: Flush output cache.
    fItem.Invalidate;

    fItem.Position := fCamera.RenderPosition;
    fItem.RenderingTime := fItem.RenderingTime + Integer(GetTickCount - fRenderingStartTime);

    // INFO, PJ: If not stopped half-way, the item was fully rendered.
    if fCamera.ShouldStop then
        fItem.Status := risPaused
    else
        fItem.Status := risDone;

    FreeAndNil(fCamera);
    FreeAndNil(world);
end;

// INFO, PJ: Only method that might be called from a different thread.
procedure TRaytraceManager.Stop;
var camera: TRTCamera;
begin
    camera := fCamera;
    if Assigned( camera ) then begin
        camera.ShouldStop := TRUE;
    end;
end;

// ----------------------------------------------------------------------------

constructor TRaytraceTexture.Create( const aData: TTextureData );
begin
    inherited Create;

    fTextureData := aData;
end;

function TRaytraceTexture.SampleColor( const aShadeRec: TRTShadeRec ): TRGBColor;
var color: PTexturePixel24;
    x, y: Integer;
    index: Integer;
    
begin
    x := Trunc( aShadeRec.UVCoordinate.U * fTextureData.Width );
    y := Trunc( aShadeRec.UVCoordinate.V * fTextureData.Height );

    // INFO, PJ: Need an "unsigned mod" operator here. Or unsigned Frac().
    x := x mod fTextureData.Width;
    if x < 0 then x := x + fTextureData.Width;

    y := y mod fTextureData.Height;
    if y < 0 then y := y + fTextureData.Height;

    index := ( y * fTextureData.Width + x ) * ( fTextureData.BitsPerPixel div 8 );
//    if fTextureData.BitsPerPixel = 32 then Inc( index );
    color := @fTextureData.Pixels[index];

    try
        result.R := color[2] / 255;
        result.G := color[1] / 255;
        result.B := color[0] / 255;
    except
        result.R := 0;
    end;
end;

// ----------------------------------------------------------------------------

procedure TestRaytracer;
var
  manager: TRaytraceManager;
  item: TRaytracingItem;
  options: TRaytracingOptions;
//  form: TForm;
  
begin
  try
      dir_raytracing := 'D:\scratch\raytracing\';
      dir_texture := 'D:\Applications\Graphics\DeleD PRO\Textures\';

      item := TRaytracingItem.Create;

      item.SceneName := 'test';
      item.SceneFileName := 'CastleSiege_Lighted.dxs';
      item.ThumbFileName := 'CastleSiege_Lighted_Thump.jpg';

      item.Camera.FOV := 85;
      item.Camera.Position.X := 39.12697;
      item.Camera.Position.Y := 660.629;
      item.Camera.Position.Z := 639.9786;
      item.Camera.Rotation.X := 48;
      item.Camera.Rotation.Y := 343;
      item.Camera.Rotation.Z := 0;

      options.Width := 128;
      options.Height := 128;
      options.NumSamples := 1;
      options.MaxDepth := 8;
      options.CreaseAngle := 66;
      options.BackgroundColor := 0;

      item.Options := options;

      manager := TRaytraceManager.Create( item );
      manager.RayTrace;

      FreeAndNil( manager );

//      form := TfrmRaytracingPreview.Create( item );
//      form.ShowModal;
//      FreeAndNil( form );

      FreeAndNil( item );

  except
      on e: Exception do
      begin
          ShowMessage( e.Message );
      end;
  end;
end;


end.
