{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_RaytracingThread;

{$MODE Delphi}

interface

uses
    Classes, Contnrs, Graphics, FileUtil,
    unit_Scene, unit_Cameras, unit_XMLReader, unit_XMLWriter, unit_Types,
    unit_TextureTypes, unit_RaytraceManager,
    frmRaytracingOptionsForm,
    RTRenderWorldCl;

type
    TRaytracingThread = class( TThread )
    private
        fShouldStop: boolean;

        fQueue: TObjectList;
        fQueueJumper: TRaytracingItem;
        fManager: TRaytraceManager;

        fProgressHandler: TRTProgressHandler;

        procedure Activate;
        procedure ProcessQueue;

        function MakeFilenameForScene( const aScene: TScene ): string;

        procedure StoreQueue;
        procedure RestoreQueue;

        function GetNextQueueItem: TRaytracingItem;
        function GetItemForIndex( const anIndex: integer ): TRaytracingItem;

        procedure RenderItem( const anItem: TRaytracingItem );
        procedure StopItemEx( const anItem: TRaytracingItem );
        procedure TerminateAndWait; // TODO, PJ: Move to destructor?

        procedure HandleProgress( const aPosition: integer; const aTexture: TFloatTextureData );

    public
        constructor Create;
        destructor Destroy; override;

        procedure Execute; override;

        procedure AddToQueue( const aScene: TScene; const aCamera: TCamera3D; const anOptions: TRaytracingOptions; const aThumbnail: TTextureData );
        procedure RemoveFromQueue( const anItemIndex: integer );

        // Full start/stop of the raytracing thread.
        procedure Start;
        procedure Stop;

        procedure StartItem( const anItem: TRaytracingItem );
        procedure StopItem( const anItem: TRaytracingItem );

        function NumItems: integer;

        property OnProgress: TRTProgressHandler read fProgressHandler write fProgressHandler;
        property Items[const anIndex: integer]: TRaytracingItem read GetItemForIndex;
    end;

// ----------------------------------------------------------------------------


    TRaytracingQueueReader = class( TObject )
    private
        procedure AppendQueue( const aNode: TXMLNode; const aQueue: TObjectList );
        procedure AppendItem( const aNode: TXMLNode; const aQueue: TObjectList );
        function ReadOptions( const aNode: TXMLNode ): TRaytracingOptions;
        procedure AppendCamera( const aNode: TXMLNode; const anItem: TRaytracingItem );
        procedure ReadVertex( const aNode: TXMLNode; const aVertex: TVertice );
        function ReadColor( const aNode: TXMLNode ): TColor;
        procedure UpdateStatus( const aQueue: TObjectList );
        function StatusForDescription( const aDescription: string ): TRaytracingItemStatus;

    public
        procedure ReadFromFile( const aFileName: string; const aQueue: TObjectList );
    end;

// ----------------------------------------------------------------------------

    TRaytracingQueueWriter = class( TObject )
    private
        procedure AppendQueue( const aQueue: TObjectList; const aWriter: TXMLWriter );
        procedure AppendItem( const anItem: TRaytracingItem; const aWriter: TXMLWriter );
        procedure AppendOptions( const anOptions: TRaytracingOptions; const aWriter: TXMLWriter );

    public
        procedure WriteToFile( const aFileName: string; const aQueue: TObjectList );
    end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
    SysUtils, Windows, Dialogs,
    unit_Constants, unit_Variables, unit_TextureWriter, unit_TextureLoader,
    unit_XMLSceneReader, unit_XMLSceneWriter, unit_Functions,frmMainForm;

const
    QUEUE_FILENAME = 'queue.xml';
    STATUS_DESCRIPTION: Array[TRaytracingItemStatus] of string = ( 'new', 'processing', 'paused', 'done' );

constructor TRaytracingThread.Create;
begin
    inherited Create( FALSE );

    fQueue := TObjectList.Create( TRUE );
end;

destructor TRaytracingThread.Destroy;
begin
    TerminateAndWait;

    FreeAndNil( fQueue );
    FreeAndNil( fManager );

    inherited Destroy;
end;

procedure TRaytracingThread.Execute;
begin
    RestoreQueue;
//    CleanupDirectory;

    try
        while not Terminated do begin
            ProcessQueue;

            Suspend;
        end;

    except
        on e: Exception do begin
            showMessage(
                'The raytracing thread terminated unexpectedly. You are advised to close and restart the application.' + CRLF +
                CRLF +
                'Error: ' + e.Message );
        end;
    end;

    StoreQueue;
end;


procedure TRaytracingThread.TerminateAndWait;
begin
    inherited Terminate;
    fShouldStop := TRUE;
    Stop;
    Activate;
    WaitFor;
end;

// Wakes up the thread (only if it is sleeping).
procedure TRaytracingThread.Activate;
begin
    if Suspended then inherited Resume;
end;

function TRaytracingThread.MakeFilenameForScene( const aScene: TScene ): string;
var index: integer;
begin
    index := 1;
    repeat
        result := Format( '%s.%.3d.dxs', [ChangeFileExt( ExtractFileName( aScene.FileName ), '' ), index] );
        index := index + 1;
    until not FileExistsUTF8(dir_raytracing + result );
end;

procedure TRaytracingThread.AddToQueue( const aScene: TScene; const aCamera: TCamera3D; const anOptions: TRaytracingOptions; const aThumbnail: TTextureData );
var item: TRaytracingItem;
    fileName: string;
    thumbFilename: string;
    writer: TXMLSceneWriter;
    textureWriter: TTextureWriter;

begin
    // Resolve Raytracing folder (bit hacky last-minute resort)
    CreateDirUTF8(dir_raytracing );

    // Copy scene to Raytracing folder
    fileName := MakeFileNameForScene( aScene );
    thumbFilename := ChangeFileExt( fileName, ''  ) + '_thumb' + '.jpg';

    writer := TXMLSceneWriter.Create;
    try
        writer.WriteToFile( aScene, dir_raytracing + fileName,frmMain.ViewPortManager );
    finally
        FreeAndNil( writer );
    end;

    // Copy thumbnail to Raytracing folder
    textureWriter := TTextureWriter.Create;
    try
        textureWriter.CompressionQuality := 90;
        textureWriter.SaveTextureData( aThumbnail, dir_raytracing + thumbFilename );
    finally
        FreeAndNil( textureWriter );
    end;

    item := TRaytracingItem.Create;
    item.Camera.Assign( aCamera );
    item.SceneFileName := fileName;
    item.ThumbFilename := thumbFilename;
    item.SceneName := aScene.Name;
    item.Options := anOptions;

    // INFO, PJ: This might have to be locked: don't read from the queue while an item is being added.
    fQueue.Add( item );


    // INFO, PJ:
    //   If "we are doing nothing" (all items pauzed), we expect the new item to start immediately,
    //   as opposed to one of the pauzed items continuing.
    if Suspended then fQueueJumper := item;

    fShouldStop := FALSE;
    Activate;
end;

procedure TRaytracingThread.RemoveFromQueue( const anItemIndex: integer );
var item: TRaytracingItem;
begin
    // INFO, PJ:
    //   Run from the main thread, and not very threadsafe.
    //   Running item could change half-way this method (refa).
    if (anItemIndex >= 0) and (anItemIndex < fQueue.Count) then
    begin
        item := fQueue[anItemIndex] as TRaytracingItem;

        if item.Status = risProcessing then
        begin
            StopItem( item );
        end;

        DeleteFileUTF8(dir_raytracing + item.SceneFileName );
        DeleteFileUTF8(dir_raytracing + item.OutputFileName );
        DeleteFileUTF8(dir_raytracing + item.ThumbFileName );

        // REFA, PJ: We should probably make _sure_ the queue gets properly saved after this.
        // INFO, PJ: This might have to be locked: don't remove the item
        fQueue.Remove( item );
    end;
end;

// INFO, PJ: Run from inside the render thread. Make sure to keep things that way!
procedure TRaytracingThread.RenderItem( const anItem: TRaytracingItem );
begin
  try

        // INFO, PJ: Create a new one and start it
        fManager := TRaytraceManager.Create( anItem );

        try
            fManager.OnProgress := HandleProgress;
            fManager.Raytrace;

        finally
            FreeAndNil( fManager );
        end;
  except
      // INFO, PJ: Repress any errors, and try (really hard) to continu
  end;
end;

procedure TRaytracingThread.Start;
begin
    // INFO, PJ: Full (Re)start of the thread, possibly after a full stop.
    fQueue.Clear;
    RestoreQueue;
    fShouldStop := FALSE;
    Activate;
end;

procedure TRaytracingThread.Stop;
var i: integer;
    item: TRaytracingItem;

begin
    // INFO, PJ:
    //   Stops all activity.
    //   We disable all items that are set to processing (some erroneously)
    for i := 0 to NumItems - 1 do begin
        item := Items[i] as TRaytracingItem;

        if{ (item <> anItem) and} (item.Status = risProcessing) then
            StopItemEx( Items[i] );
    end;

    // INFO, PJ:
    //   Store the queue. We might be closing DeleD, or changing dir_raytracing.
    StoreQueue;
end;

// INFO, PJ: Called from Outside the thread(!)
procedure TRaytracingThread.StartItem( const anItem: TRaytracingItem );
begin
    Stop;

    // INFO, PJ: The item we explicitly started jumps the queue.
    fQueueJumper := anItem;
    fShouldStop := FALSE;
    Activate;
end;

// INFO, PJ: Called from Outside the thread(!)
procedure TRaytracingThread.StopItem( const anItem: TRaytracingItem );
begin
    StopItemEx( anItem );
end;

procedure TRaytracingThread.StopItemEx( const anItem: TRaytracingItem );
begin
    fShouldStop := TRUE;

    if Assigned( fManager ) then
    begin
        fManager.Stop;

      // INFO, PJ: Wait for the manager (which is owned by the running thread, not by us) to be released
        while Assigned( fManager ) do begin
            Sleep( 50 );
        end;
    end;

    anItem.Status := risPaused;
    anItem.Invalidate;
end;

// REFA, PJ: This should probably be in a critical section
function TRaytracingThread.GetNextQueueItem: TRaytracingItem;
var i: integer;
    item: TRaytracingItem;
begin
    result := fQueueJumper;

    if result <> nil then begin
        fQueueJumper := nil;
        exit;
    end;


    for i := 0 to fQueue.Count - 1 do begin
        item := fQueue[i] as TRaytracingItem;

        if item.Status <> risDone then begin
            result := item;
            break;
        end;
    end;
end;

procedure TRaytracingThread.ProcessQueue;
var item: TRaytracingItem;
begin
    item := GetNextQueueItem;
    while (item <> nil) and not fShouldStop do begin
        RenderItem( item );
        item := GetNextQueueItem;
    end;
end;

procedure TRaytracingThread.StoreQueue;
var writer: TRaytracingQueueWriter;
begin
    writer := TRaytracingQueueWriter.Create;
    try
        writer.WriteToFile( dir_raytracing + QUEUE_FILENAME, fQueue );
    finally
        FreeAndNil( writer );
    end;
end;

procedure TRaytracingThread.RestoreQueue;
var fileName: string;
    reader: TRaytracingQueueReader;
begin
    fileName := dir_raytracing + QUEUE_FILENAME;

    reader := TRaytracingQueueReader.Create;
    try
        if FileExistsUTF8(fileName ) then
            reader.ReadFromFile( fileName, fQueue );
            
    finally
        FreeAndNil( reader );
    end;
end;

procedure TRaytracingThread.HandleProgress( const aPosition: integer; const aTexture: TFloatTextureData );
begin
    if Assigned( fProgressHandler ) then begin
        fProgressHandler( aPosition, aTexture );
    end;

    fShouldStop := Terminated;
end;

function TRaytracingThread.NumItems: integer;
begin
    result := fQueue.Count;
end;

function TRaytracingThread.GetItemForIndex( const anIndex: integer ): TRaytracingItem;
begin
    result := fQueue[anIndex] as TRaytracingItem;
end;

// ----------------------------------------------------------------------------

procedure TRaytracingQueueReader.ReadFromFile( const aFileName: string; const aQueue: TObjectList );
var reader: TXMLReader;
    node: TXMLNode;
begin
    reader := TXMLFileReader.Create( aFileName );

    try
        node := reader.RootNode;

        AppendQueue( node, aQueue );

        UpdateStatus( aQueue );

    finally
        FreeAndNil( reader );
        FreeAndNil( node );
    end;
end;

// INFO, PJ: We pause any item left running
procedure TRaytracingQueueReader.UpdateStatus( const aQueue: TObjectList );
var i: integer;
    item: TRaytracingItem;
begin
    for i := 0 to aQueue.Count - 1 do begin
        item := aQueue[i] as TRaytracingItem;

        if item.Status = risProcessing then
            item.Status := risPaused;
    end;
end;

function TRaytracingQueueReader.StatusForDescription( const aDescription: string ): TRaytracingItemStatus;
begin
    for result := High(TRaytracingItemStatus) downto Low(TRaytracingItemStatus) do begin
        if STATUS_DESCRIPTION[result] = aDescription then
            break;
    end;
end;

procedure TRaytracingQueueReader.AppendQueue( const aNode: TXMLNode; const aQueue: TObjectList );
var i: integer;
    node: TXMLNode;

begin
    if aNode.name <> 'queue' then
        raise Exception.Create( Format( 'Unexpected root node %s encountered.', [aNode.name] ));

    for i := 0 to aNode.NumChildNodes - 1 do begin
        node := aNode.ChildNodes[i];

        if node.Name = 'item' then
            AppendItem( node, aQueue );
    end;
end;

procedure TRaytracingQueueReader.AppendItem( const aNode: TXMLNode; const aQueue: TObjectList );
var item: TRaytracingItem;
    i: integer;
    node: TXMLNode;
begin
    item := TRaytracingItem.Create;

    item.SceneName := aNode.Properties['scene'];
    item.SceneFileName := aNode.Properties['scenefile'];
    item.ThumbFilename := aNode.Properties['thumbfile'];
    item.Status := StatusForDescription( aNode.Properties['status'] );

    if aNode.HasProperty( 'renderingTime' ) then
        item.RenderingTime := aNode.IntegerProperties['renderingTime'];

    if aNode.HasProperty( 'position' ) then
        item.Position := aNode.IntegerProperties['position'];

    if aNode.HasProperty( 'outputfile' ) then
        item.OutputFilename := aNode.Properties['outputfile'];

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'camera' then
            AppendCamera( node, item )
        else if node.Name = 'options' then
            item.Options := ReadOptions( node );
    end;

    aQueue.Add( item )
end;

procedure TRaytracingQueueReader.AppendCamera( const aNode: TXMLNode; const anItem: TRaytracingItem );
var i: integer;
    node: TXMLNode;
begin
    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        anItem.Camera.FOV := aNode.FloatProperties['fov'];

        if node.Name = 'position' then
            ReadVertex( node, anItem.Camera.Position )
        else if node.Name = 'rotation' then
            ReadVertex( node, anItem.Camera.Rotation );
    end;
end;

function TRaytracingQueueReader.ReadOptions( const aNode: TXMLNode ): TRaytracingOptions;
var i: Integer;
    node: TXMLNode;
begin
    result.Width := aNode.IntegerProperties['width'];
    result.Height := aNode.IntegerProperties['height'];
    result.MaintainAspectRatio := aNode.BooleanProperties['maintainAspectRatio'];
    result.CreaseAngle := aNode.IntegerProperties['creaseAngle'];
    result.NumSamples := aNode.IntegerProperties['numSamples'];
    result.MaxDepth := aNode.IntegerProperties['maxDepth'];

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];
        if node.Name = 'backgroundColor' then
            result.BackgroundColor := ReadColor( node );
    end;

//    result.UseAmbientOcclusion := aNode.BooleanProperties['useAmbientOcclusion'];
//    result.AmbientOcclusionNumSamples := aNode.IntegerProperties['ambientOcclusionNumSamples'];
//    result.AmbientOcclusionSampleDistance := aNode.FloatProperties['ambientOcclusionSampleDistance'];
//    result.AmbientOcclusionIntensity := aNode.IntegerProperties['ambientOcclusionIntensity'];
end;

function TRaytracingQueueReader.ReadColor( const aNode: TXMLNode ): TColor;
begin
    result := aNode.IntegerProperties['r'];
    result := result or (aNode.IntegerProperties['g'] shl 8);
    result := result or (aNode.IntegerProperties['b'] shl 16);
    result := result or Integer($FF000000);
end;

procedure TRaytracingQueueReader.ReadVertex( const aNode: TXMLNode; const aVertex: TVertice );
begin
    aVertex.X := aNode.FloatProperties['x'];
    aVertex.Y := aNode.FloatProperties['y'];
    aVertex.Z := aNode.FloatProperties['z'];
end;

// ----------------------------------------------------------------------------

procedure TRaytracingQueueWriter.WriteToFile( const aFileName: string; const aQueue: TObjectList );
var writer: TXMLWriter;
begin
    writer := TXMLFileWriter.Create( aFileName );

    try
        AppendQueue( aQueue, writer );
    finally
        FreeAndNil( writer );
    end;
end;

procedure TRaytracingQueueWriter.AppendQueue( const aQueue: TObjectList; const aWriter: TXMLWriter );
var i: integer;
    item: TRaytracingItem;
begin
    aWriter.AppendOpenTag( 'queue' );

    for i := 0 to aQueue.Count - 1 do begin
        item := aQueue[i] as TRaytracingItem;

        AppendItem( item, aWriter );
    end;

    aWriter.AppendCloseTag( 'queue' );
end;

procedure TRaytracingQueueWriter.AppendItem( const anItem: TRaytracingItem; const aWriter: TXMLWriter );
begin
    aWriter.AppendOpenTag( 'item' );

    aWriter.AppendProperty( 'scene', anItem.SceneName );
    aWriter.AppendProperty( 'sceneFile', anItem.SceneFileName );
    aWriter.AppendProperty( 'thumbFile', anItem.ThumbFilename );
    aWriter.AppendProperty( 'status', STATUS_DESCRIPTION[anItem.Status] );
    aWriter.AppendProperty( 'renderingTime', anItem.RenderingTime );

    if anItem.Status in [risProcessing, risPaused] then
        aWriter.AppendProperty( 'position', anItem.Position );

    aWriter.AppendProperty( 'outputfile', anItem.OutputFilename );

    aWriter.AppendOpenTag( 'camera' );
    aWriter.AppendProperty( 'fov', anItem.Camera.FOV );
    aWriter.AppendOpenTag( 'position' );
    aWriter.AppendProperty( 'x', anItem.Camera.Position.X );
    aWriter.AppendProperty( 'y', anItem.Camera.Position.Y );
    aWriter.AppendProperty( 'z', anItem.Camera.Position.Z );
    aWriter.AppendCloseTag( 'position' );
    aWriter.AppendOpenTag( 'rotation' );
    aWriter.AppendProperty( 'x', anItem.Camera.Rotation.X );
    aWriter.AppendProperty( 'y', anItem.Camera.Rotation.Y );
    aWriter.AppendProperty( 'z', anItem.Camera.Rotation.Z );
    aWriter.AppendCloseTag( 'rotation' );
    aWriter.AppendCloseTag( 'camera' );

    AppendOptions( anItem.Options, aWriter );

    aWriter.AppendCloseTag( 'item' );
end;

procedure TRaytracingQueueWriter.AppendOptions( const anOptions: TRaytracingOptions; const aWriter: TXMLWriter );
begin
    aWriter.AppendOpenTag( 'options' );

    aWriter.AppendProperty( 'width', anOptions.Width );
    aWriter.AppendProperty( 'height', anOptions.Height );
    aWriter.AppendProperty( 'numSamples', anOptions.NumSamples );
    aWriter.AppendProperty( 'maxDepth', anOptions.MaxDepth );
    aWriter.AppendProperty( 'maintainAspectRatio', anOptions.MaintainAspectRatio );
    aWriter.AppendProperty( 'creaseAngle', anOptions.CreaseAngle );

    aWriter.AppendOpentag( 'backgroundColor' );
    aWriter.AppendProperty( 'r', ( anOptions.BackgroundColor and $0000FF ));
    aWriter.AppendProperty( 'g', ( anOptions.BackgroundColor and $00FF00 ) shr 8);
    aWriter.AppendProperty( 'b', ( anOptions.BackgroundColor and $FF0000 ) shr 16);
    aWriter.AppendCloseTag( 'backgroundColor' );

//    aWriter.AppendProperty( 'useAmbientOcclusion', anOptions.UseAmbientOcclusion );
//    aWriter.AppendProperty( 'ambientOcclusionNumSamples', anOptions.AmbientOcclusionNumSamples );
//    aWriter.AppendProperty( 'ambientOcclusionSampleDistance', anOptions.AmbientOcclusionSampleDistance );
//    aWriter.AppendProperty( 'ambientOcclusionIntensity', anOptions.AmbientOcclusionIntensity );

    aWriter.AppendCloseTag( 'options' );
end;

end.
