{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_RenderQueue;

{$MODE Delphi}

// Optimalisatie: RenderQueue krijgt een invalidate, die moet vanuit alle edit
//                operaties gezet worden.
//                Het netst zou zijn om Update vanuit de GetMethods aan te roepen,
//                maar ik vrees dat dat te duur wordt. Toch maar expliciet houden?
//
// 2e optimalisatie: capacity zetten van tevoren?
// 3e optimalisatie: preformatted opengl vertex arrays bufferen


interface

uses
    ContNrs, Classes,
    unit_Types,
    unit_Cameras,
    unit_Scene,
    unit_Variables,
    unit_PrimitiveList,
    unit_PathList,
    unit_Paths;

type
    TSolidRenderList = class;
    TTransparentRenderList = class;

    TRenderQueue = class(TObject)
    private
        fScene: TScene;
        fSolidFaces: TSolidRenderList;
        fTransparentFaces: TTransparentRenderList;
        fPaths: TPathList;
        fIsUpToDate: boolean;

        procedure AddPrimitives( const aPrimitives: TPrimitiveList );

    protected
        function GetSolidFace(const anIndex: Integer): TPolygon;
        function GetTransparentFace(const anIndex: Integer): TPolygon;
        function GetPath(const anIndex: Integer): TPath;

    public
        property Scene: TScene read fScene write fScene;

        property SolidFaces[const anIndex: Integer]: TPolygon read GetSolidFace;
        property TransparentFaces[const anIndex: Integer]: TPolygon read GetTransparentFace;
        property Paths[const anIndex: Integer]: TPath read GetPath;

        function NumSolidFaces: Integer;
        function NumTransparentFaces: Integer;
        function NumPaths: Integer;

        procedure Invalidate;
        procedure Update;
        procedure SortTransparentFaces(const aCamera: TCamera3d);

        constructor Create(Ascene: TScene);
        destructor Destroy; override;
    end;

// ############################################################################

    TSolidRenderList = class(TObjectList)
    protected
        function GetPolygon(const anIndex: Integer): TPolygon;

    public
        property Polygons[const anIndex: Integer]: TPolygon read GetPolygon; default;

        procedure Sort;

        constructor Create;
    end;

// ############################################################################

    TRenderListFaceInfo = record
        Polygon: TPolygon;
        SquaredDistanceToCamera: Single;
    end;
    PRenderListFaceInfo = ^TRenderListFaceInfo;

    TTransparentRenderList = class(TList)
    private
        procedure UpdateSquaredDistancesToCamera(const aCamera: TCamera3D);
    protected
        procedure Notify(aPointer: Pointer; anAction: TListNotification); override;
        function GetPolygon(const anIndex: Integer): TPolygon;

    public
        property Polygons[const anIndex: Integer]: TPolygon read GetPolygon; default;

        procedure Add(const aPolygon: TPolygon);
        procedure Sort(const aCamera: TCamera3D);
    end;

// ############################################################################

implementation

uses
    SysUtils,
    unit_Enums,
    Math,
    unit_Primitives,
    unit_Materials,
    unit_Math;

procedure TRenderQueue.AddPrimitives( const aPrimitives: TPrimitiveList );
var i, j: integer;
    primitive: TBasePrimitive;
    poly: TPolygon;
begin
    for i := 0 to aPrimitives.Count - 1 do begin
        primitive := aPrimitives.Get(i);

        if primitive.Visible then begin
          if primitive is TPath then begin
            fPaths.Add(primitive);
          end
          else
            for j := 0 to primitive.NumPolygons - 1 do begin
                poly := primitive.Polygons[j];

                if poly.Material.IsTransparent then
                    fTransparentFaces.Add(poly)
                else
                    fSolidFaces.Add(poly);
            end;
        end;
    end;
end;

function TRenderQueue.GetSolidFace(const anIndex: Integer): TPolygon;
begin
    result := fSolidFaces[anIndex];
end;

function TRenderQueue.GetTransparentFace(const anIndex: Integer): TPolygon;
begin
    result := fTransparentFaces[anIndex];
end;

function TRenderQueue.GetPath(const anIndex: Integer): TPath;
begin
  Result := fPaths[anIndex];
end;

procedure TRenderQueue.Invalidate;
begin
    fIsUpToDate := false;
end;

procedure TRenderQueue.Update;
var i: integer;
    primitives: TPrimitiveList;
begin
    if not fIsUpToDate then begin
        fTransparentFaces.Clear;
        fSolidFaces.Clear;
        fPaths.Clear;

        // get primitives
        primitives := Scene.Primitives.GetItems(TPrimitive);
        AddPrimitives( primitives );
        FreeAndNil( primitives );

        // get paths
        primitives := Scene.Primitives.GetItems(TPath);
        AddPrimitives( primitives );

        FreeAndNil( primitives );

        for i := 0 to Scene.Skeletons.Count - 1 do
            AddPrimitives( Scene.Skeletons[i].AssignedPrimitives );

        fSolidFaces.Sort;

        fIsUpToDate := true;
    end;
end;

procedure TRenderQueue.SortTransparentFaces(const aCamera: TCamera3d);
begin
    fTransparentFaces.Sort(aCamera);
end;

function TRenderQueue.NumSolidFaces: Integer;
begin
    result := fSolidFaces.Count;
end;

function TRenderQueue.NumTransparentFaces: Integer;
begin
    result := fTransparentFaces.Count;
end;

function TRenderQueue.NumPaths: Integer;
begin
  Result := fPaths.Count;
end;

constructor TRenderQueue.Create(Ascene: TScene);
begin
    inherited Create;

    fScene := AScene;
    fSolidFaces := TSolidRenderList.Create;
    fTransparentFaces := TTransparentRenderList.Create;
    fPaths := TPathList.Create(false);
end;

destructor TRenderQueue.Destroy;
begin
    FreeAndNil(fSolidFaces);
    FreeAndNil(fTransparentFaces);
    FreeAndNil(fPaths);

    inherited Destroy;
end;

// ############################################################################

function CompareSolidFaces(anItem1, anItem2: Pointer): Integer;
// Just compare on address.
var material1, material2: Integer;
begin
    material1 := integer(@TPolygon(anItem1).Material);
    material2 := integer(@TPolygon(anItem2).Material);

    if material1 < material2 then
        result := 1
    else if material1 > material2 then
        result := -1
    else
        result := 0;
end;

function TSolidRenderList.GetPolygon(const anIndex: Integer): TPolygon;
begin
    result := Items[anIndex] as TPolygon;
end;

procedure TSolidRenderList.Sort;
begin
    Inherited Sort(CompareSolidFaces);
end;

constructor TSolidRenderList.Create;
begin
    Inherited Create(false);
end;


// ############################################################################

function CompareTransparentFaces(anItem1, anItem2: Pointer): Integer;
var distance1, distance2: Single;
begin
    result := 0;

    distance1 := PRenderListFaceInfo(anItem1).SquaredDistanceToCamera;
    distance2 := PRenderListFaceInfo(anItem2).SquaredDistanceToCamera;

    // NaN comparing is undefined -> compare always returns 1.
    // Delphi's quicksort does not handle this well and starts providing bogus pointers.
    if IsNaN( distance1 ) or IsNaN( distance2 ) then
        exit;

    // Reverse sort: faces that are further away should end up in the start of the list.
    if distance1 < distance2 then
        result := 1
    else if distance1 > distance2 then
        result := -1;
end;

procedure TTransparentRenderList.Notify(aPointer: Pointer; anAction: TListNotification);
begin
    if anAction in [lnDeleted, lnExtracted] then Dispose(aPointer);
end;

function TTransparentRenderList.GetPolygon(const anIndex: Integer): TPolygon;
begin
    result := PRenderListFaceInfo(Items[anIndex]).Polygon;
end;

procedure TTransparentRenderList.UpdateSquaredDistancesToCamera(const aCamera: TCamera3D);
var i: integer;
    faceInfo: PRenderListFaceInfo;
    cameraPosition: TVertex;

begin
    cameraPosition := verticeToVertex(aCamera.Position);
    cameraPosition.X := cameraPosition.X;
    cameraPosition.Y := cameraPosition.Y;
    cameraPosition.Z := cameraPosition.Z;

    for i := 0 to Count - 1 do begin
        faceInfo := PRenderListFaceInfo(Items[i]);

        faceInfo.SquaredDistanceToCamera :=
            vertexSquaredLength( vertexSubtract( cameraPosition,
            verticeToVertex(faceInfo.Polygon.BoundingBoxCenter)));
    end;
end;

procedure TTransparentRenderList.Add(const aPolygon: TPolygon);
var faceInfo: PRenderListFaceInfo;
begin
    New(faceInfo);
    faceInfo.Polygon := aPolygon;
    faceInfo.SquaredDistanceToCamera := 0;

    Inherited Add(faceInfo);
end;

procedure TTransparentRenderList.Sort(const aCamera: TCamera3D);
begin
    UpdateSquaredDistancesToCamera(aCamera);

    Inherited Sort(CompareTransparentFaces);
end;

// ############################################################################

end.
