{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Scene;

{$MODE Delphi}

interface

uses Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
     Dialogs, Contnrs, Types, StrUtils,
     frmProgressForm,
     frmOGLContextForm,
     unit_Enums,
     unit_Types,
     unit_PluginVars,
     unit_Functions,
     unit_Lights,
     unit_Primitives,
     unit_PrimitiveList,
     unit_Paths,
     unit_PathList,
     unit_PathControlPointList,
     unit_PathSelector,
     unit_PathControlPointSelector,
     unit_IntegerList,
     unit_Materials,
     unit_MaterialList,
     unit_MaterialManager,
     unit_Textures,
     unit_UndoRedoManager,
     unit_DPFLoader,
     unit_ParamList,
     unit_Skeleton,
     unit_Joint,
     unit_Animation,
     unit_PrimitiveSelector,
     unit_JointSelector,
     unit_JointGeometrySelectingTool,
     unit_JointGeometryUpdatingTool,
     unit_PolygonSelector,
     unit_EdgeSelector,
     unit_VertexSelector,
     unit_UVSelector,
     unit_ToolFactory,
     unit_ToolExecuter;

type
  TScene = class(TObject)   // root class
  private
    fName:            string;
    fAuthor:          string;
    fComments:        string;
    fShadowOpacity:   integer;
    fAmbientColor:    TRGBColor;
    fFileName:        string;
    fMaterialManager: TMaterialManager;
    fHasChanged:      boolean;
    fFileSaved:       boolean;
    fGizmo:           TGizmo;  //Should not be here!!

    // collections
    fPrimitives:      TPrimitiveList;
    fSkeletons:       TSkeletonList;

    // selections
    fSelectedPrimitives:    TPrimitiveList;
    fSelectedSkeletons:     TSkeletonList;
    fSelectedJoints:        TJointList;
    fSelectedAnimations:    TAnimationList;
    fSelectedPolygons:      TPolygonList;
    fSelectedEdges:         TEdgeList;
    fSelectedVertices:      TVerticeList;
    fSelectedUVCoordinates: TUVList;
    fSelectedControlPoints: TPathControlPointList;

    // selectors
    fPrimitiveSelector:     TPrimitiveSelector;
    fJointSelector:         TJointSelector;
    fJointGeometrySelector: TJointGeometrySelectingTool;
    fPolygonSelector:       TPolygonSelector;
    fEdgeSelector:          TEdgeSelector;
    fVertexSelector:        TVertexSelector;
    fUVSelector:            TUVSelector;
    fPathSelector:          TPathSelector;
    fControlPointSelector:  TPathControlPointSelector;

    // tools
    fToolFactory:           TToolFactory;
    fToolExecuter:          TToolExecuter;
    fUndoRedoManagerEx:     TUndoRedoManagerEx;
    fJointGeometryUpdater:  TJointGeometryUpdatingTool;

    // events
    fOnCollectionChange:    TNotifyEvent;
    fOnSelectionChange:     TNotifyEvent;
    fOnLoadPrefab:          TNotifyEvent;
    fOnAddJoint:            TNotifyEvent;

    // chrisu: Fields for OrthographicDistanceShading.
    fMaxVertexCoords:       TVertice;
    fMinVertexCoords:       TVertice;
    // chrisu: ----

    // misc functions
    procedure Initialize;
    function GetSelectedPolygons: TPolygonList;
    function GetSelectedEdges: TEdgeList;
    function GetSelectedVertices: TVerticeList;
    function GetSelectedUVCoordinates: TUVList;
    function GetSelectedControlPoints: TPathControlPointList;
    function GetHasSelection: boolean;
//    procedure SetSelectedControlPoints(const Value: TPathControlPointList);
    function GetOrthographicDistancePercentage(ACoord, AMinCoord, AMaxCoord: Double; AInverse: Boolean): Single; // chrisu: Function for OrthographicDistanceShading.
  public
    // constructors & destructors
    constructor Create;
    destructor Destroy; override;

    // event handlers
    procedure OnHasChangedHandler(Sender: TObject);     // if a primitive has changed, the scene is notified
    procedure OnCollectionChangeHandler(Sender: TObject);
    procedure OnSelectionChangedHandler(Sender: TObject);
    procedure UpdateHandlers;
    procedure RebuildSelections;

    // various methods
    procedure UpdateGUI(AUpdateGUIOptions: TUpdateGUIOptions = ugoAllElements);
    procedure Clear;
    procedure SetupGizmo(AOGLContext: TfrmOGLContext);
    procedure ClearCollectionHandlers;
    procedure SetCollectionHandlers;
    procedure DeHighlightItems;

    // prefab export method (TODO, JC: be able to load various types of prefabs)
    procedure ExportPrefab(const aFileName: string);

    // addition methods
    procedure AddAnimation(ASkeleton: TSkeleton; ANumberOfFrames,AFramesPerSecond: integer);

    // chrisu: Procedure/Functions for OrthographicDistanceShading.
    procedure UpdateMinMaxCoords; // chrisu: Basics for OrthographicDistanceShading.
    function GetOrthographicDistancePercentageX(AVertice: TVertice): Single;
    function GetOrthographicDistancePercentageY(AVertice: TVertice): Single;
    function GetOrthographicDistancePercentageZ(AVertice: TVertice): Single;
    // chrisu: ----

    // properties
    property Name: string read fName write fName;
    property Author: string read fAuthor write fAuthor;
    property Comments: string read fComments write fComments;
    property ShadowOpacity: integer read fShadowOpacity write fShadowOpacity;
    property AmbientColor: TRGBColor read fAmbientColor write fAmbientColor;
    property FileName: string read fFileName write fFileName;
    property Gizmo: TGizmo read fGizmo write fGizmo;
    property FileSaved: boolean read fFileSaved write fFileSaved;
    property HasChanged: boolean read fHasChanged write fHasChanged;
    property HasSelection: boolean read GetHasSelection;

    // collections
    property Primitives: TPrimitiveList read fPrimitives;
    property Skeletons: TSkeletonList read fSkeletons;

    // selections
    property SelectedPrimitives: TPrimitiveList read fSelectedPrimitives;
    property SelectedSkeletons: TSkeletonList read fSelectedSkeletons;

    // NOTE: these properties should probably be part of their respective selectors...
    property SelectedJoints: TJointList read fSelectedJoints; // GetSelectedJoints;
    property SelectedAnimations: TAnimationList read fSelectedAnimations; // GetSelectedAnimations;
    property SelectedPolygons: TPolygonList read GetSelectedPolygons;
    property SelectedEdges: TEdgeList read GetSelectedEdges;
    property SelectedVertices: TVerticeList read GetSelectedVertices;
    property SelectedUVCoordinates: TUVList read GetSelectedUVCoordinates;
    property SelectedControlPoints: TPathControlPointList read GetSelectedControlPoints;

    // selectors
    property PrimitiveSelector: TPrimitiveSelector read fPrimitiveSelector write fPrimitiveSelector;
    property JointSelector: TJointSelector read fJointSelector write fJointSelector;
    property JointGeometrySelector: TJointGeometrySelectingTool read fJointGeometrySelector write fJointGeometrySelector;
    property PolygonSelector: TPolygonSelector read fPolygonSelector write fPolygonSelector;
    property EdgeSelector: TEdgeSelector read fEdgeSelector write fEdgeSelector;
    property VertexSelector: TVertexSelector read fVertexSelector write fVertexSelector;
    property UVSelector: TUVSelector read fUVSelector write fUVSelector;
    property PathSelector: TPathSelector read fPathSelector;
    property ControlPointSelector: TPathControlPointSelector read fControlPointSelector write fControlPointSelector;

    // tools
    property MaterialManager: TMaterialManager read fMaterialManager write fMaterialManager;
    property ToolFactory: TToolFactory read fToolFactory write fToolFactory;
    property ToolExecuter: TToolExecuter read fToolExecuter write fToolExecuter;
    property UndoRedoManagerEx: TUndoRedoManagerEx read fUndoRedoManagerEx write fUndoRedoManagerEx;
    property JointGeometryUpdater: TJointGeometryUpdatingTool read fJointGeometryUpdater write fJointGeometryUpdater;

    // events
    property OnCollectionChange: TNotifyEvent read fOnCollectionChange write fOnCollectionChange;
    property OnSelectionChange: TNotifyEvent read fOnSelectionChange write fOnSelectionChange;
    property OnLoadPrefab: TNotifyEvent read fOnLoadPrefab write fOnLoadPrefab;
    property OnAddJoint: TNotifyEvent read fOnAddJoint write fOnAddJoint;
  end;

implementation

uses
    Math,
    frmCreatePropertiesText,
    unit_Variables,
    unit_TextureTypes,
    unit_TextureWriter,
    unit_TextureLoader,
    unit_DMFSceneWriter,
    unit_SkeletonBuilder,
    unit_JointLocation,
    unit_JointBuilder,
    unit_Bone,
    unit_BoneBuilder,
    unit_PointLightBuilder,
    unit_DirectionalLightBuilder,
    unit_SpotLightBuilder,
    unit_AnimationBuilder,
    unit_Constants,
    unit_PrimitiveSnapperTool;

// ### constructors & destructors ##############################################

constructor TScene.Create;
begin
    inherited Create;

    // collections
    fPrimitives := TPrimitiveList.Create(true,true);
    fSkeletons := TSkeletonList.Create(true);

    // selections
    fSelectedPrimitives := TPrimitiveList.Create(false,true);
    fSelectedSkeletons := TSkeletonList.Create(false, true);

    fSelectedJoints := TJointList.Create(false,true);
    fSelectedAnimations := TAnimationList.Create(false,true);
    fSelectedPolygons := TPolygonList.Create(false,true);
    fSelectedEdges := TEdgeList.Create(false,true);
    fSelectedVertices := TVerticeList.Create(false,true);
    fSelectedUVCoordinates := TUVList.Create(false,true);
    fSelectedControlPoints := TPathControlPointList.Create(false, true);

    // hook eventhandlers;
    SetCollectionHandlers;
    //fPrimitives.OnChange := OnCollectionChangeHandler;
    //fSkeletons.OnChange := OnCollectionChangeHandler;
    //fSelectedPrimitives.OnChange := OnSelectionChangedHandler;
    //fSelectedSkeletons.OnChange := OnSelectionChangedHandler;

    // selectors
    fPrimitiveSelector := TPrimitiveSelector.Create(Primitives, SelectedPrimitives);
    fJointSelector := TJointSelector.Create(Skeletons,SelectedSkeletons);
    fJointGeometrySelector := TJointGeometrySelectingTool.Create(taNotApplicable);
    fJointGeometrySelector.Skeletons := Skeletons;
    fJointGeometrySelector.SelectedSkeletons := SelectedSkeletons;
    fPolygonSelector := TPolygonSelector.Create(SelectedPrimitives);
    fEdgeSelector := TEdgeSelector.Create(SelectedPrimitives);
    fVertexSelector := TVertexSelector.Create(SelectedPrimitives,SelectedVertices);
    fUVSelector := TUVSelector.Create(SelectedPrimitives);

    // tools (NOTE: we are slightly misusing TScene as a placeholder for these tools...)
    fMaterialManager := TMaterialManager.Create(true,Primitives);
    fToolFactory := TToolFactory.Create(Primitives, SelectedPrimitives, Skeletons, SelectedSkeletons);
    fUndoRedoManagerEx := TUndoRedoManagerEx.Create;
    fToolExecuter := TToolExecuter.Create(fUndoRedoManagerEx, fToolFactory);
    fJointGeometryUpdater := TJointGeometryUpdatingTool.Create(taNotApplicable);
    fJointGeometryUpdater.Skeletons := Skeletons;
    fJointGeometryUpdater.SelectedSkeletons := SelectedSkeletons;

    // we store the gizmo in the scene for easy reference in 3d views later
    // (instead of storing it locally in each 3d view, for example)
    fGizmo := TGizmo.Create;

    // chrisu: Constructing fields for OrthographicDistanceShading.
    fMaxVertexCoords := TVertice.Create();
    fMinVertexCoords := TVertice.Create();
    // chrisu: ----

    Initialize;
end;

destructor TScene.Destroy;
begin
    // unhook eventhandlers
    fPrimitives.OnChange := nil;
    fSelectedPrimitives.OnChange := nil;
    fSkeletons.OnChange := nil;
    fSelectedSkeletons.OnChange := nil;

    // free objects
    FreeAndNil(fMaterialManager);
    FreeAndNil(fPrimitiveSelector);
    FreeAndNil(fJointSelector);
    FreeAndNil(fJointGeometrySelector);
    FreeAndNil(fPolygonSelector);
    FreeAndNil(fEdgeSelector);
    FreeAndNil(fVertexSelector);
    FreeAndNil(fUVSelector);
    FreeAndNil(fSelectedPolygons);
    FreeAndNil(fSelectedEdges);
    FreeAndNil(fSelectedVertices);
    FreeAndNil(fSelectedUVCoordinates);
    FreeAndNil(fSelectedSkeletons);
    FreeAndNil(fSelectedJoints);
    FreeAndNil(fSelectedAnimations);
    FreeAndNil(fSelectedPrimitives);
    FreeAndNil(fSelectedControlPoints);
    FreeAndNil(fSkeletons);
    FreeAndNil(fPrimitives);
    FreeAndNil(fToolFactory);
    FreeAndNil(fToolExecuter);
    FreeAndNil(fJointGeometryUpdater);
    FreeAndNil(fUndoRedoManagerEx);
    FreeAndNil(fGizmo);
    // chrisu: Constructing fields for OrthographicDistanceShading.
    FreeAndNil(fMaxVertexCoords);
    FreeAndNil(fMinVertexCoords);
    // chrisu: ----

    inherited;
end;

procedure TScene.Initialize;
begin
    Name := NEW_SCENE_NAME;
    Author := '';
    Comments := '';
    AmbientColor := MakeRGB(1,1,1);
    ShadowOpacity := 75;
    FileName := NEW_SCENE_FILENAME;
    FileSaved := false;
    HasChanged := false;
end;

function TScene.GetSelectedPolygons: TPolygonList;
var i: integer;
begin
    fSelectedPolygons.Clear;
    for i := 0 to fSelectedPrimitives.Count-1 do
        fSelectedPolygons.AddList(fSelectedPrimitives[i].SelectedPolygons);
    Result := fSelectedPolygons;
end;

function TScene.GetSelectedControlPoints: TPathControlPointList;
var
  i: integer;
  j: Integer;
begin
    FSelectedControlPoints.Clear;
    for i := 0 to FSelectedPrimitives.Count-1 do
      if FSelectedPrimitives[i] is TPath then
        for j := 0 to TPath(FSelectedPrimitives[i]).SelectedControlPoints.Count - 1 do
          FSelectedControlPoints.Add( TObject(TPath(FSelectedPrimitives[i]).SelectedControlPoints[j]) );
    Result := FSelectedControlPoints;
end;

function TScene.GetSelectedEdges: TEdgeList;
var i: integer;
begin
    fSelectedEdges.Clear;
    for i := 0 to fSelectedPrimitives.Count-1 do
        fSelectedEdges.AddList(fSelectedPrimitives[i].SelectedEdges);
    Result := fSelectedEdges;
end;

function TScene.GetSelectedVertices: TVerticeList;
var i: integer;
begin
    fSelectedVertices.Clear;
    for i := 0 to fSelectedPrimitives.Count-1 do
        fSelectedVertices.AddList(fSelectedPrimitives[i].SelectedVertices);
    Result := fSelectedVertices;
end;

function TScene.GetSelectedUVCoordinates: TUVList;
var i, j, k: integer;
    primitive: TBasePrimitive;
    polygon: TPolygon;
begin
    fSelectedUVCoordinates.Clear;
    for i := 0 to fSelectedPrimitives.Count - 1 do begin
        primitive := fSelectedPrimitives[i];
        for j := 0 to primitive.NumPolygons - 1 do begin
            polygon := primitive.Polygons[j];
            for k := 0 to polygon.NumVertices - 1 do begin
                if polygon.UVCoordinates[0][k].Selected then
                    fSelectedUVCoordinates.Add(polygon.UVCoordinates[0][k]);
            end;
        end;
    end;
    Result := fSelectedUVCoordinates;
end;

function TScene.GetHasSelection: boolean;
begin
  Result := (SelectedPrimitives.Count > 0) or
            (SelectedSkeletons.Count > 0) or
            (SelectedJoints.Count > 0) or
            (SelectedAnimations.Count > 0) or
            (SelectedPolygons.Count > 0) or
            (SelectedEdges.Count > 0) or
            (SelectedVertices.Count > 0) or
            (SelectedUVCoordinates.Count > 0) or
            (SelectedControlPoints.Count > 0);
end;

procedure TScene.UpdateMinMaxCoords; // chrisu: Basics for OrthographicDistanceShading.
var i,j: Integer;
    obj: TBasePrimitive;
    init: Boolean;
    vert: TVertice;
begin
    // chrisu: Update of the fields for OrthographicDistanceShading.
    if (toggleOrthographicDistanceShading) then begin
      init := true;
      for i := 0 to Primitives.Count-1 do begin
        obj := Primitives[i];
        if (obj.visible) and not (obj is TLight) and not (obj is TPath) then begin
          for j := 0 to obj.NumVertices-1 do begin
            vert := obj.Vertices[j];
            if (init) then begin
              fMinVertexCoords.Assign(vert);
              fMaxVertexCoords.Assign(vert);
              init := false;
            end else begin
              if (vert.x < fMinVertexCoords.x) then fMinVertexCoords.x := vert.x;
              if (vert.y < fMinVertexCoords.y) then fMinVertexCoords.y := vert.y;
              if (vert.z < fMinVertexCoords.z) then fMinVertexCoords.z := vert.z;
              if (vert.x > fMaxVertexCoords.x) then fMaxVertexCoords.x := vert.x;
              if (vert.y > fMaxVertexCoords.y) then fMaxVertexCoords.y := vert.y;
              if (vert.z > fMaxVertexCoords.z) then fMaxVertexCoords.z := vert.z;
            end;
          end;
        end;
      end;
    end;
    // chrisu: ----
end;

procedure TScene.UpdateGUI(AUpdateGUIOptions: TUpdateGUIOptions = ugoAllElements);
begin
    // rebuild selections
    RebuildSelections;

    // update scene inspector
    if AUpdateGUIOptions in [ugoSceneInspector,ugoAllElements] then
        SendMessage(handle_main,WM_USER+103,0,0);
    // update object inspector
    if AUpdateGUIOptions in [ugoObjectInspector,ugoAllElements] then
        SendMessage(handle_main,WM_USER+106,0,0);

    SendMessage(handle_main,WM_PAINT,0,0);
end;

procedure TScene.Clear;
begin
    Initialize;

    // first clear selections
    SelectedPrimitives.Clear;
    SelectedSkeletons.Clear;
    SelectedVertices.Clear;

    // then clear primary collections
    Primitives.Clear;
    Skeletons.Clear;
    
    // clear other stuff
    MaterialManager.Clear;
    UndoRedoManagerEx.Clear;
    JointGeometrySelector.Clear;
end;

procedure TScene.SetupGizmo(AOGLContext: TfrmOGLContext);
var dpfLoader: TDPFLoader;
    compensateVector: TVertice;
    i: integer;
begin
    Gizmo.Materials.Textures.OGLContext := AOGLContext;
    Gizmo.Materials.Lightmaps.OGLContext := AOGLContext;
    dpfLoader := TDPFLoader.Create;
    dpfLoader.LoadDPF('gizmo',true,'',nil,Gizmo.Materials,Gizmo);
    FreeAndNil(dpfLoader);
    // the gizmo prefab is centered around 0,0,0 but we want
    // the yellow box to be the actual center; compensate here
    // by adding a vector to all vertices of the gizmo
    compensateVector := TVertice.Create(69.8,69.8,69.8);
    for i := 0 to Gizmo.Vertices.Count-1 do
        Gizmo.Vertices[i].Add(compensateVector);
    FreeAndNil(compensateVector);
    Gizmo.Materials.Textures.UploadTextures;
end;

procedure TScene.ClearCollectionHandlers;
begin
    fPrimitives.OnChange := nil;
    fSkeletons.OnChange := nil;
    fSelectedPrimitives.OnChange := nil;
    fSelectedSkeletons.OnChange := nil;
end;

procedure TScene.SetCollectionHandlers;
begin
    fPrimitives.OnChange := OnCollectionChangeHandler;
    fSkeletons.OnChange := OnCollectionChangeHandler;
    fSelectedPrimitives.OnChange := OnSelectionChangedHandler;
    fSelectedSkeletons.OnChange := OnSelectionChangedHandler;
end;

//procedure TScene.SetSelectedControlPoints(const Value: TPathControlPointList);
//begin
//  FSelectedControlPoints := Value;
//end;

procedure TScene.DeHighlightItems;
var i: integer;
    primitive: TBasePrimitive;
begin
    for i := 0 to Primitives.Count - 1 do begin
        primitive := Primitives[i];
        primitive.HighlightedVertices.Clear;
        primitive.HighlightedEdges.Clear;
        primitive.HighlightedPolygons.Clear;
    end;
end;

procedure TScene.ExportPrefab(const aFileName: string);
var i,j: integer;
    strings: TStringList;
    myStr: string;
    obj: TBasePrimitive;
    poly: TPolygon;
    vertex,c: TVertice;
    uvcoord: TUVCoord;
    usedMaterials: TMaterialList;
    material: TMaterial;
    dmfSceneWriter: TDMFSceneWriter;
    paramList: TParamList;
begin
    // to do this here is not a clean solution; but it does solve some problems anyway
    // NOTE: before sending data to plugins (!), the geometry should be checked too
    paramList := TParamList.Create(true);
    paramList.Add(integer(prmInvalid),'polygonRemoveMethod',true);
    ToolExecuter.Execute(ptPolygonRemover,taNotApplicable,paramList);
    FreeAndNil(paramList);

    strings := TStringList.Create;
    try
        obj := SelectedPrimitives[0];
        if not Assigned(obj) then exit;

        usedMaterials := obj.Face_GetUsedMaterials([gmoOriginalsOnly]);
        try
            strings.Add('DeleD Prefab File;');
            //strings.Add('Version 0.90;');
            strings.Add('Version 1.0;');
            strings.Add(IntToStr(usedMaterials.nummaterials)+';');
            dmfSceneWriter := TDMFSceneWriter.Create;
            try
                dmfSceneWriter.GetMaterials(usedMaterials,'090',false,strings,[gmoOriginalsOnly]);
            finally
                FreeAndNil(dmfSceneWriter);
            end;

            if Trim(obj.Name)='' then obj.Name := 'UserDefined';
            strings.Add(obj.Name+';');
            strings.Add(IntToStr(integer(obj.objecttype))+';');
            strings.Add(IntToStr(obj.NumVertices)+';');
            c := obj.Center;
            for i := 0 to obj.NumVertices-1 do begin
                vertex := obj.Vertices[i].Copy;
                vertex.Subtract(c);
                strings.Add(FloatToStr(vertex.x)+';'+
                            FloatToStr(vertex.y)+';'+
                            FloatToStr(vertex.z)+';');
                vertex.Free;
            end;

            strings.Add(IntToStr(obj.numpolygons)+';');
            for i := 0 to obj.NumPolygons-1 do begin
                poly := obj.Polygons[i];
                // only original materials!
                material := poly.Material;
                if not material.IsOriginal then
                    material := material.OriginalMaterial;
                myStr := IntToStr(poly.numvertices)+';'+IntToStr(usedMaterials.IndexOfMaterial(material))+';';

                for j := 0 to poly.numvertices-1 do
                    myStr := myStr + IntToStr(obj.Vertices.IndexOf(poly.VerticeList[j]))+';';

                for j := 0 to poly.numvertices-1 do begin
                    uvcoord := poly.UVCoordinates[0][j];
                    myStr := myStr + FloatToStr(uvcoord.u)+';'+FloatToStr(uvcoord.v)+';';
                end;
                strings.Add(myStr);
            end;
            strings.SaveToFile(aFileName);

        finally
            FreeAndNil(usedMaterials);
        end;

    finally
        FreeAndNil(strings);
    end;
end;

// ### Addition methods ########################################################

// creates and adds a new animation to a given skeleton
procedure TScene.AddAnimation(ASkeleton: TSkeleton; ANumberOfFrames,AFramesPerSecond: integer);
var animationBuilder: TAnimationBuilder;
begin
    animationBuilder := TAnimationBuilder.Create;
    animationBuilder.Build(ASkeleton,ANumberOfFrames,AFramesPerSecond);
    FreeAndNil(animationBuilder);
    if Assigned(fOnCollectionChange) then
        fOnCollectionChange(nil);
end;

// ### event handlers ##########################################################

// called whenever a change in a primitive occurs
procedure TScene.OnHasChangedHandler(Sender: TObject);
begin
    HasChanged := true;
end;

// the OnCollectionChangeHandler listens to the Primitive,
// Skeleton and Light collections and is called when an
// element is added/removed to/from on of those lists
procedure TScene.OnCollectionChangeHandler(Sender: TObject);
begin
    // notify world of change
    if Assigned(fOnCollectionChange) then
        fOnCollectionChange(self);
end;

// the OnSelectionChangedHandler listens to the SelectedPrimitives,
// SelectedSkeletons and SelectedLights collections and is called
// when such elements are added to those lists
procedure TScene.OnSelectionChangedHandler(Sender: TObject);
begin
    RebuildSelections;

    // notify world of change
    if Assigned(fOnSelectionChange) then
        fOnSelectionChange(self);
end;

// Call this after reading in your scene to make sure all the handlers are connected.
// PJ: I think we should connect these handlers implicitly upon adding new primitives,
//     but connecting update handling _during a scene update_ quickly leads to a mess.
procedure TScene.UpdateHandlers;
var i: integer;
begin
    for i := 0 to Primitives.Count - 1 do
        Primitives[i].OnHasChanged := OnHasChangedHandler;
end;

procedure TScene.RebuildSelections;
var i: integer;
begin
    // rebuild sub-selection collections
    fSelectedAnimations.Clear;
    for i := 0 to fSelectedSkeletons.Count-1 do begin
        if Assigned(fSelectedSkeletons[i].SelectedAnimation) then
            fSelectedAnimations.Add(fSelectedSkeletons[i].SelectedAnimation);
    end;

    fSelectedJoints.Clear;
    for i := 0 to fSelectedSkeletons.Count-1 do
        fSelectedJoints.AddList(fSelectedSkeletons[i].SelectedJoints);
end;

function TScene.GetOrthographicDistancePercentage(ACoord, AMinCoord, AMaxCoord: Double; AInverse: Boolean): Single; // chrisu: Function for OrthographicDistanceShading.
begin
  if AInverse then
    result := (100.0-((100.0/Abs(AMaxCoord-AMinCoord))*Abs(AMaxCoord-ACoord)))/100
  else
    result := (100.0-((100.0/Abs(AMaxCoord-AMinCoord))*(Abs(AMaxCoord-AMinCoord)-Abs(AMaxCoord-ACoord))))/100;
end;
function TScene.GetOrthographicDistancePercentageX(AVertice: TVertice): Single; // chrisu: Function for OrthographicDistanceShading.
begin
  result := GetOrthographicDistancePercentage(AVertice.x, fMinVertexCoords.x, fMaxVertexCoords.x, true);
end;
function TScene.GetOrthographicDistancePercentageY(AVertice: TVertice): Single; // chrisu: Function for OrthographicDistanceShading.
begin
  result := GetOrthographicDistancePercentage(AVertice.y, fMinVertexCoords.y, fMaxVertexCoords.y, false);
end;
function TScene.GetOrthographicDistancePercentageZ(AVertice: TVertice): Single; // chrisu: Function for OrthographicDistanceShading.
begin
  result := GetOrthographicDistancePercentage(AVertice.z, fMinVertexCoords.z, fMaxVertexCoords.z, false);
end;

end.
