{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Skeleton;

{$MODE Delphi}

interface

uses
  ComCtrls,
  unit_Types,
  unit_Enums,
  unit_Joint,
  unit_JointLocation,
  unit_AnimationFrame,
  unit_Animation,
  unit_Primitives,
  unit_PrimitiveList;

type
  TSkeleton = class(TObject)
  private
    fID: integer;
    fName: string;
    fAuthor: string;
    fVisible: boolean;
    fJoints: TJointList;
    fSelectedJoints: TJointList;
    fBaseFrame: TAnimationFrame;
    fAnimations: TAnimationList;
    fAssignedPrimitives: TPrimitiveList;
    fSelectedAnimation: TAnimation;
    fTreeNode: TTreeNode;
    fJointSelectionIndex: integer;
    fJointSize: integer;
    procedure SetID(AValue: integer);
    function GetNumberOfJoints: integer;
    function GetNumberOfAnimations: integer;
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(ASkeleton: TSkeleton);
    procedure LinkPrimitiveToSkeleton(APrimitive: TBasePrimitive);
    procedure LinkScenePrimitivesToSkeleton(APrimitiveList: TPrimitiveList);
    function GetActiveFrame: TAnimationFrame;
    procedure Invalidate;
    property TreeNode: TTreeNode read fTreeNode write fTreeNode;
    property ID: integer read fID write SetID;
    property Name: string read fName write fName;
    property Author: string read fAuthor write fAuthor;
    property Visible: boolean read fVisible write fVisible;
    property Joints: TJointList read fJoints write fJoints;
    property SelectedJoints: TJointList read fSelectedJoints write fSelectedJoints;
    property BaseFrame: TAnimationFrame read fBaseFrame write fBaseFrame;
    property Animations: TAnimationList read fAnimations write fAnimations;
    property SelectedAnimation: TAnimation read fSelectedAnimation write fSelectedAnimation;
    property AssignedPrimitives: TPrimitiveList read fAssignedPrimitives write fAssignedPrimitives;
    property NumberOfJoints: integer read GetNumberOfJoints;
    property NumberOfAnimations: integer read GetNumberOfAnimations;
    property JointSelectionIndex: integer read fJointSelectionIndex write fJointSelectionIndex;
    property JointSize: integer read fJointSize write fJointSize;
  end;

  TSkeletonList = class(TDeleDObjectList)
  private
    fSkeletonIDValue: integer;
    function GetSkeleton(anIndex: integer): TSkeleton;
  public
    procedure Clear; override;
    procedure AddSkeleton(ASkeleton: TSkeleton);
    function GetSkeletonByID(AID: integer): TSkeleton;
    property Skeletons[anIndex: integer]: TSkeleton read GetSkeleton; default;
  end;

implementation

uses SysUtils, Classes;

// ### TSkeleton ###############################################################

constructor TSkeleton.Create;
begin
    fID := -1;
    fName := 'skeleton';
    fAuthor := 'DeleD';
    fJoints := TJointList.Create(true,true);
    fSelectedJoints := TJointList.Create(false,true);
    fBaseFrame := TAnimationFrame.Create;
    fBaseFrame.IsBaseFrame := true;
    fAnimations := TAnimationList.Create(true,true);
    fAssignedPrimitives := TPrimitiveList.Create(true,true);
    fVisible := true;
    fJointSelectionIndex := 0;
    fJointSize := 64;
end;

destructor TSkeleton.Destroy;
begin
    FreeAndNil(fJoints);
    FreeAndNil(fSelectedJoints);
    FreeAndNil(fBaseFrame);
    FreeAndNil(fAnimations);
    FreeAndNil(fAssignedPrimitives);
    inherited Destroy;
end;

procedure TSkeleton.Assign(ASkeleton: TSkeleton);
var i: integer;
    animation: TAnimation;
    primitive: TBasePrimitive;
begin
    fID := -1;
    fName := ASkeleton.Name;
    fAuthor := ASkeleton.Author;
    // we copy the baseframe (holding jointlocations)
    fBaseFrame.Assign(ASkeleton.BaseFrame);
    // we can copy joints
    fJoints.Assign(ASkeleton.Joints);
    // joints hold a reference to the skeleton that owns them;
    // we need to set that reference after assigning joints
    // because the jointlist itself doesn't know its skeleton
    for i := 0 to fJoints.Count-1 do
        fJoints[i].Skeleton := self;
    // we copy animations too and fill their Skeleton property
    for i := 0 to ASkeleton.Animations.Count-1 do begin
        animation := TAnimation.Create;
        animation.Assign(ASkeleton.Animations[i]);
        animation.Skeleton := self;
        Animations.AddAnimation(animation);
    end;
    // and finally, we copy any assigned geometry and link
    // it to the skeleton
    for i := 0 to ASkeleton.AssignedPrimitives.Count-1 do begin
        primitive := ASkeleton.AssignedPrimitives[i].Copy;
        primitive.ID := ASkeleton.AssignedPrimitives[i].ID;
        primitive.SkeletonID := ID;
        LinkPrimitiveToSkeleton(primitive);
    end;
end;

procedure TSkeleton.SetID(AValue: integer);
var i: integer;
begin
    fID := AValue;
    // we need to make sure any assigned primitive receives
    // the skeleton ID too; this is particularly done when
    // loading skeletons from a file; a skeleton can receive
    // a new ID value when it is added to a scene because
    // another skeleton with the same ID value might already
    // exist; in that case, all assigned primitives need to
    // have their SkeletonID property updated as well
    for i := 0 to AssignedPrimitives.Count-1 do
        AssignedPrimitives[i].SkeletonID := fID;
end;

function TSkeleton.GetNumberOfJoints: integer;
begin
    Result := Joints.Count;
end;

function TSkeleton.GetNumberOfAnimations: integer;
begin
    Result := Animations.Count;
end;

procedure TSkeleton.LinkPrimitiveToSkeleton(APrimitive: TBasePrimitive);
var i: integer;
    vertex: TVertice;
    joint: TJoint;
begin
    // assign the primitive to the skeleton
    AssignedPrimitives.Add(APrimitive);

    // when assigning a primitive to a skeleton, we create
    // copies of the original vertices
    APrimitive.CopyToOriginalVertices;

    // now assign individual vertices of the primitive
    // to joints of the skeleton, as saved in the DXS
    for i := 0 to APrimitive.Vertices.Count-1 do begin
        vertex := APrimitive.Vertices[i];
        if vertex.JointID<>-1 then begin
            joint := Joints.GetJointByID(vertex.JointID);
            if Assigned(joint) then
                joint.AssignedVertices.Add(vertex);
        end;
    end;
end;

procedure TSkeleton.LinkScenePrimitivesToSkeleton(APrimitiveList: TPrimitiveList);
var counter: integer;
    primitive: TBasePrimitive;
begin
    counter := 0;
    while counter<APrimitiveList.Count do begin
        primitive := APrimitiveList[counter];
        if primitive.SkeletonID<>ID then
            inc(counter)
        else begin
            // link primitive to skeleton
            LinkPrimitiveToSkeleton(primitive);
            // and remove it from the scene
            APrimitiveList.Remove(primitive,true);
        end;
    end;
end;

function TSkeleton.GetActiveFrame: TAnimationFrame;
var frame: TAnimationFrame;
begin
    if Animations.Count=0 then
        // if we don't have animations for this skeleton, use the base frame
        frame := BaseFrame
    else begin
        // if we didn't select an animation for this skeleton yet, use the base frame
        if not Assigned(SelectedAnimation) then
            frame := BaseFrame
        else    // we have an animation selected, so pass the selected frame of that animation
            frame := SelectedAnimation.SelectedFrame;
    end;
    Result := frame;
end;

procedure TSkeleton.Invalidate;
var i: integer;
begin
    // Resets rendering state. Remove some day.
    for i := 0 to Joints.Count - 1 do
        Joints[i].Invalidate;
end;

// ### TSkeletonList ###########################################################

procedure TSkeletonList.Clear;
var i: integer;
begin
    for i := 0 to Count-1 do begin
        Skeletons[i].SelectedJoints.Clear;
        Skeletons[i].SelectedAnimation := nil;
    end;
    inherited Clear;
end;

function TSkeletonList.GetSkeleton(anIndex: integer): TSkeleton;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TSkeleton(Items[anIndex])
    else
        Result := nil;
end;

procedure TSkeletonList.AddSkeleton(ASkeleton: TSkeleton);
begin
    Inc(fSkeletonIDValue);
    ASkeleton.ID := fSkeletonIDValue;
    Add(ASkeleton);
end;

function TSkeletonList.GetSkeletonByID(AID: integer): TSkeleton;
var counter: integer;
    isFound: boolean;
    skeleton: TSkeleton;
begin
    counter := 0;
    isFound := false;
    skeleton := nil;
    while (counter<Count) and not (isFound) do begin
        skeleton := GetSkeleton(counter);
        if skeleton.ID = AID then
            isFound := true
        else
            inc(counter);
    end;
    Result := skeleton;
end;

end.
 
