{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_StringDictionary;

{$MODE Delphi}

interface

uses
  Classes;

type
  TStringDictionary = class( TObject )
  private
      fKeys: TStringList;
      fValues: TStringList;

      fCachedKey: string;
      fCachedIndex: integer;

      function GetValue( const aKey: string ): string;
      procedure SetValue( const aKey, aValue: string );

      function GetKeyForIndex( const anIndex: integer ): string;
      function GetValueForIndex( const anIndex: integer ): string;

  public
      constructor Create;
      destructor Destroy; override;

      function Count: integer;
      function ContainsKey( const aKey: string ): boolean;

      property Values[const aKey: string]: string read GetValue write SetValue; default;

      property AllKeys[const anIndex: integer]: string read GetKeyForIndex;
      property AllValues[const anIndex: integer]: string read GetValueForIndex;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  SysUtils, IniFiles;

const
  HASHING_THRESHOLD_COUNT = 12; // The dictionary automatically switches to hashing when this amount is reached.


constructor TStringDictionary.Create;
begin
  inherited Create;

  fKeys := TStringList.create;
  fValues := TStringList.create;
end;

// ----------------------------------------------------------------------------

destructor TStringDictionary.Destroy;
begin
  FreeAndNil( fValues );
  FreeAndNil( fKeys );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

function TStringDictionary.GetValue( const aKey: string ): string;
var
  index: integer;

begin
  if aKey = fCachedKey then
  begin
      index := fCachedIndex;
  end
  else
  begin
      index := fKeys.IndexOf( aKey );

      fCachedKey := aKey;
      fCachedIndex := index;
  end;

  if index <> - 1 then
  begin
      result := fValues[index];
  end
  else
      raise Exception.Create( Format( 'Key "%s" not found.', [aKey] ));
end;

// ----------------------------------------------------------------------------

procedure TStringDictionary.SetValue( const aKey, aValue: string );
var
  index: integer;
  keys: TStringList;

begin
  index := fKeys.IndexOf( aKey );

  if index = -1 then
  begin
      if fKeys.Count >= HASHING_THRESHOLD_COUNT - 1 then
      begin // INFO, PJ: Convert to hashing list
          keys := THashedStringList.Create;
          keys.Assign( fKeys );

          fKeys.Free;
          fKeys := keys;
      end;

      fKeys.Add( aKey );
      fValues.Add( aValue );
  end
  else
  begin
      fCachedKey := '';
      
      fValues[index] := aValue;
  end;
end;

// ----------------------------------------------------------------------------

function TStringDictionary.GetKeyForIndex( const anIndex: integer ): string;
begin
  result := fKeys[anIndex];
end;

// ----------------------------------------------------------------------------

function TStringDictionary.GetValueForIndex( const anIndex: integer ): string;
begin
  result := fValues[anIndex];
end;

// ----------------------------------------------------------------------------

function TStringDictionary.Count: integer;
begin
  result := fKeys.Count;
end;

// ----------------------------------------------------------------------------

function TStringDictionary.ContainsKey( const aKey: string ): boolean;
begin
  fCachedKey := aKey;
  fCachedIndex := fKeys.IndexOf( aKey );

  result := fCachedIndex <> -1;
end;

end.
