{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_TGALoader;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  TGA Loader
  ---------------------------------------------------------------------------- }


interface

uses
    Classes,
    unit_TextureTypes;

type
    TTGAHeader = packed record
        CommentSize: Byte;
        ColorMapType: Byte;
        ImageType: Byte;
        ColorMapSpec: Array[0..4] of Byte;
        OrigX: Word;
        OrigY: Word;
        Width: Word;
        Height: Word;
        BPP: Byte;
        ImageInfo: Byte;
    end;

// ----------------------------------------------------------------------------

    TTGALoader = class(TObject)
    private
        fStream: TStream;

        // cached header
        fHeader: TTGAHeader;

        procedure UnpackRawRGBBuffer( const aSource, aTarget: pointer; const aNrPixels, aNrBytesPerPixel: integer );
        procedure UnpackRawGrayscaleBuffer( const aSource, aTarget: pointer; const aNrPixels: integer );
        procedure UnpackRLERGBBuffer( const aSource, aTarget: pointer; const aNrPixels, aNrBytesPerPixel: integer );
        procedure UnpackRLEGrayscaleBuffer( const aSource, aTarget: pointer; const aNrPixels: integer );

    public
        constructor Create(aStream: TStream);

        function LoadHeader: TTextureData;
        function Load: TTextureData;
    end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
    Windows, SysUtils, unit_Types;

constructor TTGALoader.Create(aStream: TStream);
begin
    inherited Create;

    fStream := aStream;
end;

// ----------------------------------------------------------------------------

function TTGALoader.LoadHeader: TTextureData;
begin
  fStream.Read(fHeader, SizeOf(fHeader));

  if not fHeader.ImageType in [2, 3, 10, 11] then
      raise Exception.Create( 'TGA type not supported.' );

  if fHeader.ColorMapType <> 0 then
      raise Exception.Create( 'Colormapped TGA files not supported.' );

  result.Width := fHeader.Width;
  result.Height := fHeader.Height;

  if not (fHeader.BPP in [8, 24, 32]) then
      raise Exception.Create( 'Only 8, 24 and 32 bit TGA files supported.' );

  if fHeader.BPP = 8 then
      result.BitsPerPixel := 24
  else
      result.BitsPerPixel := fHeader.BPP;

  // Skip comment (convenient when rest of file contents are read in afterwards)
  fStream.Seek( fHeader.CommentSize, soCurrent );
end;

// ----------------------------------------------------------------------------

function TTGALoader.Load: TTextureData;
var buffer, imageData: pointer;
    size: cardinal;
    nrBytesPerPixel, imageSize: integer;
    sourceIndex, targetIndex: integer;
    x, y, i: integer;
    pixel: PTexturePixel;

begin
    result := LoadHeader;

    // Read in rest of file
    size := fStream.Size - fStream.Position;
    GetMem(buffer, size);
    fStream.Read(buffer^, size);

    nrBytesPerPixel := result.BitsPerPixel div 8;
    imageSize := result.Width * result.Height;

    GetMem(imageData, imageSize * nrBytesPerPixel);

    case fHeader.ImageType of
        2: UnpackRawRGBBuffer( buffer, imageData, imageSize, nrBytesPerPixel );
        3: UnpackRawGrayscaleBuffer( buffer, imageData, imageSize );
        10: UnpackRLERGBBuffer( buffer, imageData, imageSize, nrBytesPerPixel);
        11: UnpackRLEGrayscaleBuffer( buffer, imageData, imageSize );
    end;

    FreeMem(buffer);

    SetLength(result.Pixels, imageSize * nrBytesPerPixel);

    if ( fHeader.ImageInfo and 32 ) = 32 then // Do we need to flip?
        sourceIndex := 0
    else
        sourceIndex := ( result.Height - 1 ) * result.Width * nrBytesPerPixel;

    // Swp R,G,B and flip (optional)
    targetIndex := 0;
    for y := 0 to result.Height - 1 do begin
        for x := 0 to result.Width - 1 do begin
            pixel := @result.Pixels[targetIndex];

            // This should really be a 'move' or possible an assembler instruction
            for i := 0 to nrBytesPerPixel - 1 do
              pixel[i] := THugeByteArray(imageData^)[sourceIndex + i];

            sourceIndex := sourceIndex + nrBytesPerPixel;
            targetIndex := targetIndex + nrBytesPerPixel;
        end;
        if ( fHeader.ImageInfo and 32 ) = 0 then
            sourceIndex := sourceIndex - (2 * nrBytesPerPixel * result.Width);
    end;

    FreeMem(imageData);
end;

// ----------------------------------------------------------------------------

procedure TTGALoader.UnpackRawRGBBuffer( const aSource, aTarget: pointer; const aNrPixels, aNrBytesPerPixel: integer );
begin
    Move( aSource^, aTarget^, aNrPixels * aNrBytesPerPixel);
end;

// ----------------------------------------------------------------------------

procedure TTGALoader.UnpackRawGrayscaleBuffer( const aSource, aTarget: pointer; const aNrPixels: integer );
var source, target: PByte;
    index: integer;

begin
    source := aSource;
    target := aTarget;

    for index := 0 to aNrPixels - 1 do
    begin
        target^ := source^;
        inc(target);
        target^ := source^;
        inc(target);
        target^ := source^;
        inc(target);
        inc(source);
    end;
end;

// ----------------------------------------------------------------------------

procedure TTGALoader.UnpackRLERGBBuffer( const aSource, aTarget: pointer; const aNrPixels, aNrBytesPerPixel: integer );
var remainingPixels, packetIndex, index: integer;
    source, target, packetHeader: PByte;
    nrPixels: byte;
begin
    source := aSource;
    target := aTarget;
    remainingPixels := aNrPixels;
    while remainingPixels > 0  do begin
        packetHeader := source;
        inc(source);
        nrPixels := (packetHeader^ and 127) + 1;
        if ( packetHeader^ and 128 ) = 128 then begin
            for packetIndex := 0 to nrPixels - 1 do begin
                for index := 0 to aNrBytesPerPixel - 1 do begin
                    target^ := source^;
                    inc(target);
                    inc(source);
                end;
                dec(source, aNrBytesPerPixel);
            end;
            inc(source, aNrBytesPerPixel);
        end
        else begin
            for packetIndex := 0 to nrPixels - 1 do begin
                for index := 0 to aNrBytesPerPixel - 1 do begin
                    target^ := source^;
                    inc(target);
                    inc(source);
                end;
            end;
        end;
        dec(remainingPixels, nrPixels);
    end;
end;

// ----------------------------------------------------------------------------

procedure TTGALoader.UnpackRLEGrayscaleBuffer( const aSource, aTarget: pointer; const aNrPixels: integer );
var remainingPixels, packetIndex: integer;
    source, target, packetHeader: PByte;
    nrPixels: byte;
begin
    source := aSource;
    target := aTarget;
    remainingPixels := aNrPixels;
    while remainingPixels > 0  do begin
        packetHeader := source;
        inc(source);
        nrPixels := (packetHeader^ and 127) + 1;
        if ( packetHeader^ and 128 ) = 128 then begin
            for packetIndex := 0 to nrPixels * 3 - 1 do begin
                target^ := source^;
                inc(target);
            end;
            inc(source);
        end
        else begin
            for packetIndex := 0 to nrPixels - 1 do begin
                target^ := source^;
                inc(target);
                target^ := source^;
                inc(target);
                target^ := source^;
                inc(target);
                inc(source);
            end;
        end;
        dec(remainingPixels, nrPixels);
    end;
end;

// ----------------------------------------------------------------------------

end.


