{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_TextureLoader;

{$MODE Delphi}

// TTextureLoader returns either 24 bits or 32 bits (including alpha) images
// right-side up (data starts in upper left corner),
// continous data (no scanline alignment)

interface

uses
    GR32,
    Windows, Graphics,IntfGraphics, Classes, ExtCtrls, dglOpenGL,
    unit_Types,
    unit_TextureTypes;

type
    TTextureLoader = class(TObject)
    private
        function LoadBMPTextureData(const aStream: TStream; const aLoadInfoOnly: Boolean): TTextureData;
        function LoadJPGTextureData(const aStream: TStream; const aLoadInfoOnly: Boolean): TTextureData;
        function LoadTGATextureData(const aStream: TStream; const aLoadInfoOnly: Boolean): TTextureData;
        function LoadDDSTextureData(const aStream: TStream; const aLoadInfoOnly: Boolean): TTextureData;
        function LoadPNGTextureData(const aStream: TStream; const aLoadInfoOnly: Boolean): TTextureData;

        function LoadTextureDataEx(const aFileName: String; const aLoadInfoOnly: Boolean): TTextureData;
    public
        function LoadTextureData(const aFileName: String): TTextureData;
        function LoadTextureInfo(const aFileName: String): TTextureData;

        class function SupportsFileFormat(const anExtension: String): boolean;
        class function SupportedFilesMask: String;                              
    end;

// #############################################################################

    TImageTextureLoader = class(TTextureLoader)
    private
        fImage: TImage;


    protected
        property Image: TImage read fImage;

    public
        procedure LoadTextureIntoImage(const aTextureData: TTextureData);
        procedure LoadTexture(const aFileName: String);
        procedure LoadColor(const aColor: TColor);

        constructor Create(const anImage: TImage);
    end;

// #############################################################################

    TOpenGLTextureLoader = class(TTextureLoader)
    private
        fDeviceContext: HDC;
        fRenderingContext: HGLRC;

        function LoadTextureIntoOpenGL(const aTextureData: TTextureData; const aShouldFilter, aShouldMipMap: Boolean): GLuint;

    protected
        property DeviceContext: HDC read fDeviceContext;
        property RenderingContext: HGLRC read fRenderingContext;

    public
        // aShouldFilter and aShouldMipmap should probably be class properties. Change some day.
        function LoadTexture(const aFileName: String; const aShouldFilter, aShouldMipMap: Boolean): GLuint;
        function LoadColor(const aColor: TColor): GLuint;

        constructor Create(const aDeviceContext: HDC; const aRenderingContext: HGLRC);
    end;

// #############################################################################

    TBitmap32TextureLoader = class( TTextureLoader )
    private
        fBitmap: TBitmap32;

    public
        constructor Create( const aBitmap: TBitmap32 );
        procedure LoadPNGFromResource( const aResourceName: String );
    end;

implementation

uses
    SysUtils,
    unit_Variables,
    unit_TGALoader,
    unit_DDSSupport,
    unit_PNGSupport;

type
  TRGBTriplet = packed record R, G, B: Byte end;

function TTextureLoader.LoadBMPTextureData(const aStream: TStream; const aLoadInfoOnly: Boolean): TTextureData;
var fileHeader: TBitmapFileHeader;
    infoHeader: TBitmapInfoHeader;
    bitmapSize: LongWord;
    i, sourceLineSize, targetLineSize, sourceYIndex, targetYIndex: integer;
    image: Pointer;
begin
  aStream.ReadBuffer(FileHeader, SizeOf(FileHeader));  // FileHeader
  aStream.ReadBuffer(InfoHeader, SizeOf(InfoHeader));  // InfoHeader

  if InfoHeader.biBitCount <> 24 then
      Raise Exception.Create(Format('%d bit BMP not supported',[infoHeader.biBitCount]));

  result.Width := InfoHeader.biWidth;
  result.Height := InfoHeader.biHeight;
  result.BitsPerPixel := 24;

  if not aLoadInfoOnly then begin
      sourceLineSize := (result.Width * InfoHeader.biBitCount div 8 + 3) and $FFFFFFFC;
      bitmapSize := result.Height * sourceLineSize;

      GetMem(image, bitmapSize);
      try
          aStream.ReadBuffer(image^, bitmapSize);

          // Flip the data (bmp holds data upside-down, 4 byte aligned per line)
          targetLineSize := result.Width * result.BitsPerPixel div 8;
          SetLength(result.Pixels, result.Height * targetLineSize);
          sourceYIndex := (result.Height - 1 ) * sourceLineSize;
          targetYIndex := 0;
          for i := 0 to result.Height - 1 do begin
              Move(THugeByteArray(image^)[sourceYIndex], result.Pixels[targetYIndex], targetLineSize);
              sourceYIndex := sourceYIndex - sourceLineSize;
              targetYIndex := targetYIndex + targetLineSize;
          end;
      finally
          FreeMem(image);
      end;
  end;
end;

function TTextureLoader.LoadJPGTextureData(const aStream: TStream; const aLoadInfoOnly: Boolean): TTextureData;
var bmp: TBitmap;
    jpg: TJPEGImage;
    yIndex: Integer;
    lineData: Pointer;
    lineWidth: Integer;
    xbmp:TLazIntfImage;
begin
    jpg := TJPEGImage.Create;
    try
        jpg.LoadFromStream(aStream);

        result.Width := jpg.Width;
        result.Height := jpg.Height;
        result.BitsPerPixel := 24;

        if not aLoadInfoOnly then
        begin
            bmp := TBitmap.Create;
            try
                bmp.pixelformat := pf24bit;
                bmp.width := jpg.width;
                bmp.height := jpg.height;
                bmp.canvas.draw(0, 0, jpg);

                lineWidth := result.Width * (result.BitsPerPixel div 8);
                SetLength(result.Pixels, result.Height * lineWidth);

                // copy the contents of the bitmap to the result set (aligning per 4)
                // 9999 CodeTyphon Studio
                xbmp:= bmp.CreateIntfImage;
                for yIndex := 0 to result.Height - 1 do
                begin
                    lineData := xbmp.GetDataLineStart(yIndex);
                    Move(lineData^, result.Pixels[yIndex * lineWidth], lineWidth);
                end;
                bmp.LoadFromIntfImage(xbmp);
                xbmp.free;
            finally

                bmp.free;
            end;
        end;
    finally
        FreeAndNil(jpg);
    end;
end;

function TTextureLoader.LoadTGATextureData(const aStream: TStream; const aLoadInfoOnly: Boolean): TTextureData;
var tgaLoader: TTGALoader;
begin
    tgaLoader := TTGALoader.Create(aStream);
    try
        if aLoadInfoOnly then
            result := tgaLoader.LoadHeader
        else
            result := tgaLoader.Load;
    finally
        FreeAndNil(tgaLoader);
    end;
end;

function TTextureLoader.LoadDDSTextureData(const aStream: TStream; const aLoadInfoOnly: Boolean): TTextureData;
var ddsLoader: TDDSLoader;
begin
    ddsLoader := TDDSLoader.create(aStream);
    try
        if aLoadInfoOnly then
            result := ddsLoader.LoadHeader
        else
            result := ddsLoader.Load;
    finally
        FreeAndNil(ddsLoader);
    end;
end;

function TTextureLoader.LoadPNGTextureData(const aStream: TStream; const aLoadInfoOnly: Boolean): TTextureData;
var pngLoader: TPNGLoader;
begin
    pngLoader := TPNGLoader.Create(aStream);
    try
        if aLoadInfoOnly then
            result := pngLoader.LoadHeader
        else
            result := pngLoader.Load;
    finally
        FreeAndNil(pngLoader);
    end;
end;

function TTextureLoader.LoadTextureDataEx(const aFileName: String; const aLoadInfoOnly: Boolean): TTextureData;
var stream: TFileStream;
    extension: string;
begin
    stream := TFileStream.Create(aFileName, fmOpenRead or fmShareDenyNone);
    try
        extension := LowerCase(ExtractFileExt(aFileName));

        try
            if extension = '.bmp' then
                result := LoadBMPTextureData(stream, aLoadInfoOnly)
            else if (extension = '.jpg') or (extension = '.jpeg') then
                result := LoadJPGTextureData(stream, aLoadInfoOnly)
            else if extension = '.tga' then
                result := LoadTGATextureData(stream, aLoadInfoOnly)
            else if extension = '.png' then
                result := LoadPNGTextureData(stream, aLoadInfoOnly)
            else if extension = '.dds' then
                result := LoadDDSTextureData(stream, aLoadInfoOnly)
            else
                raise Exception.Create(Format('Unsupported texture format (%s)', [extension]));
        except
            on e: Exception do begin
                raise Exception.Create(Format('Error loading "%s": %s', [aFileName, e.Message]));
            end;
        end;
    finally
        FreeAndNil(stream);
    end;
end;

function TTextureLoader.LoadTextureData(const aFileName: String): TTextureData;
begin
    result := LoadTextureDataEx(aFileName, false);
end;

function TTextureLoader.LoadTextureInfo(const aFileName: String): TTextureData;
begin
    result := LoadTextureDataEx(aFileName, true);
end;

class function TTextureLoader.SupportsFileFormat(const anExtension: String): boolean;
var extension: string;
begin
    extension := LowerCase(anExtension);
    result := ( extension = '.bmp' ) or
              ( extension = '.tga' ) or
              ( extension = '.png' ) or
              ( extension = '.dds' ) or
              ( extension = '.jpg' ) or
              ( extension = '.jpeg');
end;

class function TTextureLoader.SupportedFilesMask: String;
begin
    result := 'All supported formats|*.bmp;*.dds;*.jpg;*.tga;*.png|Bitmap (*.bmp)|*.bmp|DDS (*.dds)|*.dds|JPEG (*.jpg)|*.jpg|(TARGA (*.tga)|*.tga|PNG (*.png)|*.png';
end;

// #############################################################################

procedure TImageTextureLoader.LoadTextureIntoImage(const aTextureData: TTextureData);
var bitmapInfo: TBitmapInfo;
    bytesPerPixel: Integer; x, y, i, sourceLineSize, targetLineSize, sourceOffset, targetOffset: integer;
    alignedData: Pointer;
    xOffset, yOffset, xFactor, yFactor: integer;
    backgroundPixel: PTexturePixel;
    alpha: integer;
begin
    Image.Picture.Bitmap.Width := Image.Width;
    Image.Picture.Bitmap.Height := Image.Height;
    Image.Picture.Bitmap.PixelFormat := pf32bit;

    bytesPerPixel := aTextureData.BitsPerPixel div 8;


    sourceLineSize := aTextureData.Width * bytesPerPixel;
    targetLineSize := (Image.Width  * bytesPerPixel + 3) and $FFFFFFFC;
    GetMem(alignedData, Image.Height * targetLineSize);

    try
        targetOffset := 0;
        xFactor := aTextureData.Width * 256 div Image.Width;
        yFactor := aTextureData.Height * 256 div Image.Height;

        yOffset := yFactor div 2;
        for y := 0 to Image.Height - 1 do begin
            xOffset := xFactor div 2;

            for x := 0 to Image.Width - 1 do begin
                sourceOffset := (yOffset div 256) * sourceLineSize + (xOffset div 256) * bytesPerPixel;

                if bytesPerPixel = 4 then begin
                    if (x div 8 + y div 8 ) mod 2 = 0 then
                        backgroundPixel := @colorTransparencyGrid1
                    else
                        backgroundPixel := @colorTransparencyGrid2;

                    // Alpha blending
                    alpha := aTextureData.Pixels[sourceOffset + 3];
                    for i := 0 to bytesPerPixel - 1 do begin
                        THugeByteArray(alignedData^)[targetOffset] :=
                            ( alpha * aTextureData.Pixels[sourceOffset + i] + ( 255 - alpha ) * backgroundPixel[i] ) div 256;
                        targetOffset := targetOffset + 1;
                    end;
                end
                else begin
                    for i := 0 to bytesPerPixel - 1 do begin
                        THugeByteArray(alignedData^)[targetOffset] := aTextureData.Pixels[sourceOffset + i];
                        targetOffset := targetOffset + 1;
                    end;
                end;

                xOffset := xOffset + xFactor;
            end;

            yOffset := yOffset + yFactor;
        end;

        FillChar(bitmapInfo, SizeOf(bitmapInfo), 0);
        bitmapInfo.bmiHeader.biSize := SizeOf(TBitmapInfoHeader);
        bitmapInfo.bmiHeader.biWidth := Image.Width;
        bitmapInfo.bmiHeader.biHeight := -Image.Height;
        bitmapInfo.bmiHeader.biPlanes := 1;
        bitmapInfo.bmiHeader.biBitCount := aTextureData.BitsPerPixel;
        bitmapInfo.bmiHeader.biCompression := BI_RGB;
        bitmapInfo.bmiHeader.biSizeImage := Image.Height * targetLineSize;
        SetDIBits(Image.Canvas.Handle, Image.Picture.Bitmap.Handle, 0, Image.Height, alignedData, bitmapInfo, DIB_RGB_COLORS);
        Image.Invalidate;

    finally
        FreeMem(alignedData);
    end;
end;

procedure TImageTextureLoader.LoadTexture(const aFileName: String);
var extension: string;
    textureData: TTextureData;
begin
    extension := LowerCase(ExtractFileExt(aFilename));
    if (extension = '.jpg' ) then begin // Don't let Borland routines handle .bmp (Photoshop incompatability)
        Image.Picture.LoadFromFile(aFilename); // Native borland format
    end
    else begin
        textureData := LoadTextureData(aFilename);

        LoadTextureIntoImage(textureData);
    end;
end;

procedure TImageTextureLoader.LoadColor(const aColor: TColor);
var bitmapInfo: TBitmapInfo;
    x, y, i, offset: integer;
    alignedData: Pointer;
    backgroundPixel: PTexturePixel;
    alpha: integer;
begin
    Image.Picture.Bitmap.Width := Image.Width;
    Image.Picture.Bitmap.Height := Image.Height;
    Image.Picture.Bitmap.PixelFormat := pf32bit;

    alpha := aColor shr 24;

    GetMem(alignedData, Image.Height * Image.Height * 4);
    try
        offset := 0;

        for y := 0 to Image.Height - 1 do begin
            for x := 0 to Image.Width - 1 do begin
                if alpha < 255 then begin
                    if (x div 8 + y div 8 ) mod 2 = 0 then
                        backgroundPixel := @colorTransparencyGrid1
                    else
                        backgroundPixel := @colorTransparencyGrid2;

                    // Alpha blending
                    for i := 2 downto 0 do begin
                        THugeByteArray(alignedData^)[offset] := ( alpha * TTexturePixel(aColor)[i] + ( 255 - alpha ) * backgroundPixel[i] ) div 256;
                        offset := offset + 1;
                    end;
                end
                else begin
                    for i := 2 downto 0 do begin
                        THugeByteArray(alignedData^)[offset] := TTexturePixel(aColor)[i];
                        offset := offset + 1;
                    end;
                end;
                offset := offset + 1;
            end;
        end;

        FillChar(bitmapInfo, SizeOf(bitmapInfo), 0);
        bitmapInfo.bmiHeader.biSize := SizeOf(TBitmapInfoHeader);
        bitmapInfo.bmiHeader.biWidth := Image.Width;
        bitmapInfo.bmiHeader.biHeight := -Image.Height;
        bitmapInfo.bmiHeader.biPlanes := 1;
        bitmapInfo.bmiHeader.biBitCount := 32;
        bitmapInfo.bmiHeader.biCompression := BI_RGB;
        bitmapInfo.bmiHeader.biSizeImage := Image.Height * Image.Height * 4;
        SetDIBits(Image.Canvas.Handle, Image.Picture.Bitmap.Handle, 0, Image.Height, alignedData, bitmapInfo, DIB_RGB_COLORS);
        Image.Invalidate;

    finally
        FreeMem(alignedData);
    end;
end;

constructor TImageTextureLoader.Create(const anImage: TImage);
begin
    Inherited Create;

    fImage := anImage;
end;

// #############################################################################

function TOpenGLTextureLoader.LoadTextureIntoOpenGL(const aTextureData: TTextureData; const aShouldFilter, aShouldMipMap: Boolean): GLuint;
var nrBytesPerPixel, format: integer;
    parameter, minParameter: GLUint;

begin
    glGenTextures(1, @result);
    glBindTexture(GL_TEXTURE_2D, result);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1); // aTextureData contains 32bpp or 24bpp pixels, not aligned
    nrBytesPerPixel := aTextureData.BitsPerPixel div 8;
    if nrBytesPerPixel = 4 then
        format := GL_BGRA
    else
        format := GL_BGR;

    if aShouldFilter then begin
        parameter := GL_LINEAR;
        minParameter := GL_LINEAR_MIPMAP_LINEAR;
    end
    else begin
        parameter := GL_NEAREST;
        minParameter := GL_NEAREST_MIPMAP_LINEAR;
    end;

    if aShouldMipMap then begin
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minParameter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, parameter);

        gluBuild2DMipmaps(GL_TEXTURE_2D, nrBytesPerPixel, aTextureData.Width, aTextureData.Height, format, GL_UNSIGNED_BYTE, @aTextureData.Pixels[0]);
    end
    else begin
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, parameter);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, parameter);

        glTexImage2D(GL_TEXTURE_2D, 0, nrBytesPerPixel, aTextureData.Width, aTextureData.Height, 0, format, GL_UNSIGNED_BYTE, @aTextureData.Pixels[0]);
    end;
end;

function TOpenGLTextureLoader.LoadTexture(const aFileName: String; const aShouldFilter, aShouldMipMap: Boolean): GLuint;
var textureData: TTextureData;
begin
    textureData := LoadTextureData(aFileName);

    result := LoadTextureIntoOpenGL(textureData, aShouldFilter, aShouldMipmap);

    // INFO, PJ: Pixelization hack.
    //result := LoadTextureIntoOpenGL(textureData, FALSE, FALSE);
end;

function TOpenGLTextureLoader.LoadColor(const aColor: TColor): GLuint;
var textureData: TTextureData;
    i: integer;
    pixel: PTexturePixel;
    bytesPerPixel: integer;
    containsAlpha: boolean;
begin
    textureData.Width := 4;
    textureData.Height := 4;

    containsAlpha := (aColor and $FF000000) <> 0;

    if containsAlpha then
        textureData.BitsPerPixel := 32
    else
        textureData.BitsPerPixel := 24;

    bytesPerPixel := textureData.BitsPerPixel div 8;

    SetLength(textureData.Pixels, 16 * bytesPerPixel);

    for i := 0 to 15 do begin
        pixel := @textureData.Pixels[i * bytesPerPixel];
        pixel[0] := (aColor shr 16) and $000000FF;
        pixel[1] := (aColor shr 8) and $000000FF;
        pixel[2] := aColor and $000000FF;
        if containsAlpha then
            pixel[3] := (aColor shr 24);
    end;

    result := LoadTextureIntoOpenGL(textureData, false, false);
end;

constructor TOpenGLTextureLoader.Create(const aDeviceContext: HDC; const aRenderingContext: HGLRC);
begin
    inherited Create;

    fDeviceContext := aDeviceContext;
    fRenderingContext := aRenderingContext;
end;

// #############################################################################

constructor TBitmap32TextureLoader.Create( const aBitmap: TBitmap32 );
begin
    inherited Create;

    fBitmap := aBitmap;
end;

procedure TBitmap32TextureLoader.LoadPNGFromResource( const aResourceName: String );
var stream: TStream;
    textureData: TTextureData;
    
begin
    // Assumes the texture is stored as RC_DATA
    stream := TResourceStream.Create( HInstance, aResourceName, RT_RCDATA );
    try
        textureData := LoadPNGTextureData( stream, FALSE );

    finally
        FreeAndNil( stream );
    end;

    // Convert texture data to bitmap32
    fBitmap.Width := textureData.Width;
    fBitmap.Height := textureData.Height;

    if textureData.BitsPerPixel <> 32 then
        raise Exception.Create( 'Currently, only 32 bpp PNGs are supported.' );

    Move( textureData.Pixels[0], fBitmap.Bits[0], textureData.Width * textureData.Height * 4 );
end;

end.

