{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
// This unit is meant to hold convenient instructions for more complicated
// construction and manipulation of texture. Each manipulator operators on
// exactly on one TTextureDataData.
unit unit_TextureManipulator;

{$MODE Delphi}

interface

uses
    Windows,
    Graphics,
    Classes,
    unit_TextureTypes,
    unit_Types;

type
    // Cohen-Sutherland outcode
    TOutCodeElement = ( oceLeft, oceRight, oceTop, oceBottom );
    TOutCode = set of TOutCodeElement;

    TTextureManipulator = class(TObject)
    private
        fTexture: TTextureData;

        function CalculateOutCode(anX, aY: integer): TOutCode;

        procedure Swap(var aVariable1, aVariable2: integer);

        function GetPixel(const anX, aY: integer): integer;
        procedure SetPixel(const anX, aY, aColor: integer);

        procedure LineEx(anX0, aY0, anX1, aY1, aColor: Integer);

    protected
        function GetBytesPerPixel: integer;

        property BytesPerPixel: integer read GetBytesPerPixel;

    public
        constructor Create(const aTexture: TTextureData);
        constructor CreateWithProperties(const aWidth, aHeight, aBitsPerPixel: integer);
        constructor CreateWithBitmap(const aBitmap: TBitmap);
        constructor CreateWithText(
            const aText: string;
            const aFontName: string;
            const aFontSize: integer;
            const aColor: TColor;
            const aShouldDrawReverse: boolean );

        procedure GenerateAlphaForColorKey(const aColorKey, aColorKeyReplacement: TColor);
        procedure AntialiasAlpha;
        procedure ScaleAlpha(const aScale: integer);

        function CopyTexture: TTextureData;
        procedure DownSample(const aReductionRate: Integer);
        procedure ConvertToGreyScale;

        procedure Draw(const anX, aY: Integer; const aTexture: TTextureData);
        procedure DrawOnBitmap(const anX, aY: integer; const aBitmap: TBitmap);
        procedure DrawStretchedOnBitmap(const aBitmap: TBitmap);
        procedure DrawOnTexture(const anX, anY: Integer; const aTexture: TTextureData);
        procedure DrawOnCanvas(const anX, aY: Integer; const aCanvas: TCanvas);


        procedure Line(anX0, aY0, anX1, aY1, aColor: Integer);
        procedure Fill(const aColor: TColor);

        property Texture: TTextureData read fTexture;
        property Pixels[const anX, aY: integer]: integer read GetPixel write SetPixel;
    end;

implementation

uses
    Math,
    Types,
    SysUtils,
    unit_Variables;

constructor TTextureManipulator.Create(const aTexture: TTextureData);
begin
    inherited Create;

    fTexture := aTexture;
end;

constructor TTextureManipulator.CreateWithProperties(const aWidth, aHeight, aBitsPerPixel: integer);
begin
    inherited Create;

    fTexture.Width := aWidth;
    fTexture.Height := aHeight;
    fTexture.BitsPerPixel := aBitsPerPixel;
    SetLength( fTexture.Pixels, fTexture.Width * fTexture.Height * BytesPerPixel);
end;

constructor TTextureManipulator.CreateWithBitmap(const aBitmap: TBitmap);
var bitsPerPixel: integer;
    bitmapInfo: TBitmapInfo;
    y, lineSize, x, index: integer;
    scanLine: array of byte;
begin
    case aBitmap.PixelFormat of
        pf24Bit: bitsPerPixel := 24;
        pf32Bit: bitsPerPixel := 32;
    else
        raise Exception.Create('Unsupported pixel format.');
    end;

    CreateWithProperties(aBitmap.Width, aBitmap.Height, bitsPerPixel);

    FillChar(bitmapInfo, sizeOf(bitmapInfo), 0);
    bitmapInfo.bmiHeader.biSize := SizeOf(TBitmapInfoHeader);
    bitmapInfo.bmiHeader.biWidth := fTexture.Width;
    bitmapInfo.bmiHeader.biHeight := fTexture.Height;
    bitmapInfo.bmiHeader.biPlanes := 1;
    bitmapInfo.bmiHeader.biBitCount := fTexture.BitsPerPixel;
    bitmapInfo.bmiHeader.biCompression := BI_RGB;
    bitmapInfo.bmiHeader.biSizeImage := fTexture.Width * fTexture.Height;

    lineSize := fTexture.Width * fTexture.BitsPerPixel div 8;
    SetLength(scanLine, lineSize);
    for y := 0 to fTexture.Height - 1 do begin
        GetDibits(aBitmap.Canvas.Handle, aBitmap.Handle, fTexture.Height - 1 - y, 1, @scanLine[0], bitmapInfo, DIB_RGB_COLORS);

        // Fix alpha values -> 255 instead of 0
        if fTexture.BitsPerPixel = 32 then begin
            index := 3;
            for x := 0 to fTexture.Width - 1 do begin
                ScanLine[index] := 255;
                index := index + 4;
            end;
        end;

        Move(scanLine[0], fTexture.Pixels[y * lineSize], lineSize);
    end;
end;

constructor TTextureManipulator.CreateWithText(
    const aText: string;
    const aFontName: string;
    const aFontSize: integer;
    const aColor: TColor;
    const aShouldDrawReverse: boolean );
var bitmap: TBitmap;
    alternateColor: TColor;
    textExtent: TSize;
begin
    bitmap := TBitmap.Create;
    try
        bitmap.PixelFormat := pf32bit;
        bitmap.Canvas.Font.Name := aFontName;
        bitmap.Canvas.Font.Size := aFontSize;
        textExtent := bitmap.Canvas.TextExtent(aText);
        bitmap.Width := textExtent.CX;
        bitmap.Height := textExtent.CY;

        if ColorToRGB(aColor) = 0 then
            alternateColor := $FFFFFF
        else
            alternateColor := $0;

        if aShouldDrawReverse then begin
            bitmap.Canvas.Brush.Color := aColor;
            bitmap.Canvas.Font.Color := alternateColor;
        end
        else begin
            bitmap.Canvas.Brush.Color := alternateColor;
            bitmap.Canvas.Font.Color := aColor;
        end;

        bitmap.Canvas.TextOut(0, 0, aText);
        CreateWithBitmap(bitmap);
        GenerateAlphaForColorKey(alternateColor, aColor);

    finally
        FreeAndNil(bitmap);
    end;
end;

function TTextureManipulator.GetBytesPerPixel: integer;
begin
    result := fTexture.BitsPerPixel div 8;
end;

procedure TTextureManipulator.GenerateAlphaForColorKey(const aColorKey, aColorKeyReplacement: TColor);
var x, y, index: integer;
    color, newColor: integer;
begin
    if fTexture.bitsPerPixel = 24 then
        raise Exception.Create( 'Not yet implemented.');

    color := aColorKey and $FFFFFF;
    newColor := aColorKeyReplacement and $FFFFFF;
    index := 0;

    for y := 0 to fTexture.Height - 1 do begin
        for x := 0 to fTexture.Width - 1 do begin
            if PInteger(@fTexture.Pixels[index])^ and $FFFFFF = color then
                PInteger(@fTexture.Pixels[index])^ := newColor 
            else
                fTexture.Pixels[index + 3] := 255;

            index := index + 4;
        end;
    end;
end;

procedure TTextureManipulator.AntialiasAlpha;
// Currently simply ignores all alpha values at the borders.
var newTexture: TTextureData;
    x, y, alpha, subY, subX: integer;
    sourceIndex, targetIndex: integer;
    bytesPerPixel, lineSize: integer;
begin
    if fTexture.BitsPerPixel = 24 then exit;

    newTexture := CopyTexture;

    bytesPerPixel := fTexture.BitsPerPixel div 8;
    lineSize := newTexture.Width * bytesPerPixel;

    targetIndex := lineSize + bytesPerPixel + 3;
    for y := 1 to newTexture.Height - 2 do begin
        for x := 1 to newTexture.Width - 2 do begin
            alpha := 0;
            sourceIndex := targetIndex - lineSize - bytesPerPixel;
            for subY := -1 to 1 do begin
                for subX := -1 to 1 do begin
                    alpha := alpha + fTexture.Pixels[sourceIndex];
                    sourceIndex := sourceIndex + bytesPerPixel;
                end;
                sourceIndex := sourceIndex + lineSize - 3 * bytesPerPixel;
            end;

            newTexture.Pixels[targetIndex] := alpha div 9;
            targetIndex := targetIndex + bytesPerPixel;
        end;

        targetIndex := targetIndex + 2 * bytesPerPixel
    end;

    fTexture := newTexture;
end;

procedure TTextureManipulator.ScaleAlpha(const aScale: integer);
var x, y, index, value: integer;
begin
    if fTexture.bitsPerPixel = 24 then exit;

    index := 3;
    for y := 0 to fTexture.Height - 1 do begin
        for x := 0 to fTexture.Width - 1 do begin
            value := fTexture.Pixels[index];

            fTexture.Pixels[index] := value * aScale div 256;
            index := index + 4;
        end;
    end;
end;

function TTextureManipulator.CopyTexture: TTextureData;
var size: integer;
begin
    result.Width := fTexture.Width;
    result.Height := fTexture.Height;
    result.BitsPerPixel := fTexture.BitsPerPixel;
    size := result.Width * result.Height * result.BitsPerPixel div 8;
    SetLength(result.Pixels, size);
    Move(fTexture.Pixels[0], result.Pixels[0], size);
end;

procedure TTextureManipulator.Draw(const anX, aY: Integer; const aTexture: TTextureData);
// Clipping is not supported yet.
var sourceStride, targetStride, minStride: integer;
    sourceIndex, targetIndex, sourceLineIndex, targetLineIndex: integer;
    x, y, width, height, index: integer;
    sourceLineSize, targetLineSize: integer;
    alpha: integer;
begin
    sourceStride := aTexture.BitsPerPixel div 8;
    targetStride := BytesPerPixel;
    minStride := Min(sourceStride, targetStride);
    sourceLineSize := sourceStride * aTexture.Width;
    targetLineSize := targetStride * fTexture.Width;

    width := fTexture.Width - anX;
    if width > aTexture.Width then width := aTexture.Width;

    height := fTexture.Height - aY;
    if height > aTexture.Height then height := aTexture.Height;

    sourceLineIndex := 0;
    targetLineIndex := aY * targetLineSize + anX * targetStride;;
    for y := 0 to height - 1 do begin
        sourceIndex := sourceLineIndex;
        targetIndex := targetLineIndex;
        if sourceStride = 3 then begin // No Alpha
            for x := 0 to width - 1 do begin
                for index := 0 to minStride - 1 do begin
                    fTexture.Pixels[targetIndex + index] := aTexture.Pixels[sourceIndex + index];
                end;
                sourceIndex := sourceIndex + sourceStride;
                targetIndex := targetIndex + targetStride;
            end;
        end
        else begin // Alpha
            for x := 0 to width - 1 do begin
                alpha := aTexture.Pixels[sourceIndex + 3];
                for index := 0 to minStride - 1 do begin
                    fTexture.Pixels[targetIndex + index] := (aTexture.Pixels[sourceIndex + index] * alpha + fTexture.Pixels[targetIndex + index] * (255 - alpha)) div 256;
                end;
                sourceIndex := sourceIndex + sourceStride;
                targetIndex := targetIndex + targetStride;
            end;
        end;
        targetLineIndex := targetLineIndex + targetLineSize;
        sourceLineIndex := sourceLineIndex + sourceLineSize;
    end;
end;

procedure TTextureManipulator.DrawOnBitmap(const anX, aY: integer; const aBitmap: TBitmap);
var bitmap: TBitmap;
    bitmapInfo, targetBitmapInfo: TBitmapInfo;
    scanline, targetScanLine: array of byte;
    lineSize, y, targetY, sourceIndex, targetIndex, index, alpha: integer;
begin
    bitmap := TBitmap.Create;
    try
        // first we draw on a temporary bitmap
        bitmap.Width := fTexture.Width;
        bitmap.Height := fTexture.Height;
        if fTexture.BitsPerPixel = 24 then
            bitmap.pixelFormat := pf24bit
        else if fTexture.BitsPerPixel = 32 then
            bitmap.PixelFormat := pf32bit
        else
            raise Exception.Create(Format('Unsupported pixel format (%d bpp).', [fTexture.BitsPerPixel]));

        FillChar(bitmapInfo, sizeOf(bitmapInfo), 0);
        bitmapInfo.bmiHeader.biSize := SizeOf(TBitmapInfoHeader);
        bitmapInfo.bmiHeader.biWidth := fTexture.Width;
        bitmapInfo.bmiHeader.biHeight := fTexture.Height;
        bitmapInfo.bmiHeader.biPlanes := 1;
        bitmapInfo.bmiHeader.biBitCount := fTexture.BitsPerPixel;
        bitmapInfo.bmiHeader.biCompression := BI_RGB;
        bitmapInfo.bmiHeader.biSize := fTexture.Width * fTexture.Height * fTexture.BitsPerPixel div 8;

        targetBitmapInfo := bitmapInfo;
        targetBitmapInfo.bmiHeader.biBitCount := 24;
        targetBitmapInfo.bmiHeader.biWidth := aBitmap.Width;
        bitmapInfo.bmiHeader.biSize := 128;

        aBitmap.PixelFormat := pf24bit;

        lineSize := fTexture.Width * fTexture.BitsPerPixel div 8;
        SetLength(scanLine, lineSize);  // Align data per scanline
        SetLength(targetScanLine, 4096);
        FillChar(targetScanLine[0], Length(targetScanLine), 0);
        for y := 0 to fTexture.Height - 1 do begin
            Move(fTexture.Pixels[y * lineSize], scanLine[0], lineSize);

            if fTexture.BitsPerPixel = 32 then begin // alpha support
                targetY := aY + y;
                if (targetY >= 0) and (targetY < aBitmap.Height) then begin
// ToDo: fix scanline
//!!!                    Move(aBitmap.Scanline[targetY]^, targetScanLine[0], 3 * aBitmap.Width);
                    sourceIndex := 0;
                    targetIndex := Max(0, anX) * 3;

                    for index := Max(0, anX) to Min(aBitmap.Width, anX + fTexture.Width) - 1 do begin
                        alpha := scanLine[sourceIndex + 3];
                        scanLine[sourceIndex] := (scanLine[sourceIndex] * alpha + targetScanLine[targetIndex] * (255 - alpha)) div 256;
                        sourceIndex := sourceIndex + 1;
                        targetIndex := targetIndex + 1;
                        scanLine[sourceIndex] := (scanLine[sourceIndex] * alpha + targetScanLine[targetIndex] * (255 - alpha)) div 256;
                        sourceIndex := sourceIndex + 1;
                        targetIndex := targetIndex + 1;
                        scanLine[sourceIndex] := (scanLine[sourceIndex] * alpha + targetScanLine[targetIndex] * (255 - alpha)) div 256;
                        sourceIndex := sourceIndex + 2;
                        targetIndex := targetIndex + 1;
                    end;
                end;
            end;

            SetDibits(bitmap.Canvas.Handle, bitmap.Handle, fTexture.Height - 1 - y, 1, @scanLine[0], bitmapInfo, DIB_RGB_COLORS);
        end;

        // then we blit that bitmap onto the canvas
        aBitmap.Canvas.Draw(anX, aY, bitmap);
    finally
        FreeAndNil(bitmap);
    end;
end;

procedure TTextureManipulator.DrawStretchedOnBitmap(const aBitmap: TBitmap);
var bitmapInfo: TBitmapInfo;
    bytesPerPixel: Integer; x, y, i, sourceLineSize, targetLineSize, sourceOffset, targetOffset: integer;
    alignedData: Pointer;
    xOffset, yOffset, xFactor, yFactor: integer;
    backgroundPixel: PTexturePixel;
    alpha: integer;
begin
    bytesPerPixel := fTexture.BitsPerPixel div 8;

    sourceLineSize := fTexture.Width * bytesPerPixel;
    targetLineSize := (aBitmap.Width  * bytesPerPixel + 3) and $FFFFFFFC;
    GetMem(alignedData, aBitmap.Height * targetLineSize);

    try
        targetOffset := 0;
        xFactor := fTexture.Width * 256 div aBitmap.Width;
        yFactor := fTexture.Height * 256 div aBitmap.Height;

        yOffset := yFactor div 2;
        for y := 0 to aBitmap.Height - 1 do begin
            xOffset := xFactor div 2;

            for x := 0 to aBitmap.Width - 1 do begin
                sourceOffset := (yOffset div 256) * sourceLineSize + (xOffset div 256) * bytesPerPixel;

                if bytesPerPixel = 4 then begin
                    if (x div 8 + y div 8 ) mod 2 = 0 then
                        backgroundPixel := @colorTransparencyGrid1
                    else
                        backgroundPixel := @colorTransparencyGrid2;

                    // Alpha blending
                    alpha := fTexture.Pixels[sourceOffset + 3];
                    for i := 0 to bytesPerPixel - 1 do begin
                        THugeByteArray(alignedData^)[targetOffset] :=
                            ( alpha * fTexture.Pixels[sourceOffset + i] + ( 255 - alpha ) * backgroundPixel[i] ) div 256;
                        targetOffset := targetOffset + 1;
                    end;
                end
                else begin
                    for i := 0 to bytesPerPixel - 1 do begin
                        THugeByteArray(alignedData^)[targetOffset] := fTexture.Pixels[sourceOffset + i];
                        targetOffset := targetOffset + 1;
                    end;
                end;

                xOffset := xOffset + xFactor;
            end;

            yOffset := yOffset + yFactor;
        end;

        FillChar(bitmapInfo, SizeOf(bitmapInfo), 0);
        bitmapInfo.bmiHeader.biSize := SizeOf(TBitmapInfoHeader);
        bitmapInfo.bmiHeader.biWidth := aBitmap.Width;
        bitmapInfo.bmiHeader.biHeight := -aBitmap.Height;
        bitmapInfo.bmiHeader.biPlanes := 1;
        bitmapInfo.bmiHeader.biBitCount := fTexture.BitsPerPixel;
        bitmapInfo.bmiHeader.biCompression := BI_RGB;
        bitmapInfo.bmiHeader.biSizeImage := aBitmap.Height * targetLineSize;
        SetDIBits(aBitmap.Canvas.Handle, aBitmap.Handle, 0, aBitmap.Height, alignedData, bitmapInfo, DIB_RGB_COLORS);

    finally
        FreeMem(alignedData);
    end;
end;

procedure TTextureManipulator.DrawOnTexture(const anX, anY: Integer; const aTexture: TTextureData);
// Convenient reverse of Draw...
var textureManipulator: TTextureManipulator;
begin
    textureManipulator := TTextureManipulator.Create(aTexture);
    try
        textureManipulator.Draw(anX, anY, fTexture);
    finally
        FreeAndNil(textureManipulator);
    end;
end;

procedure TTextureManipulator.DrawOnCanvas(const anX, aY: Integer; const aCanvas: TCanvas);
var bitmap: TBitmap;
    bitmapInfo: TBitmapInfo;
    y: integer;
    lineSize: integer;
    scanLine, targetScanLine: array of byte;

begin
    bitmap := TBitmap.Create;
    try
        // first we draw on a temporary bitmap
        bitmap.Width := fTexture.Width;
        bitmap.Height := fTexture.Height;
        if fTexture.BitsPerPixel = 24 then
            bitmap.pixelFormat := pf24bit
        else if fTexture.BitsPerPixel = 32 then
            bitmap.PixelFormat := pf32bit
        else
            raise Exception.Create(Format('Unsupported pixel format (%d bpp).', [fTexture.BitsPerPixel]));

        FillChar(bitmapInfo, sizeOf(bitmapInfo), 0);
        bitmapInfo.bmiHeader.biSize := SizeOf(TBitmapInfoHeader);
        bitmapInfo.bmiHeader.biWidth := fTexture.Width;
        bitmapInfo.bmiHeader.biHeight := fTexture.Height;
        bitmapInfo.bmiHeader.biPlanes := 1;
        bitmapInfo.bmiHeader.biBitCount := fTexture.BitsPerPixel;
        bitmapInfo.bmiHeader.biCompression := BI_RGB;
        bitmapInfo.bmiHeader.biSize := fTexture.Width * fTexture.Height * fTexture.BitsPerPixel div 8;

        bitmapInfo.bmiHeader.biSize := 128;

        lineSize := fTexture.Width * fTexture.BitsPerPixel div 8;
        SetLength(scanLine, lineSize);  // Align data per scanline
        SetLength(targetScanLine, 4096);
        FillChar(targetScanLine[0], Length(targetScanLine), 0);
        for y := 0 to fTexture.Height - 1 do begin
            Move(fTexture.Pixels[y * lineSize], scanLine[0], lineSize);

            if fTexture.BitsPerPixel = 32 then begin // alpha support
                raise Exception.Create( 'Alpha not supported yet for this draw method.' );
            end;

            SetDibits(bitmap.Canvas.Handle, bitmap.Handle, fTexture.Height - 1 - y, 1, @scanLine[0], bitmapInfo, DIB_RGB_COLORS);
        end;

        // then we blit that bitmap onto the canvas
        aCanvas.Draw(anX, aY, bitmap);
    finally
        FreeAndNil(bitmap);
    end;
end;

procedure TTextureManipulator.DownSample(const aReductionRate: Integer);
// Downsamples every rate x rate block to one pixel . Only apply to textures
// that are a multiple of rate, any pixels outside are ignored.
var newTexture: TTextureData;
    index, sourceIndex, stride, sqrRate: integer;
    x, y, subX, subY, subIndex: integer;
    sourceLineWidth: integer;
    value: array[0..3] of integer;
begin
    newTexture.Width := fTexture.Width div aReductionRate;
    newTexture.Height := fTexture.Height div aReductionRate;
    newTexture.BitsPerPixel := fTexture.BitsPerPixel;
    SetLength(newTexture.Pixels, newTexture.Width * newTexture.Height * BytesPerPixel);
    FillChar(newTexture.Pixels[0], newTexture.Width * newTexture.Height * BytesPerPixel, $FF);

    sqrRate := aReductionRate * aReductionRate;
    stride := BytesPerPixel;
    sourceLineWidth := BytesPerPixel * fTexture.Width;

    sourceIndex := 0;
    index := 0;
    for y := 0 to newTexture.Height - 1 do begin
        for x := 0 to newTexture.Width - 1 do begin
            value[0] := 0; value[1] := 0; value[2] := 0; value[3] := 0;
            for subY := 0 to aReductionRate - 1 do begin
                for subX := 0 to aReductionRate - 1 do begin
                    for subIndex := 0 to stride - 1 do begin
                        value[subIndex] := value[subIndex] + fTexture.Pixels[sourceIndex];
                        sourceIndex := sourceIndex + 1;
                    end;
                end;
                sourceIndex := sourceIndex + sourceLineWidth - aReductionRate * stride;
            end;
            sourceIndex := sourceIndex - aReductionRate * sourceLineWidth + aReductionRate * stride;

            for subIndex := 0 to stride - 1 do begin
                newTexture.Pixels[index] := value[subIndex] div sqrRate;
                index := index + 1;
            end;
        end;
        sourceIndex := sourceIndex + ( aReductionRate - 1 ) * sourceLineWidth + (fTexture.Width mod aReductionRate) * stride;
    end;

    fTexture := newTexture;
end;

procedure TTextureManipulator.ConvertToGreyScale;
var x, y, value: integer;
    pixel24: PTexturePixel24;
    pixel32: PTexturePixel;

begin
    if fTexture.bitsPerPixel = 24 then begin
        pixel24 := @fTexture.Pixels[0];
        for y := 0 to fTexture.Height - 1 do begin
            for x := 0 to fTexture.Width - 1 do begin
                value := ( pixel24[0] + pixel24[1] + pixel24[2] ) div 3;
                pixel24[0] := value;
                pixel24[1] := value;
                pixel24[2] := value;
                inc(pixel24);
            end;
        end;
    end
    else begin
        pixel32 := @fTexture.Pixels[0];
        for y := 0 to fTexture.Height - 1 do begin
            for x := 0 to fTexture.Width - 1 do begin
                value := ( pixel32[0] + pixel32[1] + pixel32[2] ) div 3;
                pixel32[0] := value;
                pixel32[1] := value;
                pixel32[2] := value;
                inc(pixel32);
            end;
        end;
    end;
end;

procedure TTextureManipulator.Swap(var aVariable1, aVariable2: integer);
var temp: integer;
begin
    temp := aVariable1;
    aVariable1 := aVariable2;
    aVariable2 := temp;
end;

// Bresenham's line algorithm with Cohen-Sutherland clipping applied.
// PJ: Original algorithm uses integer division. This is equivalent to Truncing
// the value towards zero, meaning asymmetry towards swapping start and endPoint
// of the line. Using Floor instead fixes this.
procedure TTextureManipulator.Line(anX0, aY0, anX1, aY1, aColor: Integer);
var outCode0, outCode1, outCode: TOutCode;
    x, y: integer;
begin
    outCode0 := CalculateOutCode(anX0, aY0);
    outCode1 := CalculateOutCode(anX1, aY1);

    if outCode0 * outCode1 <> [] then exit;

    x := 0;
    y := 0;
    while (outCode0 <> []) or (outCode1 <> []) do
    begin
        // What point should we clip first
        if outCode0 <> [] then
            outCode := outCode0
        else
            outCode := outCode1;
                                                 
        if oceBottom in outCode then begin
            y := fTexture.Height - 1;
          x := anX0 + Floor(( anX1 - anX0 ) * ( y - aY0 ) / ( aY1 - aY0 ));

        end
        else if oceTop in outCode then begin
            y := 0;
            x := anX0 + Floor(( anX1 - anX0 ) * ( y - aY0 ) / ( aY1 - aY0 ));
        end
        else if oceRight in outCode then begin
            x := fTexture.Width - 1;
            y := aY0 + Floor(( aY1 - aY0 ) * ( x - anX0 ) / ( anX1 - anX0 ));
        end
        else if oceLeft in outCode then begin
            x := 0;
            y := aY0 + Floor(( aY1 - aY0 ) * ( x - anX0 ) / ( anX1 - anX0 ));
        end;

        if outCode = outCode0 then begin
            anX0 := x;
            aY0 := y;
            outCode0 := CalculateOutCode(anX0, aY0);
        end
        else begin
            anX1 := x;
            aY1 := y;
            outCode1 := CalculateOutCode(anX1, aY1);
        end;
    end;

    LineEx(anX0, aY0, anX1, aY1, aColor);
end;

// Bresenham's line algorithm... non-clipping
procedure TTextureManipulator.LineEx(anX0, aY0, anX1, aY1, aColor: Integer);
var deltaX, deltaY, error, yStep, x, y: integer;
    isSteep: boolean;
begin
    isSteep := Abs(aY1 - aY0) > Abs(anX1 - anX0 );
    if isSteep then begin
        Swap(anX0, aY0);
        Swap(anX1, aY1);
    end;

    if anX0 > anX1 then begin
        Swap(anX0, anX1);
        Swap(aY0, aY1);
    end;

    deltaX := anX1 - anX0;
    deltaY := Abs(aY1 - aY0);
    error := 0;
    y := aY0;
    if aY0 < aY1 then
        yStep := 1
    else
        yStep := -1;

    for x := anX0 to anX1 do begin
        if isSteep then
            Pixels[y, x] := aColor
        else
            Pixels[x, y] := aColor;

        error := error + deltaY;

        if 2 * error >= deltaX then begin
            y := y + yStep;
            error := error - deltaX;
        end;
    end;
end;

function TTextureManipulator.GetPixel(const anX, aY: integer): integer;
var source, target: PByte;
    i, bytesPerPixel, offset: integer;
begin
    bytesPerPixel := fTexture.BitsPerPixel div 8;
    offset := ( aY * fTexture.Width + anX ) * bytesPerPixel;

    result := 0;

    source := @fTexture.Pixels[offset];
    target := @result;

    for i := 0 to bytesPerPixel - 1 do begin
        target^ := source^;

        inc(source);
        inc(target);
    end;
end;

procedure TTextureManipulator.SetPixel(const anX, aY, aColor: integer);
var source, target: PByte;
    i, bytesPerPixel, offset: integer;
begin
    bytesPerPixel := fTexture.BitsPerPixel div 8;
    offset := ( aY * fTexture.Width + anX ) * bytesPerPixel;

    source := @aColor;
    target := @fTexture.Pixels[offset];

    for i := 0 to bytesPerPixel - 1 do begin
        target^ := source^;

        inc(source);
        inc(target);
    end;
end;


// Calculates the Cohen-Sutherland outcode
function TTextureManipulator.CalculateOutCode(anX, aY: integer): TOutCode;
begin
    result := [];

    if anX < 0 then
        Include( result, oceLeft )
    else if anX > fTexture.Width - 1 then
        Include( result, oceRight );

    if aY < 0 then
        Include( result, oceTop )
    else if aY > fTexture.Height - 1 then
        Include( result, oceBottom );
end;

procedure TTextureManipulator.Fill(const aColor: TColor);
var source32, target32: PTexturePixel;
    source24: TTexturePixel24;
    target24: PTexturePixel24;
    i: integer;

begin
    if fTexture.BitsPerPixel = 24 then begin
        source24[0] := (aColor and $FF0000) shr 16;
        source24[1] := (aColor and $00FF00) shr 8;
        source24[2] :=  aColor and $0000FF;
        target24 := @fTexture.Pixels[0];
        for i := 0 to fTexture.Width * fTexture.Height - 1 do begin
            target24^ := source24;
            inc(target24);
        end;
    end
    else if fTexture.BitsPerPixel = 32 then begin
        source32 := @aColor;
        target32 := @fTexture.Pixels[0];
        for i := 0 to fTexture.Width * fTexture.Height - 1 do begin
            target32^ := source32^;
            inc(target32);
        end;
    end
    else
        raise Exception.Create( 'Unsupported pixel format for fill operation.' );
end;

end.
