{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_TextureWriter;

{$MODE Delphi}

interface

uses
    Classes, Types,
    unit_TextureTypes;

type
    TTextureWriter = class( TObject )
    private
        fCompressionQuality: integer;
        procedure SaveBMPTextureData(const aTextureData: TTextureData; const aStream: TStream);
        procedure SaveJPGTextureData(const aTextureData: TTextureData; const aStream: TStream);
        procedure SaveTGATextureData(const aTextureData: TTextureData; const aStream: TStream);
        procedure SavePNGTextureData(const aTextureData: TTextureData; const aStream: TStream);
        procedure SaveDDSTextureData(const aTextureData: TTextureData; const aStream: TStream);
    public
        constructor Create;
        procedure SaveTextureData(const aTextureData: TTextureData; const aFileName: string);

        class function SupportsFileFormat(const anExtension: String): boolean;
        class function SupportedFilesMask: String;
        class function SupportedFileExtensions: TStringDynArray;                              
        property CompressionQuality: integer read fCompressionQuality write fCompressionQuality;
    end;

// #############################################################################

implementation

uses
    SysUtils, Windows, Graphics,
    unit_TGALoader, unit_PNGSupport, unit_DDSSupport;

constructor TTextureWriter.Create;
begin
    inherited Create;

    fCompressionQuality := 100;
end;

procedure TTextureWriter.SaveBMPTextureData(const aTextureData: TTextureData; const aStream: TStream);
var size, lineSize, alignedLineSize, index: integer;
    fileHeader: TBitmapFileHeader;
    infoHeader: TBitmapInfoHeader;
    flippedData: pointer;
    source, target: PChar;

begin
    FillChar( fileHeader, sizeOf( fileHeader ), 0 );
    fileHeader.bfType := $4D42; // 'BM'

    // 4 byte alignment
    lineSize := aTextureData.Width * aTextureData.BitsPerPixel div 8;
    alignedLineSize := ( lineSize + 3 ) and $FFFFFFFC;
    size :=  alignedLineSize * aTextureData.Height;

    fileHeader.bfSize := size + sizeOf( fileHeader ) + sizeOf( infoHeader );
    fileHeader.bfOffbits := sizeOf( fileHeader ) + sizeOf( infoHeader );

    FillChar( infoHeader, sizeof( infoHeader ), 0 );
    infoHeader.biSize := sizeOf( infoHeader );
    infoHeader.biWidth := aTextureData.Width;
    infoHeader.biHeight := aTextureData.Height;
    infoHeader.biPlanes := 1;
    infoHeader.biBitCount := aTextureData.BitsPerPixel;
    infoHeader.biCompression := BI_RGB;
    infoHeader.biSizeImage := size;

    aStream.WriteBuffer( fileHeader, sizeOf( fileHeader ));
    aStream.WriteBuffer( infoHeader, sizeOf( infoHeader ));

    // flip and byte align
    GetMem( flippedData, size );
    source := PChar( @aTextureData.Pixels[0] ) + ( aTextureData.Height - 1 ) * lineSize;
    target := flippedData;
    for index := 0 to aTextureData.Height - 1 do begin
        move( source^, target^, lineSize );

        source := source - lineSize;
        target := target + alignedLineSize;
    end;

    aStream.WriteBuffer( flippedData^ , size );
    freeMem( flippedData );
end;

procedure TTextureWriter.SaveTextureData(const aTextureData: TTextureData; const aFileName: string);
var stream: TFileStream;
    extension: string;
begin
    stream := TFileStream.Create(aFileName, fmCreate);
    try
        extension := LowerCase(ExtractFileExt(aFileName));

        try
            if extension = '.bmp' then
                SaveBMPTextureData(aTextureData, stream)
            else if extension = '.jpg' then
                SaveJPGTextureData(aTextureData, stream)
            else if extension = '.tga' then
                SaveTGATextureData(aTextureData, stream)
            else if extension = '.png' then
                SavePNGTextureData(aTextureData, stream)
            else if extension = '.dds' then
                SaveDDSTextureData(aTextureData, stream)
            else
                raise Exception.Create(Format('Unsupported file format (%s)', [extension]));
        except
            on e: Exception do begin
                raise Exception.Create(Format('Error saving "%s": %s', [aFileName, e.Message]));
            end;
        end;

    finally
        FreeAndNil(stream);
    end;
end;

procedure TTextureWriter.SaveJPGTextureData(const aTextureData: TTextureData; const aStream: TStream);
var bitmap: TBitmap;
    image: TJPEGImage;
    bitmapInfo: TBitmapInfo;
    scanline: array of byte;
    y, lineSize: integer;
begin
    bitmap := TBitmap.Create;
    if aTextureData.BitsPerPixel = 24 then
        bitmap.pixelFormat := pf24bit
    else if aTextureData.BitsPerPixel = 32 then
        bitmap.PixelFormat := pf32bit
    else
        raise Exception.Create(Format('Unsupported pixel format (%d bpp).', [aTextureData.BitsPerPixel]));

    bitmap.Width := aTextureData.Width;
    bitmap.Height := aTextureData.Height;

    FillChar(bitmapInfo, SizeOf(bitmapInfo), 0);
    bitmapInfo.bmiHeader.biSize := SizeOf(TBitmapInfoHeader);
    bitmapInfo.bmiHeader.biWidth := aTextureData.Width;
    bitmapInfo.bmiHeader.biHeight := aTextureData.Height;
    bitmapInfo.bmiHeader.biPlanes := 1;
    bitmapInfo.bmiHeader.biBitCount := aTextureData.BitsPerPixel;
    bitmapInfo.bmiHeader.biCompression := BI_RGB;
    bitmapInfo.bmiHeader.biSizeImage := aTextureData.Width * aTextureData.Height;

    lineSize := aTextureData.Width * aTextureData.BitsPerPixel div 8;
    SetLength( scanLine, lineSize );  // Align data per scanline
    for y := 0 to aTextureData.Height - 1 do begin
        Move(aTextureData.Pixels[y * lineSize], scanLine[0], lineSize);
        SetDibits(bitmap.Canvas.Handle, bitmap.Handle, aTextureData.Height - 1 - y, 1, @scanLine[0], bitmapInfo, DIB_RGB_COLORS);
    end;

    image := TJPEGImage.create;
    image.CompressionQuality := CompressionQuality;
    image.Assign(bitmap);
    image.SaveToStream(aStream);
    FreeAndNil(image);
    FreeAndNil(bitmap);
end;

procedure TTextureWriter.SaveTGATextureData(const aTextureData: TTextureData; const aStream: TStream);
var tgaHeader: TTGAHeader;
begin
    fillChar(tgaHeader, SizeOf(tgaHeader), 0);
    tgaHeader.ImageType := 2;
    tgaHeader.Width := aTextureData.Width;
    tgaHeader.Height := aTextureData.Height;
    tgaHeader.ImageInfo := 32; // flip
    tgaHeader.BPP := aTextureData.BitsPerPixel;

    aStream.Write(tgaHeader, SizeOf(tgaHeader));
    aStream.Write(aTextureData.Pixels[0], aTextureData.Height *
        aTextureData.Width * aTextureData.BitsPerPixel div 8);
end;

procedure TTextureWriter.SavePNGTextureData(const aTextureData: TTextureData; const aStream: TStream);
var writer: TPNGWriter;
begin
    writer := TPNGWriter.Create(aStream);
    try
        writer.Write(aTextureData);
    finally
        FreeAndNil(writer);
    end;
end;

procedure TTextureWriter.SaveDDSTextureData(const aTextureData: TTextureData; const aStream: TStream);
var writer: TDDSWriter;
begin
    writer := TDDSWriter.Create(aStream);
    try
        writer.Write(aTextureData);
    finally
        FreeAndNil(writer);
    end;
end;

class function TTextureWriter.SupportsFileFormat(const anExtension: String): boolean;
var extension: string;
    supportedExtensions: TStringDynArray;
    i: integer;
begin
    extension := LowerCase(anExtension);
    result := false;

    supportedExtensions := SupportedFileExtensions;
    for i := Low(supportedExtensions) to High(supportedExtensions) do begin
        if extension = supportedExtensions[i] then begin
            result := true;
            exit;
        end;
    end;
end;

class function TTextureWriter.SupportedFilesMask: String;
begin
    result := 'Bitmap (*.bmp)|*.bmp|DirectDraw Surface (DDS) (*.dds)|*.dds|JPEG (*.jpg)|*.jpg|TARGA (*.tga)|*.tga';
end;

// Note, PJ: Keep this manually synched with SupportedFilesMask!
//           (Needed for TSaveDialog automatic extension completion support)
class function TTextureWriter.SupportedFileExtensions: TStringDynArray;
begin
    SetLength( result, 5 );
    result[0] := '.bmp';
    result[1] := '.jpg';
    result[2] := '.dds';
    result[3] := '.tga';
    result[4] := '.png';
end;

end.
