{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Textures;

{$MODE Delphi}

interface

uses
  Variants, Types, StrUtils,
  Windows, Messages, SysUtils,
  Classes, Graphics, Dialogs, ComCtrls, StdCtrls, dglOpenGL, Contnrs, ExtCtrls,
  frmOGLContextForm,
  unit_IntegerList,
  unit_TextureTypes;

type
  TTexture = class(TObject)
  private
    fTextureID: integer;
    fCategoryID: integer;
    fGraphicsID: cardinal;                // for OpenGL
    fFilename: string;                    // complete filename of texture in Texture directory
    fWidth: integer;
    fHeight: integer;                     // widht and height of the texture
    fHDC: HDC;
    fHRC: HGLRC;
    fIsColorTexture: boolean;
    fColor: TColor;
    fSelected: boolean;
    ltgc: boolean;                        // flag: texture loaded to graphicscard
    fIsValid: boolean;
    fMipmap: boolean;
    fShouldFilter: boolean;
    fAddressingMode: TTextureAddressingMode;
    function GetShortFilename: string;
  public
    constructor Create(f: string; textureHDC: HDC; textureHRC: HGLRC);
    constructor CreateColor(aColor: TColor; textureHDC: HDC; textureHRC: HGLRC);
    destructor Destroy; override;
    function Copy: TTexture;
    function LoadToGraphicsCard: boolean;
    procedure RemoveFromGraphicsCard;
    function Equals(ATexture: TTexture): boolean;
    function IsValid: boolean;
    procedure Invalidate;
    property TextureID: integer read fTextureID write fTextureID;
    property CategoryID: integer read fCategoryID write fCategoryID;
    property GraphicsID: cardinal read fGraphicsID write fGraphicsID;
    property FileName: string read fFilename write fFilename;
    property ShortFilename: string read GetShortFilename;   // short filename, relative to texturedirectory
    property Width: integer read fWidth write fWidth;
    property Height: integer read fHeight write fHeight;
    property H_DC: HDC read fHDC write fHDC;        // device context for this texture
    property H_RC: HGLRC read fHRC write fHRC;      // rendering context for this texture
    property IsColorTexture: boolean read fisColorTexture write fisColorTexture default false;
    property Color: TColor read fColor write fColor;
    property Selected: boolean read fSelected write fSelected default false;
    property MipMap: boolean read fMipmap write fMipmap default true;
    property AddressingMode: TTextureAddressingMode read fAddressingMode write fAddressingMode;
    property ShouldFilter: boolean read fShouldFilter write fShouldFilter;
  end;

  TTextureList = class(TObjectList)
  private
    fOGLContext: TfrmOGLContext;
    fAutoUpload: boolean;
    fMipmap: boolean;
    fAddressingMode: TTextureAddressingMode;
    fShouldFilter: boolean;
  protected
    function GetTexture(const anIndex: integer): TTexture;
    function GetSelectedTextures: TTextureList;
  public
    constructor Create(AOwnsObjects: boolean); overload;
    constructor Create(AOwnsObjects: boolean; AOGLContext: TfrmOGLContext); overload;
    function InList(const texture: TTexture): boolean; overload;
    function InList(const filename: string): TTexture; overload;
    function AddTexture(ATexture: TTexture; AHDC: HDC; AHRC: HGLRC): integer; overload;
    function AddTexture(ATexture: TTexture): integer; overload;
    function AddTexture(AFilename: string; AHDC: HDC; AHRC: HGLRC): TTexture; overload;
    function AddTexture(AFilename: string): TTexture; overload;
    function AddTexture(AColor: TColor; AHDC: HDC; AHRC: HGLRC): TTexture; overload;
    function AddTexture(AColor: TColor): TTexture; overload;
    procedure UploadTextures(AHDC: HDC; AHRC: HGLRC); overload;
    procedure UploadTextures; overload;
    procedure InvalidateTextures;
    property Textures[const anIndex: integer]: TTexture read GetTexture; default;
    property SelectedTextures: TTextureList read GetSelectedTextures;
    property OGLContext: TfrmOGLContext read fOGLContext write fOGLContext;     // OpenGL rendering context to which textures are uploaded
    property AutoUpload: boolean read fAutoUpload write fAutoUpload;            // should new textures be uploaded automatically to perspective views etc?
    property MipMap: boolean read fMipmap write fMipmap default true;           // passed to any texture owned by the list
    property AddressingMode: TTextureAddressingMode read fAddressingMode write fAddressingMode; // passed to any texture owned by the list
    property ShouldFilter: boolean read fShouldFilter write fShouldFilter;      // passed to any texture owned by the list
  end;

  procedure glDeleteTextures(n: GLsizei; var textures: GLuint); stdcall; external opengl32;

implementation

uses unit_TextureLoader,
     unit_Functions,
     unit_Variables,
     unit_Constants;

// ### TTexture ################################################################

// returns the filename and main directory of the texture within the DeleD texturedirectory
function TTexture.GetShortFilename: string;
begin
    Result := ExtractRelativePath(dir_texture, FileName);
end;

function TTexture.Equals(ATexture: TTexture): boolean;
begin
    Result :=
        (IsColorTexture = ATexture.IsColorTexture) and
        (Color = ATexture.Color) and
        (Uppercase(Filename)=Uppercase(ATexture.Filename)) and
        (IsValid = ATexture.IsValid);
end;

function TTexture.IsValid: Boolean;
begin
    Result := fIsValid;
end;

procedure TTexture.Invalidate;
begin
    fIsValid := false;
end;

constructor TTexture.Create(f: string; textureHDC: HDC; textureHRC: HGLRC);     // HDC+HRC should not be here?
var textureLoader: TTextureLoader;
    textureInfo: TTextureData;
begin
    isColorTexture := false;
    Color := -1;
    Filename := f;
    {$IFDEF DELED_AUTOGENERATE_TEXTURES}
    if not FileExists(FileName) then begin
        Width := 2;
        Height := 2;
    end else
    {$ENDIF}

    fselected := false;
    fTextureID := -1;
    fCategoryID := -1;
    fGraphicsID := 0;
    H_DC := textureHDC;
    H_RC := textureHRC;
    fMipMap := true;
    fIsValid := true;
    fShouldFilter := TRUE;
    fAddressingMode := tamWrap;                                                 

    textureLoader := TTextureLoader.Create;
    try
        textureInfo := textureLoader.LoadTextureInfo(f);
        Width := textureInfo.Width;
        Height := textureInfo.Height;
    finally
        FreeAndNil(textureLoader);
    end;
end;

constructor TTexture.CreateColor(aColor: TColor; textureHDC: HDC; textureHRC: HGLRC);
begin
    isColorTexture := true;
    Color := aColor;
    Filename := ColorToHexString(color);
    Width := 2;
    Height := 2;
    fSelected := false;
    fTextureID := -1;
    fCategoryID := -1;
    fGraphicsID := 0;
    H_DC := textureHDC;
    H_RC := textureHRC;
    fMipMap := false;
    fIsValid := true;
end;

destructor TTexture.Destroy;
begin
    if ltgc then RemoveFromGraphicsCard;
    inherited;
end;

// TODO: should use Assign procedure
function TTexture.Copy: TTexture;
var texture: TTexture;
begin
    texture := TTexture.Create(Filename, H_DC,H_RC);
    texture.TextureID := -1;
    texture.CategoryID := -1;
    texture.GraphicsID := GraphicsID;
    texture.Selected := Selected;
    Result := texture;
end;

function TTexture.LoadToGraphicsCard: boolean;
var textureLoader: TOpenGLTextureLoader;
begin
    try
        if not ltgc then begin
            textureLoader := TOpenGLTextureLoader.Create(H_DC, H_RC);
            try
                if not wglMakeCurrent(H_DC, H_RC) then raise Exception.Create('Unable to access OpenGL context');
                if not isColorTexture then
                    GraphicsID := textureLoader.LoadTexture(Filename, ShouldFilter, MipMap)
                else
                    GraphicsID := textureLoader.LoadColor(Color);

                if AddressingMode <> tamWrap then begin
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, OPENGL_TEXTURE_ADDRESSING_MODES[fAddressingMode]);
                    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, OPENGL_TEXTURE_ADDRESSING_MODES[fAddressingMode]);
                end;

                ltgc := true;
            finally
                FreeAndNil(textureLoader);
            end;
        end;
    except
        on E: Exception do raise Exception.Create(Format('Error loading texture to graphics card: %s',[e.Message]));
    end;
    Result := ltgc;
end;

procedure TTexture.RemoveFromGraphicsCard;
begin
    try
        if ltgc then begin
            wglMakeCurrent(H_DC, H_RC);
            glDeleteTextures(1,fGraphicsID);

            ltgc := false;
        end;
    except
        on E: Exception do Exception.Create('Error removing texture from graphics card.');
    end;
end;

// ### TTextureList ############################################################

constructor TTextureList.Create(AOwnsObjects: boolean);
begin
    inherited Create(AOwnsObjects); // JuMa: call to inherited constructor was missing
    fAutoUpload := false;
end;

constructor TTextureList.Create(AOwnsObjects: boolean; AOGLContext: TfrmOGLContext);
begin
    inherited Create(AOwnsObjects); // JuMa: call to inherited constructor was missing
    OGLContext := AOGLContext;
    fAutoUpload := false;
end;
{
procedure TTextureList.SetHDC(value: HDC);
var i: integer;
begin
    fHDC := value;
    for i := 0 to Count-1 do
        Textures[i].H_DC := value;
end;

procedure TTextureList.SetHRC(value: HGLRC);
var i: integer;
begin
    fHRC := value;
    for i := 0 to Count-1 do
        Textures[i].H_RC := value;
end;    }

function TTextureList.InList(const texture: TTexture): boolean;
var i: integer;
begin
    Result := false;
    for i := 0 to Count-1 do begin
        if Textures[i].Equals(texture) then begin
            Result := true;
            break;
        end;
    end;
end;

function TTextureList.InList(const filename: string): TTexture;
var i: integer;
begin
    Result := nil;
    for i := 0 to Count-1 do begin
        if Textures[i].Filename = filename then begin
            Result := Textures[i];
            break;
        end;
    end;
end;

function TTextureList.AddTexture(ATexture: TTexture; AHDC: HDC; AHRC: HGLRC): integer;
begin
    if AutoUpload then begin
        ATexture.H_DC := AHDC;
        ATexture.H_RC := AHRC;
        if OwnsObjects then begin
            ATexture.MipMap := MipMap;
            ATexture.AddressingMode := AddressingMode;
            ATexture.ShouldFilter := ShouldFilter;
        end;
        ATexture.LoadToGraphicsCard;
    end;
    Result := Add(ATexture);
end;

// adds a texture to te list and returns its index
function TTextureList.AddTexture(ATexture: TTexture): integer;
begin
    if AutoUpload then begin
        ATexture.H_DC := OGLContext.H_DC;
        ATexture.H_RC := OGLContext.H_RC;
        if OwnsObjects then begin
            ATexture.MipMap := MipMap;
            ATexture.AddressingMode := AddressingMode;
            ATexture.ShouldFilter := ShouldFilter;
        end;
        ATexture.LoadToGraphicsCard;
    end;
    Result := Add(ATexture);
end;

function TTextureList.AddTexture(AFilename: string; AHDC: HDC; AHRC: HGLRC): TTexture;
var texture: TTexture;
begin
    texture := InList(AFilename);

    // if texture does not exist, create a new one
    if not Assigned(texture) then begin
        texture := TTexture.Create(AFilename,AHDC,AHRC);
        if OwnsObjects then begin
            texture.MipMap := MipMap;
            texture.AddressingMode := AddressingMode;
            texture.ShouldFilter := ShouldFilter;
        end;
        if AutoUpload then texture.LoadToGraphicsCard;
        Add(texture);
    end;
    Result := texture;
end;

// if the texture does not exist in the list, create a texture object given a filename
// and add it to the list + return created texture; if it does already exist, simply
// return that particular texture object
function TTextureList.AddTexture(AFilename: string): TTexture;
var texture: TTexture;
begin
    texture := InList(AFilename);

    // if texture does not exist, create a new one
    if not Assigned(texture) then begin
        if Assigned(OGLContext) then begin
            texture := TTexture.Create(AFilename,OGLContext.H_DC,OGLContext.H_RC);
        end
        else
            texture := TTexture.Create(AFilename, 0, 0 );
        if OwnsObjects then begin
            texture.MipMap := MipMap;
            texture.AddressingMode := AddressingMode;
            texture.ShouldFilter := ShouldFilter;
        end;
        if AutoUpload then texture.LoadToGraphicsCard;
        Add(texture);
    end;
    Result := texture;
end;

function TTextureList.AddTexture(AColor: TColor; AHDC: HDC; AHRC: HGLRC): TTexture;
var texture: TTexture;
begin
    texture := InList(ColorToHexString(AColor));

    // if texture does not exist, create a new one
    if not Assigned(texture) then begin
        texture := TTexture.CreateColor(AColor,AHDC,AHRC);
        if AutoUpload then
            texture.LoadToGraphicsCard;
        Add(texture);
    end;
    Result := texture;
end;

// checks if a color-texture exists and if not, create and adds it to the list
function TTextureList.AddTexture(AColor: TColor): TTexture;
var texture: TTexture;
begin
    texture := InList(ColorToHexString(AColor));

    // if texture does not exist, create a new one
    if not Assigned(texture) then begin
        if Assigned( OGLContext ) then
            texture := TTexture.CreateColor(AColor,OGLContext.H_DC,OGLContext.H_RC)
        else
            texture := TTexture.CreateColor(AColor,0,0);
        if AutoUpload then
            texture.LoadToGraphicsCard;
        Add(texture);
    end;
    Result := texture;
end;

procedure TTextureList.UploadTextures(AHDC: HDC; AHRC: HGLRC);
var i: integer;
    texture: TTexture;
begin
    for i := 0 to Count-1 do begin
        texture := Textures[i];
        texture.RemoveFromGraphicsCard;     // the texture might already have been uploaded to another device-/rendering context so remove it first
        texture.H_DC := AHDC;               // all textures now receive this particular device-/rendering context
        texture.H_RC := AHRC;
        texture.LoadToGraphicsCard;         // and are uploaded to that device-/rendering context
    end;
end;

procedure TTextureList.UploadTextures;
var i: integer;
    texture: TTexture;
begin
    for i := 0 to Count-1 do begin
        texture := Textures[i];
        texture.H_DC := OGLContext.H_DC;
        texture.H_RC := OGLContext.H_RC;
        texture.LoadToGraphicsCard;
    end;
end;

procedure TTextureList.InvalidateTextures;
var i: integer;
    texture: TTexture;
begin
    for i := 0 to Count-1 do begin
        texture := Textures[i];
        texture.Invalidate;
    end;
end;

function TTextureList.GetTexture(const anIndex: integer): TTexture;
begin
    if anIndex > Count then
        Result := nil
    else
        Result := TTexture(GetItem(anIndex));
end;

// returns all selected textures in a new list
// note that the new list does NOT own the textures
function TTextureList.GetSelectedTextures: TTextureList;
var i: integer;
    list: TTextureList;
begin
    list := TTextureList.Create(false);
    for i := 0 to Count-1 do begin
        if Textures[i].selected then
            list.AddTexture(Textures[i]);
    end;
    Result := list;
end;

end.

