{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Tiler;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  INFO, PJ: Generic tiler. Usage
  - Feed it TileInfo's, with a reference to your own custom objects
  - Call Tile when you are done
  - Process the information from the returned TileInfo's into your own objects.

  Algorithm description (in Dutch):
    Tiling algorithm voor N windows:
    - Bepaal de slots
    - Gooi de windows in een queue
    - Zolang nog windows in de queue, pak een window
      - Sorteer slots op afstand. Bepaal dichtstbijzijnde doel-slot (op basis van middelpunt).
      - als doel-slot al gevuld, vergelijk afstand.
          - Als die afstand kleiner of gelijk, door naar volgende slot
          - Als die afstand groter,
      - als doel-slot niet gevuld, stop window in slot (met afstand) en haal uit queue.

    Bepalen van de slots:
    - Afhankelijk v/d hoogte of de breedte, werk met rijen/kolommen. Uitgaande van kolommen:
    - Bepaal de wortel v/h aantal windows. Trunc. Zoveel kolommen
    - Deel aantal windows door aantal kolommen. Trunc. Zoveel windows per kolom, behalve de [remainder] laatste kolommen:
    daar 1 window meer.
 ---------------------------------------------------------------------------- }

interface

uses
    Contnrs;

type
    TTileInfo = class;
    TSlotInfo = class;

// ----------------------------------------------------------------------------

    TTiler = class( TObject )
    private
        fLeft: integer;
        fTop: integer;
        fWidth: integer;
        fHeight: integer;

        fInputTiles: TObjectList;
        fTiles: TObjectList;

        function GetTile( const anIndex: integer ): TTileInfo;

        function GenerateSlots( const aLeft, aTop, aWidth, aHeight, aNumTiles: integer ): TObjectList;
        procedure AssignSlots( const aTiles, aSlots: TObjectList );
        function SortedSlotsForTile( const aSlots: TObjectList; const aTile: TTileInfo ): TObjectList;

    public
        constructor Create( const aLeft, aTop, aWidth, aHeight: integer );
        destructor Destroy; override;

        procedure AddTileInfo( const aLeft, aTop, aWidth, aHeight: integer; const aTileObject: TObject );

        procedure Tile;

        function NumTiles: integer;

        property Tiles[const anIndex: integer]: TTileInfo read GetTile;
    end;

// ----------------------------------------------------------------------------

    TTileInfo = class( TObject )
    private
        fLeft: integer;
        fTop: integer;
        fWidth: integer;
        fHeight: integer;
        fTileObject: TObject;
        
    public
        constructor Create( const aLeft, aTop, aWidth, aHeight: integer; const aTileObject: TObject );

        property Left: integer read fLeft;
        property Top: integer read fTop;
        property Width: integer read fWidth;
        property Height: integer read fHeight;
        property TileObject: TObject read fTileObject;
    end;

// ----------------------------------------------------------------------------

    TSlotInfo = class( TObject )
    private
        fTile: TTileInfo;
        fTileDistance: integer;
        fCurrentDistance: integer;

        fLeft: integer;
        fTop: integer;
        fWidth: integer;
        fHeight: integer;

    public
        constructor Create( const aLeft, aTop, aWidth, aHeight: integer );

        function IsEmpty: boolean;

        property Left: integer read fLeft;
        property Top: integer read fTop;
        property Width: integer read fWidth;
        property Height: integer read fHeight;
        property Tile: TTileInfo read fTile write fTile;
        property TileDistance: integer read fTileDistance write fTileDistance;
        property CurrentDistance: integer read fCurrentDistance write fCurrentDistance;
    end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
    SysUtils, Types;

constructor TTiler.Create( const aLeft, aTop, aWidth, aHeight: integer );
begin
    inherited Create;

    fLeft := aLeft;
    fTop := aTop;
    fWidth := aWidth;
    fHeight := aHeight;

    fInputTiles := TObjectList.Create( TRUE );
end;

// ----------------------------------------------------------------------------

destructor TTiler.Destroy;
begin
    FreeAndNil( fInputTiles );
    FreeAndNil( fTiles );

    inherited Destroy;
end;

// ----------------------------------------------------------------------------

procedure TTiler.AddTileInfo( const aLeft, aTop, aWidth, aHeight: integer; const aTileObject: TObject );
var tile: TTileInfo;
begin
    tile := TTileInfo.Create( aLeft, aTop, aWidth, aHeight, aTileObject );

    fInputTiles.Add( tile );
end;

// ----------------------------------------------------------------------------

function TTiler.NumTiles: integer;
begin
    result := fInputTiles.Count;
end;

// ----------------------------------------------------------------------------

function TTiler.GetTile( const anIndex: integer ): TTileInfo;
begin
    if not Assigned( fTiles ) then begin
        raise Exception.Create( 'Cannot return tile information. Tile has not been called.' );
    end;

    result := fTiles[anIndex] as TTileInfo;
end;

// ----------------------------------------------------------------------------

procedure TTiler.Tile;
var i: integer;
    tile: TTileInfo;
    slots: TObjectList;
    slot: TSlotInfo;
begin
    if Assigned( fTiles ) then
        raise Exception.Create( 'Tile has already been called.' );

    slots := GenerateSlots( fLeft, fTop, fWidth, fHeight, NumTiles );
    AssignSlots( fInputTiles, slots );

    fTiles := TObjectList.Create( TRUE );
    for i := 0 to slots.Count - 1 do begin
        slot := slots[i] as TSlotInfo;

        if slot.IsEmpty then
            raise Exception.Create( 'Tiling failed, empty slot encountered.' );

        tile := TTileInfo.Create( slot.Left, slot.Top, slot.Width, slot.Height, slot.Tile.tileObject );
        fTiles.add( tile );
    end;

    FreeAndNil( slots );
end;

// ----------------------------------------------------------------------------

function TTiler.GenerateSlots( const aLeft, aTop, aWidth, aHeight, aNumTiles: integer ): TObjectList;
var numColumns: integer;
    numTilesPerColumn: integer;
    numRemainingTiles: integer;
    x, y: integer;
    numTiles: integer;
    slot: TSlotInfo;
    left, top, width, height: integer;
    originalTileWidth, originalTileHeight: integer;
    tileLeft, tileTop, tileWidth, tileHeight: integer;


begin
    result := TObjectList.Create( TRUE );

    if aNumTiles = 0 then exit;

    // INFO, PJ: Swap columns/rows if aspect ratio calls for it.
    if aWidth > aHeight * 1.2 then begin
        left := aTop;
        top := aLeft;
        width := aHeight;
        height := aWidth;
    end
    else
    begin
        left := aLeft;
        top := aTop;
        width := aWidth;
        height := aHeight;
    end;

    numColumns := Trunc( Sqrt( aNumTiles ));
    numTilesPerColumn := aNumTiles div numColumns;
    numRemainingTiles := aNumTiles mod numColumns;

    originalTileWidth := width div numColumns;

    for x := 0 to numColumns - 1 do begin
        numTiles := numTilesPerColumn;

        if x >= (numColumns - numRemainingTiles) then
            numTiles := numTiles + 1;

        tileWidth := originalTileWidth;
        if x = numColumns - 1 then
            tileWidth := tileWidth + ( width mod numColumns );

        originalTileHeight := height div numTiles;

        for y := 0 to numTiles - 1 do begin
            tileLeft := left + x * originalTileWidth;
            tileTop := top + y * originalTileHeight;

            tileHeight := originalTileHeight;
            if y = numTiles - 1 then
                tileHeight := tileHeight + ( height mod numTiles );

            if width = aWidth then
                slot := TSlotInfo.Create( tileLeft, tileTop, tileWidth, tileHeight )
            else // INFO, PJ: Swapped columns/rows...
                slot := TSlotInfo.Create( tileTop, tileLeft, tileHeight, tileWidth );

            result.Add( slot );
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TTiler.AssignSlots( const aTiles, aSlots: TObjectList );
var i: integer;
    slot: TSlotInfo;
    tiles: TObjectList;
    tile: TTileInfo;
    slots: TObjectList;
begin
    if aTiles.Count <> aSlots.Count then
        raise Exception.Create( 'Number of tiles does not match number of slots.' );

    // INFO, PJ: We create a queue of tiles.
    tiles := TObjectList.Create( FALSE );
    tiles.Assign( fInputTiles );

    while tiles.Count > 0 do begin
        tile := tiles[0] as TTileInfo;

        slots := SortedSlotsForTile( aSlots, Tile );
        for i := 0 to slots.Count - 1 do begin
            slot := slots[i] as TSlotInfo;

            if not slot.isEmpty and (slot.CurrentDistance < slot.TileDistance) then begin
                // INFO, PJ: Slot is occupied, but current tile is closer than the one in the slot
                //           We remove that tile from the slot and put it back onto the queue
                tiles.Add( slot.Tile );
                slot.Tile := nil;
            end;

            if slot.isEmpty then begin
                // INFO, PJ: Assign tile to slot and remove from queue
                slot.Tile := tile;
                slot.TileDistance := slot.CurrentDistance;
                tiles.Delete( 0 );
                break;
            end;
        end;

        FreeAndNil( slots );
    end;

    FreeAndNil( tiles );

    // INFO, PJ: HAck, for testing purposes only.
    for i := 0 to aTiles.Count - 1 do begin
        slot := aSlots[i] as TSlotInfo;
        slot.Tile := aTiles[i] as TTileInfo;
    end;
end;

// ----------------------------------------------------------------------------

// INFO, PJ: Helper method to sort slots.
function CompareSlotDistance( aSlot1, aSlot2: Pointer ): integer;
var distance1, distance2: integer;
begin
    distance1 := TSlotInfo( aSlot1 ).CurrentDistance;
    distance2 := TSlotInfo( aslot2 ).CurrentDistance;

    if distance1 < distance2 then
        result := -1
    else if distance1 > distance2 then
        result := 1
    else
        result := 0;
end;

function TTiler.SortedSlotsForTile( const aSlots: TObjectList; const aTile: TTileInfo ): TObjectList;
var i: integer;
    slot: TSlotInfo;
    tileCenter, slotCenter: TPoint;
begin
    result := TObjectList.Create( FALSE );
    result.Assign( aSlots );

    tileCenter.X := aTile.Left + aTile.Width div 2;
    tileCenter.Y := aTile.Top + aTile.Height div 2;

    // INFO, PJ: Assign square distances
    for i := 0 to result.Count - 1 do begin
        slot := result[i] as TSlotInfo;

        slotCenter.X := slot.Left + slot.Width div 2;
        slotCenter.Y := slot.Top + slot.Height div 2;

        slot.CurrentDistance := Sqr( tileCenter.X - slotCenter.X ) + Sqr( tileCenter.Y - slotCenter.Y );
    end;

    // INFO, PJ: Sort
    result.Sort( CompareSlotDistance );
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TTileInfo.Create( const aLeft, aTop, aWidth, aHeight: integer; const aTileObject: TObject );
begin
    inherited Create;

    fLeft := aLeft;
    fTop := aTop;
    fWidth := aWidth;
    fHeight := aHeight;

    fTileObject := aTileObject;
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TSlotInfo.Create( const aLeft, aTop, aWidth, aHeight: integer );
begin
    inherited Create;

    fLeft := aLeft;
    fTop := aTop;
    fWidth := aWidth;
    fHeight := aHeight;
end;

// ----------------------------------------------------------------------------

function TSlotInfo.IsEmpty: boolean;
begin
    result := not Assigned( fTile );
end;

end.
