{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_ToolExecuter;

{$MODE Delphi}

interface

uses
  unit_Types,
  unit_BaseTool,
  unit_ParamList,
  unit_ToolFactory,
  unit_Enums,
  unit_UndoRedoManager;

type
  TToolExecuter = class(TObject)
  private
    fUndoRedoManager: TUndoRedoManagerEx;
    fToolFactory: TToolFactory;
  public
    constructor Create(AUndoRedoManager: TUndoRedoManagerEx; AToolFactory: TToolFactory);
    procedure Execute(APrimitiveToolEnum: TPrimitiveToolEnum; AToolActionEnum: TToolAction; AParamList: TParamList); overload;
    procedure Execute(ASkeletonToolEnum: TSkeletonToolEnum; AToolActionEnum: TToolAction; AParamList: TParamList); overload;
    procedure Execute(AEditTool: TEditTool; AEditMode: TEditMode; AParamList: TParamList); overload;
    procedure Execute(var ABaseTool: TBaseTool; AParamList: TParamList); overload;
  end;

implementation

uses
  SysUtils, Dialogs, unit_PrimitiveTool, unit_Primitives;

// ### TToolExecuter ###########################################################

constructor TToolExecuter.Create(AUndoRedoManager: TUndoRedoManagerEx; AToolFactory: TToolFactory);
begin
    fUndoRedoManager := AUndoRedoManager;
    fToolFactory := AToolFactory;
end;

// NOTE: tools are always (!) executed by the ToolExecuter. Therefor, the
// ToolExecuter may free a tool if it is not registered with the Undo system.
// The Undo system owns the tools it keeps on its stacks so the tools are
// automatically freed when closing DeleD.

procedure TToolExecuter.Execute(var ABaseTool: TBaseTool; AParamList: TParamList);
var i: integer;
    primitiveTool: TPrimitiveTool;
    primitive: TBasePrimitive;
begin
    if not Assigned(ABaseTool) then
        // raise exception?
    else begin
        if not ABaseTool.CanExecute(AParamList) then begin
            if ABaseTool.ErrorMessage <> '' then
                MessageDlg(ABaseTool.ErrorMessage,mtError,[mbOK],0);
            FreeAndNil(ABaseTool);
        end else begin
            // Make sure everything is not highlighted anymore. Highlighted items
            // might get altered by the tool in question (because they are actually
            // selected), and trying to render them after the tool is finished
            // might lead to unexpected resutls (av's).
            // We do this here so we are sure this happens before any (!) tool
            // gets executed. Therefor, tools don't need to check on this.
            if (ABaseTool is TPrimitiveTool) then begin
                primitiveTool := ABaseTool as TPrimitiveTool;
                for i := 0 to primitiveTool.SelectedPrimitives.Count-1 do begin
                    primitive := primitiveTool.SelectedPrimitives[i];
                    primitive.HighlightedVertices.Clear;
                    primitive.HighlightedEdges.Clear;
                    primitive.HighlightedPolygons.Clear;
                end;
            end;

            // Execute the tool.
            ABaseTool.Execute(AParamList);
            if ABaseTool.ToolAction<>taNotApplicable then
                fUndoRedoManager.AddToolForUndo(ABaseTool)
            else
                FreeAndNil(ABaseTool);
        end;
    end;
end;

procedure TToolExecuter.Execute(APrimitiveToolEnum: TPrimitiveToolEnum; AToolActionEnum: TToolAction; AParamList: TParamList);
var baseTool: TBaseTool;
begin
    baseTool := fToolFactory.CreateTool(APrimitiveToolEnum, AToolActionEnum);
    Execute(baseTool,AParamList);
end;

procedure TToolExecuter.Execute(AEditTool: TEditTool; AEditMode: TEditMode; AParamList: TParamList);
var baseTool: TBaseTool;
begin
    baseTool := fToolFactory.CreateTool(AEditTool, AEditMode);
    Execute(baseTool,AParamList);
end;

procedure TToolExecuter.Execute(ASkeletonToolEnum: TSkeletonToolEnum; AToolActionEnum: TToolAction; AParamList: TParamList);
var baseTool: TBaseTool;
begin
    baseTool := fToolFactory.CreateTool(ASkeletonToolEnum, AToolActionEnum);
    Execute(baseTool,AParamList);
end;

end.

