{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_TrackBarControl;

{$MODE Delphi}

// Floating Point precision trackbar with onhint information

interface

uses
  StdCtrls, ComCtrls, Classes;

type
  TTrackBarType = ( tbLinear, tbExponential );
  TTrackBarControl = class( TObject )
  private
      fTrackbar: TTrackBar;
      fType: TTrackBarType;
      fMin: single;
      fMax: single;
      fPosition: single;
      fEnabled: Boolean;
      fUpdating: Boolean;
      fEdit: TEdit;
      fOriginalChangeHandler: TNotifyEvent;
      fOriginalEditChangeHandler: TNotifyEvent;
//      fOriginalExitHandler: TNotifyEvent;

      procedure HandleChange(Sender: TObject);
      procedure HandleEditChange(Sender: TObject);
//      procedure HandleExit(Sender: TObject);
      procedure Initialize;
      procedure UpdateEdit;
      procedure UpdateHint;
      procedure UpdateTrackbar;

      procedure SetPosition(const aPosition: single);
      procedure SetEdit(const anEdit: TEdit);
      procedure SetEnabled(const aFlag: Boolean);

  public
      constructor Create(
        const aTrackBar: TTrackBar;
        const aMin: single = 0;
        const aMax: single = 100 );

      property Edit: TEdit read fEdit write SetEdit; // Optional Linked Edit Control
      property Enabled: Boolean read fEnabled write SetEnabled;
      property Position: single read fPosition write SetPosition;
      property TrackBarType: TTrackBarType read fType write fType;
  end;

implementation

uses
  Graphics, SysUtils, Windows, CommCtrl, Forms;

constructor TTrackBarControl.Create(
  const aTrackBar: TTrackBar;
  const aMin: single = 0;
  const aMax: single = 100 );
begin
    inherited create;

    fTrackBar := aTrackBar;
    fTrackBar.Min := 0;
    fTrackbar.Max := 100;

    fMin := aMin;
    fMax := aMax;

    initialize;
end;

procedure TTrackBarControl.Initialize;
var style: DWord;
begin
    style := GetWindowLong(fTrackBar.Handle, GWL_STYLE);
    SetWindowLong(fTrackBar.Handle, GWL_STYLE, style and not TBS_ENABLESELRANGE);

    fOriginalChangeHandler := fTrackBar.OnChange;

    fTrackBar.OnChange := handleChange;
end;

//procedure TTrackBarControl.HandleExit(Sender: TObject);
//var index: Integer;
//begin
//    Val( fEdit.Text, fPosition, index );
//
//    UpdateEdit;
//    UpdateTrackbar;
//end;

procedure TTrackBarControl.SetPosition(const aPosition: single);
begin
    fPosition := aPosition;

    UpdateEdit;
    UpdateHint;
    UpdateTrackbar;
end;

procedure TTrackBarControl.UpdateEdit;
var format: String;
begin
    if Assigned( fEdit ) then begin
        fUpdating := TRUE;

        try
            if fType = tbExponential then begin
                if fPosition > 10 then
                    format := '0'
                else if fPosition > 1 then
                    format := '0.##'
                else
                    format := '0.###';
            end
            else
                format := '0.###';

            fEdit.Text := FormatFloat( format, fPosition );
        finally
            fUpdating := FALSE;
        end;
    end;
end;

procedure TTrackBarControl.UpdateHint;
var point: TPoint;
begin
    if not Assigned( fEdit ) then begin
        fTrackBar.Hint := FormatFloat( '0.##', fPosition );
        fTrackBar.ShowHint := TRUE;
        GetCursorPos(point);
        Application.ActivateHint(point) ;
    end;
end;

procedure TTrackBarControl.UpdateTrackbar;
var position: Single;
begin
    fUpdating := TRUE;

    try
        position := fPosition;

        // INFO, PJ: Correct position for min, max the trackbar can display.
        if position < fMin then position := fMin;
        if position > fMax then position := fMax;
        if (position = 0) and (fType = tbExponential) then position := 0.00000001;

        if fType = tbLinear then
            fTrackBar.Position := Round( (position - fMin) * 100 / (fMax - fMin))
        else
            fTrackBar.Position := Round( 50 * ln( position ) / ln (fMax ) + 50 );

    finally
        fUpdating := FALSE;
    end;
end;

procedure TTrackBarControl.SetEdit(const anEdit: TEdit);
begin
    if Assigned( fEdit ) then
    begin
//        fEdit.OnExit := fOriginalExitHandler;
        fEdit.OnChange := fOriginalEditChangeHandler;
    end;

    fEdit := anEdit;
//    fOriginalExitHandler := fEdit.OnExit;
    fOriginalEditChangeHandler := fEdit.OnChange;
    fEdit.OnChange := HandleEditChange;
//    fEdit.OnExit := HandleExit;
end;

procedure TTrackBarControl.SetEnabled(const aFlag: Boolean);
begin
    if Assigned( fEdit ) then begin
        if aFlag then
            fEdit.Color := clWindow
        else
            fEdit.Color := clBtnFace;

        fEdit.Enabled := aFlag;
    end;

    fTrackbar.Enabled := aFlag;
end;

procedure TTrackBarControl.HandleChange(Sender: TObject);
begin
    if not fUpdating then begin
        if fType = tbExponential then
            fPosition := exp( ( fTrackBar.Position - 50 ) * ln( fMax ) / 50 )
        else
            fPosition := fMin + (fTrackBar.Position / 100) * (fMax - fMin);

        UpdateEdit;
        UpdateHint;

        if Assigned( fOriginalChangeHandler ) then
            fOriginalChangeHandler( Sender );
    end;
end;

procedure TTrackBarControl.HandleEditChange(Sender: TObject);
var index: Integer;
begin
    if not fUpdating then begin
        Val( fEdit.Text, fPosition, index );

        UpdateTrackbar;

        if Assigned( fOriginalEditChangeHandler ) then
            fOriginalEditChangeHandler( Sender );
    end;
end;


end.
