{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_Types;

{$MODE Delphi}

interface

uses Classes,Math,Contnrs,Dialogs,SysUtils,Types,Graphics,Windows,
     unit_Enums,
     unit_IntegerList,
     unit_HashedList,
     unit_Textures,
     unit_Materials;

type
    TSkyboxSide = ( skyFront, skyBack, skyLeft, skyRight, skyUp, skyDown );

    TPolyPlane = ( ppYZ, ppXZ, ppXY );

    TFogMode = (fmLinear, fmExponential, fmExponential2);

    TOpenGLColor4 = array[0..3] of single;

    TDimensions = record
        Left, Top: integer;
        Width, Height: integer;
    end;

    TDCRC = record
        H_DC: HDC;
        H_RC: HGLRC;
    end;

  // Lightweight alternative for TVertice class (for convenience in mathematical applications)
    TVertex = record
        case boolean of
            false:  ( X, Y, Z: single );
            true:   ( Elements: Array[0..2] of single );
    end;
    TVertexArray = Array of TVertex;

  // Lightweight alternative for TUVCoord class (for convenience in mathematical application)
    TUVCoordinate = record
        case boolean of
            false:  ( u, v: single );
            true:   ( Elements: Array[0..1] of single );
    end;

  // Lightweight alternative for TMatrix class (for convenience and performance in mathematical applications)
  TMatrix4x4 = Array[0..3, 0..3] of single;
  TSymMatrix3x3 = Array[0..5] of single; // Symmetrical 3x3 matrix
  TVertex4 = Array[0..3] of single;

  // Data structure for generic raytracing applications (like 3D object picking)
  TRay = record
      Origin: TVertex;
      Destination: TVertex;
      Direction: TVertex; // Normalized (Destination-Origin), for efficiency purposes
  end;

  TViewPort = packed record
      Left, Top: Integer;
      Width, Height: Integer;
  end;

  // Data structure for the cohen-sutherland line-clipping algorithm.
  TClippingCode = (coClipRight, coClipLeft, coClipTop, coClipBottom, coClipFront, coClipBack);
  TClippingOutcode = Set of TClippingCode;

  TPoint2D = record                                                             // It's ok for any 2D point, not specifically vertices.
    X, Y: double;
  end;
  TPoint2DArray = array of TPoint2D;

  TEdgeList = class;                                                            // forward declarations (used for circular reference in datastructure)
  TPolygonList = class;

  TDeleDObjectList = class(TObjectList)                                         // main objectlist class; allmost all geometry lists are derived from this one
  private
    fUniquesOnly: boolean;
    fOnAdd: TNotifyEvent;
    fOnChange: TNotifyEvent;
  public
    constructor Create(AOwnsObjects, AUniquesOnly: boolean); overload;
    // methods
    procedure Clear; override;
    function Add(AObject: TObject): integer; overload;  // this Add method can make sure only unique pointers are stored in the list
    function Add(AObject: TObject; ARemoveOnExist: boolean): integer; overload;
    function AddList(list: TDeleDObjectList): integer;  // TODO: copyObjects: boolean; invertList: boolean = false);
    procedure RemoveList(list: TDeleDObjectList);
    function Remove(AObject: TObject): integer; overload;
    function Remove(AObject: TObject; AIgnoreOwnership: boolean): integer; overload;
    function RemoveWithoutChangeEvent(AObject: TObject): integer;
    function IsNotIn(otherList: TDeleDObjectList; resultList: TDeleDObjectList = nil): integer;
    function IsIn(otherList: TDeleDObjectList; resultList: TDeleDObjectList = nil): integer;
    function IsInList(AObject: TObject): boolean;
    function NumberOfOccurances(AObject: TObject): integer;
    // properties & evente
    property OnAdd: TNotifyEvent read fOnAdd write fOnAdd;
    property OnChange: TNotifyEvent read fOnChange write fOnChange;
    property UniquesOnly: boolean read fUniquesOnly write fUniquesOnly;         // TODO, JC: overwrite Assign method to include the UniqueOnly mechanism
  end;

  //forward declaration
  TVerticeList = class;
  TEdge = class;
  TPolygon = class;

  TVertice = class(TPersistent)
  private
    fEdgeList: TEdgeList;
    fTag: string;
    fColor: TColor;
    fOpenGLColor: TOpenGLColor4;  // INFO, PJ: Cache for quick rendering.
    fJointID: integer;
    function GetLength: double;
    function GetIsInfinite: boolean;
    function GetIsZero: boolean;
    function GetIsZeroOrOne: boolean;
    procedure SetColor( const aColor: TColor );
  public
    // xyz values
    x,y,z: double;
    // copies of xyz values, copioed when assigning primitives to skeletons and used by animation routines (e.g. JointLocationRotatorTool)
    ox,oy,oz: double;
    // constructor/destructor
    constructor Create; overload;
    constructor Create(vx,vy,vz: double); overload;
    destructor Destroy; override;
    // methods
    procedure Assign(Source: TPersistent); overload; override;
    procedure Assign(vx,vy,vz: double); reintroduce; overload;
    procedure Assign(Source: TVertex); reintroduce; overload;  
    function Copy: TVertice;
    procedure Normalize;
    procedure Reset;
    procedure ResetToInfinite;
    procedure Snap(ASnapValue: integer);
    procedure RoundValue;
    procedure SimpleRoundValue;
    procedure DivideByScalar(scalar: integer);
    procedure MultiplyByScalar(scalar: integer; doCheckForZero: boolean = true); overload;
    procedure MultiplyByScalar(scalar: double; doCheckForZero: boolean = true); overload;
    procedure MultiplyByVertex(vertex: TVertice);
    procedure Add(vertex: TVertice);
    procedure Subtract(vertex: TVertice);
    function Inverse: TVertice;
    function Equals(AVertex: TVertice; AnEpsilon: double = 0.001): boolean;
    function OnEdge(p1,p2: TVertice): boolean;
    function BetweenPoints(p1,p2: TVertice): boolean;
    function DistanceToLine(p1,p2: TVertice): double;
    function ProjectToScreen(const aModelViewMatrix, aProjectionMatrix: TMatrix4x4; const aViewport: TViewport): TPoint2D;
    function Interpolate(const aVertex: TVertice; const t: double): TVertice;
    function ApplyEditDimension(const aEditDimension: TEditDimension): TVertice;

    //Mesh topology methods
    procedure GetAdjacent(APolygonList: TPolygonList; ASelection: TPolygonList = nil); overload;
    procedure GetAdjacent(AVertexList: TVerticeList; ASelection: TVerticeList =  nil); overload;
    procedure GetAdjacent(AEdgeList: TEdgeList; ASelection: TEdgeList = nil); overload;
    procedure GetOrderedEdges(AOrderedEdges: TEdgeList; ANoOpenEdges: Boolean);
    function UsedByNumberOfEdges(AEdges: TEdgeList): Integer;
    function GetNormal(): TVertice;
    function GetEdge(AOtherVertice: TVertice): TEdge;

    // properties
    property Tag: string read fTag write fTag;
    property Color: TColor read fColor write SetColor;
    property OpenGLColor: TOpenGLColor4 read fOpenGLColor;
    property IsInfinite: boolean read GetIsInfinite;
    property IsZero: boolean read GetIsZero;
    property IsZeroOrOne: boolean read GetIsZeroOrOne;
    property EdgeList: TEdgeList read fEdgeList;                                // edgelist of edges that use this vertex
    property Length: double read GetLength;
    property JointID: integer read fJointID write fJointID;
  end;

  TVerticeList = class(TDeleDObjectList)
  private
    fCenter: TVertice;
    fNormal: TVertice;
    function GetCenter: TVertice;
    function GetNormal: TVertice;
    function GetVertice(anIndex: integer): TVertice;
    procedure SetVertice(anIndex: integer; vertex: TVertice);
  public
    constructor Create(AOwnsObjects, AUniquesOnly: boolean);
    destructor Destroy; override;
    // methods
    function InList(AVertex: TVertice): integer;
    function AddVertex(AVertex: TVertice; freeOnExist: boolean = false): TVertice;
    function IsStrictlyConvex: boolean;
    procedure InsertList(index: integer; list: TVerticeList);
    procedure Reverse;

    function GetNext(aVertex: TVertice): TVertice;
    function GetPrev(aVertex: TVertice): TVertice;
    function Copy: TVerticeList; // chrisu: OrthographicDistanceShading.


    //Mesh topology methods
    procedure GetAdjacent(APolygonList: TPolygonList; ASelection: TPolygonList = nil); overload;
    procedure GetAdjacent(AEdgeList: TEdgeList; ASelection: TEdgeList = nil); overload;

    // properties
    property Vertices[anIndex: integer]: TVertice read GetVertice write SetVertice; default;
    property Center: TVertice read GetCenter;
    property Normal: TVertice read GetNormal;
  end;

  TEdge = class(TObject)
  private
    fStartVertex: TVertice;
    fEndVertex: TVertice;
    fPolygonList: TPolygonList;
    fCenterVertex: TVertice;
    fDirectionVertor: TVertice;
    function GetCenterVertex: TVertice;
    function GetDirectionVertor: TVertice;
    function GetLength: double;
  public
    // constructor/destructor
    constructor Create(v0,v1: TVertice; polyList: TPolygonList = nil);
    destructor Destroy; override;
    // methods
    function Copy: TEdge;
    procedure Snap(ASnapValue: integer);

    //Mesh topology methods
    function GetSharedVertex(AOtherEdge: TEdge): TVertice;
    procedure GetAdjacent(APolygonList: TPolygonList; ASelection: TPolygonList = nil); overload;
    procedure GetAdjacent(AEdgeList: TEdgeList); overload;
    procedure GetAdjacentEdgesForEdgeLoop(edgeList: TEdgeList; useStartVertex: boolean);
    procedure GetOneAdjacentEdgeForEdgeRing(edgeList: TEdgeList; startLeftPolygon: boolean; polygon: TPolygon = nil);
    function IsUsingVertex(vertex: TVertice): boolean;
    function GetOtherVertex(vertex: TVertice): TVertice;
    function IsAdjacent(AEdge: TEdge): Boolean;

    // properties
    property StartVertex: TVertice read fStartVertex write fStartVertex;
    property EndVertex: TVertice read fEndVertex write fEndVertex;
    property CenterVertex: TVertice read GetCenterVertex;
    property DirectionVector: TVertice read GetDirectionVertor;
    property Length: double read GetLength;
    property PolygonList: TPolygonList read fPolygonList write fPolygonList;
  end;

  TEdgeList = class(TDeleDObjectList)
  private
    function GetEdge(anIndex: integer): TEdge;
  public
    procedure AddEdge(anEdge: TEdge);

    //Mesh topology methods
    procedure GetAdjacent(AVertexList: TVerticeList); overload;
    procedure GetAdjacent(APolygonList: TPolygonList); overload;
    function IsVertexUsed(AVertex: TVertice): Boolean;
    procedure GetStartEndVertices(AVertexList: TVerticeList);
    function isContinuousSet(): Boolean;
    function Copy: TEdgeList; // chrisu: OrthographicDistanceShading.

    property Edges[anIndex: integer]: TEdge read GetEdge; default;
  end;

  TUVCoord = class(TObject)
  private
    fU: double;
    fV: double;
    fSelected: boolean;
    function GetCoordinate: TUVCoordinate;
  public
    constructor Create(vu,vv: double);
    function Copy: TUVcoord;
    procedure Assign(uvcoord: TUVCoord);
    procedure Move(const aDelta: TUVCoordinate);
    procedure ApplyEditDimension(const anEditDimension: TEditUVDimension);
    property U: double read fU write fU;
    property V: double read fV write fV;
    property Coordinate: TUVCoordinate read GetCoordinate;
    property Selected: boolean read fSelected write fSelected;
  end;

  TUVList = class(TDeleDObjectList)
  private
    function GetUV(anIndex: integer): TUVCoord;
    function GetCenter: TUVCoordinate;
  public
    procedure Assign(const aList: TUVList);
    function Add(const aU, aV: double): integer; overload;
    procedure Insert(const anIndex: integer; const aU, aV: double); overload;
    function Copy: TUVList; overload;
    function Copy(index: integer): TUVCoord; overload;
    function InList(uvcoord: TUVCoord): integer;
    procedure Reverse;
    function Delete(index: integer): boolean;

    property Center: TUVCoordinate read GetCenter;
    property UVCoords[anIndex: integer]: TUVCoord read GetUV; default;
  end;

  TPlane = class(TObject)       // TODO: create an Assign method for TPlane
  private
    fNormal: TVertice;
    fDistance: double;
  public
    constructor Create; overload;
    constructor Create(v,n: TVertice); overload;
    constructor Create(v0,v1,v2: TVertice); overload;
    destructor Destroy; override;
    property Distance: double read fDistance write fDistance;
    function Equals(plane: TPlane): boolean;
    property Normal: TVertice read fNormal write fNormal;
  end;

  TPolygon = class(TObject)
  private
    fNormal: TVertice;
    fMaterial: TMaterial;
    fVerticeList: TVerticeList;
    fHighlightedUVCoordinates: TUVList;
    fUVCoords: TObjectList;
    fEdgeList: TEdgeList;
    fBoundingBoxCenter: TVertice;
    fAllowLightmap: boolean;
    fVertexNormals: TObjectList;
    fCenter: TVertice;
    fTag: string;
    //fIsPathPolygon: Boolean;
    function Area: double;
    function GetNumVertices: integer;
    function GetNumHighlightedUVCoordinates: integer;
    function GetIsClockWise: boolean;
    function GetNormal: TVertice;
    function GetDistanceToOrigin: double;
    function GetPlane: TPolyPlane;
    function GetUVCoordsForIndex( const anIndex: Integer ): TUVList;
    procedure SetUVCoordsForIndex(const anIndex: Integer; const aUVCoords: TUVList);
    function GetPlaneDistance: double;
    function GetBoundingBoxCenter: TVertice;
    function GetVertexNormalForIndex(const anIndex: Integer): TVertice;
    function GetCenter: TVertice;
  public
    // constructor/destructor
    constructor Create;
    destructor Destroy; override;

    // methods
    function Copy: TPolygon;
    function HasUVCoordinates: boolean;
    function HasVertexNormals: boolean;
    procedure UpdateUVCoordinates(aFitToFace: boolean = false);
    procedure AddCopiedUVCoordinates(const aPoly: TPolygon; const anIndex: Integer);
    procedure AddUVCoordinatesForVertex(const aVertex: TVertice);
    procedure AddEmptyUVCoordinatesForVertex(const aVertex: TVertice);
    procedure InsertUVCoordinatesForVertex(const anIndex: integer; const aVertex: TVertice);
    procedure DeleteUVCoordinates(const anIndex: integer);
    procedure AdjustUVCoordinatesToMaterial;
    procedure Flip;
    procedure Snap(aSnapValue: integer);
    procedure ClearVertexNormals;
    procedure AddVertexNormal(const aNormal: TVertice);
    procedure Assign(polygon: TPolygon);
    function IsPointInTriangle(p,a,b,c: TVertice; EPS: double = 0): boolean;
    function IsPointInPolygon(aVertex: TVertice; EPS: double; ResetComponents: boolean = true): boolean;
    function IsUVCoordinateInPolygon(const aUVCoordinate: TUVCoordinate): boolean;
    function IsPointInFront(aVertex: TVertice): boolean;
    function IsIntersectedByRay(const aRay: TRay): boolean; overload;
    function IsIntersectedByRay(const aRay: TRay; out aDistance: single): boolean; overload;
    function IsCoplanar(aPlane: TPlane): boolean; overload;
    function IsCoplanar(aPolygon: TPolygon): boolean; overload;
    function IsEqualNormal(aPolygon: TPolygon): boolean;
    function IsStrictlyConvex: boolean;
    function HasIdenticalVertexNormals: boolean;
    procedure ProjectToScreen(const aModelViewMatrix, aProjectionMatrix: TMatrix4x4; const aViewport: TViewport;
                              var aPolygonPoints: TPoint2DArray);

    //mesh topology methods
    function GetPreviousEdge(AStartVertex: TVertice): TEdge;
    function GetNextEdge(AStartVertex: TVertice): TEdge;
    procedure GetAdjacent(APolygonList: TPolygonList; ASelection: TPolygonList = nil; MustBeCoPlanar: Boolean = false); overload;
    procedure GetAdjacent(AEdgeList: TEdgeList); overload;
    procedure GetAdjacent(AVertexList: TVerticeList; ASelection: TVerticeList = nil; UsedOnlyByThisPolygon: Boolean = false); overload;
    procedure GetSharedEdges(AOtherPolygon: TPolygon; AEdgeList: TEdgeList; clearlist: boolean);
    procedure GetSharedVertices(AOtherPolygon: TPolygon; AVertexList: TVerticeList; clearlist: boolean);

    // properties
    property NumVertices: integer read GetNumVertices;
    property NumHighlightedUVCoordinates: integer read GetNumHighlightedUVCoordinates;
    property IsClockWise: boolean read GetIsClockWise;
    property Normal: TVertice read GetNormal;
    property DistanceToOrigin: double read GetDistanceToOrigin;
    property Plane: TPolyPlane read GetPlane;
    property PlaneDistance: double read GetPlaneDistance;
    property Material: TMaterial read fMaterial write fMaterial;
    property BoundingBoxCenter: TVertice read GetBoundingBoxCenter;
    property Center: TVertice read GetCenter;
    property Tag: string read fTag write fTag;
    property AllowLightmap: boolean read fAllowLightmap write fAllowLightmap;
    property VerticeList: TVerticeList read fVerticeList write fVerticeList;
    property HighlightedUVCoordinates: TUVList read fHighlightedUVCoordinates write fHighlightedUVCoordinates;
    property VertexNormals[const anIndex: Integer]: TVertice read GetVertexNormalForIndex;
    property UVCoordinates[const anIndex: Integer]: TUVList read GetUVCoordsForIndex write SetUVCoordsForIndex;
    property EdgeList: TEdgeList read fEdgeList write fEdgeList;
    //property IsPathPolygon: Boolean read fIsPathPolygon write fIsPathPolygon;
  end;

  TPolygonList = class(TDeleDObjectList)
  private
    function GetPolygon(anIndex: integer): TPolygon;
  public
    //Mesh topology methods
    function GetAveragedNormal(): TVertice;
    procedure GetAdjacent(AVerticeList: TVerticeList); overload;
    procedure GetAdjacent(AEdgeList: TEdgeList); overload;
    function Copy: TPolygonList; // chrisu: OrthographicDistanceShading.

    property Polygons[anIndex: integer]: TPolygon read GetPolygon; default;
  end;

  TLookupTableEntry = record
    U, V: single;
    diffuse: integer;
    specular: integer;
  end;
  PLookupTableEntry = ^TLookupTableEntry;
  TLookupTable = array of TLookupTableEntry;

  THugeByteArray = array[0..MaxInt-1] of Byte;

  TFontQuality = (fqLow, fqMedium, fqHigh, fqPerfect);
  TFontInfo = record
      Name: String;
      Width, Height: Integer;
      Style: TFontStyles;
      Extrusion: Integer;
      Quality: TFontQuality;
  end;

  TTextDimensions = record
      Width: Single;
      PositiveHeight: Single;     // Height above baseline
      NegativeHeight: Single;     // Height below baseline
  end;

  TTextExtent = record
      Offset: Integer;
      Width, Height: Integer;
  end;

  TGlyphMetricsArray = array[0..255] of TGlyphMetricsFloat;

implementation

uses unit_Primitives,
     unit_Math,
     unit_Constants,
     unit_Functions,
     unit_Variables;

// ### TDeleDObjectList ########################################################

function TDeleDObjectList.Add(AObject: TObject): integer;
begin
    Result := Add(AObject, false);  // Call the method below.
end;

// if UniquesOnly is true, the list can only contain unique pointers;
// used heavily in the geometry routines
function TDeleDObjectList.Add(AObject: TObject; ARemoveOnExist: boolean): integer;
var index: integer;
begin
    if UniquesOnly then begin
        index := IndexOf(AObject);
        if index=-1 then begin          // object isn't stored yet
            Result := inherited Add(AObject);
            if Assigned(fOnAdd) then
                fOnAdd(AObject);
        end else begin
            if ARemoveOnExist then begin    // object already stored in list, remove if wanted
                Remove(AObject);
                Result := -1;
            end else
                Result := index;
        end;
    end else begin
        if ARemoveOnExist then begin    // if it already exists, remove it
            Remove(AObject);
            Result := -1;
        end else begin                  // store
            Result := inherited Add(AObject);
            if Assigned(fOnAdd) then
                fOnAdd(AObject);
        end;
    end;
    if Assigned(fOnChange) then
        fOnChange(nil);
end;

// NOTE: what happens if the provided list owns its objects and the current
// list does too?
// NOTE: returns the # of items actually added to the list ---> TODO: implement!
function TDeleDObjectList.AddList(list: TDeleDObjectList): integer;
var i,counter: integer;
begin
    counter := 0;
    if Assigned(list) then begin
        for i := 0 to list.Count-1 do begin
            if IndexOf(list[i])=-1 then
                inc(counter);       // sort of a hack: we know the item will be added to the list
            Add(list[i]);
        end;
    end;
    Result := counter;
end;

procedure TDeleDObjectList.RemoveList(list: TDeleDObjectList);
var i: integer;
begin
    if Assigned(list) then begin
        for i := 0 to list.Count-1 do
            Remove(list[i]);
    end;
end;

function TDeleDObjectList.Remove(AObject: TObject): integer;
var index: integer;
begin
    index := inherited Remove(AObject);
    if Assigned(fOnChange) then
        fOnChange(nil);
    Result := index;
end;

function TDeleDObjectList.Remove(AObject: TObject; AIgnoreOwnership: boolean): integer;
var doesOwnObjects: boolean;
begin
    doesOwnObjects := OwnsObjects;

    if AIgnoreOwnership then
        OwnsObjects := false;

    Result := Remove(AObject);

    if AIgnoreOwnership then
        OwnsObjects := doesOwnObjects;
end;

function TDeleDObjectList.RemoveWithoutChangeEvent(AObject: TObject): integer;
var index: integer;
begin
    index := inherited Remove(AObject);
    Result := index;
end;

// copies the items of the current list that do NOT occur in the "otherList" to the provided resultList (if available)
// the function returns the number of items in that list (0 if no items where saved to the result)
function TDeleDObjectList.IsNotIn(otherList: TDeleDObjectList; resultList: TDeleDObjectList = nil): integer;
var i,counter: integer;
begin
    counter := 0;

    if Assigned(resultList) then
        resultList.Clear;

    for i := 0 to Count-1 do begin

        if otherList.IndexOf(Items[i])=-1 then begin        // item not found
            if Assigned(resultList) then
                resultList.Add(Items[i]);                   // put in resultlist
            inc(counter);
        end;

    end;
    Result := counter;
end;

// counts how many items of the list are also in the other list and returns them in a result list
function TDeleDObjectList.IsIn(otherList: TDeleDObjectList; resultList: TDeleDObjectList = nil): integer;
var i,counter: integer;
begin
    counter := 0;

    if Assigned(resultList) then
        resultList.Clear;

    for i := 0 to Count-1 do begin

        if otherList.IndexOf(Items[i])<>-1 then begin        // item found
            if Assigned(resultList) then
                resultList.Add(Items[i]);                   // put in resultlist
            inc(counter);
        end;

    end;
    Result := counter;
end;

function TDeleDObjectList.IsInList(AObject: TObject): boolean;
begin
    Result := IndexOf(AObject)<>-1;
end;

function TDeleDObjectList.NumberOfOccurances(AObject: TObject): integer;
var i,counter: integer;
begin
    counter := 0;
    for i := 0 to Count-1 do begin
        if Items[i]=AObject then
            inc(counter);
    end;
    Result := counter;
end;

constructor TDeleDObjectList.Create(AOwnsObjects, AUniquesOnly: boolean);
begin
    inherited Create(AOwnsObjects); // JuMa: call to inherited constructor was missing
    UniquesOnly := AUniquesOnly;
end;

procedure TDeleDObjectList.Clear;
begin
    inherited Clear;
    if Assigned(fOnChange) then
        fOnChange(nil);
end;

// ### TVertice ################################################################

constructor TVertice.Create;
begin
    inherited Create;

    x := 0;
    y := 0;
    z := 0;
    Color := TColor($FFFFFFFF);
    fEdgeList := TEdgeList.Create(false,true);
    fJointID := -1;
end;

constructor TVertice.Create(vx, vy, vz: double);
begin
    Create;

    x := vx;
    y := vy;
    z := vz;
end;

destructor TVertice.Destroy;
begin
    FreeAndNil(fEdgeList);
    inherited;
end;

procedure TVertice.Assign(Source: TVertex);
begin
    X := Source.X;
    Y := Source.Y;
    Z := Source.Z;
end;

function TVertice.GetLength: Double;
begin
    Result := sqrt(sqr(x) + sqr(y) + sqr(z));
end;

//Get's the edge inbetween this vertex and AOtherVertice
//If none found, NIL is returned
function TVertice.GetEdge(AOtherVertice: TVertice): TEdge;
var i: integer;
    edge: TEdge;
begin
    // loop through the edgelist of this vertex looking for an
    // edge that consists of this vertex and AOtherVertice; if found, return that edge
    // note: there is no need to check the edgelist of the provided AOtherVertice
    Result := nil;
    for i := 0 to EdgeList.Count-1 do
    begin
      edge := EdgeList[i];

      // see if the provided edge consists of SV and EV
      if (edge.StartVertex = AOtherVertice) or
         (edge.EndVertex = AOtherVertice) then
      begin
          Result := edge;
          exit;
      end;
    end;
end;

//Puts all vertices that are adjaent to this vertex in AVertexList
//If Aselection is assigned, only vertices in aSelection are put in AVertexList
//!!!!!!!! WARNING: AVertexList does NOT include the vertex itsself
procedure TVertice.GetAdjacent(AVertexList, ASelection: TVerticeList);
var i: integer;
    edge: TEdge;
begin
    if not Assigned(AVertexList) then Exit;

    for i := 0 to EdgeList.Count-1 do
    begin
        edge := EdgeList[i];

        if edge.StartVertex <> Self then
        begin
          //Add startvertex of edge
          if Assigned(ASelection) then
          begin
            if ASelection.IndexOf(edge.StartVertex) <> -1 then
              AVertexList.Add(edge.StartVertex);
          end
          else
            AVertexList.Add(edge.StartVertex);
        end
        else
        begin
          //Add endvertex of edge
          if Assigned(ASelection) then
          begin
            if ASelection.IndexOf(edge.EndVertex) <> -1 then
              AVertexList.Add(edge.EndVertex);
          end
          else
            AVertexList.Add(edge.EndVertex);
        end;
    end;
end;

//Puts all polygons that are adjaent to this vertex in APolygonList
//If Aselection is assigned, only polygons in aSelection are put in APolygonList
procedure TVertice.GetAdjacent(APolygonList, ASelection: TPolygonList);
var i,j: integer;
    edge: TEdge;
begin
    if Assigned(APolygonList) then begin

        APolygonList.Clear;

        for i := 0 to EdgeList.Count-1 do begin
            edge := EdgeList[i];
            for j := 0 to edge.PolygonList.Count-1 do begin

                if APolygonList.IndexOf(edge.PolygonList[j])=-1 then begin
                    if not Assigned(ASelection) then
                        APolygonList.Add(edge.PolygonList[j])
                    else begin
                        if ASelection.IndexOf(edge.PolygonList[j])<>-1 then
                            APolygonList.Add(edge.PolygonList[j]);
                    end;
                end;

            end;
        end;
    end;
end;


//Puts all edges that are adjacent to this vertex in AEdgeList
//When ASelection is assigned, only vertices that are also inside ASelection
//will be added to AEdgeList
procedure TVertice.GetAdjacent(AEdgeList, ASelection: TEdgeList);
var
  I: Integer;
begin
  if Assigned(ASelection) then
    for I := 0 to fEdgeList.Count - 1 do
      if ASelection.IndexOf(fEdgeList[I]) <> -1 then
        AEdgeList.Add(fEdgeList[I])
  else
    AEdgeList.AddList(fEdgeList)
end;

//Computes the vertex normal
function TVertice.GetNormal: TVertice;
var
  I:           Integer;
  polygonList: TPolygonList;
begin
    Result := TVertice.Create(0,0,0);
    polygonList := TPolygonList.Create(false,true);
    GetAdjacent(polygonList);
    for i := 0 to polygonList.Count-1 do
      Result.Add(PolygonList[i].Normal);
    Result.DivideByScalar(polygonList.Count);
    Result.Normalize;
    FreeAndNil(polygonList);
end;

// returns an ordered edgelist for this vertex
// The edges may only be adjacent to 1 or 2 polygons.
// However, if ANoOpenEdges is true, no "open" edges are allowed
//(so all must be adjacent to 2 polygons)
procedure TVertice.GetOrderedEdges(AOrderedEdges: TEdgeList; ANoOpenEdges: Boolean);
var currentEdge,nextEdge: TEdge;
    currentPolygon: TPolygon;
    vertexEdges: TEdgeList;
    i,counter,index: integer;
begin
    // let's order some edges!
    if Assigned(AOrderedEdges) then
    begin
        AOrderedEdges.Clear;

        // each edge should be shared by only 1 or 2 polygons; we quit if this is not the case;
        // the calling routine can check orderedEdges to see if the ordering was succesfull
        for i := 0 to EdgeList.Count-1 do begin
            if ANoOpenEdges then begin
                if EdgeList[i].PolygonList.Count<>2 then exit;   // used by Doo-Sabin smoothing
            end else begin
                if EdgeList[i].PolygonList.Count>2 then exit;    // used by Vertex Bevel
            end;
        end;

        // if the vertex has 3 edges, ordering them doesn't really make sense
        if EdgeList.Count=3 then begin
            AOrderedEdges.Assign(EdgeList);
            exit;
        end;

        // copy the original vertex.edgelist into a templist
        vertexEdges := TEdgeList.Create(false,true);
        vertexEdges.Assign(EdgeList);

        // start with the very first edge and polygon
        currentEdge := vertexEdges[0];
        currentPolygon := currentEdge.PolygonList[0];
        vertexEdges.Remove(currentEdge);
        AOrderedEdges.Add(currentEdge);

        while vertexEdges.Count>0 do begin

            // currentEdge is used by currentPolygon; now find another edge that
            // is also used by currentPolygon (we search the index of that edge here)
            counter := 0;
            index := currentPolygon.EdgeList.IndexOf(vertexEdges[counter]);
            while (index=-1) and (counter<vertexEdges.Count) do begin
                inc(counter);
                index := currentPolygon.EdgeList.IndexOf(vertexEdges[counter])
            end;

            // if we found an edge...
            if index<>-1 then begin

                nextEdge := vertexEdges[counter];

                // the polygonlist of the next edge should have 2 polygons;
                // one of those is equal to the current polygon (because we know
                // current polygon uses this edge); find the other (!) polygon
                if nextEdge.PolygonList[0]<>currentPolygon then
                    currentPolygon := nextEdge.PolygonList[0]
                else
                    currentPolygon := nextEdge.PolygonList[1];

                // store in ordered list and remove from the templist
                AOrderedEdges.Add(nextEdge);
                vertexEdges.Remove(nextEdge);
            end;
        end;

        FreeAndNil(vertexEdges);
    end;
end;

//Returns the number of edges that use this vertex
function TVertice.UsedByNumberOfEdges(AEdges: TEdgeList): Integer;
var
  I: Integer;
begin
    Result := 0;
    if Assigned(AEdges) then
      for I := 0 to AEdges.Count-1 do
        if AEdges[i].IsUsingVertex(self) then
          inc(Result);
end;

function TVertice.GetIsInfinite: boolean;
begin
    Result := (x=infinite) and (y=infinite) and (z=infinite);
end;

function TVertice.GetIsZero: boolean;
begin
    Result := (x=0) and (y=0) and (z=0);
end;

function TVertice.GetIsZeroOrOne: boolean;
begin
    Result := ((x=0) or (x=1)) and
              ((y=0) or (y=1)) and
              ((z=0) or (z=1));
end;

procedure TVertice.SetColor( const aColor: TColor );
begin
    fColor := aColor;
    fOpenGLColor := ColorToOpenGL( aColor );
end;

procedure TVertice.Assign(Source: TPersistent);
begin
    if Source is TVertice then begin
        x := TVertice(Source).x;
        y := TVertice(Source).y;
        z := TVertice(Source).z;
        Color := TVertice(Source).Color;
        // this needs to be replaced by a set of joints someday (for weighted joints)
        JointID := TVertice(Source).JointID;
    end else
        inherited;
end;

procedure TVertice.Assign(vx,vy,vz: double);
begin
    x := vx;
    y := vy;
    z := vz;
end;

function TVertice.Copy: TVertice;
begin
    Result := TVertice.Create;
    Result.Assign(self);
end;

procedure TVertice.Add(vertex: TVertice);
begin
    x := x+vertex.x;
    y := y+vertex.y;
    z := z+vertex.z;
end;

procedure TVertice.Subtract(vertex: TVertice);
begin
    x := x-vertex.x;
    y := y-vertex.y;
    z := z-vertex.z;
end;

function TVertice.Inverse: TVertice;
begin
    Result := TVertice.Create(-x,-y,-z);
end;

procedure TVertice.Normalize;
var length: double;
begin
    length := sqrt(sqr(x) + sqr(y) + sqr(z));
    if length=0 then length := 1;
    x := x/length;
    y := y/length;
    z := z/length;
end;

procedure TVertice.Reset;
begin
    x := 0;
    y := 0;
    z := 0;
end;

procedure TVertice.ResetToInfinite;
begin
    x := infinite;
    y := infinite;
    z := infinite;
end;

procedure TVertice.Snap(ASnapValue: integer); // chrisu: Made snapping to consider dimension-locks.
begin
    if editDimension = edX then
      x := round(x / ASnapValue) * ASnapValue
    else if editDimension = edY then
      y := round(y / ASnapValue) * ASnapValue
    else if editDimension = edZ then
      z := round(z / ASnapValue) * ASnapValue
    else begin
      x := round(x / ASnapValue) * ASnapValue;
      y := round(y / ASnapValue) * ASnapValue;
      z := round(z / ASnapValue) * ASnapValue;
    end;
end;

procedure TVertice.RoundValue;
begin
    x := Round(x);
    y := Round(y);
    z := Round(z);
end;

procedure TVertice.SimpleRoundValue;
begin
    x := SimpleRoundTo(x,-2);
    y := SimpleRoundTo(y,-2);
    z := SimpleRoundTo(z,-2);
end;

procedure TVertice.DivideByScalar(scalar: integer);
begin
    if scalar<>0 then begin
        x := x / scalar;
        y := y / scalar;
        z := z / scalar;
    end;
end;

procedure TVertice.MultiplyByScalar(scalar: integer; doCheckForZero: boolean = true);
begin
    if (doCheckForZero) and (scalar=0) then
        exit
    else begin
        x := x * scalar;
        y := y * scalar;
        z := z * scalar;
    end;
end;

procedure TVertice.MultiplyByScalar(scalar: double; doCheckForZero: boolean = true);
begin
    if (doCheckForZero) and (scalar=0) then
        exit
    else begin
        x := x * scalar;
        y := y * scalar;
        z := z * scalar;
    end;
end;

procedure TVertice.MultiplyByVertex(vertex: TVertice);
begin
    if Assigned(vertex) then begin
        x := x * vertex.x;
        y := y * vertex.y;
        z := z * vertex.z;
    end;
end;

function TVertice.Equals(AVertex: TVertice; AnEpsilon: double = 0.001): boolean;
begin
    Result := (Abs(x-AVertex.x)<AnEpsilon) and
              (Abs(y-AVertex.y)<AnEpsilon) and
              (Abs(z-AVertex.z)<AnEpsilon);
end;

// returns wether the vertex lies on the edge p1->p2
function TVertice.OnEdge(p1,p2: TVertice): boolean;
const EPS=0.001;
var qp1,p2p1: TVertice;
    lengthP2P1,dp: double;
begin
    Result := false;
    if DistanceToLine(p1,p2)<EPS then begin
        // je moet nog steeds bepalen of Q ook echt tussen P1 en P2 in ligt
        // Dat is het geval als de de lengte van de projectie van de lijn p1 -> Q op de edge
        // kleiner is dan de lengte van de edge zelf, dus
        // t = ( Q - P1 ) * ( P2 - P1 ) / || P2 - P1 ||   <    || P2 - P1 || - EPS

        qp1 := vectorsub(self,p1);

        p2p1 := vectorsub(p2,p1);

        lengthP2P1 := p2p1.Length;

        dp := dotproduct(qp1,p2p1)/lengthP2P1;

        if dp>EPS then
            Result := dp < lengthP2P1-EPS
        else
            Result := false;

        FreeAndNil(qp1);
        FreeAndNil(p2p1);
    end;
end;

function TVertice.BetweenPoints(p1,p2: TVertice): boolean;
const EPS=0.001;
var qp1,p2p1: TVertice;
    lengthP2P1,dp: double;
begin
// je moet nog steeds bepalen of Q ook echt tussen P1 en P2 in ligt
// Dat is het geval als de de lengte van de projectie van de lijn p1 -> Q op de edge
// kleiner is dan de lengte van de edge zelf, dus
// t = ( Q - P1 ) * ( P2 - P1 ) / || P2 - P1 ||   <    || P2 - P1 || - EPS

    qp1 := vectorsub(self,p1);

    p2p1 := vectorsub(p2,p1);

    lengthP2P1 := p2p1.Length;

    dp := dotproduct(qp1,p2p1)/lengthP2P1;

    if dp>EPS then
        Result := dp < lengthP2P1-EPS
    else
        Result := false;

    FreeAndNil(qp1);
    FreeAndNil(p2p1);
end;

function TVertice.DistanceToLine(p1,p2: TVertice): double;
var qp1,p2p1: TVertice;
begin
// Gegeven een edge, gedefinieerd door de eindpunten P1 en P2, willen we van een vertex Q
// weten of hij op een afstand < e ligt van die edge. De squared distance van een punt Q
// tot de lijn die loopt door P1 en P2 is gedefinieerd als
// d² = ( Q - P1 )² - (( Q - P1 ) * ( P2 - P1 ))² / ( P2 - P1 )²
// met * is dotproduct

    qp1 := vectorSub(self,p1);
    p2p1 := vectorSub(p2,p1);

    Result := dotproduct(qp1,qp1) - sqr(dotproduct(qp1,p2p1)) / dotProduct(p2p1,p2p1);

    // Suppress small rounding errors;
    if result < 0 then result := 0;

    Result := sqrt(Result);

    FreeAndNil(qp1);
    FreeAndNil(p2p1);
end;


function TVertice.ProjectToScreen(const aModelViewMatrix, aProjectionMatrix: TMatrix4x4; const aViewport: TViewport): TPoint2D;
var vertex: TVertex4;
    inverse: single;
begin
    vertex[0] := X;
    vertex[1] := Y;
    vertex[2] := Z;
    vertex[3] := 1;

    vertex := matrix4x4Multiply(aModelViewMatrix, vertex);
    vertex := matrix4x4Multiply(aProjectionMatrix, vertex);

    if vertex[3] <> 0 then
    begin
        inverse := 1 / vertex[3];

        vertex[0] := vertex[0] * inverse;
        vertex[1] := vertex[1] * inverse;
        vertex[2] := vertex[2] * inverse;
    end
    else begin
        // Degenerate matrix - it happens
        vertex[0] := 0;
        vertex[1] := 0;
        vertex[2] := 0;
    end;

    vertex[0] := vertex[0] * 0.5 + 0.5;
    vertex[1] := vertex[1] * 0.5 + 0.5;

    result.X := aViewport.Left + vertex[0] * aViewport.Width;
    result.Y := aViewport.Top + vertex[1] * aViewport.Height;

    // And convert from "view" to "screen" coordinates
    result.Y := aViewport.Height - result.Y - 0.5;
end;

// t=0..1
function TVertice.Interpolate(const aVertex: TVertice; const t: double): TVertice;
var vertex: TVertice;
begin
    vertex := vectorSub(aVertex,self);
    vertex.x := (t*vertex.x);
    vertex.y := (t*vertex.y);
    vertex.z := (t*vertex.z);
    vertex.Add(self);
    Result := vertex;
end;

function TVertice.ApplyEditDimension(const aEditDimension: TEditDimension): TVertice;
begin
    case editDimension of
        edX:
            begin
                y := 0;
                z := 0;
            end;
        edY:
            begin
                x := 0;
                z := 0;
            end;
        edZ:
            begin
                x := 0;
                y := 0;
            end;
    end;
    Result := self;
end;

// ### TVerticeList ############################################################

function TVerticeList.GetCenter: TVertice;
var i: integer;
begin
    fCenter.x := 0;
    fCenter.y := 0;
    fCenter.z := 0;
    for i := 0 to Count-1 do
        fCenter.Add(Get(i));
    fCenter.DivideByScalar(Count);
    Result := fCenter;
end;

// copy from TPolygon.GetNormal
function TVerticeList.GetNormal: TVertice;
const EPS=0.001;
var averageYZ, averageZX, averageXY: double;
    index: integer;
    vertex, previousVertice: TVertice;
    length: double;
begin
    averageYZ := 0;
    averageZX := 0;
    averageXY := 0;

    previousVertice := Vertices[Count-1];
    for index := 0 to Count-1 do begin
        vertex := Vertices[index];
        averageYZ := averageYZ + (previousVertice.Z + vertex.Z) * (vertex.Y - previousVertice.Y);
        averageZX := averageZX + (previousVertice.X + vertex.X) * (vertex.Z - previousVertice.Z);
        averageXY := averageXY + (previousVertice.Y + vertex.Y) * (vertex.X - previousVertice.X);
        previousVertice := vertex;
    end;

    length := Sqrt(averageYZ * averageYZ + averageZX * averageZX + averageXY * averageXY);

    if length < EPS then length := EPS;
    fNormal.X := averageYZ / length;
    fNormal.Y := averageZX / length;
    fNormal.Z := averageXY / length;

    Result := fNormal;
end;

//Get next vertex of list
function TVerticeList.GetNext(aVertex: TVertice): TVertice;
var
  I: Integer;
begin
  Result := nil;
  I := IndexOf(aVertex);
  if I = -1 then Exit;
  I := (I + 1) mod Count;
  Result := Vertices[I];
end;

//Get previous vertex of list
function TVerticeList.GetPrev(aVertex: TVertice): TVertice;
var
  I: Integer;
begin
  Result := nil;
  I := IndexOf(aVertex);
  if I = -1 then Exit;

  if I > 0 then I := (I - 1) mod Count
  else          I := Count -1;

  Result := Vertices[I];
end;

function TVerticeList.GetVertice(anIndex: integer): TVertice;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TVertice(Items[anIndex])
    else
        Result := nil;
end;

procedure TVerticeList.SetVertice(anIndex: integer; vertex: TVertice);
begin
    if (anIndex>=0) and (anIndex<Count) then
        Items[anIndex] := vertex;
end;

constructor TVerticeList.Create(AOwnsObjects, AUniquesOnly: boolean);
begin
    inherited Create(AOwnsObjects,AUniquesOnly);
    fCenter := TVertice.Create(0,0,0);
    fNormal := TVertice.Create(0,0,0);
end;

destructor TVerticeList.Destroy;
begin
    FreeAndNil(fNormal);
    FreeAndNil(fCenter);
    inherited;
end;

//Retrieves the list of polygons that are adjacent to these vertices.
//If ASelection is not nil, only polygons in ASelection will be put on the APolygon List
procedure TVerticeList.GetAdjacent(APolygonList, ASelection: TPolygonList);
var
  I: integer;
  TempPolygons: TPolygonList;
begin
    if Assigned(APolygonList) then begin

        APolygonList.Clear;

        tempPolygons := TPolygonList.Create(false,true);

        for i := 0 to Count-1 do begin
            tempPolygons.Clear;

            Vertices[I].GetAdjacent(tempPolygons, ASelection);
            APolygonList.AddList(tempPolygons);
        end;

        FreeAndNil(tempPolygons);
    end;
end;

//Returns list of edges that are adjacent to this vertex-list.
//If ASelection is assigned, only edges that are in ASelection are returned
procedure TVerticeList.GetAdjacent(AEdgeList, ASelection: TEdgeList);
var
  I: integer;
  J: Integer;
begin
    if Assigned(AEdgeList) then
    begin
        AEdgeList.Clear;
        if not Assigned(ASelection) then
        begin
          for I := 0 to Count-1 do
            AEdgeList.AddList(Vertices[i].EdgeList);
        end
        else
          for I := 0 to Count-1 do
            for J := 0 to Vertices[i].EdgeList.Count - 1 do
              if ASelection.IndexOf(Vertices[i].EdgeList[J]) <> -1 then
                AEdgeList.Add(Vertices[i].EdgeList[J]);
    end;
end;


// returns the index of the first vertex that has the same x,y,z values as the provided vertex
function TVerticeList.InList(AVertex: TVertice): integer;
var i: integer;
    vertex: TVertice;
begin
    Result := -1;
    for i := 0 to Count-1 do begin
        vertex := GetVertice(i);
        if vertex.Equals(AVertex) then begin
            Result := i;
            break;
        end;
    end;
end;

// if a simular vertex doesn't exist in the list yet (based on x,y,z values!),
// the provided vertex is added; if a simular vertex already exists, that vertex
// is returned instead and, if wanted, the provided vertex is destroyed
// NOTE: rather expensive routine when used on large primitives
// TODO: change method name
function TVerticeList.AddVertex(AVertex: TVertice; freeOnExist: boolean = false): TVertice;
var index: integer;
begin
    index := InList(AVertex);       // based on x,y,z: checks if a simular vertex exists in the list
    if index<>-1 then begin
        Result := GetVertice(index);
        if freeOnExist then
            FreeAndNil(AVertex);
    end else begin
        Add(AVertex);
        Result := AVertex;
    end;
end;

function TVerticeList.IsStrictlyConvex: boolean;
var previousEdge, edge, previousCrossProduct, crossProduct: TVertex;
    i: integer;
begin
    result := true;

    if Count > 3 then begin

        previousEdge := VertexSubtract(VerticeToVertex(Vertices[1]), VerticeToVertex(Vertices[0]));
        edge := VertexSubtract(VerticeToVertex(Vertices[2]), VerticeToVertex(Vertices[1]));
        previousCrossProduct := vertexCrossProduct(previousEdge, edge);
        previousEdge := edge;

        for i := 3 to Count - 1 do begin
            edge := VertexSubtract(VerticeToVertex(Vertices[i]), VerticeToVertex(Vertices[i-1]));

            crossProduct := vertexCrossProduct(previousEdge, edge); // plane direction, or 'not normalized plane normal'
            if vertexDotProduct(crossProduct, previousCrossProduct) <= 0 then begin
                result := false;
                break;
            end;
 
            previousCrossProduct := crossProduct;
            previousEdge := edge;
        end;
    end;
end;

procedure TVerticeList.InsertList(index: integer; list: TVerticeList);
var i: integer;
begin
    if Assigned(list) then begin
        for i := 0 to list.Count-1 do
            Insert(index+i,list[i]);
    end;
end;

procedure TVerticeList.Reverse;
var i: integer;
    tempList: TVerticeList;
begin
    tempList := TVerticeList.Create(false,true);
    tempList.Assign(self);
    Clear;                                                                      // clear list
    for i := tempList.Count-1 downto 0 do                                       // add temp to list in reverse order
        Add(tempList[i]);
    FreeAndNil(tempList);
end;

function TVerticeList.Copy: TVerticeList;
var i: integer;
    list: TVerticeList;
begin
    list := TVerticeList.Create(false, false);
    for i := 0 to Count-1 do
        list.Add(Vertices[i]);
    Result := list;
end;

// order the list of vertices based on distance of each vertex to the first vertex in the list
{
procedure TVerticeList.Sort(L,H: integer);
var t0,t1: TVertice;
begin
    t0 := vectorsub(vertices[L],vertices[0]);
    t1 := vectorsub(vertices[H],vertices[0]);

    if t1.length<t0.length then
        Exchange(L,H);

    // TODO, JC: decide what to do when length is 0

    inc(H);
    if H=Count then begin
        inc(L);
        H := L+1;
    end;

    if (H<Count) then
        Sort(L,H);

    FreeAndNil(t0);
    FreeAndNil(t1);
end;
}
// ### TEdge ###################################################################

function TEdge.GetCenterVertex: TVertice;
begin
    fCenterVertex.x := (StartVertex.x + EndVertex.x) / 2;
    fCenterVertex.y := (StartVertex.y + EndVertex.y) / 2;
    fCenterVertex.z := (StartVertex.z + EndVertex.z) / 2;
    Result := fCenterVertex;
end;

function TEdge.GetDirectionVertor: TVertice;
begin
    fDirectionVertor.x := StartVertex.x - EndVertex.x;
    fDirectionVertor.y := StartVertex.y - EndVertex.y;
    fDirectionVertor.z := StartVertex.z - EndVertex.z;
    fDirectionVertor.Normalize;
    Result := fDirectionVertor;
end;

function TEdge.GetLength: double;
var x,y,z: double;
begin
    x := StartVertex.x-EndVertex.x;
    y := StartVertex.y-EndVertex.y;
    z := StartVertex.z-EndVertex.z;
    Result := sqrt(sqr(x) + sqr(y) + sqr(z));
end;

constructor TEdge.Create(v0, v1: TVertice; polyList: TPolygonList = nil);
begin
    // set start/end vertices references
    StartVertex := v0;
    EndVertex := v1;

    // create property vertices
    fCenterVertex := TVertice.Create(0,0,0);
    fDirectionVertor := TVertice.Create(0,0,0);

    // create a polygonlist
    fPolygonList := TPolygonList.Create(false,true);

    // TODO: copy polylist to own internal polylist
    if Assigned(polyList) then begin
    end;
end;

destructor TEdge.Destroy;
begin
    // free stuff
    FreeAndNil(fCenterVertex);
    FreeAndNil(fDirectionVertor);
    FreeAndNil(fPolygonList);
    inherited;
end;

//Get's the shared vertex between this edge and AOtherEdge
function TEdge.GetSharedVertex(AOtherEdge: TEdge): TVertice;
begin
  Result := nil;

  if (StartVertex = AOtherEdge.StartVertex) or
     (StartVertex = AOtherEdge.EndVertex) then
      Result := StartVertex;

  if (EndVertex = AOtherEdge.StartVertex) or
     (EndVertex = AOtherEdge.EndVertex) then
      Result := EndVertex;
end;

//Checks if this edge is adjacent to AEdge
function TEdge.IsAdjacent(AEdge: TEdge): Boolean;
var
  adjacentEdges: TEdgeList;
begin
  adjacentEdges := TEdgeList.Create(false,true);

  // get all adjacent edges (unique pointers)
  adjacentEdges.AddList(StartVertex.EdgeList);
  adjacentEdges.AddList(EndVertex.EdgeList);

  // and see if the provided edge is in that list
  Result := adjacentEdges.IndexOf(AEdge)<>-1;

  FreeAndNil(adjacentEdges);
end;

//Puts all adjacent polygons in APolygonList
//If ASelection is assigned, only elements in ASelection will be added
procedure TEdge.GetAdjacent(APolygonList, ASelection: TPolygonList);
var
  I: Integer;
begin
    if Assigned(APolygonList) then
    begin
      APolygonList.Clear;

      if assigned(ASelection) then
      begin
        for I := 0 to PolygonList.Count - 1 do
          if ASelection.IndexOf(PolygonList[I]) <> -1 then
            APolygonList.Add(PolygonList[I]);
      end
      else
        APolygonList.AddList(PolygonList);
    end;
end;

//Puts all adjacent edges in AEdgeList
procedure TEdge.GetAdjacent(AEdgeList: TEdgeList);
begin
    if Assigned(AEdgeList) then
    begin
        AEdgeList.Clear;
        AEdgeList.AddList(StartVertex.EdgeList);
        AEdgeList.Addlist(EndVertex.EdgeList);
        AEdgeList.Remove(Self);
        AEdgeList.Remove(Self);  // the original edge might appear twice (if doubles are allowed), as it is listed in the edgelists for both the startvertex and endvertex
    end;
end;

procedure TEdge.GetAdjacentEdgesForEdgeLoop(edgeList: TEdgeList; useStartVertex: boolean);
var startEdge, nextEdge, currentEdge: TEdge;
    currentVertex: TVertice;
    doLoop, isFound: boolean;
    counter: integer;
begin
    startEdge   := Self;
    currentEdge := startEdge;
    // the loop might be interrupted by a pole (poles are vertices where more or less
    // than four edges converge); we traverse the loop twice, once using the startvertex,
    // and the second time using the endvertex as starting point for the loop
    if useStartVertex then
        currentVertex := startEdge.StartVertex
    else
        currentVertex := startEdge.EndVertex;
    nextEdge := nil;
    doLoop := true;
    while doLoop do begin
        if currentVertex.EdgeList.Count<>4 then
            doLoop := false
        else begin
            isFound := false;
            counter := 0;
            // we examine all 4 edges adjacent to the current vertex until
            // we find an edge that does not share its polygons with the
            // current edge
            while (not isFound) and (counter<4) do begin
                // we want to know if any of the polygons adjacent to the next edge
                // are also adjacent to the current edge; if none of them are, the
                // new edge is the edge we're looking for
                nextEdge := currentVertex.EdgeList[counter];
                if nextEdge.PolygonList.IsIn(currentEdge.PolygonList)=0 then
                    isFound := true
                else
                    inc(counter);
            end;
            // if we haven't found an edge, the edgeloop stops here (probably a pole)
            if not isFound then
                doLoop := false
            else begin
                // we stop if the next edge is the same as the start edge (the loop is complete)
                if nextEdge=startEdge then
                    doLoop := false
                else begin
                    // add the found edge to a list, set it edge to be the
                    // current edge and update the currentVertex as well
                    edgeList.Add(nextEdge);
                    currentEdge := nextEdge;
                    if currentEdge.StartVertex=currentVertex then
                        currentVertex := currentEdge.EndVertex
                    else
                        currentVertex := currentEdge.StartVertex;
                end;
            end;
        end;
    end;
end;

procedure TEdge.GetOneAdjacentEdgeForEdgeRing(edgeList: TEdgeList; startLeftPolygon: boolean; polygon: TPolygon);
var currentPolygon: TPolygon;
    opposingEdge: TEdge;
    index: integer;
begin
    if PolygonList.Count=2 then begin
        if not Assigned(polygon) then begin
            if startLeftPolygon then
                currentPolygon := PolygonList[0]
            else
                currentPolygon := PolygonList[1]
        end else begin
            if PolygonList[0]=polygon then
                currentPolygon := PolygonList[1]
            else
                currentPolygon := PolygonList[0];
        end;

        if currentPolygon.NumVertices=4 then begin

            index := (currentPolygon.EdgeList.IndexOf(Self) + 2) mod 4;
            opposingEdge := currentPolygon.EdgeList[index];
            if not edgeList.IsInList(opposingEdge) then begin
                edgeList.Add(opposingEdge);

                // recursive search for next opposing edge
                opposingEdge.GetOneAdjacentEdgeForEdgeRing(edgeList, startLeftPolygon, currentPolygon);
            end;
        end;
    end;
end;

function TEdge.IsUsingVertex(vertex: TVertice): boolean;
begin
    Result := false;
    if not Assigned(vertex) then Exit;
    Result := (StartVertex=vertex) or (EndVertex=vertex);
end;

function TEdge.GetOtherVertex(vertex: TVertice): TVertice;
begin
    Result := nil;
    if StartVertex=vertex then
        Result := EndVertex;
    if EndVertex=vertex then
        Result := StartVertex;
end;

function TEdge.Copy: TEdge;
begin
    Result := TEdge.Create(StartVertex,EndVertex);
    Result.PolygonList.Assign(PolygonList);
end;

procedure TEdge.Snap(ASnapValue: integer);
begin
    StartVertex.Snap(ASnapValue);
    EndVertex.Snap(ASnapValue);
end;

// ### TEdgeList ###############################################################

function TEdgeList.GetEdge(anIndex: integer): TEdge;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TEdge(Items[anIndex])
    else
        Result := nil;
end;

procedure TEdgeList.AddEdge(anEdge: TEdge);
begin
    // we set a reference to this edge in both vertices and add the edge to the list
    anEdge.StartVertex.EdgeList.Add(anEdge);
    anEdge.EndVertex.EdgeList.Add(anEdge);
    Add(anEdge);
end;

procedure TEdgeList.GetAdjacent(APolygonList: TPolygonList);
var i: integer;
begin
    if Assigned(APolygonList) then
    begin
        APolygonList.Clear;
        for i := 0 to Count-1 do
            APolygonList.AddList(Edges[i].PolygonList);
    end;
end;

procedure TEdgeList.GetAdjacent(AVertexList: TVerticeList);
var i: integer;
begin
    if Assigned(AVertexList) then
    begin
        AVertexList.Clear;
        for i := 0 to Count-1 do
        begin
            AVertexList.Add(Edges[i].StartVertex);
            AVertexList.Add(Edges[i].EndVertex);
        end;
    end;
end;

procedure TEdgeList.GetStartEndVertices(AVertexList: TVerticeList);
var i: integer;
    tempVertices: TVerticeList;
begin
    if IsContinuousSet() then
    begin
        AVertexList.Clear;

        // get all vertices as used by the provided edgelist in a templist
        tempVertices := TVerticeList.Create(false,false);
        for i := 0 to Count-1 do begin
            tempVertices.Add(Edges[i].StartVertex);
            tempVertices.Add(Edges[i].EndVertex);
        end;

        // check how many times a vertex occurs in the templist; if it only
        // occurs once, it is either the begin- or end of the provided edges
        for i := 0 to Count-1 do begin
            if tempVertices.NumberOfOccurances(Edges[i].StartVertex)=1 then
                AVertexList.Add(Edges[i].StartVertex);
            if tempVertices.NumberOfOccurances(Edges[i].EndVertex)=1 then
                AVertexList.Add(Edges[i].EndVertex);
        end;

        FreeAndNil(tempVertices);
    end;
end;

function TEdgeList.isContinuousSet: Boolean;
var
  vertexList: TVerticeList;
begin
    // edges form a continuous set (no gaps) if all vertices but the first
    // and last are used by exactly 2 edges; the first and last vertices are
    // only used by a single edge in the list; so, if more than 2 vertices
    // are used by just one edge in the list, one or more gaps exists
    // OR (faster): the # of vertices - 1 should be equal to the # of edges
    vertexList := TVerticeList.Create(false,true);
    try
        GetAdjacent(VertexList);
        Result := vertexList.Count-1 = Count;
    finally
        FreeAndNil(vertexList);
    end;
end;

function TEdgeList.IsVertexUsed(AVertex: TVertice): Boolean;
var
    i: integer;
begin
  Result := false;
  for i:=0 to Count-1 do
     if Edges[i].IsUsingVertex(AVertex) then
     begin
         Result := true;
         Exit;
     end;
end;

function TEdgeList.Copy: TEdgeList;
var i: integer;
    list: TEdgeList;
begin
    list := TEdgeList.Create(false);
    for i := 0 to Count-1 do
        list.Add(Edges[i]);
    Result := list;
end;

// ### TPolygon ################################################################

constructor TPolygon.Create;
begin
    fNormal := TVertice.Create(0,0,0);
    fCenter := TVertice.Create(0,0,0);
    fVerticeList := TVerticeList.Create(false,true);
    fHighlightedUVCoordinates := TUVList.Create(false,false);
    fVertexNormals := TObjectList.Create(true);
    fUVCoords := TObjectList.Create(true);
    fEdgeList := TEdgeList.Create(false,true);
    fBoundingBoxCenter := TVertice.Create(0,0,0);
    fAllowLightmap := true;
    //fIsPathPolygon := false;
end;

destructor TPolygon.Destroy;
begin
    FreeAndNil(fNormal);
    FreeAndNil(fCenter);
    FreeAndNil(fVerticeList);
    FreeAndNil(fHighlightedUVCoordinates);
    FreeAndNil(fUVCoords);
    FreeAndNil(fEdgeList);
    FreeAndNil(fBoundingBoxCenter);
    FreeAndNil(fVertexNormals);
    inherited;
end;

function TPolygon.Area: double;
var i: integer;
    verti,vertj: TVertice;
    sum: double;
begin
//    2 A(P) = (N . (sum_{i=0}^{n-1} (v_i x v_{i+1})))
{    n := planenormal;
    a := TVertice.Create(0,0,0);
    for i := 0 to numvertices-1 do begin
        verti := GetVertice(i);
        vertj := GetVertice((i+1) mod numvertices);
        c := crossproduct(verti,vertj);
        a := vectoradd(a,c);
        c.Free;
    end;

    verti := GetVertice(0);
    d := dotproduct(n,verti);
    verti := TVertice.Create(0,0,0);
    if dotproduct(n,verti)>d then begin

    end;
    verti.Free;

    Result := 0.5*(dotproduct(n,a));
    a.Free;
    n.Free; }

    sum := 0;
    for i := 0 to numvertices-1 do begin
        verti := VerticeList[i];
        vertj := VerticeList[(i+1) mod numvertices];

        if verti.x=0 then
            sum := sum + ((verti.Z * vertj.Y) - (vertj.Z * verti.Y))            // 2D calculations!
        else if verti.y=0 then
            sum := sum + ((verti.X * vertj.Z) - (vertj.X * verti.Z))
        else if verti.z=0 then
            sum := sum + ((verti.X * vertj.Y) - (vertj.X * verti.Y));
    end;
    Result := 0.5*sum;      // divide by two
end;

function TPolygon.GetNumVertices: integer;
begin
    Result := VerticeList.Count;
end;

function TPolygon.GetNumHighlightedUVCoordinates: integer;
begin
  result := HighlightedUVCoordinates.Count;
end;

function TPolygon.GetIsClockWise: boolean;
begin
    Result := Area>0;
end;

function TPolygon.Copy: TPolygon;
var i: integer;
    polygon: TPolygon;
begin
    polygon := TPolygon.Create;
    polygon.Material := Material;
    //polygon.IsPathPolygon := IsPathPolygon;
    for i := 0 to numVertices-1 do begin
        polygon.VerticeList.Add(VerticeList[i]);
        //polygon.AddAllUVCoordsForIndex(self, i);
    end;
    Result := polygon;
end;

function TPolygon.HasUVCoordinates: boolean;
begin
    Result := fUVCoords.Count > 0;
end;

function TPolygon.HasVertexNormals: boolean;
begin
    Result := fVertexNormals.Count > 0;
end;

// regenerates all UV Coordinates using planar mapping algorithm
procedure TPolygon.UpdateUVCoordinates(aFitToFace: boolean = false);
var i,j,w,h: integer;
    pn,vertex: TVertice;
    uvcoord: TUVCoord;
    uvCoords: TUVList;
    polyPlane: TPolyPlane;
    minU, maxU, minV, maxV: single;
begin
    pn := Normal;       // check: should be _plane_ normal
    polyPlane := Plane;

    for j := 0 to Material.NumTextureLayers - 1 do begin
        minU := 0;
        maxU := 0;
        minV := 0;
        maxV := 0;

        for i := 0 to numVertices-1 do begin
            // create layer if it doesn't exist yet.
            if j >= fUVCoords.Count then
                fUVCoords.Add(TUVList.Create);
            uvCoords := fUVCoords[j] as TUVList;

            // get the uv coordinate we need to calculate
            if i >= uvCoords.Count then begin
                uvcoord := TUVCoord.Create(0, 0);
                uvcoords.Add(uvcoord);
            end;
            uvcoord := uvCoords.Get(i);

            // if this layer has a texture (other than the lightmap), perform planar mapping
            if  (j <> LIGHTMAP_TEXTURE_LAYER_INDEX) and
                Assigned(Material.TextureLayers[j]) and
                (Material.TextureLayers[j].LayerType = mlTexture) then begin

                w := Material.TextureLayers[j].Texture.Width;
                h := Material.TextureLayers[j].Texture.Height;

                vertex := VerticeList[i];

                case polyPlane of
                    ppYZ:
                    begin
                        uvcoord.u := - vertex.z / w * Sign(pn.x);
                        uvcoord.v := -vertex.y / h;
                    end;

                    ppXZ:
                    begin
                        uvcoord.u := vertex.x / w;
                        uvcoord.v := vertex.z / h * Sign(pn.y);
                    end;

                    ppXY:
                    begin
                        uvcoord.u := vertex.x / w * Sign(pn.z);
                        uvcoord.v := -vertex.y / h;
                    end;
                end;

                if aFitToFace then begin
                    if (i=0) and (j=0) then begin
                        minU := uvCoord.U;
                        maxU := uvCoord.U;
                        minV := uvCoord.V;
                        maxV := uvCoord.V;
                    end else begin
                        minU := Min(minU, uvCoord.U);
                        maxU := Max(maxU, uvCoord.U);
                        minV := Min(minV, uvCoord.V);
                        maxV := Max(maxV, uvCoord.V);
                    end;
                end
                else begin
                    uvcoord.U := uvCoord.U * defaultUVScale;
                    uvcoord.v := uvCoord.V * defaultUVScale;
                end;
            end else if (j <> LIGHTMAP_TEXTURE_LAYER_INDEX) then
            begin // if this layer does not have texture information (or is lightmap or color layer), set UV coordinates to (0, 0)
                uvCoord.U := 0;
                uvCoord.V := 0;
            end;
        end;

        if aFitToFace and (j <> LIGHTMAP_TEXTURE_LAYER_INDEX) and ( maxU > minU ) and ( maxV > minV ) then begin
            // we scale and translate all coordinates to fit into the 0,1 region
            for i := 0 to numVertices-1 do begin
                uvCoord := UVCoordinates[j][i];

                uvCoord.U := ( uvCoord.U - minU ) / ( maxU - minU );
                uvCoord.V := ( uvCoord.V - minV ) / ( maxV - minV );
            end;
        end;
    end;
end;

// convenience function, copies all uv coordinates (for all texture layers) for a given index;
procedure TPolygon.AddCopiedUVCoordinates(const aPoly: TPolygon; const anIndex: Integer);
var material: TMaterial;
    uvCoord: TUVCoord;
    uvCoords: TUVList;
    layerIndex: integer;
begin
    material := aPoly.Material;
    for layerIndex := 0 to material.NumTextureLayers - 1 do
    begin
        // target UVCoord layer might not exist yet, resolve
        if layerIndex >= fUVCoords.Count then
            fUVCoords.Add(TUVList.Create);

        if Assigned(material.TextureLayers[layerIndex]) then begin
            uvCoord := aPoly.UVCoordinates[layerIndex].Get(anIndex);

            if not Assigned(uvCoord) then
                raise Exception.Create( Format( 'No source uv coord found at layerindex %d for layer %d.', [anIndex, layerIndex] ));

            uvCoords := UVCoordinates[layerIndex] as TUVList;
            uvCoords.Add(uvCoord.Copy);
        end;
    end;
end;

procedure TPolygon.AddUVCoordinatesForVertex(const aVertex: TVertice);
begin
    InsertUVCoordinatesForVertex(-1, aVertex);
end;

// Calculate the UV Coordinates of a new vertex, by first calculating the
// barycentric coordinates of the vertex within the polygon;
// Bit of a hack: anIndex of -1 means 'add to end'.
// Should be called when adding new vertices to an existing polygon.
// If the polygon is 'under construction' (doesn't have any UV Coords), the method does nothing.
procedure TPolygon.InsertUVCoordinatesForVertex(const anIndex: integer; const aVertex: TVertice);
var v0, v1, v2, p, u, v, w: TVertex;
    uu, uv, wv, vv, wu, b0, b1, b2, denominator, nu, nv: single;
    uv0, uv1, uv2: TUVCoord;
    startIndex, layerIndex: integer;
begin
    if not HasUVCoordinates then
        exit;

    // We loop through all vertices until we find three subsequent vertices that
    // actually form a triangle (are not in one line). This is virtually free
    // for strictly convex polygons, at the cost of being a bit more expensive
    // for polygons that actually do have multiple vertices in a line.
    startIndex := -1;
    repeat
        startIndex := startIndex + 1;

        v0 := verticeToVertex(VerticeList[startIndex]);
        v1 := verticeToVertex(VerticeList[startIndex + 1]);
        v2 := verticeToVertex(VerticeList[startIndex + 2]);

        p := verticeToVertex(aVertex);
        u := vertexSubtract(v1, v0);
        v := vertexSubtract(v2, v0);
        w := vertexSubtract(P, v0);

        uu := vertexDotProduct(u, u);
        uv := vertexDotProduct(u, v);
        wv := vertexDotProduct(w, v);
        vv := vertexDotProduct(v, v);
        wu := vertexDotProduct(w, u);

        denominator := uv * uv - uu * vv;
    until (denominator <> 0) or (startIndex > NumVertices - 4);

    if denominator = 0 then
        denominator := 1; // degenerate poly

    // barycentric coordinates
    b1 := (uv * wv - vv * wu) / denominator;
    b2 := (uv * wu - uu * wv) / denominator;
    b0 := 1 - b1 - b2;

    // use barycentric coordinates to calculate u,v coordinates for each texturelayer in the used material
    for layerIndex := 0 to Material.NumTextureLayers - 1 do begin               // NOTE JC: including the lightmap layer?
        if Assigned(Material.TextureLayers[layerIndex]) then begin
            uv0 := UVCoordinates[layerIndex][startIndex];
            uv1 := UVCoordinates[layerIndex][startIndex + 1];
            uv2 := UVCoordinates[layerIndex][startIndex + 2];

            nu := b0 * uv0.U + b1 * uv1.U + b2 * uv2.U;
            nv := b0 * uv0.V + b1 * uv1.V + b2 * uv2.V;

            if anIndex = -1 then
                UVCoordinates[layerIndex].Add(nu, nv)
            else
                UVCoordinates[layerIndex].Insert(anIndex, nu, nv)
        end;
    end;
end;

// used in Polygon_AddVertex for adding UV coordinates to new vertices in the polygon
procedure TPolygon.AddEmptyUVCoordinatesForVertex(const aVertex: TVertice);
var layerIndex: integer;
begin
    for layerIndex := 0 to Material.NumTextureLayers - 1 do begin
        if Assigned(Material.TextureLayers[layerIndex]) then begin

            // create uvlist for this material layer if it doesn't exist yet
            if layerIndex >= fUVCoords.Count then
                fUVCoords.Add(TUVList.Create);

            // and add an empty (initialized to zeros) uv coord to this layer
            UVCoordinates[layerIndex].Add(0,0)
        end;
    end;
end;

procedure TPolygon.DeleteUVCoordinates(const anIndex: integer);
var layerIndex: integer;
begin
    for layerIndex := 0 to Material.NumTextureLayers - 1 do begin
        if Assigned(Material.TextureLayers[layerIndex]) then
            UVCoordinates[layerIndex].Delete(anIndex);
    end;
end;

// Adjust the UV Coordinate Layers to the number of layers in the current material.
// This method should solve _any_ mismatch issues between uvCoords and Material.
// Could be done implicitly during material assignment, but this is the _simple_
// solution.
procedure TPolygon.AdjustUVCoordinatesToMaterial;
var i: integer;
    uvList: TUVList;
begin
  if not Assigned(fMaterial) then exit;

  // Safe-guard for uninitialized uv-coordinates
  if fUVCoords.Count = 0 then
      UpdateUVCoordinates;

  // Add additional layers if needed.
  for i := fUVCoords.Count to fMaterial.NumTextureLayers - 1 do begin
      uvList := fUVCoords[i - 1] as TUVList;
      fUVCoords.Add(uvList.Copy);
  end;

  // Remove redundant layers if needed;
  for i := fUVCoords.Count - 1 downto fMaterial.NumTextureLayers do
      fUVCoords.Delete(i);

  // Fill previously unssigned layers if now assigned (most common case).
  for i := 1 to fMaterial.NumTextureLayers - 1 do begin
      uvList := fUVCoords[i] as TUVList;
      if  (uvList.Count = 0) and
          (i <> LIGHTMAP_TEXTURE_LAYER_INDEX) and
          Assigned(fMaterial.TextureLayers[i]) and
          (Material.TextureLayers[i].LayerType = mlTexture) then begin

          uvList.Assign(fUVCoords[i - 1] as TUVList);
      end;
  end;
end;

procedure TPolygon.Flip;
var i: integer;
begin
    VerticeList.Reverse;                                                        // reverse the vertex order

    for i := 0 to Material.NumTextureLayers - 1 do
        UVCoordinates[i].Reverse;                                               // and the corresponding uv coordinates
end;

procedure TPolygon.Snap(aSnapValue: integer);
var i: integer;
begin
    for i := 0 to VerticeList.Count-1 do
        VerticeList[i].Snap(aSnapValue);
end;

procedure TPolygon.ClearVertexNormals;
begin
    fVertexNormals.Clear;
end;

procedure TPolygon.AddVertexNormal(const aNormal: TVertice);
begin
    fVertexNormals.Add(aNormal);
end;

procedure TPolygon.Assign(polygon: TPolygon);
begin
    VerticeList.Assign(polygon.VerticeList);
    EdgeList.Assign(polygon.EdgeList);
end;

function TPolygon.IsPointInTriangle(p,a,b,c: TVertice; EPS: double = 0): boolean;    // TODO: move to math unit or some general vector unit?
var ap,bp,cp,v1,v2,v3,r: TVertice;
    length: single;
begin
    ap := vectorsub(a,p);
    bp := vectorsub(b,p);
    cp := vectorsub(c,p);
    v1 := vectornormalize(ap);  // three normalized vectors from triangle vertices to the point in question
    v2 := vectornormalize(bp);
    v3 := vectornormalize(cp);

    r := TVertice.Create(v1.x+v2.x+v3.x,v1.y+v2.y+v3.y,v1.z+v2.z+v3.z);
    length := sqrt(sqr(r.x) + sqr(r.y) + sqr(r.z));     // length of vector

    r.Free;
    v1.Free;
    v2.Free;
    v3.Free;
    ap.Free;
    bp.Free;
    cp.Free;

    Result := length<=1-EPS; // EPS has been included for the PolygonConvexionator; note: these routines should be improved somehow...
end;

// NOTE: resetting components to 0 is done for 2D selection mechanism; that
// mechanism provides a 3D vertex with one of its components set to 0 by
// default due to the 2D views which use 2 components only; this effectively
// projects the triangle onto one of the standard planes
// a better system would probably project the vertex (point) onto the polygon
// plane somehow; the other way around so to speak
function TPolygon.IsPointInPolygon(aVertex: TVertice; EPS: double; ResetComponents: boolean = true): boolean;
var i: integer;
    v0,v1,v2: TVertice;
begin
    Result := false;

    if ResetComponents then begin

        for i := 1 to numVertices-2 do begin

            v0 := VerticeList[i].Copy;
            v1 := VerticeList[(i+1)].Copy;
            v2 := VerticeList[0].Copy;
            try
                if aVertex.x=0 then begin
                    v0.x := 0;
                    v1.x := 0;
                    v2.x := 0;
                end;
                if aVertex.y=0 then begin
                    v0.y := 0;
                    v1.y := 0;
                    v2.y := 0;
                end;
                if aVertex.z=0 then begin
                    v0.z := 0;
                    v1.z := 0;
                    v2.z := 0;
                end;

                Result := IsPointInTriangle(aVertex,v0,v1,v2,EPS);
                if Result then Exit;
                
            finally
                v0.Free;
                v1.Free;
                v2.Free;
            end;
        end;

    end else begin

        for i := 1 to numVertices-2 do begin
            v0 := VerticeList[i].Copy;
            v1 := VerticeList[(i+1)].Copy;
            v2 := VerticeList[0].Copy;

            try
                Result := IsPointInTriangle(aVertex,v0,v1,v2,EPS);
                if Result then Exit;
            finally
                v0.Free;
                v1.Free;
                v2.Free;
            end;

        end;
    end;
end;

// INFO, PJ: 2D Point in Polygon test from comp.graphics.algorithms FAQ
function TPolygon.IsUVCoordinateInPolygon(const aUVCoordinate: TUVCoordinate): boolean;
var i, j: integer;
    uvCoord1, uvCoord2: TUVCoord;
begin
    result := false;

    j := NumVertices - 1;
    for i := 0 to NumVertices - 1 do begin
        uvCoord1 := UVCoordinates[0][i];
        uvCoord2 := UVCoordinates[0][j];

        if  ((( uvCoord1.V <= aUVCoordinate.V ) and ( aUVCoordinate.V < uvCoord2.V )) or (( uvCoord2.V <= aUVCoordinate.V ) and ( aUVCoordinate.V < uvCoord1.V ))
            ) and
            ( aUVCoordinate.U < ( uvCoord2.U - uvCoord1.U ) * ( aUVCoordinate.V - uvCoord1.V ) / ( uvCoord2.V - uvCoord1.V ) + uvCoord1.U ) then

            result := not result;

        j := i;
    end;
end;

function TPolygon.IsPointInFront(aVertex: TVertice): boolean;
var n: TVertice;
    d: double;
begin
    n := Normal.copy;
    if aVertex.x=0 then                                                           // 2d calculations!
        n.x := 0
    else if aVertex.y=0 then
        n.y := 0
    else if aVertex.z=0 then
        n.z := 0;

    d := -dotproduct(VerticeList[0],n);                                         // distance from polyplane to origin
    Result := (d > planeD(aVertex,n));
    FreeAndNil(n);
end;


function TPolygon.IsIntersectedByRay(const aRay: TRay; out aDistance: single): boolean;
var i: integer;
    v0, v1, v2: TVertex;
    t, u, v: Single;
begin
    result := false;
    for i := 0 to numVertices - 3 do begin
        v0 := verticeToVertex(VerticeList[i + 2]);
        v1 := verticeToVertex(VerticeList[i + 1]);
        v2 := verticeToVertex(VerticeList[0]);
        if IntersectsRayTriangle(aRay, v0, v1, v2, t, u, v) then begin
            // The IsNan-test prevents degenerate faces from returning hits
            result := ( t > EPSILON ) and (not IsNan(t));
            aDistance := t;
            exit;
        end;
    end;
end;

// sends all triangles of the polygon to IntersectsRayTriangle (reversed winding)
function TPolygon.IsIntersectedByRay(const aRay: TRay): boolean;
var t: single;
begin
    result := IsIntersectedByRay(aRay, t);
end;

function TPolygon.IsCoplanar(aPlane: TPlane): boolean;
var pn: TVertice;
    pd,q: double;
begin
    pn := Normal.copy;
    pd := planedistance;

    if (pn.x<>0) and (aPlane.normal.x<>0) then
        q := pn.x*aPlane.normal.x
    else if (pn.y<>0) and (aPlane.normal.y<>0) then
        q := pn.y*aPlane.normal.y
    else if (pn.z<>0) and (aPlane.normal.z<>0) then
        q := pn.z*aPlane.normal.z
    else if (pd<>0) and (aPlane.distance<>0) then
        q := pd*aPlane.distance
    else
        q := 1;

    pn.x := pn.x*q;
    pn.y := pn.y*q;
    pn.z := pn.z*q;
    pd := pd*q;

    // if matrix has rank one, then planes are the same
    Result := (pn.x=aPlane.normal.x) and
              (pn.y=aPlane.normal.y) and
              (pn.z=aPlane.normal.z) and
              (pd=aPlane.distance);

{    if (pn.x=aPlane.normal.x) and
       (pn.y=aPlane.normal.y) and
       (pn.z=aPlane.normal.z) and
       (pd=aPlane.distance) then  // matrix has rank one; planes are the same
        Result := true
    else
        Result := false;     }
    pn.Free;
end;

// checks if all vertices of the supplied polygon are on the plane of the polygon
// NOTE: IsEqualNormal is probably not always correct if we're only interested in
// plane (!) equality and NOT direction equality...
function TPolygon.IsCoplanar(aPolygon: TPolygon): boolean;
const EPS = 0.001;
var i: integer;
    d,pd: single;
begin
    Result := false;
    if Assigned(aPolygon) then begin
    //if IsEqualNormal(aPolygon) then begin
        d := -dotproduct(VerticeList[0],Normal);                                // distance to parralel plane passing through origin
        // check each vertex of the other polygon against d
        for i := 0 to aPolygon.VerticeList.Count-1 do begin
            pd := -dotproduct(aPolygon.VerticeList[i],Normal);
            if (pd<d-EPS) or (pd>d+EPS) then                                    // we check for equal distance
                Exit;
        end;
        Result := true;
    //end;
    end;
end;

// NOTE: page 297 Lengyel
function TPolygon.IsEqualNormal(aPolygon: TPolygon): boolean;
const EPS = 0.001;   // EPS because of inaccuracies in floating point calculations
begin
    //Result := DotProduct(Normal,aPolygon.Normal) > EPS;
    Result := Abs(Normal.x-aPolygon.Normal.x) +
              Abs(Normal.y-aPolygon.Normal.y) +
              Abs(Normal.z-aPolygon.Normal.z) < EPS;
end;

function TPolygon.IsStrictlyConvex: boolean;
var previousEdge, edge, previousCrossProduct, crossProduct: TVertex;
    i: integer;
begin
    result := true;

    if NumVertices > 3 then begin

        previousEdge := VertexSubtract(VerticeToVertex(VerticeList[1]), VerticeToVertex(VerticeList[0]));
        edge := VertexSubtract(VerticeToVertex(VerticeList[2]), VerticeToVertex(VerticeList[1]));
        previousCrossProduct := vertexCrossProduct(previousEdge, edge);
        previousEdge := edge;

        for i := 3 to NumVertices - 1 do begin
            edge := VertexSubtract(VerticeToVertex(VerticeList[i]), VerticeToVertex(VerticeList[i-1]));

            crossProduct := vertexCrossProduct(previousEdge, edge); // plane direction, or 'not normalized plane normal'
            if vertexDotProduct(crossProduct, previousCrossProduct) <= 0 then begin
                result := false;
                break;
            end;
 
            previousCrossProduct := crossProduct;
            previousEdge := edge;
        end;
    end;
end;

function TPolygon.HasIdenticalVertexNormals: boolean;
var i: integer;
begin
    result := true;
    for i := 1 to fVertexNormals.Count - 1 do begin
        if not VertexNormals[i - 1].Equals(VertexNormals[i]) then begin
            result := false;
            break;
        end;
    end;
end;

procedure TPolygon.ProjectToScreen(const aModelViewMatrix, aProjectionMatrix: TMatrix4x4; const aViewport: TViewport;
                                   var aPolygonPoints: TPoint2DArray);
var
    i: integer;
begin
    SetLength(aPolygonPoints,NumVertices);
    for i := 0 to NumVertices - 1 do
        aPolygonPoints[i] := VerticeList[i].ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewPort);
end;



function TPolygon.GetNormal: TVertice;
const EPS=0.001;
var averageYZ, averageZX, averageXY: double;
    index: integer;
    vertex, previousVertice: TVertice;
    length: double;
begin
    averageYZ := 0;
    averageZX := 0;
    averageXY := 0;

    previousVertice := VerticeList[numVertices - 1];
    for index := 0 to numVertices - 1 do begin
        vertex := VerticeList[index];
        averageYZ := averageYZ + (previousVertice.Z + vertex.Z) * (vertex.Y - previousVertice.Y);
        averageZX := averageZX + (previousVertice.X + vertex.X) * (vertex.Z - previousVertice.Z);
        averageXY := averageXY + (previousVertice.Y + vertex.Y) * (vertex.X - previousVertice.X);
        previousVertice := vertex;
    end;

    length := Sqrt(averageYZ * averageYZ + averageZX * averageZX + averageXY * averageXY);

    if length < EPS then length := EPS;
    fNormal.X := averageYZ / length;
    fNormal.Y := averageZX / length;
    fNormal.Z := averageXY / length;

    result := fNormal;
end;

function TPolygon.GetDistanceToOrigin: double;
begin
    Result := -dotproduct(VerticeList[0],Normal);
end;

function TPolygon.GetPlane;
var pn: TVertice;
begin
    // PJ: This might be a bit tricky, is normal always the plane normal by our definition?
    pn := Normal;

    if (Abs(pn.x) >= Abs(pn.y)) and (Abs(pn.x) >= Abs(pn.z)) then result := ppYZ
    else if (Abs(pn.y) >= Abs(pn.x)) and (Abs(pn.y) >= Abs(pn.z)) then result := ppXZ
    else result := ppXY;
end;

// Auto-resolves. For read-only use only, don't use for mutations to uvcoords.
// if list doesn't exist for given index, create a copy of the list before
function TPolygon.GetUVCoordsForIndex(const anIndex: integer): TUVList;
var index: Integer;
    uvlist: TUVList;
begin
    if anIndex >= fUVCoords.Count then begin
        for index := fUVCoords.Count to anIndex do begin
            if index > 0 then begin
                uvList := fUVCoords[index-1] as TUVList;
                uvList := uvList.Copy;
            end else
                uvList := TUVList.create;
            fUVCoords.Add(uvList);
        end;
    end;
    result := fUVCoords[anIndex] as TUVList;
end;

procedure TPolygon.SetUVCoordsForIndex(const anIndex: Integer; const aUVCoords: TUVList);
var index: integer;
begin
    for index := fUVCoords.Count to anIndex - 1 do
        fUVCoords.Add(TUVList.Create);

    if anIndex < fUVCoords.Count then
        fUVCoords[anIndex] := aUVCoords
    else
        fUVCoords.Add(aUVCoords);
end;

function TPolygon.GetPlaneDistance: double;
begin
    Result := -dotproduct(VerticeList[0],Normal);
end;

//Get's the previous edge from the polygon (with respect to AStartVertex)
function TPolygon.GetPreviousEdge(AStartVertex: TVertice): TEdge;
var indexPV: integer;
    pv: TVertice;
begin
    Result := nil;
    if Assigned(AStartVertex) then
    begin
        // locate and get the previous vertex in relation to sv
        indexPV := VerticeList.IndexOf(AStartVertex)-1;
        if indexPV < 0 then indexPV := NumVertices-1;
        pv := VerticeList[indexPV];

        // return edge that is made out of sv,pv
        Result := AStartVertex.GetEdge(pv);
    end;
end;

//Get's the next edge from the polygon (with respect to AStartVertex)
function TPolygon.GetNextEdge(AStartVertex: TVertice): TEdge;
var indexNV: integer;
    nv: TVertice;
begin
    Result := nil;
    if Assigned(AStartVertex) then
    begin
        // locate and get the previous vertex in relation to sv
        indexNV := (VerticeList.IndexOf(AStartVertex)+1) mod NumVertices;
        nv := VerticeList[indexNV];

        // return edge that is made out of sv,nv
        Result := AStartVertex.GetEdge(nv);
    end;
end;

//Get's the adjacent edges of this polygon and puts them in AEdgeList.
//An adjacent edge is an edge that uses one of the polygons vertices
//WARNING: The edges of the polygon will appear multiple times on this list
procedure TPolygon.GetAdjacent(AEdgeList: TEdgeList);
var i: integer;
    vertex: TVertice;
begin
    if Assigned(AEdgeList) then
    begin
        AEdgeList.Clear;

        for i := 0 to NumVertices-1 do
        begin
            vertex := VerticeList[i];
            AEdgeList.AddList(vertex.EdgeList);
        end;
    end;
end;

//Get's the adjacent polygons of this polygon and puts them in APolygonList.
//If ASelection is assigned, only polygons present in ASelection will be added.
//If MustBeCoPlanar is true, only polygons that lie on the same plane as "this" polygon
//will be added to the list
//The list will only contain uniques and not the polygon itsself
procedure TPolygon.GetAdjacent(APolygonList, ASelection: TPolygonList; MustBeCoPlanar: Boolean);
var i,j: integer;
    edge: TEdge;
begin
    // make sure we only store polygons once and clear the list
    APolygonList.UniquesOnly := true;
    APolygonList.Clear;

    // loop through edgelist, adding the polygonlist of each edge to the provided list
    for i := 0 to EdgeList.Count-1 do
    begin
        edge := EdgeList[i];

        //Check all polygons
        for j := 0 to edge.PolygonList.Count-1 do
        begin
          //Do not add "self" to APolygonList
          if edge.PolygonList[j] = Self then
            Continue;

          //Skip polygons that are not on the Selected list, when it is assigned
          if Assigned(ASelection) then
            if ASelection.IndexOf(edge.PolygonList[j]) = -1 then
              Continue;

          //If polygon is not coplanar while it should, skip it!
          if MustBeCoPlanar then
            if not IsCoPlanar(edge.PolygonList[j]) then
              Continue;

          //Add polygon
          APolygonList.Add(edge.PolygonList[j]);
        end;
    end;
end;

//Get's the adjacent vertices of this polygon and puts them in AVertexList.
//If ASelection is assigned, only vertices present in ASelection will be added.
//If UsedOnlyByThisPolygon is true, only vertices that are only used by this polygon will be added.
procedure TPolygon.GetAdjacent(AVertexList, ASelection: TVerticeList; UsedOnlyByThisPolygon: Boolean);
var i,j: integer;
    polygonList: TPolygonList;
    vertex: TVertice;
    edge: TEdge;
begin
    AVertexList.Clear;

    if UsedOnlyByThisPolygon then
    begin
      //We only want vertices that are used ONLY by this polygon

      polygonList := TPolygonList.Create(false,true);
      try

        for i := 0 to VerticeList.Count-1 do
        begin
            polygonList.Clear;

            //Get all polygons used by vertex
            vertex := VerticeList[i];
            for j := 0 to vertex.EdgeList.Count-1 do
            begin
                edge := vertex.EdgeList[j];
                polygonList.AddList(edge.PolygonList);
            end;

            //If vertex has less than 2 polygons, add it
            if polygonList.Count<2 then
                AVertexList.Add(VerticeList[i]);
        end;

      finally
        FreeAndNil(polygonList);
      end;
    end
    else
    begin

      for i := 0 to VerticeList.Count-1 do
        if Assigned(ASelection) then
        begin
          //Only add vertices that are in ASelection
          if ASelection.IndexOf(VerticeList[i]) <> -1 then
            AVertexList.Add(VerticeList[i]);
        end
        else
          AVertexList.Add(VerticeList[i]);

    end;
end;

//Gets the edges shared between this polygon and AOtherPolygon
//and put them into AEdgeList
//If ClearList is true, AEdgeList will be cleared when the routine starts
procedure TPolygon.GetSharedEdges(AOtherPolygon: TPolygon; AEdgeList: TEdgeList; clearlist: boolean);
var
  i: integer;
begin
    if clearList then
        AEdgeList.Clear;

    for i := 0 to EdgeList.Count-1 do
    begin
        if AOtherPolygon.EdgeList.IndexOf(EdgeList[i])<>-1 then
            AEdgeList.Add(EdgeList[i]);
    end;
end;

//Gets the vertices shared between this polygon and AOtherPolygon
//and put them into AVertexList
//If ClearList is true, AVertexList will be cleared when the routine starts
procedure TPolygon.GetSharedVertices(AOtherPolygon: TPolygon; AVertexList: TVerticeList; clearlist: boolean);
var i: integer;
begin
    if clearList then
        AVertexList.Clear;

    for i := 0 to VerticeList.Count-1 do begin
        if AOtherPolygon.VerticeList.IndexOf(VerticeList[i])<>-1 then
            AVertexList.Add(VerticeList[i]);
    end;
end;

function TPolygon.GetBoundingBoxCenter: TVertice;
var minX, maxX, minY, maxY, minZ, maxZ: single;
    v: TVertice;
    i: integer;
begin
    if numVertices > 0 then begin
        v := VerticeList[0];
        minX := v.X;
        minY := v.Y;
        minZ := v.Z;

        maxX := minX;
        maxY := minY;
        maxZ := minZ;

        for i := 1 to numVertices-1 do begin
            v := VerticeList[i];
            if v.X < minX then minX := v.X;
            if v.X > maxX then maxX := v.X;
            if v.Y < minY then minY := v.Y;
            if v.Y > maxY then maxY := v.Y;
            if v.Z < minZ then minZ := v.Z;
            if v.Z > maxZ then maxZ := v.Z;
        end;

        fBoundingBoxCenter.X := ( minX + maxX ) / 2;
        fBoundingBoxCenter.Y := ( minY + maxY ) / 2;
        fBoundingBoxCenter.Z := ( minZ + maxZ ) / 2;
    end;

    Result := fBoundingBoxCenter;
end;

function TPolygon.GetVertexNormalForIndex(const anIndex: Integer): TVertice;
begin
    result := fVertexNormals[anIndex] as TVertice;
end;

// center=average vertex
function TPolygon.GetCenter: TVertice;
var i: integer;
    c,v: TVertice;
begin
    c := TVertice.Create(0,0,0);
    for i := 0 to numVertices-1 do begin
        v := VerticeList[i];
        c.x := c.x + v.x;
        c.y := c.y + v.y;
        c.z := c.z + v.z;
    end;
    fCenter.x := c.x / numVertices;
    fCenter.y := c.y / numVertices;
    fCenter.z := c.z / numVertices;
    FreeAndNil(c);
    Result := fCenter;
end;

// ### TPolygonList ############################################################

//Gets all vertices used by all polygons on the list
procedure TPolygonList.GetAdjacent(AVerticeList: TVerticeList);
var
  i: integer;
begin
    if Assigned(AVerticeList) then
    begin
        for i := 0 to Count-1 do
            AVerticeList.AddList(Polygons[i].VerticeList);
    end;
end;

//Get's all edges of all polygons on the list,
//NOT including the ones sharing only one vertex with the polygon
procedure TPolygonList.GetAdjacent(AEdgeList: TEdgeList);
var
  i: integer;
begin
    if Assigned(AEdgeList) then
    begin
        AEdgeList.Clear;

        for i := 0 to Count-1 do
            AEdgeList.AddList(Polygons[i].EdgeList);
    end;
end;

//Returns the averaged normal of a set of polygons
function TPolygonList.GetAveragedNormal(): TVertice;
var i: integer;
    polygon: TPolygon;

begin
    Result := TVertice.Create(0,0,0);
    for i := 0 to Count-1 do
    begin
        polygon := Polygons[i];
        Result.Add(polygon.Normal);
    end;
    Result.DivideByScalar(Count);
    Result.Normalize;
end;

function TPolygonList.GetPolygon(anIndex: integer): TPolygon;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TPolygon(Items[anIndex])
    else
        Result := nil;
end;

function TPolygonList.Copy: TPolygonList;
var i: integer;
    list: TPolygonList;
begin
    list := TPolygonList.Create(false, false);
    for i := 0 to Count-1 do
        list.Add(Polygons[i]);
    Result := list;
end;

// ### TUVCoord ################################################################

constructor TUVCoord.Create(vu, vv: double);
begin
    u := vu;
    v := vv;
end;

function TUVCoord.GetCoordinate: TUVCoordinate;
begin
  Result.u := U;
  Result.v := V;
end;

function TUVCoord.Copy: TUVcoord;
begin
    Result := TUVCoord.Create(u,v);
end;

procedure TUVCoord.Assign(uvcoord: TUVCoord);
begin
    u := uvcoord.U;
    v := uvcoord.V;
end;

procedure TUVCoord.Move(const aDelta: TUVCoordinate);
begin
    U := U + aDelta.U;
    V := V + aDelta.V;
end;

procedure TUVCoord.ApplyEditDimension(const anEditDimension: TEditUVDimension);
begin
    case anEditDimension of
        edU:    V := 0;
        edV:    U := 0;
    end;
end;

// ### TUVList #################################################################

procedure TUVList.Assign(const aList: TUVList);
var i: integer;
begin
    Clear;
    for i := 0 to aList.Count - 1 do
        Add(aList[i].Copy);
end;

function TUVList.GetCenter: TUVCoordinate;
var
  I: integer;
begin
    Result.U := 0;
    Result.V := 0;

    for i := 0 to Count - 1 do
    begin
      Result.U := Result.U + uvCoords[i].U;
      Result.V := Result.V + uvCoords[i].V;
    end;

    if Count > 0 then
    begin
      Result.U := Result.U / Count;
      Result.V := Result.V / Count;
    end;
end;

// Convenience method: creates a TUVCoord with the given properties and adds it to the list
function TUVList.Add(const aU, aV: double): integer;
var uvCoord: TUVCoord;
begin
    uvCoord := TUVCoord.create(aU, aV);
    result := Add(uvCoord);
end;

procedure TUVList.Insert(const anIndex: integer; const aU, aV: double);
var uvCoord: TUVCoord;
begin
    uvCoord := TUVCoord.create(aU, aV);
    Insert(anIndex, uvCoord);
end;

function TUVList.Copy: TUVList;
begin
    result := TUVList.Create;
    result.Assign(self);
end;

function TUVList.GetUV(anIndex: integer): TUVCoord;
begin
    if (anIndex>=0) and (anIndex<Count) then
        Result := TUVCoord(inherited Items[anIndex])
    else
        Result := nil;
end;

function TUVList.Copy(index: integer): TUVCoord;
begin
    if (index>=0) and (index<Count) then
        Result := TUVCoord.Create(TUVCoord(Items[index]).u,TUVCoord(Items[index]).v)
    else
        Result := nil;
end;

function TUVList.InList(uvcoord: TUVCoord): integer;
begin
    Result := -1;
end;

procedure TUVList.Reverse;
var i: integer;
    temp: TUVList;
    uvcoord: TUVCoord;
begin
    temp := TUVList.Create;
    try
        for i := 0 to Count-1 do begin                                          // store in temp list
            uvcoord := Get(i);
            temp.Add(uvcoord.Copy);
        end;

        Clear;                                                                  // clear list

        for i := temp.Count-1 downto 0 do begin                                 // add temp to list in reverse order
            uvcoord := temp.Get(i);
            Add(uvcoord.Copy);
        end;
    finally
        FreeAndNil(temp);
    end;
end;

function TUVList.Delete(index: integer): boolean;
var uvcoord: TUVCoord;
begin
    if (index>=0) and (index<Count) then begin
        uvcoord := Get(index);
        Remove(uvcoord);
        Result := true;
    end else
        Result := false;
end;

// ### TPlane ##################################################################

constructor TPlane.Create;
begin
    fNormal := TVertice.Create(0,0,0);
    fDistance := 0;
end;

constructor TPlane.Create(v,n: TVertice);
begin
    fNormal := n.copy; //TVertice.Create(n.x,n.y,n.z);
    fDistance := -dotproduct(v,n);
end;

constructor TPlane.Create(v0, v1, v2: TVertice);
begin
    fNormal := plane_normal(v0,v1,v2);
    fDistance := -dotproduct(v0,fNormal);
end;

destructor TPlane.Destroy;
begin
    FreeAndNil(fNormal);
    inherited;
end;

function TPlane.Equals(plane: TPlane): boolean;
begin
    Result := (normal.x=plane.Normal.x) and
              (normal.y=plane.Normal.y) and
              (normal.z=plane.Normal.z) and
              (Distance=plane.Distance);
end;





end.

