{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_UndoRedoManager;

{$MODE Delphi}

interface

uses
  Classes,
  unit_ToolStack,
  unit_BaseTool;

type
  TUndoRedoManagerEx = class(TObject)
  private
    fUndoToolStack: TToolStack;
    fRedoToolStack: TToolStack;
    fOnChange: TNotifyEvent;
    function GetUndoCount: integer;
    function GetUndoDescription: string;
    function GetRedoCount: integer;
    function GetRedoDescription: string;
  public
    constructor Create;
    destructor Destroy; override;
    procedure AddToolForUndo(ABaseTool: TBaseTool);
    procedure AddToolForRedo(ABaseTool: TBaseTool);
    procedure Undo;
    procedure Redo;
    procedure Clear;
    property UndoCount: integer read GetUndoCount;
    property UndoDescription: string read GetUndoDescription;
    property RedoCount: integer read GetRedoCount;
    property RedoDescription: string read GetRedoDescription;
    property OnChange: TNotifyEvent read fOnChange write fOnChange;
  end;

implementation

uses
  SysUtils, Dialogs;

// ### TUndoRedoManagerEx ######################################################

constructor TUndoRedoManagerEx.Create;
begin
    fUndoToolStack := TToolStack.Create(true,true);
    fRedoToolStack := TToolStack.Create(true,true);
end;

destructor TUndoRedoManagerEx.Destroy;
begin
    FreeAndNil(fUndoToolStack);
    FreeAndNil(fRedoToolStack);
    inherited Destroy;
end;

function TUndoRedoManagerEx.GetUndoCount: integer;
begin
    Result := fUndoToolStack.Count;
end;

function TUndoRedoManagerEx.GetUndoDescription: string;
var description: string;
begin
    if fUndoToolStack.Count>0 then
        description := 'Undo ' + TBaseTool(fUndoToolStack.Last).Hint
    else
        description := 'Nothing to undo';   
    Result := description;
end;

function TUndoRedoManagerEx.GetRedoCount: integer;
begin
    Result := fRedoToolStack.Count;
end;

function TUndoRedoManagerEx.GetRedoDescription: string;
var description: string;
begin
    if fRedoToolStack.Count>0 then
        description := 'Redo ' + TBaseTool(fRedoToolStack.Last).Hint
    else
        description := 'Nothing to redo';
    Result := description;
end;

procedure TUndoRedoManagerEx.AddToolForUndo(ABaseTool: TBaseTool);
begin
    // if the provided tool isn't already on the UndoToolStack, push it on
    if (fUndoToolStack.Count=0) or (fUndoToolStack.Last <> ABaseTool) then begin
        fRedoToolStack.Clear;
        fUndoToolStack.Push(ABaseTool);
        if Assigned(fOnChange) then
            fOnChange(nil);
    end;
end;

procedure TUndoRedoManagerEx.AddToolForRedo(ABaseTool: TBaseTool);
begin
    // if the provided tool isn't already on the UndoToolStack, push it on
    if (fRedoToolStack.Count=0) or (fRedoToolStack.Last <> ABaseTool) then begin
        fRedoToolStack.Push(ABaseTool);
        if Assigned(fOnChange) then
            fOnChange(nil);
    end;
end;

procedure TUndoRedoManagerEx.Undo;
var baseTool: TBaseTool;
begin
    baseTool := fUndoToolStack.Pop;
    baseTool.Undo;
    AddToolForRedo(baseTool);
    if Assigned(fOnChange) then
        fOnChange(nil);
end;

procedure TUndoRedoManagerEx.Redo;
var baseTool: TBaseTool;
begin
    baseTool := fRedoToolStack.Pop;
    baseTool.Redo;
    AddToolForUndo(baseTool);
    if Assigned(fOnChange) then
        fOnChange(nil);
end;

procedure TUndoRedoManagerEx.Clear;
begin
    fUndoToolStack.Clear;
    fRedoToolStack.Clear;
    if Assigned(fOnChange) then
        fOnChange(nil);
end;



end.
 
