{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_VertexBevellerTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TVertexBeveller = class(TPrimitiveTool)
  private
    fEPS: double;
    fDefaultMaterial: TMaterial;
    procedure PolygonScaleAlongEdge(primitive: TBasePrimitive;
                                    polygon: TPolygon;
                                    numOfUnits: integer;
                                    var continueScale: boolean;
                                    performEdgeCheck: boolean = true);
    procedure DoBevel;
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses
  unit_PolygonAlongEdgeScalerTool;

procedure TVertexBeveller.PolygonScaleAlongEdge(primitive: TBasePrimitive; polygon: TPolygon; numOfUnits: integer; var continueScale: boolean; performEdgeCheck: boolean = true);
var i: integer;
    currentVertex,edgeVertex,dirVector: TVertice;
    deltaVectors,edgeVectors: TVerticeList;
    edgeList,adjacentEdges: TEdgeList;
    rollBack: boolean;
begin
    with primitive do begin

        edgeList := TEdgeList.Create(false,true);
        adjacentEdges := TEdgeList.Create(false,true);
        deltaVectors := TVerticeList.Create(true,true);
        edgeVectors := TVerticeList.Create(true,true);

        try
            // get adjacent edges in a list + make a copy of the edge directionvectors in a seperate list
            Polygon_GetAdjacentEdges(polygon,adjacentEdges);
            for i := 0 to adjacentEdges.Count-1 do
                edgeVectors.Add(adjacentEdges[i].DirectionVector.Copy);

            // we use the edge that is NOT part of the current polygon as direction vector
            for i := 0 to polygon.NumVertices-1 do begin

                currentVertex := polygon.VerticeList[i];

                // find out which adjacent edge(s) of the current vertex are NOT
                // part of the polygon; we use that one as direction vector
                currentVertex.EdgeList.IsNotIn(polygon.EdgeList,edgeList);

                // we found exactly one adjacent edge that isn't part of the current polygon
                if edgeList.Count=1 then begin

                    // get the other vertex of the edge in question
                    edgeVertex := edgeList[0].GetOtherVertex(currentVertex);

                    // create a normalized direction vector
                    dirVector := TVertice.Create(0,0,0);
                    dirVector.x := currentVertex.x-edgeVertex.x;
                    dirVector.y := currentVertex.y-edgeVertex.y;
                    dirVector.z := currentVertex.z-edgeVertex.z;
                    dirVector.Normalize;

                    // multiply with # of units to move to get the final direction vector
                    dirVector.MultiplyByScalar(numOfUnits);

                    // and save it for later use
                    deltaVectors.Add(dirVector);
                end;
            end;

            // perform scale
            for i := 0 to polygon.NumVertices-1 do begin
                currentVertex := polygon.VerticeList[i];
                currentVertex.Add(deltaVectors[i]);
            end;

            // perform edge-check if wanted: checks to see if any edge flipped direction
            // and rolls back the scale operation if that is the case
            if performEdgeCheck then begin
                rollBack := false;
                for i := 0 to adjacentEdges.Count-1 do begin
                    if DotProduct(edgeVectors[i],adjacentEdges[i].DirectionVector) < fEPS then begin // dot of original and changed edge
                        rollBack := true;
                        break;
                    end;
                end;

                // rollback if needed
                if rollBack then begin
                    continueScale := false;
                    for i := 0 to polygon.NumVertices-1 do begin
                        currentVertex := polygon.VerticeList[i];
                        currentVertex.Subtract(deltaVectors[i]);
                    end;
                end;
            end;


        finally
            FreeAndNil(deltaVectors);
            FreeAndNil(edgeVectors);
            FreeAndNil(adjacentEdges);
            FreeAndNil(edgeList);
        end;

        HasChanged := true;
    end;
end;

procedure TVertexBeveller.DoBevel;
const minimalEdgeLength = 3;
var i,j: integer;
    vertex,NV,vertexNormal: TVertice;
    edge: TEdge;
    polygon: TPolygon;
    orderedEdges: TEdgeList;
    newSelection: TVerticeList;
    doBevel,doScale: boolean;
begin
    orderedEdges := TEdgeList.Create(false,true);
    newSelection := TVerticeList.Create(false,true);

    for i := 0 to SelectedPrimitives.Count-1 do begin

        newSelection.Clear;

        with SelectedPrimitives[i] do begin

            if ModifyGeometry then begin

                while SelectedVertices.Count>0 do begin

                    vertex := SelectedVertices[0];

                    // we only bevel the vertex if it has more than 2 adjacent edges
                    if vertex.EdgeList.Count>2 then begin

                        // check if each adjacent edge has a minimal length
                        doBevel := true;
                        for j := 0 to vertex.EdgeList.Count-1 do begin
                            if vertex.EdgeList[j].Length <= minimalEdgeLength then begin
                                doBevel := false;
                                break;
                            end;
                        end;

                        // only bevel if allowed
                        if doBevel then begin

                            // get an ordered edgelist of the current vertex in (anti) clockwise order
                            vertex.GetOrderedEdges(orderedEdges, false);

                            // if the ordered list is empty, one or more of the edges might be shared
                            // by more than 2 polygons; we quit if that is the case

                            // TODO: even if we can bevel, check if the length of the edges is sufficient enough

                            if orderedEdges.Count>0 then begin

                                // create a new polygon
                                polygon := TPolygon.Create;
                                polygon.Material := fDefaultMaterial;

                                // loop through all edges that use the current selected vertex,
                                // calculate new vertices + add them to polygons + update current edge
                                for j := 0 to orderedEdges.Count-1 do begin
                                    edge := orderedEdges[j];

                                    // create a new vertex by copying the original one;
                                    // it will be scaled by the PolygonScaleAlongEdgeFunction during mousemove on 2D+3D forms
                                    NV := vertex.Copy;

                                    // add new vertex to primitive verticelist
                                    Vertices.Add(NV);

                                    // and add it to the new selection
                                    newSelection.Add(NV);

                                    // insert the new vertex in the current edge
                                    Edge_InsertVertex(edge,NV);

                                    // add it to our new polygon, which will create a new edge also
                                    Polygon_AddVertex(polygon,NV);
                                end;

                                // store new polygon
                                Polygon_Add(polygon);

                                // get the normal of this vertex
                                vertexNormal := vertex.GetNormal();

                                // now delete original vertex; NOTE: it is important to this here and not after ScaleAlongEdge
                                Vertex_DeleteVertex(vertex);

                                // the polygon is degenerate at the moment (because all its vertices are at one spot)
                                // so we scale its vertices along edges just a bit
                                doScale := false;    // dummy placeholder
                                PolygonScaleAlongEdge(SelectedPrimitives[i],polygon,-1,doScale,false);

                                // new polygon should point in the same direction as the original vertex;
                                // if this is not the case, flip the polygon
                                if DotProduct(vertexNormal,polygon.Normal) < -fEPS then
                                    polygon.Flip;
                                FreeAndNil(vertexNormal);

                                // and scale back by the same amount
                                //Primitive_PolygonScaleAlongEdge(SelectedPrimitives[i],polygon,1,doScale);

                                // polygon is immediately selected;
                                SelectedPolygons.Add(polygon);
                            end;
                        end;
                    end;

                    // remove vertex from selection, even if this vertex wasn't bevelled
                    SelectedVertices.Remove(vertex);
                end;

                SelectedVertices.Assign(newSelection);

                HasChanged := true;
            end;
        end;
    end;

    FreeAndNil(newSelection);
    FreeAndNil(orderedEdges);
end;

procedure TVertexBeveller.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Bevel vertices';
    Description := 'Bevels vertices';

    fEPS := TDoubleObject(AParamList[0]).Value;
    fDefaultMaterial := TMaterial(AParamList[1]);

    DoBevel;
end;

procedure TVertexBeveller.DoExecute(AParamList: TParamList);
var i,j: integer;
    primitive: TBasePrimitive;
    numberOfUnits: integer;
    continueScale: boolean;
begin
    numberOfUnits := TIntegerObject(AParamList.NamedParameters['numberOfUnits']).Value;
    continueScale := true;

    for i := 0 to SelectedPrimitives.Count-1 do begin
        primitive := SelectedPrimitives[i];
        if primitive.ModifyGeometry then begin
            for j := 0 to primitive.SelectedPolygons.Count-1 do
                TPolygonAlongEdgeScalerFunction.Execute(primitive,primitive.SelectedPolygons[j],numberOfUnits,continueScale);
            primitive.HasChanged := true;
        end;
    end;
end;


end.

