{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
// this class is basicly a wrapper around a PolygonSplitter object;
// it determines which vertices are selected + which polygons use those selected
// vertices and then splits each affected polygon by the selected vertices that
// are part of that particular polygon by feeding that info to a PolygonSplitter object
// note: this means that we only connect vertices that are part of the same polygon

unit unit_VertexConnectorTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TVertexConnectByVertexListFunction = class(TObject)
  private
    procedure DoConnect(APrimitive: TBasePrimitive; AVertexList: TVerticeList);
  public
    class procedure Execute(APrimitive: TBasePrimitive; AVertexList: TVerticeList);
  end;

  TVertexConnectByPolygonsAndVerticesFunction = class(TObject)
  private
    procedure DoConnect(APrimitive: TBasePrimitive; APolygonList: TPolygonList; AVertexList: TVerticeList);
  public
    class procedure Execute(APrimitive: TBasePrimitive; APolygonList: TPolygonList; AVertexList: TVerticeList);
  end;

  TVertexConnector = class(TPrimitiveTool)
  private
    procedure DoConnect;
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses unit_PolygonByVerticesSplitterTool;

// ### TVertexConnectByVertexListFunction ######################################

// works with a single primitive and vertexlist containing selected vertices within that primitive
procedure TVertexConnectByVertexListFunction.DoConnect(APrimitive: TBasePrimitive; AVertexList: TVerticeList);
var usedVertices: TVerticeList;
    i,j: integer;
    polygonToBeSplit: TPolygon;
    polygonList,newPolygons,deletePolygons: TPolygonList;
    paramList: TParamList;
begin
    usedVertices := TVerticeList.Create(false,true);
    polygonList := TPolygonList.Create(false,true);
    newPolygons := TPolygonList.Create(false);
    deletePolygons := TPolygonList.Create(false);
    paramList := TParamList.Create(false);

    with APrimitive do begin
        if ModifyGeometry then begin
             // get all polygons that use selected vertices
            AVertexList.GetAdjacent(polygonList);

            // loop through the affected polygons, connecting vertices
            for i := 0 to polygonList.Count-1 do begin

                // determine which of the selected vertices are used by the current polygon
                // and returns them in the usedVertices list ordered by index
                Polygon_GetUsedVertices(polygonList[i],AVertexList,usedVertices);

                // if more than 1 vertex is being used, split the polygon using
                // the selected vertices that are part of this polygon
                if usedVertices.Count>1 then begin

                    polygonToBeSplit := polygonList[i];

                    for j := 0 to usedVertices.Count-1 do begin

                        newPolygons.Clear;

                        // fill a parameter list and feed it to a PolygonSplitter command
                        paramList.Add(APrimitive, 'primitive', true, false);
                        paramList.Add(polygonToBeSplit, 'polygonToBeSplit');
                        paramList.Add(usedVertices[j], 'vertex1');
                        paramList.Add(usedVertices[(j+1) mod usedVertices.Count], 'vertex2');
                        paramList.Add(newPolygons, 'newPolygons');
                        TPolygonByVerticesSplitterFunction.Execute(paramList);

                        // if new polygons were created, mark the original polygon to be deleted
                        if newPolygons.Count>0 then begin
                            deletePolygons.Add(polygonToBeSplit);
                            polygonToBeSplit := newPolygons[1];
                        end;
                    end;
                end;
            end;

            // delete original polygons
            for i := 0 to deletePolygons.Count-1 do
                Polygon_Remove(deletePolygons[i]);

            HasChanged := true;
        end;
    end;

    // free stuff
    FreeAndNil(paramList);
    FreeAndNil(usedVertices);
    FreeAndNil(polygonList);
    FreeAndNil(newPolygons);
    FreeAndNil(deletePolygons);
end;

class procedure TVertexConnectByVertexListFunction.Execute(APrimitive: TBasePrimitive; AVertexList: TVerticeList);
begin
    with self.Create do
    try
        DoConnect(APrimitive, AVertexList);
    finally
        Free;
    end;
end;

// ### TVertexConnectByPolygonsAndVerticesFunction #############################

// works with a single primitive, a polygonlist containing affected polygons
// and vertexlist containing selected vertices within that primitive
procedure TVertexConnectByPolygonsAndVerticesFunction.DoConnect(APrimitive: TBasePrimitive; APolygonList: TPolygonList; AVertexList: TVerticeList);
var usedVertices: TVerticeList;
    i,j: integer;
    polygonToBeSplit: TPolygon;
    newPolygons,deletePolygons: TPolygonList;
    paramList: TParamList;
begin
    usedVertices := TVerticeList.Create(false,true);
    newPolygons := TPolygonList.Create(false);
    deletePolygons := TPolygonList.Create(false);
    paramList := TParamList.Create(false);

    with APrimitive do begin
        if ModifyGeometry then begin
            // loop through the affected polygons, connecting vertices
            for i := 0 to APolygonList.Count-1 do begin

                // determine which of the selected vertices are used by the current polygon
                // and returns them in the usedVertices list ordered by index
                Polygon_GetUsedVertices(APolygonList[i],AVertexList,usedVertices);

                // if more than 1 vertex is being used, split the polygon using
                // the selected vertices that are part of this polygon
                if usedVertices.Count>1 then begin

                    polygonToBeSplit := APolygonList[i];

                    for j := 0 to usedVertices.Count-1 do begin

                        newPolygons.Clear;

                        // fill a parameter list and feed it to a PolygonSplitter command
                        paramList.Add(APrimitive, 'primitive', true, false);
                        paramList.Add(polygonToBeSplit, 'polygonToBeSplit');
                        paramList.Add(usedVertices[j], 'vertex1');
                        paramList.Add(usedVertices[(j+1) mod usedVertices.Count], 'vertex2');
                        paramList.Add(newPolygons, 'newPolygons');

                        TPolygonByVerticesSplitterFunction.Execute(paramList);

                        // if new polygons were created, mark the original polygon to be deleted
                        if newPolygons.Count>0 then begin
                            deletePolygons.Add(polygonToBeSplit);
                            polygonToBeSplit := newPolygons[1];
                        end;
                    end;
                end;
            end;

            // delete original polygons
            for i := 0 to deletePolygons.Count-1 do
                Polygon_Remove(deletePolygons[i]);

            HasChanged := true;
        end;
    end;

    // free stuff
    FreeAndNil(paramList);
    FreeAndNil(usedVertices);
    FreeAndNil(newPolygons);
    FreeAndNil(deletePolygons);
end;

class procedure TVertexConnectByPolygonsAndVerticesFunction.Execute(APrimitive: TBasePrimitive; APolygonList: TPolygonList; AVertexList: TVerticeList);
begin
    with self.Create do
    try
        DoConnect(APrimitive, APolygonList, AVertexList);
    finally
        Free;
    end;
end;

// ### TVertexConnector ########################################################

// standard connect method; works with all selected primitives
// and selected vertices within each selected primitive
procedure TVertexConnector.DoConnect;
var usedVertices: TVerticeList;
    i,j,k: integer;
    polygonToBeSplit: TPolygon;
    primitive: TBasePrimitive;
    polygonList,newPolygons,deletePolygons: TPolygonList;
    paramList: TParamList;
begin
    usedVertices := TVerticeList.Create(false,true);
    polygonList := TPolygonList.Create(false,true);
    newPolygons := TPolygonList.Create(false);
    deletePolygons := TPolygonList.Create(false);
    paramList := TParamList.Create(false);

    for i := 0 to SelectedPrimitives.Count-1 do begin
        usedVertices.Clear;
        polygonList.Clear;
        newPolygons.Clear;
        deletePolygons.Clear;

        primitive := SelectedPrimitives[i];
        with primitive do begin
            if ModifyGeometry then begin
                 // get all polygons that use selected vertices
                SelectedVertices.GetAdjacent(polygonList);

                // loop through the affected polygons, connecting vertices
                for j := 0 to polygonList.Count-1 do begin

                    // determine which of the selected vertices are used by the current polygon
                    // and returns them in the usedVertices list ordered by index
                    Polygon_GetUsedVertices(polygonList[j],SelectedVertices,usedVertices);

                    // if more than 1 vertex is being used, split the polygon using
                    // the selected vertices that are part of this polygon
                    if usedVertices.Count>1 then begin

                        polygonToBeSplit := polygonList[j];

                        for k := 0 to usedVertices.Count-1 do begin

                            newPolygons.Clear;

                            // fill a parameter list and feed it to a PolygonSplitter command
                            paramList.Add(primitive, 'primitive', true, false);
                            paramList.Add(polygonToBeSplit, 'polygonToBeSplit');
                            paramList.Add(usedVertices[k], 'vertex1');
                            paramList.Add(usedVertices[(k+1) mod usedVertices.Count], 'vertex2');
                            paramList.Add(newPolygons, 'newPolygons');
                            TPolygonByVerticesSplitterFunction.Execute(paramList);

                            // if new polygons were created, mark the original polygon to be deleted
                            if newPolygons.Count>0 then begin
                                deletePolygons.Add(polygonToBeSplit);
                                polygonToBeSplit := newPolygons[1];
                            end;
                        end;
                    end;
                end;

                // delete original polygons
                for j := 0 to deletePolygons.Count-1 do
                    Polygon_Remove(deletePolygons[j]);

                HasChanged := true;
            end;
        end;
    end;

    // free stuff
    FreeAndNil(paramList);
    FreeAndNil(usedVertices);
    FreeAndNil(polygonList);
    FreeAndNil(newPolygons);
    FreeAndNil(deletePolygons);
end;

procedure TVertexConnector.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Connect vertices';
    Description := 'Connect vertices';
end;

procedure TVertexConnector.DoExecute(AParamList: TParamList);
begin
    // based on provided parameters, we execute one method or the other
    // note: the interface to using these commands should be very well documented some day
    
    // if we want different methods to be executed based on provided parameters,
    // the first parameter in the list should provide info about how to execute
    // the command
    if not Assigned(AParamList) then
        DoConnect
    else begin
        if AParamList.Count=2 then
            TVertexConnectByVertexListFunction.Execute(TBasePrimitive(AParamList[0]),TVerticeList(AParamList[1]));
        if AParamList.Count=3 then
            TVertexConnectByPolygonsAndVerticesFunction.Execute(TBasePrimitive(AParamList[0]),TPolygonList(AParamList[1]),TVerticeList(AParamList[2]));
    end;

end;

end.

