{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_VertexFillerTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_Materials,
  unit_PrimitiveTool;

type
  TVertexFillerFunction = class(TObject)
  private
    function MustFlip(APolygonList: TPolygonList; APolygon: TPolygon): boolean; overload;
    function MustFlip(APrimitive: TBasePrimitive; AVertexList: TVerticeList; APolygon: TPolygon): boolean; overload;
    procedure DoVertexFill(APrimitive: TBasePrimitive; AMaterial: TMaterial);
  public
    class procedure Execute(APrimitive: TBasePrimitive; AMaterial: TMaterial);
  end;

  TVertexFiller = class(TPrimitiveTool)
  private
    procedure DoFill(AMaterial: TMaterial);
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses unit_Lights, unit_BaseTool;

// ### TVertexFillerFunction ###################################################

procedure TVertexFillerFunction.DoVertexFill(APrimitive: TBasePrimitive; AMaterial: TMaterial);
var sortedVertices,possibleStartVertices: TVerticeList;
    vertex,otherVertex: TVertice;
    edge: TEdge;
    continueSort,vertexFound: boolean;
    j,k,counter,startVertexCounter,numOfSelectedVertices,numOfEdges: integer;
    polygon: TPolygon;
    adjacentCoPlanarPolygons: TPolygonList;
begin
    sortedVertices := TVerticeList.Create(false,true);
    possibleStartVertices := TVerticeList.Create(false,true);

    try
        with APrimitive do begin

            if not (APrimitive is TLight) and (ModifyGeometry) and (SelectedVertices.Count>0) then begin

                // find a suitable starting vertex; a vertex can be used as starting
                // vertex if it shares an edge with only 1 other vertex in the selection;
                // if all vertices in the selection have 2 or more neighboring vertices
                // that are part of the selection, an arbitrary vertex can be used as
                // starting vertex
                for j := 0 to SelectedVertices.Count-1 do begin

                    vertex := SelectedVertices[j];

                    // count how many vertices in the selection are adjacent to this
                    // vertex; if we only find one vertex, this can be our starting vertex;
                    // if we find more than one, continue search for a starting vertex
                    startVertexCounter := 0;
                    for k := 0 to vertex.EdgeList.Count-1 do begin

                        edge := vertex.EdgeList[k];
                        if edge.StartVertex=vertex then
                            otherVertex := edge.EndVertex
                        else
                            otherVertex := edge.StartVertex;

                        if SelectedVertices.IsInList(otherVertex) then
                            inc(startVertexCounter);
                    end;

                    if startVertexCounter=0 then
                        exit;

                    if startVertexCounter=1 then
                        possibleStartVertices.Add(vertex);
                end;

                // now sort the selected vertices starting with our starting vertex;
                // if no starting vertex was determined (because all selected vertices
                // are connecting to eachother by edges), use the first selected vertex;
                // we need to sort the vertices (using edges) because a polygon has to
                // have its vertices in the correct order; this means we can fill a region
                // that has at most one edge missing; for example, the edge between the
                // last and the first vertex in the ordered selection
                sortedVertices.Clear;
                numOfSelectedVertices := SelectedVertices.Count;    // remember how many vertices are selected
                if possibleStartVertices.Count>0 then begin
                    sortedVertices.Add(possibleStartVertices[0]);
                    SelectedVertices.Remove(possibleStartVertices[0]);
                    vertex := possibleStartVertices[0];
                end else begin
                    sortedVertices.Add(SelectedVertices[0]);
                    SelectedVertices.Remove(SelectedVertices[0]);
                    vertex := sortedVertices[0];
                end;

                continueSort := true;
                while (SelectedVertices.Count>0) and (continueSort) do begin

                    counter := 0;
                    numOfEdges := vertex.EdgeList.Count;
                    vertexFound := false;
                    while (counter<numOfEdges) and (not vertexFound) do begin
                        edge := vertex.EdgeList[counter];
                        if edge.StartVertex=vertex then
                            otherVertex := edge.EndVertex
                        else
                            otherVertex := edge.StartVertex;

                        if not SelectedVertices.IsInList(otherVertex) then
                            inc(counter)
                        else begin
                            vertexFound := true;
                            sortedVertices.Add(otherVertex);
                            SelectedVertices.Remove(otherVertex);

                            // now use opposite vertex of the edge as current vertex and continue search
                            vertex := otherVertex;
                        end;
                    end;
                    continueSort := counter<numOfEdges;
                end;

                // create a polygon if we have enough vertices
                if sortedVertices.Count<>numOfSelectedVertices then
                    raise Exception.Create('Unable to fill region. Selected vertices need to be connected by existing edges. Only one edge may be missing at most.')
                else begin

                    // check if a polygon containing the same vertices does not already exist
                    polygon := Polygon_Exists(sortedVertices);
                    if Assigned(polygon) then
                        raise Exception.Create('Similar polygon already exists. You might need to flip that polygon to face in the right direction.')
                    else begin
                        // create the new polygon
                        polygon := TPolygon.Create;
                        polygon.Material := AMaterial;
                        for j := 0 to sortedVertices.Count-1 do
                            Polygon_AddVertex(polygon,sortedVertices[j]);
                        Polygon_Add(polygon);

                        // flip if needed; we check with adjacent coplanar polygons
                        // first and if we can't dertermine a flip based on that,
                        // we use all adajacent polygons and their center vertex to
                        // see if we need to flip
                        adjacentCoPlanarPolygons := TPolygonList.Create(false,true);
                        try
                            Polygon_GetAdjacentCoplanarPolygons(polygon,adjacentCoPlanarPolygons,false);
                            if adjacentCoPlanarPolygons.Count>0 then begin
                                if MustFlip(adjacentCoPlanarPolygons,polygon) then
                                    polygon.Flip;
                            end else begin
                                if MustFlip(APrimitive,sortedVertices,polygon) then
                                    polygon.Flip;
                            end;
                        finally
                            FreeAndNil(adjacentCoPlanarPolygons);
                        end;

                        // signal change
                        HasChanged := true;
                    end;
                end;
                SelectedVertices.Clear;
            end;
        end;

    finally
        FreeAndNil(possibleStartVertices);
        FreeAndNil(sortedVertices);
    end;
end;

class procedure TVertexFillerFunction.Execute(APrimitive: TBasePrimitive; AMaterial: TMaterial);
begin
    with self.Create do
    try
        DoVertexFill(APrimitive, AMaterial);
    finally
        Free;
    end;
end;

function TVertexFillerFunction.MustFlip(APolygonList: TPolygonList; APolygon: TPolygon): boolean;
var i,numSameDirection,numOppositeDirection: integer;
    polygon: TPolygon;
    mustFlipPolygon: boolean;
begin
    numSameDirection := 0;
    numOppositeDirection := 0;
    if APolygonList.Count=0 then
        mustFlipPolygon := false
    else begin
        for i := 0 to APolygonList.Count-1 do begin
            polygon := APolygonList[i];
            if polygon.IsEqualNormal(APolygon) then
                inc(numSameDirection)
            else
                inc(numOppositeDirection);
        end;
        mustFlipPolygon := numOppositeDirection>numSameDirection;
    end;
    Result := mustFlipPolygon;
end;

function TVertexFillerFunction.MustFlip(APrimitive: TBasePrimitive; AVertexList: TVerticeList; APolygon: TPolygon): boolean;
var i,numPointsTowards,numPointsAWay: integer;
    averageVertex: TVertice;
    adjacentPolygons: TPolygonList;
    vertexList: TVerticeList;
    polygon: TPolygon;
    mustFlipPolygon: boolean;
begin
    // get a list of polygons adjacent to the supplied vertexlist;
    // do not take the polygon in question into account here
    adjacentPolygons := TPolygonList.Create(false,true);
    AVertexList.GetAdjacent(adjacentPolygons);
    adjacentPolygons.Remove(APolygon);

    // determine which vertices are used by these polygons and
    // save the average vertex of that list of vertices
    vertexList := TVerticeList.Create(false,true);
    APrimitive.Polygon_GetAdjacentVertices(adjacentPolygons,vertexList);
    averageVertex := vertexList.Center;

    // now check how many of these polygons point towards or away
    // from the average vertex
    numPointsTowards := 0;
    numPointsAWay := 0;
    for i := 0 to adjacentPolygons.Count-1 do begin
        polygon := adjacentPolygons[i];
        if (polygon.DistanceToOrigin > planeD(averageVertex,polygon.Normal)) then
            inc(numPointsTowards)
        else
            inc(numPointsAWay);

         {
        if (polygon.Normal.x*averageVertex.x +
            polygon.Normal.y*averageVertex.y +
            polygon.Normal.z*averageVertex.z +
            -dotproduct(averageVertex,polygon.Normal)) < 0 then
            inc(numPointsTowards)
        else
            inc(numPointsAWay);  }
    end;
    FreeAndNil(vertexList);
    FreeAndNil(adjacentPolygons);

    // we want our polygon in question to point towards the direction
    // in which most adjacent polygons point; so, if 3 out of 5 polygons
    // point towards the average vertex and our polygon in question is
    // pointing away from it, we need to flip it
    if (APolygon.DistanceToOrigin > planeD(averageVertex,APolygon.Normal)) then
        mustFlipPolygon := numPointsAWay>numPointsTowards
    else
        mustFlipPolygon := numPointsTowards>numPointsAWay;
 {
    if (APolygon.Normal.x*averageVertex.x +
        APolygon.Normal.y*averageVertex.y +
        APolygon.Normal.z*averageVertex.z +
        -dotproduct(averageVertex,APolygon.Normal)) < 0 then  // APolygon points towards average vertex
        mustFlipPolygon := numPointsAWay>numPointsTowards
    else      // APolygon points away from average vertex
        mustFlipPolygon := numPointsTowards>numPointsAWay;    }
    Result := mustFlipPolygon;
end;

// ### TVertexFiller ###########################################################

procedure TVertexFiller.DoFill(AMaterial: TMaterial);
var i: integer;
    primitive: TBasePrimitive;
begin
    for i := 0 to SelectedPrimitives.Count-1 do begin
        primitive := SelectedPrimitives[i];
        TVertexFillerFunction.Execute(primitive, AMaterial);
    end;
end;

procedure TVertexFiller.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Vertex fill';
    Description := 'Fills a region specified by selected vertices';
end;

procedure TVertexFiller.DoExecute(AParamList: TParamList);
var defaultMaterial: TMaterial;
begin
    defaultMaterial := TMaterial(AParamList[0]);
    if Assigned(defaultMaterial) then
        DoFill(defaultMaterial);
end;



end.

