{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_VertexInserterTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_Primitives,
  unit_ParamList,
  unit_PrimitiveTool;

type
  TVertexInserter = class(TPrimitiveTool)
  private
    function GetEdge(ARay: TRayObject; out APrimitive: TBasePrimitive; out AIntersectionPoint: TVertice): TEdge;
    procedure DoInsert(AParamList: TParamList);
  protected
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

uses unit_VertexConnectorTool;

function TVertexInserter.GetEdge(ARay: TRayObject; out APrimitive: TBasePrimitive; out AIntersectionPoint: TVertice): TEdge;
var i: integer;
    closestDistance, distance: single;
    closestPolygon, polygon: TPolygon;
    edge,selectedEdge: TEdge;
    edgeDistance,projectionLength: single;
    tempVertex,P,interpolatedVertex: TVertice;
begin
    closestDistance := -1;
    closestPolygon := nil;
    selectedEdge := nil;

    // first determine if a polygon was hit
    for i := 0 to SelectedPrimitives.Count-1 do begin
        APrimitive := SelectedPrimitives[i];
        if APrimitive.Polygon_Select(ARay.Value, distance, polygon) and ((closestDistance = -1) or (distance < closestDistance)) then begin
            closestPolygon := polygon;
            closestDistance := distance;
        end;
    end;

    // if a polygon was hit, find out which edge is closest to the intersection
    // between the ray and the polygon within a predefined threshold
    if Assigned(closestPolygon) then begin
        // calculate intersection point
        AIntersectionPoint := TVertice.Create(aRay.Value.Origin.X,aRay.Value.Origin.Y,aRay.Value.Origin.Z);
        tempVertex := TVertice.Create(aRay.Value.Direction.X,aRay.Value.Direction.Y,aRay.Value.Direction.Z);
        try
            tempVertex.MultiplyByScalar(closestDistance);
            AIntersectionPoint.Add(tempVertex);
        finally
            FreeAndNil(tempVertex);
        end;

        // check to which edge the intersection is closest to
        distance := 8;       // TODO: calculate "smart" value for this, based on camera distance etc
        for i := 0 to closestPolygon.EdgeList.Count-1 do begin
            edge := closestPolygon.EdgeList[i];
            edgeDistance := AIntersectionPoint.DistanceToLine(edge.StartVertex,edge.EndVertex);
            if edgeDistance<=distance then begin
                if AIntersectionPoint.BetweenPoints(edge.StartVertex,edge.EndVertex) then
                    selectedEdge := edge;
                distance := edgeDistance;
            end;
        end;

        // if we did not find an edge, free the intersection point
        if not Assigned(selectedEdge) then
            FreeAndNil(AIntersectionPoint)
        else begin
            // else adjust the intersection point so that it is _exactly_ on the edge

            // create a vector P using the starting point of the edge and the found intersection point
            P := vectorSub(selectedEdge.StartVertex,AIntersectionPoint);

            // figure out the length of the projection of P onto the selected edge (page 19 Lengyel)
            projectionLength := dotproduct(P,selectedEdge.DirectionVector)/selectedEdge.Length;

            // interpolate from the starting point of the edge
            interpolatedVertex := selectedEdge.StartVertex.Interpolate(selectedEdge.EndVertex,projectionLength);

            // assign the values to the intersection point
            AIntersectionPoint.Assign(interpolatedVertex);

            // and free stuff
            FreeAndNil(P);
            FreeAndNil(interpolatedVertex);
        end;
    end;
    Result := selectedEdge;
end;

procedure TVertexInserter.DoInsert(AParamList: TParamList);
var ray: TRayObject;
    autoConnect: boolean;
    edge: TEdge;
    edgeList: TEdgeList;
    primitive: TBasePrimitive;
    intersectionPoint: TVertice;
    isEqual: boolean;
begin
    ray := TRayObject(AParamList[0]);
    autoConnect := TBooleanObject(AParamList[1]).Value;

    intersectionPoint := nil;
    edge := GetEdge(ray,primitive,intersectionPoint);
    if Assigned(edge) then begin
        // if we get an edge, both the primitive and intersectionpoint have been assigned
        isEqual := false;

        // note: the "5" is arbitrary chosen; calculate "smart" value for this, based on camera distance etc
        if intersectionPoint.Equals(edge.StartVertex,5) then begin
            isEqual := true;
            FreeAndNil(intersectionPoint);
            intersectionPoint := edge.StartVertex;
        end else if intersectionPoint.Equals(edge.EndVertex,5) then begin
            isEqual := true;
            FreeAndNil(intersectionPoint);
            intersectionPoint := edge.EndVertex;
        end;

        if not isEqual then begin
            edgeList := TEdgeList.Create(false,true);
            primitive.Vertices.Add(intersectionPoint);
            primitive.Edge_InsertVertex(edge,intersectionPoint,edgeList);
            FreeAndNil(edgeList);
        end;

        // auto-connect vertices if wanted
        if autoConnect then begin
            primitive.SelectedVertices.Add(intersectionPoint);
            if primitive.SelectedVertices.Count=2 then begin
                TVertexConnectByVertexListFunction.Execute(primitive, primitive.SelectedVertices);
                primitive.SelectedVertices.Clear;
                primitive.SelectedVertices.Add(intersectionPoint);
            end;
        end;


    end;
end;

procedure TVertexInserter.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Insert vertex';
    Description := 'Insert a vertex by clicking on an edge in a 3D view.';
    if AParamList.Count=2 then
        DoInsert(AParamList);
end;

procedure TVertexInserter.DoExecute(AParamList: TParamList);
begin
    // empty by design
end;

end.

