{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_VertexMagnetTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_PrimitiveTool;

type
  TVertexMagnet = class(TPrimitiveTool)
  private
    fDelta: TVertice;
    fRange: integer;
    fFinalWorkload: TVerticeList;
    fDeltaPercentages: TIntegerList;
    fMultipleVerticesSelected: boolean;
    function GetAdjacentVertices(AOrigin: TVertice; ARange: integer; ACurrentVertices,AExcludedVertices: TVerticeList): TVerticeList;
  protected
    function AdjustDeltaFractionHill(ADeltaFraction: double): double;
    function AdjustDeltaFractionSmooth(ADeltaFraction: double): double;
    function AdjustDeltaFractionSharp(ADeltaFraction: double): double;
    function AdjustDeltaFractionLinear(ADeltaFraction: double): double;
  public
    constructor Create(AToolAction: TToolAction);
    destructor Destroy; override;
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

// ### TVertexMagnet ###########################################################

constructor TVertexMagnet.Create(AToolAction: TToolAction);
begin
    fFinalWorkload := TVerticeList.Create(false,true);
    fDeltaPercentages := TIntegerList.Create;
    inherited Create(AToolAction);
end;

destructor TVertexMagnet.Destroy;
begin
    FreeAndNil(fFinalWorkload);
    FreeAndNil(fDeltaPercentages);
    inherited Destroy;
end;

function TVertexMagnet.AdjustDeltaFractionHill(ADeltaFraction: double): double;
begin
    Result := 1-((3-2 * ADeltaFraction) * Sqr(ADeltaFraction));
end;

function TVertexMagnet.AdjustDeltaFractionSmooth(ADeltaFraction: double): double;
begin
    Result := 1-Sqr(ADeltaFraction);
end;

function TVertexMagnet.AdjustDeltaFractionSharp(ADeltaFraction: double): double;
begin
    Result := 1-Sqrt(ADeltaFraction);
end;

function TVertexMagnet.AdjustDeltaFractionLinear(ADeltaFraction: double): double;
begin
    Result := 1-ADeltaFraction;
end;

function TVertexMagnet.GetAdjacentVertices(AOrigin: TVertice; ARange: integer; ACurrentVertices,AExcludedVertices: TVerticeList): TVerticeList;
var adjacentVertices: TVerticeList;
    i,j: integer;
    vertex,lengthVertex: TVertice;
    edge: TEdge;
    vertexList: TVerticeList;
begin
    adjacentVertices := TVerticeList.Create(false,true);
    vertexList := TVerticeList.Create(false,true);

    for i := 0 to ACurrentVertices.Count-1 do begin
        vertex := ACurrentVertices[i];
        for j := 0 to vertex.EdgeList.Count-1 do begin
            edge := vertex.EdgeList[j];
            if (not AExcludedVertices.IsInList(edge.StartVertex)) then begin
                //if (not ACurrentVertices.IsInList(edge.StartVertex)) then
                    vertexList.Add(edge.StartVertex);
            end;

            if (not AExcludedVertices.IsInList(edge.EndVertex)) then begin
                //if (not ACurrentVertices.IsInList(edge.EndVertex)) then
                    vertexList.Add(edge.EndVertex);
            end;
        end;
    end;

    // check if the found vertices are within range
    for i := 0 to vertexList.Count-1 do begin
        lengthVertex := vectorSub(AOrigin, vertexList[i]);
        if vectorLength(lengthVertex)<ARange then
            adjacentVertices.Add(vertexList[i]);
        FreeAndNil(lengthVertex);
    end;

    FreeAndNil(vertexList);
    Result := adjacentVertices;
end;

procedure TVertexMagnet.DoPrepare(AParamList: TParamList);
var i,j,k: integer;
    adjacentVertices,currentVertices: TVerticeList;
    currentDistance,deltaPercentage: double;
    vertex,currentToOrigin: TVertice;
    primitive: TBasePrimitive;
begin
    Hint := 'Vertex magnet';
    Description := 'Moves vertices within a certain range to a certain movement vector';

    // clear data
    fDeltaPercentages.Clear;
    fFinalWorkload.Clear;

    fRange := TIntegerObject(AParamList.NamedParameters['magnetRange']).Value;
    fMultipleVerticesSelected := false;

    for i := 0 to SelectedPrimitives.Count-1 do begin
        primitive := SelectedPrimitives[i];
        if primitive.SelectedVertices.Count>1 then begin
            fMultipleVerticesSelected := true;
            break;
        end else begin
            currentVertices := TVerticeList.Create(false,true);

            for j := 0 to primitive.SelectedVertices.Count-1 do begin
                fFinalWorkload.Add(primitive.SelectedVertices[j]);
                fDeltaPercentages.Add(0);
            end;

            for j := 0 to primitive.SelectedVertices.Count-1 do begin
                vertex := primitive.SelectedVertices[j];

                currentVertices.Add(vertex);

                while currentVertices.Count > 0 do begin
                    adjacentVertices := GetAdjacentVertices(vertex, fRange, currentVertices, fFinalWorkload);

                    if adjacentVertices.Count = 0 then
                        currentVertices.Clear
                    else begin
                        currentVertices.Assign(adjacentVertices);

                        for k := 0 to currentVertices.Count-1 do begin
                            currentToOrigin := vectorSub(vertex,currentVertices[k]);
                            currentDistance := vectorLength(currentToOrigin);
                            deltaPercentage := (currentDistance/fRange) * 100;
                            fDeltaPercentages.Add(Round(deltaPercentage));
                            FreeAndNil(currentToOrigin);
                        end;

                        fFinalWorkload.AddList(currentVertices);
                    end;

                    FreeAndNil(adjacentVertices);
                end;
            end;
            FreeAndNil(currentVertices);
        end;
    end;

    // quick hack: show message when one or more primitives have more than 1 vertex selected
    // TODO: use proper Exception Handling; atm it seems handling the exception does not produce correct results (memory leaks)
    if fMultipleVerticesSelected then begin
        fDeltaPercentages.Clear;
        fFinalWorkload.Clear;
        //raise Exception.Create('Vertex Magnet does not support multiple selected vertices per primitive.');
        MessageDlg('Vertex Magnet does not support multiple selected vertices per primitive.',mtInformation,[mbOK],0);
    end;
end;

procedure TVertexMagnet.DoExecute(AParamList: TParamList);
var i: integer;
    deltaFraction: double;
    vertex,tempDelta: TVertice;
    magnetType: TMagnetType;
begin
    if not fMultipleVerticesSelected then begin
        tempDelta := TVertice.Create;
        fDelta := TVertice(AParamList.NamedParameters['delta']);
        for i := 0 to fFinalWorkload.Count-1 do begin
            vertex := fFinalWorkload[i];
            tempDelta.Assign(fDelta);
            deltaFraction := fDeltaPercentages.Values[i] / 100;
            magnetType := TMagnetType(TIntegerObject(AParamList.NamedParameters['magnetType']).Value);
            case magnetType of
                mtHill:
                    deltaFraction := AdjustDeltaFractionHill(deltaFraction);
                mtSmooth:
                    deltaFraction := AdjustDeltaFractionSmooth(deltaFraction);
                mtSharp:
                    deltaFraction := AdjustDeltaFractionSharp(deltaFraction);
                mtLinear:
                    deltaFraction := AdjustDeltaFractionLinear(deltaFraction);
            end;

            tempDelta.x := tempDelta.x * deltaFraction;
            tempDelta.y := tempDelta.y * deltaFraction;
            tempDelta.z := tempDelta.z * deltaFraction;
            vertex.Subtract(tempDelta);
        end;
        FreeAndNil(tempDelta);
    end;
end;
(*
// ### THillVertexMagnet #######################################################

function THillVertexMagnet.AdjustDeltaFraction(ADeltaFraction: double): double;
begin
    Result := 1-((3-2 * ADeltaFraction) * Sqr(ADeltaFraction));
end;

// ### TSmoothVertexMagnet #####################################################

function TSmoothVertexMagnet.AdjustDeltaFraction(ADeltaFraction: double): double;
begin
    Result := 1-Sqr(ADeltaFraction);
end;

// ### TSharpVertexMagnet ######################################################

function TSharpVertexMagnet.AdjustDeltaFraction(ADeltaFraction: double): double;
begin
    Result := 1-Sqrt(ADeltaFraction);
end;

// ### TLinearVertexMagnet #####################################################

function TLinearVertexMagnet.AdjustDeltaFraction(ADeltaFraction: double): double;
begin
    Result := 1-ADeltaFraction;
end; *)

end.
 
