{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_VertexRemoverTool;

{$MODE Delphi}

interface

uses
  Windows, SysUtils, Classes, Contnrs, StrUtils, Dialogs,
  unit_Types,
  unit_Enums,
  unit_Math,
  unit_IntegerList,
  unit_Primitives,
  unit_ParamList,
  unit_PrimitiveTool;

type
  TVertexOnEdgeRemoveFunction = class(TObject)
  private
    procedure DoRemove(APrimitive: TBasePrimitive);
  public
    class procedure Execute(APrimitive: TBasePrimitive);
  end;

  TVertexRemoveFunction = class(TObject)
  private
    procedure DoRemove(APrimitive: TBasePrimitive);
  public
    class procedure Execute(APrimitive: TBasePrimitive);
  end;

  TVertexRemover = class(TPrimitiveTool)
  private
  public
    procedure DoPrepare(AParamList: TParamList); override;
    procedure DoExecute(AParamList: TParamList); override;
  end;

implementation

// ### TVertexOnEdgeRemoveFunction #############################################

procedure TVertexOnEdgeRemoveFunction.DoRemove(APrimitive: TBasePrimitive);
var i,index,tempIndex: integer;
    polygon: TPolygon;
    adjacentPolygons: TPolygonList;
    vertexList: TVerticeList;
    vertex0,vertex1,vertex2: TVertice;
    isOnEdge: boolean;
begin
    with APrimitive do begin

        adjacentPolygons := TPolygonList.Create(false,true);
        vertexList := TVerticeList.Create(false,true);
        try
            SelectedVertices.Clear;

            // walk through each polygon, examining edges in search of vertices that can be removed
            // if 2 successive edges are a straight line (the angle between the two edges is 180;
            // the middle-vertex is on the edge formed by vertex 0 and vertex 2), then the vertex
            // used by both edges (the middle-vertex) can be removed
            for i := 0 to Polygons.Count-1 do begin
                polygon := Polygons[i];
                vertexList.Assign(polygon.VerticeList);

                index := 0;
                while index<vertexList.Count-1 do begin
                    tempIndex := index-1;
                    if tempIndex<0 then tempIndex := vertexList.Count-1;
                    vertex0 := vertexList[tempIndex];
                    vertex1 := vertexList[index];
                    vertex2 := vertexList[(index+1) mod vertexList.Count];

                    isOnEdge := vertex1.OnEdge(vertex0 ,vertex2);

                    // successive edges should form a straight line
                    if not isOnEdge then
                        inc(index)
                    else begin
                        // the middle-vertex should be used by 2 edges only
                       if vertex1.EdgeList.Count>2 then
                            inc(index)
                        else begin
                            SelectedVertices.AddVertex(vertex1);
                            vertexList.Remove(vertex1);
                        end;
                    end;
                end;
            end;

            // now remove the selected vertices from the primitive
            TVertexRemoveFunction.Execute(APrimitive);
        finally
            FreeAndNil(adjacentPolygons);
            FreeAndNil(vertexList);
        end;
    end;
end;

class procedure TVertexOnEdgeRemoveFunction.Execute(APrimitive: TBasePrimitive);
begin
    with self.Create do
    try
        DoRemove(APrimitive);
    finally
        Free;
    end;
end;

// ### TVertexRemoveFunction ###################################################

procedure TVertexRemoveFunction.DoRemove(APrimitive: TBasePrimitive);
var i: integer;
begin
    with APrimitive do begin
        // delete all selected vertices from the current primitive; this
        // updates the entire datastructure of the primitive
        for i := 0 to SelectedVertices.Count-1 do begin
            if Vertices.IsInList(SelectedVertices[i]) then
                Vertex_DeleteVertex(SelectedVertices[i]);
        end;

        HasChanged := true;
    end;
end;

class procedure TVertexRemoveFunction.Execute(APrimitive: TBasePrimitive);
begin
    with self.Create do
    try
        DoRemove(APrimitive);
    finally
        Free;
    end;
end;

// ### TVertexRemover ##########################################################

procedure TVertexRemover.DoPrepare(AParamList: TParamList);
begin
    Hint := 'Remove vertex';
    Description := 'Removes vertices from primitives';    
end; 

procedure TVertexRemover.DoExecute(AParamList: TParamList);
var index: integer;
    primitive: TBasePrimitive;
begin
    try
        index := 0;
        while index<SelectedPrimitives.Count do begin
            primitive := SelectedPrimitives[index];
            TVertexRemoveFunction.Execute(primitive);

            // if the primitive has no more polygons left, remove it itself;
            // else clear SelectedVertices list of the primitive and continue
            // with the next primitive
            if primitive.Polygons.Count>0 then begin
                primitive.SelectedVertices.Clear;
                primitive.HasChanged := true;
                inc(index);
            end else begin
                ScenePrimitives.Remove(primitive);
                SelectedPrimitives.Delete(index);
            end;
        end;
    except
        on E: Exception do MessageDlg('TVertexRemover.Execute: '+E.Message,mtError,[mbOK],0);
    end;
    
end;



end.
