{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_VertexSelector;

{$MODE Delphi}

interface

uses
  Classes,
  unit_Selector,
  unit_Types,
  unit_Primitives,
  unit_PrimitiveList,
  unit_Paths;

type
  // This class selects vertices from primitives.
  //
  // NOTES:
  //  - when a vertex is selected, it is stored in two collections:
  //      - the SelectedVertices of the primitive it belongs to
  //      - the SelectedVertices of this class
  //    this means the SelectedVertices collection of this class can
  //    hold verticesof multiple primitives
  TVertexSelector = class(TSelector)
  private
    fSelectedVertices: TVerticeList;
  protected
    function GetCenter: TVertice; override;
    procedure SelectVertexByRay(X, Y: integer; const ARay: TRay; const AModelViewMatrix, AProjectionMatrix: TMatrix4x4; const AViewport: TViewport; const AAddToSelection: boolean; const AClearSelection: boolean);
  public
    constructor Create(APrimitiveList: TPrimitiveList; ASelectedVerticeList: TVerticeList);

    procedure SelectByRange(AUpperLeftVertex, ALowerRightVertex: TVertice; AClearSelection: boolean = false);
    procedure SelectByVertex(AVertex: TVertice; ARange: double; AAddToSelection: boolean; AClearSelection: boolean = false); reintroduce; overload;
    procedure SelectByRay(ARay: TRay; AClearSelection: boolean = false); overload;
    procedure SelectByRay(X, Y: integer; const ARay: TRay; const AModelViewMatrix, AProjectionMatrix: TMatrix4x4; const AViewport: TViewport; const AAddToSelection: boolean; const AClearSelection: boolean); overload;
    procedure SelectByRectangle(const ax1,ay1,ax2,ay2: integer;
                                const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                const aViewport: TViewport;
                                const AClearSelection: boolean);
    procedure SelectAll;
    procedure DeselectAll;
    procedure ReverseSelection;
    procedure GrowSelection;
    procedure ShrinkSelection;

    property SelectedVertices: TVerticeList read fSelectedVertices write fSelectedVertices;
  end;

implementation

uses
  SysUtils,
  unit_Variables,
  unit_Math;

// ### TVertexSelector #########################################################

constructor TVertexSelector.Create(APrimitiveList: TPrimitiveList; ASelectedVerticeList: TVerticeList);
begin
    fPrimitives := APrimitiveList;
    fSelectedVertices := ASelectedVerticeList;
    inherited Create;
end;

function TVertexSelector.GetCenter: TVertice;
var i,j,counter: integer;
    primitive: TBasePrimitive;
begin
    counter := 0;
    fCenter.Assign(0,0,0);
    for i := 0 to Primitives.Count-1 do begin
        primitive := Primitives[i];
        for j := 0 to primitive.SelectedVertices.Count-1 do begin
            fCenter.Add(primitive.SelectedVertices[j]);
            inc(counter);
        end;
    end;
    if counter>0 then
        fCenter.DivideByScalar(counter)
    else
        fCenter.ResetToInfinite;
    Result := fCenter;
end;

//modify  (3d point click)
procedure TVertexSelector.SelectVertexByRay(X, Y: integer; const ARay: TRay; const AModelViewMatrix, AProjectionMatrix: TMatrix4x4; const AViewport: TViewport; const AAddToSelection: boolean; const AClearSelection: boolean);
var i, j: integer;
    primitive, hitPrimitive: TBasePrimitive;
    vertex, hitVertex: TVertice;
    point: TPoint2D;
    pointHalfSize, distance, hitDistance: single;
    Path: TPath;
    hitControlPoint: TPathControlPoint;
begin
    // INFO, PJ: First we find the closest vertex in the selection that we actually hit
    pointHalfSize := defaultPointSize / 2;
    hitDistance := -1;
    hitVertex := nil;
    hitPrimitive := nil;
    hitControlPoint := nil;
    for i := 0 to Primitives.Count-1 do begin
        primitive := Primitives[i];

        //Are we dealing with a path or a primitive

        if Primitive is TPath then
        begin
          //Do the computation with Controlpoints instaed of vertices
          Path := TPath(Primitive);

          if AClearSelection then
            Path.SelectedControlPoints.Clear();

          vertex := TVertice.Create();
          try
            for j := 0 to Path.ControlPointCount - 1 do
            begin
                vertex.x := Path.ControlPoints[j].Position.x;
                vertex.y := Path.ControlPoints[j].Position.y;
                vertex.z := Path.ControlPoints[j].Position.z;

                point := vertex.ProjectToScreen(AModelViewMatrix, AProjectionMatrix, AViewPort);

                if (X >= point.X - pointHalfSize) and (X <= point.X + pointHalfSize) and
                   (Y >= point.Y - pointHalfSize) and (Y <= point.Y + pointHalfSize) then begin
                    distance := VertexSquaredLength( VertexSubtract( ARay.Origin, VerticeToVertex( vertex )));
                    if (hitDistance = -1) or (distance < hitDistance) then begin
                        hitPrimitive := primitive;
                        hitControlPoint := Path.ControlPoints[J];
                        hitDistance := distance;
                    end;
                end;
            end;
          finally
            Vertex.Free();
          end;
        end
        else
        begin
          //The code for normal primitives

          if AClearSelection then
            primitive.SelectedVertices.Clear;
          primitive.HighlightedVertices.Clear;

          for j := 0 to primitive.NumVertices - 1 do
          begin
              vertex := primitive.Vertices[j];
              point := vertex.ProjectToScreen(AModelViewMatrix, AProjectionMatrix, AViewPort);

              if (X >= point.X - pointHalfSize) and (X <= point.X + pointHalfSize) and
                 (Y >= point.Y - pointHalfSize) and (Y <= point.Y + pointHalfSize) then begin
                  distance := VertexSquaredLength( VertexSubtract( ARay.Origin, VerticeToVertex( vertex )));
                  if (hitDistance = -1) or (distance < hitDistance) then begin
                      hitPrimitive := primitive;
                      hitVertex := vertex;
                      hitDistance := distance;
                  end;
              end;
          end;
        end;
    end;

    // INFO, PJ: Then we check to see if that particular vertex is occluded by any polygons.
    if Assigned(hitVertex) then
    begin
        if (not Assigned(SelectPolygonByRay(aRay, primitive, distance, FALSE, FALSE))) or
           ( distance > Sqrt(hitDistance) - pointHalfSize - 1 ) then
        begin
             if AAddToSelection then
               hitPrimitive.Vertex_AddSelected(hitVertex)
             else
               hitPrimitive.HighlightedVertices.Add(hitVertex);
        end;
    end;

    //In this case, a path was clicked
    if Assigned(hitControlPoint) then
    begin
        if (not Assigned(SelectPolygonByRay(aRay, primitive, distance, FALSE, FALSE))) or
           ( distance > Sqrt(hitDistance) - pointHalfSize - 1 ) then
        begin
             if AAddToSelection then
               TPath(hitPrimitive).SelectedControlPoints.Add( hitControlPoint );
        end;
    end;
end;

procedure TVertexSelector.SelectByRange(AUpperLeftVertex, ALowerRightVertex: TVertice; AClearSelection: boolean);
var i: integer;
begin
    if AClearSelection then
        SelectedVertices.Clear;

    for i := 0 to Primitives.Count-1 do
    begin
        if Primitives[i] is TPath then
          //If it's a path, we want to select its controlpoints instead
          TPath(Primitives[i]).SelectControlPoints(AUpperLeftVertex,ALowerRightVertex, AClearSelection)
        else
        begin
          if AClearSelection then
            Primitives[i].SelectedVertices.Clear;

          Primitives[i].Vertex_SelectMultiple(AUpperLeftVertex,ALowerRightVertex);
          SelectedVertices.AddList(Primitives[i].SelectedVertices);
        end;
    end;

    if Assigned(OnSelect) then
        OnSelect(SelectedVertices);
end;

procedure TVertexSelector.SelectByVertex(AVertex: TVertice; ARange: double; AAddToSelection: boolean; AClearSelection: boolean);
var i: integer;
    primitive: TBasePrimitive;
    vertex: TVertice;
begin
    if AClearSelection then
        SelectedVertices.Clear;

    for i := 0 to Primitives.Count-1 do begin
        primitive := Primitives[i];
        primitive.HighlightedVertices.Clear;

        if primitive is TPath then
        begin
          //Select controlpoints here
          //Highlighting not supported...
          if AAddToSelection then
            TPath(primitive).SelectControlPoint(AVertex, ARange, AClearSelection);
        end
        else
        begin
          if AClearSelection then
            primitive.SelectedVertices.Clear;

          //Vertex selection behaviour for normal primitives
          vertex := primitive.Vertex_SelectByPoint(AVertex, ARange);
          if Assigned(vertex) then
          begin
              if AAddToSelection then
                  primitive.Vertex_Select(vertex)
              else
                  primitive.HighlightedVertices.Add(vertex);
          end;
          SelectedVertices.AddList(Primitives[i].SelectedVertices);
        end;

    end;

    if Assigned(OnSelect) then
        OnSelect(SelectedVertices);
end;

procedure TVertexSelector.SelectByRay(ARay: TRay; AClearSelection: boolean);
begin
    raise Exception.Create('TVertexSelector.SelectByRay(ARay, AClearSelection) not implemented.');
end;

procedure TVertexSelector.SelectByRay(X, Y: integer; const ARay: TRay; const AModelViewMatrix, AProjectionMatrix: TMatrix4x4; const AViewport: TViewport; const AAddToSelection: boolean; const AClearSelection: boolean);
begin
    SelectVertexByRay(X, Y, ARay, AModelViewMatrix, AProjectionMatrix, AViewPort, AAddToSelection, AClearSelection);
end;

procedure TVertexSelector.SelectByRectangle(const ax1,ay1,ax2,ay2: integer;
                                            const aModelViewMatrix, aProjectionMatrix: TMatrix4x4;
                                            const aViewport: TViewport;
                                            const AClearSelection: boolean);
var
    primitive: TBasePrimitive;
    polygon: TPolygon;
    vertex: TVertice;
    point: TPoint2D;
    i,j,k: integer;
    PolygonPoints: TPoint2DArray;
    vertexList: TVerticeList;
begin
    if AClearSelection then
        SelectedVertices.Clear;

    vertexList := TVerticeList.Create(false,true);
    for i := 0 to Primitives.Count - 1 do begin
        primitive := Primitives[i];

        if primitive is TPath then
        begin
          //Let a path method handle the selection process
          TPath(Primitive).SelectControlPoints(ax1,ay1,ax2,ay2,
                                               aModelViewMatrix,
                                               aProjectionMatrix,
                                               aViewport,
                                               AClearSelection);
        end
        else
        begin
          if AClearSelection then
            primitive.SelectedVertices.Clear;

          vertexList.Clear;
          for j := 0 to primitive.NumPolygons - 1 do
          begin
              polygon := primitive.Polygons[j];
              polygon.ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewport,PolygonPoints);
              if Polygon2DFacesViewer(PolygonPoints) then
              begin
                  for k := 0 to polygon.VerticeList.Count-1 do
                  begin
                      vertex := polygon.VerticeList[k];
                      point := vertex.ProjectToScreen(aModelViewMatrix,aProjectionMatrix,aViewPort);
                      if (point.X >= ax1) and (point.X <= ax2) and (point.Y >= ay1) and (point.Y <= ay2) then
                          vertexList.Add(vertex);
                  end;
              end;
          end;

          for j := 0 to vertexList.Count-1 do
          begin
              vertex := vertexList[j];
              if primitive.SelectedVertices.IndexOf(vertex)=-1 then
                  primitive.SelectedVertices.Add(vertex)
              else
                  primitive.SelectedVertices.Remove(vertex);
          end;
        end;
    end;

    FreeAndNil(vertexList);
end;

procedure TVertexSelector.SelectAll;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Vertex_SelectAll;
end;

procedure TVertexSelector.DeselectAll;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Vertex_DeselectAll;
end;

procedure TVertexSelector.ReverseSelection;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Vertex_ReverseSelection;
end;

procedure TVertexSelector.GrowSelection;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Vertex_GrowSelection;
end;

procedure TVertexSelector.ShrinkSelection;
var i: integer;
begin
    for i := 0 to Primitives.Count-1 do
        Primitives[i].Vertex_ShrinkSelection;
end;



end.
 
