{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_ViewPortManager;

{$MODE Delphi}

interface

uses
  Contnrs,Controls,Dialogs,Classes,SysUtils,Forms, FileUtil,
  unit_Enums,
  unit_PrimitiveList,
  unit_Scene,
  unit_Options,
  unit_Types,
  unit_Cameras,
  frmOGLContextForm,
  frmViewPortForm,
  frmFlatViewPortForm,
  frmTopView,
  frmFrontView,
  frmSideView,
  frm3dView,
  frmUVEditorForm;

type
  // Each viewport individually remembers camera positions for all available view modes.
  TViewPortStore = class(TObject)
  private
    fCameras: Array[TViewMode] of TCamera;
    fViewPort: TfrmViewPort;
    procedure SetViewport(const AViewport: TFrmViewPort);
  public
    destructor destroy; override;
    property ViewPort: TfrmViewPort read fViewPort write SetViewPort;
  end;

  TViewPortManager = class(TObject)
  private
    fViewportStores: TObjectList;
    fOGLContext: TfrmOGLContext;
    fLoadedFromIniFile: boolean;
    fParentForm: TForm;
    function GetViewPort(AnIndex: integer): TfrmViewPort;
    function GetActiveViewPort: TfrmViewPort;
    function GetActiveViewportStore: TViewportStore;
    function GetActive3DViewPort: Tfrm3D;
    function InstantiateViewPort(AViewMode: TViewMode; AScene: TScene; ADimensions: TDimensions; ARenderOptions: TRenderOptions; ACameraOptions: TCameraOptions = nil; AControlOptions: TControlOptions = nil): TfrmViewPort; overload;
    procedure SetupViewPort(ASettingStrings: TStringList; AScene: TScene; ARenderOptions: TRenderOptions; ACameraOptions: TCameraOptions; AControlOptions: TControlOptions);
    function GetHowManyViewPorts: integer;
    property ActiveViewportStore: TViewportStore read GetActiveViewportStore;
  public
    constructor Create( const aParentForm: TForm );
    destructor Destroy; override;
    procedure ResetCamera;
    procedure ZoomCamera(ADelta: integer);
    procedure FocusCameraOnPrimitives(APrimitiveList: TPrimitiveList);
    procedure UpdateSkyBox;
    procedure UpdateFOV;
    procedure SetPerspectiveDrawModes(ADrawMode: TDrawMode);
    procedure CreateViewPorts(AIniFileName: string; AScene: TScene; ARenderOptions: TRenderOptions; ACameraOptions: TCameraOptions; AControlOptions: TControlOptions);
    procedure SaveViewPorts(AIniFileName: string);
    function AddViewPort(AViewMode: TViewMode; AScene: TScene; ARenderOptions: TRenderOptions;
      ACameraOptions: TCameraOptions = nil; AControlOptions: TControlOptions = nil): TfrmViewPort;
    function CreateViewPort(AViewMode: TViewMode; AScene: TScene; ARenderOptions: TRenderOptions;
      ACameraOptions: TCameraOptions = nil; AControlOptions: TControlOptions = nil): TfrmViewPort;
    procedure ChangeActiveViewPort(AViewMode: TViewMode; AScene: TScene; ARenderOptions: TRenderOptions;
      ACameraOptions: TCameraOptions = nil; AControlOptions: TControlOptions = nil);
    procedure RemoveActiveViewPort;
    //procedure SetDoPaint
    procedure Draw(AImmediatePaint: boolean);
    procedure Tile;
    function IsMaximized: boolean;
    function IsMouseDown: boolean;
    function IsAnyViewportInSpecialState(): Boolean;
    property ViewPorts[AnIndex: integer]: TfrmViewPort read GetViewPort; default;
    property LoadedFromIniFile: boolean read fLoadedFromIniFile;
    property ActiveViewPort: TfrmViewPort read GetActiveViewPort;
    property Active3DViewPort: Tfrm3D read GetActive3DViewPort;
    property OGLContext: TfrmOGLContext read fOGLContext write fOGLContext;
    property Numberof : integer read  GetHowManyViewPorts;
  end;

implementation

uses
    IniFiles, Math, Types, Windows,
    unit_Functions, unit_Tiler;

// ### TViewportStore ########################################################

destructor TViewportStore.destroy;
var viewMode: TViewMode;
begin
    // We release the camera store
    for viewMode := Low(TViewMode) to High(TViewMode) do
        FreeAndNil( fCameras[viewMode] );

    inherited destroy;
end;

procedure TViewportStore.SetViewport( const aViewport: TFrmViewPort );
begin
    // Store camera position of old viewport
    if Assigned(fViewPort) then begin
        FreeAndNil(fCameras[fViewPort.ViewMode]);
        fCameras[fViewPort.ViewMode] := fViewPort.Camera.Clone;
    end;

    fViewPort := aViewPort;

    // Restore camera positino into new viewport.
    if Assigned(fViewPort) then begin
        fViewPort.AssignCamera(fCameras[fViewPort.ViewMode]);
    end;
end;

// ### TViewPortManager ########################################################

constructor TViewPortManager.Create( const aParentForm: TForm );
begin
    inherited Create;

    fParentForm := aParentForm;
    fViewportStores := TObjectList.Create(true);
end;

destructor TViewPortManager.Destroy;
var i: integer;
    store: TViewportStore;
    viewPort: TFrmViewPort;
begin
    for i := fViewPortStores.Count - 1 downto 0 do begin
        // INFO, PJ: Tricky, As freeing a form triggers another update, order is important!
        //           Also note how form.release doesn't work here (message never arrives when app is dying)
        store := fViewPortStores[i] as TViewportStore;
        viewPort := store.ViewPort;
        FreeAndNil( viewPort );
    end;

    FreeAndNil( fViewPortStores );

    inherited Destroy;
end;

function TViewPortManager.GetViewPort(AnIndex: integer): TfrmViewPort;
begin
    if (AnIndex>=0) and (AnIndex<fViewPortStores.Count) then
        Result := TViewportStore(fViewPortStores[AnIndex]).ViewPort
    else
        Result := nil;
end;

function TViewPortManager.GetActiveViewPort: TfrmViewPort;
var store: TViewportStore;
begin
    result := nil;
    store := ActiveViewPortStore;
    if Assigned(store) then result := store.ViewPort;
end;

function TViewPortManager.GetActiveViewPortStore: TViewportStore;
var i: integer;
begin
    Result := nil;
    for i := 0 to fViewPortStores.Count-1 do begin
        if ViewPorts[i].Active then begin
            Result := fViewPortStores[i] as TViewportStore;
            break;
        end;
    end;
end;

function TViewPortManager.GetHowManyViewPorts: integer;
begin
    Result := fViewportStores.Count;
end;

// Returns the active view if it is 3D, otherwise it returns any of the
// 3D views (or nil if there aren't any)
function TViewPortManager.GetActive3DViewPort: Tfrm3D;
var i: integer;
    viewPort: TFrmViewPort;
begin
    Result := nil;
    for i := 0 to fViewPortStores.Count-1 do begin
        viewPort := ViewPorts[i];
        if viewPort is Tfrm3D then begin
            result := viewPort as Tfrm3D;
            if result.Active then
                break;
        end;
    end;
end;

procedure TViewPortManager.ResetCamera;
var i: integer;
begin
    for i := 0 to fViewPortStores.Count-1 do
        ViewPorts[i].ResetCamera;
end;

procedure TViewPortManager.ZoomCamera(ADelta: integer);
var i: integer;
begin
    for i := 0 to fViewPortStores.Count-1 do
        ViewPorts[i].ZoomCamera(ADelta);
end;

procedure TViewPortManager.FocusCameraOnPrimitives(APrimitiveList: TPrimitiveList);
var i: integer;
begin
    for i := 0 to fViewPortStores.Count-1 do
        ViewPorts[i].FocusCameraOnPrimitives(APrimitiveList);
end;

procedure TViewPortManager.UpdateSkyBox;
var i: integer;
begin
    for i := 0 to fViewPortStores.Count-1 do begin
        if ViewPorts[i] is Tfrm3D then begin
            Tfrm3D(ViewPorts[i]).UpdateSkybox;
            Tfrm3D(ViewPorts[i]).SetView;
        end;
    end;
end;

procedure TViewPortManager.UpdateFOV;
var i: integer;
begin
    for i := 0 to fViewPortStores.Count-1 do begin
        if ViewPorts[i] is Tfrm3D then
            Tfrm3D(ViewPorts[i]).SetView;
    end;
end;

procedure TViewPortManager.SetPerspectiveDrawModes(ADrawMode: TDrawMode);
var i: integer;
begin
    for i := 0 to fViewPortStores.Count-1 do begin
        if ViewPorts[i] is Tfrm3D then
            Tfrm3D(ViewPorts[i]).DrawMode := ADrawMode;
    end;
end;

procedure TViewPortManager.SetupViewPort(ASettingStrings: TStringList;
                                         AScene: TScene;
                                         ARenderOptions: TRenderOptions;
                                         ACameraOptions: TCameraOptions;
                                         AControlOptions: TControlOptions);
var frmViewPort: TfrmViewPort;
    viewMode: TViewMode;
begin
    frmViewPort := nil;
    viewMode := TViewMode(StrToInt(ASettingStrings.Strings[0]));
    case viewMode of
        vmTop,vmFront,vmSide,vmUV:
            frmViewPort := AddViewPort(viewMode,AScene,ARenderOptions);
        vmPerspective:
            frmViewPort := AddViewPort(vmPerspective,AScene,ARenderOptions,ACameraOptions,AControlOptions);
    end;
    if Assigned(frmViewPort) then begin
        if ASettingStrings.Count<6 then begin     // 1.72 didn't store windowstate
            frmViewPort.WindowState := wsNormal;
            frmViewPort.Width := StrToInt(ASettingStrings[1]);
            frmViewPort.Height := StrToInt(ASettingStrings[2]);
            frmViewPort.Top := StrToInt(ASettingStrings[3]);
            frmViewPort.Left := StrToInt(ASettingStrings[4]);
        end else begin
            case StrToInt(ASettingStrings[5]) of
                0: frmViewPort.WindowState := wsMinimized;
                1: frmViewPort.WindowState := wsMaximized;
               -1:  begin
                        frmViewPort.WindowState := wsNormal;
                        frmViewPort.Width := StrToInt(ASettingStrings[1]);
                        frmViewPort.Height := StrToInt(ASettingStrings[2]);
                        frmViewPort.Top := StrToInt(ASettingStrings[3]);
                        frmViewPort.Left := StrToInt(ASettingStrings[4]);
                    end;
            end;
            if ASettingStrings.Count = 7 then   // 1.73 and prior versions didn´t store which view is _also_ supported by the view (for switching between views)
                frmViewPort.SupportedViewMode := TViewMode(StrToInt(ASettingStrings[6]));
        end;
    end;
end;

procedure TViewPortManager.CreateViewPorts(AIniFileName: string;
                                           AScene: TScene;
                                           ARenderOptions: TRenderOptions;
                                           ACameraOptions: TCameraOptions;
                                           AControlOptions: TControlOptions);
var iniFile: TIniFile;
    i,numOfViewPorts: integer;
    viewPortInfo: string;
    strValues,maximizedViews: TStringList;
begin
    if not FileExistsUTF8(AIniFileName) then begin      // default views
        AddViewPort(vmTop,AScene,ARenderOptions);
        AddViewPort(vmFront,AScene,ARenderOptions);
        AddViewPort(vmSide,AScene,ARenderOptions);
        AddViewPort(vmPerspective,AScene,ARenderOptions,ACameraOptions,AControlOptions);
        fLoadedFromIniFile := false;
    end else begin
        iniFile := TIniFile.Create(AIniFileName);
        numOfViewPorts := iniFile.ReadInteger('ViewPorts','numOfViewPorts',0);
        if numOfViewPorts=0 then begin
            AddViewPort(vmTop,AScene,ARenderOptions);
            AddViewPort(vmFront,AScene,ARenderOptions);
            AddViewPort(vmSide,AScene,ARenderOptions);
            AddViewPort(vmPerspective,AScene,ARenderOptions,ACameraOptions,AControlOptions);
            fLoadedFromIniFile := false;
        end else begin
            maximizedViews := TStringList.Create;

            // first handle normal and minimized views
            for i := 1 to numOfViewPorts do begin
                viewPortInfo := iniFile.ReadString('ViewPorts','ViewPort'+IntToStr(i),'');
                if viewPortInfo <> '' then begin
                    strValues := SubdivideStringToList(viewPortInfo,',');
                    if strValues.Count<6 then       // 1.72 was the first version to store viewport info in the INI file but it didn't store windowstate
                        SetupViewPort(strValues,AScene,ARenderOptions,ACameraOptions,AControlOptions)
                    else begin
                        if Trim(strValues[5])='-1' then
                            SetupViewPort(strValues,AScene,ARenderOptions,ACameraOptions,AControlOptions)
                        else
                            maximizedViews.Add(viewPortInfo);
                    end;
                    FreeAndNil(strValues);
                end;
            end;

            // now handle maximized views
            if maximizedViews.Count>0 then begin
                for i := 0 to maximizedViews.Count-1 do begin
                    viewPortInfo := maximizedViews.Strings[i];
                    if viewPortInfo <> '' then begin
                        strValues := SubdivideStringToList(viewPortInfo,',');
                        SetupViewPort(strValues,AScene,ARenderOptions,ACameraOptions,AControlOptions);
                        FreeAndNil(strValues);
                    end;
                end;
            end;

            FreeAndNil(maximizedViews);
            fLoadedFromIniFile := true;
        end;
        FreeAndNil(iniFile);
    end;
end;

procedure TViewPortManager.SaveViewPorts(AIniFileName: string);
var iniFile: TIniFile;
    frmViewPort: TfrmViewPort;
    i: integer;
    strValue: string;
begin
    if not FileExistsUTF8(AIniFileName) then Exit;
    iniFile := TIniFile.Create(AIniFileName);
    iniFile.EraseSection('ViewPorts');
    iniFile.WriteInteger('ViewPorts','numOfViewPorts',fViewportStores.Count);
    for i := 0 to fViewportStores.Count-1 do begin
        frmViewPort := ViewPorts[i];
        strValue := IntToStr(Integer(frmViewPort.ViewMode)) + ',';
        strValue := strValue + IntToStr(frmViewPort.Width) + ',';
        strValue := strValue + IntToStr(frmViewPort.Height) + ',';
        strValue := strValue + IntToStr(frmViewPort.Top) + ',';
        strValue := strValue + IntToStr(frmViewPort.Left) + ',';
        case frmViewPort.WindowState of
            wsMinimized:
                strValue := strValue + '0,';
            wsMaximized:
                strValue := strValue + '1,';
            else
                strValue := strValue + '-1,';
        end;
        strValue := strValue + IntToStr(Integer(frmViewPort.SupportedViewMode));
        iniFile.WriteString('ViewPorts','ViewPort'+IntToStr(i+1),strValue);
    end;
    FreeAndNil(iniFile);
end;

function TViewPortManager.AddViewPort(AViewMode: TViewMode; AScene: TScene; ARenderOptions: TRenderOptions;
               ACameraOptions: TCameraOptions = nil; AControlOptions: TControlOptions = nil): TfrmViewPort;
var
    viewportStore: TViewportStore;
begin
    Result := CreateViewPort(AViewMode,AScene,ARenderOptions,ACameraOptions,AControlOptions);
    viewportStore := TViewportStore.Create;
    viewportStore.ViewPort := Result;
    fViewPortStores.Add(viewportStore);
    Result.Activate;
end;

function TViewPortManager.CreateViewPort(AViewMode: TViewMode; AScene: TScene; ARenderOptions: TRenderOptions; ACameraOptions: TCameraOptions = nil; AControlOptions: TControlOptions = nil): TfrmViewPort;
var dimensions: TDimensions;
begin
    // Dummy dimensions, for views that are moved/tiled/whatevered afterwards anyway.
    dimensions.Left := 0;
    dimensions.Top := 0;
    dimensions.Width := 256;
    dimensions.Height := 256;

    Result := InstantiateViewPort(AViewMode, AScene, dimensions, ARenderOptions, ACameraOptions, AControlOptions);
end;

function TViewPortManager.InstantiateViewPort(AViewMode: TViewMode; AScene: TScene; ADimensions: TDimensions; ARenderOptions: TRenderOptions;
               ACameraOptions: TCameraOptions = nil; AControlOptions: TControlOptions = nil): TfrmViewPort;
var frmNewViewPort: TfrmViewPort;
begin
    case AViewMode of
        vmTop:
            frmNewViewPort := TfrmTop.Create(ADimensions, OGLContext);
        vmFront:
            frmNewViewPort := TfrmFront.Create(ADimensions, OGLContext);
        vmSide:
            frmNewViewPort := TfrmSide.Create(ADimensions, OGLContext);
        vmUV:
            frmNewViewPort := TfrmUVEditor.Create(ADimensions, OGLContext);

        else begin
            frmNewViewPort := Tfrm3D.Create(ADimensions, OGLContext);
            Tfrm3D(frmNewViewPort).CameraOptions := ACameraOptions;
            Tfrm3D(frmNewViewPort).ControlOptions := AControlOptions;
        end;
    end;

    frmNewViewPort.Scene := AScene;
    frmNewViewPort.RenderOptions := ARenderOptions;
    Result := frmNewViewPort;
end;

procedure TViewPortManager.ChangeActiveViewPort(AViewMode: TViewMode; AScene: TScene; ARenderOptions: TRenderOptions;
               ACameraOptions: TCameraOptions = nil; AControlOptions: TControlOptions = nil);
var frmActiveViewPort,frmNewViewPort: TfrmViewPort;
    dimensions: TDimensions;
    activeViewportStore: TViewportStore;
begin
    activeViewportStore := GetActiveViewportStore;
    if Assigned(activeViewportStore) then
        frmActiveViewPort := activeViewportStore.Viewport
    else
        frmActiveViewPort := nil;

    // prevent setting the same viewmode twice or changing the viewport when there are none.
    if (not Assigned(frmActiveViewPort)) or
       (frmActiveViewPort.ViewMode = AViewMode) then exit;

    dimensions.Top := frmActiveViewPort.Top;
    dimensions.Left := frmActiveViewPort.Left;
    dimensions.Width := frmActiveViewPort.Width;
    dimensions.Height := frmActiveViewPort.Height;

    frmActiveViewPort.Release;

    frmNewViewPort := InstantiateViewPort(AViewMode,AScene,dimensions,ARenderOptions,ACameraOptions,AControlOptions);

    activeViewportStore.ViewPort := frmNewViewPort;
    frmNewViewPort.Activate;
    // JCC 27-12-2007: Set2DView and Set3DView renamed to SetView and made public
    // in TfrmViewPort and called here; this makes sure the projection matrix is
    // recalculated (which is used to draw the Gizmo, amongst other things...)
    frmNewViewPort.SetView;
end;

procedure TViewPortManager.RemoveActiveViewPort;
var activeViewPortStore: TViewportStore;
begin
    activeViewPortStore := GetActiveViewPortStore;

    if Assigned(activeViewPortStore) then begin
        activeViewPortStore.ViewPort.Release;
        fViewPortStores.Remove(activeViewPortStore);
    end;
end;

procedure TViewPortManager.Draw(AImmediatePaint: boolean);
var i: integer;
begin
    for i := 0 to fViewPortStores.Count-1 do begin
        if AImmediatePaint then
            ViewPorts[i].glDraw()
        else
            ViewPorts[i].DoPaint := true;
    end;
end;

procedure TViewPortManager.Tile;
var tiler: TTiler;
    i, style: integer;
    viewPort: TfrmViewPort;
    tileInfo: TTileInfo;
    rect: TRect;
begin
    // INFO, PJ:
    //   We start by calculating the MDI client rectangle and feeding it to the tiler.
    //   Note how we explicitly disregard any space taken up by currently visible scrollbars.

    GetClientRect( fParentForm.ClientHandle, rect );
    style := getWindowLong( fParentForm.ClientHandle, GWL_STYLE );
    if style and WS_HSCROLL = WS_HSCROLL then
        rect.Bottom := rect.Bottom + GetSystemMetrics( SM_CYHSCROLL );
    if style and WS_VSCROLL = WS_VSCROLL then
        rect.Right := rect.Right + GetSystemMetrics( SM_CXVSCROLL );
    tiler := TTiler.Create( 0, 0, rect.Right, rect.Bottom );

    // INFO, PJ: The tiler is a generic tiler, so we have to convert our viewports to and from generic TileInfos's.
    for i := 0 to fViewPortStores.Count - 1 do begin
        viewPort := ViewPorts[i];
        tiler.AddTileInfo( viewport.Left, viewPort.Top, viewPort.Width, viewPort.Height, viewPort );
    end;

    tiler.Tile;

    for i := 0 to tiler.NumTiles - 1 do begin
        tileInfo := tiler.Tiles[i];
        viewPort := tileInfo.TileObject as TfrmViewPort;
        viewPort.Left := tileInfo.Left;
        viewPort.Top := tileInfo.Top;
        viewPort.Width := tileInfo.Width;
        viewPort.Height := tileInfo.Height;
        // JCC: we need to set the WindowState too, otherwise we might accidentally
        // maximize viewports (http://www.delgine.com/forum/viewtopic.php?t=3708)
        viewPort.WindowState := wsNormal;
    end;

    FreeAndNil( tiler );
end;

// indicates if one of the viewports is maximized
function TViewPortManager.IsMaximized: boolean;
var i: integer;
begin
    Result := false;
    for i := 0 to fViewPortStores.Count-1 do begin
        if ViewPorts[i].WindowState=wsMaximized then begin
            Result := true;
            break;
        end;
    end;
end;

function TViewPortManager.IsMouseDown: boolean;
var i: integer;
begin
    Result := false;
    for i := 0 to fViewPortStores.Count-1 do begin
        if ViewPorts[i].IsMouseDown then begin
            Result := true;
            break;
        end;
    end;
end;

function TViewPortManager.IsAnyViewportInSpecialState: Boolean;
var i: integer;
begin
    Result := false;
    for i := 0 to fViewPortStores.Count-1 do
    begin
        if ViewPorts[i].ViewportState <> vsNone then
        begin
            Result := true;
            break;
        end;
    end;
end;


end.

