{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_XMLDataReader;

{$MODE Delphi}

interface

uses
    frmViewPortForm,
    frmFlatViewPortForm,
    unit_Options,
    unit_ViewPortManager,
    Graphics,
    Classes,
    Dialogs, FileUtil,
    unit_Scene,
    unit_Materials,
    unit_Primitives,
    unit_Types,
    unit_Lights,
    unit_Enums,
    unit_Skeleton,
    unit_Joint,
    unit_Paths,
    unit_JointLocation,
    unit_Animation,
    unit_AnimationFrame,
    unit_XMLSceneConstants,
    unit_XMLReader,
    unit_Vectors,
    frmProgressForm;

type
    // the XMLDataReader is a baseclass for the XMLSceneReader and XMLPluginReader
    // classes; it contains methods to append scene items to a scene, as both the
    // XMLSceneReader and XMLPluginReader classes can do that; the XMLPluginReader
    // can also replace items; that behaviour is implemented in XMLPluginReader only;
    TXMLDataReader = class(TObject)
    private
        fProgressForm: TFrmProgress;
        fWarnings: TStringList;
        fShouldShowProgress: boolean;

        function ReadColor(const aNode: TXMLNode; const aColorType: TColorType): TColor;
        function ReadRGBColor(const aNode: TXMLNode): TRGBColor;
    procedure SizeBackground(var aViewport: TfrmFlatViewPort; const aNode: TXMLNode);

    protected
        function GetWarnings: TStringList;

        procedure UpdateScene(const aScene: TScene);
        procedure UpdateProgressForNode( const aNode: TXMLNode );
        procedure HandleProgress(const aPosition, aTotalSize: integer);

        function LayerTypeForDescription(const aDescription: string): TMaterialLayerType;
        function BlendOperationForDescription(const aDescription: string): TBlendOperation;
        function ObjectTypeForDescription(const aDescription: string): TObjectType;
        function SnapMethodForDescription(const aDescription: string): TSnapMethod;
        function LightClassForDescription(const aDescription: string): TLightClass;
        procedure LoadPrefabForLight(const aLight: TLight; const aScene: TScene);

        procedure ReadScene(const aNode: TXMLNode; const aScene: TScene; aViewPortManager : TViewPortManager = nil);
        procedure ProcessSceneNode(const aNode: TXMLNode; const aScene: TScene); virtual;

        procedure AppendSettings(const aNode: TXMLNode; const aScene: TScene);
        procedure AppendMaterials(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene);
        procedure ProcessPrimitives(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene); virtual; abstract;
        procedure AppendSkeletons(const aNode: TXMLNode; const aScene: TScene);
        procedure AppendLights(const aNode: TXMLNode; const aScene: TScene);
        procedure AppendPaths(const aNode: TXMLNode; const aScene: TScene);

        procedure AppendMaterialCategory(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene);
        procedure ProcessMaterial(const aNode: TXMLNode; const aCategoryName: string; const aMaterialList: TStringList; const aScene: TScene); virtual;
        procedure AppendMaterialLayer(const aNode: TXMLNode; const aMaterial: TMaterial; const aScene: TScene);
        procedure AppendLayerTexture(const aNode: TXMLNode; const aLayer: TTextureLayer; const anIsLightmap: Boolean; const aScene: TScene);
        procedure AppendMaterialRaytracingProperties(const aNode: TXMLNode; const aMaterial: TMaterial);

        procedure AppendPrimitive(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene);
        procedure AppendLightingProperties(const aNode: TXMLNode; const aPrimitive: TPrimitive);
        procedure AppendVertices(const aNode: TXMLNode; const aPrimitive: TBasePrimitive; const aVertices: TStringList);
        procedure AppendVertex(const aNode: TXMLNode; const aPrimitive: TBasePrimitive; const aVertices: TStringList);
        procedure AppendPolygons(const aNode: TXMLNode; const aPrimitive: TBasePrimitive; const aVertices: TStringList; const aMaterialList: TStringList; const aScene: TScene);
        procedure AppendPolygon(const aNode: TXMLNode; const aPrimitive: TBasePrimitive; const aVertices: TStringList; const aMaterialList: TStringList; const aScene: TScene);
        procedure AppendPolygonVertex(const aNode: TXMLNode; const aPolygon: TPolygon; const aVertices: TStringList; const aPrimitive: TBasePrimitive);

        procedure AppendSkeleton(const aNode: TXMLNode; const aScene: TScene);
        procedure AppendJoints(const aNode: TXMLNode; const aSkeleton: TSkeleton; const aScene: TScene);
        procedure AppendJoint(const aNode: TXMLNode; const aSkeleton: TSkeleton; const aParentJoint: TJoint; const aParentJointLocation: TJointLocation; const aScene: TScene);
        procedure AppendAnimations(const aNode: TXMLNode; const aSkeleton: TSkeleton);
        procedure AppendAnimation(const aNode: TXMLNode; const aSkeleton: TSkeleton);
        procedure AppendFrames(const aNode: TXMLNode; const aAnimation: TAnimation);
        procedure AppendFrame(const aNode: TXMLNode; const aAnimation: TAnimation);
        procedure AppendJointLocation(const aNode: TXMLNode; const aAnimationFrame: TAnimationFrame);

        procedure AppendLight(const aNode: TXMLNode; const aScene: TScene);
        procedure AppendLightAttenuation(const aNode: TXMLNode; const aLight: TLight);
        procedure AppendLightPosition(const aNode: TXMLNode; const aLight: TLight);
        procedure AppendLightDirection(const aNode: TXMLNode; const aLight: TLight);
        procedure AppendLightCone(const aNode: TXMLNode; const aLight: TLight);

        procedure AppendPath(const aNode: TXMLNode; const aScene: TScene);
        procedure AppendControlPoints(const aNode: TXMLNode; const aPath: TPath);
        procedure AppendControlPoint(const aNode: TXMLNode; const aPath: TPath);
        procedure ProcessPaths(const aNode: TXMLNode; const aScene: TScene); virtual; abstract;
        procedure AppendViews(const aNode: TXMLNode;  const aScene: TScene; const aViewPortManager: TViewPortManager);
        procedure AppendPerspectiveView(const aNode: TXMLNode;  const aScene: TScene; const aViewPortManager: TViewPortManager);
        procedure AppendFlatView(const aNode: TXMLNode;  const aScene: TScene; const aViewPortManager: TViewPortManager);
        procedure appendBackground(const aNode: TXMLNode; aViewport: TfrmFlatViewPort);
        procedure appendCamera(const aNode: TXMLNode;aViewPort: TfrmViewPort);

        property ProgressForm: TFrmProgress read fProgressForm;
        property Warnings: TStringList read GetWarnings;

    public
        constructor Create(); overload;
        constructor CreateWithoutProgressform(); overload; // chrisu: Quickfix for the raytracer-hang bug.
        destructor Destroy; override;

        property ShouldShowProgress: boolean read fShouldShowProgress write fShouldShowProgress;
    end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
    frm3dView,
    frmUVEditorForm,
    forms,
    frmMainForm,
    SysUtils, Windows, IniFiles,           // INFO, PJ: Added for THashedStringList.. Gotta love the VCL unit consistency ;-)
    unit_DPFLoader,
    unit_Textures,
    unit_Functions,
    unit_Variables,
    unit_TextureLoader,
    unit_MaterialMatcher,
    unit_AnimationBuilder,
    unit_JointBuilder,
    unit_Bone,
    unit_BoneBuilder,
    unit_ParamList;

constructor TXMLDataReader.Create;
begin
  inherited;

  //Create progressform here, instead of using getter function
  fProgressForm := TFrmProgress.Create( nil ); // chrisu: Commenting this out makes raytracing ok again - but DeleD crashes on opening a new scene!
end;

constructor TXMLDataReader.CreateWithoutProgressform; // chrisu: Quickfix for the raytracer-hang bug.
begin
  inherited;

  fShouldShowProgress := false;
end;

destructor TXMLDataReader.Destroy;
begin
    //Chronoz:
    //We previously called fProgressForm.Release here
    //If calling FreeAndNIL( fProgressForm ) has any
    //negative consequences, we must change it back

    FreeAndNil( fProgressForm );
    FreeAndNil( fWarnings );

    inherited destroy;
end;


// ----------------------------------------------------------------------------

procedure TXMLDataReader.ReadScene(const aNode: TXMLNode; const aScene: TScene; aViewPortManager : TViewPortManager = nil);
var i: integer;
    node: TXMLNode;
    materials: THashedStringList;
begin
    materials := THashedStringList.Create;

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'settings' then
        begin
            AppendSettings(node, aScene);
        end
        else if node.Name = 'materials' then
        begin
            AppendMaterials(node, materials, aScene);
        end
        else if node.Name = 'primitives' then
        begin
            ProcessPrimitives(node, materials, aScene);
        end
        else if node.Name = 'skeletons' then
        begin
           AppendSkeletons(node, aScene);
        end
        else if node.Name = 'lights' then
        begin
            AppendLights(node, aScene);
        end
        else if node.Name = 'paths' then
        begin
           AppendPaths(node, aScene);
        end
        else if (node.Name = 'views') and (assigned(aViewPortManager)) then
        begin
           AppendViews(node,ascene,aViewPortManager);
        end;
    end;

    FreeAndNil(materials);

    aScene.UpdateHandlers;

    if Warnings.Count > 0 then
        MessageDlg(PChar(Warnings.Text), mtWarning, [mbOK], 0);
end;


// ----------------------------------------------------------------------------

procedure TXMLDataReader.ProcessSceneNode(const aNode: TXMLNode; const aScene: TScene); 
begin
  // INFO, PJ: Empty on purpose. Override to extend scene-parsing behaviour (like plugin-specifics).
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendSettings(const aNode: TXMLNode; const aScene: TScene);
var i: integer;
    node: TXMLNode;
begin
    if ShouldShowProgress then
        ProgressForm.Description := 'Loading settings.';

    aScene.Name := aNode.Properties['name'];
    aScene.Author := aNode.Properties['author'];
    aScene.Comments := aNode.Properties['comments'];
    aScene.ShadowOpacity := aNode.IntegerProperties['shadowOpacity'];

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.name = 'ambient' then
        begin
            aScene.AmbientColor := ReadRGBColor( node );
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendMaterials(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene);
var i: integer;
    node: TXMLNode;
begin
    if ShouldShowProgress then
        ProgressForm.Description := 'Loading Materials.';

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.name = 'category' then
        begin
            AppendMaterialCategory(node, aMaterialList, aScene);
        end;
    end;

    TMaterialMatcher.Match(aScene.MaterialManager);
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendMaterialCategory(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene);
var i: integer;
    node: TXMLNode;
    categoryName: string;
begin
    categoryName := aNode.Properties['name'];

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'material' then
        begin
            ProcessMaterial(node, categoryName, aMaterialList, aScene);
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.ProcessMaterial(const aNode: TXMLNode; const aCategoryName: string; const aMaterialList: TStringList; const aScene: TScene);
begin
// implemented in derived classes:
//
//   TXMLSceneReader: adds materials to the scene while updating the material ID value when needed
//   TXMLPluginReader: adds materials to the scene or, if a material with the same ID already exists, replaces existing materials
//
// in both cases, the provided aMaterialList stringlist is appended with the local material ID as found in the XML node
// this local material ID is later used when loading polygons from the XML scene data, which refer to the local material ID
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendMaterialLayer(const aNode: TXMLNode; const aMaterial: TMaterial; const aScene: TScene);
var layer: TTextureLayer;
    node: TXMLNode;
    i: integer;
    isLightmap: boolean;
    grayScale: integer;
begin
    layer := TTextureLayer.Create;

    layer.LayerType := LayerTypeForDescription(aNode.Properties['type']);
    layer.BlendOperation := BlendOperationForDescription(aNode.Properties['blend']);

    isLightmap := aNode.Properties['type'] = XML_LAYER_TYPE_LIGHTMAP;

    if isLightmap then begin
        if aMaterial.ContainsLightmap then
            raise Exception.Create( Format( 'Multiple lightmaps encountered for material "%s"', [aMaterial.Name]));

        // INFO, PJ: Fill up with empty layers to get lightmap layer at correct position
        for i := aMaterial.NumTextureLayers to LIGHTMAP_TEXTURE_LAYER_INDEX - 1 do
            aMaterial.AddEmptyLayer;
    end
    else if aMaterial.ContainsLightmap then
        raise Exception.Create( Format( 'Layers encountered after lightmap for material "%s"', [aMaterial.Name]));

    aMaterial.AddLayer(layer);

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'color' then
        begin
            layer.LayerType := mlColor;
            layer.Color := ReadColor(node, ct32Bits);
            layer.Texture := aScene.MaterialManager.Textures.AddTexture(layer.Color);
        end
        else if node.Name = 'texture' then
        begin
            AppendLayerTexture(node, layer, isLightmap, aScene);
        end;
    end;

    if not Assigned(layer.Texture) then
    begin
        // NOTE, JC: if a texture is missing (either normal texture and/or lightmaps),
        // the file should load anyway, so we provide a random grey color here instead;
        // INFO, PJ: Warnings are handled inside AppendLayerTexture
        if isLightMap then
        begin
            grayScale := 255; // Fullbright... looks like no lightmap has been applied.
        end
        else
            grayScale := random(256);

        layer.LayerType := mlColor;
        layer.Color := TColor(grayScale + grayScale shl 8 + grayScale shl 16 + 255 shl 24);
        layer.Texture := aScene.MaterialManager.Textures.AddTexture(layer.Color);
    end;

end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendLayerTexture(const aNode: TXMLNode; const aLayer: TTextureLayer; const anIsLightmap: boolean; const aScene: TScene);
var textureList: TTextureList;
    fileName: string;
    extension: string;
begin
    // INFO, PJ: First we do some basic checks.
    fileName := dir_texture + aNode.Properties['file'];    // TODO: remove global dir_texture someday

    aLayer.InvalidTextureFileName := fileName; // INFO, PJ: We clear this if all tests pass.
        
    if not FileExistsUTF8(fileName ) then begin
        Warnings.Add( Format( 'Texture "%s" not found.', [fileName] ));
        exit;
    end;

    extension := LowerCase(ExtractFileExt( fileName ));
    if not TTextureLoader.SupportsFileFormat( extension ) then begin
        Warnings.Add( Format( 'Could not load texture "%s" (format "%s" not supported).',
            [ExtractFileName(aNode.Properties['file']), extension]));
        exit;
    end;

    // Then we try to create the texture.
    try
        if FileExistsUTF8(fileName ) then begin
            if anIsLightmap then
                textureList := aScene.MaterialManager.Lightmaps
            else
                textureList := aScene.MaterialManager.Textures;

            aLayer.Texture := textureList.AddTexture( fileName );
            aLayer.InvalidTextureFileName := '';
        end;
    except
        on e: Exception do
            Warnings.Add( Format( 'Could not load texture %s: %s.', [ExtractFileName(aNode.Properties['file']), e.message] ));
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendMaterialRaytracingProperties(const aNode: TXMLNode; const aMaterial: TMaterial);
var properties: TMaterialRaytracingProperties;
    i: Integer;
    node: TXMLNode;
begin
    properties := aMaterial.RaytracingProperties;

    if aNode.HasProperty( 'ambientReflection' ) then aMaterial.RaytracingProperties.AmbientReflection := aNode.FloatProperties['ambientReflection']; 
    if aNode.HasProperty( 'diffuseReflection' ) then aMaterial.RaytracingProperties.DiffuseReflection := aNode.FloatProperties['diffuseReflection'];
    if aNode.HasProperty( 'specularReflection' ) then aMaterial.RaytracingProperties.SpecularReflection := aNode.FloatProperties['specularReflection'];
    if aNode.HasProperty( 'specularExponent' ) then aMaterial.RaytracingProperties.SpecularExponent := aNode.FloatProperties['specularExponent'];
    if aNode.HasProperty( 'reflection' ) then aMaterial.RaytracingProperties.Reflection := aNode.FloatProperties['reflection'];
    if aNode.HasProperty( 'reflectiveExponent' ) then aMaterial.RaytracingProperties.ReflectiveExponent := aNode.FloatProperties['reflectiveExponent'];
    if aNode.HasProperty( 'transparency' ) then aMaterial.RaytracingProperties.Transparency := aNode.FloatProperties['transparency'];
    if aNode.HasProperty( 'refraction' ) then aMaterial.RaytracingProperties.Refraction := aNode.FloatProperties['refraction'];

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'specularColor' then begin
            properties.SpecularColor := ReadColor(node, ct24Bits);
        end
        else if node.Name = 'reflectiveColor' then begin
            properties.ReflectiveColor := ReadColor(node, ct24Bits);
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendPrimitive(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene);
var i: integer;
    node: TXMLNode;
    primitive: TPrimitive;
    vertices: TStringList;
begin
    UpdateProgressForNode( aNode );

    primitive := TPrimitive.Create;
    primitive.ID := aNode.IntegerProperties['id'];
    primitive.Name := aNode.Properties['name'];
    primitive.ObjectType := ObjectTypeForDescription(aNode.Properties['type']);
    primitive.Visible := aNode.BooleanProperties['visible'];
    primitive.SnapMethod := SnapMethodForDescription(aNode.Properties['snap']);
    primitive.AutoUV := aNode.BooleanProperties['autoUV'];
    if aNode.HasProperty('sphere') then
        primitive.IsTrueSphere := aNode.BooleanProperties['sphere'];
    if aNode.HasProperty('groupID') then            // added for 1.7
        primitive.GroupID := aNode.IntegerProperties['groupID'];
    if aNode.HasProperty('skeletonID') then
        primitive.SkeletonID := aNode.IntegerProperties['skeletonID'];

    vertices := THashedStringList.Create;
    try
        for i := 0 to aNode.NumChildNodes - 1 do
        begin
            node := aNode.ChildNodes[i];

            if node.Name = 'comments' then begin
                primitive.Comments := node.Text;
            end else if node.Name = 'tag' then begin
                primitive.Tag := node.Text;
            end else if node.Name = 'lighting' then begin
                AppendLightingProperties(node, primitive as TPrimitive);
            end else if node.Name = 'vertices' then begin
                AppendVertices(node, primitive, vertices);
            end else if node.Name = 'polygons' then  begin
                AppendPolygons(node, primitive, vertices, aMaterialList, aScene);
            end;
        end;
    finally
        FreeAndNil(vertices);
    end;

    // PERF, PJ: Don't trigger any events here. Do it in one go at the end.
    aScene.Primitives.AddPrimitive(primitive,false,false);
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendLightingProperties(const aNode: TXMLNode; const aPrimitive: TPrimitive);
begin
    aPrimitive.OverrideMaterialLighting := TRUE;

    if aNode.HasProperty('castShadows') then
        aPrimitive.CastShadows := aNode.BooleanProperties['castShadows'];

    if aNode.HasProperty('receiveShadows') then
        aPrimitive.ReceiveShadows := aNode.BooleanProperties['receiveShadows'];
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendVertices(const aNode: TXMLNode; const aPrimitive: TBasePrimitive; const aVertices: TStringList);
var i: integer;
    node: TXMLNode;
begin
    for i := 0 to aNode.NumChildNodes - 1 do begin
        node := aNode.ChildNodes[i];

        if node.name = 'vertex' then
            AppendVertex(node, aPrimitive, aVertices);
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendViews(const aNode: TXMLNode; const aScene: TScene; const aViewPortManager: TViewPortManager);
var
  I: Integer;
  node: TXMLNode;

begin
// *********************HERE*********************HERE*********************HERE*********************HERE*********************HERE*********************HERE
    if ShouldShowProgress then
        ProgressForm.Description := 'Loading views.';
    If (aNode.NumChildNodes = aNode.IntegerProperties['numOfViewPorts'])and (aNode.NumChildNodes >0) then Begin
        for I :=  aviewportmanager.Numberof - 1 downto 0 do begin
            aviewportmanager.ViewPorts[i].Activate ;
            aviewportmanager.RemoveActiveViewPort;
        end;
        // Deal with normal and minimised views
        for I := 0 to aNode.NumChildNodes-1 do begin
            UpdateProgressForNode( aNode );
            node:= anode.ChildNodes[i];
            if anode.ChildNodes[i].Properties['WindowState']<> '1' then begin
                case tviewmode(node.integerProperties['viewmode']) of
                    vmPerspective:
                        AppendPerspectiveView(node,aScene,aViewPortManager);
                    vmTop,vmFront,vmSide,vmUV:
                        AppendFlatView(node,aScene,aViewPortManager);
                end;
            end;
        end;
        // now deal with maximised views
        for I := 0 to aNode.NumChildNodes-1 do begin
            UpdateProgressForNode( aNode );
            node:= anode.ChildNodes[i];
            if anode.ChildNodes[i].Properties['WindowState']= '1' then begin
                case tviewmode(node.integerProperties['viewmode']) of
                    vmPerspective:
                        AppendPerspectiveView(node,aScene,aViewPortManager);
                    vmTop,vmFront,vmSide,vmUV:
                        AppendFlatView(node,aScene,aViewPortManager);
                end;
            end;            
        end;
    End;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendFlatView(const aNode: TXMLNode;
  const aScene: TScene; const aViewPortManager: TViewPortManager);
var
    thisViewPort : TfrmViewPort;
    I: Integer;
    node: TXMLNode;
begin
    thisViewPort := aViewPortManager.AddViewPort(TViewMode(aNode.integerProperties['viewmode']),aScene,frmmain.RenderOptions);
    case aNode.integerProperties['WindowState'] of
        0:
            thisViewPort.WindowState := wsMinimized;
        1:
            thisViewPort.WindowState := wsMaximized;
        -1: begin
            thisViewPort.WindowState := wsNormal;
            thisViewPort.Width := aNode.integerProperties['width'];
            thisViewPort.Height := aNode.integerProperties['height'];
            thisViewPort.Top := aNode.integerProperties['top'];
            thisViewPort.Left:= aNode.integerProperties['left'];
            end;
    end;
    thisViewPort.SupportedViewMode:= TviewMode(aNode.integerProperties['SupportedViewMode']);
    for I := 0 to aNode.NumChildNodes - 1 do begin
        node:= anode.ChildNodes[i];
        if node.Name = 'background' then
             appendBackground(node, TfrmFlatViewPort(thisViewPort))
        else if node.Name = 'camera' then
             appendCamera(node,thisViewPort);
    end;

end;

// ----------------------------------------------------------------------------


procedure TXMLDataReader.AppendPerspectiveView(const aNode: TXMLNode;
  const aScene: TScene; const aViewPortManager: TViewPortManager);
var
    thisViewPort : TfrmViewPort;
    I : Integer;
    node: TXMLNode;
begin
  UpdateProgressForNode( aNode );
  thisViewPort := aViewPortManager.AddViewPort(TViewMode(aNode.integerProperties['viewmode']), ascene,frmmain.RenderOptions,frmmain.cameraoptions,frmmain.controloptions);
    case aNode.integerProperties['WindowState'] of
        0:
            thisViewPort.WindowState := wsMinimized;
        1:
            thisViewPort.WindowState := wsMaximized;
        -1: begin
            thisViewPort.WindowState := wsNormal;
            thisViewPort.Width := aNode.integerProperties['width'];
            thisViewPort.Height := aNode.integerProperties['height'];
            thisViewPort.Top := aNode.integerProperties['top'];
            thisViewPort.Left:= aNode.integerProperties['left'];
            end;
    end;
  thisViewPort.SupportedViewMode:= TviewMode(aNode.integerProperties['SupportedViewMode']);
    for I := 0 to aNode.NumChildNodes - 1 do begin
        node:= anode.ChildNodes[i];
        if node.Name = 'camera' then
             appendCamera(node,thisViewPort);
    end;

end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.appendBackground(const aNode: TXMLNode; aViewport: TfrmFlatViewPort);
var
    Selected: Integer;
    openDialog : TOpenDialog;
begin
    UpdateProgressForNode( aNode );
    if (SetCurrentDirUTF8(extractfilepath(application.exename)))and (save_Load_Backgrounds) then begin
//!!!        forcecurrentdirectory := true;
        if  FileExistsUTF8(ExpandFileName(aNode.Properties['imagefile']))then begin
          aViewPort.PrimitiveBGI := TBackground.CreateBackground(ExpandFileNameUTF8(aNode.Properties['imagefile']), aviewport.viewmode, aviewport.H_DC, aviewport.H_RC);
          aViewPort.ShowBGI := true;
          SizeBackground(aViewport, aNode);
        end
        else begin
          Selected := MessageDlg(
          #9#9'File for background '#13#10 +
          aNode.Properties['imagefile'] + #13#10 +
          #9'Not found. Do you wish to find manually?',mtError, [mbYes, mbNo], 0);
          if Selected = 6 then begin
            openDialog := TOpenDialog.Create(aViewport);
            openDialog.InitialDir := GetCurrentDirUTF8;
            openDialog.Filter := 'JPEG(*.jpg)|*.JPG|Bitmap(*.bmp)|*.BMP';
            openDialog.Options := [ofFileMustExist ];
            if  openDialog.Execute then begin
              aViewPort.PrimitiveBGI := TBackground.CreateBackground(openDialog.FileName, aviewport.viewmode, aviewport.H_DC, aviewport.H_RC);
              aViewPort.ShowBGI := true;
              SizeBackground(aViewport, aNode);
            end;
            openDialog.Free;
          end;
        end;
    end;
end;

procedure TXMLDataReader.SizeBackground(var aViewport: TfrmFlatViewPort; const aNode: TXMLNode);
var
  node: TXMLNode;
  I,J: Integer;
begin
  //aViewPort.PrimitiveBGI := TBackground.CreateBackground(ExpandFileName(aNode.Properties['imagefile']), aviewport.viewmode, aviewport.H_DC, aviewport.H_RC);
  //aViewPort.ShowBGI := true;
  for I := 0 to aNode.NumChildNodes - 1 do
  begin
    if aNode.ChildNodes[I].Name = 'vertexpositions' then
    begin
      node := aNode.ChildNodes[I];
      for J := 0 to node.NumChildNodes - 1 do
      begin
        if node.ChildNodes[J].Name = 'vertexposition' then
        begin
          aViewPort.PrimitiveBGI.Vertices[J].x := node.ChildNodes[j].FloatProperties['x'];
          aViewPort.PrimitiveBGI.Vertices[J].y := node.ChildNodes[j].FloatProperties['y'];
          aViewPort.PrimitiveBGI.Vertices[J].z := node.ChildNodes[j].FloatProperties['z'];
        end;
      end;
    end;
  end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.appendCamera(const aNode: TXMLNode;  aViewPort: TfrmViewPort);
begin
    UpdateProgressForNode( aNode );
    if save_Load_Cameras then begin
        if aNode.Properties['type']= 'TCamera2d' then begin
            if aNode.ChildNodes[0].Name = 'position' then begin
                TfrmFlatViewPort(aViewPort).Camera.Position.x := aNode.ChildNodes[0].FloatProperties['x'];
                TfrmFlatViewPort(aViewPort).Camera.Position.y := aNode.ChildNodes[0].FloatProperties['y'];
                TfrmFlatViewPort(aViewPort).Camera.Position.z := aNode.ChildNodes[0].FloatProperties['z'];
                TfrmFlatViewPort(aViewPort).Camera.Depth := aNode.ChildNodes[0].IntegerProperties['depth'];
                aViewPort.SetView;
            end;
        end
        else if aNode.Properties['type']= 'TCameraUV' then begin
            if aNode.ChildNodes[0].Name = 'position' then begin
                TfrmUVEditor(aViewPort).Camera.CenterU := aNode.ChildNodes[0].FloatProperties['CenterU'];
                TfrmUVEditor(aViewPort).Camera.CenterV := aNode.ChildNodes[0].FloatProperties['CenterV'];
                TfrmUVEditor(aViewPort).Camera.Scale := aNode.ChildNodes[0].FloatProperties['scale'];
                aViewPort.SetView;
            end;
        end
        else if aNode.Properties['type']= 'TCamera3d' then begin
            if aNode.ChildNodes[0].Name = 'position' then begin
                Tfrm3d(aViewPort).Camera3D.Position.x := aNode.ChildNodes[0].FloatProperties['x'];
                Tfrm3d(aViewPort).Camera3D.Position.y := aNode.ChildNodes[0].FloatProperties['y'];
                Tfrm3d(aViewPort).Camera3D.Position.z := aNode.ChildNodes[0].FloatProperties['z'];
                Tfrm3d(aViewPort).Camera3D.Rotation.x := aNode.ChildNodes[0].FloatProperties['rotx'];
                Tfrm3d(aViewPort).Camera3D.Rotation.y := aNode.ChildNodes[0].FloatProperties['roty'];
                Tfrm3d(aViewPort).Camera3D.Rotation.z := aNode.ChildNodes[0].FloatProperties['rotz'];
                Tfrm3d(aViewPort).Camera3D.FOV := aNode.ChildNodes[0].FloatProperties['fov'];
                aViewPort.SetView;
            end;
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendVertex(const aNode: TXMLNode; const aPrimitive: TBasePrimitive; const aVertices: TStringList);
var vertex: TVertice;
    i: integer;
    node: TXMLNode;
begin
    UpdateProgressForNode( aNode );

    vertex := TVertice.Create;
    aVertices.AddObject(aNode.Properties['id'], vertex);

    vertex.X := aNode.FloatProperties['x'];
    vertex.Y := aNode.FloatProperties['y'];
    vertex.Z := aNode.FloatProperties['z'];

    if aNode.HasProperty('jointID') then
        vertex.JointID := aNode.IntegerProperties['jointID'];

    for i := 0 to aNode.NumChildNodes - 1 do begin
        node := aNode.ChildNodes[i];

        if node.name = 'tag' then
            vertex.Tag := node.Text
        else if node.name = 'color' then
            vertex.Color := ReadColor(node, ct32Bits);
    end;

    aPrimitive.Vertex_Add(vertex);
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendPolygons(const aNode: TXMLNode; const aPrimitive: TBasePrimitive; const aVertices: TStringList; const aMaterialList: TStringList; const aScene: TScene);
var i: integer;
    node: TXMLNode;
begin
    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.name = 'poly' then
        begin
            AppendPolygon(node, aPrimitive, aVertices, aMaterialList, aScene);
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendPolygon(const aNode: TXMLNode; const aPrimitive: TBasePrimitive; const aVertices: TStringList; const aMaterialList: TStringList; const aScene: TScene);
var i: integer;
    node: TXMLNode;
    polygon: TPolygon;
    materialID: string;
    material: TMaterial;
begin
    UpdateProgressForNode( aNode );

    polygon := TPolygon.Create;

    // the aMaterialList parameter stores the material ID values as they come from
    // the XML data; as polygons also refer to these material ID values, we use that
    // list to lookup the actual materials; do note that the materials themselfs might
    // have totally different ID values at this moment when a TXMLSceneReader is used;
    // explanation: when merging scenes, materials having the same ID as materials
    // in the scene might be added; these materials receive a new unique ID when being
    // loaded into the scene; they are stored using their old ID's (from the XML data)
    // in the aMaterialList so we can still link polygons to materials here (who, in
    // the XML, refer to the materials using the old (!) ID values);
    // note: when using a TXMLPluginReader, the provided aMaterialList might not have
    // any materials (when materials weren't send to DeleD at all or when materials are
    // send _after_ polygons have been send); if that is the case, we search for a
    // material with the correct ID in DeleD's current materiallist and, if found,
    // assign that material to the polygon in question
    materialID := aNode.Properties['mid'];
    if (materialID = '') or (StrToInt(materialID)<0) then materialID := '0';       // 1.7 bugfix: not sure if this is the way
    //if StrToInt(materialID)<0 then materialID := '0';       // 1.7 bugfix: not sure if this is the way
    if aMaterialList.IndexOf(materialID)=-1 then begin
        material := aScene.MaterialManager.GetMaterialByID(StrToInt(materialID));
        if not Assigned(material) then
            raise Exception.Create(Format('No material found for id "%s".', [materialID]));
    end else
        material := aMaterialList.Objects[aMaterialList.IndexOf(materialID)] as TMaterial;

    polygon.Material := material;

    // Initialize UV Coordinates.
    for i := 0 to material.NumTextureLayers - 1 do
        polygon.UVCoordinates[i] := TUVList.Create;

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.name = 'vertex' then
            AppendPolygonVertex(node, polygon, aVertices, aPrimitive)
        else if node.Name = 'tag' then
            polygon.Tag := node.Text;

    end;

    aPrimitive.Polygon_Add(polygon);
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendPolygonVertex(const aNode: TXMLNode; const aPolygon: TPolygon; const aVertices: TStringList; const aPrimitive: TBasePrimitive);
var i, index: integer;
    vertexID: string;
    vertex: TVertice;
    
begin
    vertexID := aNode.Properties['vid'];
    index := aVertices.IndexOf(vertexID);
    if index = -1 then
        raise Exception.Create( Format( 'No vertex found for id "%s".', [vertexID] ));

    vertex := aVertices.Objects[index] as TVertice;
    aPrimitive.Polygon_AddVertex(aPolygon, vertex);

    if Assigned( aPolygon.Material ) then
    begin
        index := 0;
        for i := 0 to aPolygon.Material.NumTextureLayers - 1 do
        begin
            if Assigned( aPolygon.Material.TextureLayers[i] ) then begin
                // Sometimes the model on disk is slightly corrupted, missing UV coordinates for lightmap layers.
                // Only remove this check if you are pretty sure this never happens anymore :)
                if aNode.HasProperty( 'u' + IntToStr(index) ) and aNode.HasProperty( 'v' + IntToStr(index) ) then begin
                    aPolygon.UVCoordinates[i].Add(
                        aNode.FloatProperties['u' + IntToStr(index)],
                        aNode.FloatProperties['v' + IntToStr(index)]
                    );
                end
                else
                    aPolygon.UVCoordinates[i].Add( 0, 0 );

                index := index + 1;
            end;
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendSkeleton(const aNode: TXMLNode; const aScene: TScene);
var i: integer;
    node: TXMLNode;
    skeleton: TSkeleton;
begin
    UpdateProgressForNode( aNode );

    skeleton := TSkeleton.Create;
    skeleton.ID := aNode.IntegerProperties['id'];
    skeleton.Name := aNode.Properties['name'];
    skeleton.Author := aNode.Properties['author'];
    skeleton.Visible := aNode.BooleanProperties['visible'];

    if aNode.HasProperty('jointsize') then begin
        skeleton.JointSize := aNode.IntegerProperties['jointsize'];
    end;

    for i := 0 to aNode.NumChildNodes - 1 do begin
        node := aNode.ChildNodes[i];

        if node.Name = 'joints' then begin
            AppendJoints(node, skeleton, aScene)
        end else if node.Name = 'animations' then begin
            AppendAnimations(node, skeleton);
        end;
    end;

    skeleton.LinkScenePrimitivesToSkeleton(aScene.Primitives);
    aScene.Skeletons.AddSkeleton(skeleton);
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendJoints(const aNode: TXMLNode; const aSkeleton: TSkeleton; const aScene: TScene);
var i: integer;
    node: TXMLNode;
begin
    for i := 0 to aNode.NumChildNodes - 1 do begin
        node := aNode.ChildNodes[i];

        if node.name = 'joint' then
            AppendJoint(node, aSkeleton, nil, nil, aScene);
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendJoint(const aNode: TXMLNode; const aSkeleton: TSkeleton; const aParentJoint: TJoint; const aParentJointLocation: TJointLocation; const aScene: TScene);
var joint: TJoint;
    jointLocation: TJointLocation;
    i: integer;
    center: TVertice;
    node, childNode: TXMLNode;
    jointBuilder: TJointBuilder;
    boneBuilder: TBoneBuilder;
    bone: TBone;
    paramList: TParamList;
begin
    UpdateProgressForNode( aNode );

    center := TVertice.Create(aNode.FloatProperties['x'],aNode.FloatProperties['y'],aNode.FloatProperties['z']);
    paramList := TParamList.Create(false);
    jointBuilder := TJointBuilder.Create;
    try
        paramList.Add(center,'center');
        paramList.Add(aScene.MaterialManager,'materialManager');
        paramList.Add(aSkeleton.JointSize, 'jointSize');
        joint := jointBuilder.Build(paramList);
    finally
        FreeAndNil(jointBuilder);
        FreeAndNil(paramList);
        FreeAndNil(center);
    end;

    if Assigned(joint) then begin
        joint.Skeleton := aSkeleton;
        joint.ID := aNode.IntegerProperties['id'];
        joint.Name := aNode.Properties['name'];
        joint.ParentJoint := aParentJoint;
        if Assigned(aParentJoint) then
            aParentJoint.ChildJoints.Add(joint);
        aSkeleton.Joints.Add(joint);    // do NOT use AddJoint here because it sets the joint.ID value

        jointLocation := TJointLocation.Create;
        jointLocation.ID := joint.ID;
        jointLocation.Position.x := aNode.FloatProperties['x'];
        jointLocation.Position.y := aNode.FloatProperties['y'];
        jointLocation.Position.z := aNode.FloatProperties['z'];
        aSkeleton.BaseFrame.AddJointLocation(jointLocation, aParentJointLocation);

        if Assigned(aParentJointLocation) then begin
            paramList := TParamList.Create(false);
            boneBuilder := TBoneBuilder.Create;
            try
                paramList.Add(aParentJointLocation, 'jointLocation1');
                paramList.Add(jointLocation, 'jointLocation2');
                paramList.Add(aScene.MaterialManager, 'materialManager');
                paramList.Add(aSkeleton.JointSize, 'boneSize');                 // NOTE, JCC: bonesize = jointsize
                bone := boneBuilder.Build(paramList);
            finally
                FreeAndNil(boneBuilder);
                FreeAndNil(paramList);
            end;
            aParentJoint.ChildBones.Add(bone);
        end;

        if aNode.NumChildNodes=1 then begin
            node := aNode.ChildNodes[0];
            if node.Name = 'childjoints' then begin
                for i := 0 to node.NumChildNodes-1 do begin
                    childNode := node.ChildNodes[i];
                    if childNode.Name = 'joint' then
                        AppendJoint(childNode, aSkeleton, joint, jointLocation, aScene);
                end;
            end;
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendAnimations(const aNode: TXMLNode; const aSkeleton: TSkeleton);
var i: integer;
    node: TXMLNode;
begin
    for i := 0 to aNode.NumChildNodes - 1 do begin
        node := aNode.ChildNodes[i];

        if node.Name = 'animation' then
            AppendAnimation(node, aSkeleton);
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendAnimation(const aNode: TXMLNode; const aSkeleton: TSkeleton);
var animationBuilder: TAnimationBuilder;
    i,NOF,FPS: integer;
    animation: TAnimation;
    node: TXMLNode;
begin
    UpdateProgressForNode( aNode );

    NOF := aNode.IntegerProperties['NOF'];
    FPS := aNode.IntegerProperties['FPS'];

    // the animationBuilder builds an animation by duplicating the baseframe
    // a specific number of times (NOF); after that, we update specific frames
    // in the animation by using keyframe information from the DXS
    animationBuilder := TAnimationBuilder.Create;
    try
        animation := animationBuilder.Build(aSkeleton,NOF,FPS)
    finally
        FreeAndNil(animationBuilder);
    end;

    if Assigned(animation) then begin
        aSkeleton.SelectedAnimation := nil;     // it was set by AnimationBuilder

        // TODO: set by AnimationBuilder too; we override the ID value but should we really???
        // We won't have to unless other objects refer to the animation by ID...
        animation.ID := aNode.IntegerProperties['id'];
        animation.Name := aNode.Properties['name'];
        animation.Author := aNode.Properties['author'];

        for i := 0 to aNode.NumChildNodes-1 do begin
            node := aNode.ChildNodes[i];
            if node.Name = 'tag' then
                animation.Tag := node.Text;
            // an animation has its keyframes stored in the 'keyframes' tag;
            // read that data and update those frames of the animation accordingly
            if node.Name = 'keyframes' then
                AppendFrames(node, animation);
        end;

        // setting the HasChanged property to true will trigger the
        // AnimationFramesInterpolatorTool so frames in between keyframes
        // will receive interpolated jointlocations (location & rotation values)
        animation.HasChanged := true;
        aSkeleton.Animations.Add(animation);
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendFrames(const aNode: TXMLNode; const aAnimation: TAnimation);
var i: integer;
    node: TXMLNode;
begin
    for i := 0 to aNode.NumChildNodes - 1 do begin
        node := aNode.ChildNodes[i];

        if node.Name = 'keyframe' then
            AppendFrame(node, aAnimation);
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendFrame(const aNode: TXMLNode; const aAnimation: Tanimation);
var i,id: integer;
    animationFrame: TAnimationFrame;
    node: TXMLNode;
begin
    // The AppendFrame method actually doesn't append a frame, it replaces
    // the JointLocation data of a frame with data from the DXS.
    // Only keyframes are stored in the DXS. When loading a file, DeleD
    // first creates an animation by copying the baseframe of the skeleton
    // a specific number of time: it uses the NumberOfFrames from the DXS
    // for that. Then it overwrites the data of keyframes with data found
    // in the DXS.

    id := aNode.IntegerProperties['id'];
    animationFrame := aAnimation.AnimationFrames.GetAnimationFrameByID(id);
    if Assigned(animationFrame) then begin
        animationFrame.IsBaseFrame := aNode.BooleanProperties['isBaseFrame'];
        animationFrame.IsKeyFrame := aNode.BooleanProperties['isKeyFrame'];

        // read jointlocations and update keyframes with them
        if aNode.NumChildNodes=1 then begin
            node := aNode.ChildNodes[0];
            if node.Name = 'jointlocations' then begin
                for i := 0 to node.NumChildNodes - 1 do
                    AppendJointLocation(node.ChildNodes[i], animationFrame);
            end;
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendJointLocation(const aNode: TXMLNode; const aAnimationFrame: TAnimationFrame);
var id: integer;
    jointLocation: TJointLocation;
begin
    id := aNode.IntegerProperties['id'];
    jointLocation := aAnimationFrame.GetJointLocationByID(id);
    if not Assigned(jointLocation) then
        raise Exception.Create( 'Unable to find specified jointlocation with ID ' + IntToStr(id) )
    else begin
        jointLocation.Location.x := aNode.FloatProperties['lx'];
        jointLocation.Location.y := aNode.FloatProperties['ly'];
        jointLocation.Location.z := aNode.FloatProperties['lz'];
        jointLocation.Rotation.x := aNode.FloatProperties['rx'];
        jointLocation.Rotation.y := aNode.FloatProperties['ry'];
        jointLocation.Rotation.z := aNode.FloatProperties['rz'];
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendSkeletons(const aNode: TXMLNode; const aScene: TScene);
var i: integer;
    node: TXMLNode;
begin
    if ShouldShowProgress then
        ProgressForm.Description := 'Loading Skeletons.';

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'skeleton' then
        begin
            AppendSkeleton(node, aScene);
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendLights(const aNode: TXMLNode; const aScene: TScene);
var i: integer;
    node: TXMLNode;
begin
    if ShouldShowProgress then
        ProgressForm.Description := 'Loading Lights.';

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'light' then
        begin
            AppendLight(node, aScene);
        end;
    end;
end;

// -----------------------------------------------------------------------------

procedure TXMLDataReader.AppendPaths(const aNode: TXMLNode; const aScene: TScene);
var i: integer;
    node: TXMLNode;
begin
    if ShouldShowProgress then
        ProgressForm.Description := 'Loading Paths.';

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'path' then
        begin
            AppendPath(node, aScene);
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendLight(const aNode: TXMLNode; const aScene: TScene);
var i: integer;
    node: TXMLNode;
    light: TLight;
    lightClass: TLightClass;
begin
    UpdateProgressForNode( aNode );

    lightClass := LightClassForDescription(aNode.Properties['type']);
    light := lightClass.Create;
    aScene.Primitives.AddPrimitive(light);

    light.ID := aNode.IntegerProperties['id'];
    light.Name := aNode.Properties['name'];
    if aNode.HasProperty('visible') then
        light.Visible := aNode.BooleanProperties['visible'];
    light.Range := aNode.IntegerProperties['range'];
    light.Active := aNode.BooleanProperties['active'];
    if aNode.HasProperty('castShadows') then
        light.CastShadows := aNode.BooleanProperties['castShadows'];
    if aNode.HasProperty('snap') then
        light.Snap := aNode.BooleanProperties['snap'];
    light.Intensity := aNode.IntegerProperties['intensity'];

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'comments' then begin
            light.Comments := node.Text;
        end
        else if node.Name = 'tag' then begin
            light.Tag := node.Text;
        end
        else if node.name = 'attenuation' then
        begin
            AppendLightAttenuation(node, light);
        end
        else if node.name = 'direction' then
        begin
            AppendLightDirection(node, light);
        end
        else if node.name = 'position' then
        begin
            AppendLightPosition(node, light);
        end
        else if node.name = 'cone' then
        begin
            AppendLightCone(node, light);
        end
        else if node.name = 'diffuse' then
        begin
            light.Diffuse := ReadRGBColor(node);
        end
        else if node.name = 'specular' then
        begin
            light.Specular := ReadRGBColor(node);
        end;
    end;

    LoadPrefabForLight(light, aScene);
    light.FaceDirection;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendLightAttenuation(const aNode: TXMLNode; const aLight: TLight);
begin
    aLight.Constant := aNode.FloatProperties['constant'];
    aLight.Linear := aNode.FloatProperties['linear'];
    aLight.Quadratic := aNode.FloatProperties['quadratic'];
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendLightPosition(const aNode: TXMLNode; const aLight: TLight);
begin
    aLight.Position.X := aNode.FloatProperties['x'];
    aLight.Position.Y := aNode.FloatProperties['y'];
    aLight.Position.Z := aNode.FloatProperties['z'];
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendLightDirection(const aNode: TXMLNode; const aLight: TLight);
begin
    aLight.Direction.X := aNode.FloatProperties['x'];
    aLight.Direction.Y := aNode.FloatProperties['y'];
    aLight.Direction.Z := aNode.FloatProperties['z'];
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.AppendLightCone(const aNode: TXMLNode; const aLight: TLight);
var spotLight: TSpotLight;
begin
    if not (aLight is TSpotLight) then
        raise Exception.Create( 'Cone properties only supported for spotlights.' );

    spotLight := aLight as TSpotLight;
    spotLight.InnerAngle := aNode.IntegerProperties['inner'];
    spotLight.OuterAngle := aNode.IntegerProperties['outer'];
    spotLight.FallOff := aNode.IntegerProperties['falloff'];
end;

//---------------------------------------------------------------

procedure TXMLDataReader.AppendPath(const aNode: TXMLNode; const aScene: TScene);
var
  Path: TPath;

    i: integer;
    node: TXMLNode;
begin
    UpdateProgressForNode( aNode );

    //Create path and read properties
    path            := TPath.Create();
    path.ID         := aNode.IntegerProperties['id'];
    path.Name       := aNode.Properties['name'];
    path.PathType   := StringToPathType( aNode.Properties['type'] );
    path.IsClosed   := aNode.BooleanProperties['isclosed'];
    path.Visible    := aNode.BooleanProperties['visible'];
    path.SnapMethod := SnapMethodForDescription(aNode.Properties['snap']);
    path.GroupID    := aNode.IntegerProperties['groupID'];

    //Read comments, tag and controlpoints
    for I := 0 to aNode.NumChildNodes - 1 do
    begin
      node := aNode.ChildNodes[i];

      if node.Name = 'comments' then path.Comments := node.Text;
      if node.Name = 'tag' then path.Tag := node.Text;
      if node.Name = 'controlpoints' then AppendControlPoints(node, path);
    end;

    //Add the path to the scene
    aScene.Primitives.AddPrimitive( Path );
end;

//Reads all controlpoints from the XML
procedure TXMLDataReader.AppendControlPoints(const aNode: TXMLNode; const aPath: TPath);
var
  I:    Integer;
  node: TXMLNode;
begin
  //Just iterate through the child nodes and create all control points
  for I := 0 to aNode.NumChildNodes - 1 do
  begin
    node := aNode.ChildNodes[I];

    AppendControlPoint(node, aPath);
  end;
end;

//Reads a single controlpoint from the XML
procedure TXMLDataReader.AppendControlPoint(const aNode: TXMLNode; const aPath: TPath);

  //Little inline function for reading vectors
  function ReadVectorFromNode(const aNode: TXMLNode): TVector3f;
  begin
    Result.x := aNode.FloatProperties['x'];
    Result.y := aNode.FloatProperties['y'];
    Result.z := aNode.FloatProperties['z'];
  end;

var
  I:     Integer;
  Point: TPathControlPoint;
  node:  TXMLNode;
begin
  Point := TPathControlPoint.Create(aPath);

  //Read tag, position, upVector and Tangent vectors for this control-point
  for I := 0 to aNode.NumChildNodes - 1 do
  begin
    node := aNode.ChildNodes[I];

    if node.Name = 'tag' then Point.Tag := node.Text;
    if node.Name = 'position' then  Point.Position := ReadVectorFromNode(node);
    if node.Name = 'up' then Point.UpVector := ReadVectorFromNode(node);
    if node.Name = 'tangent1' then
    begin
      Point.TangentVector1 := ReadVectorFromNode(node);
      Point.UserDefinedTangent1 := node.BooleanProperties['userdefined'];
    end;
    if node.Name = 'tangent2' then
    begin
      Point.TangentVector2 := ReadVectorFromNode(node);
      Point.UserDefinedTangent2 := node.BooleanProperties['userdefined'];
    end;
  end;

  //Add the controlpoint to the path
  aPath.AddControlPoint( Point );
end;

// ----------------------------------------------------------------------------

function TXMLDataReader.ReadColor(const aNode: TXMLNode; const aColorType: TColorType): TColor;
begin
    result := aNode.IntegerProperties['r'];
    result := result or (aNode.IntegerProperties['g'] shl 8);
    result := result or (aNode.IntegerProperties['b'] shl 16);

    if aColorType = ct32Bits then
    begin
        if aNode.HasProperty('a') then
        begin
            result := result or (aNode.IntegerProperties['a'] shl 24);
        end
        else
            result := result or integer($FF000000);
    end;
end;

// ----------------------------------------------------------------------------

function TXMLDataReader.ReadRGBColor(const aNode: TXMLNode): TRGBColor;
begin                                  
    result.R := aNode.IntegerProperties['r'] / 255;
    result.G := aNode.IntegerProperties['g'] / 255;
    result.B := aNode.IntegerProperties['b'] / 255;
end;

// ----------------------------------------------------------------------------

function TXMLDataReader.GetWarnings: TStringList;
begin
  if fWarnings = nil then
      fWarnings := TStringList.Create;

  result := fWarnings;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.UpdateScene(const aScene: TScene);
var i: integer;
begin
    for i := 0 to aScene.Primitives.Count-1 do
        aScene.Primitives[i].OnHasChanged := aScene.OnHasChangedHandler;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.UpdateProgressForNode( const aNode: TXMLNode );
begin
    if ShouldShowProgress then
        ProgressForm.Position := aNode.Position;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.HandleProgress(const aPosition, aTotalSize: integer);
begin
    if ShouldShowProgress then begin
        ProgressForm.Position := aPosition;
        ProgressForm.Max := aTotalSize;
    end;
end;

// ----------------------------------------------------------------------------

function TXMLDataReader.LayerTypeForDescription(const aDescription: string): TMaterialLayerType;
var found: boolean;

begin
    found := FALSE;

    for result := Low(TMaterialLayerType) to High(TMaterialLayerType) do
    begin
        if XML_MATERIAL_LAYER_TYPES[result] = aDescription then
        begin
            found := TRUE;
            break;
        end;
    end;

    if (not found) and (aDescription = XML_LAYER_TYPE_LIGHTMAP) then
    begin
        result := mlTexture;
        found := TRUE;
    end;

    if not found then
    begin
        raise Exception.Create(Format('Unknown layer type "%s".', [aDescription]));
    end;
end;

// ----------------------------------------------------------------------------

function TXMLDataReader.BlendOperationForDescription(const aDescription: string): TBlendOperation;
var found: boolean;

begin
    found := FALSE;

    for result := Low(TBlendOperation) to High(TBlendOperation) do
    begin
        if XML_BLEND_OPERATIONS[result] = aDescription then
        begin
            found := TRUE;
            break;
        end;
    end;

    if not found then
    begin
        raise Exception.Create(Format('Unknown blend operation "%s".', [aDescription]));
    end;
end;


// ----------------------------------------------------------------------------

function TXMLDataReader.ObjectTypeForDescription(const aDescription: string): TObjectType;
var found: boolean;

begin
    found := FALSE;

    for result := Low(TObjectType) to High(TObjectType) do
    begin
        if XML_PRIMITIVE_TYPES[result] = aDescription then
        begin
            found := TRUE;
            break;
        end;
    end;

    if not found then
    begin
        raise Exception.Create(Format('Unknown object type "%s".', [aDescription]));
    end;
end;

// ----------------------------------------------------------------------------

function TXMLDataReader.SnapMethodForDescription(const aDescription: string): TSnapMethod;
var found: boolean;

begin
    found := FALSE;

    for result := Low(TSnapMethod) to High(TSnapMethod) do
    begin
        if XML_SNAP_METHODS[result] = aDescription then
        begin
            found := TRUE;
            break;
        end;
    end;

    if not found then
    begin
        raise Exception.Create(Format('Unknown snap method "%s".', [aDescription]));
    end;
end;

// ----------------------------------------------------------------------------


function TXMLDataReader.LightClassForDescription(const aDescription: string): TLightClass;
var i: integer;
begin
    result := nil;

    for i := Low(XML_LIGHT_CLASS_INFOS) to High(XML_LIGHT_CLASS_INFOS) do
    begin
        if XML_LIGHT_CLASS_INFOS[i].name = LowerCase(aDescription) then
        begin
            result := XML_LIGHT_CLASS_INFOS[i].LightClass;
            break;
        end;
    end;

    if not Assigned(result) then
    begin
        raise Exception.Create(Format('Unknown light class "%s".', [aDescription]));
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataReader.LoadPrefabForLight(const aLight: TLight; const aScene: TScene);
var i: integer;
    prefab, name: string;
    dpfLoader: TDPFLoader;
begin
    for i := Low(XML_LIGHT_CLASS_INFOS) to High(XML_LIGHT_CLASS_INFOS) do
    begin
        if XML_LIGHT_CLASS_INFOS[i].LightClass = aLight.ClassType then
        begin
            prefab := XML_LIGHT_CLASS_INFOS[i].Prefab;
            break;
        end;
    end;

    if prefab = '' then
    begin
        raise Exception.Create( Format('Could not find prefab for light class "%s".', [aLight.ClassName]));
    end;

    name := aLight.Name;

    dpfLoader := TDPFLoader.Create;
    dpfLoader.LoadDPF(prefab, true, dir_texture, aLight.Position, aScene.MaterialManager, TPrimitive(aLight));
    FreeAndNil(dpfLoader);

    // INFO, PJ: Override default name assigned by LoadDPF with acutal name from xml.
    aLight.Name := name;
end;

end.

