{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_XMLDataWriter;

{$MODE Delphi}

{ ----------------------------------------------------------------------------
  TODO, PJ:
  [ ] Whitespace aan het eind van 'text' (closeTag komt pas op volgende regel)
  [ ] Testen met verschillende float formats
  ---------------------------------------------------------------------------- }


interface

uses
    Graphics,
    unit_Enums,
    unit_Scene,
    unit_Materials,
    unit_Primitives,
    unit_Skeleton,
    unit_Joint,
    unit_JointLocation,
    unit_Animation,
    unit_AnimationFrame,
    unit_Types,
    unit_Lights,
    unit_Paths,
    unit_PathList,
    unit_MaterialList,
    unit_XMLWriter,
    unit_ViewPortManager,
    frmViewPortForm,
    frmFlatViewPortForm,
    unit_Cameras,
    forms;

type
    TXMLDataWriter = class(TObject)
    private
        function DescriptionForLightClass(const aLightClass: TClass): string;
    protected
        procedure AppendSettings(const aScene: TScene; const aWriter: TXMLWriter);
        procedure AppendMaterials(const aScene: TScene; const aWriter: TXMLWriter; const aSubSet: string = 'all'; const aIDOnly: boolean = false);
        procedure AppendPrimitives(const aScene: TScene; const aWriter: TXMLWriter; const aSubSet: string = 'all'; const aIDOnly: boolean = false);
        procedure AppendSkeletons(const aScene: TScene; const aWriter: TXMLWriter; const aSubSet: string = 'all'; const aIDOnly: boolean = false);
        procedure AppendLights(const aScene: TScene; const aWriter: TXMLWriter; const aSubSet: string = 'all'; const aIDOnly: boolean = false);
        procedure AppendPaths(const aScene: TScene; const aWriter: TXMLWriter; const aSubSet: string = 'all'; const aIDOnly: boolean = false);

        procedure AppendCategory(const aCategoryName: string; const aMaterialList: TMaterialList; const aWriter: TXMLWriter);
        procedure AppendMaterial(const aMaterial: TMaterial; const aWriter: TXMLWriter);
        procedure AppendMaterialLayer(const aMaterialLayer: TTextureLayer; const anIsLightmap: Boolean; const aWriter: TXMLWriter);
        procedure AppendMaterialRaytracingProperties(const aProperties: TMaterialRaytracingProperties; const aWriter: TXMLWriter);

        procedure AppendPrimitive(const aPrimitive: TBasePrimitive; const aWriter: TXMLWriter; const aIDOnly: boolean = false);
        procedure AppendVertices(const aPrimitive: TBasePrimitive; const aWriter: TXMLWriter);
        procedure AppendPolygons(const aPrimitive: TBasePrimitive; const aWriter: TXMLWriter);
        procedure AppendPolygon(const aPrimitive: TBasePrimitive; const aPolygon: TPolygon; const aWriter: TXMLWriter);
        procedure AppendVertex(const aVertex: TVertice; const anID: integer; const aWriter: TXMLWriter);

        procedure AppendSkeleton(const aSkeleton: TSkeleton; const aWriter: TXMLWriter; const aIDOnly: boolean = false);
        procedure AppendJoints(const aSkeleton: TSkeleton; const aWriter: TXMLWriter);
        procedure AppendJoint(const aJoint: TJoint; const aJointLocation: TJointLocation; const aWriter: TXMLWriter);
        procedure AppendAnimations(const aSkeleton: TSkeleton; const aWriter: TXMLWriter);
        procedure AppendAnimation(const aAnimation: TAnimation; const aWriter: TXMLWriter);
        procedure AppendFrame(const aAnimationFrame: TAnimationFrame; const aWriter: TXMLWriter);
        procedure AppendJointLocation(const aJointLocation: TJointLocation; const aWriter: TXMLWriter);

        procedure AppendLightingProperties( const aPrimitive: TPrimitive; const aWriter: TXMLWriter);
        procedure AppendLight(const aLight: TLight; const aWriter: TXMLWriter; const aIDOnly: boolean = false);

        procedure AppendPath(aPath: TPath; const aWriter: TXMLWriter; const aIDOnly: boolean = false);
        procedure AppendControlPoint(aControlPoint: TPathControlPoint; const aWriter: TXMLWriter);

        procedure AppendComments(const aComments: string; const aWriter: TXMLWriter);
        procedure AppendTag(const aTag: String; const aWriter: TXMLWriter);
        procedure AppendColor(const aNodeName: string; const aColor: TColor; const aColorType: TColorType; const aWriter: TXMLWriter ); overload;
        procedure AppendColor(const aNodeName: string; const aColor: TRGBColor; const aWriter: TXMLWriter ); overload;

        procedure AppendViews(const aViewPortManager: TViewPortManager; const aWriter: TXMLWriter);
        procedure AppendView( aViewPort: TfrmViewPort; const aWriter: TXMLWriter; aViewName: String);
        procedure AppendBackGround(aFlatView: TfrmFlatViewPort; const awriter: TXMLWriter);
        procedure AppendBackGroundVertexPositions(aBackgroundVertice: TVertice; const awriter : TXMLWriter);
        procedure AppendCamera(aView: TfrmViewPort; const awriter: TXMLWriter);
    end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
     unit_Variables,
     Classes, SysUtils, Dialogs,
     unit_XMLSceneConstants,
     unit_Functions,
     unit_PrimitiveList;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendSettings(const aScene: TScene; const aWriter: TXMLWriter);
begin
    aWriter.AppendOpenTag('settings');
    
    try
        aWriter.AppendProperty('name', aScene.Name);
        aWriter.AppendProperty('author', aScene.Author);
        aWriter.AppendProperty('comments', aScene.Comments);
        aWriter.AppendProperty('shadowOpacity', aScene.ShadowOpacity);

        AppendColor('ambient', aScene.AmbientColor, aWriter);

    finally
        aWriter.AppendCloseTag('settings');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendMaterials(const aScene: TScene; const aWriter: TXMLWriter; const aSubSet: string; const aIDOnly: boolean);
var i: integer;
    categories: TStringList;
begin
    aWriter.AppendOpenTag('materials');

    try
        aWriter.AppendProperty('highestID', aScene.MaterialManager.HighestID);

        if aSubSet = 'all' then
        begin
            if aIDOnly then
            begin
              //We want to export only the ID's of the materials (no category tags)
              for I := 0 to aScene.MaterialManager.Count - 1 do
              begin
                aWriter.AppendOpenTag('material');
                aWriter.AppendProperty('ID', aScene.MaterialManager.Materials[I].ID);
                aWriter.AppendCloseTag('material');
              end;
            end
            else
            begin
              //We want to export all material information
              aScene.MaterialManager.UpdateMaterialUsage;

              categories := aScene.MaterialManager.ExtractCategories;
              try
                  for i := 0 to categories.Count - 1 do
                      AppendCategory(categories[i], aScene.MaterialManager, aWriter);
              finally
                  FreeAndNil(categories);
              end;
            end;
        end;

    finally
        aWriter.AppendCloseTag('materials');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendCategory(const aCategoryName: string; const aMaterialList: TMaterialList; const aWriter: TXMLWriter);
var i: integer;
    materials: TMaterialList;
begin
    aWriter.AppendOpenTag('category');

    try
        aWriter.AppendProperty('name', aCategoryName);

        materials := aMaterialList.ExtractMaterialsForCategory(aCategoryName);
        try
            for i := 0 to materials.Count - 1 do
                AppendMaterial(materials[i], aWriter);
        finally
            FreeAndNil(materials);
        end;

    finally
        aWriter.AppendCloseTag('category');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendPrimitives(const aScene: TScene; const aWriter: TXMLWriter; const aSubSet: string; const aIDOnly: boolean);
var i,j: integer;
    primitives: TPrimitiveList;
    skeleton: TSkeleton;
begin
    aWriter.AppendOpenTag('primitives');

    try
        aWriter.AppendProperty('highestID', aScene.Primitives.HighestID);

        if (aSubSet = 'selection') or (aSubSet = 'all') then begin
            if aSubSet = 'selection' then
                primitives := aScene.SelectedPrimitives.GetItems(TPrimitive)
            else
                primitives := aScene.Primitives.GetItems(TPrimitive);
            for i := 0 to primitives.Count - 1 do
                AppendPrimitive(primitives[i], aWriter, aIDOnly);
            FreeAndNil(primitives);

            // save primitives assigned to skeletons
            // TODO: take care of things when saving a selection only...
            for i := 0 to aScene.Skeletons.Count-1 do begin
                skeleton := aScene.Skeletons[i];
                for j := 0 to skeleton.AssignedPrimitives.Count-1 do
                    AppendPrimitive(skeleton.AssignedPrimitives[j], aWriter, aIDOnly);
            end;

        end;

    finally
        aWriter.AppendCloseTag('primitives');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendSkeletons(const aScene: TScene; const aWriter: TXMLWriter; const aSubSet: string = 'all'; const aIDOnly: boolean = false);
var i: integer;
    skeletons: TSkeletonList;
begin
    aWriter.AppendOpenTag('skeletons');

    try
        //aWriter.AppendProperty('highestID', aScene.Primitives.HighestID);

        if (aSubSet = 'selection') or (aSubSet = 'all') then begin
            if aSubSet = 'selection' then
                skeletons := aScene.SelectedSkeletons
            else
                skeletons := aScene.Skeletons;
            for i := 0 to skeletons.Count - 1 do
                AppendSkeleton(skeletons[i], aWriter, aIDOnly);
        end;

    finally
        aWriter.AppendCloseTag('skeletons');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendLights(const aScene: TScene; const aWriter: TXMLWriter; const aSubSet: string; const aIDOnly: boolean);
var i: integer;
    lights: TPrimitiveList;
begin
    aWriter.AppendOpenTag('lights');

    try
        aWriter.AppendProperty('highestID', aScene.Primitives.HighestID);      // lights are in the same collection as primitives

        if (aSubSet = 'selection') or (aSubSet = 'all') then begin
            if aSubSet = 'selection' then
                lights := aScene.SelectedPrimitives.GetItems(TLight)
            else
                lights := aScene.Primitives.GetItems(TLight);
            for i := 0 to lights.Count - 1 do
                AppendLight(TLight(lights[i]), aWriter, aIDOnly);
            FreeAndNil(lights);
        end;

    finally
        aWriter.AppendCloseTag('lights');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendMaterial(const aMaterial: TMaterial; const aWriter: TXMLWriter);
var i: integer;
    textureLayer: TTextureLayer;
begin
    aWriter.AppendOpenTag('material');

    try
        aWriter.AppendProperty('id', aMaterial.ID);
        aWriter.AppendProperty('name', aMaterial.Name);
        aWriter.AppendProperty('used', aMaterial.IsUsed);
        aWriter.AppendProperty('lightmap', aMaterial.EnableLightmap);
        aWriter.AppendProperty('castShadows', aMaterial.CastShadows);
        aWriter.AppendProperty('receiveShadows', aMaterial.ReceiveShadows);

        if aMaterial.HasRaytracingProperties then
            AppendMaterialRaytracingProperties(aMaterial.RaytracingProperties, aWriter);

        AppendTag(aMaterial.Tag, aWriter);
        AppendComments(aMaterial.Comments, aWriter);

        for i := 0 to aMaterial.NumTextureLayers - 1 do begin
            textureLayer := aMaterial.TextureLayers[i];

            if Assigned( textureLayer ) then
                AppendMaterialLayer(aMaterial.TextureLayers[i], i = LIGHTMAP_TEXTURE_LAYER_INDEX, aWriter);
        end;

    finally
        aWriter.AppendCloseTag('material');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendMaterialLayer(const aMaterialLayer: TTextureLayer; const anIsLightmap: Boolean; const aWriter: TXMLWriter);
begin
    aWriter.AppendOpenTag('layer');

    try
        if anIsLightmap then
            aWriter.AppendProperty('type', XML_LAYER_TYPE_LIGHTMAP)
        else
            aWriter.AppendProperty('type', XML_MATERIAL_LAYER_TYPES[aMaterialLayer.LayerType]);

        aWriter.AppendProperty('blend', XML_BLEND_OPERATIONS[aMaterialLayer.BlendOperation]);

        if aMaterialLayer.LayerType = mlTexture then begin
            aWriter.AppendOpenTag('texture');
            aWriter.AppendProperty('file', aMaterialLayer.Texture.ShortFilename);
            aWriter.AppendCloseTag('texture');
        end
        else
            AppendColor('color', aMaterialLayer.Color, ct32Bits, aWriter);

    finally
        aWriter.AppendCloseTag('layer');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendMaterialRaytracingProperties(const aProperties: TMaterialRaytracingProperties; const aWriter: TXMLWriter);
begin
    aWriter.AppendOpenTag( 'raytracing' );
    try
        aWriter.AppendConditionalProperty( 'ambientReflection', aProperties.AmbientReflection, aProperties.HasAmbientReflection  );
        aWriter.AppendConditionalProperty( 'diffuseReflection', aProperties.DiffuseReflection, aProperties.HasDiffuseReflection );
        aWriter.AppendConditionalProperty( 'specularReflection', aProperties.SpecularReflection, aProperties.HasSpecularReflection );
        aWriter.AppendConditionalProperty( 'specularExponent', aProperties.SpecularExponent, aProperties.HasSpecularExponent );
        aWriter.AppendConditionalProperty( 'reflection', aProperties.Reflection, aProperties.HasReflection );
        aWriter.AppendConditionalProperty( 'reflectiveExponent', aProperties.ReflectiveExponent, aProperties.HasReflectiveExponent );
        aWriter.AppendConditionalProperty( 'transparency', aProperties.Transparency, aProperties.HasTransparency );
        aWriter.AppendConditionalProperty( 'refraction', aProperties.Refraction, aProperties.HasRefraction );

        AppendColor( 'specularColor', aProperties.SpecularColor, ct24Bits, aWriter );
        AppendColor( 'reflectiveColor', aProperties.ReflectiveColor, ct24Bits, aWriter );
    finally
        aWriter.AppendCloseTag( 'raytracing' );
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendPrimitive(const aPrimitive: TBasePrimitive; const aWriter: TXMLWriter; const aIDOnly: boolean);
var primitive: TPrimitive;
begin
    aWriter.AppendOpenTag('primitive');

    try
        aWriter.AppendProperty('id', aPrimitive.ID);

        if not aIDOnly then begin
            aWriter.AppendProperty('name', aPrimitive.Name);
            aWriter.AppendProperty('type', XML_PRIMITIVE_TYPES[aPrimitive.ObjectType]);
            aWriter.AppendProperty('visible', aPrimitive.Visible);
            aWriter.AppendProperty('snap', XML_SNAP_METHODS[aPrimitive.SnapMethod]);
            aWriter.AppendProperty('autoUV', aPrimitive.AutoUV);
            aWriter.AppendProperty('groupID', aPrimitive.GroupID);  // added for 1.7
            aWriter.AppendProperty('skeletonID', aPrimitive.SkeletonID);

            primitive := aPrimitive as TPrimitive;
            if primitive.IsTrueSphere then
                aWriter.AppendProperty('sphere', primitive.IsTrueSphere);

            if primitive.OverrideMaterialLighting then
            begin
                AppendLightingProperties(primitive, aWriter);
            end;

            AppendComments(aPrimitive.Comments, aWriter);
            AppendTag(aPrimitive.Tag, aWriter);

            AppendVertices(aPrimitive, aWriter);
            AppendPolygons(aPrimitive, aWriter);
        end;
    finally
        aWriter.AppendCloseTag('primitive');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendLightingProperties( const aPrimitive: TPrimitive; const aWriter: TXMLWriter);
begin
    aWriter.AppendOpenTag('lighting');

    try
        aWriter.AppendProperty( 'castShadows', aPrimitive.CastShadows );
        aWriter.AppendProperty( 'receiveShadows', aPrimitive.ReceiveShadows );

    finally
        aWriter.AppendCloseTag('lighting');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendVertices(const aPrimitive: TBasePrimitive; const aWriter: TXMLWriter);
var i: integer;
    vertex: TVertice;
begin
    aWriter.AppendOpenTag('vertices');

    try
        for i := 0 to aPrimitive.NumVertices - 1 do
        begin
            vertex := aPrimitive.Vertices[i];

            try
                AppendVertex(vertex, i, aWriter);
            except
            end;
        end;

    finally
        aWriter.AppendCloseTag('vertices');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendViews(const aViewPortManager: TViewPortManager; const aWriter: TXMLWriter);
var
  I: Integer;

begin
    aWriter.AppendOpenTag('views');

    try
        aWriter.AppendProperty('numOfViewPorts', aViewPortManager.Numberof);
        for I := 0 to aViewPortManager.Numberof - 1 do             
            AppendView( aViewPortManager.ViewPorts[I], aWriter,'ViewPort'+IntToStr(I+1));

    finally
        aWriter.AppendCloseTag('views');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendView(aViewPort: TfrmViewPort;  const aWriter: TXMLWriter;aViewName: String);
begin
    aWriter.AppendOpenTag(aViewName);
    try
        aWriter.AppendProperty('viewmode', Integer(aViewPort.ViewMode));
        aWriter.AppendProperty('width', aViewPort.Width);
        aWriter.AppendProperty('height', aViewPort.Height);
        aWriter.AppendProperty('top', aViewPort.Top);
        aWriter.AppendProperty('left', aViewPort.Left);
        case aViewPort.WindowState of
            wsMinimized:
                aWriter.AppendProperty('WindowState','0');
            wsMaximized:
                aWriter.AppendProperty('WindowState', '1');
            else
               aWriter.AppendProperty('WindowState', '-1');
        end;
        aWriter.AppendProperty('SupportedViewMode',Integer(aViewPort.SupportedViewMode));
        if not((aviewport.ClassName = 'Tfrm3d') or (aviewport.ClassName = 'TfrmUVEditor')) then
            if (TfrmFlatViewPort(aviewport).ShowBGI) and (save_Load_Backgrounds) then
                AppendBackGround(TfrmFlatViewPort(aviewport),awriter);
            if save_Load_Cameras then
                AppendCamera(aviewport,awriter);
    finally
        aWriter.AppendCloseTag(aViewName);
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendBackGround(aFlatView: TfrmFlatViewPort; const awriter: TXMLWriter);
Var
    RelativePathStr : String;
  I: Integer;
begin

    aWriter.AppendOpenTag('background');
    try
        RelativePathStr := ExtractRelativePath (extractfilepath(application.exename)
                ,aFlatView.PrimitiveBGI.Material.TextureLayers[0].Texture.FileName);
        aWriter.Appendproperty('imagefile',RelativePathStr);
        aWriter.AppendopenTag('vertexPositions');
        try
            for I := 0 to aFlatView.PrimitiveBGI.Vertices.Count - 1 do
                AppendBackGroundVertexPositions(aFlatView.PrimitiveBGI.Vertices.Vertices[i],awriter);
        finally
            aWriter.AppendcloseTag('vertexPositions');
        end;

    finally
    aWriter.AppendCloseTag('background');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendBackGroundVertexPositions(aBackgroundVertice: TVertice; const awriter: TXMLWriter);
begin
    aWriter.AppendOpenTag('vertexPosition');
    try
        aWriter.AppendProperty('x', aBackgroundVertice.x);
        aWriter.AppendProperty('y', aBackgroundVertice.y);
        aWriter.AppendProperty('z', aBackgroundVertice.z);
    finally
        aWriter.AppendCloseTag('vertexPosition');
    end;
end;

// ----------------------------------------------------------------------------


procedure TXMLDataWriter.AppendCamera(aView: TfrmViewPort;  const awriter: TXMLWriter);
var
    CameraTypeStr: String;
begin
    CameraTypeStr := aView.Camera.ClassName;
    aWriter.AppendOpenTag('camera');
    aWriter.Appendproperty('Type',CameraTypeStr);
    try
        aWriter.AppendOpenTag('position');
        try
        //********HERE********HERE********HERE********HERE********HERE********HERE********HERE********HERE

        if CameraTypeStr = 'TCamera2d' then begin
            aWriter.AppendProperty('x',TCamera2d(aView.Camera).Position.x);
            aWriter.AppendProperty('y',TCamera2d(aView.Camera).Position.y);
            aWriter.AppendProperty('z',TCamera2d(aView.Camera).Position.z);
            aWriter.AppendProperty('Depth',TCamera2d(aView.Camera).Depth);
        end;

        if CameraTypeStr = 'TCameraUV' then begin
            aWriter.AppendProperty('CenterU',TCameraUV(aView.Camera).CenterU);
            aWriter.AppendProperty('CenterV',TCameraUV(aView.Camera).CenterV);
            aWriter.AppendProperty('scale',TCameraUV(aView.Camera).Scale);
        end;

        if CameraTypeStr = 'TCamera3d' then begin
            aWriter.AppendProperty('x',TCamera3d(aView.Camera).Position.x);
            aWriter.AppendProperty('y',TCamera3d(aView.Camera).Position.y);
            aWriter.AppendProperty('z',TCamera3d(aView.Camera).Position.z);
            aWriter.AppendProperty('fov',TCamera3d(aView.Camera).FOV);
            aWriter.AppendProperty('rotx',TCamera3d(aView.Camera).Rotation.x);
            aWriter.AppendProperty('roty',TCamera3d(aView.Camera).Rotation.y);
            aWriter.AppendProperty('rotz',TCamera3d(aView.Camera).Rotation.z);
        end;

        finally
            aWriter.AppendcloseTag('position');
        end;
    finally
        aWriter.AppendcloseTag('camera');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendPolygons(const aPrimitive: TBasePrimitive; const aWriter: TXMLWriter);
var i: integer;
begin
    aWriter.AppendOpenTag('polygons');

    try
        for i := 0 to aPrimitive.NumPolygons - 1 do
            AppendPolygon(aPrimitive, aPrimitive.Polygons[i], aWriter);

    finally
        aWriter.AppendCloseTag('polygons');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendVertex(const aVertex: TVertice; const anID: integer; const aWriter: TXMLWriter);
begin
    aWriter.AppendOpenTag('vertex');

    try
        aWriter.AppendProperty('id', anID);
        // if the vertex has been assigned to a joint for animation purposes,
        // write the original values, else write the regular x,y,z values
        if aVertex.JointID <> -1 then begin
            aWriter.AppendProperty('x', aVertex.OX);
            aWriter.AppendProperty('y', aVertex.OY);
            aWriter.AppendProperty('z', aVertex.OZ);
        end else begin
            aWriter.AppendProperty('x', aVertex.X);
            aWriter.AppendProperty('y', aVertex.Y);
            aWriter.AppendProperty('z', aVertex.Z);
        end;
        aWriter.AppendProperty('jointID', aVertex.JointID);

        AppendTag(aVertex.Tag, aWriter);

        if aVertex.Color <> TColor($FFFFFFFF) then
            AppendColor( 'color', aVertex.Color, ct32Bits, aWriter );

    finally
        aWriter.AppendCloseTag('vertex');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendSkeleton(const aSkeleton: TSkeleton; const aWriter: TXMLWriter; const aIDOnly: boolean = false);
begin
    aWriter.AppendOpenTag('skeleton');

    try
        aWriter.AppendProperty('id', aSkeleton.ID);

        if not aIDOnly then begin
            aWriter.AppendProperty('name', aSkeleton.Name);
            aWriter.AppendProperty('author', aSkeleton.Author);
            aWriter.AppendProperty('visible', aSkeleton.Visible);
            aWriter.AppendProperty('jointsize', aSkeleton.JointSize);

            AppendJoints(aSkeleton, aWriter);
            AppendAnimations(aSkeleton, aWriter);
        end;
    finally
        aWriter.AppendCloseTag('skeleton');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendJoints(const aSkeleton: TSkeleton; const aWriter: TXMLWriter);
var joint: TJoint;
    jointLocation: TJointLocation;
begin
    aWriter.AppendOpenTag('joints');

    try
        joint := aSkeleton.Joints[0];
        jointLocation := aSkeleton.BaseFrame.JointLocations[0];

        try
            AppendJoint(joint, jointLocation, aWriter);
        except
        end;

    finally
        aWriter.AppendCloseTag('joints');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendJoint(const aJoint: TJoint; const aJointLocation: TJointLocation; const aWriter: TXMLWriter);
var i: integer;
begin
    aWriter.AppendOpenTag('joint');

    try
        aWriter.AppendProperty('id', aJoint.ID);
        aWriter.AppendProperty('name', aJoint.Name);
        aWriter.AppendProperty('x', aJointLocation.Position.x);
        aWriter.AppendProperty('y', aJointLocation.Position.y);
        aWriter.AppendProperty('z', aJointLocation.Position.z);

        // joints form a tree; recursively write all childjoints to the DXS
        if aJoint.ChildJoints.Count>0 then begin
            aWriter.AppendOpenTag('childJoints');
            try
                for i := 0 to aJoint.ChildJoints.Count-1 do
                    AppendJoint(aJoint.ChildJoints[i], aJointLocation.ChildLocations[i], aWriter);
            finally
                aWriter.AppendCloseTag('childJoints');
            end;
        end;

    finally
        aWriter.AppendCloseTag('joint');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendAnimations(const aSkeleton: TSkeleton; const aWriter: TXMLWriter);
var i: integer;
begin
    aWriter.AppendOpenTag('animations');

    try
        for i := 0 to aSkeleton.NumberOfAnimations-1 do
            AppendAnimation(aSkeleton.Animations[i], aWriter);

    finally
        aWriter.AppendCloseTag('animations');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendAnimation(const aAnimation: TAnimation; const aWriter: TXMLWriter);
var i: integer;
begin
    aWriter.AppendOpenTag('animation');

    try
        aAnimation.AnimationFrames.ResetIDValues;

        aWriter.AppendProperty('id', aAnimation.ID);
        aWriter.AppendProperty('name', aAnimation.Name);
        aWriter.AppendProperty('author', aAnimation.Author);
        aWriter.AppendProperty('NOF', aAnimation.NumberOfFrames);
        aWriter.AppendProperty('FPS', aAnimation.FPS);

        AppendTag(aAnimation.Tag, aWriter);

        aWriter.AppendOpenTag('keyframes');
        try
            aAnimation.GotoFirstFrame;
            for i := 0 to aAnimation.NumberOfKeyFrames-1 do begin
                AppendFrame(aAnimation.SelectedFrame, aWriter);
                aAnimation.GotoNextKeyFrame;
            end;
        finally
            aWriter.AppendCloseTag('keyframes');
        end;

    finally
        aWriter.AppendCloseTag('animation');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendFrame(const aAnimationFrame: TAnimationFrame; const aWriter: TXMLWriter);
var i: integer;
begin
    aWriter.AppendOpenTag('keyframe');
    try
        aWriter.AppendProperty('id', aAnimationFrame.ID);
        aWriter.AppendProperty('isBaseFrame', aAnimationFrame.IsBaseFrame);
        aWriter.AppendProperty('isKeyFrame', aAnimationFrame.IsKeyFrame);

        aWriter.AppendOpenTag('jointLocations');
        try
            for i := 0 to aAnimationFrame.JointLocations.Count-1 do
                AppendJointLocation(aAnimationFrame.JointLocations[i], aWriter);
        finally
            aWriter.AppendCloseTag('jointLocations');
        end;

    finally
        aWriter.AppendCloseTag('keyframe');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendJointLocation(const aJointLocation: TJointLocation; const aWriter: TXMLWriter);
begin
    aWriter.AppendOpenTag('jointLocation');
    try
        aWriter.AppendProperty('id', aJointLocation.ID);
        aWriter.AppendProperty('lx', aJointLocation.Location.x);
        aWriter.AppendProperty('ly', aJointLocation.Location.y);
        aWriter.AppendProperty('lz', aJointLocation.Location.z);
        aWriter.AppendProperty('rx', aJointLocation.Rotation.x);
        aWriter.AppendProperty('ry', aJointLocation.Rotation.y);
        aWriter.AppendProperty('rz', aJointLocation.Rotation.z);

        // joints form a tree; recursively write all childjoints to the DXS
        {if aJointLocation.ChildLocations.Count>0 then begin
            aWriter.AppendOpenTag('childJointLocations');
            try
                for i := 0 to aJointLocation.ChildLocations.Count-1 do
                    AppendJointLocation(aJointLocation.ChildLocations[i], aWriter);
            finally
                aWriter.AppendCloseTag('childJointLocations');
            end;
        end; }

    finally
        aWriter.AppendCloseTag('jointLocation');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendPolygon(const aPrimitive: TBasePrimitive; const aPolygon: TPolygon; const aWriter: TXMLWriter);
var i, j, layerIndex: integer;
    uvCoordinate: TUVCoord;
    canAppend: boolean;
begin
    // NOTE, JCC: sometimes a vertex referenced by the polygon cannot be found in the vertex list of the primitive.
    // In this case, we skip this vertex here. Of course, this should not happen in the first place, but it has happened
    // with several users. DXS files referencing unknown vertices (id=-1) won't load. To avoid this problem, this little
    // hack is introduced. If all 3D editing tools are robust and 100% fail-safe, this hack shouldn't be needed.
    canAppend := true;
    for i := 0 to aPolygon.NumVertices - 1 do begin
        canAppend := aPrimitive.Vertices.IndexOf(aPolygon.VerticeList[i])<>-1;
        if not canAppend then
            Exit;
    end;

    if canAppend then begin
        aWriter.AppendOpenTag('poly');

        try
            aWriter.AppendProperty('mid', aPolygon.Material.ID);

            for i := 0 to aPolygon.NumVertices - 1 do begin
                aWriter.AppendOpenTag('vertex');
                try
                    try
                        aWriter.AppendProperty('vid', aPrimitive.Vertices.IndexOf(aPolygon.VerticeList[i]));    // vid should never be -1

                        layerIndex := 0;
                        for j := 0 to aPolygon.Material.NumTextureLayers - 1 do begin
                            if Assigned( aPolygon.Material.TextureLayers[j] ) then begin
                                uvCoordinate := aPolygon.UVCoordinates[j][i];

                                aWriter.AppendProperty('u' + IntToStr(layerIndex), uvCoordinate.U);
                                aWriter.AppendProperty('v' + IntToStr(layerIndex), uvCoordinate.V);

                                layerIndex := layerIndex + 1;
                            end;
                        end;
                    except
                        on E : Exception do
                          ShowMessage(E.ClassName+' error occured. Please notify DeleD developers with message : '+E.Message);
                    end;
                finally
                    aWriter.AppendCloseTag('vertex');
                end;
            end;

            AppendTag(aPolygon.Tag, aWriter);

        finally
            aWriter.AppendCloseTag('poly');
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendLight(const aLight: TLight; const aWriter: TXMLWriter; const aIDOnly: boolean = false);
var spotLight: TSpotLight;
begin
    aWriter.AppendOpenTag('light');
    try
        aWriter.AppendProperty('id', aLight.ID);

        if not aIDOnly then begin
            aWriter.AppendProperty('type', DescriptionForLightClass(aLight.ClassType));
            aWriter.AppendProperty('name', aLight.Name);
            aWriter.AppendProperty('visible', aLight.Visible);
            aWriter.AppendProperty('range', aLight.Range);
            aWriter.AppendProperty('active', aLight.Active);
            aWriter.AppendProperty('castShadows', aLight.CastShadows);
            aWriter.AppendProperty('snap', aLight.Snap);
            aWriter.AppendProperty('intensity', aLight.Intensity);

            AppendTag(aLight.Tag, aWriter);
            AppendComments(aLight.Comments, aWriter);

            AppendColor('diffuse', aLight.Diffuse, aWriter);
            AppendColor('specular', aLight.Specular, aWriter);

            aWriter.AppendOpenTag('position');
            aWriter.AppendProperty('x', aLight.Position.X);
            aWriter.AppendProperty('y', aLight.Position.Y);
            aWriter.AppendProperty('z', aLight.Position.Z);
            aWriter.AppendCloseTag('position');

            if (aLight is TSpotLight) or (aLight is TDirectionalLight) then begin
                aWriter.AppendOpenTag('direction');
                aWriter.AppendProperty('x', aLight.Direction.X);
                aWriter.AppendProperty('y', aLight.Direction.Y);
                aWriter.AppendProperty('z', aLight.Direction.Z);
                aWriter.AppendCloseTag('direction');
            end;

            if (aLight is TPointLight) or (aLight is TSpotLight) then begin
                aWriter.AppendOpenTag('attenuation');
                aWriter.AppendProperty('constant', aLight.Constant);
                aWriter.AppendProperty('linear', aLight.Linear);
                aWriter.AppendProperty('quadratic', aLight.Quadratic);
                aWriter.AppendCloseTag('attenuation');
            end;

            if aLight is TSpotLight then begin
                spotLight := aLight as TSpotLight;

                aWriter.AppendOpenTag('cone');
                aWriter.AppendProperty('inner', spotLight.InnerAngle);
                aWriter.AppendProperty('outer', spotLight.OuterAngle);
                aWriter.AppendProperty('falloff', spotLight.Falloff);
                aWriter.AppendCloseTag('cone');
            end;
        end;

    finally
        aWriter.AppendCloseTag('light');
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendColor(const aNodeName: string; const aColor: TColor; const aColorType: TColorType; const aWriter: TXMLWriter );
begin
    aWriter.AppendOpenTag( aNodeName );

    try
        aWriter.AppendProperty( 'r', ( aColor and $0000FF ));
        aWriter.AppendProperty( 'g', ( aColor and $00FF00 ) shr 8);
        aWriter.AppendProperty( 'b', ( aColor and $FF0000 ) shr 16);

        if aColorType = ct32Bits then
        begin
            aWriter.AppendProperty( 'a', integer((aColor and $FF000000) shr 24));
        end;

    finally
        aWriter.AppendCloseTag( aNodeName );
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendColor(const aNodeName: string; const aColor: TRGBColor; const aWriter: TXMLWriter );
begin
    aWriter.AppendOpenTag( aNodeName );
    try
        aWriter.AppendProperty( 'r', Integer(Round( aColor.R * 255 )));
        aWriter.AppendProperty( 'g', Integer(Round( aColor.G * 255 )));
        aWriter.AppendProperty( 'b', Integer(Round( aColor.B * 255 )));
    finally
        aWriter.AppendCloseTag( aNodeName );
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendTag(const aTag: String; const aWriter: TXMLWriter);
begin
    if aTag <> '' then begin
        aWriter.AppendOpenTag('tag');
        try
            aWriter.AppendText(aTag);
        finally
            aWriter.AppendCloseTag('tag');
        end;
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendComments(const aComments: string; const aWriter: TXMLWriter);
begin
    if aComments <> '' then begin
        aWriter.AppendOpenTag('comments');
        try
            aWriter.AppendText(aComments);
        finally
            aWriter.AppendCloseTag('comments');
        end;
    end;
end;

// ----------------------------------------------------------------------------

function TXMLDataWriter.DescriptionForLightClass(const aLightClass: TClass): string;
var i: integer;
begin
    result := '';

    for i := Low(XML_LIGHT_CLASS_INFOS) to High(XML_LIGHT_CLASS_INFOS) do
    begin
        if XML_LIGHT_CLASS_INFOS[i].LightClass = aLightClass then
        begin
            result := XML_LIGHT_CLASS_INFOS[i].Name;
            break;
        end;
    end;

    if result = '' then
    begin
        raise Exception.Create(Format('Unknown light class "%s".', [aLightClass.ClassName]));
    end;
end;

// ----------------------------------------------------------------------------

procedure TXMLDataWriter.AppendPaths(const aScene: TScene;
  const aWriter: TXMLWriter; const aSubSet: string; const aIDOnly: boolean);
var
  I: Integer;
  Paths: TPrimitiveList;
begin
    aWriter.AppendOpenTag('paths');

    try
        aWriter.AppendProperty('highestID', aScene.Primitives.HighestID);

        if (aSubSet = 'selection') or (aSubSet = 'all') then
        begin
            if aSubSet = 'selection' then
                Paths := aScene.SelectedPrimitives.GetItems(TPath)
            else
                Paths := aScene.Primitives.GetItems(TPath);

            for i := 0 to Paths.Count - 1 do
                AppendPath(TPath(Paths[i]), aWriter, aIDOnly);
        end;

    finally
        aWriter.AppendCloseTag('paths');
        FreeAndNil(Paths);
    end;
end;

procedure TXMLDataWriter.AppendPath(aPath: TPath; const aWriter: TXMLWriter; const aIDOnly: boolean);
var
  I: Integer;
begin
    aWriter.AppendOpenTag('path');

    try
        aWriter.AppendProperty('id', aPath.ID);

        if not aIDOnly then
        begin
            aWriter.AppendProperty('name', aPath.Name );
            aWriter.AppendProperty('type', PathTypeToString( aPath.PathType ) );
            aWriter.AppendProperty('isclosed', aPath.IsClosed);
            aWriter.AppendProperty('visible', aPath.Visible);
            aWriter.AppendProperty('snap', XML_SNAP_METHODS[aPath.SnapMethod]);
            aWriter.AppendProperty('groupID', aPath.GroupID);

            //Add comments and the Tag
            AppendComments(aPath.Comments, aWriter);
            AppendTag(aPath.Tag, aWriter);

            //Write controlpoint data
            aWriter.AppendOpenTag('controlpoints');
            try
              for I := 0 to aPath.ControlPointCount - 1 do
                AppendControlPoint(aPath.ControlPoints[I], aWriter);
            finally
              aWriter.AppendCloseTag('controlpoints');
            end;

        end;
    finally
        aWriter.AppendCloseTag('path');
    end;
end;

procedure TXMLDataWriter.AppendControlPoint(aControlPoint: TPathControlPoint; const aWriter: TXMLWriter);
begin
  aWriter.AppendOpenTag('controlpoint');
  try
     aWriter.AppendProperty('ID', aControlPoint.ID);

     //Add tag
     AppendTag(aControlPoint.Tag, aWriter);

     //Write position
     aWriter.AppendOpenTag('position');
     aWriter.AppendProperty('X',aControlPoint.Position.x);
     aWriter.AppendProperty('Y',aControlPoint.Position.y);
     aWriter.AppendProperty('Z',aControlPoint.Position.z);
     aWriter.AppendCloseTag('position');

     //Write upvector
     aWriter.AppendOpenTag('up');
     aWriter.AppendProperty('X',aControlPoint.UpVector.x);
     aWriter.AppendProperty('Y',aControlPoint.UpVector.y);
     aWriter.AppendProperty('Z',aControlPoint.UpVector.z);
     aWriter.AppendCloseTag('up');

     //Write first tangent
     aWriter.AppendOpenTag('tangent1');
     aWriter.AppendProperty('X',aControlPoint.TangentVector1.x);
     aWriter.AppendProperty('Y',aControlPoint.TangentVector1.y);
     aWriter.AppendProperty('Z',aControlPoint.TangentVector1.z);
     aWriter.AppendProperty('userdefined', aControlPoint.UserDefinedTangent1);
     aWriter.AppendCloseTag('tangent1');

     //write second tangent
     aWriter.AppendOpenTag('tangent2');
     aWriter.AppendProperty('X',aControlPoint.TangentVector2.x);
     aWriter.AppendProperty('Y',aControlPoint.TangentVector2.y);
     aWriter.AppendProperty('Z',aControlPoint.TangentVector2.z);
     aWriter.AppendProperty('userdefined', aControlPoint.UserDefinedTangent2);
     aWriter.AppendCloseTag('tangent2');     
  finally
    aWriter.AppendCloseTag('controlpoint');
  end;
end;

end.

