{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_XMLPluginReader;

{$MODE Delphi}

interface

uses
    Classes,
    unit_Scene,
    unit_XMLSceneConstants,
    unit_XMLReader,
    unit_XMLDataReader,
    unit_Primitives,
    unit_Materials,
    unit_Paths;

type
    // the XMLPluginReader processes XML scene data originating from plugins;
    // it can add and replace scene items like Primitives, light, materials etc
    TXMLPluginReader = class(TXMLDataReader)
    private
        procedure ReplacePrimitive(const aNode: TXMLNode; const aPrimitive: TBasePrimitive; const aMaterialList: TStringList; const aScene: TScene);
        procedure ReplaceMaterial(const aNode: TXMLNode; const aOldMaterial: TMaterial; const aCategoryName: string; const aMaterialList: TStringList; const aScene: TScene);
        procedure ReplacePath(const aNode: TXMLNode; const aOldPath: TPath; const aScene: TScene);
        function AppendMaterial(const aNode: TXMLNode; const aCategoryName: string; const aMaterialList: TStringList; const aScene: TScene): TMaterial;
    protected
        procedure ProcessPrimitives(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene); override;
        procedure ProcessMaterial(const aNode: TXMLNode; const aCategoryName: string; const aMaterialList: TStringList; const aScene: TScene); override;
        procedure ProcessPaths(const aNode: TXMLNode; const aScene: TScene); override;
    public
        procedure ReadFromString(const aXMLString: string; const aScene: TScene);
    end;

implementation

uses
    SysUtils, Windows, IniFiles,
    unit_MaterialList,
    unit_ParamList,
    unit_Enums,
    unit_FaceMaterialReplacerTool;

// ### TXMLPluginReader ########################################################

procedure TXMLPluginReader.ReadFromString(const aXMLString: string; const aScene: TScene);
var reader: TXMLStringReader;
    node: TXMLNode;
begin
    node := nil;
    reader := TXMLStringReader.Create(aXMLString);
    try
        reader.OnProgress := HandleProgress;

        node := reader.RootNode;
        ReadScene(node, aScene);
    finally
        FreeAndNil(reader);
        FreeAndNil(node);
    end;
end;

procedure TXMLPluginReader.ProcessPrimitives(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene);
var i: integer;
    node: TXMLNode;
    primitive: TBasePrimitive;
begin
    ProgressForm.Description := 'Loading Primitives.';

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'primitive' then
        begin
            primitive := aScene.Primitives.GetPrimitiveByID(node.IntegerProperties['id']);
            if Assigned(primitive) then
                ReplacePrimitive(node, primitive, aMaterialList, aScene)
            else
                AppendPrimitive(node, aMaterialList, aScene);
        end;
    end;
end;

procedure TXMLPluginReader.ProcessPaths(const aNode: TXMLNode; const aScene: TScene);
var i: integer;
    node: TXMLNode;
    path: TBasePrimitive;
begin
    ProgressForm.Description := 'Loading Paths.';

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'path' then
        begin
            //If path allready existed, replace it with the version returned by the plugin
            path := aScene.Primitives.GetPrimitiveByID(node.IntegerProperties['id']);
            if Assigned(path) then
                ReplacePath(node, TPath(path), aScene)
            else
                AppendPath(node, aScene);
        end;
    end;
end;

procedure TXMLPluginReader.ReplacePrimitive(const aNode: TXMLNode; const aPrimitive: TBasePrimitive; const aMaterialList: TStringList; const aScene: TScene);
var i: integer;
    node: TXMLNode;
    vertices: TStringList;
begin
    aPrimitive.Clear;
    aPrimitive.ID := aNode.IntegerProperties['id'];
    aPrimitive.Name := aNode.Properties['name'];
    aPrimitive.ObjectType := ObjectTypeForDescription(aNode.Properties['type']);
    aPrimitive.Visible := aNode.BooleanProperties['visible'];
    aPrimitive.SnapMethod := SnapMethodForDescription(aNode.Properties['snap']);
    aPrimitive.AutoUV := aNode.BooleanProperties['autoUV'];

    vertices := THashedStringList.Create;
    try
        for i := 0 to aNode.NumChildNodes - 1 do
        begin
            node := aNode.ChildNodes[i];

            if node.Name = 'comments' then begin
                aPrimitive.Comments := node.Text;
            end else if node.Name = 'tag' then begin
                aPrimitive.Tag := node.Text;
            end else if node.Name = 'vertices' then begin
                AppendVertices(node, aPrimitive, vertices);
            end else if node.Name = 'polygons' then  begin
                AppendPolygons(node, aPrimitive, vertices, aMaterialList, aScene);
            end;
        end;
    finally
        FreeAndNil(vertices);
    end;
end;

procedure TXMLPluginReader.ProcessMaterial(const aNode: TXMLNode; const aCategoryName: string; const aMaterialList: TStringList; const aScene: TScene);
var material: TMaterial;
begin
    UpdateProgressForNode( aNode );

    material := aScene.MaterialManager.GetMaterialByID(aNode.IntegerProperties['id']);
    if Assigned(material) then
        ReplaceMaterial(aNode, material, aCategoryName, aMaterialList, aScene)
    else
        AppendMaterial(aNode, aCategoryName, aMaterialList, aScene);
end;

procedure TXMLPluginReader.ReplaceMaterial(const aNode: TXMLNode; const aOldMaterial: TMaterial; const aCategoryName: string; const aMaterialList: TStringList; const aScene: TScene);
var newMaterial: TMaterial;
    materialList: TMaterialList;
    fParamList: TParamList;
begin
    newMaterial := AppendMaterial(aNode, aCategoryName, aMaterialList, aScene);

    // replacing a material with another one means that all polygons using the
    // original material are to be updated with the new material
    materialList := TMaterialList.Create(false);
    materialList.Add(aOldMaterial);

    fParamList := TParamList.Create(false);
    fParamList.Add(materialList,'currentMaterials',true);
    fParamList.Add(newMaterial,'newMaterial');
    fParamList.Add(false,'doSelectionOnly');
    aScene.ToolExecuter.Execute(ptFaceMaterialReplacer,taNotApplicable,fParamList);
    FreeAndNil(fParamList);

    FreeAndNil(materialList);
    aScene.MaterialManager.Remove(aOldMaterial);
end;

procedure TXMLPluginReader.ReplacePath(const aNode: TXMLNode;
                                       const aOldPath: TPath;
                                       const aScene: TScene);
var i: integer;
    node: TXMLNode;
begin
    aOldPath.Clear;

    //Re-write all path data
    aOldPath.ID         := aNode.IntegerProperties['id'];
    aOldPath.Name       := aNode.Properties['name'];
    aOldPath.PathType   := StringToPathType( aNode.Properties['type'] );
    aOldPath.IsClosed   := aNode.BooleanProperties['isclosed'];
    aOldPath.Visible    := aNode.BooleanProperties['visible'];
    aOldPath.SnapMethod := SnapMethodForDescription(aNode.Properties['snap']);
    aOldPath.GroupID    := aNode.IntegerProperties['groupID'];

    //Read comments, tag and controlpoints
    for I := 0 to aNode.NumChildNodes - 1 do
    begin
      node := aNode.ChildNodes[i];

      if node.Name = 'comments' then aOldPath.Comments := node.Text;
      if node.Name = 'tag' then aOldPath.Tag := node.Text;
      if node.Name = 'controlpoints' then AppendControlPoints(node, aOldPath);
    end;
end;


// Keep in synch with TXMLSceneReader.ProcessMaterial! (TODO, PJ: Small rewrite needed, 'keep in synch' = evil :P)
function TXMLPluginReader.AppendMaterial(const aNode: TXMLNode; const aCategoryName: string; const aMaterialList: TStringList; const aScene: TScene): TMaterial;
var i: integer;
    node: TXMLNode;
    material: TMaterial;
begin
    material := TMaterial.Create;

     // save the material under its ID value coming from the XML
    aMaterialList.AddObject(aNode.Properties['id'],material);

    material.ID := aNode.IntegerProperties['id'];
    material.Category := aCategoryName;
    material.Name := aNode.Properties['name'];
    material.IsUsed := aNode.BooleanProperties['used'];

    if aNode.HasProperty('lightmap') then
        material.EnableLightmap := aNode.BooleanProperties['lightmap']; 

    if aNode.HasProperty('castShadows') then
        material.CastShadows := aNode.BooleanProperties['castShadows'];

    if aNode.HasProperty('receiveShadows') then
        material.ReceiveShadows := aNode.BooleanProperties['receiveShadows'];

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'comments' then begin
            material.Comments := node.Text;
        end
        else if node.Name = 'tag' then begin
            material.Tag := node.Text;
        end
        else if node.Name = 'raytracer' then begin
            AppendMaterialRaytracingProperties(node, material);
        end
        else if node.name = 'layer' then
        begin
            AppendMaterialLayer(node, material, aScene);
        end;
    end;
    aScene.MaterialManager.Add(material);

    Result := material;
end;

end.
