{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_XMLReader;

{$MODE Delphi}

(* ==========================================================================
   Basic XML Reader. The reader consists of the following components:

   (for public use)
   - TXMLReader: Invokes the parser/scanner and returns the root XML node
   - TXMLFileReader: Convenience implementation of TXMLReader for file access
   - TXMLNode: Hierarchical representation of an XML node. This is the
               component the calling application will need to process.

   (for private use)
   - TXMLScanner: Splits the XML text up into separate tokens ('tokenizer')
   - TXMLParser: Processes the tokens into XML Nodes

   - Usage:
     1) Create a reader
     2) Read the rootnode
     3) You may now discard the reader
     4) parse the information from the node structure
     5) Free the rootnode!

   - Notes:
     In most DOM implementations, 'text' is stored in  separate childnode.
     However, we choose to flatten this out and store the text in the
     containing node.
   ========================================================================== *)

interface

uses
  Contnrs, Classes,
  unit_StringDictionary;

type
  TToken = record
      ID: integer;
      Text: string;
      Line: integer;
      Column: integer;
      Position: integer;
  end;

  TXMLToken = (
      xmltError,
      xmltTag,
      xmltPropertyName,
      xmltPropertyValue,
      xmltText,
      xmltTagEnd,
      xmltTagSingle
  );

  TProcessCharacterMethod = procedure( const aCharacter: char ) of object;
  TProgressHandler = procedure( const aPosition, aTotalSize: integer ) of object;

  TXMLNode = class;
// ----------------------------------------------------------------------------

  TXMLReader = class( TObject )
  private
      fStream: TStream;
      fTotalSize: integer;
      fOnProgress: TProgressHandler;

      fRootNode: TXMLNode;

      procedure Resolve;

      procedure HandleProgress( const aPosition, aTotalSize: integer );
      
  public
      constructor Create( const aStream: TStream );

      function RootNode: TXMLNode;

      property TotalSize: integer read fTotalSize;
      property OnProgress: TProgressHandler read fOnProgress write fOnProgress;
  end;

// ----------------------------------------------------------------------------

  TXMLFileReader = class( TXMLReader )
  private
      fStream: TStream;

  public
      constructor Create( const aFileName: string );
      destructor Destroy; override;
  end;

// ----------------------------------------------------------------------------

  TXMLStringReader = class( TXMLReader )
  private
      fStream: TStream;

  public
      constructor Create( const aString: string );
      destructor Destroy; override;
  end;

// ----------------------------------------------------------------------------

  TXMLNode = class( TObject )
  private
      fName: string;
      fText: string;
      fPosition: integer;

      fChildNodes: TObjectList;
      fProperties: TStringDictionary;

      function GetChildNode( const anIndex: integer ): TXMLNode;

      function GetProperty( const aName: string ): string;
      procedure SetProperty( const aName, aValue: string );

      function GetBooleanProperty( const aName: string ): boolean;
      function GetIntegerProperty( const aName: string ): integer;
      function GetFloatProperty( const aName: string ): single;

  public
      constructor Create( const aName: string; const aPosition: integer );
      destructor Destroy; override;

      procedure AddChildNode( const aChildNode: TXMLNode );
      function NumChildNodes: integer;

      function HasProperty( const aName: string ): boolean;

      property Name: string read fName;
      property Text: string read fText write fText;
      property Position: integer read fPosition; // position of node inside original stream

      property ChildNodes[const anIndex: integer]: TXMLNode read GetChildNode;
      property Properties[const aName: string]: string read GetProperty write SetProperty;
      property BooleanProperties[const aName: string]: boolean read GetBooleanProperty;
      property IntegerProperties[const aName: string]: integer read GetIntegerProperty;
      property FloatProperties[const aName: string]: single read GetFloatProperty;
  end;

// ----------------------------------------------------------------------------

  TXMLScanner = class( TObject )
  private
      fText: string;
      fTextSize: integer;
      fIndex: integer;
      fTokenID: integer;
      fLine: integer;
      fColumn: integer;
      fPosition: integer;
      fTokenStart: integer;
      fTextCharacterFound: boolean;

      fProcessCharacter: TProcessCharacterMethod;
      fStartMethod: TProcessCharacterMethod;
      fErrorMethod: TProcessCharacterMethod;

      function HasRemainingCharacters: Boolean;
      function NextCharacter: Char;

      procedure StartToken;
      function TokenText: string;

      function IsInErrorState: boolean;
      function IsInEndState: boolean;
      function IsIgnoreSymbol( const aCharacter: char ): boolean;
      function IsTagNameSymbol( const aCharacter: char ): boolean;
      function IsPropertyNameSymbol( const aCharacter: char ): boolean;
      function IsPropertyValueSymbol( const aCharacter: char ): boolean;

      procedure ProcessStart( const aCharacter: char );
      procedure ProcessError( const aCharacter: char );

      procedure ProcessTagStart( const aCharacter: char );
      procedure ProcessTagName( const aCharacter: char );
      procedure ProcessCloseTagNameStart( const aCharacter: char );
      procedure ProcessCloseTagName( const aCharacter: char );
      procedure ProcessCloseSingleTagName( const aCharacter: char );
      procedure ProcessTagNameEnd( const aCharacter: char );

      procedure ProcessPropertyName( const aCharacter: char );
      procedure ProcessPropertyAssign( const aCharacter: char );
      procedure ProcessPropertyAssignEnd( const aCharacter: char );
      procedure ProcessPropertyValueStart( const aCharacter: char );
      procedure ProcessPropertyValue( const aCharacter: char );
      procedure ProcessPropertyValueEnd( const aCharacter: char );

      procedure ProcessCommentStart( const aCharacter: char );
      procedure ProcessCommentStart1( const aCharacter: char );
      procedure ProcessComment( const aCharacter: char );
      procedure ProcessCommentEnd( const aCharacter: char );
      procedure ProcessCommentEnd1( const aCharacter: char );
      procedure ProcessText( const aCharacter: char );

      // INFO, PJ: This 'processCharacter' method represents the scanner state
      property ProcessCharacter: TProcessCharacterMethod read fProcessCharacter write fProcessCharacter;

  public
      constructor Create( const aText: string );

      function NextToken: TToken;
      property TextSize: integer read fTextSize;
  end;

// ----------------------------------------------------------------------------

  TXMLParser = class( TObject )
  private
      fScanner: TXMLScanner;
      fRootNode: TXMLNode;
      fNodes: TObjectStack;
      fOnProgress: TProgressHandler;
      fUpdateCount: integer;

      fPropertyName: string;

      procedure Resolve;
      procedure HandleToken( const aToken: TToken );

      procedure UpdateProgress( const aPosition, aTotalSize: integer ); 

      procedure AddTreeNode( const aNode: TXMLNode );

      function XMLDecode( const aText: string ): string;
      function ConvertSpecialSequence( const aSequence: string ): string;

  public
      constructor Create( const aScanner: TXMLScanner );
      destructor Destroy; override;

      function RootNode: TXMLNode;

      property OnProgress: TProgressHandler read fOnProgress write fOnProgress;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

implementation

uses
  Math, SysUtils, Windows;

const
  NO_TOKEN = -1;
  TAB = #9;
  LF = #10;
  CR = #13;
  SPACE = #32;

  LOWERCASE_CHARACTER_SET = ['a'..'z'];
  UPPERCASE_CHARACTER_SET = ['A'..'Z'];
  DECIMAL_DIGITS_CHARACTER_SET = ['0'..'9'];

constructor TXMLReader.Create( const aStream: TStream );
begin
  inherited Create;

  fStream := aStream;
  fTotalSize := aStream.Size;
end;

// ----------------------------------------------------------------------------

procedure TXMLReader.Resolve;
var
  text: string;
  scanner: TXMLScanner;
  parser: TXMLParser;

begin
  if not Assigned( fRootNode ) then
  begin
      SetLength( text, fStream.Size );
      fStream.Read( text[1], fStream.Size );

      scanner := TXMLScanner.Create( Trim( text ));
      try
          parser := TXMLParser.Create( scanner );
          parser.OnProgress := HandleProgress;
          try
              fRootNode := parser.RootNode;
          finally
              FreeAndNil( parser );
          end;
      finally
          FreeAndNil( scanner );
      end;

      if fRootNode = nil then
          raise Exception.create( 'No rootnode found.' );
  end;
end;

// ----------------------------------------------------------------------------

procedure TXMLReader.HandleProgress( const aPosition, aTotalSize: integer );
begin
    if Assigned( fOnProgress ) then
        fOnProgress( aPosition, aTotalSize );
end;

// ----------------------------------------------------------------------------

function TXMLReader.RootNode: TXMLNode;
begin
  Resolve;

  result := fRootNode;
end;

// ----------------------------------------------------------------------------

constructor TXMLFileReader.Create( const aFileName: string );
begin
  fStream := TFileStream.Create( aFileName, fmOpenRead {or fmShareDenyNone} );

  inherited Create( fStream );
end;

// ----------------------------------------------------------------------------

destructor TXMLFileReader.Destroy;
begin
  FreeAndNil( fStream );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

constructor TXMLStringReader.Create(const aString: string);
begin
    fStream := TStringStream.Create(aString);

    inherited Create(fStream);
end;

destructor TXMLStringReader.Destroy;
begin
    FreeAndNil(fStream);

    inherited Destroy;
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TXMLNode.Create( const aName: string; const aPosition: integer );
begin
  inherited Create;

  fName := aName;
  fPosition := aPosition;
end;

// ----------------------------------------------------------------------------

destructor TXMLNode.Destroy;
begin
  FreeAndNil( fChildNodes );
  FreeAndNil( fProperties );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

function TXMLNode.GetChildNode( const anIndex: integer ): TXMLNode;
begin
  if Assigned( fChildNodes ) then
  begin
      result := fChildNodes[anIndex] as TXMLNode;
  end
  else
      result := nil;
end;

// ----------------------------------------------------------------------------

function TXMLNode.GetProperty( const aName: string ): string;
begin
  if  (not Assigned( fProperties )) or
      (not fProperties.ContainsKey( aName )) then
  begin
      raise Exception.Create( Format( 'Property "%s" not found for node "%s".', [aName, Name] ));
  end
  else
      result := fProperties[aName];
end;

// ----------------------------------------------------------------------------

procedure TXMLNode.SetProperty( const aName, aValue: string );
begin
  if not Assigned( fProperties ) then
  begin
      fProperties := TStringDictionary.Create;
  end;

  fProperties[aName] := aValue;  
end;

// ----------------------------------------------------------------------------

function TXMLNode.GetBooleanProperty( const aName: string ): boolean;
var
  stringValue: string;

begin
  stringValue := LowerCase( Properties[aName] );

  result :=
      (stringValue = 'true') or
      (stringValue = '1') or
      (stringValue = 'yes')
end;

// ----------------------------------------------------------------------------

function TXMLNode.GetIntegerProperty( const aName: string ): integer;
begin
    result := StrToInt( Properties[aName] );
end;

// ----------------------------------------------------------------------------

function TXMLNode.GetFloatProperty( const aName: string ): single;
var value: string;
begin
    value := LowerCase( Properties[aName] );

    if (value = 'nan') or (value = 'inf') or (value = '-inf') then
        result := 0
    else
        result := StrToFloat( value );
end;

// ----------------------------------------------------------------------------

procedure TXMLNode.AddChildNode( const aChildNode: TXMLNode );
begin
  if not Assigned( fChildNodes ) then
  begin
      fChildNodes := TObjectList.Create( TRUE );
  end;

  fChildNodes.Add( aChildNode );
end;

// ----------------------------------------------------------------------------

function TXMLNode.NumChildNodes: integer;
begin
  if Assigned( fChildNodes ) then
  begin
      result := fChildNodes.Count;
  end
  else
      result := 0;
end;

// ----------------------------------------------------------------------------

function TXMLNode.HasProperty( const aName: string ): boolean;
begin
    result :=
        Assigned( fProperties ) and
        fProperties.ContainsKey( aName );
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TXMLScanner.Create( const aText: string );
begin
  inherited Create;

  fText := aText;
  fTextSize := Length( fText );
  
  fIndex := 1;
  fLine := 1;
  fColumn := 1;
  fPosition := 1;

  fStartMethod := processStart;
  fErrorMethod := processError;

  fProcessCharacter := fStartMethod;
end;

// ----------------------------------------------------------------------------

function TXMLScanner.NextToken: TToken;
var
  character: char;
  errorText: string;

begin
  fTokenID := NO_TOKEN;

  while HasRemainingCharacters and (fTokenID = NO_TOKEN) do
  begin
      character := NextCharacter;

      if character = CR then
      begin
          fLine := fLine + 1;
          fColumn := 1;
      end
      else if character <> LF then
      begin
          fColumn := fColumn + 1;
      end;

      fPosition := fPosition + 1;

      ProcessCharacter( character );

      if IsInErrorState then
      begin
          errorText := Copy( fText, Max( fIndex - 10, 0 ), 10 );

          raise Exception.create(
              'Unexpected character ''' + character + ''' in input (...' + errorText + '), position (' +
              IntToStr( fLine ) + ', ' + IntToStr( fColumn ) + ').' );
      end;
  end;

  if not IsInEndState and not HasRemainingCharacters then
  begin
      raise Exception.create( Format( 'Unexpected end of input (%d, %d).', [fLine, fColumn] ));
  end;

  result.id := fTokenID;

  if fText <> '' then
  begin
      result.Text := TokenText;
  end
  else
      result.Text := '';

  result.Line := fLine;
  result.Column := fColumn;
  result.Position := fPosition;
end;

// ----------------------------------------------------------------------------

function TXMLScanner.HasRemainingCharacters: Boolean;
begin
  result := ( fIndex <= fTextSize );
end;

// ----------------------------------------------------------------------------

function TXMLScanner.NextCharacter: Char;
begin
	result := fText[fIndex];
	fIndex := fIndex + 1;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.StartToken;
begin
  fTokenStart := fIndex - 1;
end;

// ----------------------------------------------------------------------------

function TXMLScanner.TokenText: string;
begin
  result := Copy( fText, fTokenStart, fIndex - fTokenStart - 1 );
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessStart( const aCharacter: Char );
begin
  StartToken;

  if aCharacter = '<' then
  begin
      fProcessCharacter := ProcessTagStart;
  end
  else
  begin
      fProcessCharacter := ProcessText;
      fTextCharacterFound := not IsIgnoreSymbol( aCharacter );
  end;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessError( const aCharacter: Char );
begin
  // TODO, PJ: Implement;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessTagName( const aCharacter: char );
begin
  if IsTagNameSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessTagName;
  end
  else if aCharacter = '>' then
  begin
      fTokenID := ord( xmltTag );

      fProcessCharacter := ProcessStart;
  end
  else if IsIgnoreSymbol( aCharacter ) then
  begin
      fTokenID := ord( xmltTag );

      fProcessCharacter := ProcessTagNameEnd;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessTagStart( const aCharacter: char );
begin
  StartToken;

  if IsTagNameSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessTagName;
  end
  else if aCharacter = '/' then
  begin
      fProcessCharacter := ProcessCloseTagNameStart;
  end
  else if aCharacter = '!' then
  begin
      fProcessCharacter := ProcessCommentStart;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessTagNameEnd( const aCharacter: char );
begin
  StartToken;

  if aCharacter = '/' then
  begin
      fTokenID := ord( xmltTagSingle );

      fProcessCharacter := ProcessCloseSingleTagName;
  end
  else if IsPropertyNameSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessPropertyName;
  end
  else if aCharacter = '>' then
  begin
      fTokenID := ord( xmltTag );

      fProcessCharacter := ProcessStart;
  end
  else if IsIgnoreSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessTagNameEnd;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessCloseTagName( const aCharacter: char );
begin
  if IsTagNameSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessCloseTagName;
  end
  else if aCharacter = '>' then
  begin
      fTokenID := ord( xmltTagEnd );

      fProcessCharacter := ProcessStart;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessCloseSingleTagName( const aCharacter: char );
begin
  if aCharacter = '>' then
  begin
      fProcessCharacter := ProcessStart;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessCloseTagNameStart( const aCharacter: char );
begin
  StartToken;

  fProcessCharacter := ProcessCloseTagName;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessPropertyName( const aCharacter: char );
begin
  if IsPropertyNameSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessPropertyName;
  end
  else if IsIgnoreSymbol( aCharacter ) then
  begin
      fTokenID := ord( xmltPropertyName );

      fProcessCharacter := ProcessPropertyAssign;
  end
  else if aCharacter = '=' then
  begin
      fTokenID := ord( xmltPropertyName );

      fProcessCharacter := ProcessPropertyAssignEnd;
  end
  else
      fProcessCharacter := processError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessPropertyAssign( const aCharacter: char );
begin
  if IsIgnoreSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessPropertyAssign;
  end
  else if aCharacter = '=' then
  begin
      fProcessCharacter := ProcessPropertyAssignEnd;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessPropertyAssignEnd( const aCharacter: char );
begin
  if IsIgnoreSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessPropertyAssignEnd;
  end
  else if aCharacter = '"' then
  begin
      fProcessCharacter := ProcessPropertyValueStart;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessPropertyValue( const aCharacter: char );
begin
  if IsPropertyValueSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessPropertyValue;
  end
  else if aCharacter = '"' then
  begin
      fTokenID := ord( xmltPropertyValue );

      fProcessCharacter := ProcessPropertyValueEnd;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessPropertyValueStart( const aCharacter: char );
begin
  StartToken;

  if aCharacter = '"' then
  begin
      fTokenID := ord( xmltPropertyValue );
      fProcessCharacter := ProcessPropertyValueEnd;
  end
  else
      fProcessCharacter := ProcessPropertyValue;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessPropertyValueEnd( const aCharacter: char );
begin
  StartToken;

  if IsPropertyNameSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessPropertyName;
  end
  else if IsIgnoreSymbol( aCharacter ) then
  begin
      fProcessCharacter := ProcessPropertyValueEnd;
  end
  else if aCharacter = '/' then
  begin
      fTokenID := ord( xmltTagEnd );

      fProcessCharacter := ProcessCloseSingleTagName;
  end
  else if aCharacter = '>' then
  begin
      fProcessCharacter := ProcessStart;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessComment( const aCharacter: char );
begin
  if aCharacter = '-' then
  begin
      fProcessCharacter := ProcessCommentEnd1;
  end
  else
      fProcessCharacter := ProcessComment;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessCommentStart( const aCharacter: char );
begin
  if aCharacter = '-' then
  begin
      fProcessCharacter := ProcessCommentStart1;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessCommentStart1( const aCharacter: char );
begin
  if aCharacter = '-' then
  begin
      fProcessCharacter := ProcessComment;
  end
  else
      fProcessCharacter := ProcessError;
end;

// ----------------------------------------------------------------------------

procedure TXMLScanner.ProcessCommentEnd1( const aCharacter: char );
begin
  if aCharacter = '-' then
  begin
      fProcessCharacter := ProcessCommentEnd;
  end
  else
      fProcessCharacter := ProcessComment;
end;

(* ========================================================================= *)

procedure TXMLScanner.ProcessCommentEnd( const aCharacter: char );
begin
  if aCharacter = '>' then
  begin
      fProcessCharacter := ProcessStart;
  end
  else
      fProcessCharacter := ProcessComment;
end;

(* ========================================================================= *)

procedure TXMLScanner.ProcessText( const aCharacter: char );
begin
  if (aCharacter <> '<') and HasRemainingCharacters then
  begin
      if not IsIgnoreSymbol( aCharacter ) then
      begin
          fTextCharacterFound := TRUE;
      end;

      fProcessCharacter := ProcessText;
  end
  else
  begin
      if fTextCharacterFound then
      begin
          fTokenID := ord( xmltText );
      end;

      if HasRemainingCharacters then
      begin
          fProcessCharacter := ProcessTagStart;
      end
      else // INFO, PJ: 'processStart' = end.
          fProcessCharacter := ProcessStart;
  end;
end;

// ----------------------------------------------------------------------------

function TXMLScanner.IsIgnoreSymbol( const aCharacter: Char ): boolean;
begin
  result := aCharacter in [SPACE, CR, LF, TAB];
end;

// ----------------------------------------------------------------------------

function TXMLScanner.IsTagNameSymbol( const aCharacter: char ): boolean;
begin
  result :=
      (aCharacter in LOWERCASE_CHARACTER_SET) or
      (aCharacter in UPPERCASE_CHARACTER_SET) or
      (aCharacter in DECIMAL_DIGITS_CHARACTER_SET) or
      (aCharacter = '_' ) or
      (aCharacter = '.' ) or
      (aCharacter = '-' );
end;

// ----------------------------------------------------------------------------

function TXMLScanner.IsPropertyNameSymbol( const aCharacter: char ): boolean;
begin
  result := IsTagNameSymbol( aCharacter );
end;

// ----------------------------------------------------------------------------

function TXMLScanner.IsPropertyValueSymbol( const aCharacter: char ): boolean;
begin
  result := (aCharacter <> '"');
end;

// ----------------------------------------------------------------------------

// INFO, PJ: ProcessCharacter method stores scanner state.
//           This convenience method checks this state.
function TXMLScanner.IsInErrorState: boolean;
begin
  result := (TMethod(fProcessCharacter).code = TMethod(fErrorMethod).code);
end;

// ----------------------------------------------------------------------------

// INFO, PJ: ProcessCharacter method stores scanner state.
//           This convenience method checks this state.
function TXMLScanner.IsInEndState: boolean;
begin
  result := (TMethod(fProcessCharacter).code = TMethod(fStartMethod).code);
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TXMLParser.Create( const aScanner: TXMLScanner );
begin
  inherited Create;

  fNodes := TObjectStack.Create;
  fScanner := aScanner;
end;

// ----------------------------------------------------------------------------

destructor TXMLParser.Destroy;
begin
  FreeAndNil( fNodes );

  inherited Destroy;
end;

// ----------------------------------------------------------------------------

function TXMLParser.RootNode: TXMLNode;
var
  node: TXMLNode;

begin
  Resolve;

  if fNodes.Count > 0 then
  begin
      node := fNodes.Pop as TXMLNode;

      raise Exception.create( 'Closing tag missing for opening tag ' + node.Name );
  end;

  result := fRootNode;
end;

// ----------------------------------------------------------------------------

procedure TXMLParser.Resolve;
var
  token: TToken;

begin
  token := fScanner.NextToken;

  while token.ID <> NO_TOKEN do
  begin
      HandleToken( token );

      Token := fScanner.NextToken;
  end;
end;

// ----------------------------------------------------------------------------

procedure TXMLParser.HandleToken( const aToken: TToken );
var
  name: string;
  node: TXMLNode;

begin
  // PERF, PJ: Update once every 32 tokens (updating is relatively expensive)
  fUpdateCount := fUpdateCount + 1;
  if fUpdateCount = 32 then
  begin
      UpdateProgress( aToken.Position, fScanner.TextSize );

      fUpdateCount := 0;
  end;

  case TXMLToken( aToken.ID ) of

      xmltTag:
      begin
          name := LowerCase( aToken.text );
          node := TXMLNode.Create( name, aToken.position );

          AddTreeNode( node );

          fNodes.Push( node );
      end;

      xmltTagEnd:
      begin
          name := LowerCase( aToken.text );
          node := fNodes.Pop as TXMLNode;

          if ( name <> '' ) and ( node.Name <> name ) then
          begin
              raise Exception.create( 'Opening tag "' + node.name + '" doesn''t mach closing tag "' + name + '" (' + IntToStr( aToken.Line ) + ',' + IntToStr( aToken.Column ) + ')' );
          end;
      end;

      xmltTagSingle:
      begin
          fNodes.Pop;
      end;

      xmltPropertyName:
      begin
          fPropertyName := LowerCase( aToken.text );
      end;

      xmltPropertyValue:
      begin
          node := fNodes.Peek as TXMLNode;
          node.Properties[fPropertyName] := XMLDecode( aToken.text );
      end;

      xmltText:
      begin
          if fNodes.Count = 0 then
              raise Exception.Create( Format( 'Unexpected text encountered outside root node (%d, %d).', [aToken.Line, aToken.Column] ));
            
          node := fNodes.Peek as TXMLNode;
          node.Text := node.Text + XMLDecode( aToken.text );
      end;
  else
      raise Exception.create( Format( 'Unexpected token (%d, %d).', [aToken.Line, aToken.Column] ));
  end;
end;

// ----------------------------------------------------------------------------

procedure TXMLParser.UpdateProgress( const aPosition, aTotalSize: integer );
begin
  if Assigned( fOnProgress ) then
    fOnProgress( aPosition, aTotalSize );
end;

// ----------------------------------------------------------------------------

procedure TXMLParser.AddTreeNode( const aNode: TXMLNode );
var
  node: TXMLNode;

begin
	if fNodes.Count > 0 then
	begin
		node := fNodes.Peek as TXMLNode;
      node.AddChildNode( aNode );
	end
	else if fRootNode = nil then
  begin
		fRootNode := aNode;
  end
  else
      raise Exception.create( 'Only 1 root node supported.' );
end;

// ----------------------------------------------------------------------------

function TXMLParser.XMLDecode( const aText: string ): string;
var
  i: integer;
  character: char;
  specialSequence: string;

begin
  result := '';
  specialSequence := '';

  for i := 1 to Length( aText ) do
  begin
      character := aText[i];

      if character = '&' then
      begin
          result := result + specialSequence;
          specialSequence := '&';
      end
      else if specialSequence <> '' then
      begin
          specialSequence := specialSequence + character;

          if ( character = ';' ) then
          begin
              result := result + ConvertSpecialSequence( specialSequence );

              specialSequence := '';
          end;
      end
      else
          result := result + character;
  end;

  result := result + specialSequence;
end;

// ----------------------------------------------------------------------------

function TXMLParser.ConvertSpecialSequence( const aSequence: string ): string;
var
  isValid: boolean;
  i: integer;
  character: char;

begin
  if aSequence = '&amp;' then
  begin
      result := '&';
  end
  else if aSequence = '&lt;' then
  begin
      result := '<';
  end
  else if aSequence = '&gt;' then
  begin
      result := '>';
  end
  else if aSequence = '&quot;' then
  begin
      result := '"';
  end
  else if aSequence = '&apos;' then
  begin
      result := '''';
  end
  else if aSequence[2] = '#' then
  begin
      isValid := TRUE;

      for i := 3 to Length( aSequence ) do
      begin
          character := aSequence[i];

          if not ( character in DECIMAL_DIGITS_CHARACTER_SET ) then
          begin
              isValid := FALSE;

              break;
          end
      end;

      if isValid then
      begin
          result := Chr( StrToInt( Copy( aSequence, 3, Length( aSequence ) - 3 + 1  )) mod 256 );
      end
      else
          result := aSequence;
  end
  else
      result := aSequence;
end;


end.
