{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_XMLSceneReader;

{$MODE Delphi}

interface

uses
    unit_ViewPortManager,
    Classes,
    unit_Materials,
    unit_MaterialList,
    unit_Scene,
    unit_XMLSceneConstants,
    unit_XMLReader,
    unit_XMLDataReader;

type
    // the XMLSceneReader processes XML scene data as read from a file;
    // it adds data to a scene but does not replace items
    TXMLSceneReader = class(TXMLDataReader)
    private
        function MakeUniqueMaterial( const aMaterial: TMaterial; const aMaterials: TMaterialList ): TMaterial;
        function MakeUniqueName( const aName: string; const aMaterials: TMaterialList ): string;
    protected
        procedure ProcessPrimitives(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene); override;
        procedure ProcessMaterial(const aNode: TXMLNode; const aCategoryName: string; const aMaterialList: TStringList; const aScene: TScene); override;
    public
        procedure ReadFromFile(const aFileName: string; const aScene: TScene; Aviewportmanager: tviewportmanager = nil);
    end;

var xmlReadStartTime: Cardinal;
    xmlReadEndTime: Cardinal;
    xmlParseStartTime: Cardinal;
    xmlParseEndTime: Cardinal;

implementation

uses

    SysUtils, Windows, IniFiles,           // INFO, PJ: Added for THashedStringList.. Gotta love the VCL unit consistency ;-)
    unit_Primitives;

// ### TXMLSceneReader #########################################################

procedure TXMLSceneReader.ReadFromFile(const aFileName: string; const aScene: TScene; Aviewportmanager: tviewportmanager = nil);
var reader: TXMLFileReader;
    node: TXMLNode;
begin
    xmlReadStartTime := GetTickCount;
    node := nil;

    if ShouldShowProgress then begin
        progressForm.Title := 'Loading...';
        progressForm.SubTitle := 'Loading ' + ExtractFileName(aFileName);
        ProgressForm.Description := 'Loading XML';
        ProgressForm.Show;
    end;

    reader := TXMLFileReader.Create(aFileName);
    try
        reader.OnProgress := HandleProgress;

        node := reader.RootNode;

        xmlReadEndTime := GetTickCount;
        xmlParseStartTime := xmlReadEndTime;

        ReadScene(node, aScene, Aviewportmanager);

        xmlParseEndTime := GetTickCount;
    finally
        FreeAndNil(reader);
        FreeAndNil(node);
    end;
end;

//This makes sure that there are no duplicate ID's
procedure TXMLSceneReader.ProcessPrimitives(const aNode: TXMLNode; const aMaterialList: TStringList; const aScene: TScene);
var i: integer;
    node: TXMLNode;
    primitive: TBasePrimitive;
begin
    if ShouldShowProgress then
        ProgressForm.Description := 'Loading Primitives.';

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'primitive' then
        begin
            primitive := aScene.Primitives.GetPrimitiveByID(node.IntegerProperties['id']);
            // a primitive with the same ID already exists, so we assign a unique
            // ID to the new primitive before appending it to the scene
            if Assigned(primitive) then
                node.Properties['id'] := IntToStr(aScene.Primitives.HighestID+1);
            AppendPrimitive(node, aMaterialList, aScene);
        end;
    end;
end;

// directly adds the material to the scene; updates the ID value to the highest
// ID value + 1 if a material with the same ID already exists
// Besides ID merging (always needed), also merges on properties/name (only when merging scenes).
// Keep in synch with TXMLPluginReader.AppendMaterial! (TODO, PJ: Small rewrite needed, 'keep in synch' = evil :P)
procedure TXMLSceneReader.ProcessMaterial(const aNode: TXMLNode; const aCategoryName: string; const aMaterialList: TStringList; const aScene: TScene);
var i: integer;
    node: TXMLNode;
    material, existingMaterial, uniqueMaterial: TMaterial;
    localID: string;
begin
    UpdateProgressForNode( aNode );

    material := TMaterial.Create;

    localID := aNode.Properties['id'];

    // check if a material with the same ID already exist; if so, reset the ID value
    // as found in the XML to the highest ID in the scene + 1; provide that new ID
    // to the material in-the-making
    existingMaterial := aScene.MaterialManager.GetMaterialByID(aNode.IntegerProperties['id']);
    if Assigned(existingMaterial) then
        aNode.Properties['id'] := IntToStr(aScene.MaterialManager.HighestID+1);

    material.ID := aNode.IntegerProperties['id'];
    material.Category := aCategoryName;
    material.Name := aNode.Properties['name'];
    material.IsUsed := aNode.BooleanProperties['used'];

    if aNode.HasProperty('lightmap') then
        material.EnableLightmap := aNode.BooleanProperties['lightmap'];

    if aNode.HasProperty('castShadows') then
        material.CastShadows := aNode.BooleanProperties['castShadows'];

    if aNode.HasProperty('receiveShadows') then
        material.ReceiveShadows := aNode.BooleanProperties['receiveShadows'];

    for i := 0 to aNode.NumChildNodes - 1 do
    begin
        node := aNode.ChildNodes[i];

        if node.Name = 'comments' then begin
            material.Comments := node.Text;
        end
        else if node.Name = 'tag' then begin
            material.Tag := node.Text;
        end
        else if node.name = 'layer' then begin
            AppendMaterialLayer(node, material, aScene);
        end
        else if node.name = 'raytracing' then begin
            AppendMaterialRaytracingProperties(node, material);
        end;
    end;

    uniqueMaterial := MakeUniqueMaterial( material, aScene.MaterialManager );

    if uniqueMaterial = material then
        aScene.MaterialManager.Add( material )
    else
        material.Free;

     // Register the material under its ID value coming from the XML; this might very
     // well be a totally different ID than the material will have in the scene itself
    aMaterialList.AddObject(localID, uniqueMaterial);
end;

// Merges the material into the scene in a way that is specific to the XMLSceneReader (merging)
// For each material A, find an exising material B with the same name.
//    - If doesn't exist, use A
//    - If does exist, check wheter A equals B
//      - If so, use B (destroy A).
//      - If not, generate a new (and unique) name for A and use A.
// When the existing set of materials is empty (i.e. loading a new scene), this behaves as expected.
// REFA, PJ: Slightly different behaviour from TMaterialList.AddMaterial .
//           Perhaps TMaterialList.addMaterial should do exactly this?
function TXMLSceneReader.MakeUniqueMaterial( const aMaterial: TMaterial; const aMaterials: TMaterialList ): TMaterial;
var material: TMaterial;
begin         
    result := aMaterial;

    material := aMaterials.GetMaterialByName( aMaterial.Name );
    if material <> nil then begin
        if aMaterial.Equals( material ) then begin
            result := material;
        end
        else
            aMaterial.Name := MakeUniqueName( aMaterial.Name, aMaterials );
    end;
end;

// NOTE, PJ: This is not the same functionality as MaterialList.MakeUniqueName
// REFA, PJ: But perhaps MaterialList.MakeUniqueName should be more like this?
function TXMLSceneReader.MakeUniqueName( const aName: string; const aMaterials: TMaterialList ): string;
var index: integer;
begin
    index := 2;
    result := aName;

    if not aMaterials.NameIsUnique( result ) then begin
        repeat
            result := Format( '%s (%d)', [aName, index] );
            index := index + 1;
        until aMaterials.NameIsUnique( result );
    end;
end;

end.

