{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_XMLWriter;

{$MODE Delphi}

(* ==========================================================================
   Basic XML Writer. Doesn't do any validation.
   ========================================================================== *)

interface

uses
  Classes,
  unit_BufferedStringStream;

type
  TXMLWriter = class( TObject )
  private
      fStream: TStream;
      fText: string;
      fPrefix: string;
      fIsTagOpen: Boolean;
      fLastOpentag: string;

      function XMLEncode( const aString: string ): string;
      procedure AppendString( const aString: string );

  public
      constructor Create( const aStream: TStream );

      procedure AppendOpenTag( const aName: string );
      procedure AppendCloseTag( const aName: string );

      procedure AppendProcessingInstruction( const anInstruction: string );
      procedure AppendText( const aText: string );
      procedure AppendComment( const aText: string );

      procedure AppendProperty( const aName, aValue: string ); overload;
      procedure AppendProperty( const aName: string; const aValue: Boolean ); overload;
      procedure AppendProperty( const aName: string; const aValue: Single ); overload;
      procedure AppendProperty( const aName: string; const aValue: Integer ); overload;

      procedure AppendConditionalProperty( const aName: string; const aValue: Single; const aCondition: Boolean ); 
  end;

// ----------------------------------------------------------------------------

  TXMLFileWriter = class( TXMLWriter )
  private
      fStream: TStream;
  public
      constructor Create( const aFileName: string );
      destructor Destroy; override;
  end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

  TXMLStringWriter = class(TXMLWriter)
  private
      fStream: TBufferedStringStream; // INFO, PJ: Important! Using TStringStream results in _wrong_ datastring being called. Design flaw.
      function GetXMLString: string;
  public
      constructor Create;
      destructor Destroy; override;
      property XMLString: string read GetXMLString;
  end;


implementation

uses
  SysUtils,
  unit_BufferedFileStream;

const
  CRLF = #13#10;
  XML_BOOLEAN_VALUE: array[Boolean] of string = ('false',  'true');

constructor TXMLWriter.Create( const aStream: TStream );
begin
  inherited Create;

  fStream := aStream;
end;

// ----------------------------------------------------------------------------

// XML Encode, optimized for speed, at the cost of a little bit of readability...
function TXMLWriter.XMLEncode( const aString: string ): string;
type
  TCharSet = Set of Char;

const
  INVALID_CHARACTERS: TCharSet = ['''', '"', '&', '<', '>'];
  CHARACTER_CONVERSION_TABLE: Array[Char] of String = (
      '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '',
      '', '', '', '', '', '', '', '', '', '', '', '', '', '', '', '',
      ' ','!','&quot;','#','$','%','&amp;','&apos;','(',')','*','+',',','-','.','/',
      '0','1','2','3','4','5','6','7','8','9',':',';','&lt;','=','&gt;','?',
      '@','A','B','C','D','E','F','G','H','I','J','K','L','M','N','O',
      'P','Q','R','S','T','U','V','W','X','Y','Z','[','\',']','^','_',
      '`','a','b','c','d','e','f','g','h','i','j','k','l','m','n','o',
      'p','q','r','s','t','u','v','w','x','y','z','{','|','}','~','',
      '€','','‚','ƒ','„','…','†','‡','ˆ','‰','Š','‹','Œ','','Ž','',
      '','‘','’','“','”','•','–','—','˜','™','š','›','œ','','ž','Ÿ',
      ' ','¡','¢','£','¤','¥','¦','§','¨','©','ª','«','¬','­','®','¯',
      '°','±','²','³','´','µ','¶','·','¸','¹','º','»','¼','½','¾','¿',
      'À','Á','Â','Ã','Ä','Å','Æ','Ç','È','É','Ê','Ë','Ì','Í','Î','Ï',
      'Ð','Ñ','Ò','Ó','Ô','Õ','Ö','×','Ø','Ù','Ú','Û','Ü','Ý','Þ','ß',
      'à','á','â','ã','ä','å','æ','ç','è','é','ê','ë','ì','í','î','ï',
      'ð','ñ','ò','ó','ô','õ','ö','÷','ø','ù','ú','û','ü','ý','þ','ÿ'
  );

var
  i, j, size, index: integer;
  character: char;
  value: string;

begin
  // First pass: Determine length
  size := 0;
  for i := 1 to Length( aString ) do
  begin
      character := aString[i];
      if character in INVALID_CHARACTERS then
      begin
          size := size + Length( CHARACTER_CONVERSION_TABLE[character] );
      end
      else
          size := size + 1;
  end;

  SetLength( result, size );

  // Second pass: fill in character data
  index := 1;
  for i := 1 to Length( aString ) do
  begin
      character := aString[i];

      if character in INVALID_CHARACTERS then
      begin
          value := CHARACTER_CONVERSION_TABLE[character];

          for j := 1 to Length( value ) do
          begin
              result[index] := value[j];
              index := index + 1;
          end;
      end
      else
      begin
          result[index] := character;
          index := index + 1;
      end;
  end;
end;
//}
// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendString( const aString: string );
begin
  if Length( aString ) > 0 then
  begin
      fStream.Write( aString[1], Length( aString ));
  end;
end;

// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendOpenTag( const aName: string );
begin
  // Close any open tags
  if fIsTagOpen then
  begin
      AppendString( '>' );
  end;

  // Append any pending text
  if fText <> '' then
  begin
      AppendString( fText );
      fText := '';
  end;

  if fPrefix <> '' then
  begin
      AppendString( CRLF );
      AppendString( fPrefix );
  end;

  AppendString( '<' + aName );

  fIsTagOpen := TRUE;
  fLastOpenTag := aName;
  fPrefix := fPrefix + '  ';
end;

// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendCloseTag( const aName: string );
begin
  Delete( fPrefix, 1, 2 );

  if  fIsTagOpen and
      ( aName = fLastOpenTag ) and
      ( fText = '' ) then
  begin // <tag />
      AppendString( ' />' );
  end
  else
  begin // </tag>
      if fIsTagOpen then
      begin
          AppendString( '>' );
      end;

      if fText <> '' then
      begin
          AppendString( fText );
          fText := '';
      end
      else
      begin
          AppendString( CRLF );
          AppendString( fPrefix );
      end;

      AppendString( '</' + aName + '>' );
  end;

  fIsTagOpen := FALSE;
  fLastOpenTag := '';
end;

// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendProcessingInstruction( const anInstruction: string );
begin
  fText := fText + '<?' + anInstruction + ' ?>';
end;

// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendText( const aText: string );
begin
  fText := fText + XMLEncode( aText );
end;

// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendComment( const aText: string );
begin
  ftext := fText + CRLF + CRLF + fPrefix + '<!-- ' + aText + ' -->' + CRLF + fPrefix;
end;

// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendProperty( const aName, aValue: string );
begin
  if fIsTagOpen then
  begin
      AppendString( Format( ' %s="%s"', [aName, XMLEncode( aValue ) ] ));
  end
  else
      raise Exception.Create( Format( 'Couldn''t add property "%s". No tag opened.', [aName] ));
end;

// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendProperty( const aName: string; const aValue: Integer ); 
begin
    AppendProperty( aName, IntToStr( aValue ));
end;

// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendProperty( const aName: string; const aValue: Boolean );
begin
    AppendProperty( aName, XML_BOOLEAN_VALUE[aValue] );
end;

// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendProperty( const aName: string; const aValue: Single );
begin
    // INFO, PJ: Not a very generic formatting rule. Perhaps make the
    //           formatting the application's responsibility, or at least make
    //           it configurable.
    AppendProperty( aName, FloatToStrF( aValue, ffGeneral, 7, 7 ));
end;

// ----------------------------------------------------------------------------

procedure TXMLWriter.AppendConditionalProperty( const aName: string; const aValue: Single; const aCondition: Boolean );
begin
    if aCondition then AppendProperty( aName, aValue );
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TXMLFileWriter.Create( const aFileName: string );
begin
    fStream := TBufferedFileStream.Create( aFileName, fmCreate or fmShareDenyWrite );
    inherited Create( fStream );
end;

// ----------------------------------------------------------------------------

destructor TXMLFileWriter.Destroy;
begin
    FreeAndNil( fStream );
    inherited Destroy;
end;

// ----------------------------------------------------------------------------
// ----------------------------------------------------------------------------

constructor TXMLStringWriter.Create;
begin
    fStream := TBufferedStringStream.Create;
    inherited Create(fStream);
end;

destructor TXMLStringWriter.Destroy;
begin
    FreeAndNil(fStream);
    inherited Destroy;
end;

function TXMLStringWriter.GetXMLString: string;
begin
    Result := fStream.DataString;
end;


end.
