{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_AttenuationGraph;

{$MODE Delphi}

interface

uses
    Types,
    Graphics,
    Classes,
    Controls;

type
    TAttenuationGraph = class(TCustomControl)
    private
        fAttenuationConst: double;
        fAttenuationLinear: double;
        fAttenuationQuadratic: double;
        fAttenuationRange: double;

        fGridColorMajor: TColor;
        fGridColorMinor: TColor;
        fRangeColor: TColor;

        fGraphRange: Integer;

        procedure InitializeMenu;
        procedure SetAttenuationConst(const aValue: Double);
        procedure SetAttenuationLinear(const aValue: Double);
        procedure SetAttenuationQuadratic(const aValue: Double);
        procedure SetAttenuationRange(const aValue: Double);

        procedure HandleMenuClick(aSender: TObject);

    protected
        procedure Paint; override;

    public
        property AttenuationConst: double read fAttenuationConst write SetAttenuationConst;
        property AttenuationLinear: double read fAttenuationLinear write SetAttenuationLinear;
        property AttenuationQuadratic: double read fAttenuationQuadratic write SetAttenuationQuadratic;
        property AttenuationRange: double read fAttenuationRange write SetAttenuationRange;

        property GridColorMajor: TColor read fGridColorMajor write fGridColorMajor;
        property GridColorMinor: TColor read fGridColorMinor write fGridColorMinor;
        property RangeColor: TColor read fRangeColor write fRangeColor;

        property Color;
        property Font;

        procedure SetAttenuation(const aConst, aLinear, aQuadratic, aRange: double);

        constructor Create(aOwner: TComponent); override;
    end;

implementation

uses
    Menus,
    SysUtils;

const
    STEP_SIZE = 2;  // Might want to adjust this when making the graph larger (performance!)
    GRID_SIZE = 8;

procedure TAttenuationGraph.InitializeMenu;
var i, range: integer;
    menu: TPopupMenu;
    item: TMenuItem;
    rangeItem: TMenuItem;

begin
    menu := TPopupMenu.Create(Owner);
    item := TMenuItem.Create(menu);
    menu.Items.Add(item);
    item.Caption := 'Set Graph Range';

    rangeItem := TMenuItem.Create(menu);
    item.Add(rangeItem);
    rangeItem.Caption := 'Attenuation Range';
    rangeItem.Tag := -1;
    rangeItem.OnClick := HandleMenuClick;

    rangeItem := TMenuItem.Create(menu);
    item.Add(rangeItem);
    rangeItem.Caption := '-';

    range := 64;
    for i := 0 to 7 do begin
      rangeItem := TMenuItem.Create(menu);
      item.Add(rangeItem);
      rangeItem.Caption := IntToStr(range);
      rangeItem.Tag := range;
    rangeItem.OnClick := HandleMenuClick;

      range := range * 2;
    end;

    popupMenu := menu;
end;

procedure TAttenuationGraph.SetAttenuationConst(const aValue: Double);
begin
    fAttenuationConst := aValue;
    Paint;
end;

procedure TAttenuationGraph.SetAttenuationLinear(const aValue: Double);
begin
    fAttenuationLinear := aValue;
    Paint;
end;

procedure TAttenuationGraph.SetAttenuationQuadratic(const aValue: Double);
begin
    fAttenuationQuadratic := aValue;
    Paint;
end;

procedure TAttenuationGraph.SetAttenuationRange(const aValue: Double);
begin
    fAttenuationRange := aValue;
    Paint;
end;

procedure TAttenuationGraph.Paint;
var i, x, y, graphWidth,graphHeight,graphLeft,graphRight,graphBottom,graphTop: integer;
    text: string;
    distance, attenuation: double;
begin
    Canvas.Brush.Style := bsSolid;
    Canvas.Brush.Color := color;
    Canvas.FillRect( Rect( 0, 0, Width, Height ));

    graphLeft := 8;
    graphTop := 1;
    graphWidth := Width - 10;
    graphHeight := Height - 13;

    graphBottom := graphTop + graphHeight;
    graphRight := graphLeft + graphWidth;

    // draw fine grid
    Canvas.Pen.Style := psSolid;
    Canvas.Pen.Width := 1;
    Canvas.Pen.Color := fGridColorMinor;

    x := GraphLeft;
    for i := 0 to GraphWidth div GRID_SIZE do begin
        Canvas.PenPos := Point(x,GraphTop);
        Canvas.LineTo(x, GraphBottom);
        x := x + GRID_SIZE;
    end;

    y := GraphBottom;
    for i := 0 to GraphHeight div GRID_SIZE do begin
        Canvas.PenPos := Point(GraphLeft,y);
        Canvas.LineTo( GraphRight, y);
        y := y - GRID_SIZE;
    end;

    // draw thick major lines at grid borders
    Canvas.Pen.Color := fGridColorMajor;

    Canvas.PenPos := Point(GraphLeft,GraphBottom);
    Canvas.LineTo(GraphLeft, GraphTop);

    Canvas.PenPos := Point(GraphLeft,GraphBottom);
    Canvas.LineTo(GraphRight, GraphBottom);

    // draw ticks
    x := GraphLeft + GraphWidth div 2;
    Canvas.PenPos := Point( x, GraphBottom - 2);
    Canvas.LineTo(x, GraphBottom + 3);
    Canvas.PenPos := Point(GraphLeft - 2, GraphTop);
    Canvas.LineTo(GraphLeft + 3, GraphTop);
    Canvas.PenPos := Point(GraphRight, GraphBottom - 2);
    Canvas.LineTo(GraphRight, GraphBottom + 3);

    Canvas.Font := Font;
    Canvas.Pen.Color := Font.Color;
    Canvas.Brush.Style := bsClear;
    Canvas.TextOut(GraphLeft - 6, GraphBottom - 1, '0');
    Canvas.TextOut(GraphLeft - 7, GraphTop - 2, '1');
    text := IntToStr( fGraphRange div 2 );
    Canvas.TextOut(GraphLeft + GraphWidth div 2 - Canvas.TextWidth( Text ) div 2, GraphBottom + 1, text );
    text := IntToStr( fGraphRange );
    Canvas.TextOut(GraphLeft + GraphWidth - Canvas.TextWidth( Text ) + 1 , GraphBottom + 1, text );

    // draw graph
    Canvas.Pen.Color := clBLack;
    Canvas.Pen.Width := 2;
    x := GraphLeft;
    for i := 0 to GraphWidth div STEP_SIZE do begin
        distance := STEP_SIZE * i * fGraphRange / GraphWidth;
        attenuation := 1 / (AttenuationConst + AttenuationLinear * distance + AttenuationQuadratic * Sqr(distance));

        y := graphBottom - Round(attenuation * GraphHeight);

        if i = 0 then
            Canvas.PenPos := Point(x, y)
        else
            Canvas.LineTo(x, y);

        x := x + STEP_SIZE;
    end;

    // draw range
    Canvas.Pen.Style := psDot;
    Canvas.Pen.Width := 1;
    Canvas.Pen.Color := RangeColor;
    x := GraphLeft + Round(GraphWidth * AttenuationRange / fGraphRange);
    Canvas.PenPos := Point(x, GraphTop);
    Canvas.LineTo(x, GraphBottom);
end;

procedure TAttenuationGraph.SetAttenuation(const aConst, aLinear, aQuadratic, aRange: double);
begin
    fAttenuationConst := aConst;
    fAttenuationLinear := aLinear;
    fAttenuationQuadratic := aQuadratic;
    fAttenuationRange := aRange;

    Paint;
end;

procedure TAttenuationGraph.HandleMenuClick(aSender: TObject);
var range: integer;
begin
    range := TMenuItem(aSender).Tag;
    if range = -1 then
        fGraphRange := Round(fAttenuationRange)
    else
        fGraphRange := range;
        
    Paint;
end;

constructor TAttenuationGraph.Create(aOwner: TComponent);
begin
    inherited Create(aOwner);

    fAttenuationConst := 1;
    fAttenuationLinear := 0.0005;
    fAttenuationQuadratic := 0.000003;
    fAttenuationRange := 1000;
    fGraphRange := 1024;

    fGridColorMajor := clBlack;
    fGridColorMinor := clGray;
    fRangeColor := clRed;

    Font.Height := -10;
    Font.Name := 'Arial';

    Color := clBtnface;

    Enabled := true;
    TabStop := true;

    InitializeMenu;
end;


end.

