{ ----------------------------------------------------------------------------
    Copyright (C) 2005-2010 Jeroen Commandeur and Paul-Jan Pauptit

    This file is part of DeleD CE (Community Edition)

    DeleD CE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Deled CE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Deled CE. If not, see <http://www.gnu.org/licenses/>.
  ---------------------------------------------------------------------------- }
unit unit_ScreenshotManager;

{$MODE Delphi}

interface

uses
    Graphics,
    Controls,
    frmOGLForm,
    frmScreenshotForm,
    unit_TextureTypes,
    unit_Types,
    unit_Options;

type
    TScreenshotManager = class(TObject)
    private
        fForm: TOGLForm;
        fScreenshotForm: TScreenshotForm;
        fOptions: TScreenshotOptions;
        fParentControl: TControl;

        function GenerateScreenshotData: TTextureData;
        function GenerateTileScreenshot(const anX, aY: integer; const aTotalWidth, aTotalHeight: integer): TTextureData;

        procedure ApplyWatermarkToScreenshot(const aScreenshot: TTextureData);
        function DownSampledScreenshot(const aScreenshot: TTextureData): TTextureData;
        procedure SaveScreenshot(const aScreenshot: TTextureData);
        function GetTileSize: integer;

        property TileSize: integer read GetTileSize;

    protected
        function GetScreenshotForm: TScreenshotForm;

        property ScreenshotForm: TScreenshotForm read GetScreenshotForm;

    public
        constructor Create(const aForm: TOGLForm; const anOptions: TScreenshotOptions; const aParentControl: TControl = nil);
        destructor Destroy; override;

        function GenerateScreenshot: TTextureData;
        procedure TakeScreenshot(const aShouldShowOptions: boolean);
        class function MakeWatermarkTexture(const aWatermarkColor: TColor; const anOpacity: integer): TTextureData;
    end;

implementation

uses
    Math,
    Types,
    ExtCtrls,
    SysUtils,
    dglOpenGL,
    Dialogs,
    Forms,
    Windows,
    unit_TextureLoader,
    unit_TextureWriter,
    unit_TextureManipulator,
    unit_VersionInfo,
    unit_Constants,
    frmScreenshotOptionsForm;

constructor TScreenshotManager.Create(const aForm: TOGLForm; const anOptions: TScreenshotOptions; const aParentControl: TControl);
begin
    inherited Create;

    fForm := aForm;
    fOptions := anOptions;
    fParentControl := aParentControl;
end;

destructor TScreenshotManager.Destroy;
begin
    FreeAndNil(fScreenshotForm);

    inherited Destroy;
end;

function TScreenshotManager.GetScreenshotForm: TScreenshotForm;
begin
    if not Assigned(fScreenshotForm) then begin
        fScreenshotForm := TScreenshotForm.Create(fForm, fOptions, fParentControl);
        fScreenshotForm.Show;

        SetWindowPos(fScreenshotForm.Handle, HWND_TOPMOST, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);

        Application.ProcessMessages;                  
    end;

    result := fScreenshotForm;
end;

function TScreenshotManager.GenerateScreenshot: TTextureData;
begin
    result := GenerateScreenshotData;
    result := DownSampledScreenshot(result);
end;

procedure TScreenshotManager.TakeScreenshot(const aShouldShowOptions: boolean);
var optionsForm: TFrmScreenshotOptions;
    shouldTakeScreenshot: boolean;
    screenshot: TTextureData;
    cursor: TCursor;
begin
    if aShouldShowOptions then begin
        optionsForm := TFrmScreenshotOptions.Create(nil);
        try
            shouldTakeScreenshot := (optionsForm.ShowModal = mrOK);
        finally
            FreeAndNil(optionsForm);
        end;
    end
    else
        shouldTakeScreenshot := TRUE;

    if shouldTakeScreenshot then begin
        cursor := Screen.Cursor;
        screen.Cursor := crHourGlass;
        try
            screenshot := GenerateScreenshotData;

            if fOptions.UseSuperSampling then
                screenshot := DownSampledScreenshot(screenshot);

            ApplyWatermarkToScreenshot(screenshot);

        finally
            Screen.Cursor := cursor;
        end;

        SaveScreenshot(screenshot);
    end;
end;

class function TScreenshotManager.MakeWatermarkTexture(const aWatermarkColor: TColor; const anOpacity: integer): TTextureData;
var textureManipulator: TTextureManipulator;
begin
    try
        textureManipulator := TTextureManipulator.CreateWithText(' DeleD ' + _versionstring + ' screenshot ', WATERMARK_FONT_NAME, WATERMARK_FONT_SIZE, aWatermarkColor, true);
        textureManipulator.ScaleAlpha(anOpacity * 255 div 100);

        result := textureManipulator.Texture;
    finally
        FreeAndNil(textureManipulator);
    end;
end;

function TScreenshotManager.GenerateScreenshotData: TTextureData;
var width, height: integer;
    nrHorizontalTiles, nrVerticalTiles: integer;
    x, y: integer;
    textureManipulator: TTextureManipulator;
    screenshot: TTextureData;

begin
    with fOptions do begin
        width := screenshotWidth;
        height := screenshotHeight;

        if width = -1 then width := fForm.ClientWidth;
        if height = -1 then height := fForm.clientHeight;
 
        if UseSuperSampling then begin
            width := width * 3;
            height := height * 3;
        end;

        nrHorizontalTiles := ( width + TileSize - 1 ) div TileSize;
        nrVerticalTiles := ( height + TileSize - 1 ) div TileSize;

        textureManipulator := TTextureManipulator.CreateWithProperties(width, height, 24);
        try
            for y := 0 to nrVerticalTiles - 1 do begin
                for x := 0 to nrHorizontalTiles - 1 do begin
                    // ... generate screenshot data ...
                    screenshot := GenerateTileScreenshot(x * TileSize, y * TileSize, width, height);

                    textureManipulator.Draw(x * TileSize, y * TileSize, screenshot);
                end;
            end;

            result := textureManipulator.texture;
        finally
            FreeAndNil(textureManipulator);
        end;
    end;
end;

function TScreenshotManager.GenerateTileScreenshot(const anX, aY: integer; const aTotalWidth, aTotalHeight: integer): TTextureData;
var bytesPerPixel: integer;
    lineWidth: integer;
    buffer: array of byte;
    y: integer;
begin
    result.Width := Min(TileSize, aTotalWidth - anX );
    result.Height := Min(TileSize, aTotalHeight - aY );
    result.BitsPerPixel := 24;
    SetLength(result.Pixels, result.Width * result.Height * 24 );

    screenshotForm.UpdateTiledProjection( anX, aY, aTotalWidth, aTotalHeight );
    screenshotForm.glDraw;

    bytesPerPixel := result.BitsPerPixel div 8;
    lineWidth := result.Width * bytesPerPixel;

    SetLength(buffer, Length(result.Pixels));

    glPixelStorei(GL_PACK_ALIGNMENT,1);
    glReadBuffer(GL_BACK);
    // INFO, PJ: Reading pixels from the backbuffer after a swap is not completely safe
    //
    glReadPixels(0, TileSize-result.Height, result.Width, result.Height, GL_BGR_EXT, GL_UNSIGNED_BYTE, @buffer[0]);

    for y := 0 to result.Height - 1 do begin
        Move(buffer[y * lineWidth], result.Pixels[(result.Height - y - 1) * lineWidth], lineWidth);
    end;
end;

procedure TScreenshotManager.ApplyWatermarkToScreenshot(const aScreenshot: TTextureData);
var textureManipulator: TTextureManipulator;
    opacity: integer;
begin
    opacity := fOptions.screenshotWatermarkOpacity;
    textureManipulator := TTextureManipulator.Create(
        MakeWatermarkTexture( fOptions.screenshotWatermarkColor, opacity ));
    textureManipulator.DrawOnTexture(10, 10, aScreenshot);

    FreeAndNil(textureManipulator);
end;

function TScreenshotManager.DownSampledScreenshot(const aScreenshot: TTextureData): TTextureData;
// Reduces any 3x3 block of pixels to a single pixel
var textureManipulator: TTextureManipulator;
begin
    textureManipulator := TTextureManipulator.Create(aScreenshot);
    try
        textureManipulator.Downsample(3);
        result := textureManipulator.Texture;
    finally
        FreeAndNil(textureManipulator);
    end;
end;

procedure TScreenshotManager.SaveScreenshot(const aScreenshot: TTextureData);
var saveDialog: TSaveDialog;
    textureWriter: TTextureWriter;
    filter: string;
begin
    saveDialog := TSaveDialog.Create(nil);
    try
        saveDialog.InitialDir := ExtractFilePath(Application.ExeName);
        saveDialog.FileName := '';
        filter := 'JPEG (*.jpg)|*.jpg|TARGA (*.tga)|*.tga|Bitmap (*.bmp)|*.bmp';

        saveDialog.Filter := filter;
        saveDialog.DefaultExt := 'jpg';                                         // default
        if saveDialog.Execute then begin
            textureWriter := TTextureWriter.Create;
            try
                textureWriter.CompressionQuality := fOptions.screenshotCompressionQuality;
                textureWriter.SaveTextureData(aScreenshot, saveDialog.Filename);
            finally
                FreeAndNil(textureWriter);
            end;
        end;
    finally
        FreeAndNil(saveDialog);
    end;
end;

function TScreenshotManager.GetTileSize: integer;
begin
    result := ScreenshotForm.ClientWidth;
end;

end.
